
#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 2203
 * transforms:
 *    ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul', 'a@8', 128) => ('ishl', 'a', 7)
 *    ('imul', 'a@16', 32768) => ('ishl', 'a', 15)
 *    ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31)
 *    ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63)
 *    ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b'))
 *    ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b')))
 *    ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b'))
 *    ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b'))
 *    ('udiv', 'a', 1) => a
 *    ('idiv', 'a', 1) => a
 *    ('umod', 'a', 1) => 0
 *    ('imod', 'a', 1) => 0
 *    ('imod', 'a', -1) => 0
 *    ('irem', 'a', 1) => 0
 *    ('irem', 'a', -1) => 0
 *    ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b'))
 *    ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b')))
 *    ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b')))))
 *    ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b'))
 *    ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b')))
 *    ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b'))
 *    ('~fmul', ('fsign', 'a'), ('ffloor', ('fadd', ('fabs', 'a'), 0.5))) => ('ftrunc', ('fadd', 'a', ('fmul', ('fsign', 'a'), 0.5)))
 *    ('~fneg', ('fneg', 'a')) => a
 *    ('ineg', ('ineg', 'a')) => a
 *    ('fabs', ('fneg', 'a')) => ('fabs', 'a')
 *    ('fabs', ('u2f', 'a')) => ('u2f', 'a')
 *    ('iabs', ('iabs', 'a')) => ('iabs', 'a')
 *    ('iabs', ('ineg', 'a')) => ('iabs', 'a')
 *    ('~fadd', 'a', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a
 *    ('iadd', 'a', 0) => a
 *    ('iadd_sat', 'a', 0) => a
 *    ('isub_sat', 'a', 0) => a
 *    ('uadd_sat', 'a', 0) => a
 *    ('usub_sat', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', -1) => -1
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd'))
 *    ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('iadd', ('ishl', 'b', 'a'), ('ishl', 'c', 'a')) => ('ishl', ('iadd', 'b', 'c'), 'a')
 *    ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c'))
 *    ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c'))
 *    ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ine', ('iand', 'a', 'b'), 0)
 *    ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ieq', ('iand', 'a', 'b'), 0)
 *    ('ieq', ('ushr(is_used_once)', 'a', '#b'), 0) => ('ult', 'a', ('ishl', 1, 'b'))
 *    ('ine', ('ushr(is_used_once)', 'a', '#b'), 0) => ('uge', 'a', ('ishl', 1, 'b'))
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a'))
 *    ('~fmul', 'a', 0.0) => 0.0
 *    ('fmul', 'a@16', 0.0) => 0.0
 *    ('fmul', 'a@32', 0.0) => 0.0
 *    ('fmulz', 'a', 0.0) => 0.0
 *    ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a', 'a') => ('fmul', 'a', 'a')
 *    ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b')
 *    ('imul', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', -1) => a
 *    ('~fmul', 'a', 1.0) => a
 *    ('~fmulz', 'a', 1.0) => a
 *    ('fmul(is_only_used_as_float)', 'a', 1.0) => a
 *    ('imul', 'a', 1) => a
 *    ('fmul', 'a', -1.0) => ('fneg', 'a')
 *    ('imul', 'a', -1) => ('ineg', 'a')
 *    ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a')
 *    ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a')
 *    ('~ffma', 0.0, 'a', 'b') => b
 *    ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b')
 *    ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b')
 *    ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~flrp', 'a', 'b', 0.0) => a
 *    ('~flrp', 'a', 'b', 1.0) => b
 *    ('~flrp', 'a', 'a', 'b') => a
 *    ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b')
 *    ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('sdot_4x8_iadd', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd', 'a', 0, 'b') => b
 *    ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd_sat', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 0, 'a', 'b') => b
 *    ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b
 *    ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c')
 *    ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b')
 *    ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b')
 *    ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 1.0, ('fexp2', ('fmul@32', 'a', 'b'))) => ('fexp2', ('fmulz', 'a', 'b'))
 *    ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a')
 *    ('ftrunc@16', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ftrunc@32', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ftrunc@64', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('fadd@16', 'a', ('fadd@16', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@16', 'b', ('ffloor', 'a'))
 *    ('fadd@32', 'a', ('fadd@32', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@32', 'b', ('ffloor', 'a'))
 *    ('fadd@64', 'a', ('fadd@64', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@64', 'b', ('ffloor', 'a'))
 *    ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a')))
 *    ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e'))
 *    ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b')
 *    ('fdot4', 'a', 0.0) => 0.0
 *    ('fdot3', 'a', 0.0) => 0.0
 *    ('fdot2', 'a', 0.0) => 0.0
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y')
 *    ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b'))
 *    ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b'))
 *    ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b'))
 *    ('fsum2', 'a') => ('fadd', 'a.x', 'a.y')
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd'))
 *    ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c'))
 *    ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7))
 *    ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15))
 *    ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31))
 *    ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63))
 *    ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b'))
 *    ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b'))
 *    ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b'))
 *    ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b'))
 *    ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b')
 *    ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b')
 *    ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1)
 *    ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1)
 *    ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2)
 *    ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2)
 *    ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16)
 *    ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16)
 *    ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16)
 *    ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24)
 *    ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24)
 *    ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd'))
 *    ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd'))
 *    ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b')
 *    ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b')
 *    ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b')
 *    ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b')
 *    ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b')
 *    ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b')
 *    ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b')
 *    ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b')
 *    ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b')
 *    ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False
 *    ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False
 *    ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False
 *    ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False
 *    ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('ieq', ('ineg', 'a'), 0) => ('ieq', 'a', 0)
 *    ('ine', ('ineg', 'a'), 0) => ('ine', 'a', 0)
 *    ('ieq', ('iabs', 'a'), 0) => ('ieq', 'a', 0)
 *    ('ine', ('iabs', 'a'), 0) => ('ine', 'a', 0)
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a')
 *    ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a')
 *    ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b')
 *    ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b')
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b')
 *    ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b')
 *    ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b')
 *    ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b')
 *    ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0)
 *    ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0)
 *    ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b'))
 *    ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0)
 *    ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0)
 *    ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0)
 *    ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0)
 *    ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0)
 *    ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0)
 *    ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a')
 *    ('fneu', ('b2f', 'a@1'), 0.0) => a
 *    ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a')
 *    ('ine', ('b2i', 'a@1'), 0) => a
 *    ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('u2f', 'a'), 0.0) => True
 *    ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a')
 *    ('flt', ('u2f', 'a'), 0.0) => False
 *    ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a')
 *    ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0)
 *    ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a')
 *    ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0)
 *    ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a')
 *    ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0)
 *    ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0)
 *    ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0)
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a'))
 *    ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b'))
 *    ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b')))
 *    ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b')))
 *    ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0))
 *    ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b')
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('bcsel', ('ult', 'b', 'a'), 'b', 'a') => ('umin', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 'b', 'a') => ('umax', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), 'b', 'a') => ('umin', 'a', 'b')
 *    ('bcsel', ('uge', 'b', 'a'), 'b', 'a') => ('umax', 'a', 'b')
 *    ('bcsel', ('ilt', 'b', 'a'), 'b', 'a') => ('imin', 'a', 'b')
 *    ('bcsel', ('ilt', 'a', 'b'), 'b', 'a') => ('imax', 'a', 'b')
 *    ('bcsel', ('ige', 'a', 'b'), 'b', 'a') => ('imin', 'a', 'b')
 *    ('bcsel', ('ige', 'b', 'a'), 'b', 'a') => ('imax', 'a', 'b')
 *    ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b')
 *    ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b')
 *    ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b')
 *    ('~fmin', 'a', 'a') => a
 *    ('~fmax', 'a', 'a') => a
 *    ('imin', 'a', 'a') => a
 *    ('imax', 'a', 'a') => a
 *    ('umin', 'a', 'a') => a
 *    ('umin', 'a', 0) => 0
 *    ('umin', 'a', -1) => a
 *    ('umax', 'a', 'a') => a
 *    ('umax', 'a', 0) => a
 *    ('umax', 'a', -1) => -1
 *    ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b')
 *    ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b')
 *    ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b')
 *    ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b')
 *    ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b')
 *    ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b')
 *    ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c')
 *    ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c')
 *    ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c')
 *    ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c')
 *    ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c')
 *    ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c')
 *    ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b')
 *    ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0)
 *    ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0)
 *    ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0)
 *    ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0)
 *    ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a'))
 *    ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a'))
 *    ('~fmin', 'a', ('fabs', 'a')) => a
 *    ('imin', 'a', ('iabs', 'a')) => a
 *    ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a
 *    ('imax', 'a', ('ineg', ('iabs', 'a'))) => a
 *    ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a')
 *    ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a')
 *    ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a')
 *    ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a')
 *    ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a')
 *    ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a'))
 *    ('fsat', ('b2f', 'a')) => ('b2f', 'a')
 *    ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0)
 *    ('fsat', ('fsat', 'a')) => ('fsat', 'a')
 *    ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b')))
 *    ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b')))
 *    ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c')
 *    ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c')
 *    ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c')
 *    ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b'))
 *    ('fmax', ('fsat(is_used_once)', 'a'), ('fsat(is_used_once)', 'b')) => ('fsat', ('fmax', 'a', 'b'))
 *    ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b'))
 *    ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a'))
 *    ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a'))
 *    ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b'))
 *    ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255)
 *    ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c'))
 *    ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c')
 *    ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c'))
 *    ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c')
 *    ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c'))
 *    ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c'))
 *    ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c'))
 *    ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c')
 *    ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c'))
 *    ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c')
 *    ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c'))
 *    ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c'))
 *    ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b')
 *    ('umin', ('iand', 'a', '#b(is_pos_power_of_two)'), ('iand', 'c', 'b')) => ('iand', ('iand', 'a', 'b'), ('iand', 'c', 'b'))
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('feq(is_used_once)', 'a@16', 0.0), ('feq', 'b@16', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('ior', ('fneu(is_used_once)', 'a@16', 0.0), ('fneu', 'b@16', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('feq(is_used_once)', 'a@16', 0.0), ('feq', 'b@16', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('fneu(is_used_once)', 'a@16', 0.0), ('fneu', 'b@16', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a')
 *    ('~f2i16', ('i2f', 'a@16')) => a
 *    ('~f2i16', ('u2f', 'a@16')) => a
 *    ('~f2u16', ('i2f', 'a@16')) => a
 *    ('~f2u16', ('u2f', 'a@16')) => a
 *    ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a'))
 *    ('~f2f16', ('f2f', 'a')) => ('f2f16', 'a')
 *    ('~f2f16', ('u2f', 'a')) => ('u2f16', 'a')
 *    ('~f2f16', ('i2f', 'a')) => ('i2f16', 'a')
 *    ('~f2u16', ('f2f', 'a')) => ('f2u16', 'a')
 *    ('~f2i16', ('f2f', 'a')) => ('f2i16', 'a')
 *    ('i2f16', ('f2i', ('fsign', 'a@16'))) => ('fsign', 'a')
 *    ('~f2f16', ('f2f32', 'a@16')) => a
 *    ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a')
 *    ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a')
 *    ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a')
 *    ('~f2f16', ('f2f64', 'a@16')) => a
 *    ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a')
 *    ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('feq(is_used_once)', 'a@32', 0.0), ('feq', 'b@32', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('ior', ('fneu(is_used_once)', 'a@32', 0.0), ('fneu', 'b@32', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('feq(is_used_once)', 'a@32', 0.0), ('feq', 'b@32', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('fneu(is_used_once)', 'a@32', 0.0), ('fneu', 'b@32', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a')
 *    ('~f2i32', ('i2f', 'a@32')) => a
 *    ('~f2i32', ('u2f', 'a@32')) => a
 *    ('~f2u32', ('i2f', 'a@32')) => a
 *    ('~f2u32', ('u2f', 'a@32')) => a
 *    ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a'))
 *    ('~f2f32', ('f2f', 'a')) => ('f2f32', 'a')
 *    ('~f2f32', ('u2f', 'a')) => ('u2f32', 'a')
 *    ('~f2f32', ('i2f', 'a')) => ('i2f32', 'a')
 *    ('~f2u32', ('f2f', 'a')) => ('f2u32', 'a')
 *    ('~f2i32', ('f2f', 'a')) => ('f2i32', 'a')
 *    ('i2f32', ('f2i', ('fsign', 'a@32'))) => ('fsign', 'a')
 *    ('~f2f32', ('f2f64', 'a@32')) => a
 *    ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a')
 *    ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a')
 *    ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('feq(is_used_once)', 'a@64', 0.0), ('feq', 'b@64', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('ior', ('fneu(is_used_once)', 'a@64', 0.0), ('fneu', 'b@64', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('feq(is_used_once)', 'a@64', 0.0), ('feq', 'b@64', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('iand', ('fneu(is_used_once)', 'a@64', 0.0), ('fneu', 'b@64', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0)
 *    ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a')
 *    ('~f2i64', ('i2f', 'a@64')) => a
 *    ('~f2i64', ('u2f', 'a@64')) => a
 *    ('~f2u64', ('i2f', 'a@64')) => a
 *    ('~f2u64', ('u2f', 'a@64')) => a
 *    ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a'))
 *    ('~f2f64', ('f2f', 'a')) => ('f2f64', 'a')
 *    ('~f2f64', ('u2f', 'a')) => ('u2f64', 'a')
 *    ('~f2f64', ('i2f', 'a')) => ('i2f64', 'a')
 *    ('~f2u64', ('f2f', 'a')) => ('f2u64', 'a')
 *    ('~f2i64', ('f2f', 'a')) => ('f2i64', 'a')
 *    ('i2f64', ('f2i', ('fsign', 'a@64'))) => ('fsign', 'a')
 *    ('i2i1', ('i2i8', 'a@1')) => a
 *    ('u2u1', ('u2u8', 'a@1')) => a
 *    ('i2i1', ('i2i16', 'a@1')) => a
 *    ('u2u1', ('u2u16', 'a@1')) => a
 *    ('u2f16', ('u2u8', 'a@1')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i8', 'a@1')) => ('i2f16', 'a')
 *    ('u2f16', ('u2u16', 'a@1')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i16', 'a@1')) => ('i2f16', 'a')
 *    ('u2f16', ('u2u32', 'a@1')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i32', 'a@1')) => ('i2f16', 'a')
 *    ('u2f16', ('u2u64', 'a@1')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i64', 'a@1')) => ('i2f16', 'a')
 *    ('i2i1', ('i2i32', 'a@1')) => a
 *    ('u2u1', ('u2u32', 'a@1')) => a
 *    ('u2f32', ('u2u8', 'a@1')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i8', 'a@1')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u16', 'a@1')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i16', 'a@1')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u32', 'a@1')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i32', 'a@1')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u64', 'a@1')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i64', 'a@1')) => ('i2f32', 'a')
 *    ('i2i1', ('i2i64', 'a@1')) => a
 *    ('u2u1', ('u2u64', 'a@1')) => a
 *    ('u2f64', ('u2u8', 'a@1')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i8', 'a@1')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u16', 'a@1')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i16', 'a@1')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u32', 'a@1')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i32', 'a@1')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u64', 'a@1')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@1')) => ('i2f64', 'a')
 *    ('i2i8', ('i2i16', 'a@8')) => a
 *    ('u2u8', ('u2u16', 'a@8')) => a
 *    ('u2f16', ('u2u16', 'a@8')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i16', 'a@8')) => ('i2f16', 'a')
 *    ('u2f16', ('u2u32', 'a@8')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i32', 'a@8')) => ('i2f16', 'a')
 *    ('u2f16', ('u2u64', 'a@8')) => ('u2f16', 'a')
 *    ('i2f16', ('i2i64', 'a@8')) => ('i2f16', 'a')
 *    ('i2i8', ('i2i32', 'a@8')) => a
 *    ('u2u8', ('u2u32', 'a@8')) => a
 *    ('u2f32', ('u2u16', 'a@8')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i16', 'a@8')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u32', 'a@8')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i32', 'a@8')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u64', 'a@8')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i64', 'a@8')) => ('i2f32', 'a')
 *    ('i2i8', ('i2i64', 'a@8')) => a
 *    ('u2u8', ('u2u64', 'a@8')) => a
 *    ('u2f64', ('u2u16', 'a@8')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i16', 'a@8')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u32', 'a@8')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i32', 'a@8')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u64', 'a@8')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@8')) => ('i2f64', 'a')
 *    ('i2i16', ('i2i32', 'a@16')) => a
 *    ('u2u16', ('u2u32', 'a@16')) => a
 *    ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a')
 *    ('u2f32', ('u2u64', 'a@16')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i64', 'a@16')) => ('i2f32', 'a')
 *    ('i2i16', ('i2i64', 'a@16')) => a
 *    ('u2u16', ('u2u64', 'a@16')) => a
 *    ('u2f64', ('u2u32', 'a@16')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i32', 'a@16')) => ('i2f64', 'a')
 *    ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a')
 *    ('i2i32', ('i2i64', 'a@32')) => a
 *    ('u2u32', ('u2u64', 'a@32')) => a
 *    ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a')
 *    ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a')
 *    ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a')
 *    ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a')
 *    ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a')
 *    ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a')
 *    ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a')
 *    ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a')
 *    ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a')
 *    ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b')
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b')
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b')
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i8', 'a@8')) => a
 *    ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2))
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i16', 'a@16')) => a
 *    ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3))
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i32', 'a@32')) => a
 *    ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4))
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i64', 'a@64')) => a
 *    ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5))
 *    ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a')
 *    ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a')
 *    ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a')
 *    ('ior', 'a', ('ieq', 'a', False)) => True
 *    ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b')
 *    ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b'))
 *    ('ishl', ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)), '#c') => ('bcsel', ('ige', ('iand', 'c', 31), ('find_lsb', 'b')), ('ishl', ('iand', 'a', 'b'), ('iadd', ('iand', 'c', 31), ('ineg', ('find_lsb', 'b')))), ('ushr', ('iand', 'a', 'b'), ('iadd', ('ineg', ('iand', 'c', 31)), ('find_lsb', 'b'))))
 *    ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)) => ('ushr', ('iand', 'a', 'b'), ('find_lsb', 'b'))
 *    ('ior', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('ior', ('b2i', 'a'), 'b'), 1)
 *    ('iand', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('b2i', 'a'), 'b')
 *    ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0))
 *    ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0))
 *    ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b')
 *    ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a')
 *    ('~fge', 'a', ('fmin', 'b', 'a')) => True
 *    ('~fge', ('fmax', 'a', 'b'), 'a') => True
 *    ('flt', 'a', ('fmin', 'b', 'a')) => False
 *    ('flt', ('fmax', 'a', 'b'), 'a') => False
 *    ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b')
 *    ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a')
 *    ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b')
 *    ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a')
 *    ('ige', 'a', ('imin', 'b', 'a')) => True
 *    ('ige', ('imax', 'a', 'b'), 'a') => True
 *    ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b')
 *    ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a')
 *    ('uge', 'a', ('umin', 'b', 'a')) => True
 *    ('uge', ('umax', 'a', 'b'), 'a') => True
 *    ('ilt', 'a', ('imin', 'b', 'a')) => False
 *    ('ilt', ('imax', 'a', 'b'), 'a') => False
 *    ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b')
 *    ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a')
 *    ('ult', 'a', ('umin', 'b', 'a')) => False
 *    ('ult', ('umax', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b')
 *    ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a')
 *    ('ult', 'a', ('iand', 'b', 'a')) => False
 *    ('ult', ('ior', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('iand', 'b', 'a')) => True
 *    ('uge', ('ior', 'a', 'b'), 'a') => True
 *    ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b'))
 *    ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b')
 *    ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b')
 *    ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b')
 *    ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b')
 *    ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b')
 *    ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b'))
 *    ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b'))
 *    ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b'))
 *    ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b'))
 *    ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y'))
 *    ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0)
 *    ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0)
 *    ('fall_equal8', 'a', 'b') => ('seq', ('fany_nequal8', 'a', 'b'), 0.0)
 *    ('fall_equal16', 'a', 'b') => ('seq', ('fany_nequal16', 'a', 'b'), 0.0)
 *    ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y'))
 *    ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal8', 'a', 'b') => ('fsat', ('fdot8', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal16', 'a', 'b') => ('fsat', ('fdot16', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y'))
 *    ('ball_iequal3', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('iand', ('ieq', 'a.y', 'b.y'), ('ieq', 'a.z', 'b.z')))
 *    ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w')))
 *    ('ball_iequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h'))))
 *    ('ball_iequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('ieq', 'a.i', 'b.i'), ('ieq', 'a.j', 'b.j')), ('iand', ('ieq', 'a.k', 'b.k'), ('ieq', 'a.l', 'b.l'))), ('iand', ('iand', ('ieq', 'a.m', 'b.m'), ('ieq', 'a.n', 'b.n')), ('iand', ('ieq', 'a.o', 'b.o'), ('ieq', 'a.p', 'b.p')))))
 *    ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y'))
 *    ('ball_fequal3', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('iand', ('feq', 'a.y', 'b.y'), ('feq', 'a.z', 'b.z')))
 *    ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w')))
 *    ('ball_fequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h'))))
 *    ('ball_fequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('feq', 'a.i', 'b.i'), ('feq', 'a.j', 'b.j')), ('iand', ('feq', 'a.k', 'b.k'), ('feq', 'a.l', 'b.l'))), ('iand', ('iand', ('feq', 'a.m', 'b.m'), ('feq', 'a.n', 'b.n')), ('iand', ('feq', 'a.o', 'b.o'), ('feq', 'a.p', 'b.p')))))
 *    ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y'))
 *    ('bany_inequal3', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ior', ('ine', 'a.y', 'b.y'), ('ine', 'a.z', 'b.z')))
 *    ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w')))
 *    ('bany_inequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h'))))
 *    ('bany_inequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('ine', 'a.i', 'b.i'), ('ine', 'a.j', 'b.j')), ('ior', ('ine', 'a.k', 'b.k'), ('ine', 'a.l', 'b.l'))), ('ior', ('ior', ('ine', 'a.m', 'b.m'), ('ine', 'a.n', 'b.n')), ('ior', ('ine', 'a.o', 'b.o'), ('ine', 'a.p', 'b.p')))))
 *    ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y'))
 *    ('bany_fnequal3', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('ior', ('fneu', 'a.y', 'b.y'), ('fneu', 'a.z', 'b.z')))
 *    ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w')))
 *    ('bany_fnequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h'))))
 *    ('bany_fnequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('fneu', 'a.i', 'b.i'), ('fneu', 'a.j', 'b.j')), ('ior', ('fneu', 'a.k', 'b.k'), ('fneu', 'a.l', 'b.l'))), ('ior', ('ior', ('fneu', 'a.m', 'b.m'), ('fneu', 'a.n', 'b.n')), ('ior', ('fneu', 'a.o', 'b.o'), ('fneu', 'a.p', 'b.p')))))
 *    ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0)
 *    ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0)
 *    ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b'))
 *    ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b'))
 *    ('iand', 'a@bool16', 1.0) => ('b2f', 'a')
 *    ('iand', 'a@bool32', 1.0) => ('b2f', 'a')
 *    ('flt', ('fneg', ('b2f', 'a@1')), 0) => a
 *    ('ilt', 'a', 'a') => False
 *    ('ige', 'a', 'a') => True
 *    ('ieq', 'a', 'a') => True
 *    ('ine', 'a', 'a') => False
 *    ('ult', 'a', 'a') => False
 *    ('uge', 'a', 'a') => True
 *    ('flt', 'a', 'a') => False
 *    ('fge', 'a(is_a_number)', 'a') => True
 *    ('feq', 'a(is_a_number)', 'a') => True
 *    ('fneu', 'a(is_a_number)', 'a') => False
 *    ('iand', 'a', 'a') => a
 *    ('iand', 'a', 0) => 0
 *    ('iand', 'a', -1) => a
 *    ('iand', 'a', ('inot', 'a')) => 0
 *    ('ior', 'a', 'a') => a
 *    ('ior', 'a', 0) => a
 *    ('ior', 'a', -1) => -1
 *    ('ior', 'a', ('inot', 'a')) => -1
 *    ('ixor', 'a', 'a') => 0
 *    ('ixor', 'a', 0) => a
 *    ('ixor', 'a', ('ixor', 'a', 'b')) => b
 *    ('ixor', 'a', -1) => ('inot', 'a')
 *    ('inot', ('inot', 'a')) => a
 *    ('ior', ('iand', 'a', 'b'), 'b') => b
 *    ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b')
 *    ('iand', ('ior', 'a', 'b'), 'b') => b
 *    ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b')
 *    ('iand', ('iand(is_used_once)', 'a', 'b'), ('iand(is_used_once)', 'a', 'c')) => ('iand', 'a', ('iand', 'b', 'c'))
 *    ('iand@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', 0, ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('iand@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), 0)
 *    ('iand@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('iand@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('ior@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ior@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('ior@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', -1, ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ior@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), -1)
 *    ('ixor@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ixor', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ixor@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ixor', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b'))
 *    ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b'))
 *    ('ishl', 0, 'a') => 0
 *    ('ishl', 'a', 0) => a
 *    ('ishr', 0, 'a') => 0
 *    ('ishr', -1, 'a') => -1
 *    ('ishr', 'a', 0) => a
 *    ('ushr', 0, 'a') => 0
 *    ('ushr', 'a', 0) => a
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b')
 *    ('urol@8', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 8, 'b')))
 *    ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b')))
 *    ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b')))
 *    ('urol@64', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 64, 'b')))
 *    ('uror@8', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 8, 'b')))
 *    ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b')))
 *    ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b')))
 *    ('uror@64', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 64, 'b')))
 *    ('bfi', 4294967295, 'a', 'b') => a
 *    ('bfi', 0, 'a', 'b') => b
 *    ('bfi', ('ineg', ('b2i', 'a@1')), 'b', 'c') => ('bcsel', 'a', 'b', 'c')
 *    ('bfi', '#a(is_odd)', 'a', 'b') => ('ior', 'a', 'b')
 *    ('bfi', '#a(is_odd)', 'b', 0) => ('iand', 'a', 'b')
 *    ('u2f32', ('bfi', '#a(is_pos_power_of_two)', 'b', 0)) => ('bcsel', ('ieq', ('iand', 'b', 1), 0), ('iand', 'b', 1), ('u2f', 'a'))
 *    ('u2f', ('bfi', '#a(is_pos_power_of_two)', 'b', 0)) => ('bcsel', ('ieq', ('iand', 'b', 1), 0), 0, ('u2f', 'a'))
 *    ('~fexp2', ('flog2', 'a')) => a
 *    ('~flog2', ('fexp2', 'a')) => a
 *    ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b')
 *    ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 'a', 1.0) => a
 *    ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a')
 *    ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 2.0, 'a') => ('fexp2', 'a')
 *    ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a
 *    ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a')
 *    ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a'))
 *    ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a'))
 *    ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a'))
 *    ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a'))
 *    ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a'))
 *    ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a'))
 *    ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a'))
 *    ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b'))
 *    ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0))
 *    ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fdiv', 1.0, 'a') => ('frcp', 'a')
 *    ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b'))
 *    ('~frcp', ('frcp', 'a')) => a
 *    ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a')
 *    ('fsqrt', 'a') => ('frcp', ('frsq', 'a'))
 *    ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a')
 *    ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))
 *    ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))
 *    ('ieq', 'a', True) => a
 *    ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a')
 *    ('ine', 'a', False) => a
 *    ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a')
 *    ('bcsel', 'a', True, False) => a
 *    ('bcsel', 'a', False, True) => ('inot', 'a')
 *    ('bcsel', True, 'b', 'c') => b
 *    ('bcsel', False, 'b', 'c') => c
 *    ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel', 'a', 'b', 'b') => b
 *    ('~fcsel', 'a', 'b', 'b') => b
 *    ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1'))
 *    ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a')))
 *    ('bcsel', 'a', 1, 0) => ('b2i', 'a@1')
 *    ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a'))
 *    ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a
 *    ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a')
 *    ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a')
 *    ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a
 *    ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a
 *    ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a')
 *    ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a')
 *    ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('umax', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b'))
 *    ('umin', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('ine', ('umin', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('iand', 'a', ('ine', 'b', 0))
 *    ('ine', ('umax', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('ior', 'a', ('ine', 'b', 0))
 *    ('f2i', ('ftrunc', 'a')) => ('f2i', 'a')
 *    ('f2u', ('ftrunc', 'a')) => ('f2u', 'a')
 *    ('f2fmp', ('f2f32', 'a@16')) => a
 *    ('i2imp', ('i2i32', 'a@16')) => a
 *    ('i2imp', ('u2u32', 'a@16')) => a
 *    ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a')
 *    ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a')
 *    ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a')
 *    ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a')
 *    ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('f2f32', ('f2fmp', 'a@32')) => a
 *    ('i2i32', ('i2imp', 'a@32')) => a
 *    ('u2u32', ('i2imp', 'a@32')) => a
 *    ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a')
 *    ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a')
 *    ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a')
 *    ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a')
 *    ('f2i32', ('f2fmp', 'a@32')) => ('f2i32', 'a')
 *    ('f2u32', ('f2fmp', 'a@32')) => ('f2u32', 'a')
 *    ('i2f32', ('i2imp', 'a@32')) => ('i2f32', 'a')
 *    ('ffloor', 'a(is_integral)') => a
 *    ('fceil', 'a(is_integral)') => a
 *    ('ftrunc', 'a(is_integral)') => a
 *    ('fround_even', 'a(is_integral)') => a
 *    ('~ffract', 'a(is_integral)') => 0.0
 *    ('fabs', 'a(is_not_negative)') => a
 *    ('iabs', 'a(is_not_negative)') => a
 *    ('fsat', 'a(is_not_positive)') => 0.0
 *    ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a')
 *    ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0)
 *    ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0)
 *    ('fneu', 'a(is_not_zero)', 0.0) => True
 *    ('feq', 'a(is_not_zero)', 0.0) => False
 *    ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True
 *    ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True
 *    ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True
 *    ('ine', 'a(is_not_zero)', 0) => True
 *    ('ieq', 'a(is_not_zero)', 0) => False
 *    ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True
 *    ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True
 *    ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True
 *    ('ult', 0, 'a(is_gt_zero)') => True
 *    ('ult', 'a', 0) => False
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y
 *    ('unpack_64_2x32_split_x', ('u2u64', 'a@32')) => a
 *    ('unpack_64_2x32_split_y', ('u2u64', 'a')) => 0
 *    ('unpack_64_2x32_split_x', ('i2i64', 'a@32')) => a
 *    ('unpack_64_2x32_split_y', ('i2i64(is_used_once)', 'a@32')) => ('ishr', 'a', 31)
 *    ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b')
 *    ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a
 *    ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a
 *    ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a
 *    ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a
 *    ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a
 *    ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a
 *    ('unpack_64_4x16', ('pack_64_4x16', 'a')) => a
 *    ('unpack_64_4x16', ('pack_64_2x32', ('vec2', ('pack_32_2x16_split', 'a', 'b'), ('pack_32_2x16_split', 'c', 'd')))) => ('vec4', 'a', 'b', 'c', 'd')
 *    ('unpack_64_4x16', ('pack_64_2x32_split', ('pack_32_2x16_split', 'a', 'b'), ('pack_32_2x16_split', 'c', 'd'))) => ('vec4', 'a', 'b', 'c', 'd')
 *    ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c'))
 *    ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1)
 *    ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3)
 *    ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7)
 *    ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1)
 *    ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3)
 *    ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7)
 *    ('iand', 255, 'a') => ('extract_u8', 'a', 0)
 *    ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2))
 *    ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1))
 *    ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b'))
 *    ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b'))
 *    ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0)
 *    ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1)
 *    ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2)
 *    ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3)
 *    ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0)
 *    ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1)
 *    ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2)
 *    ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3)
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0)
 *    ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1)
 *    ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0)
 *    ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1)
 *    ('iand', 65535, 'a') => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1)
 *    ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0)
 *    ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1)
 *    ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a'))
 *    ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32))
 *    ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16))
 *    ('pack_half_2x16_split', 'a', 'b') => ('pack_half_2x16_rtz_split', 'a', 'b')
 *    ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a')
 *    ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32))
 *    ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16))
 *    ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a')
 *    ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a')
 *    ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a')
 *    ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a'))
 *    ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0))
 *    ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a')
 *    ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0)
 *    ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0)
 *    ('ishl', ('pack_half_2x16_rtz_split', 'a', 0), 16) => ('pack_half_2x16_rtz_split', 0, 'a')
 *    ('ushr', ('pack_half_2x16_rtz_split', 0, 'a'), 16) => ('pack_half_2x16_rtz_split', 'a', 0)
 *    ('extract_u16', ('pack_half_2x16_rtz_split', 0, 'a'), 1) => ('pack_half_2x16_rtz_split', 'a', 0)
 *    ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('iadd', ('pack_half_2x16_rtz_split', 'a', 0), ('pack_half_2x16_rtz_split', 0, 'b')) => ('pack_half_2x16_rtz_split', 'a', 'b')
 *    ('ior', ('pack_half_2x16_rtz_split', 'a', 0), ('pack_half_2x16_rtz_split', 0, 'b')) => ('pack_half_2x16_rtz_split', 'a', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd')
 *    ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6)
 *    ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5)
 *    ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4)
 *    ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3)
 *    ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0)
 *    ('ussub_4x8_vc4', 'a', 0) => a
 *    ('ussub_4x8_vc4', 'a', -1) => 0
 *    ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b'))
 *    ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b'))
 *    ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('bitz', 'a', 'b') => ('inot', ('bitnz', 'a', 'b'))
 *    ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b'))
 *    ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b'))
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b'))
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c')
 *    ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b'))
 *    ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b')
 *    ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b')
 *    ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b')
 *    ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a
 *    ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ige', ('ifind_msb_rev', 'a'), 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ige', ('ufind_msb_rev', 'a'), 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', 'a', -1), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), ('ifind_msb', 'a')) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), ('ifind_msb', 'a'), ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), ('ifind_msb_rev', 'a')) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), ('ifind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('find_lsb', ('bitfield_reverse', 'a')) => ('ufind_msb_rev', 'a')
 *    ('ufind_msb_rev', ('bitfield_reverse', 'a')) => ('find_lsb', 'a')
 *    ('ifind_msb', ('f2i32(is_used_once)', 'a')) => ('ufind_msb', ('f2i32', ('fabs', 'a')))
 *    ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('fabs', ('bcsel(is_used_once)', 'b', ('fneg', 'a'), 'a')) => ('fabs', 'a')
 *    ('fabs', ('bcsel(is_used_once)', 'b', 'a', ('fneg', 'a'))) => ('fabs', 'a')
 *    ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd')
 *    ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b'))))
 *    ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b'))))
 *    ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a'))
 *    ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b'))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base'))
 *    ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0))
 *    ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0))
 *    ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807))
 *    ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base'))
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits'))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits'))
 *    ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b')
 *    ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c')
 *    ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits')
 *    ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits')
 *    ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset')
 *    ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset')
 *    ('ult', 'a', ('umin', ('iand', 'a', 'b'), 'c')) => False
 *    ('ult', 31, ('umin', '#bits(is_ult_32)', 'a')) => False
 *    ('ubfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ubfe', 'value', 'offset', 'width')
 *    ('ibfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ibfe', 'value', 'offset', 'width')
 *    ('bfm', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset')))), 'offset') => ('bfm', 'width', 'offset')
 *    ('iadd@32', ('ishl', 1, 'a'), -1) => ('bfm', 'a', 0)
 *    ('ishl', ('bfm', 'a', 0), 'b') => ('bfm', 'a', 'b')
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b')))
 *    ('b2i32', ('ine', ('ubfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1)
 *    ('b2i32', ('ine', ('ibfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1)
 *    ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits')))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1)))
 *    ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value'))
 *    ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value'))
 *    ('ifind_msb', 'value') => ('isub', 31, ('uclz', ('ixor', 'value', ('ishr', 'value', 31))))
 *    ('ufind_msb', 'value@32') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value'))
 *    ('ufind_msb', 'value@32') => ('isub', 31, ('uclz', 'value'))
 *    ('uclz', 'a') => ('umin', 32, ('ufind_msb_rev', 'a'))
 *    ('find_lsb', 'value@64') => ('ufind_msb', ('iand', 'value', ('ineg', 'value')))
 *    ('find_lsb', 'value') => ('ufind_msb', ('u2u32', ('iand', 'value', ('ineg', 'value'))))
 *    ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24)
 *    ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255)
 *    ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16)
 *    ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535)
 *    ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0))))
 *    ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0))))
 *    ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0))))
 *    ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0))))
 *    ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0)
 *    ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0)
 *    ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0)))
 *    ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0)))
 *    ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a')))
 *    ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a'))
 *    ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16)))
 *    ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a'))
 *    ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16))
 *    ('isign', 'a') => ('imin', ('imax', 'a', -1), 1)
 *    ('imin', ('imax', 'a', -1), 1) => ('isign', 'a')
 *    ('imax', ('imin', 'a', 1), -1) => ('isign', 'a')
 *    ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0)))
 *    ('amul', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215))
 *    ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c')
 *    ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b')
 *    ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c')
 *    ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b')
 *    ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b')
 *    ('imad24_ir3', 'a', 0, 'c') => c
 *    ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c')
 *    ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c')
 *    ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('imul24', 'a', 0) => 0
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b')))
 *    ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b')))
 *    ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b')))
 *    ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b')))
 *    ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b')))
 *    ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b')))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b')))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b')))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('f2f', ('b2f', 'a')) => ('b2f', 'a')
 *    ('f2u', ('b2f', 'a')) => ('b2i', 'a')
 *    ('f2i', ('b2f', 'a')) => ('b2i', 'a')
 *    ('u2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('u2u', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('i2i', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2i1', 'a@1') => a
 *    ('i2i8', 'a@8') => a
 *    ('i2i16', 'a@16') => a
 *    ('i2i32', 'a@32') => a
 *    ('i2i64', 'a@64') => a
 *    ('u2u1', 'a@1') => a
 *    ('u2u8', 'a@8') => a
 *    ('u2u16', 'a@16') => a
 *    ('u2u32', 'a@32') => a
 *    ('u2u64', 'a@64') => a
 *    ('f2f16', 'a@16') => a
 *    ('f2f32', 'a@32') => a
 *    ('f2f64', 'a@64') => a
 *    ('b2b1', 'a@1') => a
 *    ('b2b8', 'a@8') => a
 *    ('b2b16', 'a@16') => a
 *    ('b2b32', 'a@32') => a
 *    ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b')))
 *    ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b')))
 *    ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a')
 *    ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a')
 *    ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a')
 *    ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a')
 *    ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10)))
 *    ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23))
 *    ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20)))
 *    ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x')
 *    ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x')
 *    ('bcsel', ('iand', ('ine', ('fabs', 'x'), 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', 'x', 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', 'x', 0), 4294967295), ('pack_half_2x16_rtz_split', 'x', 0)) => ('pack_half_2x16_rtz_split', 'x', 0)
 *    ('bcsel', ('iand', ('ine', 'x', 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0), 4294967295), ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0)) => ('pack_half_2x16_rtz_split', 'x', 0)
 *    ('bcsel', ('iand', ('ine', ('fabs', 'x'), 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0), 4294967295), ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0)) => ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0)
 *    ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b')
 *    ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b')
 *    ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b')
 *    ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b')
 *    ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b')
 *    ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b')
 *    ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b')
 *    ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b')
 *    ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b')
 *    ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b')
 *    ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b')
 *    ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b')
 *    ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b')
 *    ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b')
 *    ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b')
 *    ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b')
 *    ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b')
 *    ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b')
 *    ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b')
 *    ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a')
 *    ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd'))
 *    ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd'))
 *    ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd'))
 *    ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd'))
 *    ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd'))
 *    ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c'))
 *    ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd'))
 *    ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c'))
 *    ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd'))
 *    ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c'))
 *    ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd'))
 *    ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c'))
 *    ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd'))
 *    ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c'))
 *    ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd'))
 *    ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c'))
 *    ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd'))
 *    ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd'))
 *    ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd'))
 *    ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd'))
 *    ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd'))
 *    ('fddx', 'a') => 0.0
 *    ('fddx_fine', 'a') => 0.0
 *    ('fddx_coarse', 'a') => 0.0
 *    ('fddy', 'a') => 0.0
 *    ('fddy_fine', 'a') => 0.0
 *    ('fddy_coarse', 'a') => 0.0
 *    ('umul_low', '#a(is_lower_half_zero)', 'b') => 0
 *    ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c
 *    ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin_mdg(is_used_once)', 'a'), ('fsin_mdg(is_used_once)', 'b')) => ('fsin_mdg', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos_mdg(is_used_once)', 'a'), ('fcos_mdg(is_used_once)', 'b')) => ('fcos_mdg', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin_agx(is_used_once)', 'a'), ('fsin_agx(is_used_once)', 'b')) => ('fsin_agx', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b'))
 *    ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b'))
 *    ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b'))
 *    ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b'))
 *    ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048))
 *    ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216))
 *    ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992))
 *    ('bfi', 4294967295, 'a', '#b') => ('iadd', ('ishl', 'a', 0), ('iand', 'b', -4294967296))
 *    ('bfi', 8589934590, 'a', '#b') => ('iadd', ('ishl', 'a', 1), ('iand', 'b', -8589934591))
 *    ('bfi', 17179869180, 'a', '#b') => ('iadd', ('ishl', 'a', 2), ('iand', 'b', -17179869181))
 *    ('bfi', 34359738360, 'a', '#b') => ('iadd', ('ishl', 'a', 3), ('iand', 'b', -34359738361))
 *    ('bfi', 68719476720, 'a', '#b') => ('iadd', ('ishl', 'a', 4), ('iand', 'b', -68719476721))
 *    ('bfi', 137438953440, 'a', '#b') => ('iadd', ('ishl', 'a', 5), ('iand', 'b', -137438953441))
 *    ('bfi', 274877906880, 'a', '#b') => ('iadd', ('ishl', 'a', 6), ('iand', 'b', -274877906881))
 *    ('bfi', 549755813760, 'a', '#b') => ('iadd', ('ishl', 'a', 7), ('iand', 'b', -549755813761))
 *    ('bfi', 1099511627520, 'a', '#b') => ('iadd', ('ishl', 'a', 8), ('iand', 'b', -1099511627521))
 *    ('bfi', 2199023255040, 'a', '#b') => ('iadd', ('ishl', 'a', 9), ('iand', 'b', -2199023255041))
 *    ('bfi', 4398046510080, 'a', '#b') => ('iadd', ('ishl', 'a', 10), ('iand', 'b', -4398046510081))
 *    ('bfi', 8796093020160, 'a', '#b') => ('iadd', ('ishl', 'a', 11), ('iand', 'b', -8796093020161))
 *    ('bfi', 17592186040320, 'a', '#b') => ('iadd', ('ishl', 'a', 12), ('iand', 'b', -17592186040321))
 *    ('bfi', 35184372080640, 'a', '#b') => ('iadd', ('ishl', 'a', 13), ('iand', 'b', -35184372080641))
 *    ('bfi', 70368744161280, 'a', '#b') => ('iadd', ('ishl', 'a', 14), ('iand', 'b', -70368744161281))
 *    ('bfi', 140737488322560, 'a', '#b') => ('iadd', ('ishl', 'a', 15), ('iand', 'b', -140737488322561))
 *    ('bfi', 281474976645120, 'a', '#b') => ('iadd', ('ishl', 'a', 16), ('iand', 'b', -281474976645121))
 *    ('bfi', 562949953290240, 'a', '#b') => ('iadd', ('ishl', 'a', 17), ('iand', 'b', -562949953290241))
 *    ('bfi', 1125899906580480, 'a', '#b') => ('iadd', ('ishl', 'a', 18), ('iand', 'b', -1125899906580481))
 *    ('bfi', 2251799813160960, 'a', '#b') => ('iadd', ('ishl', 'a', 19), ('iand', 'b', -2251799813160961))
 *    ('bfi', 4503599626321920, 'a', '#b') => ('iadd', ('ishl', 'a', 20), ('iand', 'b', -4503599626321921))
 *    ('bfi', 9007199252643840, 'a', '#b') => ('iadd', ('ishl', 'a', 21), ('iand', 'b', -9007199252643841))
 *    ('bfi', 18014398505287680, 'a', '#b') => ('iadd', ('ishl', 'a', 22), ('iand', 'b', -18014398505287681))
 *    ('bfi', 36028797010575360, 'a', '#b') => ('iadd', ('ishl', 'a', 23), ('iand', 'b', -36028797010575361))
 *    ('bfi', 72057594021150720, 'a', '#b') => ('iadd', ('ishl', 'a', 24), ('iand', 'b', -72057594021150721))
 *    ('bfi', 144115188042301440, 'a', '#b') => ('iadd', ('ishl', 'a', 25), ('iand', 'b', -144115188042301441))
 *    ('bfi', 288230376084602880, 'a', '#b') => ('iadd', ('ishl', 'a', 26), ('iand', 'b', -288230376084602881))
 *    ('bfi', 576460752169205760, 'a', '#b') => ('iadd', ('ishl', 'a', 27), ('iand', 'b', -576460752169205761))
 *    ('bfi', 1152921504338411520, 'a', '#b') => ('iadd', ('ishl', 'a', 28), ('iand', 'b', -1152921504338411521))
 *    ('bfi', 2305843008676823040, 'a', '#b') => ('iadd', ('ishl', 'a', 29), ('iand', 'b', -2305843008676823041))
 *    ('bfi', 4611686017353646080, 'a', '#b') => ('iadd', ('ishl', 'a', 30), ('iand', 'b', -4611686017353646081))
 *    ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b'))
 *    ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b'))
 *    ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b'))
 *    ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b'))
 *    ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b'))
 *    ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'))
 *    ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))
 *    ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'))
 *    ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'))
 *    ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'))
 *    ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd'))
 *    ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd'))
 *    ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd'))
 *    ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd'))
 *    ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd'))
 */


static const nir_search_value_union nir_opt_algebraic_values[] = {
   /* ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('imul', 'a@8', 128) => ('ishl', 'a', 7) */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x80 /* 128 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 6, 7 },
      -1,
   } },

   /* replace1_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 9 },
      -1,
   } },

   /* ('imul', 'a@16', 32768) => ('ishl', 'a', 15) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x8000 /* 32768 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 12 },
      -1,
   } },

   /* replace2_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf /* 15 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 14 },
      -1,
   } },

   /* ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x80000000 /* 2147483648 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 17 },
      -1,
   } },

   /* replace3_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1f /* 31 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 19 },
      -1,
   } },

   /* ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x8000000000000000 /* 9223372036854775808 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 22 },
      -1,
   } },

   /* replace4_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3f /* 63 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 24 },
      -1,
   } },

   /* ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search5_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 26 },
      -1,
   } },

   /* replace5_0_0 -> 0 in the cache */
   /* replace5_0_1_0_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 28 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 30 },
      -1,
   } },

   /* ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 33 },
      -1,
   } },

   /* replace6_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 37 },
      -1,
   } },

   /* ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b'))) */
   /* search7_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 39 },
      -1,
   } },

   /* replace7_0_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 42 },
      -1,
   } },
   /* replace7_1_0 -> 21 in the cache */
   /* replace7_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 44 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 43, 45 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search8_0_0 -> 16 in the cache */
   /* search8_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 47 },
      -1,
   } },

   /* replace8_0 -> 16 in the cache */
   /* replace8_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search9_0_0 -> 16 in the cache */
   /* search9_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 50 },
      -1,
   } },

   /* replace9_0 -> 16 in the cache */
   /* replace9_1 -> 36 in the cache */
   /* replace9 -> 49 in the cache */

   /* ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b')) */
   /* search10_0 -> 16 in the cache */
   /* search10_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace10_0_0 -> 16 in the cache */
   /* replace10_0_1 -> 36 in the cache */
   /* replace10_0 -> 49 in the cache */
   /* replace10_1_0 -> 16 in the cache */
   /* replace10_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 53 },
      -1,
   } },

   /* ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b')) */
   /* search11_0 -> 16 in the cache */
   /* search11_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace11_0_0 -> 16 in the cache */
   /* replace11_0_1 -> 36 in the cache */
   /* replace11_0 -> 49 in the cache */
   /* replace11_1_0 -> 16 in the cache */
   /* replace11_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 56 },
      -1,
   } },

   /* ('udiv', 'a', 1) => a */
   /* search12_0 -> 32 in the cache */
   /* search12_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace12 -> 32 in the cache */

   /* ('idiv', 'a', 1) => a */
   /* search13_0 -> 32 in the cache */
   /* search13_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace13 -> 32 in the cache */

   /* ('umod', 'a', 1) => 0 */
   /* search14_0 -> 32 in the cache */
   /* search14_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('imod', 'a', 1) => 0 */
   /* search15_0 -> 32 in the cache */
   /* search15_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace15 -> 61 in the cache */

   /* ('imod', 'a', -1) => 0 */
   /* search16_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace16 -> 61 in the cache */

   /* ('irem', 'a', 1) => 0 */
   /* search17_0 -> 32 in the cache */
   /* search17_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace17 -> 61 in the cache */

   /* ('irem', 'a', -1) => 0 */
   /* search18_0 -> 32 in the cache */
   /* search18_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace18 -> 61 in the cache */

   /* ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b')) */
   /* search19_0 -> 0 in the cache */
   /* search19_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace19_0 -> 0 in the cache */
   /* replace19_1_0 -> 3 in the cache */
   /* replace19_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b'))) */
   /* search20_0 -> 0 in the cache */
   /* search20_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace20_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_1_0 -> 3 in the cache */
   /* replace20_1_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 4 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 72 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b'))))) */
   /* search21_0 -> 0 in the cache */
   /* search21_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace21_0_0_0 -> 0 in the cache */
   /* replace21_0_0 -> 70 in the cache */
   /* replace21_0_1_0_0 -> 0 in the cache */
   /* replace21_0_1_0 -> 71 in the cache */
   /* replace21_0_1_1_0_0 -> 3 in the cache */
   /* replace21_0_1_1_0 -> 28 in the cache */
   /* replace21_0_1_1 -> 29 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 75 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 76 },
      -1,
   } },

   /* ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search22_0 -> 0 in the cache */
   /* search22_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace22_0 -> 0 in the cache */
   /* replace22_1_0 -> 3 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3, 79 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 80 },
      -1,
   } },

   /* ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search23_0 -> 0 in the cache */
   /* search23_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace23_0 -> 0 in the cache */
   /* replace23_1_0 -> 3 in the cache */
   /* replace23_1_1 -> 79 in the cache */
   /* replace23_1 -> 80 in the cache */
   /* replace23 -> 81 in the cache */

   /* ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b')) */
   /* search24_0 -> 0 in the cache */
   /* search24_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace24_0_0_0 -> 0 in the cache */
   /* replace24_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace24_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 84, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace24_2_0 -> 0 in the cache */
   /* replace24_2_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 85, 86, 87 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b'))) */
   /* search25_0 -> 0 in the cache */
   /* search25_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace25_0 -> 0 in the cache */
   /* replace25_1_0_0_0 -> 0 in the cache */
   /* replace25_1_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 86 },
      -1,
   } },
   /* replace25_1_0_1_0 -> 0 in the cache */
   /* replace25_1_0_1_1_0 -> 3 in the cache */
   /* replace25_1_0_1_1_1 -> 79 in the cache */
   /* replace25_1_0_1_1 -> 80 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 80 },
      -1,
   } },
   /* replace25_1_0_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 90, 91, 0 },
      -1,
   } },
   /* replace25_1_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 92, 93 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 0, 94 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b')) */
   /* search26_0 -> 0 in the cache */
   /* search26_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace26_0 -> 0 in the cache */
   /* replace26_1_0 -> 3 in the cache */
   /* replace26_1 -> 28 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 28 },
      -1,
   } },

   /* ('~fmul', ('fsign', 'a'), ('ffloor', ('fadd', ('fabs', 'a'), 0.5))) => ('ftrunc', ('fadd', 'a', ('fmul', ('fsign', 'a'), 0.5))) */
   /* search27_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search27_1_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 32 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 99, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 101 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 98, 102 },
      -1,
   } },

   /* replace27_0_0 -> 32 in the cache */
   /* replace27_0_1_0_0 -> 32 in the cache */
   /* replace27_0_1_0 -> 98 in the cache */
   /* replace27_0_1_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 98, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 32, 104 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 2,
      { 105 },
      -1,
   } },

   /* ('~fneg', ('fneg', 'a')) => a */
   /* search28_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 107 },
      -1,
   } },

   /* replace28 -> 32 in the cache */

   /* ('ineg', ('ineg', 'a')) => a */
   /* search29_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 109 },
      -1,
   } },

   /* replace29 -> 32 in the cache */

   /* ('fabs', ('fneg', 'a')) => ('fabs', 'a') */
   /* search30_0_0 -> 32 in the cache */
   /* search30_0 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 107 },
      -1,
   } },

   /* replace30_0 -> 32 in the cache */
   /* replace30 -> 99 in the cache */

   /* ('fabs', ('u2f', 'a')) => ('u2f', 'a') */
   /* search31_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 112 },
      -1,
   } },

   /* replace31_0 -> 32 in the cache */
   /* replace31 -> 112 in the cache */

   /* ('iabs', ('iabs', 'a')) => ('iabs', 'a') */
   /* search32_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 114 },
      -1,
   } },

   /* replace32_0 -> 32 in the cache */
   /* replace32 -> 114 in the cache */

   /* ('iabs', ('ineg', 'a')) => ('iabs', 'a') */
   /* search33_0_0 -> 32 in the cache */
   /* search33_0 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 109 },
      -1,
   } },

   /* replace33_0 -> 32 in the cache */
   /* replace33 -> 114 in the cache */

   /* ('~fadd', 'a', 0.0) => a */
   /* search34_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* replace34 -> 32 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a */
   /* search35_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 119 },
      1,
   } },

   /* replace35 -> 11 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a */
   /* search36_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 121 },
      1,
   } },

   /* replace36 -> 16 in the cache */

   /* ('iadd', 'a', 0) => a */
   /* search37_0 -> 32 in the cache */
   /* search37_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace37 -> 32 in the cache */

   /* ('iadd_sat', 'a', 0) => a */
   /* search38_0 -> 32 in the cache */
   /* search38_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace38 -> 32 in the cache */

   /* ('isub_sat', 'a', 0) => a */
   /* search39_0 -> 32 in the cache */
   /* search39_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace39 -> 32 in the cache */

   /* ('uadd_sat', 'a', 0) => a */
   /* search40_0 -> 32 in the cache */
   /* search40_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace40 -> 32 in the cache */

   /* ('usub_sat', 'a', 0) => a */
   /* search41_0 -> 32 in the cache */
   /* search41_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace41 -> 32 in the cache */

   /* ('usadd_4x8_vc4', 'a', 0) => a */
   /* search42_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 128 },
      -1,
   } },

   /* replace42 -> 16 in the cache */

   /* ('usadd_4x8_vc4', 'a', -1) => -1 */
   /* search43_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 130 },
      -1,
   } },

   /* replace43 -> 130 in the cache */

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* search44_1_0 -> 132 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 134, 136 },
      -1,
   } },

   /* replace44_0 -> 132 in the cache */
   /* replace44_1_0 -> 133 in the cache */
   /* replace44_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 132, 138 },
      -1,
   } },

   /* ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search45_0_0 -> 16 in the cache */
   /* search45_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* search45_1_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 16, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 140, 142 },
      -1,
   } },

   /* replace45_0 -> 16 in the cache */
   /* replace45_1_0 -> 36 in the cache */
   /* replace45_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 16, 144 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search46_0 -> 0 in the cache */
   /* search46_1 -> 3 in the cache */
   /* search46_2_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 0, 146, 147 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 148 },
      -1,
   } },

   /* replace46_0 -> 0 in the cache */
   /* replace46_1_0 -> 3 in the cache */
   /* replace46_1_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3, 146 },
      -1,
   } },
   /* replace46_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 150, 147 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search47_0 -> 0 in the cache */
   /* search47_1 -> 3 in the cache */
   /* search47_2_0 -> 0 in the cache */
   /* search47_2_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 146 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 152 },
      -1,
   } },

   /* replace47_0 -> 0 in the cache */
   /* replace47_1_0 -> 3 in the cache */
   /* replace47_1_1 -> 146 in the cache */
   /* replace47_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 0, 150 },
      -1,
   } },

   /* ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 155, 156 },
      0,
   } },
   /* search48_1_0 -> 155 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 155, 158, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 157, 160 },
      -1,
   } },

   /* replace48_0 -> 155 in the cache */
   /* replace48_1_0 -> 156 in the cache */
   /* replace48_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 156, 158 },
      -1,
   } },
   /* replace48_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 155, 162, 159 },
      -1,
   } },

   /* ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd')) */
   /* search49_0 -> 155 in the cache */
   /* search49_1_0 -> 156 in the cache */
   /* search49_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 156, 158 },
      0,
   } },
   /* search49_2_0 -> 156 in the cache */
   /* search49_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 155, 164, 165 },
      -1,
   } },

   /* replace49_0 -> 156 in the cache */
   /* replace49_1_0 -> 155 in the cache */
   /* replace49_1_1 -> 158 in the cache */
   /* replace49_1_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 155, 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 156, 167 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search50_0 -> 16 in the cache */
   /* search50_1 -> 36 in the cache */
   /* search50_2_0 -> 16 in the cache */
   /* search50_2_1 -> 141 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 141, 169 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 170 },
      -1,
   } },

   /* replace50_0 -> 16 in the cache */
   /* replace50_1_0 -> 36 in the cache */
   /* replace50_1_1 -> 141 in the cache */
   /* replace50_1 -> 144 in the cache */
   /* replace50_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 144, 169 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search51_0 -> 16 in the cache */
   /* search51_1 -> 36 in the cache */
   /* search51_2_0 -> 16 in the cache */
   /* search51_2_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 141 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 173 },
      -1,
   } },

   /* replace51_0 -> 16 in the cache */
   /* replace51_1_0 -> 36 in the cache */
   /* replace51_1_1 -> 141 in the cache */
   /* replace51_1 -> 144 in the cache */
   /* replace51 -> 145 in the cache */

   /* ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search52_0_0 -> 16 in the cache */
   /* search52_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      0,
   } },
   /* search52_1_0 -> 16 in the cache */
   /* search52_1_1 -> 141 in the cache */
   /* search52_1_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 16, 141, 169 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 175, 176 },
      -1,
   } },

   /* replace52_0 -> 16 in the cache */
   /* replace52_1_0 -> 36 in the cache */
   /* replace52_1_1 -> 141 in the cache */
   /* replace52_1 -> 144 in the cache */
   /* replace52_2 -> 169 in the cache */
   /* replace52 -> 172 in the cache */

   /* ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd')) */
   /* search53_0 -> 16 in the cache */
   /* search53_1_0 -> 36 in the cache */
   /* search53_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 141 },
      0,
   } },
   /* search53_2_0 -> 36 in the cache */
   /* search53_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 36, 169 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 16, 178, 179 },
      -1,
   } },

   /* replace53_0 -> 36 in the cache */
   /* replace53_1_0 -> 16 in the cache */
   /* replace53_1_1 -> 141 in the cache */
   /* replace53_1_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 36, 181 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search54_0_0 -> 132 in the cache */
   /* search54_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* search54_1_0 -> 132 in the cache */
   /* search54_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 183, 184 },
      -1,
   } },

   /* replace54_0 -> 132 in the cache */
   /* replace54_1_0 -> 133 in the cache */
   /* replace54_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 132, 186 },
      -1,
   } },

   /* ('iadd', ('ishl', 'b', 'a'), ('ishl', 'c', 'a')) => ('ishl', ('iadd', 'b', 'c'), 'a') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 188, 189 },
      -1,
   } },
   /* search55_1_0 -> 135 in the cache */
   /* search55_1_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 135, 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 190, 191 },
      -1,
   } },

   /* replace55_0_0 -> 188 in the cache */
   /* replace55_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 188, 135 },
      -1,
   } },
   /* replace55_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 193, 189 },
      -1,
   } },

   /* ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c')) */
   /* search56_0_0 -> 132 in the cache */
   /* search56_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* search56_1_0 -> 132 in the cache */
   /* search56_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 195, 196 },
      -1,
   } },

   /* replace56_0 -> 132 in the cache */
   /* replace56_1_0 -> 133 in the cache */
   /* replace56_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 132, 198 },
      -1,
   } },

   /* ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c')) */
   /* search57_0_0 -> 132 in the cache */
   /* search57_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* search57_1_0 -> 132 in the cache */
   /* search57_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 200, 201 },
      -1,
   } },

   /* replace57_0 -> 132 in the cache */
   /* replace57_1_0 -> 133 in the cache */
   /* replace57_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 132, 203 },
      -1,
   } },

   /* ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ine', ('iand', 'a', 'b'), 0) */
   /* search58_0_0 -> 0 in the cache */
   /* search58_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search58_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 205, 3 },
      -1,
   } },

   /* replace58_0_0 -> 0 in the cache */
   /* replace58_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace58_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 207, 86 },
      -1,
   } },

   /* ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ieq', ('iand', 'a', 'b'), 0) */
   /* search59_0_0 -> 0 in the cache */
   /* search59_0_1 -> 1 in the cache */
   /* search59_0 -> 205 in the cache */
   /* search59_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 205, 3 },
      -1,
   } },

   /* replace59_0_0 -> 0 in the cache */
   /* replace59_0_1 -> 3 in the cache */
   /* replace59_0 -> 207 in the cache */
   /* replace59_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 207, 86 },
      -1,
   } },

   /* ('ieq', ('ushr(is_used_once)', 'a', '#b'), 0) => ('ult', 'a', ('ishl', 1, 'b')) */
   /* search60_0_0 -> 32 in the cache */
   /* search60_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 33 },
      0,
   } },
   /* search60_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 211, 61 },
      -1,
   } },

   /* replace60_0 -> 32 in the cache */
   /* replace60_1_0 -> 35 in the cache */
   /* replace60_1_1 -> 36 in the cache */
   /* replace60_1 -> 37 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 37 },
      -1,
   } },

   /* ('ine', ('ushr(is_used_once)', 'a', '#b'), 0) => ('uge', 'a', ('ishl', 1, 'b')) */
   /* search61_0_0 -> 32 in the cache */
   /* search61_0_1 -> 33 in the cache */
   /* search61_0 -> 211 in the cache */
   /* search61_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 211, 61 },
      -1,
   } },

   /* replace61_0 -> 32 in the cache */
   /* replace61_1_0 -> 35 in the cache */
   /* replace61_1_1 -> 36 in the cache */
   /* replace61_1 -> 37 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 37 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   /* search62_0_0 -> 32 in the cache */
   /* search62_0 -> 107 in the cache */
   /* search62_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 107, 32 },
      -1,
   } },

   /* replace62 -> 117 in the cache */

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search63_0_0 -> 32 in the cache */
   /* search63_0 -> 109 in the cache */
   /* search63_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 109, 32 },
      -1,
   } },

   /* replace63 -> 61 in the cache */

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   /* search64_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search64_1_0 -> 0 in the cache */
   /* search64_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 218, 219 },
      -1,
   } },

   /* replace64 -> 3 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search65_0 -> 0 in the cache */
   /* search65_1_0_0 -> 0 in the cache */
   /* search65_1_0 -> 218 in the cache */
   /* search65_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 218, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 0, 221 },
      -1,
   } },

   /* replace65 -> 3 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search66_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search66_1_0 -> 0 in the cache */
   /* search66_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 223, 224 },
      -1,
   } },

   /* replace66 -> 3 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search67_0 -> 0 in the cache */
   /* search67_1_0_0 -> 0 in the cache */
   /* search67_1_0 -> 223 in the cache */
   /* search67_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 223, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 0, 226 },
      -1,
   } },

   /* replace67 -> 3 in the cache */

   /* ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a')) */
   /* search68_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search68_1_0_0 -> 32 in the cache */
   /* search68_1_0 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 107 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 228, 229 },
      -1,
   } },

   /* replace68_0_0 -> 32 in the cache */
   /* replace68_0 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 99 },
      -1,
   } },

   /* ('~fmul', 'a', 0.0) => 0.0 */
   /* search69_0 -> 32 in the cache */
   /* search69_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* replace69 -> 117 in the cache */

   /* ('fmul', 'a@16', 0.0) => 0.0 */
   /* search70_0 -> 11 in the cache */
   /* search70_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 119 },
      -1,
   } },

   /* replace70 -> 119 in the cache */

   /* ('fmul', 'a@32', 0.0) => 0.0 */
   /* search71_0 -> 16 in the cache */
   /* search71_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 121 },
      -1,
   } },

   /* replace71 -> 121 in the cache */

   /* ('fmulz', 'a', 0.0) => 0.0 */
   /* search72_0 -> 16 in the cache */
   /* search72_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 121 },
      -1,
   } },

   /* replace72 -> 121 in the cache */

   /* ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b') */
   /* search73_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 236 },
      -1,
   } },

   /* replace73_0 -> 16 in the cache */
   /* replace73_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 239, 240 },
      -1,
   } },

   /* replace74_0 -> 16 in the cache */
   /* replace74_1 -> 36 in the cache */
   /* replace74 -> 238 in the cache */

   /* ('fmulz', 'a', 'a') => ('fmul', 'a', 'a') */
   /* search75_0 -> 16 in the cache */
   /* search75_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* replace75_0 -> 16 in the cache */
   /* replace75_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search76_0 -> 16 in the cache */
   /* search76_1 -> 236 in the cache */
   /* search76_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 236, 141 },
      -1,
   } },

   /* replace76_0 -> 16 in the cache */
   /* replace76_1 -> 36 in the cache */
   /* replace76_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search77_0 -> 239 in the cache */
   /* search77_1 -> 240 in the cache */
   /* search77_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 239, 240, 141 },
      -1,
   } },

   /* replace77_0 -> 16 in the cache */
   /* replace77_1 -> 36 in the cache */
   /* replace77_2 -> 141 in the cache */
   /* replace77 -> 245 in the cache */

   /* ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b') */
   /* search78_0 -> 16 in the cache */
   /* search78_1 -> 16 in the cache */
   /* search78_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* replace78_0 -> 16 in the cache */
   /* replace78_1 -> 16 in the cache */
   /* replace78_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* ('imul', 'a', 0) => 0 */
   /* search79_0 -> 32 in the cache */
   /* search79_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace79 -> 61 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', 0) => 0 */
   /* search80_0 -> 16 in the cache */
   /* search80_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 128 },
      -1,
   } },

   /* replace80 -> 128 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', -1) => a */
   /* search81_0 -> 16 in the cache */
   /* search81_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 130 },
      -1,
   } },

   /* replace81 -> 16 in the cache */

   /* ('~fmul', 'a', 1.0) => a */
   /* search82_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 252 },
      -1,
   } },

   /* replace82 -> 32 in the cache */

   /* ('~fmulz', 'a', 1.0) => a */
   /* search83_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 254 },
      -1,
   } },

   /* replace83 -> 16 in the cache */

   /* ('fmul(is_only_used_as_float)', 'a', 1.0) => a */
   /* search84_0 -> 32 in the cache */
   /* search84_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 252 },
      1,
   } },

   /* replace84 -> 32 in the cache */

   /* ('imul', 'a', 1) => a */
   /* search85_0 -> 32 in the cache */
   /* search85_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 35 },
      -1,
   } },

   /* replace85 -> 32 in the cache */

   /* ('fmul', 'a', -1.0) => ('fneg', 'a') */
   /* search86_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 258 },
      -1,
   } },

   /* replace86_0 -> 32 in the cache */
   /* replace86 -> 107 in the cache */

   /* ('imul', 'a', -1) => ('ineg', 'a') */
   /* search87_0 -> 32 in the cache */
   /* search87_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace87_0 -> 32 in the cache */
   /* replace87 -> 109 in the cache */

   /* ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a') */
   /* search88_0_0 -> 32 in the cache */
   /* search88_0 -> 98 in the cache */
   /* search88_1_0 -> 32 in the cache */
   /* search88_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 98, 261 },
      -1,
   } },

   /* replace88_0_0 -> 32 in the cache */
   /* replace88_0 -> 99 in the cache */
   /* replace88_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 99, 32 },
      -1,
   } },

   /* ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a') */
   /* search89_0_0_0 -> 32 in the cache */
   /* search89_0_0 -> 98 in the cache */
   /* search89_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 98, 32 },
      -1,
   } },
   /* search89_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 264, 32 },
      -1,
   } },

   /* replace89_0_0 -> 32 in the cache */
   /* replace89_0 -> 99 in the cache */
   /* replace89_1 -> 32 in the cache */
   /* replace89 -> 263 in the cache */

   /* ('~ffma', 0.0, 'a', 'b') => b */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search90_1 -> 0 in the cache */
   /* search90_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 266, 0, 3 },
      -1,
   } },

   /* replace90 -> 3 in the cache */

   /* ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search91_0 -> 119 in the cache */
   /* search91_1 -> 11 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 119, 11, 268 },
      1,
   } },

   /* replace91 -> 268 in the cache */

   /* ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search92_0 -> 121 in the cache */
   /* search92_1 -> 16 in the cache */
   /* search92_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 121, 16, 36 },
      1,
   } },

   /* replace92 -> 36 in the cache */

   /* ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b') */
   /* search93_0 -> 121 in the cache */
   /* search93_1 -> 16 in the cache */
   /* search93_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 121, 16, 36 },
      -1,
   } },

   /* replace93_0 -> 121 in the cache */
   /* replace93_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 121, 36 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search94_0 -> 0 in the cache */
   /* search94_1 -> 3 in the cache */
   /* search94_2 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 0, 3, 266 },
      -1,
   } },

   /* replace94_0 -> 0 in the cache */
   /* replace94_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search95_0 -> 11 in the cache */
   /* search95_1 -> 268 in the cache */
   /* search95_2 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 268, 119 },
      -1,
   } },

   /* replace95_0 -> 11 in the cache */
   /* replace95_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 268 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search96_0 -> 16 in the cache */
   /* search96_1 -> 36 in the cache */
   /* search96_2 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 121 },
      -1,
   } },

   /* replace96_0 -> 16 in the cache */
   /* replace96_1 -> 36 in the cache */
   /* replace96 -> 238 in the cache */

   /* ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b') */
   /* search97_0 -> 16 in the cache */
   /* search97_1 -> 36 in the cache */
   /* search97_2 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 121 },
      -1,
   } },

   /* replace97_0 -> 16 in the cache */
   /* replace97_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search98_1 -> 0 in the cache */
   /* search98_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 280, 0, 3 },
      -1,
   } },

   /* replace98_0 -> 0 in the cache */
   /* replace98_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   /* search99_0 -> 254 in the cache */
   /* search99_1 -> 16 in the cache */
   /* search99_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 254, 16, 36 },
      -1,
   } },

   /* replace99_0 -> 16 in the cache */
   /* replace99_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search100_1 -> 0 in the cache */
   /* search100_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 285, 0, 3 },
      -1,
   } },

   /* replace100_0_0 -> 0 in the cache */
   /* replace100_0 -> 223 in the cache */
   /* replace100_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 223, 3 },
      -1,
   } },

   /* ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search101_1 -> 16 in the cache */
   /* search101_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 288, 16, 36 },
      -1,
   } },

   /* replace101_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 16 },
      -1,
   } },
   /* replace101_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 290, 36 },
      -1,
   } },

   /* ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search102_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 292, 293, 135 },
      -1,
   } },

   /* replace102_0_0 -> 132 in the cache */
   /* replace102_0_1 -> 133 in the cache */
   /* replace102_0 -> 134 in the cache */
   /* replace102_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 134, 135 },
      -1,
   } },

   /* ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search103_1 -> 33 in the cache */
   /* search103_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 296, 33, 141 },
      -1,
   } },

   /* replace103_0_0 -> 16 in the cache */
   /* replace103_0_1 -> 36 in the cache */
   /* replace103_0 -> 140 in the cache */
   /* replace103_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 140, 141 },
      -1,
   } },

   /* ('~flrp', 'a', 'b', 0.0) => a */
   /* search104_0 -> 0 in the cache */
   /* search104_1 -> 3 in the cache */
   /* search104_2 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 266 },
      -1,
   } },

   /* replace104 -> 0 in the cache */

   /* ('~flrp', 'a', 'b', 1.0) => b */
   /* search105_0 -> 0 in the cache */
   /* search105_1 -> 3 in the cache */
   /* search105_2 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 280 },
      -1,
   } },

   /* replace105 -> 3 in the cache */

   /* ('~flrp', 'a', 'a', 'b') => a */
   /* search106_0 -> 0 in the cache */
   /* search106_1 -> 0 in the cache */
   /* search106_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 0, 3 },
      -1,
   } },

   /* replace106 -> 0 in the cache */

   /* ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b') */
   /* search107_0 -> 266 in the cache */
   /* search107_1 -> 0 in the cache */
   /* search107_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 266, 0, 3 },
      -1,
   } },

   /* replace107_0 -> 0 in the cache */
   /* replace107_1 -> 3 in the cache */
   /* replace107 -> 274 in the cache */

   /* ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search108_0 -> 132 in the cache */
   /* search108_1_0 -> 132 in the cache */
   /* search108_1_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 132, 133 },
      0,
   } },
   /* search108_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 132, 303, 135 },
      -1,
   } },

   /* replace108_0_0 -> 133 in the cache */
   /* replace108_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 133, 135 },
      -1,
   } },
   /* replace108_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 305, 132 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 0, 'b') => b */
   /* search109_0 -> 16 in the cache */
   /* search109_1 -> 128 in the cache */
   /* search109_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace109 -> 36 in the cache */

   /* ('udot_4x8_uadd', 'a', 0, 'b') => b */
   /* search110_0 -> 16 in the cache */
   /* search110_1 -> 128 in the cache */
   /* search110_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace110 -> 36 in the cache */

   /* ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search111_0 -> 16 in the cache */
   /* search111_1 -> 128 in the cache */
   /* search111_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace111 -> 36 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 0, 'b') => b */
   /* search112_0 -> 16 in the cache */
   /* search112_1 -> 128 in the cache */
   /* search112_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace112 -> 36 in the cache */

   /* ('sdot_2x16_iadd', 'a', 0, 'b') => b */
   /* search113_0 -> 16 in the cache */
   /* search113_1 -> 128 in the cache */
   /* search113_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace113 -> 36 in the cache */

   /* ('udot_2x16_uadd', 'a', 0, 'b') => b */
   /* search114_0 -> 16 in the cache */
   /* search114_1 -> 128 in the cache */
   /* search114_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace114 -> 36 in the cache */

   /* ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b */
   /* search115_0 -> 16 in the cache */
   /* search115_1 -> 128 in the cache */
   /* search115_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace115 -> 36 in the cache */

   /* ('udot_2x16_uadd_sat', 'a', 0, 'b') => b */
   /* search116_0 -> 16 in the cache */
   /* search116_1 -> 128 in the cache */
   /* search116_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 128, 36 },
      -1,
   } },

   /* replace116 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 'a', 0, 'b') => b */
   /* search117_0 -> 16 in the cache */
   /* search117_1 -> 128 in the cache */
   /* search117_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 128, 36 },
      -1,
   } },

   /* replace117 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 0, 'a', 'b') => b */
   /* search118_0 -> 128 in the cache */
   /* search118_1 -> 16 in the cache */
   /* search118_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 128, 16, 36 },
      -1,
   } },

   /* replace118 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search119_0 -> 16 in the cache */
   /* search119_1 -> 128 in the cache */
   /* search119_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 128, 36 },
      -1,
   } },

   /* replace119 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b */
   /* search120_0 -> 128 in the cache */
   /* search120_1 -> 16 in the cache */
   /* search120_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 128, 16, 36 },
      -1,
   } },

   /* replace120 -> 36 in the cache */

   /* ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search121_0_0 -> 16 in the cache */
   /* search121_0_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 319 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 320, 321 },
      -1,
   } },

   /* replace121_0 -> 16 in the cache */
   /* replace121_1 -> 36 in the cache */
   /* replace121_2_0 -> 141 in the cache */
   /* replace121_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 2,
      { 16, 36, 323 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search122_0_0 -> 16 in the cache */
   /* search122_0_1 -> 36 in the cache */
   /* search122_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 319 },
      0,
   } },
   /* search122_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 325, 321 },
      -1,
   } },

   /* replace122_0 -> 16 in the cache */
   /* replace122_1 -> 36 in the cache */
   /* replace122_2_0 -> 141 in the cache */
   /* replace122_2_1 -> 169 in the cache */
   /* replace122_2 -> 323 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 2,
      { 16, 36, 323 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search123_0_0 -> 16 in the cache */
   /* search123_0_1 -> 36 in the cache */
   /* search123_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 319 },
      0,
   } },
   /* search123_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 328, 321 },
      -1,
   } },

   /* replace123_0 -> 16 in the cache */
   /* replace123_1 -> 36 in the cache */
   /* replace123_2_0 -> 141 in the cache */
   /* replace123_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 1,
      { 16, 36, 330 },
      -1,
   } },

   /* ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search124_0_0 -> 16 in the cache */
   /* search124_0_1 -> 36 in the cache */
   /* search124_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 319 },
      0,
   } },
   /* search124_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 332, 321 },
      -1,
   } },

   /* replace124_0 -> 16 in the cache */
   /* replace124_1 -> 36 in the cache */
   /* replace124_2_0 -> 141 in the cache */
   /* replace124_2_1 -> 169 in the cache */
   /* replace124_2 -> 323 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 2,
      { 16, 36, 323 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search125_0_0 -> 16 in the cache */
   /* search125_0_1 -> 36 in the cache */
   /* search125_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 319 },
      0,
   } },
   /* search125_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 335, 321 },
      -1,
   } },

   /* replace125_0 -> 16 in the cache */
   /* replace125_1 -> 36 in the cache */
   /* replace125_2_0 -> 141 in the cache */
   /* replace125_2_1 -> 169 in the cache */
   /* replace125_2 -> 323 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 2,
      { 16, 36, 323 },
      -1,
   } },

   /* ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search126_0_1 -> 36 in the cache */
   /* search126_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 338, 36, 128 },
      -1,
   } },
   /* search126_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 339, 141 },
      -1,
   } },

   /* replace126_0 -> 16 in the cache */
   /* replace126_1 -> 36 in the cache */
   /* replace126_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search127_0_0 -> 338 in the cache */
   /* search127_0_1 -> 36 in the cache */
   /* search127_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 338, 36, 128 },
      -1,
   } },
   /* search127_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 342, 141 },
      -1,
   } },

   /* replace127_0 -> 16 in the cache */
   /* replace127_1 -> 36 in the cache */
   /* replace127_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search128_0_0 -> 338 in the cache */
   /* search128_0_1 -> 36 in the cache */
   /* search128_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 338, 36, 128 },
      -1,
   } },
   /* search128_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 345, 141 },
      -1,
   } },

   /* replace128_0 -> 16 in the cache */
   /* replace128_1 -> 36 in the cache */
   /* replace128_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 141 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search129_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search129_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 348, 128 },
      -1,
   } },
   /* search129_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 349, 141 },
      -1,
   } },

   /* replace129_0 -> 16 in the cache */
   /* replace129_1 -> 36 in the cache */
   /* replace129_2 -> 141 in the cache */
   /* replace129 -> 347 in the cache */

   /* ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c') */
   /* search130_0_0 -> 338 in the cache */
   /* search130_0_1 -> 36 in the cache */
   /* search130_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 338, 36, 128 },
      -1,
   } },
   /* search130_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 351, 141 },
      -1,
   } },

   /* replace130_0 -> 16 in the cache */
   /* replace130_1 -> 36 in the cache */
   /* replace130_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c') */
   /* search131_0_0 -> 338 in the cache */
   /* search131_0_1 -> 36 in the cache */
   /* search131_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 338, 36, 128 },
      -1,
   } },
   /* search131_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 354, 141 },
      -1,
   } },

   /* replace131_0 -> 16 in the cache */
   /* replace131_1 -> 36 in the cache */
   /* replace131_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search132_0 -> 296 in the cache */
   /* search132_1 -> 33 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace132_0_0 -> 16 in the cache */
   /* replace132_0_1 -> 36 in the cache */
   /* replace132_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 128 },
      -1,
   } },
   /* replace132_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 359, 141 },
      -1,
   } },

   /* ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search133_0 -> 296 in the cache */
   /* search133_1 -> 33 in the cache */
   /* search133_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace133_0_0 -> 16 in the cache */
   /* replace133_0_1 -> 36 in the cache */
   /* replace133_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 128 },
      -1,
   } },
   /* replace133_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 362, 141 },
      -1,
   } },

   /* ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search134_0 -> 296 in the cache */
   /* search134_1 -> 33 in the cache */
   /* search134_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 296, 33, 357 },
      -1,
   } },

   /* replace134_0_0 -> 16 in the cache */
   /* replace134_0_1 -> 36 in the cache */
   /* replace134_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 128 },
      -1,
   } },
   /* replace134_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 365, 141 },
      -1,
   } },

   /* ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search135_0 -> 296 in the cache */
   /* search135_1 -> 33 in the cache */
   /* search135_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace135_0_0 -> 16 in the cache */
   /* replace135_0_1 -> 36 in the cache */
   /* replace135_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 128 },
      -1,
   } },
   /* replace135_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 368, 141 },
      -1,
   } },

   /* ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search136_0 -> 296 in the cache */
   /* search136_1 -> 33 in the cache */
   /* search136_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace136_0_0 -> 16 in the cache */
   /* replace136_0_1 -> 36 in the cache */
   /* replace136_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 128 },
      -1,
   } },
   /* replace136_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 371, 141 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search137_0 -> 296 in the cache */
   /* search137_1 -> 33 in the cache */
   /* search137_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace137_0_0 -> 16 in the cache */
   /* replace137_0_1 -> 36 in the cache */
   /* replace137_0_2 -> 128 in the cache */
   /* replace137_0 -> 359 in the cache */
   /* replace137_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 359, 141 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search138_0 -> 296 in the cache */
   /* search138_1 -> 33 in the cache */
   /* search138_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace138_0_0 -> 16 in the cache */
   /* replace138_0_1 -> 36 in the cache */
   /* replace138_0_2 -> 128 in the cache */
   /* replace138_0 -> 362 in the cache */
   /* replace138_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 362, 141 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search139_0 -> 296 in the cache */
   /* search139_1 -> 33 in the cache */
   /* search139_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 296, 33, 357 },
      -1,
   } },

   /* replace139_0_0 -> 16 in the cache */
   /* replace139_0_1 -> 36 in the cache */
   /* replace139_0_2 -> 128 in the cache */
   /* replace139_0 -> 365 in the cache */
   /* replace139_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 365, 141 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search140_0 -> 296 in the cache */
   /* search140_1 -> 33 in the cache */
   /* search140_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace140_0_0 -> 16 in the cache */
   /* replace140_0_1 -> 36 in the cache */
   /* replace140_0_2 -> 128 in the cache */
   /* replace140_0 -> 368 in the cache */
   /* replace140_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 368, 141 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search141_0 -> 296 in the cache */
   /* search141_1 -> 33 in the cache */
   /* search141_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 296, 33, 357 },
      -1,
   } },

   /* replace141_0_0 -> 16 in the cache */
   /* replace141_0_1 -> 36 in the cache */
   /* replace141_0_2 -> 128 in the cache */
   /* replace141_0 -> 371 in the cache */
   /* replace141_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 371, 141 },
      -1,
   } },

   /* ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search142_0_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 383, 121 },
      -1,
   } },
   /* search142_0_1 -> 121 in the cache */
   /* search142_0_2 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 384, 121, 189 },
      -1,
   } },
   /* search142_1_0_0 -> 189 in the cache */
   /* search142_1_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      2, 1,
      { 189, 121 },
      -1,
   } },
   /* search142_1_1 -> 121 in the cache */
   /* search142_1_2 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 386, 121, 383 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 385, 387 },
      -1,
   } },

   /* replace142_0 -> 189 in the cache */
   /* replace142_1 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 189, 383 },
      -1,
   } },

   /* ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b') */
   /* search143_0 -> 16 in the cache */
   /* search143_1_0_0 -> 16 in the cache */
   /* search143_1_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 16, 121 },
      -1,
   } },
   /* search143_1_1 -> 121 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 390, 121, 391 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 16, 392 },
      -1,
   } },

   /* replace143_0 -> 16 in the cache */
   /* replace143_1 -> 36 in the cache */
   /* replace143 -> 279 in the cache */

   /* ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search144_0_0_0 -> 383 in the cache */
   /* search144_0_0_1 -> 121 in the cache */
   /* search144_0_0 -> 384 in the cache */
   /* search144_0_1 -> 121 in the cache */
   /* search144_0_2 -> 189 in the cache */
   /* search144_0 -> 385 in the cache */
   /* search144_1_0_0 -> 189 in the cache */
   /* search144_1_0_1 -> 121 in the cache */
   /* search144_1_0 -> 386 in the cache */
   /* search144_1_1 -> 121 in the cache */
   /* search144_1_2 -> 383 in the cache */
   /* search144_1 -> 387 in the cache */
   /* search144_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 385, 387, 141 },
      -1,
   } },

   /* replace144_0 -> 189 in the cache */
   /* replace144_1 -> 383 in the cache */
   /* replace144_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 189, 383, 141 },
      -1,
   } },

   /* ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search145_0 -> 16 in the cache */
   /* search145_1_0_0 -> 16 in the cache */
   /* search145_1_0_1 -> 121 in the cache */
   /* search145_1_0 -> 390 in the cache */
   /* search145_1_1 -> 121 in the cache */
   /* search145_1_2 -> 391 in the cache */
   /* search145_1 -> 392 in the cache */
   /* search145_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 16, 392, 141 },
      -1,
   } },

   /* replace145_0 -> 16 in the cache */
   /* replace145_1 -> 36 in the cache */
   /* replace145_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 1.0, ('fexp2', ('fmul@32', 'a', 'b'))) => ('fexp2', ('fmulz', 'a', 'b')) */
   /* search146_0_0 -> 383 in the cache */
   /* search146_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      0, 1,
      { 383, 121 },
      -1,
   } },
   /* search146_1 -> 254 in the cache */
   /* search146_2_0_0 -> 189 in the cache */
   /* search146_2_0_1 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 189, 383 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 399 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 398, 254, 400 },
      -1,
   } },

   /* replace146_0_0 -> 189 in the cache */
   /* replace146_0_1 -> 383 in the cache */
   /* replace146_0 -> 389 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 389 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search147_0 -> 16 in the cache */
   /* search147_1 -> 36 in the cache */
   /* search147_2 -> 141 in the cache */
   /* search147 -> 341 in the cache */

   /* replace147_0_0_0_0_0 -> 16 in the cache */
   /* replace147_0_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace147_0_0_0_1_0 -> 36 in the cache */
   /* replace147_0_0_0_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 403, 404 },
      -1,
   } },
   /* replace147_0_0_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 406 },
      -1,
   } },
   /* replace147_0_0_1_1_0 -> 36 in the cache */
   /* replace147_0_0_1_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 407, 408 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 405, 409 },
      -1,
   } },
   /* replace147_0_1_0_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 411 },
      -1,
   } },
   /* replace147_0_1_0_1_0 -> 36 in the cache */
   /* replace147_0_1_0_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 411 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 412, 413 },
      -1,
   } },
   /* replace147_0_1_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 415 },
      -1,
   } },
   /* replace147_0_1_1_1_0 -> 36 in the cache */
   /* replace147_0_1_1_1_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 416, 417 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 414, 418 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 410, 419 },
      -1,
   } },
   /* replace147_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 420, 141 },
      -1,
   } },

   /* ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search148_0 -> 16 in the cache */
   /* search148_1 -> 36 in the cache */
   /* search148_2 -> 141 in the cache */
   /* search148 -> 344 in the cache */

   /* replace148_0_0_0_0_0 -> 16 in the cache */
   /* replace148_0_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace148_0_0_0_1_0 -> 36 in the cache */
   /* replace148_0_0_0_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 422, 423 },
      -1,
   } },
   /* replace148_0_0_1_0_0 -> 16 in the cache */
   /* replace148_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 406 },
      -1,
   } },
   /* replace148_0_0_1_1_0 -> 36 in the cache */
   /* replace148_0_0_1_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 425, 426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 424, 427 },
      -1,
   } },
   /* replace148_0_1_0_0_0 -> 16 in the cache */
   /* replace148_0_1_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 411 },
      -1,
   } },
   /* replace148_0_1_0_1_0 -> 36 in the cache */
   /* replace148_0_1_0_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 411 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 429, 430 },
      -1,
   } },
   /* replace148_0_1_1_0_0 -> 16 in the cache */
   /* replace148_0_1_1_0_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 415 },
      -1,
   } },
   /* replace148_0_1_1_1_0 -> 36 in the cache */
   /* replace148_0_1_1_1_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 432, 433 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 431, 434 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 428, 435 },
      -1,
   } },
   /* replace148_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 436, 141 },
      -1,
   } },

   /* ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search149_0 -> 16 in the cache */
   /* search149_1 -> 36 in the cache */
   /* search149_2 -> 141 in the cache */
   /* search149 -> 347 in the cache */

   /* replace149_0_0_0_0_0 -> 16 in the cache */
   /* replace149_0_0_0_0_1 -> 128 in the cache */
   /* replace149_0_0_0_0 -> 403 in the cache */
   /* replace149_0_0_0_1_0 -> 36 in the cache */
   /* replace149_0_0_0_1_1 -> 128 in the cache */
   /* replace149_0_0_0_1 -> 423 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 403, 423 },
      -1,
   } },
   /* replace149_0_0_1_0_0 -> 16 in the cache */
   /* replace149_0_0_1_0_1 -> 406 in the cache */
   /* replace149_0_0_1_0 -> 407 in the cache */
   /* replace149_0_0_1_1_0 -> 36 in the cache */
   /* replace149_0_0_1_1_1 -> 406 in the cache */
   /* replace149_0_0_1_1 -> 426 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 407, 426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 438, 439 },
      -1,
   } },
   /* replace149_0_1_0_0_0 -> 16 in the cache */
   /* replace149_0_1_0_0_1 -> 411 in the cache */
   /* replace149_0_1_0_0 -> 412 in the cache */
   /* replace149_0_1_0_1_0 -> 36 in the cache */
   /* replace149_0_1_0_1_1 -> 411 in the cache */
   /* replace149_0_1_0_1 -> 430 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 412, 430 },
      -1,
   } },
   /* replace149_0_1_1_0_0 -> 16 in the cache */
   /* replace149_0_1_1_0_1 -> 415 in the cache */
   /* replace149_0_1_1_0 -> 416 in the cache */
   /* replace149_0_1_1_1_0 -> 36 in the cache */
   /* replace149_0_1_1_1_1 -> 415 in the cache */
   /* replace149_0_1_1_1 -> 433 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 416, 433 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 441, 442 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 440, 443 },
      -1,
   } },
   /* replace149_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 444, 141 },
      -1,
   } },

   /* ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search150_0 -> 16 in the cache */
   /* search150_1 -> 36 in the cache */
   /* search150_2 -> 141 in the cache */
   /* search150 -> 353 in the cache */

   /* replace150_0_0_0_0 -> 16 in the cache */
   /* replace150_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace150_0_0_1_0 -> 36 in the cache */
   /* replace150_0_0_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 446, 447 },
      -1,
   } },
   /* replace150_0_1_0_0 -> 16 in the cache */
   /* replace150_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 406 },
      -1,
   } },
   /* replace150_0_1_1_0 -> 36 in the cache */
   /* replace150_0_1_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 449, 450 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 448, 451 },
      -1,
   } },
   /* replace150_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 452, 141 },
      -1,
   } },

   /* ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search151_0 -> 16 in the cache */
   /* search151_1 -> 36 in the cache */
   /* search151_2 -> 141 in the cache */
   /* search151 -> 356 in the cache */

   /* replace151_0_0_0_0 -> 16 in the cache */
   /* replace151_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace151_0_0_1_0 -> 36 in the cache */
   /* replace151_0_0_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 454, 455 },
      -1,
   } },
   /* replace151_0_1_0_0 -> 16 in the cache */
   /* replace151_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 406 },
      -1,
   } },
   /* replace151_0_1_1_0 -> 36 in the cache */
   /* replace151_0_1_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 457, 458 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 456, 459 },
      -1,
   } },
   /* replace151_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 460, 141 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search152_0 -> 16 in the cache */
   /* search152_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      7,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 462 },
      -1,
   } },

   /* replace152_0 -> 16 in the cache */
   /* replace152_1 -> 36 in the cache */
   /* replace152_2 -> 141 in the cache */
   /* replace152 -> 344 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search153_0 -> 16 in the cache */
   /* search153_1 -> 36 in the cache */
   /* search153_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace153_0_0_0_0_0 -> 16 in the cache */
   /* replace153_0_0_0_0_1 -> 128 in the cache */
   /* replace153_0_0_0_0 -> 422 in the cache */
   /* replace153_0_0_0_1_0 -> 36 in the cache */
   /* replace153_0_0_0_1_1 -> 128 in the cache */
   /* replace153_0_0_0_1 -> 423 in the cache */
   /* replace153_0_0_0 -> 424 in the cache */
   /* replace153_0_0_1_0_0 -> 16 in the cache */
   /* replace153_0_0_1_0_1 -> 406 in the cache */
   /* replace153_0_0_1_0 -> 425 in the cache */
   /* replace153_0_0_1_1_0 -> 36 in the cache */
   /* replace153_0_0_1_1_1 -> 406 in the cache */
   /* replace153_0_0_1_1 -> 426 in the cache */
   /* replace153_0_0_1 -> 427 in the cache */
   /* replace153_0_0 -> 428 in the cache */
   /* replace153_0_1_0_0_0 -> 16 in the cache */
   /* replace153_0_1_0_0_1 -> 411 in the cache */
   /* replace153_0_1_0_0 -> 429 in the cache */
   /* replace153_0_1_0_1_0 -> 36 in the cache */
   /* replace153_0_1_0_1_1 -> 411 in the cache */
   /* replace153_0_1_0_1 -> 430 in the cache */
   /* replace153_0_1_0 -> 431 in the cache */
   /* replace153_0_1_1_0_0 -> 16 in the cache */
   /* replace153_0_1_1_0_1 -> 415 in the cache */
   /* replace153_0_1_1_0 -> 432 in the cache */
   /* replace153_0_1_1_1_0 -> 36 in the cache */
   /* replace153_0_1_1_1_1 -> 415 in the cache */
   /* replace153_0_1_1_1 -> 433 in the cache */
   /* replace153_0_1_1 -> 434 in the cache */
   /* replace153_0_1 -> 435 in the cache */
   /* replace153_0 -> 436 in the cache */
   /* replace153_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 8,
      { 436, 141 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search154_0 -> 16 in the cache */
   /* search154_1 -> 36 in the cache */
   /* search154_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace154_0_0_0_0_0 -> 16 in the cache */
   /* replace154_0_0_0_0_1 -> 128 in the cache */
   /* replace154_0_0_0_0 -> 403 in the cache */
   /* replace154_0_0_0_1_0 -> 36 in the cache */
   /* replace154_0_0_0_1_1 -> 128 in the cache */
   /* replace154_0_0_0_1 -> 404 in the cache */
   /* replace154_0_0_0 -> 405 in the cache */
   /* replace154_0_0_1_0_0 -> 16 in the cache */
   /* replace154_0_0_1_0_1 -> 406 in the cache */
   /* replace154_0_0_1_0 -> 407 in the cache */
   /* replace154_0_0_1_1_0 -> 36 in the cache */
   /* replace154_0_0_1_1_1 -> 406 in the cache */
   /* replace154_0_0_1_1 -> 408 in the cache */
   /* replace154_0_0_1 -> 409 in the cache */
   /* replace154_0_0 -> 410 in the cache */
   /* replace154_0_1_0_0_0 -> 16 in the cache */
   /* replace154_0_1_0_0_1 -> 411 in the cache */
   /* replace154_0_1_0_0 -> 412 in the cache */
   /* replace154_0_1_0_1_0 -> 36 in the cache */
   /* replace154_0_1_0_1_1 -> 411 in the cache */
   /* replace154_0_1_0_1 -> 413 in the cache */
   /* replace154_0_1_0 -> 414 in the cache */
   /* replace154_0_1_1_0_0 -> 16 in the cache */
   /* replace154_0_1_1_0_1 -> 415 in the cache */
   /* replace154_0_1_1_0 -> 416 in the cache */
   /* replace154_0_1_1_1_0 -> 36 in the cache */
   /* replace154_0_1_1_1_1 -> 415 in the cache */
   /* replace154_0_1_1_1 -> 417 in the cache */
   /* replace154_0_1_1 -> 418 in the cache */
   /* replace154_0_1 -> 419 in the cache */
   /* replace154_0 -> 420 in the cache */
   /* replace154_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 420, 141 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search155_0 -> 16 in the cache */
   /* search155_1 -> 36 in the cache */
   /* search155_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 36, 141 },
      -1,
   } },

   /* replace155_0_0_0_0_0 -> 16 in the cache */
   /* replace155_0_0_0_0_1 -> 128 in the cache */
   /* replace155_0_0_0_0 -> 403 in the cache */
   /* replace155_0_0_0_1_0 -> 36 in the cache */
   /* replace155_0_0_0_1_1 -> 128 in the cache */
   /* replace155_0_0_0_1 -> 423 in the cache */
   /* replace155_0_0_0 -> 438 in the cache */
   /* replace155_0_0_1_0_0 -> 16 in the cache */
   /* replace155_0_0_1_0_1 -> 406 in the cache */
   /* replace155_0_0_1_0 -> 407 in the cache */
   /* replace155_0_0_1_1_0 -> 36 in the cache */
   /* replace155_0_0_1_1_1 -> 406 in the cache */
   /* replace155_0_0_1_1 -> 426 in the cache */
   /* replace155_0_0_1 -> 439 in the cache */
   /* replace155_0_0 -> 440 in the cache */
   /* replace155_0_1_0_0_0 -> 16 in the cache */
   /* replace155_0_1_0_0_1 -> 411 in the cache */
   /* replace155_0_1_0_0 -> 412 in the cache */
   /* replace155_0_1_0_1_0 -> 36 in the cache */
   /* replace155_0_1_0_1_1 -> 411 in the cache */
   /* replace155_0_1_0_1 -> 430 in the cache */
   /* replace155_0_1_0 -> 441 in the cache */
   /* replace155_0_1_1_0_0 -> 16 in the cache */
   /* replace155_0_1_1_0_1 -> 415 in the cache */
   /* replace155_0_1_1_0 -> 416 in the cache */
   /* replace155_0_1_1_1_0 -> 36 in the cache */
   /* replace155_0_1_1_1_1 -> 415 in the cache */
   /* replace155_0_1_1_1 -> 433 in the cache */
   /* replace155_0_1_1 -> 442 in the cache */
   /* replace155_0_1 -> 443 in the cache */
   /* replace155_0 -> 444 in the cache */
   /* replace155_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 444, 141 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search156_0 -> 16 in the cache */
   /* search156_1 -> 36 in the cache */
   /* search156_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace156_0_0_0_0 -> 16 in the cache */
   /* replace156_0_0_0_1 -> 128 in the cache */
   /* replace156_0_0_0 -> 454 in the cache */
   /* replace156_0_0_1_0 -> 36 in the cache */
   /* replace156_0_0_1_1 -> 128 in the cache */
   /* replace156_0_0_1 -> 455 in the cache */
   /* replace156_0_0 -> 456 in the cache */
   /* replace156_0_1_0_0 -> 16 in the cache */
   /* replace156_0_1_0_1 -> 406 in the cache */
   /* replace156_0_1_0 -> 457 in the cache */
   /* replace156_0_1_1_0 -> 36 in the cache */
   /* replace156_0_1_1_1 -> 406 in the cache */
   /* replace156_0_1_1 -> 458 in the cache */
   /* replace156_0_1 -> 459 in the cache */
   /* replace156_0 -> 460 in the cache */
   /* replace156_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 4,
      { 460, 141 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search157_0 -> 16 in the cache */
   /* search157_1 -> 36 in the cache */
   /* search157_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace157_0_0_0_0 -> 16 in the cache */
   /* replace157_0_0_0_1 -> 128 in the cache */
   /* replace157_0_0_0 -> 446 in the cache */
   /* replace157_0_0_1_0 -> 36 in the cache */
   /* replace157_0_0_1_1 -> 128 in the cache */
   /* replace157_0_0_1 -> 447 in the cache */
   /* replace157_0_0 -> 448 in the cache */
   /* replace157_0_1_0_0 -> 16 in the cache */
   /* replace157_0_1_0_1 -> 406 in the cache */
   /* replace157_0_1_0 -> 449 in the cache */
   /* replace157_0_1_1_0 -> 36 in the cache */
   /* replace157_0_1_1_1 -> 406 in the cache */
   /* replace157_0_1_1 -> 450 in the cache */
   /* replace157_0_1 -> 451 in the cache */
   /* replace157_0 -> 452 in the cache */
   /* replace157_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 4,
      { 452, 141 },
      -1,
   } },

   /* ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search158_0 -> 11 in the cache */
   /* search158_1 -> 268 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 268, 475 },
      -1,
   } },

   /* replace158_0 -> 474 in the cache */
   /* replace158_1 -> 268 in the cache */
   /* replace158_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 474, 268, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search159_0 -> 11 in the cache */
   /* search159_1_0 -> 11 in the cache */
   /* search159_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 268 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 478, 479 },
      -1,
   } },

   /* replace159_0_0 -> 268 in the cache */
   /* replace159_0_1 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 268, 479 },
      -1,
   } },
   /* replace159_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 481, 11 },
      -1,
   } },

   /* ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search160_0_0 -> 11 in the cache */
   /* search160_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 268 },
      0,
   } },
   /* search160_1_0 -> 11 in the cache */
   /* search160_1_1 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 11, 479 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 483, 484, 485 },
      -1,
   } },

   /* replace160_0_0 -> 268 in the cache */
   /* replace160_0_1 -> 479 in the cache */
   /* replace160_0_2 -> 485 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 268, 479, 485 },
      -1,
   } },
   /* replace160_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 487, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search161_0 -> 11 in the cache */
   /* search161_1_0 -> 11 in the cache */
   /* search161_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 268 },
      0,
   } },
   /* search161_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 489, 479 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace161_0_1 -> 268 in the cache */
   /* replace161_0_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 491, 268, 479 },
      -1,
   } },
   /* replace161_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 492, 11 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search162_0_0 -> 11 in the cache */
   /* search162_0_1_0 -> 491 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 494 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 491, 495 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 496 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search162_1_1 -> 494 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 498, 494 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 497, 499 },
      -1,
   } },

   /* replace162_0 -> 11 in the cache */
   /* replace162_1 -> 498 in the cache */
   /* replace162_2 -> 494 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 498, 494 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search163_0_0 -> 11 in the cache */
   /* search163_0_1_0_0 -> 491 in the cache */
   /* search163_0_1_0_1_0 -> 494 in the cache */
   /* search163_0_1_0_1 -> 495 in the cache */
   /* search163_0_1_0 -> 496 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 496 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 502 },
      -1,
   } },
   /* search163_1_0 -> 498 in the cache */
   /* search163_1_1_0 -> 494 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 494 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 498, 504 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 503, 505 },
      -1,
   } },

   /* replace163_0 -> 11 in the cache */
   /* replace163_1 -> 498 in the cache */
   /* replace163_2_0 -> 494 in the cache */
   /* replace163_2 -> 504 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 498, 504 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search164_0 -> 11 in the cache */
   /* search164_1_0 -> 494 in the cache */
   /* search164_1_1_0 -> 498 in the cache */
   /* search164_1_1_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 498, 508 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 494, 509 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 510 },
      -1,
   } },

   /* replace164_0 -> 11 in the cache */
   /* replace164_1 -> 498 in the cache */
   /* replace164_2 -> 494 in the cache */
   /* replace164 -> 501 in the cache */

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search165_0_0 -> 11 in the cache */
   /* search165_0_1_0 -> 491 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 512 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 491, 514 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 515 },
      -1,
   } },
   /* search165_1_0 -> 498 in the cache */
   /* search165_1_1_0 -> 512 in the cache */
   /* search165_1_1 -> 513 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 498, 513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 516, 517 },
      -1,
   } },

   /* replace165_0 -> 512 in the cache */
   /* replace165_1 -> 498 in the cache */
   /* replace165_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 498, 11 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search166_0 -> 11 in the cache */
   /* search166_1_0_0 -> 512 in the cache */
   /* search166_1_0 -> 513 in the cache */
   /* search166_1_1_0 -> 498 in the cache */
   /* search166_1_1_1_0 -> 11 in the cache */
   /* search166_1_1_1 -> 508 in the cache */
   /* search166_1_1 -> 509 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 513, 509 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 520 },
      -1,
   } },

   /* replace166_0 -> 512 in the cache */
   /* replace166_1 -> 498 in the cache */
   /* replace166_2 -> 11 in the cache */
   /* replace166 -> 519 in the cache */

   /* ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search167_0 -> 11 in the cache */
   /* search167_1_0 -> 491 in the cache */
   /* search167_1_1_0_0 -> 512 in the cache */
   /* search167_1_1_0 -> 513 in the cache */
   /* search167_1_1 -> 514 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 491, 514 },
      -1,
   } },
   /* search167_2_0 -> 498 in the cache */
   /* search167_2_1_0 -> 512 in the cache */
   /* search167_2_1 -> 513 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 498, 513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 11, 522, 523 },
      -1,
   } },

   /* replace167_0 -> 512 in the cache */
   /* replace167_1 -> 498 in the cache */
   /* replace167_2 -> 11 in the cache */
   /* replace167 -> 519 in the cache */

   /* ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search168_1_0 -> 512 in the cache */
   /* search168_1 -> 513 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 526 },
      -1,
   } },
   /* search168_2_1_0 -> 512 in the cache */
   /* search168_2_1 -> 513 in the cache */
   /* search168_2_2 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 527, 513, 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 525, 513, 528 },
      -1,
   } },

   /* replace168_0 -> 512 in the cache */
   /* replace168_1 -> 525 in the cache */
   /* replace168_2 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 525, 526 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 531 },
      -1,
   } },
   /* search169_1_0 -> 268 in the cache */
   /* search169_1_1_0 -> 526 in the cache */
   /* search169_1_1 -> 527 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 268, 527 },
      -1,
   } },
   /* search169_2 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 532, 533, 526 },
      -1,
   } },

   /* replace169_0 -> 531 in the cache */
   /* replace169_1 -> 268 in the cache */
   /* replace169_2 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 268, 526 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search170_0_0 -> 531 in the cache */
   /* search170_0 -> 532 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 536 },
      -1,
   } },
   /* search170_1_1 -> 498 in the cache */
   /* search170_1_2 -> 485 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 537, 498, 485 },
      -1,
   } },
   /* search170_2_0 -> 536 in the cache */
   /* search170_2_1 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 536, 498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 532, 538, 539 },
      -1,
   } },

   /* replace170_0 -> 531 in the cache */
   /* replace170_1 -> 485 in the cache */
   /* replace170_2_0 -> 536 in the cache */
   /* replace170_2_1 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 536, 498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 531, 485, 541 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search171_0 -> 491 in the cache */
   /* search171_1_0_0_0 -> 491 in the cache */
   /* search171_1_0_0_1_0 -> 11 in the cache */
   /* search171_1_0_0_1 -> 508 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 491, 508 },
      -1,
   } },
   /* search171_1_0_1_0 -> 491 in the cache */
   /* search171_1_0_1_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 491, 544 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 543, 545 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 546 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 491, 547 },
      -1,
   } },

   /* replace171_0 -> 268 in the cache */
   /* replace171_1 -> 491 in the cache */
   /* replace171_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 268, 491, 11 },
      -1,
   } },

   /* ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search172_0 -> 16 in the cache */
   /* search172_1 -> 36 in the cache */
   /* search172_2_0 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 36, 550 },
      -1,
   } },

   /* replace172_0 -> 474 in the cache */
   /* replace172_1 -> 36 in the cache */
   /* replace172_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 474, 36, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search173_0 -> 16 in the cache */
   /* search173_1_0 -> 16 in the cache */
   /* search173_1_1 -> 36 in the cache */
   /* search173_1 -> 284 in the cache */
   /* search173_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 284, 141 },
      -1,
   } },

   /* replace173_0_0 -> 36 in the cache */
   /* replace173_0_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 36, 141 },
      -1,
   } },
   /* replace173_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 554, 16 },
      -1,
   } },

   /* ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search174_0_0 -> 16 in the cache */
   /* search174_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search174_1_0 -> 16 in the cache */
   /* search174_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 141 },
      0,
   } },
   /* search174_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 556, 557, 169 },
      -1,
   } },

   /* replace174_0_0 -> 36 in the cache */
   /* replace174_0_1 -> 141 in the cache */
   /* replace174_0_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 141, 169 },
      -1,
   } },
   /* replace174_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 559, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search175_0 -> 16 in the cache */
   /* search175_1_0 -> 16 in the cache */
   /* search175_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search175_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 561, 141 },
      -1,
   } },

   /* replace175_0_0 -> 254 in the cache */
   /* replace175_0_1 -> 36 in the cache */
   /* replace175_0_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 254, 36, 141 },
      -1,
   } },
   /* replace175_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 563, 16 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search176_0_0 -> 16 in the cache */
   /* search176_0_1_0 -> 254 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 565 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 254, 566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 567 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search176_1_1 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 569, 565 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 568, 570 },
      -1,
   } },

   /* replace176_0 -> 16 in the cache */
   /* replace176_1 -> 569 in the cache */
   /* replace176_2 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 569, 565 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search177_0_0 -> 16 in the cache */
   /* search177_0_1_0_0 -> 254 in the cache */
   /* search177_0_1_0_1_0 -> 565 in the cache */
   /* search177_0_1_0_1 -> 566 in the cache */
   /* search177_0_1_0 -> 567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 567 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 573 },
      -1,
   } },
   /* search177_1_0 -> 569 in the cache */
   /* search177_1_1_0 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 565 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 569, 575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 574, 576 },
      -1,
   } },

   /* replace177_0 -> 16 in the cache */
   /* replace177_1 -> 569 in the cache */
   /* replace177_2_0 -> 565 in the cache */
   /* replace177_2 -> 575 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 569, 575 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search178_0 -> 16 in the cache */
   /* search178_1_0 -> 565 in the cache */
   /* search178_1_1_0 -> 569 in the cache */
   /* search178_1_1_1_0 -> 16 in the cache */
   /* search178_1_1_1 -> 290 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 569, 290 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 565, 579 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 580 },
      -1,
   } },

   /* replace178_0 -> 16 in the cache */
   /* replace178_1 -> 569 in the cache */
   /* replace178_2 -> 565 in the cache */
   /* replace178 -> 572 in the cache */

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search179_0_0 -> 16 in the cache */
   /* search179_0_1_0 -> 254 in the cache */
   /* search179_0_1_1_0_0 -> 512 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 512 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 254, 583 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 584 },
      -1,
   } },
   /* search179_1_0 -> 569 in the cache */
   /* search179_1_1_0 -> 512 in the cache */
   /* search179_1_1 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 569, 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 585, 586 },
      -1,
   } },

   /* replace179_0 -> 512 in the cache */
   /* replace179_1 -> 569 in the cache */
   /* replace179_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 569, 16 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search180_0 -> 16 in the cache */
   /* search180_1_0_0 -> 512 in the cache */
   /* search180_1_0 -> 582 in the cache */
   /* search180_1_1_0 -> 569 in the cache */
   /* search180_1_1_1_0 -> 16 in the cache */
   /* search180_1_1_1 -> 290 in the cache */
   /* search180_1_1 -> 579 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 582, 579 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 589 },
      -1,
   } },

   /* replace180_0 -> 512 in the cache */
   /* replace180_1 -> 569 in the cache */
   /* replace180_2 -> 16 in the cache */
   /* replace180 -> 588 in the cache */

   /* ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search181_0 -> 16 in the cache */
   /* search181_1_0 -> 254 in the cache */
   /* search181_1_1_0_0 -> 512 in the cache */
   /* search181_1_1_0 -> 582 in the cache */
   /* search181_1_1 -> 583 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 254, 583 },
      -1,
   } },
   /* search181_2_0 -> 569 in the cache */
   /* search181_2_1_0 -> 512 in the cache */
   /* search181_2_1 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 569, 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 16, 591, 592 },
      -1,
   } },

   /* replace181_0 -> 512 in the cache */
   /* replace181_1 -> 569 in the cache */
   /* replace181_2 -> 16 in the cache */
   /* replace181 -> 588 in the cache */

   /* ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   /* search182_0 -> 383 in the cache */
   /* search182_1_0 -> 512 in the cache */
   /* search182_1 -> 582 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 594 },
      -1,
   } },
   /* search182_2_1_0 -> 512 in the cache */
   /* search182_2_1 -> 582 in the cache */
   /* search182_2_2 -> 594 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 595, 582, 594 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 383, 582, 596 },
      -1,
   } },

   /* replace182_0 -> 512 in the cache */
   /* replace182_1 -> 383 in the cache */
   /* replace182_2 -> 594 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 383, 594 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search183_0_0 -> 531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 531 },
      -1,
   } },
   /* search183_1_0 -> 36 in the cache */
   /* search183_1_1_0 -> 594 in the cache */
   /* search183_1_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 595 },
      -1,
   } },
   /* search183_2 -> 594 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 599, 600, 594 },
      -1,
   } },

   /* replace183_0 -> 531 in the cache */
   /* replace183_1 -> 36 in the cache */
   /* replace183_2 -> 594 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 36, 594 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search184_0_0 -> 531 in the cache */
   /* search184_0 -> 599 in the cache */
   /* search184_1_0_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 189 },
      -1,
   } },
   /* search184_1_1 -> 569 in the cache */
   /* search184_1_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 603, 569, 169 },
      -1,
   } },
   /* search184_2_0 -> 189 in the cache */
   /* search184_2_1 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 189, 569 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 599, 604, 605 },
      -1,
   } },

   /* replace184_0 -> 531 in the cache */
   /* replace184_1 -> 169 in the cache */
   /* replace184_2_0 -> 189 in the cache */
   /* replace184_2_1 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 189, 569 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 531, 169, 607 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search185_0 -> 254 in the cache */
   /* search185_1_0_0_0 -> 254 in the cache */
   /* search185_1_0_0_1_0 -> 16 in the cache */
   /* search185_1_0_0_1 -> 290 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 254, 290 },
      -1,
   } },
   /* search185_1_0_1_0 -> 254 in the cache */
   /* search185_1_0_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 254, 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 609, 611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 612 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 254, 613 },
      -1,
   } },

   /* replace185_0 -> 36 in the cache */
   /* replace185_1 -> 254 in the cache */
   /* replace185_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 254, 16 },
      -1,
   } },

   /* ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search186_0 -> 21 in the cache */
   /* search186_1 -> 41 in the cache */
   /* search186_2_0 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 41, 616 },
      -1,
   } },

   /* replace186_0 -> 474 in the cache */
   /* replace186_1 -> 41 in the cache */
   /* replace186_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 474, 41, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search187_0 -> 21 in the cache */
   /* search187_1_0 -> 21 in the cache */
   /* search187_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 619, 620 },
      -1,
   } },

   /* replace187_0_0 -> 41 in the cache */
   /* replace187_0_1 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 41, 620 },
      -1,
   } },
   /* replace187_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 622, 21 },
      -1,
   } },

   /* ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search188_0_0 -> 21 in the cache */
   /* search188_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search188_1_0 -> 21 in the cache */
   /* search188_1_1 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 21, 620 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 624, 625, 626 },
      -1,
   } },

   /* replace188_0_0 -> 41 in the cache */
   /* replace188_0_1 -> 620 in the cache */
   /* replace188_0_2 -> 626 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 620, 626 },
      -1,
   } },
   /* replace188_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 628, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search189_0 -> 21 in the cache */
   /* search189_1_0 -> 21 in the cache */
   /* search189_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search189_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 630, 620 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace189_0_1 -> 41 in the cache */
   /* replace189_0_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 632, 41, 620 },
      -1,
   } },
   /* replace189_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 633, 21 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search190_0_0 -> 21 in the cache */
   /* search190_0_1_0 -> 632 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 632, 636 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 637 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search190_1_1 -> 635 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 639, 635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 638, 640 },
      -1,
   } },

   /* replace190_0 -> 21 in the cache */
   /* replace190_1 -> 639 in the cache */
   /* replace190_2 -> 635 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 639, 635 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search191_0_0 -> 21 in the cache */
   /* search191_0_1_0_0 -> 632 in the cache */
   /* search191_0_1_0_1_0 -> 635 in the cache */
   /* search191_0_1_0_1 -> 636 in the cache */
   /* search191_0_1_0 -> 637 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 637 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 643 },
      -1,
   } },
   /* search191_1_0 -> 639 in the cache */
   /* search191_1_1_0 -> 635 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 639, 645 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 644, 646 },
      -1,
   } },

   /* replace191_0 -> 21 in the cache */
   /* replace191_1 -> 639 in the cache */
   /* replace191_2_0 -> 635 in the cache */
   /* replace191_2 -> 645 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 639, 645 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search192_0 -> 21 in the cache */
   /* search192_1_0 -> 635 in the cache */
   /* search192_1_1_0 -> 639 in the cache */
   /* search192_1_1_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 639, 649 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 635, 650 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 651 },
      -1,
   } },

   /* replace192_0 -> 21 in the cache */
   /* replace192_1 -> 639 in the cache */
   /* replace192_2 -> 635 in the cache */
   /* replace192 -> 642 in the cache */

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search193_0_0 -> 21 in the cache */
   /* search193_0_1_0 -> 632 in the cache */
   /* search193_0_1_1_0_0 -> 512 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 512 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 653 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 632, 654 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 655 },
      -1,
   } },
   /* search193_1_0 -> 639 in the cache */
   /* search193_1_1_0 -> 512 in the cache */
   /* search193_1_1 -> 653 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 639, 653 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 656, 657 },
      -1,
   } },

   /* replace193_0 -> 512 in the cache */
   /* replace193_1 -> 639 in the cache */
   /* replace193_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 639, 21 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search194_0 -> 21 in the cache */
   /* search194_1_0_0 -> 512 in the cache */
   /* search194_1_0 -> 653 in the cache */
   /* search194_1_1_0 -> 639 in the cache */
   /* search194_1_1_1_0 -> 21 in the cache */
   /* search194_1_1_1 -> 649 in the cache */
   /* search194_1_1 -> 650 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 653, 650 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 660 },
      -1,
   } },

   /* replace194_0 -> 512 in the cache */
   /* replace194_1 -> 639 in the cache */
   /* replace194_2 -> 21 in the cache */
   /* replace194 -> 659 in the cache */

   /* ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search195_0 -> 21 in the cache */
   /* search195_1_0 -> 632 in the cache */
   /* search195_1_1_0_0 -> 512 in the cache */
   /* search195_1_1_0 -> 653 in the cache */
   /* search195_1_1 -> 654 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 632, 654 },
      -1,
   } },
   /* search195_2_0 -> 639 in the cache */
   /* search195_2_1_0 -> 512 in the cache */
   /* search195_2_1 -> 653 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 639, 653 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 21, 662, 663 },
      -1,
   } },

   /* replace195_0 -> 512 in the cache */
   /* replace195_1 -> 639 in the cache */
   /* replace195_2 -> 21 in the cache */
   /* replace195 -> 659 in the cache */

   /* ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search196_1_0 -> 512 in the cache */
   /* search196_1 -> 653 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 666 },
      -1,
   } },
   /* search196_2_1_0 -> 512 in the cache */
   /* search196_2_1 -> 653 in the cache */
   /* search196_2_2 -> 666 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 667, 653, 666 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 665, 653, 668 },
      -1,
   } },

   /* replace196_0 -> 512 in the cache */
   /* replace196_1 -> 665 in the cache */
   /* replace196_2 -> 666 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 512, 665, 666 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search197_0_0 -> 531 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 531 },
      -1,
   } },
   /* search197_1_0 -> 41 in the cache */
   /* search197_1_1_0 -> 666 in the cache */
   /* search197_1_1 -> 667 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 41, 667 },
      -1,
   } },
   /* search197_2 -> 666 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 671, 672, 666 },
      -1,
   } },

   /* replace197_0 -> 531 in the cache */
   /* replace197_1 -> 41 in the cache */
   /* replace197_2 -> 666 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 41, 666 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search198_0_0 -> 531 in the cache */
   /* search198_0 -> 671 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 675 },
      -1,
   } },
   /* search198_1_1 -> 639 in the cache */
   /* search198_1_2 -> 626 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 676, 639, 626 },
      -1,
   } },
   /* search198_2_0 -> 675 in the cache */
   /* search198_2_1 -> 639 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 675, 639 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 671, 677, 678 },
      -1,
   } },

   /* replace198_0 -> 531 in the cache */
   /* replace198_1 -> 626 in the cache */
   /* replace198_2_0 -> 675 in the cache */
   /* replace198_2_1 -> 639 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 675, 639 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 531, 626, 680 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search199_0 -> 632 in the cache */
   /* search199_1_0_0_0 -> 632 in the cache */
   /* search199_1_0_0_1_0 -> 21 in the cache */
   /* search199_1_0_0_1 -> 649 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 632, 649 },
      -1,
   } },
   /* search199_1_0_1_0 -> 632 in the cache */
   /* search199_1_0_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 632, 683 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 682, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 685 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 632, 686 },
      -1,
   } },

   /* replace199_0 -> 41 in the cache */
   /* replace199_1 -> 632 in the cache */
   /* replace199_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 632, 21 },
      -1,
   } },

   /* ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search200_0_0 -> 155 in the cache */
   /* search200_0_1 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 155, 156 },
      0,
   } },
   /* search200_1_0 -> 155 in the cache */
   /* search200_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 155, 158 },
      0,
   } },
   /* search200_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 689, 690, 159 },
      -1,
   } },

   /* replace200_0_0 -> 156 in the cache */
   /* replace200_0_1 -> 158 in the cache */
   /* replace200_0_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 156, 158, 159 },
      -1,
   } },
   /* replace200_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 692, 155 },
      -1,
   } },

   /* ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a') */
   /* search201_0 -> 0 in the cache */
   /* search201_1 -> 266 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 266, 694 },
      -1,
   } },

   /* replace201_0_0_0 -> 0 in the cache */
   /* replace201_0_0 -> 223 in the cache */
   /* replace201_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 223, 694 },
      -1,
   } },
   /* replace201_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 696, 0 },
      -1,
   } },

   /* ('ftrunc@16', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search202_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace202_0_0 -> 11 in the cache */
   /* replace202_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 119 },
      -1,
   } },
   /* replace202_1_0_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 700 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 701 },
      -1,
   } },
   /* replace202_2_0_0 -> 11 in the cache */
   /* replace202_2_0 -> 700 in the cache */
   /* replace202_2 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 699, 702, 701 },
      -1,
   } },

   /* ('ftrunc@32', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search203_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace203_0_0 -> 16 in the cache */
   /* replace203_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 121 },
      -1,
   } },
   /* replace203_1_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 706 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 707 },
      -1,
   } },
   /* replace203_2_0_0 -> 16 in the cache */
   /* replace203_2_0 -> 706 in the cache */
   /* replace203_2 -> 707 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 705, 708, 707 },
      -1,
   } },

   /* ('ftrunc@64', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search204_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace204_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 711 },
      -1,
   } },
   /* replace204_1_0_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 713 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 714 },
      -1,
   } },
   /* replace204_2_0_0 -> 21 in the cache */
   /* replace204_2_0 -> 713 in the cache */
   /* replace204_2 -> 714 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 712, 715, 714 },
      -1,
   } },

   /* ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search205_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace205_0 -> 11 in the cache */
   /* replace205_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 718 },
      -1,
   } },

   /* ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search206_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace206_0 -> 16 in the cache */
   /* replace206_1_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 721 },
      -1,
   } },

   /* ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search207_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace207_0 -> 21 in the cache */
   /* replace207_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 724 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fadd@16', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@16', 'b', ('ffloor', 'a')) */
   /* search208_0 -> 11 in the cache */
   /* search208_1_0 -> 268 in the cache */
   /* search208_1_1_0_0 -> 11 in the cache */
   /* search208_1_1_0 -> 718 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 718 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 268, 726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 11, 727 },
      -1,
   } },

   /* replace208_0 -> 268 in the cache */
   /* replace208_1_0 -> 11 in the cache */
   /* replace208_1 -> 717 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 268, 717 },
      -1,
   } },

   /* ('fadd@32', 'a', ('fadd@32', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@32', 'b', ('ffloor', 'a')) */
   /* search209_0 -> 16 in the cache */
   /* search209_1_0 -> 36 in the cache */
   /* search209_1_1_0_0 -> 16 in the cache */
   /* search209_1_1_0 -> 721 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 721 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 730 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 16, 731 },
      -1,
   } },

   /* replace209_0 -> 36 in the cache */
   /* replace209_1_0 -> 16 in the cache */
   /* replace209_1 -> 720 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 36, 720 },
      -1,
   } },

   /* ('fadd@64', 'a', ('fadd@64', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@64', 'b', ('ffloor', 'a')) */
   /* search210_0 -> 21 in the cache */
   /* search210_1_0 -> 41 in the cache */
   /* search210_1_1_0_0 -> 21 in the cache */
   /* search210_1_1_0 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 724 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 41, 734 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 21, 735 },
      -1,
   } },

   /* replace210_0 -> 41 in the cache */
   /* replace210_1_0 -> 21 in the cache */
   /* replace210_1 -> 723 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 41, 723 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search211_0 -> 11 in the cache */
   /* search211_1_0_0 -> 11 in the cache */
   /* search211_1_0 -> 718 in the cache */
   /* search211_1 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 726 },
      -1,
   } },

   /* replace211_0 -> 11 in the cache */
   /* replace211 -> 717 in the cache */

   /* ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search212_0 -> 16 in the cache */
   /* search212_1_0_0 -> 16 in the cache */
   /* search212_1_0 -> 721 in the cache */
   /* search212_1 -> 730 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 730 },
      -1,
   } },

   /* replace212_0 -> 16 in the cache */
   /* replace212 -> 720 in the cache */

   /* ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search213_0 -> 21 in the cache */
   /* search213_1_0_0 -> 21 in the cache */
   /* search213_1_0 -> 724 in the cache */
   /* search213_1 -> 734 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 734 },
      -1,
   } },

   /* replace213_0 -> 21 in the cache */
   /* replace213 -> 723 in the cache */

   /* ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search214_0 -> 11 in the cache */
   /* search214 -> 718 in the cache */

   /* replace214_0 -> 11 in the cache */
   /* replace214_1_0 -> 11 in the cache */
   /* replace214_1 -> 717 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 717 },
      -1,
   } },

   /* ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search215_0 -> 16 in the cache */
   /* search215 -> 721 in the cache */

   /* replace215_0 -> 16 in the cache */
   /* replace215_1_0 -> 16 in the cache */
   /* replace215_1 -> 720 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 720 },
      -1,
   } },

   /* ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search216_0 -> 21 in the cache */
   /* search216 -> 724 in the cache */

   /* replace216_0 -> 21 in the cache */
   /* replace216_1_0 -> 21 in the cache */
   /* replace216_1 -> 723 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 723 },
      -1,
   } },

   /* ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a'))) */
   /* search217_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace217_0_0_0 -> 32 in the cache */
   /* replace217_0_0 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 107 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 745 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search218_0 -> 11 in the cache */
   /* search218_1 -> 268 in the cache */
   /* search218_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 268, 479 },
      -1,
   } },

   /* replace218_0_0 -> 11 in the cache */
   /* replace218_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 11, 268 },
      -1,
   } },
   /* replace218_1 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 748, 479 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search219_0 -> 16 in the cache */
   /* search219_1 -> 36 in the cache */
   /* search219_2 -> 141 in the cache */
   /* search219 -> 245 in the cache */

   /* replace219_0_0 -> 16 in the cache */
   /* replace219_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace219_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 750, 141 },
      -1,
   } },

   /* ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search220_0 -> 21 in the cache */
   /* search220_1 -> 41 in the cache */
   /* search220_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 620 },
      -1,
   } },

   /* replace220_0_0 -> 21 in the cache */
   /* replace220_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace220_1 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 753, 620 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search221_0 -> 16 in the cache */
   /* search221_1 -> 36 in the cache */
   /* search221_2 -> 141 in the cache */
   /* search221 -> 397 in the cache */

   /* replace221_0_0 -> 16 in the cache */
   /* replace221_0_1 -> 36 in the cache */
   /* replace221_0 -> 140 in the cache */
   /* replace221_1 -> 141 in the cache */
   /* replace221 -> 298 in the cache */

   /* ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search222_0 -> 11 in the cache */
   /* search222_1 -> 268 in the cache */
   /* search222_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 268, 479 },
      -1,
   } },

   /* replace222_0_0 -> 11 in the cache */
   /* replace222_0_1 -> 268 in the cache */
   /* replace222_0 -> 748 in the cache */
   /* replace222_1 -> 479 in the cache */
   /* replace222 -> 749 in the cache */

   /* ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search223_0 -> 16 in the cache */
   /* search223_1 -> 36 in the cache */
   /* search223_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace223_0_0 -> 16 in the cache */
   /* replace223_0_1 -> 36 in the cache */
   /* replace223_0 -> 750 in the cache */
   /* replace223_1 -> 141 in the cache */
   /* replace223 -> 751 in the cache */

   /* ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search224_0 -> 21 in the cache */
   /* search224_1 -> 41 in the cache */
   /* search224_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 620 },
      -1,
   } },

   /* replace224_0_0 -> 21 in the cache */
   /* replace224_0_1 -> 41 in the cache */
   /* replace224_0 -> 753 in the cache */
   /* replace224_1 -> 620 in the cache */
   /* replace224 -> 754 in the cache */

   /* ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search225_0 -> 16 in the cache */
   /* search225_1 -> 36 in the cache */
   /* search225_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace225_0_0 -> 16 in the cache */
   /* replace225_0_1 -> 36 in the cache */
   /* replace225_0 -> 140 in the cache */
   /* replace225_1 -> 141 in the cache */
   /* replace225 -> 298 in the cache */

   /* ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 759 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 760 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 762, 763 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_iand,
      2, 2,
      { 761, 764 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 3,
      { 765, 766 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 4,
      { 767, 768 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace226_1_0_0_0 -> 762 in the cache */
   /* replace226_1_0_0_1 -> 763 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 762, 763 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 771, 772 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 773, 774 },
      -1,
   } },
   /* replace226_2_0 -> 772 in the cache */
   /* replace226_2_1 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 772, 774 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 770, 775, 776 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b') */
   /* search227_0 -> 0 in the cache */
   /* search227_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   /* replace227_0_3 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 779, 780, 781, 280 },
      -1,
   } },
   /* replace227_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 782, 3 },
      -1,
   } },

   /* ('fdot4', 'a', 0.0) => 0.0 */
   /* search228_0 -> 32 in the cache */
   /* search228_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* replace228 -> 117 in the cache */

   /* ('fdot3', 'a', 0.0) => 0.0 */
   /* search229_0 -> 32 in the cache */
   /* search229_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* replace229 -> 117 in the cache */

   /* ('fdot2', 'a', 0.0) => 0.0 */
   /* search230_0 -> 32 in the cache */
   /* search230_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* replace230 -> 117 in the cache */

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search231_0_0 -> 155 in the cache */
   /* search231_0_1 -> 156 in the cache */
   /* search231_0_2 -> 158 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 155, 156, 158, 787 },
      -1,
   } },
   /* search231_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 788, 159 },
      -1,
   } },

   /* replace231_0_0 -> 155 in the cache */
   /* replace231_0_1 -> 156 in the cache */
   /* replace231_0_2 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 155, 156, 158 },
      -1,
   } },
   /* replace231_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 790, 159 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search232_0_0 -> 0 in the cache */
   /* search232_0_1 -> 266 in the cache */
   /* search232_0_2 -> 266 in the cache */
   /* search232_0_3 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 0, 266, 266, 266 },
      -1,
   } },
   /* search232_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 792, 3 },
      -1,
   } },

   /* replace232_0 -> 0 in the cache */
   /* replace232_1 -> 3 in the cache */
   /* replace232 -> 274 in the cache */

   /* ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search233_0_0 -> 132 in the cache */
   /* search233_0_1 -> 133 in the cache */
   { .constant = {
      { nir_search_value_constant, -3 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search233_0_3 -> 794 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 132, 133, 794, 794 },
      -1,
   } },
   /* search233_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 795, 135 },
      -1,
   } },

   /* replace233_0_0 -> 132 in the cache */
   /* replace233_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 132, 133 },
      -1,
   } },
   /* replace233_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 797, 135 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search234_0_0 -> 155 in the cache */
   /* search234_0_1 -> 156 in the cache */
   /* search234_0_2 -> 158 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 155, 156, 158, 799 },
      -1,
   } },
   /* search234_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 800, 159 },
      -1,
   } },

   /* replace234_0_0 -> 155 in the cache */
   /* replace234_0_1 -> 156 in the cache */
   /* replace234_0_2 -> 158 in the cache */
   /* replace234_0 -> 790 in the cache */
   /* replace234_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 790, 159 },
      -1,
   } },

   /* ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search235_0_0 -> 0 in the cache */
   /* search235_0_1 -> 266 in the cache */
   /* search235_0_2 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 0, 266, 266 },
      -1,
   } },
   /* search235_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 803, 3 },
      -1,
   } },

   /* replace235_0 -> 0 in the cache */
   /* replace235_1 -> 3 in the cache */
   /* replace235 -> 274 in the cache */

   /* ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search236_0_0 -> 132 in the cache */
   /* search236_0_1 -> 133 in the cache */
   /* search236_0_2 -> 794 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 132, 133, 794 },
      -1,
   } },
   /* search236_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 805, 135 },
      -1,
   } },

   /* replace236_0_0 -> 132 in the cache */
   /* replace236_0_1 -> 133 in the cache */
   /* replace236_0 -> 797 in the cache */
   /* replace236_1 -> 135 in the cache */
   /* replace236 -> 798 in the cache */

   /* ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search237_0_0 -> 0 in the cache */
   /* search237_0_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 0, 266 },
      -1,
   } },
   /* search237_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 807, 3 },
      -1,
   } },

   /* replace237_0 -> 0 in the cache */
   /* replace237_1 -> 3 in the cache */
   /* replace237 -> 274 in the cache */

   /* ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y') */
   /* search238_0 -> 32 in the cache */
   /* search238_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 32, 252 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      -1, 0,
      { 810, 811 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b')) */
   /* search239_0 -> 0 in the cache */
   /* search239_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace239_0_0 -> 0 in the cache */
   /* replace239_0_1 -> 3 in the cache */
   /* replace239_0 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 1,
      { 274 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b')) */
   /* search240_0 -> 0 in the cache */
   /* search240_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace240_0_0 -> 0 in the cache */
   /* replace240_0_1 -> 3 in the cache */
   /* replace240_0 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum3,
      -1, 1,
      { 274 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b')) */
   /* search241_0 -> 0 in the cache */
   /* search241_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace241_0_0 -> 0 in the cache */
   /* replace241_0_1 -> 3 in the cache */
   /* replace241_0 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum4,
      -1, 1,
      { 274 },
      -1,
   } },

   /* ('fsum2', 'a') => ('fadd', 'a.x', 'a.y') */
   /* search242_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace242_0 -> 810 in the cache */
   /* replace242_1 -> 811 in the cache */
   /* replace242 -> 812 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 820 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 821 },
      0,
   } },
   /* search243_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 822, 252 },
      -1,
   } },

   /* replace243_0_0 -> 252 in the cache */
   /* replace243_0_1_0 -> 32 in the cache */
   /* replace243_0_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 252, 107 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 824 },
      -1,
   } },

   /* ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd')) */
   /* search244_0_0_0 -> 132 in the cache */
   /* search244_0_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 132, 293 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 826, 827 },
      -1,
   } },
   /* search244_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 828, 321 },
      -1,
   } },

   /* replace244_0_0 -> 132 in the cache */
   /* replace244_0_1_0 -> 133 in the cache */
   /* replace244_0_1_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 133, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 132, 830 },
      -1,
   } },
   /* replace244_1_0 -> 135 in the cache */
   /* replace244_1_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 135, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 831, 832 },
      -1,
   } },

   /* ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c')) */
   /* search245_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 834 },
      -1,
   } },
   /* search245_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 835, 319 },
      -1,
   } },

   /* replace245_0 -> 0 in the cache */
   /* replace245_1_0 -> 3 in the cache */
   /* replace245_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 837 },
      -1,
   } },

   /* ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search246_0_0 -> 6 in the cache */
   /* search246_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search246_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 839, 319 },
      -1,
   } },

   /* replace246_0_0_0_0 -> 36 in the cache */
   /* replace246_0_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 9 },
      -1,
   } },
   /* replace246_0_0_1_0 -> 141 in the cache */
   /* replace246_0_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 141, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 841, 842 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 843, 844 },
      -1,
   } },
   /* replace246_1_0 -> 6 in the cache */
   /* replace246_1_1_0 -> 36 in the cache */
   /* replace246_1_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 36, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 846 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 845, 847, 848 },
      -1,
   } },

   /* ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search247_0_0 -> 6 in the cache */
   /* search247_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search247_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 850, 319 },
      -1,
   } },

   /* replace247_0_0_0_0 -> 36 in the cache */
   /* replace247_0_0_0_1 -> 9 in the cache */
   /* replace247_0_0_0 -> 841 in the cache */
   /* replace247_0_0_1_0 -> 141 in the cache */
   /* replace247_0_0_1_1 -> 9 in the cache */
   /* replace247_0_0_1 -> 842 in the cache */
   /* replace247_0_0 -> 843 in the cache */
   /* replace247_0_1 -> 844 in the cache */
   /* replace247_0 -> 845 in the cache */
   /* replace247_1_0 -> 6 in the cache */
   /* replace247_1_1_0 -> 36 in the cache */
   /* replace247_1_1_1 -> 141 in the cache */
   /* replace247_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 846 },
      -1,
   } },
   /* replace247_2 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 845, 852, 848 },
      -1,
   } },

   /* ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7)) */
   /* search248_0_0 -> 6 in the cache */
   /* search248_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search248_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 854, 319 },
      -1,
   } },

   /* replace248_0 -> 6 in the cache */
   /* replace248_1_0_0_0 -> 36 in the cache */
   /* replace248_1_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 9 },
      -1,
   } },
   /* replace248_1_0_1_0 -> 141 in the cache */
   /* replace248_1_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 141, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 856, 857 },
      -1,
   } },
   /* replace248_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 858, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 6, 859 },
      -1,
   } },

   /* ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search249_0_0 -> 11 in the cache */
   /* search249_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search249_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 861, 319 },
      -1,
   } },

   /* replace249_0_0_0_0 -> 36 in the cache */
   /* replace249_0_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 14 },
      -1,
   } },
   /* replace249_0_0_1_0 -> 141 in the cache */
   /* replace249_0_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 141, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 863, 864 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 865, 866 },
      -1,
   } },
   /* replace249_1_0 -> 11 in the cache */
   /* replace249_1_1_0 -> 36 in the cache */
   /* replace249_1_1_1 -> 141 in the cache */
   /* replace249_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 846 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 867, 868, 869 },
      -1,
   } },

   /* ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search250_0_0 -> 11 in the cache */
   /* search250_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search250_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 871, 319 },
      -1,
   } },

   /* replace250_0_0_0_0 -> 36 in the cache */
   /* replace250_0_0_0_1 -> 14 in the cache */
   /* replace250_0_0_0 -> 863 in the cache */
   /* replace250_0_0_1_0 -> 141 in the cache */
   /* replace250_0_0_1_1 -> 14 in the cache */
   /* replace250_0_0_1 -> 864 in the cache */
   /* replace250_0_0 -> 865 in the cache */
   /* replace250_0_1 -> 866 in the cache */
   /* replace250_0 -> 867 in the cache */
   /* replace250_1_0 -> 11 in the cache */
   /* replace250_1_1_0 -> 36 in the cache */
   /* replace250_1_1_1 -> 141 in the cache */
   /* replace250_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 846 },
      -1,
   } },
   /* replace250_2 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 867, 873, 869 },
      -1,
   } },

   /* ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15)) */
   /* search251_0_0 -> 11 in the cache */
   /* search251_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search251_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 875, 319 },
      -1,
   } },

   /* replace251_0 -> 11 in the cache */
   /* replace251_1_0_0_0 -> 36 in the cache */
   /* replace251_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 14 },
      -1,
   } },
   /* replace251_1_0_1_0 -> 141 in the cache */
   /* replace251_1_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 141, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 877, 878 },
      -1,
   } },
   /* replace251_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 879, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 11, 880 },
      -1,
   } },

   /* ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search252_0_0 -> 16 in the cache */
   /* search252_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search252_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 882, 319 },
      -1,
   } },

   /* replace252_0_0_0_0 -> 36 in the cache */
   /* replace252_0_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 19 },
      -1,
   } },
   /* replace252_0_0_1_0 -> 141 in the cache */
   /* replace252_0_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 141, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 884, 885 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x20 /* 32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 886, 887 },
      -1,
   } },
   /* replace252_1_0 -> 16 in the cache */
   /* replace252_1_1_0 -> 36 in the cache */
   /* replace252_1_1_1 -> 141 in the cache */
   /* replace252_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 846 },
      -1,
   } },
   /* replace252_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 888, 889, 128 },
      -1,
   } },

   /* ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search253_0_0 -> 16 in the cache */
   /* search253_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search253_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 891, 319 },
      -1,
   } },

   /* replace253_0_0_0_0 -> 36 in the cache */
   /* replace253_0_0_0_1 -> 19 in the cache */
   /* replace253_0_0_0 -> 884 in the cache */
   /* replace253_0_0_1_0 -> 141 in the cache */
   /* replace253_0_0_1_1 -> 19 in the cache */
   /* replace253_0_0_1 -> 885 in the cache */
   /* replace253_0_0 -> 886 in the cache */
   /* replace253_0_1 -> 887 in the cache */
   /* replace253_0 -> 888 in the cache */
   /* replace253_1_0 -> 16 in the cache */
   /* replace253_1_1_0 -> 36 in the cache */
   /* replace253_1_1_1 -> 141 in the cache */
   /* replace253_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 846 },
      -1,
   } },
   /* replace253_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 888, 893, 128 },
      -1,
   } },

   /* ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31)) */
   /* search254_0_0 -> 16 in the cache */
   /* search254_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search254_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 895, 319 },
      -1,
   } },

   /* replace254_0 -> 16 in the cache */
   /* replace254_1_0_0_0 -> 36 in the cache */
   /* replace254_1_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 19 },
      -1,
   } },
   /* replace254_1_0_1_0 -> 141 in the cache */
   /* replace254_1_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 141, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 897, 898 },
      -1,
   } },
   /* replace254_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 899, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 16, 900 },
      -1,
   } },

   /* ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search255_0_0 -> 21 in the cache */
   /* search255_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search255_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 902, 319 },
      -1,
   } },

   /* replace255_0_0_0_0 -> 36 in the cache */
   /* replace255_0_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 24 },
      -1,
   } },
   /* replace255_0_0_1_0 -> 141 in the cache */
   /* replace255_0_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 141, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 904, 905 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x40 /* 64 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 906, 907 },
      -1,
   } },
   /* replace255_1_0 -> 21 in the cache */
   /* replace255_1_1_0 -> 36 in the cache */
   /* replace255_1_1_1 -> 141 in the cache */
   /* replace255_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 846 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 908, 909, 910 },
      -1,
   } },

   /* ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search256_0_0 -> 21 in the cache */
   /* search256_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search256_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 912, 319 },
      -1,
   } },

   /* replace256_0_0_0_0 -> 36 in the cache */
   /* replace256_0_0_0_1 -> 24 in the cache */
   /* replace256_0_0_0 -> 904 in the cache */
   /* replace256_0_0_1_0 -> 141 in the cache */
   /* replace256_0_0_1_1 -> 24 in the cache */
   /* replace256_0_0_1 -> 905 in the cache */
   /* replace256_0_0 -> 906 in the cache */
   /* replace256_0_1 -> 907 in the cache */
   /* replace256_0 -> 908 in the cache */
   /* replace256_1_0 -> 21 in the cache */
   /* replace256_1_1_0 -> 36 in the cache */
   /* replace256_1_1_1 -> 141 in the cache */
   /* replace256_1_1 -> 846 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 846 },
      -1,
   } },
   /* replace256_2 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 908, 914, 910 },
      -1,
   } },

   /* ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63)) */
   /* search257_0_0 -> 21 in the cache */
   /* search257_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search257_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 916, 319 },
      -1,
   } },

   /* replace257_0 -> 21 in the cache */
   /* replace257_1_0_0_0 -> 36 in the cache */
   /* replace257_1_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 24 },
      -1,
   } },
   /* replace257_1_0_1_0 -> 141 in the cache */
   /* replace257_1_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 141, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 918, 919 },
      -1,
   } },
   /* replace257_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 920, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 21, 921 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search258_0_0 -> 6 in the cache */
   /* search258_0_1 -> 33 in the cache */
   /* search258_0 -> 850 in the cache */
   /* search258_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 850, 36 },
      -1,
   } },

   /* replace258_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace258_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 924, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 925 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search259_0_0 -> 6 in the cache */
   /* search259_0_1 -> 33 in the cache */
   /* search259_0 -> 854 in the cache */
   /* search259_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 854, 36 },
      -1,
   } },

   /* replace259_0 -> 6 in the cache */
   /* replace259_1_0 -> 924 in the cache */
   /* replace259_1_1 -> 36 in the cache */
   /* replace259_1 -> 925 in the cache */
   /* replace259 -> 926 in the cache */

   /* ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b')) */
   /* search260_0_0 -> 6 in the cache */
   /* search260_0_1 -> 33 in the cache */
   /* search260_0 -> 839 in the cache */
   /* search260_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 839, 36 },
      -1,
   } },

   /* replace260_0 -> 6 in the cache */
   /* replace260_1_0 -> 924 in the cache */
   /* replace260_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 924, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 929 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search261_0_0 -> 11 in the cache */
   /* search261_0_1 -> 33 in the cache */
   /* search261_0 -> 871 in the cache */
   /* search261_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 871, 36 },
      -1,
   } },

   /* replace261_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace261_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 932, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 933 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search262_0_0 -> 11 in the cache */
   /* search262_0_1 -> 33 in the cache */
   /* search262_0 -> 875 in the cache */
   /* search262_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 875, 36 },
      -1,
   } },

   /* replace262_0 -> 11 in the cache */
   /* replace262_1_0 -> 932 in the cache */
   /* replace262_1_1 -> 36 in the cache */
   /* replace262_1 -> 933 in the cache */
   /* replace262 -> 934 in the cache */

   /* ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b')) */
   /* search263_0_0 -> 11 in the cache */
   /* search263_0_1 -> 33 in the cache */
   /* search263_0 -> 861 in the cache */
   /* search263_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 861, 36 },
      -1,
   } },

   /* replace263_0 -> 11 in the cache */
   /* replace263_1_0 -> 932 in the cache */
   /* replace263_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 932, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 937 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search264_0_0 -> 16 in the cache */
   /* search264_0_1 -> 33 in the cache */
   /* search264_0 -> 891 in the cache */
   /* search264_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 891, 36 },
      -1,
   } },

   /* replace264_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace264_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 940, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 941 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search265_0_0 -> 16 in the cache */
   /* search265_0_1 -> 33 in the cache */
   /* search265_0 -> 895 in the cache */
   /* search265_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 895, 36 },
      -1,
   } },

   /* replace265_0 -> 16 in the cache */
   /* replace265_1_0 -> 940 in the cache */
   /* replace265_1_1 -> 36 in the cache */
   /* replace265_1 -> 941 in the cache */
   /* replace265 -> 942 in the cache */

   /* ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b')) */
   /* search266_0_0 -> 16 in the cache */
   /* search266_0_1 -> 33 in the cache */
   /* search266_0 -> 882 in the cache */
   /* search266_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 882, 36 },
      -1,
   } },

   /* replace266_0 -> 16 in the cache */
   /* replace266_1_0 -> 940 in the cache */
   /* replace266_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 940, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 945 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search267_0_0 -> 21 in the cache */
   /* search267_0_1 -> 33 in the cache */
   /* search267_0 -> 912 in the cache */
   /* search267_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 912, 36 },
      -1,
   } },

   /* replace267_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffffffffffff /* 18446744073709551615 */ },
   } },
   /* replace267_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 948, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 949 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search268_0_0 -> 21 in the cache */
   /* search268_0_1 -> 33 in the cache */
   /* search268_0 -> 916 in the cache */
   /* search268_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 916, 36 },
      -1,
   } },

   /* replace268_0 -> 21 in the cache */
   /* replace268_1_0 -> 948 in the cache */
   /* replace268_1_1 -> 36 in the cache */
   /* replace268_1 -> 949 in the cache */
   /* replace268 -> 950 in the cache */

   /* ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b')) */
   /* search269_0_0 -> 21 in the cache */
   /* search269_0_1 -> 33 in the cache */
   /* search269_0 -> 902 in the cache */
   /* search269_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 902, 36 },
      -1,
   } },

   /* replace269_0 -> 21 in the cache */
   /* replace269_1_0 -> 948 in the cache */
   /* replace269_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 948, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 953 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b') */
   /* search270_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      9,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 955 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 956, 957 },
      -1,
   } },

   /* replace270_0 -> 16 in the cache */
   /* replace270_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b') */
   /* search271_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 960 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 961, 962 },
      -1,
   } },

   /* replace271_0 -> 0 in the cache */
   /* replace271_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b') */
   /* search272_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      11,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 965 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffe /* 4294967294 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 966, 967 },
      -1,
   } },

   /* replace272_0_0 -> 16 in the cache */
   /* replace272_0_1 -> 967 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 967 },
      -1,
   } },
   /* replace272_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 969, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b') */
   /* search273_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 971 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffc /* 4294967292 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 972, 973 },
      -1,
   } },

   /* replace273_0_0 -> 16 in the cache */
   /* replace273_0_1 -> 973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 973 },
      -1,
   } },
   /* replace273_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 975, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b') */
   /* search274_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      12,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 977 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff8 /* 4294967288 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 978, 979 },
      -1,
   } },

   /* replace274_0_0 -> 16 in the cache */
   /* replace274_0_1 -> 979 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 979 },
      -1,
   } },
   /* replace274_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 981, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b') */
   /* search275_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      13,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 983 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff0 /* 4294967280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 984, 985 },
      -1,
   } },

   /* replace275_0_0 -> 16 in the cache */
   /* replace275_0_1 -> 985 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 985 },
      -1,
   } },
   /* replace275_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 987, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b') */
   /* search276_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      14,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 989 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffe0 /* 4294967264 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 990, 991 },
      -1,
   } },

   /* replace276_0_0 -> 16 in the cache */
   /* replace276_0_1 -> 991 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 991 },
      -1,
   } },
   /* replace276_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 993, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b') */
   /* search277_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      15,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 995 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffc0 /* 4294967232 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 996, 997 },
      -1,
   } },

   /* replace277_0_0 -> 16 in the cache */
   /* replace277_0_1 -> 997 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 997 },
      -1,
   } },
   /* replace277_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 999, 36 },
      -1,
   } },

   /* ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1) */
   /* search278_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1001 },
      -1,
   } },
   /* search278_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1002, 406 },
      -1,
   } },

   /* replace278_0 -> 16 in the cache */
   /* replace278_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 406 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1) */
   /* search279_0_0 -> 16 in the cache */
   /* search279_0_1 -> 406 in the cache */
   /* search279_0 -> 1004 in the cache */
   /* search279_1 -> 967 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1004, 967 },
      -1,
   } },

   /* replace279_0 -> 16 in the cache */
   /* replace279_1 -> 406 in the cache */
   /* replace279 -> 1004 in the cache */

   /* ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1) */
   /* search280_0_0 -> 16 in the cache */
   /* search280_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 406 },
      -1,
   } },
   /* search280_1 -> 1001 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1006, 1001 },
      -1,
   } },

   /* replace280_0 -> 16 in the cache */
   /* replace280_1 -> 406 in the cache */
   /* replace280 -> 1006 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2) */
   /* search281_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffff /* 1073741823 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1008 },
      -1,
   } },
   /* search281_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1009, 411 },
      -1,
   } },

   /* replace281_0 -> 16 in the cache */
   /* replace281_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 411 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2) */
   /* search282_0_0 -> 16 in the cache */
   /* search282_0_1 -> 411 in the cache */
   /* search282_0 -> 1011 in the cache */
   /* search282_1 -> 973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1011, 973 },
      -1,
   } },

   /* replace282_0 -> 16 in the cache */
   /* replace282_1 -> 411 in the cache */
   /* replace282 -> 1011 in the cache */

   /* ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2) */
   /* search283_0_0 -> 16 in the cache */
   /* search283_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 411 },
      -1,
   } },
   /* search283_1 -> 1008 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1013, 1008 },
      -1,
   } },

   /* replace283_0 -> 16 in the cache */
   /* replace283_1 -> 411 in the cache */
   /* replace283 -> 1013 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16) */
   /* search284_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1015 },
      -1,
   } },
   /* search284_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1016, 866 },
      -1,
   } },

   /* replace284_0 -> 16 in the cache */
   /* replace284_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 866 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16) */
   /* search285_0_0 -> 16 in the cache */
   /* search285_0_1 -> 866 in the cache */
   /* search285_0 -> 1018 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff0000 /* 4294901760 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1018, 1019 },
      -1,
   } },

   /* replace285_0 -> 16 in the cache */
   /* replace285_1 -> 866 in the cache */
   /* replace285 -> 1018 in the cache */

   /* ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16) */
   /* search286_0_0 -> 16 in the cache */
   /* search286_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 866 },
      -1,
   } },
   /* search286_1 -> 1015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1021, 1015 },
      -1,
   } },

   /* replace286_0 -> 16 in the cache */
   /* replace286_1 -> 866 in the cache */
   /* replace286 -> 1021 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24) */
   /* search287_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1023 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1024, 1025 },
      -1,
   } },

   /* replace287_0 -> 16 in the cache */
   /* replace287_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 1025 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24) */
   /* search288_0_0 -> 16 in the cache */
   /* search288_0_1 -> 1025 in the cache */
   /* search288_0 -> 1027 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1027, 1028 },
      -1,
   } },

   /* replace288_0 -> 16 in the cache */
   /* replace288_1 -> 1025 in the cache */
   /* replace288 -> 1027 in the cache */

   /* ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24) */
   /* search289_0_0 -> 16 in the cache */
   /* search289_0_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 1025 },
      -1,
   } },
   /* search289_1 -> 1023 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1030, 1023 },
      -1,
   } },

   /* replace289_0 -> 16 in the cache */
   /* replace289_1 -> 1025 in the cache */
   /* replace289 -> 1030 in the cache */

   /* ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c')) */
   /* search290_0_0 -> 0 in the cache */
   /* search290_0_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 834 },
      -1,
   } },
   /* search290_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1032, 319 },
      -1,
   } },

   /* replace290_0_0 -> 0 in the cache */
   /* replace290_0_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 141 },
      -1,
   } },
   /* replace290_1_0 -> 3 in the cache */
   /* replace290_1_1 -> 141 in the cache */
   /* replace290_1 -> 837 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1034, 837 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c')) */
   /* search291_0_0 -> 132 in the cache */
   /* search291_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 132, 293 },
      0,
   } },
   /* search291_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 1036, 827 },
      -1,
   } },

   /* replace291_0_0 -> 132 in the cache */
   /* replace291_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace291_1_0 -> 133 in the cache */
   /* replace291_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 1038, 1039 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd')) */
   /* search292_0_0_0 -> 155 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 155, 1041 },
      0,
   } },
   /* search292_0_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 2,
      { 1042, 158 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 1043, 1044 },
      -1,
   } },

   /* replace292_0_0_0 -> 155 in the cache */
   /* replace292_0_0_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 155, 158 },
      -1,
   } },
   /* replace292_0_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 2,
      { 1046, 159 },
      -1,
   } },
   /* replace292_1_0 -> 156 in the cache */
   /* replace292_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 1047, 1048 },
      -1,
   } },

   /* ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd')) */
   /* search293_0_0_0 -> 132 in the cache */
   /* search293_0_0_1 -> 293 in the cache */
   /* search293_0_0 -> 1036 in the cache */
   /* search293_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1036, 135 },
      0,
   } },
   /* search293_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 1050, 321 },
      -1,
   } },

   /* replace293_0_0_0 -> 132 in the cache */
   /* replace293_0_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace293_0_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1052, 169 },
      -1,
   } },
   /* replace293_1_0 -> 133 in the cache */
   /* replace293_1_1 -> 169 in the cache */
   /* replace293_1 -> 830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1053, 830 },
      -1,
   } },

   /* ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1055, 1056 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1057 },
      -1,
   } },

   /* replace294_0 -> 0 in the cache */
   /* replace294_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b') */
   /* search295_0_0 -> 1055 in the cache */
   /* search295_0_1 -> 1056 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1055, 1056 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1060 },
      -1,
   } },

   /* replace295_0 -> 0 in the cache */
   /* replace295_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b') */
   /* search296_0_0 -> 0 in the cache */
   /* search296_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1063 },
      -1,
   } },

   /* replace296_0 -> 0 in the cache */
   /* replace296_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b') */
   /* search297_0_0 -> 0 in the cache */
   /* search297_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1066 },
      -1,
   } },

   /* replace297_0 -> 0 in the cache */
   /* replace297_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b') */
   /* search298_0_0 -> 0 in the cache */
   /* search298_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1069 },
      -1,
   } },

   /* replace298_0 -> 0 in the cache */
   /* replace298_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b') */
   /* search299_0_0 -> 0 in the cache */
   /* search299_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1072 },
      -1,
   } },

   /* replace299_0 -> 0 in the cache */
   /* replace299_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b') */
   /* search300_0_0 -> 0 in the cache */
   /* search300_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1075 },
      -1,
   } },

   /* replace300_0 -> 0 in the cache */
   /* replace300_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b') */
   /* search301_0_0 -> 0 in the cache */
   /* search301_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1078 },
      -1,
   } },

   /* replace301_0 -> 0 in the cache */
   /* replace301_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b') */
   /* search302_0_0 -> 0 in the cache */
   /* search302_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1081 },
      -1,
   } },

   /* replace302_0 -> 0 in the cache */
   /* replace302_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b') */
   /* search303_0_0 -> 0 in the cache */
   /* search303_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1084 },
      -1,
   } },

   /* replace303_0 -> 0 in the cache */
   /* replace303_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False */
   /* search304_0_0 -> 0 in the cache */
   /* search304_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search304_1_0 -> 0 in the cache */
   /* search304_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1087, 1088 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_FALSE /* False */ },
   } },

   /* ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False */
   /* search305_0_0 -> 0 in the cache */
   /* search305_0_1 -> 3 in the cache */
   /* search305_0 -> 1062 in the cache */
   /* search305_1_0 -> 3 in the cache */
   /* search305_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1062, 1091 },
      -1,
   } },

   /* replace305 -> 1090 in the cache */

   /* ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False */
   /* search306_0_0 -> 0 in the cache */
   /* search306_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search306_1_0 -> 0 in the cache */
   /* search306_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1093, 1094 },
      -1,
   } },

   /* replace306 -> 1090 in the cache */

   /* ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False */
   /* search307_0_0 -> 0 in the cache */
   /* search307_0_1 -> 3 in the cache */
   /* search307_0 -> 1077 in the cache */
   /* search307_1_0 -> 3 in the cache */
   /* search307_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1077, 1096 },
      -1,
   } },

   /* replace307 -> 1090 in the cache */

   /* ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False */
   /* search308_0_0 -> 0 in the cache */
   /* search308_0_1 -> 3 in the cache */
   /* search308_0 -> 1080 in the cache */
   /* search308_1_0 -> 3 in the cache */
   /* search308_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1080, 1098 },
      -1,
   } },

   /* replace308 -> 1090 in the cache */

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search309_0_0 -> 0 in the cache */
   /* search309_0 -> 223 in the cache */
   /* search309_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 223, 1100 },
      -1,
   } },

   /* replace309_0 -> 3 in the cache */
   /* replace309_1 -> 0 in the cache */
   /* replace309 -> 1091 in the cache */

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search310_0_0 -> 0 in the cache */
   /* search310_0 -> 223 in the cache */
   /* search310_1_0 -> 3 in the cache */
   /* search310_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 223, 1100 },
      -1,
   } },

   /* replace310_0 -> 3 in the cache */
   /* replace310_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search311_0_0 -> 0 in the cache */
   /* search311_0 -> 223 in the cache */
   /* search311_1_0 -> 3 in the cache */
   /* search311_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 223, 1100 },
      -1,
   } },

   /* replace311_0 -> 3 in the cache */
   /* replace311_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search312_0_0 -> 0 in the cache */
   /* search312_0 -> 223 in the cache */
   /* search312_1_0 -> 3 in the cache */
   /* search312_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 223, 1100 },
      -1,
   } },

   /* replace312_0 -> 3 in the cache */
   /* replace312_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search313_0_0 -> 32 in the cache */
   /* search313_0 -> 107 in the cache */
   /* search313_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 107, 258 },
      -1,
   } },

   /* replace313_0 -> 252 in the cache */
   /* replace313_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 252, 32 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search314_0 -> 258 in the cache */
   /* search314_1_0 -> 32 in the cache */
   /* search314_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 258, 107 },
      -1,
   } },

   /* replace314_0 -> 32 in the cache */
   /* replace314_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 252 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search315_0_0 -> 32 in the cache */
   /* search315_0 -> 107 in the cache */
   /* search315_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 107, 258 },
      -1,
   } },

   /* replace315_0 -> 252 in the cache */
   /* replace315_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 252, 32 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search316_0 -> 258 in the cache */
   /* search316_1_0 -> 32 in the cache */
   /* search316_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 258, 107 },
      -1,
   } },

   /* replace316_0 -> 32 in the cache */
   /* replace316_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 252 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search317_0_0 -> 32 in the cache */
   /* search317_0 -> 107 in the cache */
   /* search317_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 107, 258 },
      -1,
   } },

   /* replace317_0 -> 252 in the cache */
   /* replace317_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 252, 32 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search318_0 -> 258 in the cache */
   /* search318_1_0 -> 32 in the cache */
   /* search318_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 258, 107 },
      -1,
   } },

   /* replace318_0 -> 32 in the cache */
   /* replace318_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 252 },
      -1,
   } },

   /* ('ieq', ('ineg', 'a'), 0) => ('ieq', 'a', 0) */
   /* search319_0_0 -> 32 in the cache */
   /* search319_0 -> 109 in the cache */
   /* search319_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 109, 61 },
      -1,
   } },

   /* replace319_0 -> 32 in the cache */
   /* replace319_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('ine', ('ineg', 'a'), 0) => ('ine', 'a', 0) */
   /* search320_0_0 -> 32 in the cache */
   /* search320_0 -> 109 in the cache */
   /* search320_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 109, 61 },
      -1,
   } },

   /* replace320_0 -> 32 in the cache */
   /* replace320_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('ieq', ('iabs', 'a'), 0) => ('ieq', 'a', 0) */
   /* search321_0_0 -> 32 in the cache */
   /* search321_0 -> 114 in the cache */
   /* search321_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 114, 61 },
      -1,
   } },

   /* replace321_0 -> 32 in the cache */
   /* replace321_1 -> 61 in the cache */
   /* replace321 -> 1121 in the cache */

   /* ('ine', ('iabs', 'a'), 0) => ('ine', 'a', 0) */
   /* search322_0_0 -> 32 in the cache */
   /* search322_0 -> 114 in the cache */
   /* search322_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 114, 61 },
      -1,
   } },

   /* replace322_0 -> 32 in the cache */
   /* replace322_1 -> 61 in the cache */
   /* replace322 -> 1123 in the cache */

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1127 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1126, 1128 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace323_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search324_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1132, 1133 },
      -1,
   } },

   /* replace324_0 -> 0 in the cache */
   /* replace324_1 -> 3 in the cache */
   /* replace324 -> 1059 in the cache */

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search325_0_0 -> 0 in the cache */
   /* search325_0 -> 1132 in the cache */
   /* search325_1 -> 1133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1132, 1133 },
      -1,
   } },

   /* replace325_0 -> 0 in the cache */
   /* replace325_1 -> 3 in the cache */
   /* replace325 -> 1068 in the cache */

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search326_0_0 -> 0 in the cache */
   /* search326_0 -> 1132 in the cache */
   /* search326_1 -> 1133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1132, 1133 },
      -1,
   } },

   /* replace326_0 -> 0 in the cache */
   /* replace326_1 -> 3 in the cache */
   /* replace326 -> 1065 in the cache */

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search327_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      0,
   } },
   /* search327_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1137, 252 },
      -1,
   } },

   /* replace327_0 -> 32 in the cache */
   /* replace327_1 -> 252 in the cache */
   /* replace327 -> 1115 in the cache */

   /* ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a') */
   /* search328_0 -> 117 in the cache */
   /* search328_1_0 -> 32 in the cache */
   /* search328_1 -> 1137 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 1137 },
      -1,
   } },

   /* replace328_0 -> 117 in the cache */
   /* replace328_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 32 },
      -1,
   } },

   /* ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a') */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search329_1_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1141, 1142 },
      -1,
   } },

   /* replace329_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a') */
   /* search330_0_0_0 -> 770 in the cache */
   /* search330_0_0 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1142 },
      -1,
   } },
   /* search330_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1145, 1141 },
      -1,
   } },

   /* replace330_0 -> 770 in the cache */
   /* replace330 -> 1144 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search331_0_0_0 -> 770 in the cache */
   /* search331_0_0 -> 1142 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1142, 1148 },
      -1,
   } },
   /* search331_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1149, 1141 },
      -1,
   } },

   /* replace331_0 -> 770 in the cache */
   /* replace331_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 770, 1147 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search332_0_0 -> 770 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search332_0_2_0 -> 1147 in the cache */
   /* search332_0_2 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1152, 1148 },
      -1,
   } },
   /* search332_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1153, 1141 },
      -1,
   } },

   /* replace332_0 -> 770 in the cache */
   /* replace332_1 -> 1147 in the cache */
   /* replace332 -> 1151 in the cache */

   /* ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search333_0_0 -> 770 in the cache */
   /* search333_0 -> 1142 in the cache */
   /* search333_1_0_0 -> 1147 in the cache */
   /* search333_1_0 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1148 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1142, 1155 },
      -1,
   } },

   /* replace333_0 -> 770 in the cache */
   /* replace333_1 -> 1147 in the cache */
   /* replace333 -> 1151 in the cache */

   /* ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b') */
   /* search334_0_0_0 -> 770 in the cache */
   /* search334_0_0 -> 1142 in the cache */
   /* search334_0_1_0 -> 1147 in the cache */
   /* search334_0_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 1142, 1148 },
      -1,
   } },
   /* search334_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1157, 1141 },
      -1,
   } },

   /* replace334_0 -> 770 in the cache */
   /* replace334_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 770, 1147 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b') */
   /* search335_0_0 -> 770 in the cache */
   /* search335_0_1_0 -> 1147 in the cache */
   /* search335_0_1 -> 1148 in the cache */
   /* search335_0_2 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1148, 1141 },
      -1,
   } },
   /* search335_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1160, 1141 },
      -1,
   } },

   /* replace335_0 -> 770 in the cache */
   /* replace335_1 -> 1147 in the cache */
   /* replace335 -> 1159 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b') */
   /* search336_0_0_0 -> 770 in the cache */
   /* search336_0_0 -> 1142 in the cache */
   /* search336_0_1_0_0 -> 1147 in the cache */
   /* search336_0_1_0 -> 1148 in the cache */
   /* search336_0_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1142, 1155 },
      -1,
   } },
   /* search336_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1162, 1141 },
      -1,
   } },

   /* replace336_0 -> 770 in the cache */
   /* replace336_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 770, 1147 },
      -1,
   } },

   /* ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b') */
   /* search337_0_0 -> 770 in the cache */
   /* search337_0 -> 1142 in the cache */
   /* search337_1_0 -> 1147 in the cache */
   /* search337_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1142, 1148 },
      -1,
   } },

   /* replace337_0 -> 770 in the cache */
   /* replace337_1 -> 1147 in the cache */
   /* replace337 -> 1164 in the cache */

   /* ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b') */
   /* search338_0_0_0 -> 770 in the cache */
   /* search338_0_0 -> 1142 in the cache */
   /* search338_0 -> 1145 in the cache */
   /* search338_1_0_0 -> 1147 in the cache */
   /* search338_1_0 -> 1148 in the cache */
   /* search338_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1145, 1155 },
      -1,
   } },

   /* replace338_0 -> 770 in the cache */
   /* replace338_1 -> 1147 in the cache */
   /* replace338 -> 1164 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search339_0_0_0 -> 770 in the cache */
   /* search339_0_0 -> 1142 in the cache */
   /* search339_0_1_0 -> 1147 in the cache */
   /* search339_0_1 -> 1148 in the cache */
   /* search339_0 -> 1149 in the cache */
   /* search339_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1149, 1141 },
      -1,
   } },

   /* replace339_0_0 -> 770 in the cache */
   /* replace339_0_1 -> 1147 in the cache */
   /* replace339_0 -> 1151 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1151 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search340_0_0 -> 770 in the cache */
   /* search340_0_1 -> 1152 in the cache */
   /* search340_0_2_0 -> 1147 in the cache */
   /* search340_0_2 -> 1148 in the cache */
   /* search340_0 -> 1153 in the cache */
   /* search340_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1153, 1141 },
      -1,
   } },

   /* replace340_0_0 -> 770 in the cache */
   /* replace340_0_1 -> 1147 in the cache */
   /* replace340_0 -> 1151 in the cache */
   /* replace340 -> 1168 in the cache */

   /* ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search341_0_0 -> 770 in the cache */
   /* search341_0 -> 1142 in the cache */
   /* search341_1_0_0 -> 1147 in the cache */
   /* search341_1_0 -> 1148 in the cache */
   /* search341_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1142, 1155 },
      -1,
   } },

   /* replace341_0_0 -> 770 in the cache */
   /* replace341_0_1 -> 1147 in the cache */
   /* replace341_0 -> 1151 in the cache */
   /* replace341 -> 1168 in the cache */

   /* ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search342_0_0_0 -> 770 in the cache */
   /* search342_0_0 -> 1142 in the cache */
   /* search342_0_1_0 -> 1147 in the cache */
   /* search342_0_1 -> 1148 in the cache */
   /* search342_0 -> 1157 in the cache */
   /* search342_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1157, 1141 },
      -1,
   } },

   /* replace342_0_0 -> 770 in the cache */
   /* replace342_0_1 -> 1147 in the cache */
   /* replace342_0 -> 1159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1159 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search343_0_0 -> 770 in the cache */
   /* search343_0_1_0 -> 1147 in the cache */
   /* search343_0_1 -> 1148 in the cache */
   /* search343_0_2 -> 1141 in the cache */
   /* search343_0 -> 1160 in the cache */
   /* search343_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1160, 1141 },
      -1,
   } },

   /* replace343_0_0 -> 770 in the cache */
   /* replace343_0_1 -> 1147 in the cache */
   /* replace343_0 -> 1159 in the cache */
   /* replace343 -> 1172 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b') */
   /* search344_0_0_0 -> 770 in the cache */
   /* search344_0_0 -> 1142 in the cache */
   /* search344_0_1_0_0 -> 1147 in the cache */
   /* search344_0_1_0 -> 1148 in the cache */
   /* search344_0_1 -> 1155 in the cache */
   /* search344_0 -> 1162 in the cache */
   /* search344_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1162, 1141 },
      -1,
   } },

   /* replace344_0 -> 770 in the cache */
   /* replace344_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 770, 1147 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b') */
   /* search345_0_0 -> 770 in the cache */
   /* search345_0 -> 1142 in the cache */
   /* search345_1_0 -> 1147 in the cache */
   /* search345_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1142, 1148 },
      -1,
   } },

   /* replace345_0 -> 770 in the cache */
   /* replace345_1 -> 1147 in the cache */
   /* replace345 -> 1175 in the cache */

   /* ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b') */
   /* search346_0_0_0 -> 770 in the cache */
   /* search346_0_0 -> 1142 in the cache */
   /* search346_0 -> 1145 in the cache */
   /* search346_1_0_0 -> 1147 in the cache */
   /* search346_1_0 -> 1148 in the cache */
   /* search346_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1145, 1155 },
      -1,
   } },

   /* replace346_0 -> 770 in the cache */
   /* replace346_1 -> 1147 in the cache */
   /* replace346 -> 1175 in the cache */

   /* ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b') */
   /* search347_0_0_0_0 -> 770 in the cache */
   /* search347_0_0_0 -> 1142 in the cache */
   /* search347_0_0_1_0 -> 1147 in the cache */
   /* search347_0_0_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1142, 1148 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1178 },
      -1,
   } },
   /* search347_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1179, 1141 },
      -1,
   } },

   /* replace347_0 -> 770 in the cache */
   /* replace347_1 -> 1147 in the cache */
   /* replace347 -> 1151 in the cache */

   /* ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search348_0 -> 1141 in the cache */
   /* search348_1_0_0 -> 770 in the cache */
   /* search348_1_0 -> 1142 in the cache */
   /* search348_1_1_0 -> 1147 in the cache */
   /* search348_1_1 -> 1148 in the cache */
   /* search348_1 -> 1178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1141, 1178 },
      -1,
   } },

   /* replace348_0 -> 770 in the cache */
   /* replace348_1 -> 1147 in the cache */
   /* replace348 -> 1151 in the cache */

   /* ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search349_0_0_0_0 -> 770 in the cache */
   /* search349_0_0_0 -> 1142 in the cache */
   /* search349_0_0_1_0 -> 1147 in the cache */
   /* search349_0_0_1 -> 1148 in the cache */
   /* search349_0_0 -> 1178 in the cache */
   /* search349_0 -> 1179 in the cache */
   /* search349_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1179, 1141 },
      -1,
   } },

   /* replace349_0_0 -> 770 in the cache */
   /* replace349_0_1 -> 1147 in the cache */
   /* replace349_0 -> 1151 in the cache */
   /* replace349 -> 1168 in the cache */

   /* ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search350_0 -> 1141 in the cache */
   /* search350_1_0_0 -> 770 in the cache */
   /* search350_1_0 -> 1142 in the cache */
   /* search350_1_1_0 -> 1147 in the cache */
   /* search350_1_1 -> 1148 in the cache */
   /* search350_1 -> 1178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1141, 1178 },
      -1,
   } },

   /* replace350_0_0 -> 770 in the cache */
   /* replace350_0_1 -> 1147 in the cache */
   /* replace350_0 -> 1151 in the cache */
   /* replace350 -> 1168 in the cache */

   /* ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0) */
   /* search351_0 -> 32 in the cache */
   /* search351_1_0 -> 32 in the cache */
   /* search351_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 107 },
      -1,
   } },

   /* replace351_0 -> 32 in the cache */
   /* replace351_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 117 },
      -1,
   } },

   /* ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0) */
   /* search352_0 -> 32 in the cache */
   /* search352_1_0 -> 32 in the cache */
   /* search352_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 107 },
      -1,
   } },

   /* replace352_0 -> 32 in the cache */
   /* replace352_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 117 },
      -1,
   } },

   /* ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1189 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1191 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1190, 1192 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1193 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1188, 1194 },
      -1,
   } },
   /* search353_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 1195, 117 },
      -1,
   } },

   /* replace353_0_0 -> 1188 in the cache */
   /* replace353_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1188, 117 },
      -1,
   } },
   /* replace353_1_0 -> 1189 in the cache */
   /* replace353_1_1 -> 1191 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1189, 1191 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1197, 1198 },
      -1,
   } },

   /* ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0) */
   /* search354_0_0 -> 0 in the cache */
   /* search354_0_1 -> 3 in the cache */
   /* search354_0 -> 282 in the cache */
   /* search354_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 282, 0 },
      -1,
   } },

   /* replace354_0 -> 3 in the cache */
   /* replace354_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 266 },
      -1,
   } },

   /* ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0) */
   /* search355_0_0 -> 0 in the cache */
   /* search355_0_1 -> 3 in the cache */
   /* search355_0 -> 282 in the cache */
   /* search355_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 282, 0 },
      -1,
   } },

   /* replace355_0 -> 3 in the cache */
   /* replace355_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 266 },
      -1,
   } },

   /* ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0) */
   /* search356_0_0 -> 0 in the cache */
   /* search356_0_1 -> 3 in the cache */
   /* search356_0 -> 224 in the cache */
   /* search356_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 224, 0 },
      -1,
   } },

   /* replace356_0 -> 3 in the cache */
   /* replace356_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 266 },
      -1,
   } },

   /* ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0) */
   /* search357_0_0 -> 0 in the cache */
   /* search357_0_1 -> 3 in the cache */
   /* search357_0 -> 224 in the cache */
   /* search357_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 224, 0 },
      -1,
   } },

   /* replace357_0 -> 3 in the cache */
   /* replace357_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 266 },
      -1,
   } },

   /* ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search358_0_0 -> 132 in the cache */
   /* search358_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 132, 293 },
      0,
   } },
   /* search358_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1208, 827 },
      -1,
   } },

   /* replace358_0 -> 132 in the cache */
   /* replace358_1_0 -> 135 in the cache */
   /* replace358_1_1_0 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 133 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 135, 1210 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 132, 1211 },
      -1,
   } },

   /* ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search359_0_0_0 -> 132 in the cache */
   /* search359_0_0_1 -> 293 in the cache */
   /* search359_0_0 -> 1208 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1208 },
      0,
   } },
   /* search359_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1213, 827 },
      -1,
   } },

   /* replace359_0_0_0 -> 135 in the cache */
   /* replace359_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 135, 133 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1215 },
      -1,
   } },
   /* replace359_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1216, 132 },
      -1,
   } },

   /* ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search360_0_0 -> 132 in the cache */
   /* search360_0_1 -> 293 in the cache */
   /* search360_0 -> 1208 in the cache */
   /* search360_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1208, 827 },
      -1,
   } },

   /* replace360_0 -> 132 in the cache */
   /* replace360_1_0 -> 135 in the cache */
   /* replace360_1_1_0 -> 133 in the cache */
   /* replace360_1_1 -> 1210 in the cache */
   /* replace360_1 -> 1211 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 132, 1211 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search361_0_0_0 -> 132 in the cache */
   /* search361_0_0_1 -> 293 in the cache */
   /* search361_0_0 -> 1208 in the cache */
   /* search361_0 -> 1213 in the cache */
   /* search361_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1213, 827 },
      -1,
   } },

   /* replace361_0_0_0 -> 135 in the cache */
   /* replace361_0_0_1 -> 133 in the cache */
   /* replace361_0_0 -> 1215 in the cache */
   /* replace361_0 -> 1216 in the cache */
   /* replace361_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1216, 132 },
      -1,
   } },

   /* ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search362_0_0 -> 132 in the cache */
   /* search362_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 132, 293 },
      0,
   } },
   /* search362_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1222, 827 },
      -1,
   } },

   /* replace362_0 -> 132 in the cache */
   /* replace362_1_0 -> 135 in the cache */
   /* replace362_1_1_0 -> 133 in the cache */
   /* replace362_1_1 -> 1210 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 135, 1210 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 132, 1224 },
      -1,
   } },

   /* ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search363_0_0_0 -> 132 in the cache */
   /* search363_0_0_1 -> 293 in the cache */
   /* search363_0_0 -> 1222 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1222 },
      0,
   } },
   /* search363_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1226, 827 },
      -1,
   } },

   /* replace363_0_0_0 -> 135 in the cache */
   /* replace363_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 135, 133 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1228 },
      -1,
   } },
   /* replace363_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1229, 132 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search364_0_0 -> 132 in the cache */
   /* search364_0_1 -> 293 in the cache */
   /* search364_0 -> 1222 in the cache */
   /* search364_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1222, 827 },
      -1,
   } },

   /* replace364_0 -> 132 in the cache */
   /* replace364_1_0 -> 135 in the cache */
   /* replace364_1_1_0 -> 133 in the cache */
   /* replace364_1_1 -> 1210 in the cache */
   /* replace364_1 -> 1224 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 132, 1224 },
      -1,
   } },

   /* ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search365_0_0_0 -> 132 in the cache */
   /* search365_0_0_1 -> 293 in the cache */
   /* search365_0_0 -> 1222 in the cache */
   /* search365_0 -> 1226 in the cache */
   /* search365_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1226, 827 },
      -1,
   } },

   /* replace365_0_0_0 -> 135 in the cache */
   /* replace365_0_0_1 -> 133 in the cache */
   /* replace365_0_0 -> 1228 in the cache */
   /* replace365_0 -> 1229 in the cache */
   /* replace365_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1229, 132 },
      -1,
   } },

   /* ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0) */
   /* search366_0_0 -> 0 in the cache */
   /* search366_0_1 -> 3 in the cache */
   /* search366_0 -> 219 in the cache */
   /* search366_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 219, 0 },
      -1,
   } },

   /* replace366_0 -> 3 in the cache */
   /* replace366_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0) */
   /* search367_0_0 -> 0 in the cache */
   /* search367_0_1 -> 3 in the cache */
   /* search367_0 -> 219 in the cache */
   /* search367_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 219, 0 },
      -1,
   } },

   /* replace367_0 -> 3 in the cache */
   /* replace367_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a') */
   /* search368_0_0 -> 770 in the cache */
   /* search368_0 -> 1142 in the cache */
   /* search368_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1142, 1141 },
      -1,
   } },

   /* replace368_0 -> 770 in the cache */
   /* replace368 -> 1144 in the cache */

   /* ('fneu', ('b2f', 'a@1'), 0.0) => a */
   /* search369_0_0 -> 770 in the cache */
   /* search369_0 -> 1142 in the cache */
   /* search369_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1142, 1141 },
      -1,
   } },

   /* replace369 -> 770 in the cache */

   /* ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a') */
   /* search370_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 770 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1241, 1242 },
      -1,
   } },

   /* replace370_0 -> 770 in the cache */
   /* replace370 -> 1144 in the cache */

   /* ('ine', ('b2i', 'a@1'), 0) => a */
   /* search371_0_0 -> 770 in the cache */
   /* search371_0 -> 1241 in the cache */
   /* search371_1 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1241, 1242 },
      -1,
   } },

   /* replace371 -> 770 in the cache */

   /* ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search372_0_0 -> 32 in the cache */
   /* search372_0 -> 112 in the cache */
   /* search372_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 112, 1141 },
      -1,
   } },

   /* replace372_0 -> 32 in the cache */
   /* replace372_1 -> 61 in the cache */
   /* replace372 -> 1123 in the cache */

   /* ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search373_0_0 -> 32 in the cache */
   /* search373_0 -> 112 in the cache */
   /* search373_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 112, 1141 },
      -1,
   } },

   /* replace373_0 -> 32 in the cache */
   /* replace373_1 -> 61 in the cache */
   /* replace373 -> 1121 in the cache */

   /* ('fge', ('u2f', 'a'), 0.0) => True */
   /* search374_0_0 -> 32 in the cache */
   /* search374_0 -> 112 in the cache */
   /* search374_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 112, 1141 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_TRUE /* True */ },
   } },

   /* ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a') */
   /* search375_0 -> 1141 in the cache */
   /* search375_1_0 -> 32 in the cache */
   /* search375_1 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1141, 112 },
      -1,
   } },

   /* replace375_0 -> 61 in the cache */
   /* replace375_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('u2f', 'a'), 0.0) => False */
   /* search376_0_0 -> 32 in the cache */
   /* search376_0 -> 112 in the cache */
   /* search376_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 112, 1141 },
      -1,
   } },

   /* replace376 -> 1090 in the cache */

   /* ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a') */
   /* search377_0 -> 1141 in the cache */
   /* search377_1_0 -> 32 in the cache */
   /* search377_1 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1141, 112 },
      -1,
   } },

   /* replace377_0 -> 61 in the cache */
   /* replace377_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search378_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search378_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1254, 1141 },
      -1,
   } },

   /* replace378_0 -> 32 in the cache */
   /* replace378_1 -> 61 in the cache */
   /* replace378 -> 1123 in the cache */

   /* ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search379_0_0 -> 32 in the cache */
   /* search379_0 -> 1254 in the cache */
   /* search379_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1254, 1141 },
      -1,
   } },

   /* replace379_0 -> 32 in the cache */
   /* replace379_1 -> 61 in the cache */
   /* replace379 -> 1121 in the cache */

   /* ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0) */
   /* search380_0_0 -> 32 in the cache */
   /* search380_0 -> 1254 in the cache */
   /* search380_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1254, 1141 },
      -1,
   } },

   /* replace380_0 -> 32 in the cache */
   /* replace380_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a') */
   /* search381_0 -> 1141 in the cache */
   /* search381_1_0 -> 32 in the cache */
   /* search381_1 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1141, 1254 },
      -1,
   } },

   /* replace381_0 -> 61 in the cache */
   /* replace381_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0) */
   /* search382_0_0 -> 32 in the cache */
   /* search382_0 -> 1254 in the cache */
   /* search382_1 -> 1141 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1254, 1141 },
      -1,
   } },

   /* replace382_0 -> 32 in the cache */
   /* replace382_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a') */
   /* search383_0 -> 1141 in the cache */
   /* search383_1_0 -> 32 in the cache */
   /* search383_1 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1141, 1254 },
      -1,
   } },

   /* replace383_0 -> 61 in the cache */
   /* replace383_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0) */
   /* search384_0 -> 117 in the cache */
   /* search384_1_0 -> 32 in the cache */
   /* search384_1 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 99 },
      -1,
   } },

   /* replace384_0 -> 32 in the cache */
   /* replace384_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0) */
   /* search385_0_0_0 -> 32 in the cache */
   /* search385_0_0 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 99 },
      -1,
   } },
   /* search385_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1267, 117 },
      -1,
   } },

   /* replace385_0 -> 32 in the cache */
   /* replace385_1 -> 117 in the cache */
   /* replace385 -> 1266 in the cache */

   /* ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0) */
   /* search386_0 -> 117 in the cache */
   /* search386_1_0 -> 32 in the cache */
   /* search386_1 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 117, 99 },
      -1,
   } },

   /* replace386_0 -> 32 in the cache */
   /* replace386_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 117 },
      -1,
   } },

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search387_0_0_0 -> 32 in the cache */
   /* search387_0_0 -> 99 in the cache */
   /* search387_0 -> 1267 in the cache */
   /* search387_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1267, 117 },
      -1,
   } },

   /* replace387_0 -> 32 in the cache */
   /* replace387_1 -> 117 in the cache */
   /* replace387 -> 1270 in the cache */

   /* ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a')) */
   /* search388_0_0 -> 32 in the cache */
   /* search388_0_1 -> 117 in the cache */
   /* search388_0 -> 1187 in the cache */
   /* search388_1_0 -> 252 in the cache */
   /* search388_1_1 -> 32 in the cache */
   /* search388_1 -> 1113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1187, 1113 },
      -1,
   } },

   /* replace388_0 -> 32 in the cache */
   /* replace388_1_0 -> 32 in the cache */
   /* replace388_1 -> 228 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 228 },
      -1,
   } },

   /* ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b')) */
   /* search389_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      0,
   } },
   /* search389_1_0 -> 1147 in the cache */
   /* search389_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1274, 1148 },
      -1,
   } },

   /* replace389_0_0 -> 770 in the cache */
   /* replace389_0_1 -> 1147 in the cache */
   /* replace389_0 -> 1151 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1151 },
      -1,
   } },

   /* ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b'))) */
   /* search390_0_0_0 -> 770 in the cache */
   /* search390_0_0 -> 1274 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1274 },
      0,
   } },
   /* search390_1_0_0 -> 1147 in the cache */
   /* search390_1_0 -> 1148 in the cache */
   /* search390_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1277, 1155 },
      -1,
   } },

   /* replace390_0_0_0 -> 770 in the cache */
   /* replace390_0_0_1 -> 1147 in the cache */
   /* replace390_0_0 -> 1159 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1279 },
      -1,
   } },

   /* ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search391_0_0 -> 770 in the cache */
   /* search391_0 -> 1274 in the cache */
   /* search391_1_0 -> 1147 in the cache */
   /* search391_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1274, 1148 },
      -1,
   } },

   /* replace391_0_0 -> 770 in the cache */
   /* replace391_0_1 -> 1147 in the cache */
   /* replace391_0 -> 1159 in the cache */
   /* replace391 -> 1279 in the cache */

   /* ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b'))) */
   /* search392_0_0_0 -> 770 in the cache */
   /* search392_0_0 -> 1274 in the cache */
   /* search392_0 -> 1277 in the cache */
   /* search392_1_0_0 -> 1147 in the cache */
   /* search392_1_0 -> 1148 in the cache */
   /* search392_1 -> 1155 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1277, 1155 },
      -1,
   } },

   /* replace392_0_0_0 -> 770 in the cache */
   /* replace392_0_0_1 -> 1147 in the cache */
   /* replace392_0_0 -> 1151 in the cache */
   /* replace392_0 -> 1276 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1276 },
      -1,
   } },

   /* ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0)) */
   /* search393_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search393_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1284, 834 },
      -1,
   } },

   /* replace393_0 -> 770 in the cache */
   /* replace393_1_0 -> 3 in the cache */
   /* replace393_1_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 3, 280 },
      -1,
   } },
   /* replace393_2_0 -> 3 in the cache */
   /* replace393_2_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 3, 266 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 1286, 1287 },
      -1,
   } },

   /* ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b') */
   /* search394_0_0 -> 0 in the cache */
   /* search394_0_1_0 -> 3 in the cache */
   /* search394_0_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1100 },
      0,
   } },
   /* search394_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1289, 266 },
      -1,
   } },

   /* replace394_0 -> 0 in the cache */
   /* replace394_1 -> 3 in the cache */
   /* replace394 -> 1062 in the cache */

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search395_0_0_0 -> 32 in the cache */
   /* search395_0_0 -> 99 in the cache */
   /* search395_0 -> 1267 in the cache */
   /* search395_1 -> 117 in the cache */
   /* search395 -> 1271 in the cache */

   /* replace395_0 -> 32 in the cache */
   /* replace395_1 -> 117 in the cache */
   /* replace395 -> 1270 in the cache */

   /* ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search396_0_0 -> 1130 in the cache */
   /* search396_0_1 -> 1127 in the cache */
   /* search396_0 -> 1131 in the cache */
   /* search396_1 -> 1130 in the cache */
   /* search396_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 1130, 1127 },
      -1,
   } },

   /* replace396_0 -> 1127 in the cache */
   /* replace396_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search397_0_0 -> 0 in the cache */
   /* search397_0_1 -> 3 in the cache */
   /* search397_0 -> 1062 in the cache */
   /* search397_1 -> 3 in the cache */
   /* search397_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1062, 3, 0 },
      -1,
   } },

   /* replace397_0 -> 0 in the cache */
   /* replace397_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search398_0_0 -> 0 in the cache */
   /* search398_0_1 -> 3 in the cache */
   /* search398_0 -> 1059 in the cache */
   /* search398_1 -> 3 in the cache */
   /* search398_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1059, 3, 0 },
      -1,
   } },

   /* replace398_0 -> 0 in the cache */
   /* replace398_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search399_0_0 -> 1130 in the cache */
   /* search399_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },
   /* search399_1 -> 1130 in the cache */
   /* search399_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1297, 1130, 1127 },
      -1,
   } },

   /* replace399_0 -> 1127 in the cache */
   /* replace399_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('bcsel', ('ult', 'b', 'a'), 'b', 'a') => ('umin', 'a', 'b') */
   /* search400_0_0 -> 1130 in the cache */
   /* search400_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },
   /* search400_1 -> 1130 in the cache */
   /* search400_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1300, 1130, 1127 },
      -1,
   } },

   /* replace400_0 -> 1127 in the cache */
   /* replace400_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('bcsel', ('ult', 'a', 'b'), 'b', 'a') => ('umax', 'a', 'b') */
   /* search401_0_0 -> 0 in the cache */
   /* search401_0_1 -> 3 in the cache */
   /* search401_0 -> 1080 in the cache */
   /* search401_1 -> 3 in the cache */
   /* search401_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1080, 3, 0 },
      -1,
   } },

   /* replace401_0 -> 0 in the cache */
   /* replace401_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), 'b', 'a') => ('umin', 'a', 'b') */
   /* search402_0_0 -> 0 in the cache */
   /* search402_0_1 -> 3 in the cache */
   /* search402_0 -> 1074 in the cache */
   /* search402_1 -> 3 in the cache */
   /* search402_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1074, 3, 0 },
      -1,
   } },

   /* replace402_0 -> 0 in the cache */
   /* replace402_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('uge', 'b', 'a'), 'b', 'a') => ('umax', 'a', 'b') */
   /* search403_0_0 -> 1130 in the cache */
   /* search403_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },
   /* search403_1 -> 1130 in the cache */
   /* search403_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1307, 1130, 1127 },
      -1,
   } },

   /* replace403_0 -> 1127 in the cache */
   /* replace403_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'b', 'a'), 'b', 'a') => ('imin', 'a', 'b') */
   /* search404_0_0 -> 1130 in the cache */
   /* search404_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },
   /* search404_1 -> 1130 in the cache */
   /* search404_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1310, 1130, 1127 },
      -1,
   } },

   /* replace404_0 -> 1127 in the cache */
   /* replace404_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a', 'b'), 'b', 'a') => ('imax', 'a', 'b') */
   /* search405_0_0 -> 0 in the cache */
   /* search405_0_1 -> 3 in the cache */
   /* search405_0 -> 1077 in the cache */
   /* search405_1 -> 3 in the cache */
   /* search405_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1077, 3, 0 },
      -1,
   } },

   /* replace405_0 -> 0 in the cache */
   /* replace405_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('ige', 'a', 'b'), 'b', 'a') => ('imin', 'a', 'b') */
   /* search406_0_0 -> 0 in the cache */
   /* search406_0_1 -> 3 in the cache */
   /* search406_0 -> 1071 in the cache */
   /* search406_1 -> 3 in the cache */
   /* search406_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1071, 3, 0 },
      -1,
   } },

   /* replace406_0 -> 0 in the cache */
   /* replace406_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('ige', 'b', 'a'), 'b', 'a') => ('imax', 'a', 'b') */
   /* search407_0_0 -> 1130 in the cache */
   /* search407_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1130, 1127 },
      -1,
   } },
   /* search407_1 -> 1130 in the cache */
   /* search407_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1317, 1130, 1127 },
      -1,
   } },

   /* replace407_0 -> 1127 in the cache */
   /* replace407_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 1127, 1130 },
      -1,
   } },

   /* ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b') */
   /* search408_0_0 -> 770 in the cache */
   /* search408_0 -> 1144 in the cache */
   /* search408_1 -> 133 in the cache */
   /* search408_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1144, 133, 135 },
      -1,
   } },

   /* replace408_0 -> 770 in the cache */
   /* replace408_1 -> 135 in the cache */
   /* replace408_2 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 135, 133 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd') */
   /* search409_0 -> 770 in the cache */
   /* search409_1_0 -> 770 in the cache */
   /* search409_1_1 -> 156 in the cache */
   /* search409_1_2 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 156, 158 },
      -1,
   } },
   /* search409_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1322, 159 },
      -1,
   } },

   /* replace409_0 -> 770 in the cache */
   /* replace409_1 -> 156 in the cache */
   /* replace409_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 156, 159 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd') */
   /* search410_0 -> 770 in the cache */
   /* search410_1 -> 156 in the cache */
   /* search410_2_0 -> 770 in the cache */
   /* search410_2_1 -> 158 in the cache */
   /* search410_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 156, 1325 },
      -1,
   } },

   /* replace410_0 -> 770 in the cache */
   /* replace410_1 -> 156 in the cache */
   /* replace410_2 -> 159 in the cache */
   /* replace410 -> 1324 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search411_0 -> 770 in the cache */
   /* search411_1_0 -> 1147 in the cache */
   /* search411_1_1 -> 763 in the cache */
   /* search411_1_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 763, 772 },
      -1,
   } },
   /* search411_2_0 -> 1147 in the cache */
   /* search411_2_1 -> 763 in the cache */
   /* search411_2_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 763, 774 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1327, 1328 },
      -1,
   } },

   /* replace411_0 -> 1147 in the cache */
   /* replace411_1 -> 763 in the cache */
   /* replace411_2_0 -> 770 in the cache */
   /* replace411_2_1 -> 772 in the cache */
   /* replace411_2_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 772, 774 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 763, 1330 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search412_0 -> 770 in the cache */
   /* search412_1_0 -> 1147 in the cache */
   /* search412_1_1 -> 763 in the cache */
   /* search412_1_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 763, 772 },
      0,
   } },
   /* search412_2_0 -> 1147 in the cache */
   /* search412_2_1 -> 763 in the cache */
   /* search412_2_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 763, 774 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1332, 1333 },
      -1,
   } },

   /* replace412_0 -> 1147 in the cache */
   /* replace412_1 -> 763 in the cache */
   /* replace412_2_0 -> 770 in the cache */
   /* replace412_2_1 -> 772 in the cache */
   /* replace412_2_2 -> 774 in the cache */
   /* replace412_2 -> 1330 in the cache */
   /* replace412 -> 1331 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search413_0 -> 770 in the cache */
   /* search413_1_0 -> 1147 in the cache */
   /* search413_1_1 -> 158 in the cache */
   /* search413_1_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 158, 159 },
      -1,
   } },
   /* search413_2_0 -> 1147 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search413_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 1336, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1335, 1337 },
      -1,
   } },

   /* replace413_0 -> 1147 in the cache */
   /* replace413_1_0 -> 770 in the cache */
   /* replace413_1_1 -> 158 in the cache */
   /* replace413_1_2 -> 1336 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 158, 1336 },
      -1,
   } },
   /* replace413_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 1339, 159 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search414_0 -> 770 in the cache */
   /* search414_1_0 -> 1147 in the cache */
   /* search414_1_1 -> 158 in the cache */
   /* search414_1_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 158, 159 },
      0,
   } },
   /* search414_2_0 -> 1147 in the cache */
   /* search414_2_1 -> 1336 in the cache */
   /* search414_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1147, 1336, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1341, 1342 },
      -1,
   } },

   /* replace414_0 -> 1147 in the cache */
   /* replace414_1_0 -> 770 in the cache */
   /* replace414_1_1 -> 158 in the cache */
   /* replace414_1_2 -> 1336 in the cache */
   /* replace414_1 -> 1339 in the cache */
   /* replace414_2 -> 159 in the cache */
   /* replace414 -> 1340 in the cache */

   /* ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b') */
   /* search415_0 -> 770 in the cache */
   /* search415_1 -> 1248 in the cache */
   /* search415_2 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1248, 1147 },
      -1,
   } },

   /* replace415_0 -> 770 in the cache */
   /* replace415_1 -> 1147 in the cache */
   /* replace415 -> 1151 in the cache */

   /* ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b') */
   /* search416_0 -> 770 in the cache */
   /* search416_1 -> 770 in the cache */
   /* search416_2 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 770, 1147 },
      -1,
   } },

   /* replace416_0 -> 770 in the cache */
   /* replace416_1 -> 1147 in the cache */
   /* replace416 -> 1151 in the cache */

   /* ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b') */
   /* search417_0 -> 770 in the cache */
   /* search417_1 -> 1147 in the cache */
   /* search417_2 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1147, 1090 },
      -1,
   } },

   /* replace417_0 -> 770 in the cache */
   /* replace417_1 -> 1147 in the cache */
   /* replace417 -> 1159 in the cache */

   /* ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b') */
   /* search418_0 -> 770 in the cache */
   /* search418_1 -> 1147 in the cache */
   /* search418_2 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1147, 770 },
      -1,
   } },

   /* replace418_0 -> 770 in the cache */
   /* replace418_1 -> 1147 in the cache */
   /* replace418 -> 1159 in the cache */

   /* ('~fmin', 'a', 'a') => a */
   /* search419_0 -> 32 in the cache */
   /* search419_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace419 -> 32 in the cache */

   /* ('~fmax', 'a', 'a') => a */
   /* search420_0 -> 32 in the cache */
   /* search420_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace420 -> 32 in the cache */

   /* ('imin', 'a', 'a') => a */
   /* search421_0 -> 32 in the cache */
   /* search421_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace421 -> 32 in the cache */

   /* ('imax', 'a', 'a') => a */
   /* search422_0 -> 32 in the cache */
   /* search422_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace422 -> 32 in the cache */

   /* ('umin', 'a', 'a') => a */
   /* search423_0 -> 32 in the cache */
   /* search423_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace423 -> 32 in the cache */

   /* ('umin', 'a', 0) => 0 */
   /* search424_0 -> 32 in the cache */
   /* search424_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace424 -> 61 in the cache */

   /* ('umin', 'a', -1) => a */
   /* search425_0 -> 32 in the cache */
   /* search425_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace425 -> 32 in the cache */

   /* ('umax', 'a', 'a') => a */
   /* search426_0 -> 32 in the cache */
   /* search426_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace426 -> 32 in the cache */

   /* ('umax', 'a', 0) => a */
   /* search427_0 -> 32 in the cache */
   /* search427_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace427 -> 32 in the cache */

   /* ('umax', 'a', -1) => -1 */
   /* search428_0 -> 32 in the cache */
   /* search428_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace428 -> 63 in the cache */

   /* ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b') */
   /* search429_0_0 -> 0 in the cache */
   /* search429_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search429_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1358, 3 },
      -1,
   } },

   /* replace429_0 -> 0 in the cache */
   /* replace429_1 -> 3 in the cache */
   /* replace429 -> 1294 in the cache */

   /* ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b') */
   /* search430_0_0 -> 0 in the cache */
   /* search430_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search430_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1360, 3 },
      -1,
   } },

   /* replace430_0 -> 0 in the cache */
   /* replace430_1 -> 3 in the cache */
   /* replace430 -> 1304 in the cache */

   /* ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b') */
   /* search431_0_0 -> 0 in the cache */
   /* search431_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search431_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1362, 3 },
      -1,
   } },

   /* replace431_0 -> 0 in the cache */
   /* replace431_1 -> 3 in the cache */
   /* replace431 -> 1314 in the cache */

   /* ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b') */
   /* search432_0_0 -> 0 in the cache */
   /* search432_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search432_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1364, 3 },
      -1,
   } },

   /* replace432_0 -> 0 in the cache */
   /* replace432_1 -> 3 in the cache */
   /* replace432 -> 1296 in the cache */

   /* ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b') */
   /* search433_0_0 -> 0 in the cache */
   /* search433_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search433_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1366, 3 },
      -1,
   } },

   /* replace433_0 -> 0 in the cache */
   /* replace433_1 -> 3 in the cache */
   /* replace433 -> 1306 in the cache */

   /* ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b') */
   /* search434_0_0 -> 0 in the cache */
   /* search434_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search434_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1368, 3 },
      -1,
   } },

   /* replace434_0 -> 0 in the cache */
   /* replace434_1 -> 3 in the cache */
   /* replace434 -> 1316 in the cache */

   /* ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c') */
   /* search435_0_0_0 -> 132 in the cache */
   /* search435_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search435_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 2,
      { 1370, 135 },
      -1,
   } },
   /* search435_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 1371, 132 },
      -1,
   } },

   /* replace435_0_0 -> 132 in the cache */
   /* replace435_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace435_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1373, 135 },
      -1,
   } },

   /* ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c') */
   /* search436_0_0_0 -> 132 in the cache */
   /* search436_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search436_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 2,
      { 1375, 135 },
      -1,
   } },
   /* search436_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 3,
      { 1376, 132 },
      -1,
   } },

   /* replace436_0_0 -> 132 in the cache */
   /* replace436_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace436_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1378, 135 },
      -1,
   } },

   /* ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c') */
   /* search437_0_0_0 -> 132 in the cache */
   /* search437_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search437_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 2,
      { 1380, 135 },
      -1,
   } },
   /* search437_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 3,
      { 1381, 132 },
      -1,
   } },

   /* replace437_0_0 -> 132 in the cache */
   /* replace437_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace437_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1383, 135 },
      -1,
   } },

   /* ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c') */
   /* search438_0_0_0 -> 132 in the cache */
   /* search438_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search438_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 1385, 135 },
      -1,
   } },
   /* search438_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 1386, 132 },
      -1,
   } },

   /* replace438_0_0 -> 132 in the cache */
   /* replace438_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace438_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1388, 135 },
      -1,
   } },

   /* ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c') */
   /* search439_0_0_0 -> 132 in the cache */
   /* search439_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search439_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 2,
      { 1390, 135 },
      -1,
   } },
   /* search439_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 1391, 132 },
      -1,
   } },

   /* replace439_0_0 -> 132 in the cache */
   /* replace439_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace439_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1393, 135 },
      -1,
   } },

   /* ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c') */
   /* search440_0_0_0 -> 132 in the cache */
   /* search440_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 1,
      { 132, 133 },
      -1,
   } },
   /* search440_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 2,
      { 1395, 135 },
      -1,
   } },
   /* search440_1 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 3,
      { 1396, 132 },
      -1,
   } },

   /* replace440_0_0 -> 132 in the cache */
   /* replace440_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 132, 133 },
      -1,
   } },
   /* replace440_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1398, 135 },
      -1,
   } },

   /* ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b') */
   /* search441_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search441_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1400, 1401 },
      -1,
   } },

   /* replace441_0 -> 770 in the cache */
   /* replace441_1 -> 1147 in the cache */
   /* replace441 -> 1175 in the cache */

   /* ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b') */
   /* search442_0_0 -> 770 in the cache */
   /* search442_0 -> 1400 in the cache */
   /* search442_1_0 -> 1147 in the cache */
   /* search442_1 -> 1401 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1400, 1401 },
      -1,
   } },

   /* replace442_0 -> 770 in the cache */
   /* replace442_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 770, 1147 },
      -1,
   } },

   /* ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search443_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search443_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1405, 1406 },
      -1,
   } },

   /* replace443_0 -> 770 in the cache */
   /* replace443_1 -> 1147 in the cache */
   /* replace443 -> 1175 in the cache */

   /* ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b') */
   /* search444_0_0 -> 770 in the cache */
   /* search444_0 -> 1405 in the cache */
   /* search444_1_0 -> 1147 in the cache */
   /* search444_1 -> 1406 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1405, 1406 },
      -1,
   } },

   /* replace444_0 -> 770 in the cache */
   /* replace444_1 -> 1147 in the cache */
   /* replace444 -> 1404 in the cache */

   /* ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search445_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search445_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1409, 1410 },
      -1,
   } },

   /* replace445_0 -> 770 in the cache */
   /* replace445_1 -> 1147 in the cache */
   /* replace445 -> 1175 in the cache */

   /* ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b') */
   /* search446_0_0 -> 770 in the cache */
   /* search446_0 -> 1409 in the cache */
   /* search446_1_0 -> 1147 in the cache */
   /* search446_1 -> 1410 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1409, 1410 },
      -1,
   } },

   /* replace446_0 -> 770 in the cache */
   /* replace446_1 -> 1147 in the cache */
   /* replace446 -> 1404 in the cache */

   /* ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search447_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search447_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1413, 1414 },
      -1,
   } },

   /* replace447_0 -> 770 in the cache */
   /* replace447_1 -> 1147 in the cache */
   /* replace447 -> 1175 in the cache */

   /* ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b') */
   /* search448_0_0 -> 770 in the cache */
   /* search448_0 -> 1413 in the cache */
   /* search448_1_0 -> 1147 in the cache */
   /* search448_1 -> 1414 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1413, 1414 },
      -1,
   } },

   /* replace448_0 -> 770 in the cache */
   /* replace448_1 -> 1147 in the cache */
   /* replace448 -> 1404 in the cache */

   /* ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search449_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search449_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1417, 1418 },
      -1,
   } },

   /* replace449_0 -> 770 in the cache */
   /* replace449_1 -> 1147 in the cache */
   /* replace449 -> 1175 in the cache */

   /* ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b') */
   /* search450_0_0 -> 770 in the cache */
   /* search450_0 -> 1417 in the cache */
   /* search450_1_0 -> 1147 in the cache */
   /* search450_1 -> 1418 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1417, 1418 },
      -1,
   } },

   /* replace450_0 -> 770 in the cache */
   /* replace450_1 -> 1147 in the cache */
   /* replace450 -> 1404 in the cache */

   /* ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search451_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search451_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1421, 1422 },
      -1,
   } },

   /* replace451_0 -> 770 in the cache */
   /* replace451_1 -> 1147 in the cache */
   /* replace451 -> 1175 in the cache */

   /* ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b') */
   /* search452_0_0 -> 770 in the cache */
   /* search452_0 -> 1421 in the cache */
   /* search452_1_0 -> 1147 in the cache */
   /* search452_1 -> 1422 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1421, 1422 },
      -1,
   } },

   /* replace452_0 -> 770 in the cache */
   /* replace452_1 -> 1147 in the cache */
   /* replace452 -> 1404 in the cache */

   /* ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search453_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 770 },
      -1,
   } },
   /* search453_1_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1425, 1426 },
      -1,
   } },

   /* replace453_0 -> 770 in the cache */
   /* replace453_1 -> 1147 in the cache */
   /* replace453 -> 1175 in the cache */

   /* ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b') */
   /* search454_0_0 -> 770 in the cache */
   /* search454_0 -> 1425 in the cache */
   /* search454_1_0 -> 1147 in the cache */
   /* search454_1 -> 1426 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1425, 1426 },
      -1,
   } },

   /* replace454_0 -> 770 in the cache */
   /* replace454_1 -> 1147 in the cache */
   /* replace454 -> 1404 in the cache */

   /* ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0) */
   /* search455_0 -> 6 in the cache */
   /* search455_1_0_0 -> 6 in the cache */
   /* search455_1_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1429 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 1430 },
      -1,
   } },

   /* replace455_0 -> 6 in the cache */
   /* replace455_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 6, 848 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0) */
   /* search456_0_0 -> 6 in the cache */
   /* search456_0_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 6, 7 },
      -1,
   } },
   /* search456_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1433, 848 },
      -1,
   } },

   /* replace456_0 -> 6 in the cache */
   /* replace456_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 848 },
      -1,
   } },

   /* ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0) */
   /* search457_0_0 -> 6 in the cache */
   /* search457_0_1 -> 7 in the cache */
   /* search457_0 -> 1433 in the cache */
   /* search457_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1433, 7 },
      -1,
   } },

   /* replace457_0 -> 6 in the cache */
   /* replace457_1 -> 848 in the cache */
   /* replace457 -> 1435 in the cache */

   /* ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0) */
   /* search458_0_0 -> 6 in the cache */
   /* search458_0_1 -> 7 in the cache */
   /* search458_0 -> 1433 in the cache */
   /* search458_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1433, 848 },
      -1,
   } },

   /* replace458_0 -> 6 in the cache */
   /* replace458_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 848 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0) */
   /* search459_0_0 -> 6 in the cache */
   /* search459_0_1 -> 7 in the cache */
   /* search459_0 -> 1433 in the cache */
   /* search459_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1433, 7 },
      -1,
   } },

   /* replace459_0 -> 6 in the cache */
   /* replace459_1 -> 848 in the cache */
   /* replace459 -> 1438 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search460_0_0 -> 6 in the cache */
   /* search460_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   /* search460_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1440, 848 },
      -1,
   } },

   /* replace460_0 -> 6 in the cache */
   /* replace460_1 -> 848 in the cache */
   /* replace460 -> 1438 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search461_0_0 -> 6 in the cache */
   /* search461_0_1 -> 9 in the cache */
   /* search461_0 -> 1440 in the cache */
   /* search461_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1440, 848 },
      -1,
   } },

   /* replace461_0 -> 6 in the cache */
   /* replace461_1 -> 848 in the cache */
   /* replace461 -> 1435 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0) */
   /* search462_0_0 -> 6 in the cache */
   /* search462_0_1 -> 9 in the cache */
   /* search462_0 -> 1440 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1440, 1443 },
      -1,
   } },

   /* replace462_0 -> 6 in the cache */
   /* replace462_1 -> 848 in the cache */
   /* replace462 -> 1438 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0) */
   /* search463_0_0 -> 6 in the cache */
   /* search463_0_1 -> 9 in the cache */
   /* search463_0 -> 1440 in the cache */
   /* search463_1 -> 1443 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1440, 1443 },
      -1,
   } },

   /* replace463_0 -> 6 in the cache */
   /* replace463_1 -> 848 in the cache */
   /* replace463 -> 1435 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search464_0_0 -> 6 in the cache */
   /* search464_0_1 -> 9 in the cache */
   /* search464_0 -> 1429 in the cache */
   /* search464_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1429, 848 },
      -1,
   } },

   /* replace464_0 -> 6 in the cache */
   /* replace464_1 -> 848 in the cache */
   /* replace464 -> 1438 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search465_0_0 -> 6 in the cache */
   /* search465_0_1 -> 9 in the cache */
   /* search465_0 -> 1429 in the cache */
   /* search465_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1429, 848 },
      -1,
   } },

   /* replace465_0 -> 6 in the cache */
   /* replace465_1 -> 848 in the cache */
   /* replace465 -> 1435 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0) */
   /* search466_0_0 -> 6 in the cache */
   /* search466_0_1 -> 9 in the cache */
   /* search466_0 -> 1429 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1429, 1448 },
      -1,
   } },

   /* replace466_0 -> 6 in the cache */
   /* replace466_1 -> 848 in the cache */
   /* replace466 -> 1438 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0) */
   /* search467_0_0 -> 6 in the cache */
   /* search467_0_1 -> 9 in the cache */
   /* search467_0 -> 1429 in the cache */
   /* search467_1 -> 1448 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1429, 1448 },
      -1,
   } },

   /* replace467_0 -> 6 in the cache */
   /* replace467_1 -> 848 in the cache */
   /* replace467 -> 1435 in the cache */

   /* ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0) */
   /* search468_0 -> 11 in the cache */
   /* search468_1_0_0 -> 11 in the cache */
   /* search468_1_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1451 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 1452 },
      -1,
   } },

   /* replace468_0 -> 11 in the cache */
   /* replace468_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 11, 869 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0) */
   /* search469_0_0 -> 11 in the cache */
   /* search469_0_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 11, 12 },
      -1,
   } },
   /* search469_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1455, 869 },
      -1,
   } },

   /* replace469_0 -> 11 in the cache */
   /* replace469_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 869 },
      -1,
   } },

   /* ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0) */
   /* search470_0_0 -> 11 in the cache */
   /* search470_0_1 -> 12 in the cache */
   /* search470_0 -> 1455 in the cache */
   /* search470_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1455, 12 },
      -1,
   } },

   /* replace470_0 -> 11 in the cache */
   /* replace470_1 -> 869 in the cache */
   /* replace470 -> 1457 in the cache */

   /* ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0) */
   /* search471_0_0 -> 11 in the cache */
   /* search471_0_1 -> 12 in the cache */
   /* search471_0 -> 1455 in the cache */
   /* search471_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1455, 869 },
      -1,
   } },

   /* replace471_0 -> 11 in the cache */
   /* replace471_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 869 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0) */
   /* search472_0_0 -> 11 in the cache */
   /* search472_0_1 -> 12 in the cache */
   /* search472_0 -> 1455 in the cache */
   /* search472_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1455, 12 },
      -1,
   } },

   /* replace472_0 -> 11 in the cache */
   /* replace472_1 -> 869 in the cache */
   /* replace472 -> 1460 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search473_0_0 -> 11 in the cache */
   /* search473_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   /* search473_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1462, 869 },
      -1,
   } },

   /* replace473_0 -> 11 in the cache */
   /* replace473_1 -> 869 in the cache */
   /* replace473 -> 1460 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search474_0_0 -> 11 in the cache */
   /* search474_0_1 -> 14 in the cache */
   /* search474_0 -> 1462 in the cache */
   /* search474_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1462, 869 },
      -1,
   } },

   /* replace474_0 -> 11 in the cache */
   /* replace474_1 -> 869 in the cache */
   /* replace474 -> 1457 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0) */
   /* search475_0_0 -> 11 in the cache */
   /* search475_0_1 -> 14 in the cache */
   /* search475_0 -> 1462 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1462, 1465 },
      -1,
   } },

   /* replace475_0 -> 11 in the cache */
   /* replace475_1 -> 869 in the cache */
   /* replace475 -> 1460 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0) */
   /* search476_0_0 -> 11 in the cache */
   /* search476_0_1 -> 14 in the cache */
   /* search476_0 -> 1462 in the cache */
   /* search476_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1462, 1465 },
      -1,
   } },

   /* replace476_0 -> 11 in the cache */
   /* replace476_1 -> 869 in the cache */
   /* replace476 -> 1457 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search477_0_0 -> 11 in the cache */
   /* search477_0_1 -> 14 in the cache */
   /* search477_0 -> 1451 in the cache */
   /* search477_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1451, 869 },
      -1,
   } },

   /* replace477_0 -> 11 in the cache */
   /* replace477_1 -> 869 in the cache */
   /* replace477 -> 1460 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search478_0_0 -> 11 in the cache */
   /* search478_0_1 -> 14 in the cache */
   /* search478_0 -> 1451 in the cache */
   /* search478_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1451, 869 },
      -1,
   } },

   /* replace478_0 -> 11 in the cache */
   /* replace478_1 -> 869 in the cache */
   /* replace478 -> 1457 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0) */
   /* search479_0_0 -> 11 in the cache */
   /* search479_0_1 -> 14 in the cache */
   /* search479_0 -> 1451 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1451, 1470 },
      -1,
   } },

   /* replace479_0 -> 11 in the cache */
   /* replace479_1 -> 869 in the cache */
   /* replace479 -> 1460 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0) */
   /* search480_0_0 -> 11 in the cache */
   /* search480_0_1 -> 14 in the cache */
   /* search480_0 -> 1451 in the cache */
   /* search480_1 -> 1470 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1451, 1470 },
      -1,
   } },

   /* replace480_0 -> 11 in the cache */
   /* replace480_1 -> 869 in the cache */
   /* replace480 -> 1457 in the cache */

   /* ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0) */
   /* search481_0 -> 16 in the cache */
   /* search481_1_0_0 -> 16 in the cache */
   /* search481_1_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1474 },
      -1,
   } },

   /* replace481_0 -> 16 in the cache */
   /* replace481_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 16, 128 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0) */
   /* search482_0_0 -> 16 in the cache */
   /* search482_0_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 17 },
      -1,
   } },
   /* search482_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1477, 128 },
      -1,
   } },

   /* replace482_0 -> 16 in the cache */
   /* replace482_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 128 },
      -1,
   } },

   /* ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0) */
   /* search483_0_0 -> 16 in the cache */
   /* search483_0_1 -> 17 in the cache */
   /* search483_0 -> 1477 in the cache */
   /* search483_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1477, 17 },
      -1,
   } },

   /* replace483_0 -> 16 in the cache */
   /* replace483_1 -> 128 in the cache */
   /* replace483 -> 1479 in the cache */

   /* ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0) */
   /* search484_0_0 -> 16 in the cache */
   /* search484_0_1 -> 17 in the cache */
   /* search484_0 -> 1477 in the cache */
   /* search484_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1477, 128 },
      -1,
   } },

   /* replace484_0 -> 16 in the cache */
   /* replace484_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 128 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0) */
   /* search485_0_0 -> 16 in the cache */
   /* search485_0_1 -> 17 in the cache */
   /* search485_0 -> 1477 in the cache */
   /* search485_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1477, 17 },
      -1,
   } },

   /* replace485_0 -> 16 in the cache */
   /* replace485_1 -> 128 in the cache */
   /* replace485 -> 1482 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search486_0_0 -> 16 in the cache */
   /* search486_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   /* search486_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1484, 128 },
      -1,
   } },

   /* replace486_0 -> 16 in the cache */
   /* replace486_1 -> 128 in the cache */
   /* replace486 -> 1482 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search487_0_0 -> 16 in the cache */
   /* search487_0_1 -> 19 in the cache */
   /* search487_0 -> 1484 in the cache */
   /* search487_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1484, 128 },
      -1,
   } },

   /* replace487_0 -> 16 in the cache */
   /* replace487_1 -> 128 in the cache */
   /* replace487 -> 1479 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0) */
   /* search488_0_0 -> 16 in the cache */
   /* search488_0_1 -> 19 in the cache */
   /* search488_0 -> 1484 in the cache */
   /* search488_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1484, 406 },
      -1,
   } },

   /* replace488_0 -> 16 in the cache */
   /* replace488_1 -> 128 in the cache */
   /* replace488 -> 1482 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0) */
   /* search489_0_0 -> 16 in the cache */
   /* search489_0_1 -> 19 in the cache */
   /* search489_0 -> 1484 in the cache */
   /* search489_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1484, 406 },
      -1,
   } },

   /* replace489_0 -> 16 in the cache */
   /* replace489_1 -> 128 in the cache */
   /* replace489 -> 1479 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search490_0_0 -> 16 in the cache */
   /* search490_0_1 -> 19 in the cache */
   /* search490_0 -> 1473 in the cache */
   /* search490_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1473, 128 },
      -1,
   } },

   /* replace490_0 -> 16 in the cache */
   /* replace490_1 -> 128 in the cache */
   /* replace490 -> 1482 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search491_0_0 -> 16 in the cache */
   /* search491_0_1 -> 19 in the cache */
   /* search491_0 -> 1473 in the cache */
   /* search491_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1473, 128 },
      -1,
   } },

   /* replace491_0 -> 16 in the cache */
   /* replace491_1 -> 128 in the cache */
   /* replace491 -> 1479 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0) */
   /* search492_0_0 -> 16 in the cache */
   /* search492_0_1 -> 19 in the cache */
   /* search492_0 -> 1473 in the cache */
   /* search492_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1473, 130 },
      -1,
   } },

   /* replace492_0 -> 16 in the cache */
   /* replace492_1 -> 128 in the cache */
   /* replace492 -> 1482 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0) */
   /* search493_0_0 -> 16 in the cache */
   /* search493_0_1 -> 19 in the cache */
   /* search493_0 -> 1473 in the cache */
   /* search493_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1473, 130 },
      -1,
   } },

   /* replace493_0 -> 16 in the cache */
   /* replace493_1 -> 128 in the cache */
   /* replace493 -> 1479 in the cache */

   /* ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0) */
   /* search494_0 -> 21 in the cache */
   /* search494_1_0_0 -> 21 in the cache */
   /* search494_1_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1493 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 1494 },
      -1,
   } },

   /* replace494_0 -> 21 in the cache */
   /* replace494_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 910 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0) */
   /* search495_0_0 -> 21 in the cache */
   /* search495_0_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 22 },
      -1,
   } },
   /* search495_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1497, 910 },
      -1,
   } },

   /* replace495_0 -> 21 in the cache */
   /* replace495_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 910 },
      -1,
   } },

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0) */
   /* search496_0_0 -> 21 in the cache */
   /* search496_0_1 -> 22 in the cache */
   /* search496_0 -> 1497 in the cache */
   /* search496_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1497, 22 },
      -1,
   } },

   /* replace496_0 -> 21 in the cache */
   /* replace496_1 -> 910 in the cache */
   /* replace496 -> 1499 in the cache */

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0) */
   /* search497_0_0 -> 21 in the cache */
   /* search497_0_1 -> 22 in the cache */
   /* search497_0 -> 1497 in the cache */
   /* search497_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1497, 910 },
      -1,
   } },

   /* replace497_0 -> 21 in the cache */
   /* replace497_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 910 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0) */
   /* search498_0_0 -> 21 in the cache */
   /* search498_0_1 -> 22 in the cache */
   /* search498_0 -> 1497 in the cache */
   /* search498_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1497, 22 },
      -1,
   } },

   /* replace498_0 -> 21 in the cache */
   /* replace498_1 -> 910 in the cache */
   /* replace498 -> 1502 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search499_0_0 -> 21 in the cache */
   /* search499_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   /* search499_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1504, 910 },
      -1,
   } },

   /* replace499_0 -> 21 in the cache */
   /* replace499_1 -> 910 in the cache */
   /* replace499 -> 1502 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search500_0_0 -> 21 in the cache */
   /* search500_0_1 -> 24 in the cache */
   /* search500_0 -> 1504 in the cache */
   /* search500_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1504, 910 },
      -1,
   } },

   /* replace500_0 -> 21 in the cache */
   /* replace500_1 -> 910 in the cache */
   /* replace500 -> 1499 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0) */
   /* search501_0_0 -> 21 in the cache */
   /* search501_0_1 -> 24 in the cache */
   /* search501_0 -> 1504 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1504, 1507 },
      -1,
   } },

   /* replace501_0 -> 21 in the cache */
   /* replace501_1 -> 910 in the cache */
   /* replace501 -> 1502 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0) */
   /* search502_0_0 -> 21 in the cache */
   /* search502_0_1 -> 24 in the cache */
   /* search502_0 -> 1504 in the cache */
   /* search502_1 -> 1507 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1504, 1507 },
      -1,
   } },

   /* replace502_0 -> 21 in the cache */
   /* replace502_1 -> 910 in the cache */
   /* replace502 -> 1499 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search503_0_0 -> 21 in the cache */
   /* search503_0_1 -> 24 in the cache */
   /* search503_0 -> 1493 in the cache */
   /* search503_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1493, 910 },
      -1,
   } },

   /* replace503_0 -> 21 in the cache */
   /* replace503_1 -> 910 in the cache */
   /* replace503 -> 1502 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search504_0_0 -> 21 in the cache */
   /* search504_0_1 -> 24 in the cache */
   /* search504_0 -> 1493 in the cache */
   /* search504_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1493, 910 },
      -1,
   } },

   /* replace504_0 -> 21 in the cache */
   /* replace504_1 -> 910 in the cache */
   /* replace504 -> 1499 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0) */
   /* search505_0_0 -> 21 in the cache */
   /* search505_0_1 -> 24 in the cache */
   /* search505_0 -> 1493 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1493, 1512 },
      -1,
   } },

   /* replace505_0 -> 21 in the cache */
   /* replace505_1 -> 910 in the cache */
   /* replace505 -> 1502 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0) */
   /* search506_0_0 -> 21 in the cache */
   /* search506_0_1 -> 24 in the cache */
   /* search506_0 -> 1493 in the cache */
   /* search506_1 -> 1512 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1493, 1512 },
      -1,
   } },

   /* replace506_0 -> 21 in the cache */
   /* replace506_1 -> 910 in the cache */
   /* replace506 -> 1499 in the cache */

   /* ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a')) */
   /* search507_0 -> 32 in the cache */
   /* search507_1_0 -> 32 in the cache */
   /* search507_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 107 },
      -1,
   } },

   /* replace507_0_0 -> 32 in the cache */
   /* replace507_0 -> 99 in the cache */
   /* replace507 -> 1267 in the cache */

   /* ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a')) */
   /* search508_0 -> 32 in the cache */
   /* search508_1_0 -> 32 in the cache */
   /* search508_1 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 109 },
      -1,
   } },

   /* replace508_0_0 -> 32 in the cache */
   /* replace508_0 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 114 },
      -1,
   } },

   /* ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a')) */
   /* search509_0 -> 32 in the cache */
   /* search509_1_0_0 -> 32 in the cache */
   /* search509_1_0 -> 99 in the cache */
   /* search509_1 -> 1267 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 1267 },
      -1,
   } },

   /* replace509_0_0 -> 32 in the cache */
   /* replace509_0 -> 99 in the cache */
   /* replace509 -> 1267 in the cache */

   /* ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a')) */
   /* search510_0 -> 32 in the cache */
   /* search510_1_0_0 -> 32 in the cache */
   /* search510_1_0 -> 114 in the cache */
   /* search510_1 -> 1517 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 1517 },
      -1,
   } },

   /* replace510_0_0 -> 32 in the cache */
   /* replace510_0 -> 114 in the cache */
   /* replace510 -> 1517 in the cache */

   /* ('~fmin', 'a', ('fabs', 'a')) => a */
   /* search511_0 -> 32 in the cache */
   /* search511_1_0 -> 32 in the cache */
   /* search511_1 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 99 },
      -1,
   } },

   /* replace511 -> 32 in the cache */

   /* ('imin', 'a', ('iabs', 'a')) => a */
   /* search512_0 -> 32 in the cache */
   /* search512_1_0 -> 32 in the cache */
   /* search512_1 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 114 },
      -1,
   } },

   /* replace512 -> 32 in the cache */

   /* ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a */
   /* search513_0 -> 32 in the cache */
   /* search513_1_0_0 -> 32 in the cache */
   /* search513_1_0 -> 99 in the cache */
   /* search513_1 -> 1267 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 1267 },
      -1,
   } },

   /* replace513 -> 32 in the cache */

   /* ('imax', 'a', ('ineg', ('iabs', 'a'))) => a */
   /* search514_0 -> 32 in the cache */
   /* search514_1_0_0 -> 32 in the cache */
   /* search514_1_0 -> 114 in the cache */
   /* search514_1 -> 1517 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 1517 },
      -1,
   } },

   /* replace514 -> 32 in the cache */

   /* ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a') */
   /* search515_0 -> 32 in the cache */
   /* search515_1_0 -> 32 in the cache */
   /* search515_1 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 99 },
      -1,
   } },

   /* replace515_0 -> 32 in the cache */
   /* replace515 -> 99 in the cache */

   /* ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a') */
   /* search516_0 -> 32 in the cache */
   /* search516_1_0 -> 32 in the cache */
   /* search516_1 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 114 },
      -1,
   } },

   /* replace516_0 -> 32 in the cache */
   /* replace516 -> 114 in the cache */

   /* ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a') */
   /* search517_0 -> 32 in the cache */
   /* search517_1_0 -> 32 in the cache */
   /* search517_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 107 },
      -1,
   } },

   /* replace517_0 -> 32 in the cache */
   /* replace517 -> 99 in the cache */

   /* ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a') */
   /* search518_0 -> 32 in the cache */
   /* search518_1_0 -> 32 in the cache */
   /* search518_1 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 109 },
      -1,
   } },

   /* replace518_0 -> 32 in the cache */
   /* replace518 -> 114 in the cache */

   /* ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a') */
   /* search519_0_0 -> 32 in the cache */
   /* search519_0 -> 99 in the cache */
   /* search519_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 99, 117 },
      -1,
   } },

   /* replace519_0 -> 32 in the cache */
   /* replace519 -> 99 in the cache */

   /* ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a') */
   /* search520_0_0 -> 32 in the cache */
   /* search520_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 117 },
      -1,
   } },
   /* search520_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1529, 252 },
      -1,
   } },

   /* replace520_0 -> 32 in the cache */
   /* replace520 -> 228 in the cache */

   /* ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a') */
   /* search521_0_0 -> 32 in the cache */
   /* search521_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 252 },
      -1,
   } },
   /* search521_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1531, 117 },
      -1,
   } },

   /* replace521_0 -> 32 in the cache */
   /* replace521 -> 228 in the cache */

   /* ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search522_0_0 -> 32 in the cache */
   /* search522_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 258 },
      -1,
   } },
   /* search522_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1533, 117 },
      -1,
   } },

   /* replace522_0_0_0 -> 32 in the cache */
   /* replace522_0_0 -> 107 in the cache */
   /* replace522_0 -> 229 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 229 },
      -1,
   } },

   /* ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search523_0_0 -> 32 in the cache */
   /* search523_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 117 },
      -1,
   } },
   /* search523_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1536, 258 },
      -1,
   } },

   /* replace523_0_0_0 -> 32 in the cache */
   /* replace523_0_0 -> 107 in the cache */
   /* replace523_0 -> 229 in the cache */
   /* replace523 -> 1535 in the cache */

   /* ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a')) */
   /* search524_0_0 -> 32 in the cache */
   /* search524_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace524_0_0 -> 117 in the cache */
   /* replace524_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1539 },
      -1,
   } },

   /* ('fsat', ('b2f', 'a')) => ('b2f', 'a') */
   /* search525_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1541 },
      -1,
   } },

   /* replace525_0 -> 32 in the cache */
   /* replace525 -> 1541 in the cache */

   /* ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0) */
   /* search526_0 -> 32 in the cache */
   /* search526 -> 228 in the cache */

   /* replace526_0_0 -> 32 in the cache */
   /* replace526_0_1 -> 117 in the cache */
   /* replace526_0 -> 1529 in the cache */
   /* replace526_1 -> 252 in the cache */
   /* replace526 -> 1530 in the cache */

   /* ('fsat', ('fsat', 'a')) => ('fsat', 'a') */
   /* search527_0_0 -> 32 in the cache */
   /* search527_0 -> 228 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 228 },
      -1,
   } },

   /* replace527_0 -> 32 in the cache */
   /* replace527 -> 228 in the cache */

   /* ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b'))) */
   /* search528_0_0_0 -> 0 in the cache */
   /* search528_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1544 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1545 },
      -1,
   } },

   /* replace528_0_0_0 -> 0 in the cache */
   /* replace528_0_0 -> 223 in the cache */
   /* replace528_0_1_0 -> 3 in the cache */
   /* replace528_0_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 223, 1100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1547 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b')) */
   /* search529_0_0_0 -> 0 in the cache */
   /* search529_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1549 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1550 },
      -1,
   } },

   /* replace529_0_0_0 -> 0 in the cache */
   /* replace529_0_0 -> 223 in the cache */
   /* replace529_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 223, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1552 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b')) */
   /* search530_0_0_0 -> 16 in the cache */
   /* search530_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1554 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1555 },
      -1,
   } },

   /* replace530_0_0_0 -> 16 in the cache */
   /* replace530_0_0 -> 290 in the cache */
   /* replace530_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 290, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1557 },
      -1,
   } },

   /* ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b'))) */
   /* search531_0_0_0 -> 0 in the cache */
   /* search531_0_0_1 -> 3 in the cache */
   /* search531_0_0 -> 1549 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 1549 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1559 },
      -1,
   } },

   /* replace531_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace531_0_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1561, 1562 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1563 },
      -1,
   } },

   /* ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c') */
   /* search532_0_0_0_0 -> 132 in the cache */
   /* search532_0_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      3, 1,
      { 132, 133 },
      -1,
   } },
   /* search532_0_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 2,
      { 1565, 135 },
      -1,
   } },
   /* search532_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 3,
      { 1566, 133 },
      -1,
   } },
   /* search532_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 4,
      { 1567, 135 },
      -1,
   } },

   /* replace532_0_0 -> 132 in the cache */
   /* replace532_0_1 -> 133 in the cache */
   /* replace532_0 -> 1373 in the cache */
   /* replace532_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1373, 135 },
      -1,
   } },

   /* ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c') */
   /* search533_0_0_0_0 -> 132 in the cache */
   /* search533_0_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      3, 1,
      { 132, 133 },
      -1,
   } },
   /* search533_0_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 2,
      { 1570, 135 },
      -1,
   } },
   /* search533_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 3,
      { 1571, 133 },
      -1,
   } },
   /* search533_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 1572, 135 },
      -1,
   } },

   /* replace533_0_0 -> 132 in the cache */
   /* replace533_0_1 -> 133 in the cache */
   /* replace533_0 -> 1383 in the cache */
   /* replace533_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1383, 135 },
      -1,
   } },

   /* ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c') */
   /* search534_0_0_0_0 -> 132 in the cache */
   /* search534_0_0_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      3, 1,
      { 132, 133 },
      -1,
   } },
   /* search534_0_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 2,
      { 1575, 135 },
      -1,
   } },
   /* search534_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 3,
      { 1576, 133 },
      -1,
   } },
   /* search534_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 4,
      { 1577, 135 },
      -1,
   } },

   /* replace534_0_0 -> 132 in the cache */
   /* replace534_0_1 -> 133 in the cache */
   /* replace534_0 -> 1378 in the cache */
   /* replace534_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1378, 135 },
      -1,
   } },

   /* ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b')) */
   /* search535_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      18,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1580, 1581 },
      -1,
   } },

   /* replace535_0_0 -> 0 in the cache */
   /* replace535_0_1 -> 3 in the cache */
   /* replace535_0 -> 1294 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1294 },
      -1,
   } },

   /* ('fmax', ('fsat(is_used_once)', 'a'), ('fsat(is_used_once)', 'b')) => ('fsat', ('fmax', 'a', 'b')) */
   /* search536_0_0 -> 0 in the cache */
   /* search536_0 -> 1132 in the cache */
   /* search536_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1132, 1584 },
      -1,
   } },

   /* replace536_0_0 -> 0 in the cache */
   /* replace536_0_1 -> 3 in the cache */
   /* replace536_0 -> 1294 in the cache */
   /* replace536 -> 1583 in the cache */

   /* ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b')) */
   /* search537_0_0 -> 0 in the cache */
   /* search537_0 -> 1580 in the cache */
   /* search537_1 -> 1581 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1580, 1581 },
      -1,
   } },

   /* replace537_0_0 -> 0 in the cache */
   /* replace537_0_1 -> 3 in the cache */
   /* replace537_0 -> 1296 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1296 },
      -1,
   } },

   /* ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a')) */
   /* search538_0_0_0 -> 1130 in the cache */
   /* search538_0_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 1130, 1127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1588 },
      -1,
   } },
   /* search538_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1589, 1130 },
      -1,
   } },

   /* replace538_0_0 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 1130 },
      -1,
   } },
   /* replace538_1_0 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1591, 1592 },
      -1,
   } },

   /* ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a')) */
   /* search539_0_0_0 -> 1130 in the cache */
   /* search539_0_0_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 1130, 1127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1594 },
      -1,
   } },
   /* search539_1 -> 1130 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1595, 1130 },
      -1,
   } },

   /* replace539_0_0_0 -> 1130 in the cache */
   /* replace539_0_0 -> 1591 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1591 },
      -1,
   } },
   /* replace539_1_0 -> 1127 in the cache */
   /* replace539_1 -> 1592 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1597, 1592 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1599 },
      -1,
   } },
   /* search540_0_1 -> 1581 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1600, 1581 },
      0,
   } },
   /* search540_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1601, 266 },
      -1,
   } },

   /* replace540_0_0_0 -> 0 in the cache */
   /* replace540_0_0 -> 223 in the cache */
   /* replace540_0_1 -> 3 in the cache */
   /* replace540_0 -> 287 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 287 },
      -1,
   } },

   /* ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255) */
   /* search541_0_0_0 -> 32 in the cache */
   /* search541_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 61 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1604, 1605 },
      -1,
   } },
   /* search541_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 2,
      { 1606, 61 },
      -1,
   } },

   /* replace541_0_0 -> 32 in the cache */
   /* replace541_0_1 -> 61 in the cache */
   /* replace541_0 -> 1604 in the cache */
   /* replace541_1 -> 1605 in the cache */
   /* replace541 -> 1606 in the cache */

   /* ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search542_0_0 -> 132 in the cache */
   /* search542_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search542_1_0 -> 132 in the cache */
   /* search542_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1608, 1609 },
      -1,
   } },

   /* replace542_0 -> 132 in the cache */
   /* replace542_1_0 -> 133 in the cache */
   /* replace542_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 132, 1611 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   /* search543_0_0 -> 0 in the cache */
   /* search543_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 694 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search543_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1613, 1615 },
      -1,
   } },

   /* replace543_0_0 -> 0 in the cache */
   /* replace543_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace543_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1617, 694 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search544_0_0 -> 132 in the cache */
   /* search544_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search544_1_0 -> 132 in the cache */
   /* search544_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1619, 1620 },
      -1,
   } },

   /* replace544_0 -> 132 in the cache */
   /* replace544_1_0 -> 133 in the cache */
   /* replace544_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 132, 1622 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search545_0_0 -> 0 in the cache */
   /* search545_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 694 },
      0,
   } },
   /* search545_1_0 -> 1614 in the cache */
   /* search545_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1624, 1625 },
      -1,
   } },

   /* replace545_0_0 -> 0 in the cache */
   /* replace545_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace545_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1627, 694 },
      -1,
   } },

   /* ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search546_0_0 -> 132 in the cache */
   /* search546_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 293 },
      -1,
   } },
   /* search546_1_0 -> 132 in the cache */
   /* search546_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1629, 1630 },
      -1,
   } },

   /* replace546_0 -> 132 in the cache */
   /* replace546_1_0 -> 133 in the cache */
   /* replace546_1_1 -> 135 in the cache */
   /* replace546_1 -> 1611 in the cache */
   /* replace546 -> 1612 in the cache */

   /* ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search547_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1632, 694 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search547_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1634, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1633, 1635 },
      -1,
   } },

   /* replace547_0_0 -> 0 in the cache */
   /* replace547_0_1 -> 1614 in the cache */
   /* replace547_0 -> 1617 in the cache */
   /* replace547_1 -> 694 in the cache */
   /* replace547 -> 1618 in the cache */

   /* ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search548_0_0 -> 132 in the cache */
   /* search548_0_1 -> 293 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 293 },
      -1,
   } },
   /* search548_1_0 -> 132 in the cache */
   /* search548_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1637, 1638 },
      -1,
   } },

   /* replace548_0 -> 132 in the cache */
   /* replace548_1_0 -> 133 in the cache */
   /* replace548_1_1 -> 135 in the cache */
   /* replace548_1 -> 1622 in the cache */
   /* replace548 -> 1623 in the cache */

   /* ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search549_0_0 -> 1632 in the cache */
   /* search549_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1632, 694 },
      -1,
   } },
   /* search549_1_0 -> 1634 in the cache */
   /* search549_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1634, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1640, 1641 },
      -1,
   } },

   /* replace549_0_0 -> 0 in the cache */
   /* replace549_0_1 -> 1614 in the cache */
   /* replace549_0 -> 1627 in the cache */
   /* replace549_1 -> 694 in the cache */
   /* replace549 -> 1628 in the cache */

   /* ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search550_0_0 -> 132 in the cache */
   /* search550_0_1 -> 133 in the cache */
   /* search550_0 -> 1608 in the cache */
   /* search550_1_0 -> 132 in the cache */
   /* search550_1_1 -> 135 in the cache */
   /* search550_1 -> 1609 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1608, 1609 },
      -1,
   } },

   /* replace550_0 -> 132 in the cache */
   /* replace550_1_0 -> 133 in the cache */
   /* replace550_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 132, 1644 },
      -1,
   } },

   /* ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   /* search551_0_0 -> 0 in the cache */
   /* search551_0_1 -> 694 in the cache */
   /* search551_0 -> 1613 in the cache */
   /* search551_1_0 -> 1614 in the cache */
   /* search551_1_1 -> 694 in the cache */
   /* search551_1 -> 1615 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1613, 1615 },
      -1,
   } },

   /* replace551_0_0 -> 0 in the cache */
   /* replace551_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace551_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1647, 694 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search552_0_0 -> 132 in the cache */
   /* search552_0_1 -> 133 in the cache */
   /* search552_0 -> 1619 in the cache */
   /* search552_1_0 -> 132 in the cache */
   /* search552_1_1 -> 135 in the cache */
   /* search552_1 -> 1620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1619, 1620 },
      -1,
   } },

   /* replace552_0 -> 132 in the cache */
   /* replace552_1_0 -> 133 in the cache */
   /* replace552_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 132, 1650 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search553_0_0 -> 0 in the cache */
   /* search553_0_1 -> 694 in the cache */
   /* search553_0 -> 1624 in the cache */
   /* search553_1_0 -> 1614 in the cache */
   /* search553_1_1 -> 694 in the cache */
   /* search553_1 -> 1625 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1624, 1625 },
      -1,
   } },

   /* replace553_0_0 -> 0 in the cache */
   /* replace553_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace553_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1653, 694 },
      -1,
   } },

   /* ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search554_0_0 -> 132 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 1655 },
      -1,
   } },
   /* search554_1_0 -> 132 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 1657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1656, 1658 },
      -1,
   } },

   /* replace554_0 -> 132 in the cache */
   /* replace554_1_0 -> 133 in the cache */
   /* replace554_1_1 -> 135 in the cache */
   /* replace554_1 -> 1644 in the cache */
   /* replace554 -> 1645 in the cache */

   /* ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search555_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1660, 694 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search555_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1662, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1661, 1663 },
      -1,
   } },

   /* replace555_0_0 -> 0 in the cache */
   /* replace555_0_1 -> 1614 in the cache */
   /* replace555_0 -> 1647 in the cache */
   /* replace555_1 -> 694 in the cache */
   /* replace555 -> 1648 in the cache */

   /* ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search556_0_0 -> 132 in the cache */
   /* search556_0_1 -> 1655 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 1655 },
      -1,
   } },
   /* search556_1_0 -> 132 in the cache */
   /* search556_1_1 -> 1657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 132, 1657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1665, 1666 },
      -1,
   } },

   /* replace556_0 -> 132 in the cache */
   /* replace556_1_0 -> 133 in the cache */
   /* replace556_1_1 -> 135 in the cache */
   /* replace556_1 -> 1650 in the cache */
   /* replace556 -> 1651 in the cache */

   /* ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search557_0_0 -> 1660 in the cache */
   /* search557_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1660, 694 },
      -1,
   } },
   /* search557_1_0 -> 1662 in the cache */
   /* search557_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1662, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1668, 1669 },
      -1,
   } },

   /* replace557_0_0 -> 0 in the cache */
   /* replace557_0_1 -> 1614 in the cache */
   /* replace557_0 -> 1653 in the cache */
   /* replace557_1 -> 694 in the cache */
   /* replace557 -> 1654 in the cache */

   /* ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c')) */
   /* search558_0_0 -> 132 in the cache */
   /* search558_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search558_1_0 -> 132 in the cache */
   /* search558_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1671, 1672 },
      -1,
   } },

   /* replace558_0 -> 132 in the cache */
   /* replace558_1_0 -> 133 in the cache */
   /* replace558_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 132, 1674 },
      -1,
   } },

   /* ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c') */
   /* search559_0_0 -> 0 in the cache */
   /* search559_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 694 },
      0,
   } },
   /* search559_1_0 -> 1614 in the cache */
   /* search559_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1676, 1677 },
      -1,
   } },

   /* replace559_0_0 -> 0 in the cache */
   /* replace559_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace559_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1679, 694 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c')) */
   /* search560_0_0 -> 132 in the cache */
   /* search560_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search560_1_0 -> 132 in the cache */
   /* search560_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1681, 1682 },
      -1,
   } },

   /* replace560_0 -> 132 in the cache */
   /* replace560_1_0 -> 133 in the cache */
   /* replace560_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 132, 1684 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c') */
   /* search561_0_0 -> 0 in the cache */
   /* search561_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 694 },
      0,
   } },
   /* search561_1_0 -> 1614 in the cache */
   /* search561_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1686, 1687 },
      -1,
   } },

   /* replace561_0_0 -> 0 in the cache */
   /* replace561_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace561_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1689, 694 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c')) */
   /* search562_0_0 -> 132 in the cache */
   /* search562_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search562_1_0 -> 132 in the cache */
   /* search562_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1691, 1692 },
      -1,
   } },

   /* replace562_0 -> 132 in the cache */
   /* replace562_1_0 -> 133 in the cache */
   /* replace562_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 132, 1694 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c') */
   /* search563_0_0 -> 0 in the cache */
   /* search563_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 694 },
      0,
   } },
   /* search563_1_0 -> 1614 in the cache */
   /* search563_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1696, 1697 },
      -1,
   } },

   /* replace563_0_0 -> 0 in the cache */
   /* replace563_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace563_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1699, 694 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c')) */
   /* search564_0_0 -> 132 in the cache */
   /* search564_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 132, 133 },
      0,
   } },
   /* search564_1_0 -> 132 in the cache */
   /* search564_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 132, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1701, 1702 },
      -1,
   } },

   /* replace564_0 -> 132 in the cache */
   /* replace564_1_0 -> 133 in the cache */
   /* replace564_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 132, 1704 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c') */
   /* search565_0_0 -> 0 in the cache */
   /* search565_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 694 },
      0,
   } },
   /* search565_1_0 -> 1614 in the cache */
   /* search565_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1614, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1706, 1707 },
      -1,
   } },

   /* replace565_0_0 -> 0 in the cache */
   /* replace565_0_1 -> 1614 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 1614 },
      -1,
   } },
   /* replace565_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1709, 694 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c')) */
   /* search566_0_0 -> 132 in the cache */
   /* search566_0_1 -> 133 in the cache */
   /* search566_0 -> 1671 in the cache */
   /* search566_1_0 -> 132 in the cache */
   /* search566_1_1 -> 135 in the cache */
   /* search566_1 -> 1672 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1671, 1672 },
      -1,
   } },

   /* replace566_0 -> 132 in the cache */
   /* replace566_1_0 -> 133 in the cache */
   /* replace566_1_1 -> 135 in the cache */
   /* replace566_1 -> 1684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 132, 1684 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c') */
   /* search567_0_0 -> 0 in the cache */
   /* search567_0_1 -> 694 in the cache */
   /* search567_0 -> 1676 in the cache */
   /* search567_1_0 -> 1614 in the cache */
   /* search567_1_1 -> 694 in the cache */
   /* search567_1 -> 1677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1676, 1677 },
      -1,
   } },

   /* replace567_0_0 -> 0 in the cache */
   /* replace567_0_1 -> 1614 in the cache */
   /* replace567_0 -> 1689 in the cache */
   /* replace567_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1689, 694 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c')) */
   /* search568_0_0 -> 132 in the cache */
   /* search568_0_1 -> 133 in the cache */
   /* search568_0 -> 1681 in the cache */
   /* search568_1_0 -> 132 in the cache */
   /* search568_1_1 -> 135 in the cache */
   /* search568_1 -> 1682 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1681, 1682 },
      -1,
   } },

   /* replace568_0 -> 132 in the cache */
   /* replace568_1_0 -> 133 in the cache */
   /* replace568_1_1 -> 135 in the cache */
   /* replace568_1 -> 1674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 132, 1674 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c') */
   /* search569_0_0 -> 0 in the cache */
   /* search569_0_1 -> 694 in the cache */
   /* search569_0 -> 1686 in the cache */
   /* search569_1_0 -> 1614 in the cache */
   /* search569_1_1 -> 694 in the cache */
   /* search569_1 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1686, 1687 },
      -1,
   } },

   /* replace569_0_0 -> 0 in the cache */
   /* replace569_0_1 -> 1614 in the cache */
   /* replace569_0 -> 1679 in the cache */
   /* replace569_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1679, 694 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c')) */
   /* search570_0_0 -> 132 in the cache */
   /* search570_0_1 -> 133 in the cache */
   /* search570_0 -> 1691 in the cache */
   /* search570_1_0 -> 132 in the cache */
   /* search570_1_1 -> 135 in the cache */
   /* search570_1 -> 1692 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1691, 1692 },
      -1,
   } },

   /* replace570_0 -> 132 in the cache */
   /* replace570_1_0 -> 133 in the cache */
   /* replace570_1_1 -> 135 in the cache */
   /* replace570_1 -> 1704 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 132, 1704 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c') */
   /* search571_0_0 -> 0 in the cache */
   /* search571_0_1 -> 694 in the cache */
   /* search571_0 -> 1696 in the cache */
   /* search571_1_0 -> 1614 in the cache */
   /* search571_1_1 -> 694 in the cache */
   /* search571_1 -> 1697 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1696, 1697 },
      -1,
   } },

   /* replace571_0_0 -> 0 in the cache */
   /* replace571_0_1 -> 1614 in the cache */
   /* replace571_0 -> 1709 in the cache */
   /* replace571_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1709, 694 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c')) */
   /* search572_0_0 -> 132 in the cache */
   /* search572_0_1 -> 133 in the cache */
   /* search572_0 -> 1701 in the cache */
   /* search572_1_0 -> 132 in the cache */
   /* search572_1_1 -> 135 in the cache */
   /* search572_1 -> 1702 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1701, 1702 },
      -1,
   } },

   /* replace572_0 -> 132 in the cache */
   /* replace572_1_0 -> 133 in the cache */
   /* replace572_1_1 -> 135 in the cache */
   /* replace572_1 -> 1694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 132, 1694 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c') */
   /* search573_0_0 -> 0 in the cache */
   /* search573_0_1 -> 694 in the cache */
   /* search573_0 -> 1706 in the cache */
   /* search573_1_0 -> 1614 in the cache */
   /* search573_1_1 -> 694 in the cache */
   /* search573_1 -> 1707 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1706, 1707 },
      -1,
   } },

   /* replace573_0_0 -> 0 in the cache */
   /* replace573_0_1 -> 1614 in the cache */
   /* replace573_0 -> 1699 in the cache */
   /* replace573_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1699, 694 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search574_0_0_0 -> 0 in the cache */
   /* search574_0_0_1 -> 694 in the cache */
   /* search574_0_0 -> 1613 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1613, 1727 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search574_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1729, 694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1728, 1730 },
      -1,
   } },

   /* replace574_0_0_0 -> 0 in the cache */
   /* replace574_0_0_1 -> 1729 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 1729 },
      -1,
   } },
   /* replace574_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1732, 694 },
      -1,
   } },
   /* replace574_1 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1733, 1727 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search575_0_0_0 -> 0 in the cache */
   /* search575_0_0_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 694 },
      -1,
   } },
   /* search575_0_1 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1735, 1727 },
      0,
   } },
   /* search575_1_0 -> 1729 in the cache */
   /* search575_1_1 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1729, 694 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1736, 1737 },
      -1,
   } },

   /* replace575_0_0_0 -> 0 in the cache */
   /* replace575_0_0_1 -> 1729 in the cache */
   /* replace575_0_0 -> 1732 in the cache */
   /* replace575_0_1 -> 694 in the cache */
   /* replace575_0 -> 1733 in the cache */
   /* replace575_1 -> 1727 in the cache */
   /* replace575 -> 1734 in the cache */

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search576_0_0_0 -> 155 in the cache */
   /* search576_0_0_1 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 155, 156 },
      0,
   } },
   /* search576_0_1 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1739, 1727 },
      0,
   } },
   /* search576_1_0 -> 155 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 155, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1740, 1742 },
      -1,
   } },

   /* replace576_0_0 -> 155 in the cache */
   /* replace576_0_1_0 -> 156 in the cache */
   /* replace576_0_1_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      true,
      false,
      nir_op_fmax,
      1, 1,
      { 156, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 155, 1744 },
      -1,
   } },
   /* replace576_1 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1745, 1727 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search577_0_0_0 -> 155 in the cache */
   /* search577_0_0_1 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 155, 156 },
      -1,
   } },
   /* search577_0_1 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1747, 1727 },
      0,
   } },
   /* search577_1_0 -> 155 in the cache */
   /* search577_1_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 155, 1741 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1748, 1749 },
      -1,
   } },

   /* replace577_0_0 -> 155 in the cache */
   /* replace577_0_1_0 -> 156 in the cache */
   /* replace577_0_1_1 -> 1741 in the cache */
   /* replace577_0_1 -> 1744 in the cache */
   /* replace577_0 -> 1745 in the cache */
   /* replace577_1 -> 1727 in the cache */
   /* replace577 -> 1746 in the cache */

   /* ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b') */
   /* search578_0_0 -> 1055 in the cache */
   /* search578_0_1 -> 1056 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1055, 1056 },
      -1,
   } },
   /* search578_1_0 -> 3 in the cache */
   /* search578_1_1 -> 0 in the cache */
   /* search578_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1751, 1091 },
      -1,
   } },

   /* replace578_0 -> 0 in the cache */
   /* replace578_1 -> 3 in the cache */
   /* replace578 -> 1065 in the cache */

   /* ('umin', ('iand', 'a', '#b(is_pos_power_of_two)'), ('iand', 'c', 'b')) => ('iand', ('iand', 'a', 'b'), ('iand', 'c', 'b')) */
   /* search579_0_0 -> 0 in the cache */
   /* search579_0_1 -> 1 in the cache */
   /* search579_0 -> 205 in the cache */
   /* search579_1_0 -> 146 in the cache */
   /* search579_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 146, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 205, 1753 },
      -1,
   } },

   /* replace579_0_0 -> 0 in the cache */
   /* replace579_0_1 -> 3 in the cache */
   /* replace579_0 -> 207 in the cache */
   /* replace579_1_0 -> 146 in the cache */
   /* replace579_1_1 -> 3 in the cache */
   /* replace579_1 -> 1753 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 207, 1753 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search580_0_0 -> 119 in the cache */
   /* search580_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 119, 11 },
      0,
   } },
   /* search580_1_0 -> 268 in the cache */
   /* search580_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 268, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1756, 1757 },
      -1,
   } },

   /* replace580_0 -> 119 in the cache */
   /* replace580_1_0 -> 11 in the cache */
   /* replace580_1_1_0 -> 268 in the cache */
   /* replace580_1_1 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 11, 544 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 119, 1759 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search581_0_0 -> 119 in the cache */
   /* search581_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 119, 11 },
      -1,
   } },
   /* search581_1_0 -> 268 in the cache */
   /* search581_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 268, 119 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1761, 1762 },
      -1,
   } },

   /* replace581_0 -> 119 in the cache */
   /* replace581_1_0 -> 11 in the cache */
   /* replace581_1_1_0 -> 268 in the cache */
   /* replace581_1_1 -> 544 in the cache */
   /* replace581_1 -> 1759 in the cache */
   /* replace581 -> 1760 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search582_0_0 -> 119 in the cache */
   /* search582_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 119, 11 },
      0,
   } },
   /* search582_1_0 -> 268 in the cache */
   /* search582_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 268, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1764, 1765 },
      -1,
   } },

   /* replace582_0 -> 119 in the cache */
   /* replace582_1_0 -> 11 in the cache */
   /* replace582_1_1_0 -> 268 in the cache */
   /* replace582_1_1 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 11, 544 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 119, 1767 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search583_0_0 -> 119 in the cache */
   /* search583_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 119, 11 },
      -1,
   } },
   /* search583_1_0 -> 268 in the cache */
   /* search583_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 268, 119 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1769, 1770 },
      -1,
   } },

   /* replace583_0 -> 119 in the cache */
   /* replace583_1_0 -> 11 in the cache */
   /* replace583_1_1_0 -> 268 in the cache */
   /* replace583_1_1 -> 544 in the cache */
   /* replace583_1 -> 1767 in the cache */
   /* replace583 -> 1768 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search584_0_0 -> 119 in the cache */
   /* search584_0_1 -> 11 in the cache */
   /* search584_0 -> 1756 in the cache */
   /* search584_1_0 -> 268 in the cache */
   /* search584_1_1 -> 119 in the cache */
   /* search584_1 -> 1757 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1756, 1757 },
      -1,
   } },

   /* replace584_0 -> 119 in the cache */
   /* replace584_1_0 -> 11 in the cache */
   /* replace584_1_1_0 -> 268 in the cache */
   /* replace584_1_1 -> 544 in the cache */
   /* replace584_1 -> 1767 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 119, 1767 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search585_0_0 -> 119 in the cache */
   /* search585_0_1 -> 11 in the cache */
   /* search585_0 -> 1761 in the cache */
   /* search585_1_0 -> 268 in the cache */
   /* search585_1_1 -> 119 in the cache */
   /* search585_1 -> 1762 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1761, 1762 },
      -1,
   } },

   /* replace585_0 -> 119 in the cache */
   /* replace585_1_0 -> 11 in the cache */
   /* replace585_1_1_0 -> 268 in the cache */
   /* replace585_1_1 -> 544 in the cache */
   /* replace585_1 -> 1767 in the cache */
   /* replace585 -> 1773 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search586_0_0 -> 119 in the cache */
   /* search586_0_1 -> 11 in the cache */
   /* search586_0 -> 1764 in the cache */
   /* search586_1_0 -> 268 in the cache */
   /* search586_1_1 -> 119 in the cache */
   /* search586_1 -> 1765 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1764, 1765 },
      -1,
   } },

   /* replace586_0 -> 119 in the cache */
   /* replace586_1_0 -> 11 in the cache */
   /* replace586_1_1_0 -> 268 in the cache */
   /* replace586_1_1 -> 544 in the cache */
   /* replace586_1 -> 1759 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 119, 1759 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search587_0_0 -> 119 in the cache */
   /* search587_0_1 -> 11 in the cache */
   /* search587_0 -> 1769 in the cache */
   /* search587_1_0 -> 268 in the cache */
   /* search587_1_1 -> 119 in the cache */
   /* search587_1 -> 1770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1769, 1770 },
      -1,
   } },

   /* replace587_0 -> 119 in the cache */
   /* replace587_1_0 -> 11 in the cache */
   /* replace587_1_1_0 -> 268 in the cache */
   /* replace587_1_1 -> 544 in the cache */
   /* replace587_1 -> 1759 in the cache */
   /* replace587 -> 1776 in the cache */

   /* ('ior', ('feq(is_used_once)', 'a@16', 0.0), ('feq', 'b@16', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search588_0_0 -> 11 in the cache */
   /* search588_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 11, 119 },
      0,
   } },
   /* search588_1_0 -> 268 in the cache */
   /* search588_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 268, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1778, 1779 },
      -1,
   } },

   /* replace588_0_0_0 -> 11 in the cache */
   /* replace588_0_0 -> 700 in the cache */
   /* replace588_0_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 700, 1781 },
      -1,
   } },
   /* replace588_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1782, 119 },
      -1,
   } },

   /* ('ior', ('fneu(is_used_once)', 'a@16', 0.0), ('fneu', 'b@16', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search589_0_0 -> 11 in the cache */
   /* search589_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 11, 119 },
      0,
   } },
   /* search589_1_0 -> 268 in the cache */
   /* search589_1_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 268, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1784, 1785 },
      -1,
   } },

   /* replace589_0_0_0 -> 11 in the cache */
   /* replace589_0_0 -> 700 in the cache */
   /* replace589_0_1_0 -> 268 in the cache */
   /* replace589_0_1 -> 1781 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 700, 1781 },
      -1,
   } },
   /* replace589_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1787, 119 },
      -1,
   } },

   /* ('iand', ('feq(is_used_once)', 'a@16', 0.0), ('feq', 'b@16', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search590_0_0 -> 11 in the cache */
   /* search590_0_1 -> 119 in the cache */
   /* search590_0 -> 1778 in the cache */
   /* search590_1_0 -> 268 in the cache */
   /* search590_1_1 -> 119 in the cache */
   /* search590_1 -> 1779 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1778, 1779 },
      -1,
   } },

   /* replace590_0_0_0 -> 11 in the cache */
   /* replace590_0_0 -> 700 in the cache */
   /* replace590_0_1_0 -> 268 in the cache */
   /* replace590_0_1 -> 1781 in the cache */
   /* replace590_0 -> 1787 in the cache */
   /* replace590_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1787, 119 },
      -1,
   } },

   /* ('iand', ('fneu(is_used_once)', 'a@16', 0.0), ('fneu', 'b@16', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search591_0_0 -> 11 in the cache */
   /* search591_0_1 -> 119 in the cache */
   /* search591_0 -> 1784 in the cache */
   /* search591_1_0 -> 268 in the cache */
   /* search591_1_1 -> 119 in the cache */
   /* search591_1 -> 1785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1784, 1785 },
      -1,
   } },

   /* replace591_0_0_0 -> 11 in the cache */
   /* replace591_0_0 -> 700 in the cache */
   /* replace591_0_1_0 -> 268 in the cache */
   /* replace591_0_1 -> 1781 in the cache */
   /* replace591_0 -> 1782 in the cache */
   /* replace591_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1782, 119 },
      -1,
   } },

   /* ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search592_0_0 -> 11 in the cache */
   /* search592_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 119 },
      -1,
   } },
   /* search592_1 -> 491 in the cache */
   /* search592_2_0_0_0_0 -> 119 in the cache */
   /* search592_2_0_0_0_1 -> 11 in the cache */
   /* search592_2_0_0_0 -> 1761 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1761 },
      -1,
   } },
   /* search592_2_0_1_0_0_0 -> 11 in the cache */
   /* search592_2_0_1_0_0_1 -> 119 in the cache */
   /* search592_2_0_1_0_0 -> 699 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 699 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1795 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1794, 1796 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1793, 491, 1798 },
      -1,
   } },

   /* replace592_0_0_0_0 -> 11 in the cache */
   /* replace592_0_0_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1800 },
      -1,
   } },
   /* replace592_0_1_0_0_0 -> 11 in the cache */
   /* replace592_0_1_0_0_1 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1802 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1803 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1801, 1804 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1805 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search593_0 -> 770 in the cache */
   /* search593_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 268 },
      0,
   } },
   /* search593_2_0 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 479 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1807, 1808 },
      -1,
   } },

   /* replace593_0_0 -> 770 in the cache */
   /* replace593_0_1 -> 268 in the cache */
   /* replace593_0_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 268, 479 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1810 },
      -1,
   } },

   /* ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a') */
   /* search594_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 1812 },
      -1,
   } },

   /* replace594_0 -> 11 in the cache */
   /* replace594 -> 698 in the cache */

   /* ('~f2i16', ('i2f', 'a@16')) => a */
   /* search595_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1814 },
      -1,
   } },

   /* replace595 -> 11 in the cache */

   /* ('~f2i16', ('u2f', 'a@16')) => a */
   /* search596_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1816 },
      -1,
   } },

   /* replace596 -> 11 in the cache */

   /* ('~f2u16', ('i2f', 'a@16')) => a */
   /* search597_0_0 -> 11 in the cache */
   /* search597_0 -> 1814 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1814 },
      -1,
   } },

   /* replace597 -> 11 in the cache */

   /* ('~f2u16', ('u2f', 'a@16')) => a */
   /* search598_0_0 -> 11 in the cache */
   /* search598_0 -> 1816 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1816 },
      -1,
   } },

   /* replace598 -> 11 in the cache */

   /* ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a') */
   /* search599_0_0_0 -> 119 in the cache */
   /* search599_0_0_1 -> 11 in the cache */
   /* search599_0_0 -> 1761 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1761 },
      -1,
   } },
   /* search599_1_0_0_0 -> 11 in the cache */
   /* search599_1_0_0_1 -> 119 in the cache */
   /* search599_1_0_0 -> 699 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 699 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1821 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1820, 1822 },
      -1,
   } },

   /* replace599_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a')) */
   /* search600_0_0_0 -> 869 in the cache */
   /* search600_0_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 869, 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1825 },
      -1,
   } },
   /* search600_1_0_0_0 -> 11 in the cache */
   /* search600_1_0_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 869 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1828 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1826, 1829 },
      -1,
   } },

   /* replace600_0_0 -> 11 in the cache */
   /* replace600_0 -> 1824 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1824 },
      -1,
   } },

   /* ('~f2f16', ('f2f', 'a')) => ('f2f16', 'a') */
   /* search601_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace601_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2f16', ('u2f', 'a')) => ('u2f16', 'a') */
   /* search602_0_0 -> 32 in the cache */
   /* search602_0 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 112 },
      -1,
   } },

   /* replace602_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2f16', ('i2f', 'a')) => ('i2f16', 'a') */
   /* search603_0_0 -> 32 in the cache */
   /* search603_0 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1254 },
      -1,
   } },

   /* replace603_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2u16', ('f2f', 'a')) => ('f2u16', 'a') */
   /* search604_0_0 -> 32 in the cache */
   /* search604_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace604_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2i16', ('f2f', 'a')) => ('f2i16', 'a') */
   /* search605_0_0 -> 32 in the cache */
   /* search605_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace605_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2f16', ('f2i', ('fsign', 'a@16'))) => ('fsign', 'a') */
   /* search606_0_0_0 -> 11 in the cache */
   /* search606_0_0 -> 1824 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 1824 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 1843 },
      -1,
   } },

   /* replace606_0 -> 11 in the cache */
   /* replace606 -> 1824 in the cache */

   /* ('~f2f16', ('f2f32', 'a@16')) => a */
   /* search607_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace607 -> 11 in the cache */

   /* ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a') */
   /* search608_0_0 -> 11 in the cache */
   /* search608_0 -> 1845 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace608_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a') */
   /* search609_0_0 -> 11 in the cache */
   /* search609_0 -> 1845 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace609_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a') */
   /* search610_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1851 },
      -1,
   } },

   /* replace610_0 -> 32 in the cache */
   /* replace610 -> 1836 in the cache */

   /* ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a') */
   /* search611_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1853 },
      -1,
   } },

   /* replace611_0 -> 32 in the cache */
   /* replace611 -> 1838 in the cache */

   /* ('~f2f16', ('f2f64', 'a@16')) => a */
   /* search612_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1855 },
      -1,
   } },

   /* replace612 -> 11 in the cache */

   /* ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a') */
   /* search613_0_0 -> 11 in the cache */
   /* search613_0 -> 1855 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1855 },
      -1,
   } },

   /* replace613_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a') */
   /* search614_0_0 -> 11 in the cache */
   /* search614_0 -> 1855 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1855 },
      -1,
   } },

   /* replace614_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a') */
   /* search615_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1861 },
      -1,
   } },

   /* replace615_0 -> 32 in the cache */
   /* replace615 -> 1836 in the cache */

   /* ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a') */
   /* search616_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1863 },
      -1,
   } },

   /* replace616_0 -> 32 in the cache */
   /* replace616 -> 1838 in the cache */

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search617_0_0 -> 121 in the cache */
   /* search617_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 121, 16 },
      0,
   } },
   /* search617_1_0 -> 36 in the cache */
   /* search617_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1865, 1866 },
      -1,
   } },

   /* replace617_0 -> 121 in the cache */
   /* replace617_1_0 -> 16 in the cache */
   /* replace617_1_1_0 -> 36 in the cache */
   /* replace617_1_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 16, 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 121, 1868 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search618_0_0 -> 121 in the cache */
   /* search618_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 121, 16 },
      -1,
   } },
   /* search618_1_0 -> 36 in the cache */
   /* search618_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 121 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1870, 1871 },
      -1,
   } },

   /* replace618_0 -> 121 in the cache */
   /* replace618_1_0 -> 16 in the cache */
   /* replace618_1_1_0 -> 36 in the cache */
   /* replace618_1_1 -> 610 in the cache */
   /* replace618_1 -> 1868 in the cache */
   /* replace618 -> 1869 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search619_0_0 -> 121 in the cache */
   /* search619_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 121, 16 },
      0,
   } },
   /* search619_1_0 -> 36 in the cache */
   /* search619_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1873, 1874 },
      -1,
   } },

   /* replace619_0 -> 121 in the cache */
   /* replace619_1_0 -> 16 in the cache */
   /* replace619_1_1_0 -> 36 in the cache */
   /* replace619_1_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 16, 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 121, 1876 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search620_0_0 -> 121 in the cache */
   /* search620_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 121, 16 },
      -1,
   } },
   /* search620_1_0 -> 36 in the cache */
   /* search620_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 121 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1878, 1879 },
      -1,
   } },

   /* replace620_0 -> 121 in the cache */
   /* replace620_1_0 -> 16 in the cache */
   /* replace620_1_1_0 -> 36 in the cache */
   /* replace620_1_1 -> 610 in the cache */
   /* replace620_1 -> 1876 in the cache */
   /* replace620 -> 1877 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search621_0_0 -> 121 in the cache */
   /* search621_0_1 -> 16 in the cache */
   /* search621_0 -> 1865 in the cache */
   /* search621_1_0 -> 36 in the cache */
   /* search621_1_1 -> 121 in the cache */
   /* search621_1 -> 1866 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1865, 1866 },
      -1,
   } },

   /* replace621_0 -> 121 in the cache */
   /* replace621_1_0 -> 16 in the cache */
   /* replace621_1_1_0 -> 36 in the cache */
   /* replace621_1_1 -> 610 in the cache */
   /* replace621_1 -> 1876 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 121, 1876 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search622_0_0 -> 121 in the cache */
   /* search622_0_1 -> 16 in the cache */
   /* search622_0 -> 1870 in the cache */
   /* search622_1_0 -> 36 in the cache */
   /* search622_1_1 -> 121 in the cache */
   /* search622_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1870, 1871 },
      -1,
   } },

   /* replace622_0 -> 121 in the cache */
   /* replace622_1_0 -> 16 in the cache */
   /* replace622_1_1_0 -> 36 in the cache */
   /* replace622_1_1 -> 610 in the cache */
   /* replace622_1 -> 1876 in the cache */
   /* replace622 -> 1882 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search623_0_0 -> 121 in the cache */
   /* search623_0_1 -> 16 in the cache */
   /* search623_0 -> 1873 in the cache */
   /* search623_1_0 -> 36 in the cache */
   /* search623_1_1 -> 121 in the cache */
   /* search623_1 -> 1874 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1873, 1874 },
      -1,
   } },

   /* replace623_0 -> 121 in the cache */
   /* replace623_1_0 -> 16 in the cache */
   /* replace623_1_1_0 -> 36 in the cache */
   /* replace623_1_1 -> 610 in the cache */
   /* replace623_1 -> 1868 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 121, 1868 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search624_0_0 -> 121 in the cache */
   /* search624_0_1 -> 16 in the cache */
   /* search624_0 -> 1878 in the cache */
   /* search624_1_0 -> 36 in the cache */
   /* search624_1_1 -> 121 in the cache */
   /* search624_1 -> 1879 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1878, 1879 },
      -1,
   } },

   /* replace624_0 -> 121 in the cache */
   /* replace624_1_0 -> 16 in the cache */
   /* replace624_1_1_0 -> 36 in the cache */
   /* replace624_1_1 -> 610 in the cache */
   /* replace624_1 -> 1868 in the cache */
   /* replace624 -> 1885 in the cache */

   /* ('ior', ('feq(is_used_once)', 'a@32', 0.0), ('feq', 'b@32', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search625_0_0 -> 16 in the cache */
   /* search625_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 16, 121 },
      0,
   } },
   /* search625_1_0 -> 36 in the cache */
   /* search625_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 36, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1887, 1888 },
      -1,
   } },

   /* replace625_0_0_0 -> 16 in the cache */
   /* replace625_0_0 -> 706 in the cache */
   /* replace625_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 706, 1890 },
      -1,
   } },
   /* replace625_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1891, 121 },
      -1,
   } },

   /* ('ior', ('fneu(is_used_once)', 'a@32', 0.0), ('fneu', 'b@32', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search626_0_0 -> 16 in the cache */
   /* search626_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 16, 121 },
      0,
   } },
   /* search626_1_0 -> 36 in the cache */
   /* search626_1_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 36, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1893, 1894 },
      -1,
   } },

   /* replace626_0_0_0 -> 16 in the cache */
   /* replace626_0_0 -> 706 in the cache */
   /* replace626_0_1_0 -> 36 in the cache */
   /* replace626_0_1 -> 1890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 706, 1890 },
      -1,
   } },
   /* replace626_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1896, 121 },
      -1,
   } },

   /* ('iand', ('feq(is_used_once)', 'a@32', 0.0), ('feq', 'b@32', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search627_0_0 -> 16 in the cache */
   /* search627_0_1 -> 121 in the cache */
   /* search627_0 -> 1887 in the cache */
   /* search627_1_0 -> 36 in the cache */
   /* search627_1_1 -> 121 in the cache */
   /* search627_1 -> 1888 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1887, 1888 },
      -1,
   } },

   /* replace627_0_0_0 -> 16 in the cache */
   /* replace627_0_0 -> 706 in the cache */
   /* replace627_0_1_0 -> 36 in the cache */
   /* replace627_0_1 -> 1890 in the cache */
   /* replace627_0 -> 1896 in the cache */
   /* replace627_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1896, 121 },
      -1,
   } },

   /* ('iand', ('fneu(is_used_once)', 'a@32', 0.0), ('fneu', 'b@32', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search628_0_0 -> 16 in the cache */
   /* search628_0_1 -> 121 in the cache */
   /* search628_0 -> 1893 in the cache */
   /* search628_1_0 -> 36 in the cache */
   /* search628_1_1 -> 121 in the cache */
   /* search628_1 -> 1894 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1893, 1894 },
      -1,
   } },

   /* replace628_0_0_0 -> 16 in the cache */
   /* replace628_0_0 -> 706 in the cache */
   /* replace628_0_1_0 -> 36 in the cache */
   /* replace628_0_1 -> 1890 in the cache */
   /* replace628_0 -> 1891 in the cache */
   /* replace628_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1891, 121 },
      -1,
   } },

   /* ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search629_0_0 -> 16 in the cache */
   /* search629_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 121 },
      -1,
   } },
   /* search629_1 -> 254 in the cache */
   /* search629_2_0_0_0_0 -> 121 in the cache */
   /* search629_2_0_0_0_1 -> 16 in the cache */
   /* search629_2_0_0_0 -> 1870 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1870 },
      -1,
   } },
   /* search629_2_0_1_0_0_0 -> 16 in the cache */
   /* search629_2_0_1_0_0_1 -> 121 in the cache */
   /* search629_2_0_1_0_0 -> 705 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 705 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1904 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1903, 1905 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1906 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1902, 254, 1907 },
      -1,
   } },

   /* replace629_0_0_0_0 -> 16 in the cache */
   /* replace629_0_0_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1909 },
      -1,
   } },
   /* replace629_0_1_0_0_0 -> 16 in the cache */
   /* replace629_0_1_0_0_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1911 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1910, 1913 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1914 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search630_0 -> 770 in the cache */
   /* search630_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 36 },
      0,
   } },
   /* search630_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1916, 1917 },
      -1,
   } },

   /* replace630_0_0 -> 770 in the cache */
   /* replace630_0_1 -> 36 in the cache */
   /* replace630_0_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 36, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1919 },
      -1,
   } },

   /* ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a') */
   /* search631_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1921 },
      -1,
   } },

   /* replace631_0 -> 16 in the cache */
   /* replace631 -> 704 in the cache */

   /* ('~f2i32', ('i2f', 'a@32')) => a */
   /* search632_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1923 },
      -1,
   } },

   /* replace632 -> 16 in the cache */

   /* ('~f2i32', ('u2f', 'a@32')) => a */
   /* search633_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1925 },
      -1,
   } },

   /* replace633 -> 16 in the cache */

   /* ('~f2u32', ('i2f', 'a@32')) => a */
   /* search634_0_0 -> 16 in the cache */
   /* search634_0 -> 1923 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1923 },
      -1,
   } },

   /* replace634 -> 16 in the cache */

   /* ('~f2u32', ('u2f', 'a@32')) => a */
   /* search635_0_0 -> 16 in the cache */
   /* search635_0 -> 1925 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1925 },
      -1,
   } },

   /* replace635 -> 16 in the cache */

   /* ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a') */
   /* search636_0_0_0 -> 121 in the cache */
   /* search636_0_0_1 -> 16 in the cache */
   /* search636_0_0 -> 1870 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1870 },
      -1,
   } },
   /* search636_1_0_0_0 -> 16 in the cache */
   /* search636_1_0_0_1 -> 121 in the cache */
   /* search636_1_0_0 -> 705 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 705 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1930 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1929, 1931 },
      -1,
   } },

   /* replace636_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a')) */
   /* search637_0_0_0 -> 128 in the cache */
   /* search637_0_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1934 },
      -1,
   } },
   /* search637_1_0_0_0 -> 16 in the cache */
   /* search637_1_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1936 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1937 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1935, 1938 },
      -1,
   } },

   /* replace637_0_0 -> 16 in the cache */
   /* replace637_0 -> 1933 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1933 },
      -1,
   } },

   /* ('~f2f32', ('f2f', 'a')) => ('f2f32', 'a') */
   /* search638_0_0 -> 32 in the cache */
   /* search638_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace638_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2f32', ('u2f', 'a')) => ('u2f32', 'a') */
   /* search639_0_0 -> 32 in the cache */
   /* search639_0 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 112 },
      -1,
   } },

   /* replace639_0 -> 32 in the cache */
   /* replace639 -> 1851 in the cache */

   /* ('~f2f32', ('i2f', 'a')) => ('i2f32', 'a') */
   /* search640_0_0 -> 32 in the cache */
   /* search640_0 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1254 },
      -1,
   } },

   /* replace640_0 -> 32 in the cache */
   /* replace640 -> 1853 in the cache */

   /* ('~f2u32', ('f2f', 'a')) => ('f2u32', 'a') */
   /* search641_0_0 -> 32 in the cache */
   /* search641_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace641_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2i32', ('f2f', 'a')) => ('f2i32', 'a') */
   /* search642_0_0 -> 32 in the cache */
   /* search642_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace642_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2f32', ('f2i', ('fsign', 'a@32'))) => ('fsign', 'a') */
   /* search643_0_0_0 -> 16 in the cache */
   /* search643_0_0 -> 1933 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 1933 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1949 },
      -1,
   } },

   /* replace643_0 -> 16 in the cache */
   /* replace643 -> 1933 in the cache */

   /* ('~f2f32', ('f2f64', 'a@32')) => a */
   /* search644_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1951 },
      -1,
   } },

   /* replace644 -> 16 in the cache */

   /* ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a') */
   /* search645_0_0 -> 16 in the cache */
   /* search645_0 -> 1951 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1951 },
      -1,
   } },

   /* replace645_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a') */
   /* search646_0_0 -> 16 in the cache */
   /* search646_0 -> 1951 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1951 },
      -1,
   } },

   /* replace646_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a') */
   /* search647_0_0 -> 32 in the cache */
   /* search647_0 -> 1861 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1861 },
      -1,
   } },

   /* replace647_0 -> 32 in the cache */
   /* replace647 -> 1851 in the cache */

   /* ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a') */
   /* search648_0_0 -> 32 in the cache */
   /* search648_0 -> 1863 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1863 },
      -1,
   } },

   /* replace648_0 -> 32 in the cache */
   /* replace648 -> 1853 in the cache */

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search649_0_0 -> 711 in the cache */
   /* search649_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 711, 21 },
      0,
   } },
   /* search649_1_0 -> 41 in the cache */
   /* search649_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1959, 1960 },
      -1,
   } },

   /* replace649_0 -> 711 in the cache */
   /* replace649_1_0 -> 21 in the cache */
   /* replace649_1_1_0 -> 41 in the cache */
   /* replace649_1_1 -> 683 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 21, 683 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 711, 1962 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search650_0_0 -> 711 in the cache */
   /* search650_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 711, 21 },
      -1,
   } },
   /* search650_1_0 -> 41 in the cache */
   /* search650_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 711 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1964, 1965 },
      -1,
   } },

   /* replace650_0 -> 711 in the cache */
   /* replace650_1_0 -> 21 in the cache */
   /* replace650_1_1_0 -> 41 in the cache */
   /* replace650_1_1 -> 683 in the cache */
   /* replace650_1 -> 1962 in the cache */
   /* replace650 -> 1963 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search651_0_0 -> 711 in the cache */
   /* search651_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 711, 21 },
      0,
   } },
   /* search651_1_0 -> 41 in the cache */
   /* search651_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1967, 1968 },
      -1,
   } },

   /* replace651_0 -> 711 in the cache */
   /* replace651_1_0 -> 21 in the cache */
   /* replace651_1_1_0 -> 41 in the cache */
   /* replace651_1_1 -> 683 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 21, 683 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 711, 1970 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search652_0_0 -> 711 in the cache */
   /* search652_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 711, 21 },
      -1,
   } },
   /* search652_1_0 -> 41 in the cache */
   /* search652_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 711 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1972, 1973 },
      -1,
   } },

   /* replace652_0 -> 711 in the cache */
   /* replace652_1_0 -> 21 in the cache */
   /* replace652_1_1_0 -> 41 in the cache */
   /* replace652_1_1 -> 683 in the cache */
   /* replace652_1 -> 1970 in the cache */
   /* replace652 -> 1971 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search653_0_0 -> 711 in the cache */
   /* search653_0_1 -> 21 in the cache */
   /* search653_0 -> 1959 in the cache */
   /* search653_1_0 -> 41 in the cache */
   /* search653_1_1 -> 711 in the cache */
   /* search653_1 -> 1960 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1959, 1960 },
      -1,
   } },

   /* replace653_0 -> 711 in the cache */
   /* replace653_1_0 -> 21 in the cache */
   /* replace653_1_1_0 -> 41 in the cache */
   /* replace653_1_1 -> 683 in the cache */
   /* replace653_1 -> 1970 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 711, 1970 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search654_0_0 -> 711 in the cache */
   /* search654_0_1 -> 21 in the cache */
   /* search654_0 -> 1964 in the cache */
   /* search654_1_0 -> 41 in the cache */
   /* search654_1_1 -> 711 in the cache */
   /* search654_1 -> 1965 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1964, 1965 },
      -1,
   } },

   /* replace654_0 -> 711 in the cache */
   /* replace654_1_0 -> 21 in the cache */
   /* replace654_1_1_0 -> 41 in the cache */
   /* replace654_1_1 -> 683 in the cache */
   /* replace654_1 -> 1970 in the cache */
   /* replace654 -> 1976 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search655_0_0 -> 711 in the cache */
   /* search655_0_1 -> 21 in the cache */
   /* search655_0 -> 1967 in the cache */
   /* search655_1_0 -> 41 in the cache */
   /* search655_1_1 -> 711 in the cache */
   /* search655_1 -> 1968 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1967, 1968 },
      -1,
   } },

   /* replace655_0 -> 711 in the cache */
   /* replace655_1_0 -> 21 in the cache */
   /* replace655_1_1_0 -> 41 in the cache */
   /* replace655_1_1 -> 683 in the cache */
   /* replace655_1 -> 1962 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 711, 1962 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search656_0_0 -> 711 in the cache */
   /* search656_0_1 -> 21 in the cache */
   /* search656_0 -> 1972 in the cache */
   /* search656_1_0 -> 41 in the cache */
   /* search656_1_1 -> 711 in the cache */
   /* search656_1 -> 1973 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1972, 1973 },
      -1,
   } },

   /* replace656_0 -> 711 in the cache */
   /* replace656_1_0 -> 21 in the cache */
   /* replace656_1_1_0 -> 41 in the cache */
   /* replace656_1_1 -> 683 in the cache */
   /* replace656_1 -> 1962 in the cache */
   /* replace656 -> 1979 in the cache */

   /* ('ior', ('feq(is_used_once)', 'a@64', 0.0), ('feq', 'b@64', 0.0)) => ('feq', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search657_0_0 -> 21 in the cache */
   /* search657_0_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 21, 711 },
      0,
   } },
   /* search657_1_0 -> 41 in the cache */
   /* search657_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 41, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1981, 1982 },
      -1,
   } },

   /* replace657_0_0_0 -> 21 in the cache */
   /* replace657_0_0 -> 713 in the cache */
   /* replace657_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 713, 1984 },
      -1,
   } },
   /* replace657_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1985, 711 },
      -1,
   } },

   /* ('ior', ('fneu(is_used_once)', 'a@64', 0.0), ('fneu', 'b@64', 0.0)) => ('fneu', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search658_0_0 -> 21 in the cache */
   /* search658_0_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 21, 711 },
      0,
   } },
   /* search658_1_0 -> 41 in the cache */
   /* search658_1_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 41, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1987, 1988 },
      -1,
   } },

   /* replace658_0_0_0 -> 21 in the cache */
   /* replace658_0_0 -> 713 in the cache */
   /* replace658_0_1_0 -> 41 in the cache */
   /* replace658_0_1 -> 1984 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 713, 1984 },
      -1,
   } },
   /* replace658_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1990, 711 },
      -1,
   } },

   /* ('iand', ('feq(is_used_once)', 'a@64', 0.0), ('feq', 'b@64', 0.0)) => ('feq', ('fadd', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search659_0_0 -> 21 in the cache */
   /* search659_0_1 -> 711 in the cache */
   /* search659_0 -> 1981 in the cache */
   /* search659_1_0 -> 41 in the cache */
   /* search659_1_1 -> 711 in the cache */
   /* search659_1 -> 1982 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1981, 1982 },
      -1,
   } },

   /* replace659_0_0_0 -> 21 in the cache */
   /* replace659_0_0 -> 713 in the cache */
   /* replace659_0_1_0 -> 41 in the cache */
   /* replace659_0_1 -> 1984 in the cache */
   /* replace659_0 -> 1990 in the cache */
   /* replace659_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1990, 711 },
      -1,
   } },

   /* ('iand', ('fneu(is_used_once)', 'a@64', 0.0), ('fneu', 'b@64', 0.0)) => ('fneu', ('fmin', ('fabs', 'a'), ('fabs', 'b')), 0.0) */
   /* search660_0_0 -> 21 in the cache */
   /* search660_0_1 -> 711 in the cache */
   /* search660_0 -> 1987 in the cache */
   /* search660_1_0 -> 41 in the cache */
   /* search660_1_1 -> 711 in the cache */
   /* search660_1 -> 1988 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1987, 1988 },
      -1,
   } },

   /* replace660_0_0_0 -> 21 in the cache */
   /* replace660_0_0 -> 713 in the cache */
   /* replace660_0_1_0 -> 41 in the cache */
   /* replace660_0_1 -> 1984 in the cache */
   /* replace660_0 -> 1985 in the cache */
   /* replace660_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1985, 711 },
      -1,
   } },

   /* ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search661_0_0 -> 21 in the cache */
   /* search661_0_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 21, 711 },
      -1,
   } },
   /* search661_1 -> 632 in the cache */
   /* search661_2_0_0_0_0 -> 711 in the cache */
   /* search661_2_0_0_0_1 -> 21 in the cache */
   /* search661_2_0_0_0 -> 1964 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1964 },
      -1,
   } },
   /* search661_2_0_1_0_0_0 -> 21 in the cache */
   /* search661_2_0_1_0_0_1 -> 711 in the cache */
   /* search661_2_0_1_0_0 -> 712 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 712 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1998 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1997, 1999 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 2000 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1996, 632, 2001 },
      -1,
   } },

   /* replace661_0_0_0_0 -> 21 in the cache */
   /* replace661_0_0_0_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 21, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 2003 },
      -1,
   } },
   /* replace661_0_1_0_0_0 -> 21 in the cache */
   /* replace661_0_1_0_0_1 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 2005 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2006 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 2004, 2007 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 2008 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search662_0 -> 770 in the cache */
   /* search662_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 41 },
      0,
   } },
   /* search662_2_0 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 2010, 2011 },
      -1,
   } },

   /* replace662_0_0 -> 770 in the cache */
   /* replace662_0_1 -> 41 in the cache */
   /* replace662_0_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 41, 620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 2013 },
      -1,
   } },

   /* ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a') */
   /* search663_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2015 },
      -1,
   } },

   /* replace663_0 -> 21 in the cache */
   /* replace663 -> 710 in the cache */

   /* ('~f2i64', ('i2f', 'a@64')) => a */
   /* search664_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 2017 },
      -1,
   } },

   /* replace664 -> 21 in the cache */

   /* ('~f2i64', ('u2f', 'a@64')) => a */
   /* search665_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 2019 },
      -1,
   } },

   /* replace665 -> 21 in the cache */

   /* ('~f2u64', ('i2f', 'a@64')) => a */
   /* search666_0_0 -> 21 in the cache */
   /* search666_0 -> 2017 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 2017 },
      -1,
   } },

   /* replace666 -> 21 in the cache */

   /* ('~f2u64', ('u2f', 'a@64')) => a */
   /* search667_0_0 -> 21 in the cache */
   /* search667_0 -> 2019 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 2019 },
      -1,
   } },

   /* replace667 -> 21 in the cache */

   /* ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a') */
   /* search668_0_0_0 -> 711 in the cache */
   /* search668_0_0_1 -> 21 in the cache */
   /* search668_0_0 -> 1964 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1964 },
      -1,
   } },
   /* search668_1_0_0_0 -> 21 in the cache */
   /* search668_1_0_0_1 -> 711 in the cache */
   /* search668_1_0_0 -> 712 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 712 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2024 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 2023, 2025 },
      -1,
   } },

   /* replace668_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a')) */
   /* search669_0_0_0 -> 910 in the cache */
   /* search669_0_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 910, 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 2028 },
      -1,
   } },
   /* search669_1_0_0_0 -> 21 in the cache */
   /* search669_1_0_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 2030 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2031 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 2029, 2032 },
      -1,
   } },

   /* replace669_0_0 -> 21 in the cache */
   /* replace669_0 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 2027 },
      -1,
   } },

   /* ('~f2f64', ('f2f', 'a')) => ('f2f64', 'a') */
   /* search670_0_0 -> 32 in the cache */
   /* search670_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace670_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2f64', ('u2f', 'a')) => ('u2f64', 'a') */
   /* search671_0_0 -> 32 in the cache */
   /* search671_0 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 112 },
      -1,
   } },

   /* replace671_0 -> 32 in the cache */
   /* replace671 -> 1861 in the cache */

   /* ('~f2f64', ('i2f', 'a')) => ('i2f64', 'a') */
   /* search672_0_0 -> 32 in the cache */
   /* search672_0 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 1254 },
      -1,
   } },

   /* replace672_0 -> 32 in the cache */
   /* replace672 -> 1863 in the cache */

   /* ('~f2u64', ('f2f', 'a')) => ('f2u64', 'a') */
   /* search673_0_0 -> 32 in the cache */
   /* search673_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace673_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~f2i64', ('f2f', 'a')) => ('f2i64', 'a') */
   /* search674_0_0 -> 32 in the cache */
   /* search674_0 -> 1832 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1832 },
      -1,
   } },

   /* replace674_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2f64', ('f2i', ('fsign', 'a@64'))) => ('fsign', 'a') */
   /* search675_0_0_0 -> 21 in the cache */
   /* search675_0_0 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 2027 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2043 },
      -1,
   } },

   /* replace675_0 -> 21 in the cache */
   /* replace675 -> 2027 in the cache */

   /* ('i2i1', ('i2i8', 'a@1')) => a */
   /* search676_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace676 -> 770 in the cache */

   /* ('u2u1', ('u2u8', 'a@1')) => a */
   /* search677_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace677 -> 770 in the cache */

   /* ('i2i1', ('i2i16', 'a@1')) => a */
   /* search678_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace678 -> 770 in the cache */

   /* ('u2u1', ('u2u16', 'a@1')) => a */
   /* search679_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace679 -> 770 in the cache */

   /* ('u2f16', ('u2u8', 'a@1')) => ('u2f16', 'a') */
   /* search680_0_0 -> 770 in the cache */
   /* search680_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace680_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('i2f16', ('i2i8', 'a@1')) => ('i2f16', 'a') */
   /* search681_0_0 -> 770 in the cache */
   /* search681_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace681_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('u2f16', ('u2u16', 'a@1')) => ('u2f16', 'a') */
   /* search682_0_0 -> 770 in the cache */
   /* search682_0 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace682_0 -> 770 in the cache */
   /* replace682 -> 2054 in the cache */

   /* ('i2f16', ('i2i16', 'a@1')) => ('i2f16', 'a') */
   /* search683_0_0 -> 770 in the cache */
   /* search683_0 -> 2049 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace683_0 -> 770 in the cache */
   /* replace683 -> 2056 in the cache */

   /* ('u2f16', ('u2u32', 'a@1')) => ('u2f16', 'a') */
   /* search684_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2059 },
      -1,
   } },

   /* replace684_0 -> 770 in the cache */
   /* replace684 -> 2054 in the cache */

   /* ('i2f16', ('i2i32', 'a@1')) => ('i2f16', 'a') */
   /* search685_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2061 },
      -1,
   } },

   /* replace685_0 -> 770 in the cache */
   /* replace685 -> 2056 in the cache */

   /* ('u2f16', ('u2u64', 'a@1')) => ('u2f16', 'a') */
   /* search686_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2063 },
      -1,
   } },

   /* replace686_0 -> 770 in the cache */
   /* replace686 -> 2054 in the cache */

   /* ('i2f16', ('i2i64', 'a@1')) => ('i2f16', 'a') */
   /* search687_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2065 },
      -1,
   } },

   /* replace687_0 -> 770 in the cache */
   /* replace687 -> 2056 in the cache */

   /* ('i2i1', ('i2i32', 'a@1')) => a */
   /* search688_0_0 -> 770 in the cache */
   /* search688_0 -> 2061 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 2061 },
      -1,
   } },

   /* replace688 -> 770 in the cache */

   /* ('u2u1', ('u2u32', 'a@1')) => a */
   /* search689_0_0 -> 770 in the cache */
   /* search689_0 -> 2059 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 2059 },
      -1,
   } },

   /* replace689 -> 770 in the cache */

   /* ('u2f32', ('u2u8', 'a@1')) => ('u2f32', 'a') */
   /* search690_0_0 -> 770 in the cache */
   /* search690_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace690_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('i2f32', ('i2i8', 'a@1')) => ('i2f32', 'a') */
   /* search691_0_0 -> 770 in the cache */
   /* search691_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace691_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('u2f32', ('u2u16', 'a@1')) => ('u2f32', 'a') */
   /* search692_0_0 -> 770 in the cache */
   /* search692_0 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace692_0 -> 770 in the cache */
   /* replace692 -> 2070 in the cache */

   /* ('i2f32', ('i2i16', 'a@1')) => ('i2f32', 'a') */
   /* search693_0_0 -> 770 in the cache */
   /* search693_0 -> 2049 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace693_0 -> 770 in the cache */
   /* replace693 -> 2072 in the cache */

   /* ('u2f32', ('u2u32', 'a@1')) => ('u2f32', 'a') */
   /* search694_0_0 -> 770 in the cache */
   /* search694_0 -> 2059 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2059 },
      -1,
   } },

   /* replace694_0 -> 770 in the cache */
   /* replace694 -> 2070 in the cache */

   /* ('i2f32', ('i2i32', 'a@1')) => ('i2f32', 'a') */
   /* search695_0_0 -> 770 in the cache */
   /* search695_0 -> 2061 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2061 },
      -1,
   } },

   /* replace695_0 -> 770 in the cache */
   /* replace695 -> 2072 in the cache */

   /* ('u2f32', ('u2u64', 'a@1')) => ('u2f32', 'a') */
   /* search696_0_0 -> 770 in the cache */
   /* search696_0 -> 2063 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2063 },
      -1,
   } },

   /* replace696_0 -> 770 in the cache */
   /* replace696 -> 2070 in the cache */

   /* ('i2f32', ('i2i64', 'a@1')) => ('i2f32', 'a') */
   /* search697_0_0 -> 770 in the cache */
   /* search697_0 -> 2065 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2065 },
      -1,
   } },

   /* replace697_0 -> 770 in the cache */
   /* replace697 -> 2072 in the cache */

   /* ('i2i1', ('i2i64', 'a@1')) => a */
   /* search698_0_0 -> 770 in the cache */
   /* search698_0 -> 2065 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 2065 },
      -1,
   } },

   /* replace698 -> 770 in the cache */

   /* ('u2u1', ('u2u64', 'a@1')) => a */
   /* search699_0_0 -> 770 in the cache */
   /* search699_0 -> 2063 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 2063 },
      -1,
   } },

   /* replace699 -> 770 in the cache */

   /* ('u2f64', ('u2u8', 'a@1')) => ('u2f64', 'a') */
   /* search700_0_0 -> 770 in the cache */
   /* search700_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace700_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('i2f64', ('i2i8', 'a@1')) => ('i2f64', 'a') */
   /* search701_0_0 -> 770 in the cache */
   /* search701_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace701_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('u2f64', ('u2u16', 'a@1')) => ('u2f64', 'a') */
   /* search702_0_0 -> 770 in the cache */
   /* search702_0 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace702_0 -> 770 in the cache */
   /* replace702 -> 2082 in the cache */

   /* ('i2f64', ('i2i16', 'a@1')) => ('i2f64', 'a') */
   /* search703_0_0 -> 770 in the cache */
   /* search703_0 -> 2049 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace703_0 -> 770 in the cache */
   /* replace703 -> 2084 in the cache */

   /* ('u2f64', ('u2u32', 'a@1')) => ('u2f64', 'a') */
   /* search704_0_0 -> 770 in the cache */
   /* search704_0 -> 2059 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2059 },
      -1,
   } },

   /* replace704_0 -> 770 in the cache */
   /* replace704 -> 2082 in the cache */

   /* ('i2f64', ('i2i32', 'a@1')) => ('i2f64', 'a') */
   /* search705_0_0 -> 770 in the cache */
   /* search705_0 -> 2061 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2061 },
      -1,
   } },

   /* replace705_0 -> 770 in the cache */
   /* replace705 -> 2084 in the cache */

   /* ('u2f64', ('u2u64', 'a@1')) => ('u2f64', 'a') */
   /* search706_0_0 -> 770 in the cache */
   /* search706_0 -> 2063 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2063 },
      -1,
   } },

   /* replace706_0 -> 770 in the cache */
   /* replace706 -> 2082 in the cache */

   /* ('i2f64', ('i2i64', 'a@1')) => ('i2f64', 'a') */
   /* search707_0_0 -> 770 in the cache */
   /* search707_0 -> 2065 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2065 },
      -1,
   } },

   /* replace707_0 -> 770 in the cache */
   /* replace707 -> 2084 in the cache */

   /* ('i2i8', ('i2i16', 'a@8')) => a */
   /* search708_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace708 -> 6 in the cache */

   /* ('u2u8', ('u2u16', 'a@8')) => a */
   /* search709_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace709 -> 6 in the cache */

   /* ('u2f16', ('u2u16', 'a@8')) => ('u2f16', 'a') */
   /* search710_0_0 -> 6 in the cache */
   /* search710_0 -> 2093 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace710_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2f16', ('i2i16', 'a@8')) => ('i2f16', 'a') */
   /* search711_0_0 -> 6 in the cache */
   /* search711_0 -> 2091 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace711_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2f16', ('u2u32', 'a@8')) => ('u2f16', 'a') */
   /* search712_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2099 },
      -1,
   } },

   /* replace712_0 -> 6 in the cache */
   /* replace712 -> 2096 in the cache */

   /* ('i2f16', ('i2i32', 'a@8')) => ('i2f16', 'a') */
   /* search713_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2101 },
      -1,
   } },

   /* replace713_0 -> 6 in the cache */
   /* replace713 -> 2098 in the cache */

   /* ('u2f16', ('u2u64', 'a@8')) => ('u2f16', 'a') */
   /* search714_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2103 },
      -1,
   } },

   /* replace714_0 -> 6 in the cache */
   /* replace714 -> 2096 in the cache */

   /* ('i2f16', ('i2i64', 'a@8')) => ('i2f16', 'a') */
   /* search715_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2105 },
      -1,
   } },

   /* replace715_0 -> 6 in the cache */
   /* replace715 -> 2098 in the cache */

   /* ('i2i8', ('i2i32', 'a@8')) => a */
   /* search716_0_0 -> 6 in the cache */
   /* search716_0 -> 2101 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2101 },
      -1,
   } },

   /* replace716 -> 6 in the cache */

   /* ('u2u8', ('u2u32', 'a@8')) => a */
   /* search717_0_0 -> 6 in the cache */
   /* search717_0 -> 2099 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2099 },
      -1,
   } },

   /* replace717 -> 6 in the cache */

   /* ('u2f32', ('u2u16', 'a@8')) => ('u2f32', 'a') */
   /* search718_0_0 -> 6 in the cache */
   /* search718_0 -> 2093 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace718_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2f32', ('i2i16', 'a@8')) => ('i2f32', 'a') */
   /* search719_0_0 -> 6 in the cache */
   /* search719_0 -> 2091 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace719_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2f32', ('u2u32', 'a@8')) => ('u2f32', 'a') */
   /* search720_0_0 -> 6 in the cache */
   /* search720_0 -> 2099 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2099 },
      -1,
   } },

   /* replace720_0 -> 6 in the cache */
   /* replace720 -> 2110 in the cache */

   /* ('i2f32', ('i2i32', 'a@8')) => ('i2f32', 'a') */
   /* search721_0_0 -> 6 in the cache */
   /* search721_0 -> 2101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2101 },
      -1,
   } },

   /* replace721_0 -> 6 in the cache */
   /* replace721 -> 2112 in the cache */

   /* ('u2f32', ('u2u64', 'a@8')) => ('u2f32', 'a') */
   /* search722_0_0 -> 6 in the cache */
   /* search722_0 -> 2103 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2103 },
      -1,
   } },

   /* replace722_0 -> 6 in the cache */
   /* replace722 -> 2110 in the cache */

   /* ('i2f32', ('i2i64', 'a@8')) => ('i2f32', 'a') */
   /* search723_0_0 -> 6 in the cache */
   /* search723_0 -> 2105 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2105 },
      -1,
   } },

   /* replace723_0 -> 6 in the cache */
   /* replace723 -> 2112 in the cache */

   /* ('i2i8', ('i2i64', 'a@8')) => a */
   /* search724_0_0 -> 6 in the cache */
   /* search724_0 -> 2105 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2105 },
      -1,
   } },

   /* replace724 -> 6 in the cache */

   /* ('u2u8', ('u2u64', 'a@8')) => a */
   /* search725_0_0 -> 6 in the cache */
   /* search725_0 -> 2103 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2103 },
      -1,
   } },

   /* replace725 -> 6 in the cache */

   /* ('u2f64', ('u2u16', 'a@8')) => ('u2f64', 'a') */
   /* search726_0_0 -> 6 in the cache */
   /* search726_0 -> 2093 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace726_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2f64', ('i2i16', 'a@8')) => ('i2f64', 'a') */
   /* search727_0_0 -> 6 in the cache */
   /* search727_0 -> 2091 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace727_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2f64', ('u2u32', 'a@8')) => ('u2f64', 'a') */
   /* search728_0_0 -> 6 in the cache */
   /* search728_0 -> 2099 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2099 },
      -1,
   } },

   /* replace728_0 -> 6 in the cache */
   /* replace728 -> 2120 in the cache */

   /* ('i2f64', ('i2i32', 'a@8')) => ('i2f64', 'a') */
   /* search729_0_0 -> 6 in the cache */
   /* search729_0 -> 2101 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2101 },
      -1,
   } },

   /* replace729_0 -> 6 in the cache */
   /* replace729 -> 2122 in the cache */

   /* ('u2f64', ('u2u64', 'a@8')) => ('u2f64', 'a') */
   /* search730_0_0 -> 6 in the cache */
   /* search730_0 -> 2103 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2103 },
      -1,
   } },

   /* replace730_0 -> 6 in the cache */
   /* replace730 -> 2120 in the cache */

   /* ('i2f64', ('i2i64', 'a@8')) => ('i2f64', 'a') */
   /* search731_0_0 -> 6 in the cache */
   /* search731_0 -> 2105 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2105 },
      -1,
   } },

   /* replace731_0 -> 6 in the cache */
   /* replace731 -> 2122 in the cache */

   /* ('i2i16', ('i2i32', 'a@16')) => a */
   /* search732_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace732 -> 11 in the cache */

   /* ('u2u16', ('u2u32', 'a@16')) => a */
   /* search733_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace733 -> 11 in the cache */

   /* ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a') */
   /* search734_0_0 -> 11 in the cache */
   /* search734_0 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace734_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a') */
   /* search735_0_0 -> 11 in the cache */
   /* search735_0 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace735_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('u2f32', ('u2u64', 'a@16')) => ('u2f32', 'a') */
   /* search736_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2135 },
      -1,
   } },

   /* replace736_0 -> 11 in the cache */
   /* replace736 -> 2132 in the cache */

   /* ('i2f32', ('i2i64', 'a@16')) => ('i2f32', 'a') */
   /* search737_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2137 },
      -1,
   } },

   /* replace737_0 -> 11 in the cache */
   /* replace737 -> 2134 in the cache */

   /* ('i2i16', ('i2i64', 'a@16')) => a */
   /* search738_0_0 -> 11 in the cache */
   /* search738_0 -> 2137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2137 },
      -1,
   } },

   /* replace738 -> 11 in the cache */

   /* ('u2u16', ('u2u64', 'a@16')) => a */
   /* search739_0_0 -> 11 in the cache */
   /* search739_0 -> 2135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2135 },
      -1,
   } },

   /* replace739 -> 11 in the cache */

   /* ('u2f64', ('u2u32', 'a@16')) => ('u2f64', 'a') */
   /* search740_0_0 -> 11 in the cache */
   /* search740_0 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace740_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f64', ('i2i32', 'a@16')) => ('i2f64', 'a') */
   /* search741_0_0 -> 11 in the cache */
   /* search741_0 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace741_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a') */
   /* search742_0_0 -> 11 in the cache */
   /* search742_0 -> 2135 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2135 },
      -1,
   } },

   /* replace742_0 -> 11 in the cache */
   /* replace742 -> 2142 in the cache */

   /* ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a') */
   /* search743_0_0 -> 11 in the cache */
   /* search743_0 -> 2137 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2137 },
      -1,
   } },

   /* replace743_0 -> 11 in the cache */
   /* replace743 -> 2144 in the cache */

   /* ('i2i32', ('i2i64', 'a@32')) => a */
   /* search744_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2147 },
      -1,
   } },

   /* replace744 -> 16 in the cache */

   /* ('u2u32', ('u2u64', 'a@32')) => a */
   /* search745_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2149 },
      -1,
   } },

   /* replace745 -> 16 in the cache */

   /* ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a') */
   /* search746_0_0 -> 16 in the cache */
   /* search746_0 -> 2149 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 2149 },
      -1,
   } },

   /* replace746_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a') */
   /* search747_0_0 -> 16 in the cache */
   /* search747_0 -> 2147 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 2147 },
      -1,
   } },

   /* replace747_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a') */
   /* search748_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 2155 },
      -1,
   } },

   /* replace748_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a') */
   /* search749_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 2158 },
      -1,
   } },

   /* replace749_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a') */
   /* search750_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 2161 },
      -1,
   } },

   /* replace750_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a') */
   /* search751_0_0 -> 16 in the cache */
   /* search751_0 -> 2161 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 2161 },
      -1,
   } },

   /* replace751_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a') */
   /* search752_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2166 },
      -1,
   } },

   /* replace752_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a') */
   /* search753_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2169 },
      -1,
   } },

   /* replace753_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a') */
   /* search754_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2172 },
      -1,
   } },

   /* replace754_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a') */
   /* search755_0_0 -> 16 in the cache */
   /* search755_0 -> 2172 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2172 },
      -1,
   } },

   /* replace755_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255)) */
   /* search756_0_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2177 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2178, 2179 },
      -1,
   } },

   /* replace756_0 -> 11 in the cache */
   /* replace756_1_0 -> 268 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 268, 2181 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 11, 2182 },
      -1,
   } },

   /* ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255)) */
   /* search757_0_0_0 -> 11 in the cache */
   /* search757_0_0_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 2179 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2184 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 1,
      { 2185 },
      -1,
   } },

   /* replace757_0 -> 11 in the cache */
   /* replace757_1_0 -> 268 in the cache */
   /* replace757_1_1 -> 2181 in the cache */
   /* replace757_1 -> 2182 in the cache */
   /* replace757 -> 2183 in the cache */

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search758_0_0_0_0 -> 11 in the cache */
   /* search758_0_0_0 -> 2177 in the cache */
   /* search758_0_0 -> 2178 in the cache */
   /* search758_0_1_0_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2187 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2178, 2188 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2189 },
      -1,
   } },

   /* replace758_0_0 -> 11 in the cache */
   /* replace758_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2191 },
      -1,
   } },

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search759_0_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2193 },
      -1,
   } },
   /* search759_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2195 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2194, 2196 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2197 },
      -1,
   } },

   /* replace759_0_0 -> 16 in the cache */
   /* replace759_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2199 },
      -1,
   } },

   /* ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b') */
   /* search760_0_0 -> 0 in the cache */
   /* search760_0_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 834 },
      -1,
   } },
   /* search760_1_0 -> 146 in the cache */
   /* search760_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 146, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2201, 2202 },
      -1,
   } },

   /* replace760_0_0 -> 0 in the cache */
   /* replace760_0_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 146 },
      -1,
   } },
   /* replace760_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2204, 3 },
      -1,
   } },

   /* ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b') */
   /* search761_0_0 -> 0 in the cache */
   /* search761_0_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 834 },
      -1,
   } },
   /* search761_1_0 -> 146 in the cache */
   /* search761_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 146, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2206, 2207 },
      -1,
   } },

   /* replace761_0_0 -> 0 in the cache */
   /* replace761_0_1 -> 146 in the cache */
   /* replace761_0 -> 2204 in the cache */
   /* replace761_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2204, 3 },
      -1,
   } },

   /* ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b') */
   /* search762_0_0 -> 0 in the cache */
   /* search762_0_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 834 },
      -1,
   } },
   /* search762_1_0 -> 146 in the cache */
   /* search762_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 146, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2210, 2211 },
      -1,
   } },

   /* replace762_0_0 -> 0 in the cache */
   /* replace762_0_1 -> 146 in the cache */
   /* replace762_0 -> 2204 in the cache */
   /* replace762_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2204, 3 },
      -1,
   } },

   /* ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b') */
   /* search763_0_0 -> 0 in the cache */
   /* search763_0_1 -> 834 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 834 },
      -1,
   } },
   /* search763_1_0 -> 146 in the cache */
   /* search763_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 146, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2214, 2215 },
      -1,
   } },

   /* replace763_0_0 -> 0 in the cache */
   /* replace763_0_1 -> 146 in the cache */
   /* replace763_0 -> 2204 in the cache */
   /* replace763_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2204, 3 },
      -1,
   } },

   /* ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b') */
   /* search764_0_0 -> 132 in the cache */
   /* search764_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 132, 33 },
      0,
   } },
   /* search764_1_0 -> 135 in the cache */
   /* search764_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 135, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2218, 2219 },
      -1,
   } },

   /* replace764_0_0 -> 132 in the cache */
   /* replace764_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 132, 135 },
      -1,
   } },
   /* replace764_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2221, 36 },
      -1,
   } },

   /* ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b') */
   /* search765_0_0 -> 132 in the cache */
   /* search765_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 132, 33 },
      0,
   } },
   /* search765_1_0 -> 135 in the cache */
   /* search765_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 135, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2223, 2224 },
      -1,
   } },

   /* replace765_0_0 -> 132 in the cache */
   /* replace765_0_1 -> 135 in the cache */
   /* replace765_0 -> 2221 in the cache */
   /* replace765_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2221, 36 },
      -1,
   } },

   /* ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b') */
   /* search766_0_0 -> 132 in the cache */
   /* search766_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 132, 33 },
      0,
   } },
   /* search766_1_0 -> 135 in the cache */
   /* search766_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 135, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2227, 2228 },
      -1,
   } },

   /* replace766_0_0 -> 132 in the cache */
   /* replace766_0_1 -> 135 in the cache */
   /* replace766_0 -> 2221 in the cache */
   /* replace766_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2221, 36 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search767_0_0_0_0 -> 11 in the cache */
   /* search767_0_0_0 -> 2177 in the cache */
   /* search767_0_0 -> 2178 in the cache */
   /* search767_0_1_0_0 -> 268 in the cache */
   /* search767_0_1_0 -> 2187 in the cache */
   /* search767_0_1 -> 2188 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2178, 2188 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2231 },
      -1,
   } },

   /* replace767_0_0 -> 11 in the cache */
   /* replace767_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 11, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2233 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search768_0_0_0_0 -> 16 in the cache */
   /* search768_0_0_0 -> 2193 in the cache */
   /* search768_0_0 -> 2194 in the cache */
   /* search768_0_1_0_0 -> 36 in the cache */
   /* search768_0_1_0 -> 2195 in the cache */
   /* search768_0_1 -> 2196 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2194, 2196 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2235 },
      -1,
   } },

   /* replace768_0_0 -> 16 in the cache */
   /* replace768_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2237 },
      -1,
   } },

   /* ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b') */
   /* search769_0_0 -> 0 in the cache */
   /* search769_0_1 -> 834 in the cache */
   /* search769_0 -> 2201 in the cache */
   /* search769_1_0 -> 146 in the cache */
   /* search769_1_1 -> 3 in the cache */
   /* search769_1 -> 2202 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2201, 2202 },
      -1,
   } },

   /* replace769_0_0 -> 0 in the cache */
   /* replace769_0_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 146 },
      -1,
   } },
   /* replace769_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2240, 3 },
      -1,
   } },

   /* ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b') */
   /* search770_0_0 -> 0 in the cache */
   /* search770_0_1 -> 834 in the cache */
   /* search770_0 -> 2206 in the cache */
   /* search770_1_0 -> 146 in the cache */
   /* search770_1_1 -> 3 in the cache */
   /* search770_1 -> 2207 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2206, 2207 },
      -1,
   } },

   /* replace770_0_0 -> 0 in the cache */
   /* replace770_0_1 -> 146 in the cache */
   /* replace770_0 -> 2240 in the cache */
   /* replace770_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2240, 3 },
      -1,
   } },

   /* ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b') */
   /* search771_0_0 -> 0 in the cache */
   /* search771_0_1 -> 834 in the cache */
   /* search771_0 -> 2210 in the cache */
   /* search771_1_0 -> 146 in the cache */
   /* search771_1_1 -> 3 in the cache */
   /* search771_1 -> 2211 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2210, 2211 },
      -1,
   } },

   /* replace771_0_0 -> 0 in the cache */
   /* replace771_0_1 -> 146 in the cache */
   /* replace771_0 -> 2240 in the cache */
   /* replace771_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2240, 3 },
      -1,
   } },

   /* ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b') */
   /* search772_0_0 -> 0 in the cache */
   /* search772_0_1 -> 834 in the cache */
   /* search772_0 -> 2214 in the cache */
   /* search772_1_0 -> 146 in the cache */
   /* search772_1_1 -> 3 in the cache */
   /* search772_1 -> 2215 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2214, 2215 },
      -1,
   } },

   /* replace772_0_0 -> 0 in the cache */
   /* replace772_0_1 -> 146 in the cache */
   /* replace772_0 -> 2240 in the cache */
   /* replace772_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2240, 3 },
      -1,
   } },

   /* ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b') */
   /* search773_0_0 -> 132 in the cache */
   /* search773_0_1 -> 33 in the cache */
   /* search773_0 -> 2218 in the cache */
   /* search773_1_0 -> 135 in the cache */
   /* search773_1_1 -> 36 in the cache */
   /* search773_1 -> 2219 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2218, 2219 },
      -1,
   } },

   /* replace773_0_0 -> 132 in the cache */
   /* replace773_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 132, 135 },
      -1,
   } },
   /* replace773_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2249, 36 },
      -1,
   } },

   /* ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b') */
   /* search774_0_0 -> 132 in the cache */
   /* search774_0_1 -> 33 in the cache */
   /* search774_0 -> 2223 in the cache */
   /* search774_1_0 -> 135 in the cache */
   /* search774_1_1 -> 36 in the cache */
   /* search774_1 -> 2224 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2223, 2224 },
      -1,
   } },

   /* replace774_0_0 -> 132 in the cache */
   /* replace774_0_1 -> 135 in the cache */
   /* replace774_0 -> 2249 in the cache */
   /* replace774_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2249, 36 },
      -1,
   } },

   /* ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b') */
   /* search775_0_0 -> 132 in the cache */
   /* search775_0_1 -> 33 in the cache */
   /* search775_0 -> 2227 in the cache */
   /* search775_1_0 -> 135 in the cache */
   /* search775_1_1 -> 36 in the cache */
   /* search775_1 -> 2228 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2227, 2228 },
      -1,
   } },

   /* replace775_0_0 -> 132 in the cache */
   /* replace775_0_1 -> 135 in the cache */
   /* replace775_0 -> 2249 in the cache */
   /* replace775_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2249, 36 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search776_0_0_0_0 -> 11 in the cache */
   /* search776_0_0_0 -> 2177 in the cache */
   /* search776_0_0 -> 2178 in the cache */
   /* search776_0_1_0_0 -> 268 in the cache */
   /* search776_0_1_0 -> 2187 in the cache */
   /* search776_0_1 -> 2188 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2178, 2188 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2255 },
      -1,
   } },

   /* replace776_0_0 -> 11 in the cache */
   /* replace776_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 11, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2257 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search777_0_0_0_0 -> 16 in the cache */
   /* search777_0_0_0 -> 2193 in the cache */
   /* search777_0_0 -> 2194 in the cache */
   /* search777_0_1_0_0 -> 36 in the cache */
   /* search777_0_1_0 -> 2195 in the cache */
   /* search777_0_1 -> 2196 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2194, 2196 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2259 },
      -1,
   } },

   /* replace777_0_0 -> 16 in the cache */
   /* replace777_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2261 },
      -1,
   } },

   /* ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b') */
   /* search778_0_0 -> 0 in the cache */
   /* search778_0_1 -> 834 in the cache */
   /* search778_0 -> 2201 in the cache */
   /* search778_1_0 -> 146 in the cache */
   /* search778_1_1 -> 3 in the cache */
   /* search778_1 -> 2202 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2201, 2202 },
      -1,
   } },

   /* replace778_0_0 -> 0 in the cache */
   /* replace778_0_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 0, 146 },
      -1,
   } },
   /* replace778_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2264, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b') */
   /* search779_0_0 -> 0 in the cache */
   /* search779_0_1 -> 834 in the cache */
   /* search779_0 -> 2206 in the cache */
   /* search779_1_0 -> 146 in the cache */
   /* search779_1_1 -> 3 in the cache */
   /* search779_1 -> 2207 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2206, 2207 },
      -1,
   } },

   /* replace779_0_0 -> 0 in the cache */
   /* replace779_0_1 -> 146 in the cache */
   /* replace779_0 -> 2264 in the cache */
   /* replace779_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2264, 3 },
      -1,
   } },

   /* ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b') */
   /* search780_0_0 -> 0 in the cache */
   /* search780_0_1 -> 834 in the cache */
   /* search780_0 -> 2210 in the cache */
   /* search780_1_0 -> 146 in the cache */
   /* search780_1_1 -> 3 in the cache */
   /* search780_1 -> 2211 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2210, 2211 },
      -1,
   } },

   /* replace780_0_0 -> 0 in the cache */
   /* replace780_0_1 -> 146 in the cache */
   /* replace780_0 -> 2264 in the cache */
   /* replace780_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2264, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b') */
   /* search781_0_0 -> 0 in the cache */
   /* search781_0_1 -> 834 in the cache */
   /* search781_0 -> 2214 in the cache */
   /* search781_1_0 -> 146 in the cache */
   /* search781_1_1 -> 3 in the cache */
   /* search781_1 -> 2215 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2214, 2215 },
      -1,
   } },

   /* replace781_0_0 -> 0 in the cache */
   /* replace781_0_1 -> 146 in the cache */
   /* replace781_0 -> 2264 in the cache */
   /* replace781_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2264, 3 },
      -1,
   } },

   /* ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b') */
   /* search782_0_0 -> 132 in the cache */
   /* search782_0_1 -> 33 in the cache */
   /* search782_0 -> 2218 in the cache */
   /* search782_1_0 -> 135 in the cache */
   /* search782_1_1 -> 36 in the cache */
   /* search782_1 -> 2219 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2218, 2219 },
      -1,
   } },

   /* replace782_0_0 -> 132 in the cache */
   /* replace782_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 132, 135 },
      -1,
   } },
   /* replace782_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2273, 36 },
      -1,
   } },

   /* ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b') */
   /* search783_0_0 -> 132 in the cache */
   /* search783_0_1 -> 33 in the cache */
   /* search783_0 -> 2223 in the cache */
   /* search783_1_0 -> 135 in the cache */
   /* search783_1_1 -> 36 in the cache */
   /* search783_1 -> 2224 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2223, 2224 },
      -1,
   } },

   /* replace783_0_0 -> 132 in the cache */
   /* replace783_0_1 -> 135 in the cache */
   /* replace783_0 -> 2273 in the cache */
   /* replace783_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2273, 36 },
      -1,
   } },

   /* ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b') */
   /* search784_0_0 -> 132 in the cache */
   /* search784_0_1 -> 33 in the cache */
   /* search784_0 -> 2227 in the cache */
   /* search784_1_0 -> 135 in the cache */
   /* search784_1_1 -> 36 in the cache */
   /* search784_1 -> 2228 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2227, 2228 },
      -1,
   } },

   /* replace784_0_0 -> 132 in the cache */
   /* replace784_0_1 -> 135 in the cache */
   /* replace784_0 -> 2273 in the cache */
   /* replace784_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2273, 36 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search785_0_0 -> 6 in the cache */
   /* search785_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 848 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search785_1_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2280, 848 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2279, 2281 },
      -1,
   } },

   /* replace785_0_0 -> 6 in the cache */
   /* replace785_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 6, 2280 },
      -1,
   } },
   /* replace785_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2283, 848 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search786_0_0 -> 6 in the cache */
   /* search786_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 848 },
      -1,
   } },
   /* search786_1_0 -> 2280 in the cache */
   /* search786_1_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 2280, 848 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2285, 2286 },
      -1,
   } },

   /* replace786_0_0 -> 6 in the cache */
   /* replace786_0_1 -> 2280 in the cache */
   /* replace786_0 -> 2283 in the cache */
   /* replace786_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2283, 848 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search787_0_0 -> 6 in the cache */
   /* search787_0_1 -> 848 in the cache */
   /* search787_0 -> 2279 in the cache */
   /* search787_1_0 -> 2280 in the cache */
   /* search787_1_1 -> 848 in the cache */
   /* search787_1 -> 2281 in the cache */
   /* search787 -> 2282 in the cache */

   /* replace787_0_0 -> 6 in the cache */
   /* replace787_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 6, 2280 },
      -1,
   } },
   /* replace787_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2289, 848 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search788_0_0 -> 6 in the cache */
   /* search788_0_1 -> 848 in the cache */
   /* search788_0 -> 2279 in the cache */
   /* search788_1_0 -> 2280 in the cache */
   /* search788_1_1 -> 848 in the cache */
   /* search788_1 -> 2281 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2279, 2281 },
      -1,
   } },

   /* replace788_0_0 -> 6 in the cache */
   /* replace788_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 6, 2280 },
      -1,
   } },
   /* replace788_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2292, 848 },
      -1,
   } },

   /* ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search789_0_0 -> 6 in the cache */
   /* search789_0_1 -> 848 in the cache */
   /* search789_0 -> 2285 in the cache */
   /* search789_1_0 -> 2280 in the cache */
   /* search789_1_1 -> 848 in the cache */
   /* search789_1 -> 2286 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2285, 2286 },
      -1,
   } },

   /* replace789_0_0 -> 6 in the cache */
   /* replace789_0_1 -> 2280 in the cache */
   /* replace789_0 -> 2292 in the cache */
   /* replace789_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2292, 848 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search790_0_0 -> 6 in the cache */
   /* search790_0_1 -> 848 in the cache */
   /* search790_0 -> 2285 in the cache */
   /* search790_1_0 -> 2280 in the cache */
   /* search790_1_1 -> 848 in the cache */
   /* search790_1 -> 2286 in the cache */
   /* search790 -> 2287 in the cache */

   /* replace790_0_0 -> 6 in the cache */
   /* replace790_0_1 -> 2280 in the cache */
   /* replace790_0 -> 2289 in the cache */
   /* replace790_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2289, 848 },
      -1,
   } },

   /* ('ineg', ('b2i8', 'a@8')) => a */
   /* search791_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2297 },
      -1,
   } },

   /* replace791 -> 6 in the cache */

   /* ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b') */
   /* search792_0 -> 6 in the cache */
   /* search792_1_0 -> 9 in the cache */
   /* search792_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 9, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 2299 },
      -1,
   } },

   /* replace792_0 -> 6 in the cache */
   /* replace792_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b') */
   /* search793_0 -> 6 in the cache */
   /* search793_1_0 -> 9 in the cache */
   /* search793_1_1 -> 36 in the cache */
   /* search793_1 -> 2299 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 6, 2299 },
      -1,
   } },

   /* replace793_0 -> 6 in the cache */
   /* replace793_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b') */
   /* search794_0 -> 6 in the cache */
   /* search794_1_0 -> 9 in the cache */
   /* search794_1_1 -> 36 in the cache */
   /* search794_1 -> 2299 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2299 },
      -1,
   } },

   /* replace794_0 -> 6 in the cache */
   /* replace794_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2)) */
   /* search795_0 -> 6 in the cache */
   /* search795_1_0_0 -> 36 in the cache */
   /* search795_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 36, 406 },
      -1,
   } },
   /* search795_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2306, 411 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2307 },
      -1,
   } },

   /* replace795_0 -> 6 in the cache */
   /* replace795_1_0 -> 36 in the cache */
   /* replace795_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 411 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 2309 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search796_0_0 -> 11 in the cache */
   /* search796_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 869 },
      -1,
   } },
   /* search796_1_0 -> 268 in the cache */
   /* search796_1_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 268, 869 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2311, 2312 },
      -1,
   } },

   /* replace796_0_0 -> 11 in the cache */
   /* replace796_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 11, 268 },
      -1,
   } },
   /* replace796_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2314, 869 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search797_0_0 -> 11 in the cache */
   /* search797_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 869 },
      -1,
   } },
   /* search797_1_0 -> 268 in the cache */
   /* search797_1_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 268, 869 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2316, 2317 },
      -1,
   } },

   /* replace797_0_0 -> 11 in the cache */
   /* replace797_0_1 -> 268 in the cache */
   /* replace797_0 -> 2314 in the cache */
   /* replace797_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2314, 869 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search798_0_0 -> 11 in the cache */
   /* search798_0_1 -> 869 in the cache */
   /* search798_0 -> 2311 in the cache */
   /* search798_1_0 -> 268 in the cache */
   /* search798_1_1 -> 869 in the cache */
   /* search798_1 -> 2312 in the cache */
   /* search798 -> 2313 in the cache */

   /* replace798_0_0 -> 11 in the cache */
   /* replace798_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 11, 268 },
      -1,
   } },
   /* replace798_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2320, 869 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search799_0_0 -> 11 in the cache */
   /* search799_0_1 -> 869 in the cache */
   /* search799_0 -> 2311 in the cache */
   /* search799_1_0 -> 268 in the cache */
   /* search799_1_1 -> 869 in the cache */
   /* search799_1 -> 2312 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2311, 2312 },
      -1,
   } },

   /* replace799_0_0 -> 11 in the cache */
   /* replace799_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 11, 268 },
      -1,
   } },
   /* replace799_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2323, 869 },
      -1,
   } },

   /* ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search800_0_0 -> 11 in the cache */
   /* search800_0_1 -> 869 in the cache */
   /* search800_0 -> 2316 in the cache */
   /* search800_1_0 -> 268 in the cache */
   /* search800_1_1 -> 869 in the cache */
   /* search800_1 -> 2317 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2316, 2317 },
      -1,
   } },

   /* replace800_0_0 -> 11 in the cache */
   /* replace800_0_1 -> 268 in the cache */
   /* replace800_0 -> 2323 in the cache */
   /* replace800_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2323, 869 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search801_0_0 -> 11 in the cache */
   /* search801_0_1 -> 869 in the cache */
   /* search801_0 -> 2316 in the cache */
   /* search801_1_0 -> 268 in the cache */
   /* search801_1_1 -> 869 in the cache */
   /* search801_1 -> 2317 in the cache */
   /* search801 -> 2318 in the cache */

   /* replace801_0_0 -> 11 in the cache */
   /* replace801_0_1 -> 268 in the cache */
   /* replace801_0 -> 2320 in the cache */
   /* replace801_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2320, 869 },
      -1,
   } },

   /* ('ineg', ('b2i16', 'a@16')) => a */
   /* search802_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2328 },
      -1,
   } },

   /* replace802 -> 11 in the cache */

   /* ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b') */
   /* search803_0 -> 11 in the cache */
   /* search803_1_0 -> 14 in the cache */
   /* search803_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 14, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2330 },
      -1,
   } },

   /* replace803_0 -> 11 in the cache */
   /* replace803_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b') */
   /* search804_0 -> 11 in the cache */
   /* search804_1_0 -> 14 in the cache */
   /* search804_1_1 -> 36 in the cache */
   /* search804_1 -> 2330 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 11, 2330 },
      -1,
   } },

   /* replace804_0 -> 11 in the cache */
   /* replace804_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b') */
   /* search805_0 -> 11 in the cache */
   /* search805_1_0 -> 14 in the cache */
   /* search805_1_1 -> 36 in the cache */
   /* search805_1 -> 2330 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2330 },
      -1,
   } },

   /* replace805_0 -> 11 in the cache */
   /* replace805_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3)) */
   /* search806_0 -> 11 in the cache */
   /* search806_1_0_0 -> 36 in the cache */
   /* search806_1_0_1 -> 406 in the cache */
   /* search806_1_0 -> 2306 in the cache */
   /* search806_1_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2306, 415 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2337 },
      -1,
   } },

   /* replace806_0 -> 11 in the cache */
   /* replace806_1_0 -> 36 in the cache */
   /* replace806_1_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2339 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search807_0_0 -> 16 in the cache */
   /* search807_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 128 },
      -1,
   } },
   /* search807_1_0 -> 36 in the cache */
   /* search807_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2341, 2342 },
      -1,
   } },

   /* replace807_0_0 -> 16 in the cache */
   /* replace807_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace807_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2344, 128 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search808_0_0 -> 16 in the cache */
   /* search808_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 128 },
      -1,
   } },
   /* search808_1_0 -> 36 in the cache */
   /* search808_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2346, 2347 },
      -1,
   } },

   /* replace808_0_0 -> 16 in the cache */
   /* replace808_0_1 -> 36 in the cache */
   /* replace808_0 -> 2344 in the cache */
   /* replace808_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2344, 128 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search809_0_0 -> 16 in the cache */
   /* search809_0_1 -> 128 in the cache */
   /* search809_0 -> 2341 in the cache */
   /* search809_1_0 -> 36 in the cache */
   /* search809_1_1 -> 128 in the cache */
   /* search809_1 -> 2342 in the cache */
   /* search809 -> 2343 in the cache */

   /* replace809_0_0 -> 16 in the cache */
   /* replace809_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace809_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2350, 128 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search810_0_0 -> 16 in the cache */
   /* search810_0_1 -> 128 in the cache */
   /* search810_0 -> 2341 in the cache */
   /* search810_1_0 -> 36 in the cache */
   /* search810_1_1 -> 128 in the cache */
   /* search810_1 -> 2342 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2341, 2342 },
      -1,
   } },

   /* replace810_0_0 -> 16 in the cache */
   /* replace810_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace810_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2353, 128 },
      -1,
   } },

   /* ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search811_0_0 -> 16 in the cache */
   /* search811_0_1 -> 128 in the cache */
   /* search811_0 -> 2346 in the cache */
   /* search811_1_0 -> 36 in the cache */
   /* search811_1_1 -> 128 in the cache */
   /* search811_1 -> 2347 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2346, 2347 },
      -1,
   } },

   /* replace811_0_0 -> 16 in the cache */
   /* replace811_0_1 -> 36 in the cache */
   /* replace811_0 -> 2353 in the cache */
   /* replace811_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2353, 128 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search812_0_0 -> 16 in the cache */
   /* search812_0_1 -> 128 in the cache */
   /* search812_0 -> 2346 in the cache */
   /* search812_1_0 -> 36 in the cache */
   /* search812_1_1 -> 128 in the cache */
   /* search812_1 -> 2347 in the cache */
   /* search812 -> 2348 in the cache */

   /* replace812_0_0 -> 16 in the cache */
   /* replace812_0_1 -> 36 in the cache */
   /* replace812_0 -> 2350 in the cache */
   /* replace812_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2350, 128 },
      -1,
   } },

   /* ('ineg', ('b2i32', 'a@32')) => a */
   /* search813_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2358 },
      -1,
   } },

   /* replace813 -> 16 in the cache */

   /* ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b') */
   /* search814_0 -> 16 in the cache */
   /* search814_1_0 -> 19 in the cache */
   /* search814_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2360 },
      -1,
   } },

   /* replace814_0 -> 16 in the cache */
   /* replace814_1 -> 36 in the cache */
   /* replace814 -> 959 in the cache */

   /* ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b') */
   /* search815_0 -> 16 in the cache */
   /* search815_1_0 -> 19 in the cache */
   /* search815_1_1 -> 36 in the cache */
   /* search815_1 -> 2360 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 16, 2360 },
      -1,
   } },

   /* replace815_0 -> 16 in the cache */
   /* replace815_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b') */
   /* search816_0 -> 16 in the cache */
   /* search816_1_0 -> 19 in the cache */
   /* search816_1_1 -> 36 in the cache */
   /* search816_1 -> 2360 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2360 },
      -1,
   } },

   /* replace816_0 -> 16 in the cache */
   /* replace816_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4)) */
   /* search817_0 -> 16 in the cache */
   /* search817_1_0_0 -> 36 in the cache */
   /* search817_1_0_1 -> 406 in the cache */
   /* search817_1_0 -> 2306 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2306, 2366 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2367 },
      -1,
   } },

   /* replace817_0 -> 16 in the cache */
   /* replace817_1_0 -> 36 in the cache */
   /* replace817_1_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2369 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search818_0_0 -> 21 in the cache */
   /* search818_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 21, 910 },
      -1,
   } },
   /* search818_1_0 -> 41 in the cache */
   /* search818_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 41, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2371, 2372 },
      -1,
   } },

   /* replace818_0_0 -> 21 in the cache */
   /* replace818_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace818_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2374, 910 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search819_0_0 -> 21 in the cache */
   /* search819_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 21, 910 },
      -1,
   } },
   /* search819_1_0 -> 41 in the cache */
   /* search819_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 41, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2376, 2377 },
      -1,
   } },

   /* replace819_0_0 -> 21 in the cache */
   /* replace819_0_1 -> 41 in the cache */
   /* replace819_0 -> 2374 in the cache */
   /* replace819_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2374, 910 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search820_0_0 -> 21 in the cache */
   /* search820_0_1 -> 910 in the cache */
   /* search820_0 -> 2371 in the cache */
   /* search820_1_0 -> 41 in the cache */
   /* search820_1_1 -> 910 in the cache */
   /* search820_1 -> 2372 in the cache */
   /* search820 -> 2373 in the cache */

   /* replace820_0_0 -> 21 in the cache */
   /* replace820_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace820_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2380, 910 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search821_0_0 -> 21 in the cache */
   /* search821_0_1 -> 910 in the cache */
   /* search821_0 -> 2371 in the cache */
   /* search821_1_0 -> 41 in the cache */
   /* search821_1_1 -> 910 in the cache */
   /* search821_1 -> 2372 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2371, 2372 },
      -1,
   } },

   /* replace821_0_0 -> 21 in the cache */
   /* replace821_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace821_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2383, 910 },
      -1,
   } },

   /* ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search822_0_0 -> 21 in the cache */
   /* search822_0_1 -> 910 in the cache */
   /* search822_0 -> 2376 in the cache */
   /* search822_1_0 -> 41 in the cache */
   /* search822_1_1 -> 910 in the cache */
   /* search822_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2376, 2377 },
      -1,
   } },

   /* replace822_0_0 -> 21 in the cache */
   /* replace822_0_1 -> 41 in the cache */
   /* replace822_0 -> 2383 in the cache */
   /* replace822_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2383, 910 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search823_0_0 -> 21 in the cache */
   /* search823_0_1 -> 910 in the cache */
   /* search823_0 -> 2376 in the cache */
   /* search823_1_0 -> 41 in the cache */
   /* search823_1_1 -> 910 in the cache */
   /* search823_1 -> 2377 in the cache */
   /* search823 -> 2378 in the cache */

   /* replace823_0_0 -> 21 in the cache */
   /* replace823_0_1 -> 41 in the cache */
   /* replace823_0 -> 2380 in the cache */
   /* replace823_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2380, 910 },
      -1,
   } },

   /* ('ineg', ('b2i64', 'a@64')) => a */
   /* search824_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2388 },
      -1,
   } },

   /* replace824 -> 21 in the cache */

   /* ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b') */
   /* search825_0 -> 21 in the cache */
   /* search825_1_0 -> 24 in the cache */
   /* search825_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 24, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 2390 },
      -1,
   } },

   /* replace825_0 -> 21 in the cache */
   /* replace825_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b') */
   /* search826_0 -> 21 in the cache */
   /* search826_1_0 -> 24 in the cache */
   /* search826_1_1 -> 36 in the cache */
   /* search826_1 -> 2390 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 21, 2390 },
      -1,
   } },

   /* replace826_0 -> 21 in the cache */
   /* replace826_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b') */
   /* search827_0 -> 21 in the cache */
   /* search827_1_0 -> 24 in the cache */
   /* search827_1_1 -> 36 in the cache */
   /* search827_1 -> 2390 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2390 },
      -1,
   } },

   /* replace827_0 -> 21 in the cache */
   /* replace827_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5)) */
   /* search828_0 -> 21 in the cache */
   /* search828_1_0_0 -> 36 in the cache */
   /* search828_1_0_1 -> 406 in the cache */
   /* search828_1_0 -> 2306 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2306, 2397 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2398 },
      -1,
   } },

   /* replace828_0 -> 21 in the cache */
   /* replace828_1_0 -> 36 in the cache */
   /* replace828_1_1 -> 2397 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2397 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2400 },
      -1,
   } },

   /* ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a') */
   /* search829_0_0 -> 32 in the cache */
   /* search829_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 61 },
      -1,
   } },
   /* search829_1_0 -> 32 in the cache */
   /* search829_1_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2402, 2403 },
      -1,
   } },

   /* replace829_0 -> 35 in the cache */
   /* replace829_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 35, 32 },
      -1,
   } },

   /* ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a') */
   /* search830_0_0 -> 35 in the cache */
   /* search830_0_1 -> 32 in the cache */
   /* search830_0 -> 2405 in the cache */
   /* search830_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2405, 2407 },
      -1,
   } },

   /* replace830_0 -> 2406 in the cache */
   /* replace830_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2406, 32 },
      -1,
   } },

   /* ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a') */
   /* search831_0_0 -> 2406 in the cache */
   /* search831_0_1 -> 32 in the cache */
   /* search831_0 -> 2409 in the cache */
   /* search831_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2410 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2409, 2411 },
      -1,
   } },

   /* replace831_0 -> 2410 in the cache */
   /* replace831_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2410, 32 },
      -1,
   } },

   /* ('ior', 'a', ('ieq', 'a', False)) => True */
   /* search832_0 -> 770 in the cache */
   /* search832_1_0 -> 770 in the cache */
   /* search832_1_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 770, 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 770, 2414 },
      -1,
   } },

   /* replace832 -> 1248 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b') */
   /* search833_0_0_0 -> 770 in the cache */
   /* search833_0_0 -> 1241 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1241 },
      -1,
   } },
   /* search833_1_0_0 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1147 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2417 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace833_0 -> 770 in the cache */
   /* replace833_1 -> 1147 in the cache */
   /* replace833 -> 1404 in the cache */

   /* ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b')) */
   /* search834_0_0 -> 770 in the cache */
   /* search834_0_1 -> 1147 in the cache */
   /* search834_0 -> 1404 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1404 },
      -1,
   } },

   /* replace834_0_0 -> 770 in the cache */
   /* replace834_0_1 -> 1147 in the cache */
   /* replace834_0 -> 1164 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1164 },
      -1,
   } },

   /* ('ishl', ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)), '#c') => ('bcsel', ('ige', ('iand', 'c', 31), ('find_lsb', 'b')), ('ishl', ('iand', 'a', 'b'), ('iadd', ('iand', 'c', 31), ('ineg', ('find_lsb', 'b')))), ('ushr', ('iand', 'a', 'b'), ('iadd', ('ineg', ('iand', 'c', 31)), ('find_lsb', 'b')))) */
   /* search835_0_0_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 2422 },
      -1,
   } },
   /* search835_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2423, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 2,
      { 2424 },
      -1,
   } },
   /* search835_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 2425, 319 },
      -1,
   } },

   /* replace835_0_0_0 -> 141 in the cache */
   /* replace835_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 141, 19 },
      -1,
   } },
   /* replace835_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2427, 2428 },
      -1,
   } },
   /* replace835_1_0_0 -> 16 in the cache */
   /* replace835_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace835_1_1_0_0 -> 141 in the cache */
   /* replace835_1_1_0_1 -> 19 in the cache */
   /* replace835_1_1_0 -> 898 in the cache */
   /* replace835_1_1_1_0_0 -> 36 in the cache */
   /* replace835_1_1_1_0 -> 2428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2428 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 2,
      { 898, 2431 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 2430, 2432 },
      -1,
   } },
   /* replace835_2_0_0 -> 16 in the cache */
   /* replace835_2_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      4, 1,
      { 16, 36 },
      -1,
   } },
   /* replace835_2_1_0_0_0 -> 141 in the cache */
   /* replace835_2_1_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      6, 1,
      { 141, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2435 },
      -1,
   } },
   /* replace835_2_1_1_0 -> 36 in the cache */
   /* replace835_2_1_1 -> 2428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 2,
      { 2436, 2428 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 3,
      { 2434, 2437 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 7,
      { 2429, 2433, 2438 },
      -1,
   } },

   /* ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)) => ('ushr', ('iand', 'a', 'b'), ('find_lsb', 'b')) */
   /* search836_0_0_0 -> 16 in the cache */
   /* search836_0_0_1 -> 2422 in the cache */
   /* search836_0_0 -> 2423 in the cache */
   /* search836_0_1 -> 128 in the cache */
   /* search836_0 -> 2424 in the cache */
   /* search836 -> 2425 in the cache */

   /* replace836_0_0 -> 16 in the cache */
   /* replace836_0_1 -> 36 in the cache */
   /* replace836_0 -> 2199 in the cache */
   /* replace836_1_0 -> 36 in the cache */
   /* replace836_1 -> 2428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2199, 2428 },
      -1,
   } },

   /* ('ior', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('ior', ('b2i', 'a'), 'b'), 1) */
   /* search837_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search837_1_0 -> 3 in the cache */
   /* search837_1_1 -> 79 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3, 79 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2441, 2442 },
      -1,
   } },

   /* replace837_0_0_0 -> 32 in the cache */
   /* replace837_0_0 -> 2441 in the cache */
   /* replace837_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 2441, 3 },
      -1,
   } },
   /* replace837_1 -> 79 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2444, 79 },
      -1,
   } },

   /* ('iand', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('b2i', 'a'), 'b') */
   /* search838_0_0 -> 32 in the cache */
   /* search838_0 -> 2441 in the cache */
   /* search838_1_0 -> 3 in the cache */
   /* search838_1_1 -> 79 in the cache */
   /* search838_1 -> 2442 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2441, 2442 },
      -1,
   } },

   /* replace838_0_0 -> 32 in the cache */
   /* replace838_0 -> 2441 in the cache */
   /* replace838_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2441, 3 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0)) */
   /* search839_0_0_0_0_0 -> 32 in the cache */
   /* search839_0_0_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 32, 61 },
      -1,
   } },
   /* search839_0_0_0_1 -> 1147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 2448, 1147 },
      -1,
   } },
   /* search839_0_0_1 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2449, 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2450 },
      -1,
   } },
   /* search839_1_0 -> 32 in the cache */
   /* search839_1_1 -> 61 in the cache */
   /* search839_1 -> 1262 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2451, 1262 },
      -1,
   } },

   /* replace839_0_0_0 -> 1147 in the cache */
   /* replace839_0_0_1 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1147, 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2453 },
      -1,
   } },
   /* replace839_1_0 -> 32 in the cache */
   /* replace839_1_1 -> 61 in the cache */
   /* replace839_1 -> 1262 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2454, 1262 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0)) */
   /* search840_0_0_0_0_0 -> 0 in the cache */
   /* search840_0_0_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      3, 1,
      { 0, 3 },
      -1,
   } },
   /* search840_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 2456, 86 },
      -1,
   } },
   /* search840_0_0_1 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2457, 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2458 },
      -1,
   } },
   /* search840_1_0 -> 0 in the cache */
   /* search840_1_1 -> 86 in the cache */
   /* search840_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2459, 90 },
      -1,
   } },

   /* replace840_0_0_0_0 -> 3 in the cache */
   /* replace840_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3, 86 },
      -1,
   } },
   /* replace840_0_0_1 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 2,
      { 2461, 474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 2,
      { 2462 },
      -1,
   } },
   /* replace840_1_0 -> 0 in the cache */
   /* replace840_1_1 -> 86 in the cache */
   /* replace840_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2463, 90 },
      -1,
   } },

   /* ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b') */
   /* search841_0 -> 32 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search841_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2466 },
      -1,
   } },

   /* replace841_0 -> 32 in the cache */
   /* replace841_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a') */
   /* search842_0_0 -> 0 in the cache */
   /* search842_0_1 -> 3 in the cache */
   /* search842_0 -> 1296 in the cache */
   /* search842_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1296, 0 },
      -1,
   } },

   /* replace842_0 -> 3 in the cache */
   /* replace842_1 -> 0 in the cache */
   /* replace842 -> 1091 in the cache */

   /* ('~fge', 'a', ('fmin', 'b', 'a')) => True */
   /* search843_0 -> 32 in the cache */
   /* search843_1_0 -> 2465 in the cache */
   /* search843_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2470 },
      -1,
   } },

   /* replace843 -> 1248 in the cache */

   /* ('~fge', ('fmax', 'a', 'b'), 'a') => True */
   /* search844_0_0 -> 0 in the cache */
   /* search844_0_1 -> 3 in the cache */
   /* search844_0 -> 1294 in the cache */
   /* search844_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1294, 0 },
      -1,
   } },

   /* replace844 -> 1248 in the cache */

   /* ('flt', 'a', ('fmin', 'b', 'a')) => False */
   /* search845_0 -> 32 in the cache */
   /* search845_1_0 -> 2465 in the cache */
   /* search845_1_1 -> 32 in the cache */
   /* search845_1 -> 2470 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2470 },
      -1,
   } },

   /* replace845 -> 1090 in the cache */

   /* ('flt', ('fmax', 'a', 'b'), 'a') => False */
   /* search846_0_0 -> 0 in the cache */
   /* search846_0_1 -> 3 in the cache */
   /* search846_0 -> 1294 in the cache */
   /* search846_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1294, 0 },
      -1,
   } },

   /* replace846 -> 1090 in the cache */

   /* ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b') */
   /* search847_0 -> 32 in the cache */
   /* search847_1_0 -> 2465 in the cache */
   /* search847_1_1 -> 32 in the cache */
   /* search847_1 -> 2466 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2466 },
      -1,
   } },

   /* replace847_0 -> 32 in the cache */
   /* replace847_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a') */
   /* search848_0_0 -> 0 in the cache */
   /* search848_0_1 -> 3 in the cache */
   /* search848_0 -> 1296 in the cache */
   /* search848_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1296, 0 },
      -1,
   } },

   /* replace848_0 -> 3 in the cache */
   /* replace848_1 -> 0 in the cache */
   /* replace848 -> 1103 in the cache */

   /* ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b') */
   /* search849_0 -> 32 in the cache */
   /* search849_1_0 -> 2465 in the cache */
   /* search849_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2478 },
      -1,
   } },

   /* replace849_0 -> 32 in the cache */
   /* replace849_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a') */
   /* search850_0_0 -> 0 in the cache */
   /* search850_0_1 -> 3 in the cache */
   /* search850_0 -> 1316 in the cache */
   /* search850_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1316, 0 },
      -1,
   } },

   /* replace850_0 -> 3 in the cache */
   /* replace850_1 -> 0 in the cache */
   /* replace850 -> 1096 in the cache */

   /* ('ige', 'a', ('imin', 'b', 'a')) => True */
   /* search851_0 -> 32 in the cache */
   /* search851_1_0 -> 2465 in the cache */
   /* search851_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2482 },
      -1,
   } },

   /* replace851 -> 1248 in the cache */

   /* ('ige', ('imax', 'a', 'b'), 'a') => True */
   /* search852_0_0 -> 0 in the cache */
   /* search852_0_1 -> 3 in the cache */
   /* search852_0 -> 1314 in the cache */
   /* search852_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1314, 0 },
      -1,
   } },

   /* replace852 -> 1248 in the cache */

   /* ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b') */
   /* search853_0 -> 32 in the cache */
   /* search853_1_0 -> 2465 in the cache */
   /* search853_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2485 },
      -1,
   } },

   /* replace853_0 -> 32 in the cache */
   /* replace853_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a') */
   /* search854_0_0 -> 0 in the cache */
   /* search854_0_1 -> 3 in the cache */
   /* search854_0 -> 1306 in the cache */
   /* search854_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1306, 0 },
      -1,
   } },

   /* replace854_0 -> 3 in the cache */
   /* replace854_1 -> 0 in the cache */
   /* replace854 -> 1098 in the cache */

   /* ('uge', 'a', ('umin', 'b', 'a')) => True */
   /* search855_0 -> 32 in the cache */
   /* search855_1_0 -> 2465 in the cache */
   /* search855_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2489 },
      -1,
   } },

   /* replace855 -> 1248 in the cache */

   /* ('uge', ('umax', 'a', 'b'), 'a') => True */
   /* search856_0_0 -> 0 in the cache */
   /* search856_0_1 -> 3 in the cache */
   /* search856_0 -> 1304 in the cache */
   /* search856_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1304, 0 },
      -1,
   } },

   /* replace856 -> 1248 in the cache */

   /* ('ilt', 'a', ('imin', 'b', 'a')) => False */
   /* search857_0 -> 32 in the cache */
   /* search857_1_0 -> 2465 in the cache */
   /* search857_1_1 -> 32 in the cache */
   /* search857_1 -> 2482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2482 },
      -1,
   } },

   /* replace857 -> 1090 in the cache */

   /* ('ilt', ('imax', 'a', 'b'), 'a') => False */
   /* search858_0_0 -> 0 in the cache */
   /* search858_0_1 -> 3 in the cache */
   /* search858_0 -> 1314 in the cache */
   /* search858_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1314, 0 },
      -1,
   } },

   /* replace858 -> 1090 in the cache */

   /* ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b') */
   /* search859_0 -> 32 in the cache */
   /* search859_1_0 -> 2465 in the cache */
   /* search859_1_1 -> 32 in the cache */
   /* search859_1 -> 2478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2478 },
      -1,
   } },

   /* replace859_0 -> 32 in the cache */
   /* replace859_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a') */
   /* search860_0_0 -> 0 in the cache */
   /* search860_0_1 -> 3 in the cache */
   /* search860_0 -> 1316 in the cache */
   /* search860_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1316, 0 },
      -1,
   } },

   /* replace860_0 -> 3 in the cache */
   /* replace860_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('umin', 'b', 'a')) => False */
   /* search861_0 -> 32 in the cache */
   /* search861_1_0 -> 2465 in the cache */
   /* search861_1_1 -> 32 in the cache */
   /* search861_1 -> 2489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2489 },
      -1,
   } },

   /* replace861 -> 1090 in the cache */

   /* ('ult', ('umax', 'a', 'b'), 'a') => False */
   /* search862_0_0 -> 0 in the cache */
   /* search862_0_1 -> 3 in the cache */
   /* search862_0 -> 1304 in the cache */
   /* search862_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1304, 0 },
      -1,
   } },

   /* replace862 -> 1090 in the cache */

   /* ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b') */
   /* search863_0 -> 32 in the cache */
   /* search863_1_0 -> 2465 in the cache */
   /* search863_1_1 -> 32 in the cache */
   /* search863_1 -> 2485 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2485 },
      -1,
   } },

   /* replace863_0 -> 32 in the cache */
   /* replace863_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 2465 },
      -1,
   } },

   /* ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a') */
   /* search864_0_0 -> 0 in the cache */
   /* search864_0_1 -> 3 in the cache */
   /* search864_0 -> 1306 in the cache */
   /* search864_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1306, 0 },
      -1,
   } },

   /* replace864_0 -> 3 in the cache */
   /* replace864_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('iand', 'b', 'a')) => False */
   /* search865_0 -> 32 in the cache */
   /* search865_1_0 -> 2465 in the cache */
   /* search865_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2504 },
      -1,
   } },

   /* replace865 -> 1090 in the cache */

   /* ('ult', ('ior', 'a', 'b'), 'a') => False */
   /* search866_0_0 -> 0 in the cache */
   /* search866_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* search866_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2506, 0 },
      -1,
   } },

   /* replace866 -> 1090 in the cache */

   /* ('uge', 'a', ('iand', 'b', 'a')) => True */
   /* search867_0 -> 32 in the cache */
   /* search867_1_0 -> 2465 in the cache */
   /* search867_1_1 -> 32 in the cache */
   /* search867_1 -> 2504 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2504 },
      -1,
   } },

   /* replace867 -> 1248 in the cache */

   /* ('uge', ('ior', 'a', 'b'), 'a') => True */
   /* search868_0_0 -> 0 in the cache */
   /* search868_0_1 -> 3 in the cache */
   /* search868_0 -> 2506 in the cache */
   /* search868_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2506, 0 },
      -1,
   } },

   /* replace868 -> 1248 in the cache */

   /* ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search869_0 -> 292 in the cache */
   /* search869_1_0 -> 293 in the cache */
   /* search869_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 293, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 292, 2510 },
      -1,
   } },

   /* replace869_0_0 -> 132 in the cache */
   /* replace869_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 132, 133 },
      -1,
   } },
   /* replace869_1_0 -> 132 in the cache */
   /* replace869_1_1 -> 135 in the cache */
   /* replace869_1 -> 1672 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2512, 1672 },
      -1,
   } },

   /* ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search870_0_0 -> 292 in the cache */
   /* search870_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 292, 133 },
      -1,
   } },
   /* search870_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2514, 827 },
      -1,
   } },

   /* replace870_0_0 -> 132 in the cache */
   /* replace870_0_1 -> 135 in the cache */
   /* replace870_0 -> 1672 in the cache */
   /* replace870_1_0 -> 133 in the cache */
   /* replace870_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1672, 2516 },
      -1,
   } },

   /* ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search871_0 -> 292 in the cache */
   /* search871_1_0 -> 293 in the cache */
   /* search871_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 293, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 292, 2518 },
      -1,
   } },

   /* replace871_0_0 -> 132 in the cache */
   /* replace871_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 132, 133 },
      -1,
   } },
   /* replace871_1_0 -> 132 in the cache */
   /* replace871_1_1 -> 135 in the cache */
   /* replace871_1 -> 1682 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2520, 1682 },
      -1,
   } },

   /* ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search872_0_0 -> 292 in the cache */
   /* search872_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 292, 133 },
      -1,
   } },
   /* search872_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2522, 827 },
      -1,
   } },

   /* replace872_0_0 -> 132 in the cache */
   /* replace872_0_1 -> 135 in the cache */
   /* replace872_0 -> 1682 in the cache */
   /* replace872_1_0 -> 133 in the cache */
   /* replace872_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1682, 2524 },
      -1,
   } },

   /* ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search873_0 -> 292 in the cache */
   /* search873_1_0 -> 293 in the cache */
   /* search873_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 293, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 292, 2526 },
      -1,
   } },

   /* replace873_0_0 -> 132 in the cache */
   /* replace873_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 132, 133 },
      -1,
   } },
   /* replace873_1_0 -> 132 in the cache */
   /* replace873_1_1 -> 135 in the cache */
   /* replace873_1 -> 1692 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2528, 1692 },
      -1,
   } },

   /* ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search874_0_0 -> 292 in the cache */
   /* search874_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 292, 133 },
      -1,
   } },
   /* search874_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2530, 827 },
      -1,
   } },

   /* replace874_0_0 -> 132 in the cache */
   /* replace874_0_1 -> 135 in the cache */
   /* replace874_0 -> 1692 in the cache */
   /* replace874_1_0 -> 133 in the cache */
   /* replace874_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1692, 2532 },
      -1,
   } },

   /* ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search875_0 -> 292 in the cache */
   /* search875_1_0 -> 293 in the cache */
   /* search875_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 293, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 292, 2534 },
      -1,
   } },

   /* replace875_0_0 -> 132 in the cache */
   /* replace875_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 132, 133 },
      -1,
   } },
   /* replace875_1_0 -> 132 in the cache */
   /* replace875_1_1 -> 135 in the cache */
   /* replace875_1 -> 1702 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2536, 1702 },
      -1,
   } },

   /* ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search876_0_0 -> 292 in the cache */
   /* search876_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 292, 133 },
      -1,
   } },
   /* search876_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2538, 827 },
      -1,
   } },

   /* replace876_0_0 -> 132 in the cache */
   /* replace876_0_1 -> 135 in the cache */
   /* replace876_0 -> 1702 in the cache */
   /* replace876_1_0 -> 133 in the cache */
   /* replace876_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 133, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1702, 2540 },
      -1,
   } },

   /* ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search877_0 -> 292 in the cache */
   /* search877_1_0 -> 293 in the cache */
   /* search877_1_1 -> 135 in the cache */
   /* search877_1 -> 2518 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 292, 2518 },
      -1,
   } },

   /* replace877_0_0 -> 132 in the cache */
   /* replace877_0_1 -> 133 in the cache */
   /* replace877_0 -> 2512 in the cache */
   /* replace877_1_0 -> 132 in the cache */
   /* replace877_1_1 -> 135 in the cache */
   /* replace877_1 -> 1672 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2512, 1672 },
      -1,
   } },

   /* ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search878_0_0 -> 292 in the cache */
   /* search878_0_1 -> 133 in the cache */
   /* search878_0 -> 2522 in the cache */
   /* search878_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2522, 827 },
      -1,
   } },

   /* replace878_0_0 -> 132 in the cache */
   /* replace878_0_1 -> 135 in the cache */
   /* replace878_0 -> 1672 in the cache */
   /* replace878_1_0 -> 133 in the cache */
   /* replace878_1_1 -> 135 in the cache */
   /* replace878_1 -> 2516 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1672, 2516 },
      -1,
   } },

   /* ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search879_0 -> 292 in the cache */
   /* search879_1_0 -> 293 in the cache */
   /* search879_1_1 -> 135 in the cache */
   /* search879_1 -> 2510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 292, 2510 },
      -1,
   } },

   /* replace879_0_0 -> 132 in the cache */
   /* replace879_0_1 -> 133 in the cache */
   /* replace879_0 -> 2520 in the cache */
   /* replace879_1_0 -> 132 in the cache */
   /* replace879_1_1 -> 135 in the cache */
   /* replace879_1 -> 1682 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2520, 1682 },
      -1,
   } },

   /* ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search880_0_0 -> 292 in the cache */
   /* search880_0_1 -> 133 in the cache */
   /* search880_0 -> 2514 in the cache */
   /* search880_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2514, 827 },
      -1,
   } },

   /* replace880_0_0 -> 132 in the cache */
   /* replace880_0_1 -> 135 in the cache */
   /* replace880_0 -> 1682 in the cache */
   /* replace880_1_0 -> 133 in the cache */
   /* replace880_1_1 -> 135 in the cache */
   /* replace880_1 -> 2524 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1682, 2524 },
      -1,
   } },

   /* ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search881_0 -> 292 in the cache */
   /* search881_1_0 -> 293 in the cache */
   /* search881_1_1 -> 135 in the cache */
   /* search881_1 -> 2534 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 292, 2534 },
      -1,
   } },

   /* replace881_0_0 -> 132 in the cache */
   /* replace881_0_1 -> 133 in the cache */
   /* replace881_0 -> 2528 in the cache */
   /* replace881_1_0 -> 132 in the cache */
   /* replace881_1_1 -> 135 in the cache */
   /* replace881_1 -> 1692 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2528, 1692 },
      -1,
   } },

   /* ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search882_0_0 -> 292 in the cache */
   /* search882_0_1 -> 133 in the cache */
   /* search882_0 -> 2538 in the cache */
   /* search882_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2538, 827 },
      -1,
   } },

   /* replace882_0_0 -> 132 in the cache */
   /* replace882_0_1 -> 135 in the cache */
   /* replace882_0 -> 1692 in the cache */
   /* replace882_1_0 -> 133 in the cache */
   /* replace882_1_1 -> 135 in the cache */
   /* replace882_1 -> 2532 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1692, 2532 },
      -1,
   } },

   /* ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search883_0 -> 292 in the cache */
   /* search883_1_0 -> 293 in the cache */
   /* search883_1_1 -> 135 in the cache */
   /* search883_1 -> 2526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 292, 2526 },
      -1,
   } },

   /* replace883_0_0 -> 132 in the cache */
   /* replace883_0_1 -> 133 in the cache */
   /* replace883_0 -> 2536 in the cache */
   /* replace883_1_0 -> 132 in the cache */
   /* replace883_1_1 -> 135 in the cache */
   /* replace883_1 -> 1702 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2536, 1702 },
      -1,
   } },

   /* ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search884_0_0 -> 292 in the cache */
   /* search884_0_1 -> 133 in the cache */
   /* search884_0 -> 2530 in the cache */
   /* search884_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2530, 827 },
      -1,
   } },

   /* replace884_0_0 -> 132 in the cache */
   /* replace884_0_1 -> 135 in the cache */
   /* replace884_0 -> 1702 in the cache */
   /* replace884_1_0 -> 133 in the cache */
   /* replace884_1_1 -> 135 in the cache */
   /* replace884_1 -> 2540 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1702, 2540 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b')) */
   /* search885_0_0 -> 32 in the cache */
   /* search885_0_1 -> 61 in the cache */
   /* search885_0 -> 1262 in the cache */
   /* search885_1_0_0 -> 32 in the cache */
   /* search885_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2558 },
      -1,
   } },
   /* search885_2_0 -> 32 in the cache */
   /* search885_2_1 -> 36 in the cache */
   /* search885_2 -> 2558 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1262, 2559, 2558 },
      -1,
   } },

   /* replace885_0_0 -> 32 in the cache */
   /* replace885_0_1 -> 36 in the cache */
   /* replace885_0 -> 2558 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2558 },
      -1,
   } },

   /* ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b') */
   /* search886_0_0_0 -> 32 in the cache */
   /* search886_0_0 -> 114 in the cache */
   /* search886_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 114, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2562 },
      -1,
   } },

   /* replace886_0_0 -> 32 in the cache */
   /* replace886_0 -> 114 in the cache */
   /* replace886_1 -> 36 in the cache */
   /* replace886 -> 2562 in the cache */

   /* ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b') */
   /* search887_0_0 -> 0 in the cache */
   /* search887_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2564 },
      -1,
   } },

   /* replace887_0 -> 0 in the cache */
   /* replace887_1 -> 3 in the cache */
   /* replace887 -> 2564 in the cache */

   /* ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b') */
   /* search888_0_0 -> 0 in the cache */
   /* search888_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2566 },
      -1,
   } },

   /* replace888_0 -> 0 in the cache */
   /* replace888_1 -> 3 in the cache */
   /* replace888 -> 2566 in the cache */

   /* ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b') */
   /* search889_0_0 -> 0 in the cache */
   /* search889_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2568 },
      -1,
   } },

   /* replace889_0 -> 0 in the cache */
   /* replace889_1 -> 3 in the cache */
   /* replace889 -> 2568 in the cache */

   /* ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b') */
   /* search890_0_0 -> 0 in the cache */
   /* search890_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2570 },
      -1,
   } },

   /* replace890_0 -> 0 in the cache */
   /* replace890_1 -> 3 in the cache */
   /* replace890 -> 2570 in the cache */

   /* ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b')) */
   /* search891_0 -> 0 in the cache */
   /* search891_1 -> 3 in the cache */
   /* search891 -> 2564 in the cache */

   /* replace891_0_0 -> 0 in the cache */
   /* replace891_0_1 -> 3 in the cache */
   /* replace891_0 -> 1062 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1062 },
      -1,
   } },

   /* ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b')) */
   /* search892_0 -> 0 in the cache */
   /* search892_1 -> 3 in the cache */
   /* search892 -> 2566 in the cache */

   /* replace892_0_0 -> 0 in the cache */
   /* replace892_0_1 -> 3 in the cache */
   /* replace892_0 -> 1059 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1059 },
      -1,
   } },

   /* ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b')) */
   /* search893_0 -> 0 in the cache */
   /* search893_1 -> 3 in the cache */
   /* search893 -> 2568 in the cache */

   /* replace893_0_0 -> 0 in the cache */
   /* replace893_0_1 -> 3 in the cache */
   /* replace893_0 -> 1068 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1068 },
      -1,
   } },

   /* ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b')) */
   /* search894_0 -> 0 in the cache */
   /* search894_1 -> 3 in the cache */
   /* search894 -> 2570 in the cache */

   /* replace894_0_0 -> 0 in the cache */
   /* replace894_0_1 -> 3 in the cache */
   /* replace894_0 -> 1065 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1065 },
      -1,
   } },

   /* ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search895_0_0 -> 0 in the cache */
   /* search895_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search895_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2576, 280 },
      -1,
   } },

   /* replace895_0 -> 0 in the cache */
   /* replace895_1 -> 3 in the cache */
   /* replace895 -> 2568 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search896_0_0 -> 0 in the cache */
   /* search896_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search896_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2578, 280 },
      -1,
   } },

   /* replace896_0 -> 0 in the cache */
   /* replace896_1 -> 3 in the cache */
   /* replace896 -> 2570 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search897_0_0 -> 0 in the cache */
   /* search897_0_1 -> 3 in the cache */
   /* search897_0 -> 2564 in the cache */
   /* search897_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2564, 280 },
      -1,
   } },

   /* replace897_0 -> 0 in the cache */
   /* replace897_1 -> 3 in the cache */
   /* replace897 -> 2564 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search898_0_0 -> 0 in the cache */
   /* search898_0_1 -> 3 in the cache */
   /* search898_0 -> 2566 in the cache */
   /* search898_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2566, 280 },
      -1,
   } },

   /* replace898_0 -> 0 in the cache */
   /* replace898_1 -> 3 in the cache */
   /* replace898 -> 2566 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search899_0_0 -> 0 in the cache */
   /* search899_0_1 -> 3 in the cache */
   /* search899_0 -> 2576 in the cache */
   /* search899_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2576, 266 },
      -1,
   } },

   /* replace899_0 -> 0 in the cache */
   /* replace899_1 -> 3 in the cache */
   /* replace899 -> 2568 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search900_0_0 -> 0 in the cache */
   /* search900_0_1 -> 3 in the cache */
   /* search900_0 -> 2578 in the cache */
   /* search900_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2578, 266 },
      -1,
   } },

   /* replace900_0 -> 0 in the cache */
   /* replace900_1 -> 3 in the cache */
   /* replace900 -> 2570 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search901_0_0 -> 0 in the cache */
   /* search901_0_1 -> 3 in the cache */
   /* search901_0 -> 2564 in the cache */
   /* search901_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2564, 266 },
      -1,
   } },

   /* replace901_0 -> 0 in the cache */
   /* replace901_1 -> 3 in the cache */
   /* replace901 -> 2564 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search902_0_0 -> 0 in the cache */
   /* search902_0_1 -> 3 in the cache */
   /* search902_0 -> 2566 in the cache */
   /* search902_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2566, 266 },
      -1,
   } },

   /* replace902_0 -> 0 in the cache */
   /* replace902_1 -> 3 in the cache */
   /* replace902 -> 2566 in the cache */

   /* ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search903_0_0 -> 0 in the cache */
   /* search903_0_1 -> 3 in the cache */
   /* search903_0 -> 2576 in the cache */
   /* search903_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2576, 266 },
      -1,
   } },

   /* replace903_0 -> 0 in the cache */
   /* replace903_1 -> 3 in the cache */
   /* replace903 -> 2570 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search904_0_0 -> 0 in the cache */
   /* search904_0_1 -> 3 in the cache */
   /* search904_0 -> 2578 in the cache */
   /* search904_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2578, 266 },
      -1,
   } },

   /* replace904_0 -> 0 in the cache */
   /* replace904_1 -> 3 in the cache */
   /* replace904 -> 2568 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search905_0_0 -> 0 in the cache */
   /* search905_0_1 -> 3 in the cache */
   /* search905_0 -> 2564 in the cache */
   /* search905_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2564, 266 },
      -1,
   } },

   /* replace905_0 -> 0 in the cache */
   /* replace905_1 -> 3 in the cache */
   /* replace905 -> 2566 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search906_0_0 -> 0 in the cache */
   /* search906_0_1 -> 3 in the cache */
   /* search906_0 -> 2566 in the cache */
   /* search906_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2566, 266 },
      -1,
   } },

   /* replace906_0 -> 0 in the cache */
   /* replace906_1 -> 3 in the cache */
   /* replace906 -> 2564 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search907_0_0 -> 0 in the cache */
   /* search907_0_1 -> 3 in the cache */
   /* search907_0 -> 2576 in the cache */
   /* search907_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2576, 280 },
      -1,
   } },

   /* replace907_0 -> 0 in the cache */
   /* replace907_1 -> 3 in the cache */
   /* replace907 -> 2570 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search908_0_0 -> 0 in the cache */
   /* search908_0_1 -> 3 in the cache */
   /* search908_0 -> 2578 in the cache */
   /* search908_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2578, 280 },
      -1,
   } },

   /* replace908_0 -> 0 in the cache */
   /* replace908_1 -> 3 in the cache */
   /* replace908 -> 2568 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search909_0_0 -> 0 in the cache */
   /* search909_0_1 -> 3 in the cache */
   /* search909_0 -> 2564 in the cache */
   /* search909_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2564, 280 },
      -1,
   } },

   /* replace909_0 -> 0 in the cache */
   /* replace909_1 -> 3 in the cache */
   /* replace909 -> 2566 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search910_0_0 -> 0 in the cache */
   /* search910_0_1 -> 3 in the cache */
   /* search910_0 -> 2566 in the cache */
   /* search910_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2566, 280 },
      -1,
   } },

   /* replace910_0 -> 0 in the cache */
   /* replace910_1 -> 3 in the cache */
   /* replace910 -> 2564 in the cache */

   /* ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y')) */
   /* search911_0 -> 16 in the cache */
   /* search911_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2595, 2596 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 2598, 2599 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      -1, 2,
      { 2597, 2600 },
      -1,
   } },

   /* ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0) */
   /* search912_0 -> 16 in the cache */
   /* search912_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace912_0_0 -> 16 in the cache */
   /* replace912_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace912_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2603, 121 },
      -1,
   } },

   /* ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0) */
   /* search913_0 -> 16 in the cache */
   /* search913_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace913_0_0 -> 16 in the cache */
   /* replace913_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace913_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2606, 121 },
      -1,
   } },

   /* ('fall_equal8', 'a', 'b') => ('seq', ('fany_nequal8', 'a', 'b'), 0.0) */
   /* search914_0 -> 16 in the cache */
   /* search914_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal8,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace914_0_0 -> 16 in the cache */
   /* replace914_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal8,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace914_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2609, 121 },
      -1,
   } },

   /* ('fall_equal16', 'a', 'b') => ('seq', ('fany_nequal16', 'a', 'b'), 0.0) */
   /* search915_0 -> 16 in the cache */
   /* search915_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal16,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace915_0_0 -> 16 in the cache */
   /* replace915_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal16,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace915_1 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2612, 121 },
      -1,
   } },

   /* ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y')) */
   /* search916_0 -> 16 in the cache */
   /* search916_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace916_0_0 -> 2595 in the cache */
   /* replace916_0_1 -> 2596 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2595, 2596 },
      -1,
   } },
   /* replace916_1_0 -> 2598 in the cache */
   /* replace916_1_1 -> 2599 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 2598, 2599 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      -1, 2,
      { 2615, 2616 },
      -1,
   } },

   /* ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search917_0 -> 16 in the cache */
   /* search917_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace917_0_0_0 -> 16 in the cache */
   /* replace917_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* replace917_0_1_0 -> 16 in the cache */
   /* replace917_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot3,
      -1, 2,
      { 2619, 2620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2621 },
      -1,
   } },

   /* ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search918_0 -> 16 in the cache */
   /* search918_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace918_0_0_0 -> 16 in the cache */
   /* replace918_0_0_1 -> 36 in the cache */
   /* replace918_0_0 -> 2619 in the cache */
   /* replace918_0_1_0 -> 16 in the cache */
   /* replace918_0_1_1 -> 36 in the cache */
   /* replace918_0_1 -> 2620 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot4,
      -1, 2,
      { 2619, 2620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2624 },
      -1,
   } },

   /* ('fany_nequal8', 'a', 'b') => ('fsat', ('fdot8', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search919_0 -> 16 in the cache */
   /* search919_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal8,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace919_0_0_0 -> 16 in the cache */
   /* replace919_0_0_1 -> 36 in the cache */
   /* replace919_0_0 -> 2619 in the cache */
   /* replace919_0_1_0 -> 16 in the cache */
   /* replace919_0_1_1 -> 36 in the cache */
   /* replace919_0_1 -> 2620 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot8,
      -1, 2,
      { 2619, 2620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2627 },
      -1,
   } },

   /* ('fany_nequal16', 'a', 'b') => ('fsat', ('fdot16', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search920_0 -> 16 in the cache */
   /* search920_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal16,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace920_0_0_0 -> 16 in the cache */
   /* replace920_0_0_1 -> 36 in the cache */
   /* replace920_0_0 -> 2619 in the cache */
   /* replace920_0_1_0 -> 16 in the cache */
   /* replace920_0_1_1 -> 36 in the cache */
   /* replace920_0_1 -> 2620 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot16,
      -1, 2,
      { 2619, 2620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2630 },
      -1,
   } },

   /* ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')) */
   /* search921_0 -> 0 in the cache */
   /* search921_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace921_0_0 -> 779 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace921_1_0 -> 780 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 780, 2635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2634, 2636 },
      -1,
   } },

   /* ('ball_iequal3', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('iand', ('ieq', 'a.y', 'b.y'), ('ieq', 'a.z', 'b.z'))) */
   /* search922_0 -> 0 in the cache */
   /* search922_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace922_0_0 -> 779 in the cache */
   /* replace922_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace922_1_0_0 -> 780 in the cache */
   /* replace922_1_0_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 780, 2635 },
      -1,
   } },
   /* replace922_1_1_0 -> 781 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 781, 2641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2640, 2642 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2639, 2643 },
      -1,
   } },

   /* ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w'))) */
   /* search923_0 -> 0 in the cache */
   /* search923_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace923_0_0_0 -> 779 in the cache */
   /* replace923_0_0_1 -> 2633 in the cache */
   /* replace923_0_0 -> 2634 in the cache */
   /* replace923_0_1_0 -> 780 in the cache */
   /* replace923_0_1_1 -> 2635 in the cache */
   /* replace923_0_1 -> 2636 in the cache */
   /* replace923_0 -> 2637 in the cache */
   /* replace923_1_0_0 -> 781 in the cache */
   /* replace923_1_0_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 781, 2641 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 2647, 2648 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2646, 2649 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2637, 2650 },
      -1,
   } },

   /* ('ball_iequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))) */
   /* search924_0 -> 0 in the cache */
   /* search924_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace924_0_0_0_0 -> 779 in the cache */
   /* replace924_0_0_0_1 -> 2633 in the cache */
   /* replace924_0_0_0 -> 2634 in the cache */
   /* replace924_0_0_1_0 -> 780 in the cache */
   /* replace924_0_0_1_1 -> 2635 in the cache */
   /* replace924_0_0_1 -> 2636 in the cache */
   /* replace924_0_0 -> 2637 in the cache */
   /* replace924_0_1_0_0 -> 781 in the cache */
   /* replace924_0_1_0_1 -> 2641 in the cache */
   /* replace924_0_1_0 -> 2646 in the cache */
   /* replace924_0_1_1_0 -> 2647 in the cache */
   /* replace924_0_1_1_1 -> 2648 in the cache */
   /* replace924_0_1_1 -> 2649 in the cache */
   /* replace924_0_1 -> 2650 in the cache */
   /* replace924_0 -> 2651 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {4},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {4},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      4, 1,
      { 2653, 2654 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {5},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {5},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      5, 1,
      { 2656, 2657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2655, 2658 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {6},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {6},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      6, 1,
      { 2660, 2661 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {7},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {7},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      7, 1,
      { 2663, 2664 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2662, 2665 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2659, 2666 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2651, 2667 },
      -1,
   } },

   /* ('ball_iequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('ieq', 'a.i', 'b.i'), ('ieq', 'a.j', 'b.j')), ('iand', ('ieq', 'a.k', 'b.k'), ('ieq', 'a.l', 'b.l'))), ('iand', ('iand', ('ieq', 'a.m', 'b.m'), ('ieq', 'a.n', 'b.n')), ('iand', ('ieq', 'a.o', 'b.o'), ('ieq', 'a.p', 'b.p'))))) */
   /* search925_0 -> 0 in the cache */
   /* search925_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace925_0_0_0_0_0 -> 779 in the cache */
   /* replace925_0_0_0_0_1 -> 2633 in the cache */
   /* replace925_0_0_0_0 -> 2634 in the cache */
   /* replace925_0_0_0_1_0 -> 780 in the cache */
   /* replace925_0_0_0_1_1 -> 2635 in the cache */
   /* replace925_0_0_0_1 -> 2636 in the cache */
   /* replace925_0_0_0 -> 2637 in the cache */
   /* replace925_0_0_1_0_0 -> 781 in the cache */
   /* replace925_0_0_1_0_1 -> 2641 in the cache */
   /* replace925_0_0_1_0 -> 2646 in the cache */
   /* replace925_0_0_1_1_0 -> 2647 in the cache */
   /* replace925_0_0_1_1_1 -> 2648 in the cache */
   /* replace925_0_0_1_1 -> 2649 in the cache */
   /* replace925_0_0_1 -> 2650 in the cache */
   /* replace925_0_0 -> 2651 in the cache */
   /* replace925_0_1_0_0_0 -> 2653 in the cache */
   /* replace925_0_1_0_0_1 -> 2654 in the cache */
   /* replace925_0_1_0_0 -> 2655 in the cache */
   /* replace925_0_1_0_1_0 -> 2656 in the cache */
   /* replace925_0_1_0_1_1 -> 2657 in the cache */
   /* replace925_0_1_0_1 -> 2658 in the cache */
   /* replace925_0_1_0 -> 2659 in the cache */
   /* replace925_0_1_1_0_0 -> 2660 in the cache */
   /* replace925_0_1_1_0_1 -> 2661 in the cache */
   /* replace925_0_1_1_0 -> 2662 in the cache */
   /* replace925_0_1_1_1_0 -> 2663 in the cache */
   /* replace925_0_1_1_1_1 -> 2664 in the cache */
   /* replace925_0_1_1_1 -> 2665 in the cache */
   /* replace925_0_1_1 -> 2666 in the cache */
   /* replace925_0_1 -> 2667 in the cache */
   /* replace925_0 -> 2668 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {8},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {8},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      8, 1,
      { 2670, 2671 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {9},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {9},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      9, 1,
      { 2673, 2674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2672, 2675 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {10},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {10},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      10, 1,
      { 2677, 2678 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {11},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {11},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      11, 1,
      { 2680, 2681 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2679, 2682 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2676, 2683 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {12},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {12},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      12, 1,
      { 2685, 2686 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {13},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {13},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      13, 1,
      { 2688, 2689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2687, 2690 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {14},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {14},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      14, 1,
      { 2692, 2693 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      15, 1,
      { 2695, 2696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2694, 2697 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2691, 2698 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2684, 2699 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 16,
      { 2668, 2700 },
      -1,
   } },

   /* ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')) */
   /* search926_0 -> 0 in the cache */
   /* search926_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace926_0_0 -> 779 in the cache */
   /* replace926_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace926_1_0 -> 780 in the cache */
   /* replace926_1_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 780, 2635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2703, 2704 },
      -1,
   } },

   /* ('ball_fequal3', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('iand', ('feq', 'a.y', 'b.y'), ('feq', 'a.z', 'b.z'))) */
   /* search927_0 -> 0 in the cache */
   /* search927_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace927_0_0 -> 779 in the cache */
   /* replace927_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace927_1_0_0 -> 780 in the cache */
   /* replace927_1_0_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 780, 2635 },
      -1,
   } },
   /* replace927_1_1_0 -> 781 in the cache */
   /* replace927_1_1_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 781, 2641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2708, 2709 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2707, 2710 },
      -1,
   } },

   /* ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w'))) */
   /* search928_0 -> 0 in the cache */
   /* search928_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace928_0_0_0 -> 779 in the cache */
   /* replace928_0_0_1 -> 2633 in the cache */
   /* replace928_0_0 -> 2703 in the cache */
   /* replace928_0_1_0 -> 780 in the cache */
   /* replace928_0_1_1 -> 2635 in the cache */
   /* replace928_0_1 -> 2704 in the cache */
   /* replace928_0 -> 2705 in the cache */
   /* replace928_1_0_0 -> 781 in the cache */
   /* replace928_1_0_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 781, 2641 },
      -1,
   } },
   /* replace928_1_1_0 -> 2647 in the cache */
   /* replace928_1_1_1 -> 2648 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 2647, 2648 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2713, 2714 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2705, 2715 },
      -1,
   } },

   /* ('ball_fequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))) */
   /* search929_0 -> 0 in the cache */
   /* search929_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace929_0_0_0_0 -> 779 in the cache */
   /* replace929_0_0_0_1 -> 2633 in the cache */
   /* replace929_0_0_0 -> 2703 in the cache */
   /* replace929_0_0_1_0 -> 780 in the cache */
   /* replace929_0_0_1_1 -> 2635 in the cache */
   /* replace929_0_0_1 -> 2704 in the cache */
   /* replace929_0_0 -> 2705 in the cache */
   /* replace929_0_1_0_0 -> 781 in the cache */
   /* replace929_0_1_0_1 -> 2641 in the cache */
   /* replace929_0_1_0 -> 2713 in the cache */
   /* replace929_0_1_1_0 -> 2647 in the cache */
   /* replace929_0_1_1_1 -> 2648 in the cache */
   /* replace929_0_1_1 -> 2714 in the cache */
   /* replace929_0_1 -> 2715 in the cache */
   /* replace929_0 -> 2716 in the cache */
   /* replace929_1_0_0_0 -> 2653 in the cache */
   /* replace929_1_0_0_1 -> 2654 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      4, 1,
      { 2653, 2654 },
      -1,
   } },
   /* replace929_1_0_1_0 -> 2656 in the cache */
   /* replace929_1_0_1_1 -> 2657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      5, 1,
      { 2656, 2657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2718, 2719 },
      -1,
   } },
   /* replace929_1_1_0_0 -> 2660 in the cache */
   /* replace929_1_1_0_1 -> 2661 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      6, 1,
      { 2660, 2661 },
      -1,
   } },
   /* replace929_1_1_1_0 -> 2663 in the cache */
   /* replace929_1_1_1_1 -> 2664 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      7, 1,
      { 2663, 2664 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2721, 2722 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2720, 2723 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2716, 2724 },
      -1,
   } },

   /* ('ball_fequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('feq', 'a.i', 'b.i'), ('feq', 'a.j', 'b.j')), ('iand', ('feq', 'a.k', 'b.k'), ('feq', 'a.l', 'b.l'))), ('iand', ('iand', ('feq', 'a.m', 'b.m'), ('feq', 'a.n', 'b.n')), ('iand', ('feq', 'a.o', 'b.o'), ('feq', 'a.p', 'b.p'))))) */
   /* search930_0 -> 0 in the cache */
   /* search930_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace930_0_0_0_0_0 -> 779 in the cache */
   /* replace930_0_0_0_0_1 -> 2633 in the cache */
   /* replace930_0_0_0_0 -> 2703 in the cache */
   /* replace930_0_0_0_1_0 -> 780 in the cache */
   /* replace930_0_0_0_1_1 -> 2635 in the cache */
   /* replace930_0_0_0_1 -> 2704 in the cache */
   /* replace930_0_0_0 -> 2705 in the cache */
   /* replace930_0_0_1_0_0 -> 781 in the cache */
   /* replace930_0_0_1_0_1 -> 2641 in the cache */
   /* replace930_0_0_1_0 -> 2713 in the cache */
   /* replace930_0_0_1_1_0 -> 2647 in the cache */
   /* replace930_0_0_1_1_1 -> 2648 in the cache */
   /* replace930_0_0_1_1 -> 2714 in the cache */
   /* replace930_0_0_1 -> 2715 in the cache */
   /* replace930_0_0 -> 2716 in the cache */
   /* replace930_0_1_0_0_0 -> 2653 in the cache */
   /* replace930_0_1_0_0_1 -> 2654 in the cache */
   /* replace930_0_1_0_0 -> 2718 in the cache */
   /* replace930_0_1_0_1_0 -> 2656 in the cache */
   /* replace930_0_1_0_1_1 -> 2657 in the cache */
   /* replace930_0_1_0_1 -> 2719 in the cache */
   /* replace930_0_1_0 -> 2720 in the cache */
   /* replace930_0_1_1_0_0 -> 2660 in the cache */
   /* replace930_0_1_1_0_1 -> 2661 in the cache */
   /* replace930_0_1_1_0 -> 2721 in the cache */
   /* replace930_0_1_1_1_0 -> 2663 in the cache */
   /* replace930_0_1_1_1_1 -> 2664 in the cache */
   /* replace930_0_1_1_1 -> 2722 in the cache */
   /* replace930_0_1_1 -> 2723 in the cache */
   /* replace930_0_1 -> 2724 in the cache */
   /* replace930_0 -> 2725 in the cache */
   /* replace930_1_0_0_0_0 -> 2670 in the cache */
   /* replace930_1_0_0_0_1 -> 2671 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      8, 1,
      { 2670, 2671 },
      -1,
   } },
   /* replace930_1_0_0_1_0 -> 2673 in the cache */
   /* replace930_1_0_0_1_1 -> 2674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      9, 1,
      { 2673, 2674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2727, 2728 },
      -1,
   } },
   /* replace930_1_0_1_0_0 -> 2677 in the cache */
   /* replace930_1_0_1_0_1 -> 2678 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      10, 1,
      { 2677, 2678 },
      -1,
   } },
   /* replace930_1_0_1_1_0 -> 2680 in the cache */
   /* replace930_1_0_1_1_1 -> 2681 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      11, 1,
      { 2680, 2681 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2730, 2731 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2729, 2732 },
      -1,
   } },
   /* replace930_1_1_0_0_0 -> 2685 in the cache */
   /* replace930_1_1_0_0_1 -> 2686 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      12, 1,
      { 2685, 2686 },
      -1,
   } },
   /* replace930_1_1_0_1_0 -> 2688 in the cache */
   /* replace930_1_1_0_1_1 -> 2689 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      13, 1,
      { 2688, 2689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2734, 2735 },
      -1,
   } },
   /* replace930_1_1_1_0_0 -> 2692 in the cache */
   /* replace930_1_1_1_0_1 -> 2693 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      14, 1,
      { 2692, 2693 },
      -1,
   } },
   /* replace930_1_1_1_1_0 -> 2695 in the cache */
   /* replace930_1_1_1_1_1 -> 2696 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      15, 1,
      { 2695, 2696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2737, 2738 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2736, 2739 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2733, 2740 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 16,
      { 2725, 2741 },
      -1,
   } },

   /* ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')) */
   /* search931_0 -> 0 in the cache */
   /* search931_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace931_0_0 -> 779 in the cache */
   /* replace931_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace931_1_0 -> 780 in the cache */
   /* replace931_1_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 780, 2635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2744, 2745 },
      -1,
   } },

   /* ('bany_inequal3', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ior', ('ine', 'a.y', 'b.y'), ('ine', 'a.z', 'b.z'))) */
   /* search932_0 -> 0 in the cache */
   /* search932_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace932_0_0 -> 779 in the cache */
   /* replace932_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace932_1_0_0 -> 780 in the cache */
   /* replace932_1_0_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 780, 2635 },
      -1,
   } },
   /* replace932_1_1_0 -> 781 in the cache */
   /* replace932_1_1_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 781, 2641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2749, 2750 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2748, 2751 },
      -1,
   } },

   /* ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w'))) */
   /* search933_0 -> 0 in the cache */
   /* search933_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace933_0_0_0 -> 779 in the cache */
   /* replace933_0_0_1 -> 2633 in the cache */
   /* replace933_0_0 -> 2744 in the cache */
   /* replace933_0_1_0 -> 780 in the cache */
   /* replace933_0_1_1 -> 2635 in the cache */
   /* replace933_0_1 -> 2745 in the cache */
   /* replace933_0 -> 2746 in the cache */
   /* replace933_1_0_0 -> 781 in the cache */
   /* replace933_1_0_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 781, 2641 },
      -1,
   } },
   /* replace933_1_1_0 -> 2647 in the cache */
   /* replace933_1_1_1 -> 2648 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 2647, 2648 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2754, 2755 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2746, 2756 },
      -1,
   } },

   /* ('bany_inequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))) */
   /* search934_0 -> 0 in the cache */
   /* search934_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace934_0_0_0_0 -> 779 in the cache */
   /* replace934_0_0_0_1 -> 2633 in the cache */
   /* replace934_0_0_0 -> 2744 in the cache */
   /* replace934_0_0_1_0 -> 780 in the cache */
   /* replace934_0_0_1_1 -> 2635 in the cache */
   /* replace934_0_0_1 -> 2745 in the cache */
   /* replace934_0_0 -> 2746 in the cache */
   /* replace934_0_1_0_0 -> 781 in the cache */
   /* replace934_0_1_0_1 -> 2641 in the cache */
   /* replace934_0_1_0 -> 2754 in the cache */
   /* replace934_0_1_1_0 -> 2647 in the cache */
   /* replace934_0_1_1_1 -> 2648 in the cache */
   /* replace934_0_1_1 -> 2755 in the cache */
   /* replace934_0_1 -> 2756 in the cache */
   /* replace934_0 -> 2757 in the cache */
   /* replace934_1_0_0_0 -> 2653 in the cache */
   /* replace934_1_0_0_1 -> 2654 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      4, 1,
      { 2653, 2654 },
      -1,
   } },
   /* replace934_1_0_1_0 -> 2656 in the cache */
   /* replace934_1_0_1_1 -> 2657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      5, 1,
      { 2656, 2657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2759, 2760 },
      -1,
   } },
   /* replace934_1_1_0_0 -> 2660 in the cache */
   /* replace934_1_1_0_1 -> 2661 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      6, 1,
      { 2660, 2661 },
      -1,
   } },
   /* replace934_1_1_1_0 -> 2663 in the cache */
   /* replace934_1_1_1_1 -> 2664 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      7, 1,
      { 2663, 2664 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2762, 2763 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2761, 2764 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2757, 2765 },
      -1,
   } },

   /* ('bany_inequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('ine', 'a.i', 'b.i'), ('ine', 'a.j', 'b.j')), ('ior', ('ine', 'a.k', 'b.k'), ('ine', 'a.l', 'b.l'))), ('ior', ('ior', ('ine', 'a.m', 'b.m'), ('ine', 'a.n', 'b.n')), ('ior', ('ine', 'a.o', 'b.o'), ('ine', 'a.p', 'b.p'))))) */
   /* search935_0 -> 0 in the cache */
   /* search935_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace935_0_0_0_0_0 -> 779 in the cache */
   /* replace935_0_0_0_0_1 -> 2633 in the cache */
   /* replace935_0_0_0_0 -> 2744 in the cache */
   /* replace935_0_0_0_1_0 -> 780 in the cache */
   /* replace935_0_0_0_1_1 -> 2635 in the cache */
   /* replace935_0_0_0_1 -> 2745 in the cache */
   /* replace935_0_0_0 -> 2746 in the cache */
   /* replace935_0_0_1_0_0 -> 781 in the cache */
   /* replace935_0_0_1_0_1 -> 2641 in the cache */
   /* replace935_0_0_1_0 -> 2754 in the cache */
   /* replace935_0_0_1_1_0 -> 2647 in the cache */
   /* replace935_0_0_1_1_1 -> 2648 in the cache */
   /* replace935_0_0_1_1 -> 2755 in the cache */
   /* replace935_0_0_1 -> 2756 in the cache */
   /* replace935_0_0 -> 2757 in the cache */
   /* replace935_0_1_0_0_0 -> 2653 in the cache */
   /* replace935_0_1_0_0_1 -> 2654 in the cache */
   /* replace935_0_1_0_0 -> 2759 in the cache */
   /* replace935_0_1_0_1_0 -> 2656 in the cache */
   /* replace935_0_1_0_1_1 -> 2657 in the cache */
   /* replace935_0_1_0_1 -> 2760 in the cache */
   /* replace935_0_1_0 -> 2761 in the cache */
   /* replace935_0_1_1_0_0 -> 2660 in the cache */
   /* replace935_0_1_1_0_1 -> 2661 in the cache */
   /* replace935_0_1_1_0 -> 2762 in the cache */
   /* replace935_0_1_1_1_0 -> 2663 in the cache */
   /* replace935_0_1_1_1_1 -> 2664 in the cache */
   /* replace935_0_1_1_1 -> 2763 in the cache */
   /* replace935_0_1_1 -> 2764 in the cache */
   /* replace935_0_1 -> 2765 in the cache */
   /* replace935_0 -> 2766 in the cache */
   /* replace935_1_0_0_0_0 -> 2670 in the cache */
   /* replace935_1_0_0_0_1 -> 2671 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      8, 1,
      { 2670, 2671 },
      -1,
   } },
   /* replace935_1_0_0_1_0 -> 2673 in the cache */
   /* replace935_1_0_0_1_1 -> 2674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      9, 1,
      { 2673, 2674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2768, 2769 },
      -1,
   } },
   /* replace935_1_0_1_0_0 -> 2677 in the cache */
   /* replace935_1_0_1_0_1 -> 2678 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      10, 1,
      { 2677, 2678 },
      -1,
   } },
   /* replace935_1_0_1_1_0 -> 2680 in the cache */
   /* replace935_1_0_1_1_1 -> 2681 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      11, 1,
      { 2680, 2681 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2771, 2772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2770, 2773 },
      -1,
   } },
   /* replace935_1_1_0_0_0 -> 2685 in the cache */
   /* replace935_1_1_0_0_1 -> 2686 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      12, 1,
      { 2685, 2686 },
      -1,
   } },
   /* replace935_1_1_0_1_0 -> 2688 in the cache */
   /* replace935_1_1_0_1_1 -> 2689 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      13, 1,
      { 2688, 2689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2775, 2776 },
      -1,
   } },
   /* replace935_1_1_1_0_0 -> 2692 in the cache */
   /* replace935_1_1_1_0_1 -> 2693 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      14, 1,
      { 2692, 2693 },
      -1,
   } },
   /* replace935_1_1_1_1_0 -> 2695 in the cache */
   /* replace935_1_1_1_1_1 -> 2696 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      15, 1,
      { 2695, 2696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2778, 2779 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2777, 2780 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2774, 2781 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 16,
      { 2766, 2782 },
      -1,
   } },

   /* ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')) */
   /* search936_0 -> 0 in the cache */
   /* search936_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace936_0_0 -> 779 in the cache */
   /* replace936_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace936_1_0 -> 780 in the cache */
   /* replace936_1_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 780, 2635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2785, 2786 },
      -1,
   } },

   /* ('bany_fnequal3', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('ior', ('fneu', 'a.y', 'b.y'), ('fneu', 'a.z', 'b.z'))) */
   /* search937_0 -> 0 in the cache */
   /* search937_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace937_0_0 -> 779 in the cache */
   /* replace937_0_1 -> 2633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 779, 2633 },
      -1,
   } },
   /* replace937_1_0_0 -> 780 in the cache */
   /* replace937_1_0_1 -> 2635 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 780, 2635 },
      -1,
   } },
   /* replace937_1_1_0 -> 781 in the cache */
   /* replace937_1_1_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 781, 2641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2790, 2791 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2789, 2792 },
      -1,
   } },

   /* ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w'))) */
   /* search938_0 -> 0 in the cache */
   /* search938_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace938_0_0_0 -> 779 in the cache */
   /* replace938_0_0_1 -> 2633 in the cache */
   /* replace938_0_0 -> 2785 in the cache */
   /* replace938_0_1_0 -> 780 in the cache */
   /* replace938_0_1_1 -> 2635 in the cache */
   /* replace938_0_1 -> 2786 in the cache */
   /* replace938_0 -> 2787 in the cache */
   /* replace938_1_0_0 -> 781 in the cache */
   /* replace938_1_0_1 -> 2641 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 781, 2641 },
      -1,
   } },
   /* replace938_1_1_0 -> 2647 in the cache */
   /* replace938_1_1_1 -> 2648 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 2647, 2648 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2795, 2796 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2787, 2797 },
      -1,
   } },

   /* ('bany_fnequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))) */
   /* search939_0 -> 0 in the cache */
   /* search939_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace939_0_0_0_0 -> 779 in the cache */
   /* replace939_0_0_0_1 -> 2633 in the cache */
   /* replace939_0_0_0 -> 2785 in the cache */
   /* replace939_0_0_1_0 -> 780 in the cache */
   /* replace939_0_0_1_1 -> 2635 in the cache */
   /* replace939_0_0_1 -> 2786 in the cache */
   /* replace939_0_0 -> 2787 in the cache */
   /* replace939_0_1_0_0 -> 781 in the cache */
   /* replace939_0_1_0_1 -> 2641 in the cache */
   /* replace939_0_1_0 -> 2795 in the cache */
   /* replace939_0_1_1_0 -> 2647 in the cache */
   /* replace939_0_1_1_1 -> 2648 in the cache */
   /* replace939_0_1_1 -> 2796 in the cache */
   /* replace939_0_1 -> 2797 in the cache */
   /* replace939_0 -> 2798 in the cache */
   /* replace939_1_0_0_0 -> 2653 in the cache */
   /* replace939_1_0_0_1 -> 2654 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      4, 1,
      { 2653, 2654 },
      -1,
   } },
   /* replace939_1_0_1_0 -> 2656 in the cache */
   /* replace939_1_0_1_1 -> 2657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      5, 1,
      { 2656, 2657 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2800, 2801 },
      -1,
   } },
   /* replace939_1_1_0_0 -> 2660 in the cache */
   /* replace939_1_1_0_1 -> 2661 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      6, 1,
      { 2660, 2661 },
      -1,
   } },
   /* replace939_1_1_1_0 -> 2663 in the cache */
   /* replace939_1_1_1_1 -> 2664 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      7, 1,
      { 2663, 2664 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2803, 2804 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2802, 2805 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2798, 2806 },
      -1,
   } },

   /* ('bany_fnequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('fneu', 'a.i', 'b.i'), ('fneu', 'a.j', 'b.j')), ('ior', ('fneu', 'a.k', 'b.k'), ('fneu', 'a.l', 'b.l'))), ('ior', ('ior', ('fneu', 'a.m', 'b.m'), ('fneu', 'a.n', 'b.n')), ('ior', ('fneu', 'a.o', 'b.o'), ('fneu', 'a.p', 'b.p'))))) */
   /* search940_0 -> 0 in the cache */
   /* search940_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace940_0_0_0_0_0 -> 779 in the cache */
   /* replace940_0_0_0_0_1 -> 2633 in the cache */
   /* replace940_0_0_0_0 -> 2785 in the cache */
   /* replace940_0_0_0_1_0 -> 780 in the cache */
   /* replace940_0_0_0_1_1 -> 2635 in the cache */
   /* replace940_0_0_0_1 -> 2786 in the cache */
   /* replace940_0_0_0 -> 2787 in the cache */
   /* replace940_0_0_1_0_0 -> 781 in the cache */
   /* replace940_0_0_1_0_1 -> 2641 in the cache */
   /* replace940_0_0_1_0 -> 2795 in the cache */
   /* replace940_0_0_1_1_0 -> 2647 in the cache */
   /* replace940_0_0_1_1_1 -> 2648 in the cache */
   /* replace940_0_0_1_1 -> 2796 in the cache */
   /* replace940_0_0_1 -> 2797 in the cache */
   /* replace940_0_0 -> 2798 in the cache */
   /* replace940_0_1_0_0_0 -> 2653 in the cache */
   /* replace940_0_1_0_0_1 -> 2654 in the cache */
   /* replace940_0_1_0_0 -> 2800 in the cache */
   /* replace940_0_1_0_1_0 -> 2656 in the cache */
   /* replace940_0_1_0_1_1 -> 2657 in the cache */
   /* replace940_0_1_0_1 -> 2801 in the cache */
   /* replace940_0_1_0 -> 2802 in the cache */
   /* replace940_0_1_1_0_0 -> 2660 in the cache */
   /* replace940_0_1_1_0_1 -> 2661 in the cache */
   /* replace940_0_1_1_0 -> 2803 in the cache */
   /* replace940_0_1_1_1_0 -> 2663 in the cache */
   /* replace940_0_1_1_1_1 -> 2664 in the cache */
   /* replace940_0_1_1_1 -> 2804 in the cache */
   /* replace940_0_1_1 -> 2805 in the cache */
   /* replace940_0_1 -> 2806 in the cache */
   /* replace940_0 -> 2807 in the cache */
   /* replace940_1_0_0_0_0 -> 2670 in the cache */
   /* replace940_1_0_0_0_1 -> 2671 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      8, 1,
      { 2670, 2671 },
      -1,
   } },
   /* replace940_1_0_0_1_0 -> 2673 in the cache */
   /* replace940_1_0_0_1_1 -> 2674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      9, 1,
      { 2673, 2674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2809, 2810 },
      -1,
   } },
   /* replace940_1_0_1_0_0 -> 2677 in the cache */
   /* replace940_1_0_1_0_1 -> 2678 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      10, 1,
      { 2677, 2678 },
      -1,
   } },
   /* replace940_1_0_1_1_0 -> 2680 in the cache */
   /* replace940_1_0_1_1_1 -> 2681 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      11, 1,
      { 2680, 2681 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2812, 2813 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2811, 2814 },
      -1,
   } },
   /* replace940_1_1_0_0_0 -> 2685 in the cache */
   /* replace940_1_1_0_0_1 -> 2686 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      12, 1,
      { 2685, 2686 },
      -1,
   } },
   /* replace940_1_1_0_1_0 -> 2688 in the cache */
   /* replace940_1_1_0_1_1 -> 2689 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      13, 1,
      { 2688, 2689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2816, 2817 },
      -1,
   } },
   /* replace940_1_1_1_0_0 -> 2692 in the cache */
   /* replace940_1_1_1_0_1 -> 2693 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      14, 1,
      { 2692, 2693 },
      -1,
   } },
   /* replace940_1_1_1_1_0 -> 2695 in the cache */
   /* replace940_1_1_1_1_1 -> 2696 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      15, 1,
      { 2695, 2696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2819, 2820 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2818, 2821 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2815, 2822 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 16,
      { 2807, 2823 },
      -1,
   } },

   /* ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search941_0_0 -> 0 in the cache */
   /* search941_0_1 -> 3 in the cache */
   /* search941_0 -> 2576 in the cache */
   /* search941_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2576, 280 },
      -1,
   } },

   /* replace941_0 -> 0 in the cache */
   /* replace941_1 -> 3 in the cache */
   /* replace941 -> 1068 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search942_0_0 -> 0 in the cache */
   /* search942_0_1 -> 3 in the cache */
   /* search942_0 -> 2578 in the cache */
   /* search942_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2578, 280 },
      -1,
   } },

   /* replace942_0 -> 0 in the cache */
   /* replace942_1 -> 3 in the cache */
   /* replace942 -> 1065 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search943_0_0 -> 0 in the cache */
   /* search943_0_1 -> 3 in the cache */
   /* search943_0 -> 2564 in the cache */
   /* search943_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2564, 280 },
      -1,
   } },

   /* replace943_0 -> 0 in the cache */
   /* replace943_1 -> 3 in the cache */
   /* replace943 -> 1062 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search944_0_0 -> 0 in the cache */
   /* search944_0_1 -> 3 in the cache */
   /* search944_0 -> 2566 in the cache */
   /* search944_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2566, 280 },
      -1,
   } },

   /* replace944_0 -> 0 in the cache */
   /* replace944_1 -> 3 in the cache */
   /* replace944 -> 1059 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search945_0_0 -> 0 in the cache */
   /* search945_0_1 -> 3 in the cache */
   /* search945_0 -> 2576 in the cache */
   /* search945_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2576, 266 },
      -1,
   } },

   /* replace945_0 -> 0 in the cache */
   /* replace945_1 -> 3 in the cache */
   /* replace945 -> 1068 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search946_0_0 -> 0 in the cache */
   /* search946_0_1 -> 3 in the cache */
   /* search946_0 -> 2578 in the cache */
   /* search946_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2578, 266 },
      -1,
   } },

   /* replace946_0 -> 0 in the cache */
   /* replace946_1 -> 3 in the cache */
   /* replace946 -> 1065 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search947_0_0 -> 0 in the cache */
   /* search947_0_1 -> 3 in the cache */
   /* search947_0 -> 2564 in the cache */
   /* search947_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2564, 266 },
      -1,
   } },

   /* replace947_0 -> 0 in the cache */
   /* replace947_1 -> 3 in the cache */
   /* replace947 -> 1062 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search948_0_0 -> 0 in the cache */
   /* search948_0_1 -> 3 in the cache */
   /* search948_0 -> 2566 in the cache */
   /* search948_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2566, 266 },
      -1,
   } },

   /* replace948_0 -> 0 in the cache */
   /* replace948_1 -> 3 in the cache */
   /* replace948 -> 1059 in the cache */

   /* ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search949_0_0 -> 0 in the cache */
   /* search949_0_1 -> 3 in the cache */
   /* search949_0 -> 2576 in the cache */
   /* search949_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2576, 266 },
      -1,
   } },

   /* replace949_0 -> 0 in the cache */
   /* replace949_1 -> 3 in the cache */
   /* replace949 -> 1065 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search950_0_0 -> 0 in the cache */
   /* search950_0_1 -> 3 in the cache */
   /* search950_0 -> 2578 in the cache */
   /* search950_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2578, 266 },
      -1,
   } },

   /* replace950_0 -> 0 in the cache */
   /* replace950_1 -> 3 in the cache */
   /* replace950 -> 1068 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search951_0_0 -> 0 in the cache */
   /* search951_0_1 -> 3 in the cache */
   /* search951_0 -> 2564 in the cache */
   /* search951_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2564, 266 },
      -1,
   } },

   /* replace951_0 -> 0 in the cache */
   /* replace951_1 -> 3 in the cache */
   /* replace951 -> 1059 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search952_0_0 -> 0 in the cache */
   /* search952_0_1 -> 3 in the cache */
   /* search952_0 -> 2566 in the cache */
   /* search952_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2566, 266 },
      -1,
   } },

   /* replace952_0 -> 0 in the cache */
   /* replace952_1 -> 3 in the cache */
   /* replace952 -> 1062 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search953_0_0 -> 0 in the cache */
   /* search953_0_1 -> 3 in the cache */
   /* search953_0 -> 2576 in the cache */
   /* search953_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2576, 280 },
      -1,
   } },

   /* replace953_0 -> 0 in the cache */
   /* replace953_1 -> 3 in the cache */
   /* replace953 -> 1065 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search954_0_0 -> 0 in the cache */
   /* search954_0_1 -> 3 in the cache */
   /* search954_0 -> 2578 in the cache */
   /* search954_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2578, 280 },
      -1,
   } },

   /* replace954_0 -> 0 in the cache */
   /* replace954_1 -> 3 in the cache */
   /* replace954 -> 1068 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search955_0_0 -> 0 in the cache */
   /* search955_0_1 -> 3 in the cache */
   /* search955_0 -> 2564 in the cache */
   /* search955_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2564, 280 },
      -1,
   } },

   /* replace955_0 -> 0 in the cache */
   /* replace955_1 -> 3 in the cache */
   /* replace955 -> 1059 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search956_0_0 -> 0 in the cache */
   /* search956_0_1 -> 3 in the cache */
   /* search956_0 -> 2566 in the cache */
   /* search956_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2566, 280 },
      -1,
   } },

   /* replace956_0 -> 0 in the cache */
   /* replace956_1 -> 3 in the cache */
   /* replace956 -> 1062 in the cache */

   /* ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0) */
   /* search957_0_0 -> 32 in the cache */
   /* search957_0 -> 107 in the cache */
   /* search957_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 107, 32 },
      -1,
   } },

   /* replace957_0 -> 32 in the cache */
   /* replace957_1 -> 117 in the cache */
   /* replace957 -> 1266 in the cache */

   /* ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0) */
   /* search958_0_0 -> 32 in the cache */
   /* search958_0 -> 107 in the cache */
   /* search958_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 107, 32 },
      -1,
   } },

   /* replace958_0 -> 32 in the cache */
   /* replace958_1 -> 117 in the cache */
   /* replace958 -> 1270 in the cache */

   /* ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search959_0_0 -> 770 in the cache */
   /* search959_0 -> 1241 in the cache */
   /* search959_1_0 -> 1147 in the cache */
   /* search959_1 -> 2417 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1241, 2417 },
      -1,
   } },

   /* replace959_0_0 -> 770 in the cache */
   /* replace959_0_1 -> 1147 in the cache */
   /* replace959_0 -> 1159 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1159 },
      -1,
   } },

   /* ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search960_0_0 -> 770 in the cache */
   /* search960_0 -> 1241 in the cache */
   /* search960_1_0 -> 1147 in the cache */
   /* search960_1 -> 2417 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1241, 2417 },
      -1,
   } },

   /* replace960_0_0 -> 770 in the cache */
   /* replace960_0_1 -> 1147 in the cache */
   /* replace960_0 -> 1159 in the cache */
   /* replace960 -> 2844 in the cache */

   /* ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b')) */
   /* search961_0_0 -> 770 in the cache */
   /* search961_0 -> 1241 in the cache */
   /* search961_1_0 -> 1147 in the cache */
   /* search961_1 -> 2417 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1241, 2417 },
      -1,
   } },

   /* replace961_0_0 -> 770 in the cache */
   /* replace961_0_1 -> 1147 in the cache */
   /* replace961_0 -> 1151 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1151 },
      -1,
   } },

   /* ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search962_0_0 -> 770 in the cache */
   /* search962_0 -> 1142 in the cache */
   /* search962_1_0 -> 1147 in the cache */
   /* search962_1 -> 1148 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1142, 1148 },
      -1,
   } },

   /* replace962_0_0 -> 770 in the cache */
   /* replace962_0_1 -> 1147 in the cache */
   /* replace962_0 -> 1159 in the cache */
   /* replace962 -> 1279 in the cache */

   /* ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b')) */
   /* search963_0_0_0 -> 770 in the cache */
   /* search963_0_0 -> 1142 in the cache */
   /* search963_0_1_0 -> 1147 in the cache */
   /* search963_0_1 -> 1148 in the cache */
   /* search963_0 -> 1178 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1178 },
      -1,
   } },

   /* replace963_0_0 -> 770 in the cache */
   /* replace963_0_1 -> 1147 in the cache */
   /* replace963_0 -> 1151 in the cache */
   /* replace963 -> 1276 in the cache */

   /* ('iand', 'a@bool16', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search964_1 -> 491 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2850, 491 },
      -1,
   } },

   /* replace964_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iand', 'a@bool32', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search965_1 -> 254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2853, 254 },
      -1,
   } },

   /* replace965_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('flt', ('fneg', ('b2f', 'a@1')), 0) => a */
   /* search966_0_0_0 -> 770 in the cache */
   /* search966_0_0 -> 1142 in the cache */
   /* search966_0 -> 1145 in the cache */
   /* search966_1 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1145, 1242 },
      -1,
   } },

   /* replace966 -> 770 in the cache */

   /* ('ilt', 'a', 'a') => False */
   /* search967_0 -> 32 in the cache */
   /* search967_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace967 -> 1090 in the cache */

   /* ('ige', 'a', 'a') => True */
   /* search968_0 -> 32 in the cache */
   /* search968_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace968 -> 1248 in the cache */

   /* ('ieq', 'a', 'a') => True */
   /* search969_0 -> 32 in the cache */
   /* search969_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace969 -> 1248 in the cache */

   /* ('ine', 'a', 'a') => False */
   /* search970_0 -> 32 in the cache */
   /* search970_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace970 -> 1090 in the cache */

   /* ('ult', 'a', 'a') => False */
   /* search971_0 -> 32 in the cache */
   /* search971_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace971 -> 1090 in the cache */

   /* ('uge', 'a', 'a') => True */
   /* search972_0 -> 32 in the cache */
   /* search972_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace972 -> 1248 in the cache */

   /* ('flt', 'a', 'a') => False */
   /* search973_0 -> 32 in the cache */
   /* search973_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace973 -> 1090 in the cache */

   /* ('fge', 'a(is_a_number)', 'a') => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search974_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2864, 32 },
      -1,
   } },

   /* replace974 -> 1248 in the cache */

   /* ('feq', 'a(is_a_number)', 'a') => True */
   /* search975_0 -> 2864 in the cache */
   /* search975_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 2864, 32 },
      -1,
   } },

   /* replace975 -> 1248 in the cache */

   /* ('fneu', 'a(is_a_number)', 'a') => False */
   /* search976_0 -> 2864 in the cache */
   /* search976_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 2864, 32 },
      -1,
   } },

   /* replace976 -> 1090 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search977_0 -> 32 in the cache */
   /* search977_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace977 -> 32 in the cache */

   /* ('iand', 'a', 0) => 0 */
   /* search978_0 -> 32 in the cache */
   /* search978_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace978 -> 61 in the cache */

   /* ('iand', 'a', -1) => a */
   /* search979_0 -> 32 in the cache */
   /* search979_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace979 -> 32 in the cache */

   /* ('iand', 'a', ('inot', 'a')) => 0 */
   /* search980_0 -> 32 in the cache */
   /* search980_1_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 2871 },
      -1,
   } },

   /* replace980 -> 61 in the cache */

   /* ('ior', 'a', 'a') => a */
   /* search981_0 -> 32 in the cache */
   /* search981_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace981 -> 32 in the cache */

   /* ('ior', 'a', 0) => a */
   /* search982_0 -> 32 in the cache */
   /* search982_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace982 -> 32 in the cache */

   /* ('ior', 'a', -1) => -1 */
   /* search983_0 -> 32 in the cache */
   /* search983_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace983 -> 63 in the cache */

   /* ('ior', 'a', ('inot', 'a')) => -1 */
   /* search984_0 -> 32 in the cache */
   /* search984_1_0 -> 32 in the cache */
   /* search984_1 -> 2871 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 2871 },
      -1,
   } },

   /* replace984 -> 63 in the cache */

   /* ('ixor', 'a', 'a') => 0 */
   /* search985_0 -> 32 in the cache */
   /* search985_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace985 -> 61 in the cache */

   /* ('ixor', 'a', 0) => a */
   /* search986_0 -> 32 in the cache */
   /* search986_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace986 -> 32 in the cache */

   /* ('ixor', 'a', ('ixor', 'a', 'b')) => b */
   /* search987_0 -> 0 in the cache */
   /* search987_1_0 -> 0 in the cache */
   /* search987_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 0, 2879 },
      -1,
   } },

   /* replace987 -> 3 in the cache */

   /* ('ixor', 'a', -1) => ('inot', 'a') */
   /* search988_0 -> 32 in the cache */
   /* search988_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace988_0 -> 32 in the cache */
   /* replace988 -> 2871 in the cache */

   /* ('inot', ('inot', 'a')) => a */
   /* search989_0_0 -> 32 in the cache */
   /* search989_0 -> 2871 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 2871 },
      -1,
   } },

   /* replace989 -> 32 in the cache */

   /* ('ior', ('iand', 'a', 'b'), 'b') => b */
   /* search990_0_0 -> 0 in the cache */
   /* search990_0_1 -> 3 in the cache */
   /* search990_0 -> 207 in the cache */
   /* search990_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 207, 3 },
      -1,
   } },

   /* replace990 -> 3 in the cache */

   /* ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b') */
   /* search991_0_0 -> 0 in the cache */
   /* search991_0_1 -> 3 in the cache */
   /* search991_0 -> 84 in the cache */
   /* search991_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace991_0 -> 0 in the cache */
   /* replace991_1 -> 3 in the cache */
   /* replace991 -> 2506 in the cache */

   /* ('iand', ('ior', 'a', 'b'), 'b') => b */
   /* search992_0_0 -> 0 in the cache */
   /* search992_0_1 -> 3 in the cache */
   /* search992_0 -> 84 in the cache */
   /* search992_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace992 -> 3 in the cache */

   /* ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b') */
   /* search993_0_0 -> 0 in the cache */
   /* search993_0_1 -> 3 in the cache */
   /* search993_0 -> 207 in the cache */
   /* search993_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 207, 3 },
      -1,
   } },

   /* replace993_0 -> 0 in the cache */
   /* replace993_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('iand(is_used_once)', 'a', 'b'), ('iand(is_used_once)', 'a', 'c')) => ('iand', 'a', ('iand', 'b', 'c')) */
   /* search994_0_0 -> 132 in the cache */
   /* search994_0_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 132, 133 },
      0,
   } },
   /* search994_1_0 -> 132 in the cache */
   /* search994_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 132, 135 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2888, 2889 },
      -1,
   } },

   /* replace994_0 -> 132 in the cache */
   /* replace994_1_0 -> 133 in the cache */
   /* replace994_1_1 -> 135 in the cache */
   /* replace994_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 132, 198 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', 0, ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search995_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      20,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2892 },
      -1,
   } },

   /* replace995_0 -> 128 in the cache */
   /* replace995_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace995_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2894, 2895 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 128, 2896 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), 0) */
   /* search996_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2898 },
      -1,
   } },

   /* replace996_0_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace996_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2900, 2901 },
      -1,
   } },
   /* replace996_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2902, 128 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search997_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      22,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2904 },
      -1,
   } },

   /* replace997_0_0 -> 21 in the cache */
   /* replace997_0 -> 2900 in the cache */
   /* replace997_1_0_0 -> 21 in the cache */
   /* replace997_1_0 -> 2894 in the cache */
   /* replace997_1_1_0 -> 41 in the cache */
   /* replace997_1_1 -> 2895 in the cache */
   /* replace997_1 -> 2896 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2900, 2896 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search998_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      23,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2907 },
      -1,
   } },

   /* replace998_0_0_0 -> 21 in the cache */
   /* replace998_0_0 -> 2900 in the cache */
   /* replace998_0_1_0 -> 41 in the cache */
   /* replace998_0_1 -> 2901 in the cache */
   /* replace998_0 -> 2902 in the cache */
   /* replace998_1_0 -> 21 in the cache */
   /* replace998_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2902, 2894 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search999_0 -> 21 in the cache */
   /* search999_1 -> 2892 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2892 },
      -1,
   } },

   /* replace999_0_0 -> 21 in the cache */
   /* replace999_0 -> 2900 in the cache */
   /* replace999_1_0_0 -> 21 in the cache */
   /* replace999_1_0 -> 2894 in the cache */
   /* replace999_1_1_0 -> 41 in the cache */
   /* replace999_1_1 -> 2895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2894, 2895 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2900, 2911 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search1000_0 -> 21 in the cache */
   /* search1000_1 -> 2898 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2898 },
      -1,
   } },

   /* replace1000_0_0_0 -> 21 in the cache */
   /* replace1000_0_0 -> 2900 in the cache */
   /* replace1000_0_1_0 -> 41 in the cache */
   /* replace1000_0_1 -> 2901 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2900, 2901 },
      -1,
   } },
   /* replace1000_1_0 -> 21 in the cache */
   /* replace1000_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2914, 2894 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', -1, ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search1001_0 -> 21 in the cache */
   /* search1001_1 -> 2904 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2904 },
      -1,
   } },

   /* replace1001_0 -> 130 in the cache */
   /* replace1001_1_0_0 -> 21 in the cache */
   /* replace1001_1_0 -> 2894 in the cache */
   /* replace1001_1_1_0 -> 41 in the cache */
   /* replace1001_1_1 -> 2895 in the cache */
   /* replace1001_1 -> 2911 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 130, 2911 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), -1) */
   /* search1002_0 -> 21 in the cache */
   /* search1002_1 -> 2907 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2907 },
      -1,
   } },

   /* replace1002_0_0_0 -> 21 in the cache */
   /* replace1002_0_0 -> 2900 in the cache */
   /* replace1002_0_1_0 -> 41 in the cache */
   /* replace1002_0_1 -> 2901 in the cache */
   /* replace1002_0 -> 2914 in the cache */
   /* replace1002_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2914, 130 },
      -1,
   } },

   /* ('ixor@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ixor', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search1003_0 -> 21 in the cache */
   /* search1003_1 -> 2892 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 21, 2892 },
      -1,
   } },

   /* replace1003_0_0 -> 21 in the cache */
   /* replace1003_0 -> 2900 in the cache */
   /* replace1003_1_0_0 -> 21 in the cache */
   /* replace1003_1_0 -> 2894 in the cache */
   /* replace1003_1_1_0 -> 41 in the cache */
   /* replace1003_1_1 -> 2895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2894, 2895 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2900, 2921 },
      -1,
   } },

   /* ('ixor@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ixor', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search1004_0 -> 21 in the cache */
   /* search1004_1 -> 2898 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 21, 2898 },
      -1,
   } },

   /* replace1004_0_0_0 -> 21 in the cache */
   /* replace1004_0_0 -> 2900 in the cache */
   /* replace1004_0_1_0 -> 41 in the cache */
   /* replace1004_0_1 -> 2901 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2900, 2901 },
      -1,
   } },
   /* replace1004_1_0 -> 21 in the cache */
   /* replace1004_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2924, 2894 },
      -1,
   } },

   /* ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b')) */
   /* search1005_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search1005_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2926, 2927 },
      -1,
   } },

   /* replace1005_0_0 -> 0 in the cache */
   /* replace1005_0_1 -> 3 in the cache */
   /* replace1005_0 -> 2506 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2506 },
      -1,
   } },

   /* ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b')) */
   /* search1006_0_0 -> 0 in the cache */
   /* search1006_0 -> 2926 in the cache */
   /* search1006_1_0 -> 3 in the cache */
   /* search1006_1 -> 2927 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2926, 2927 },
      -1,
   } },

   /* replace1006_0_0 -> 0 in the cache */
   /* replace1006_0_1 -> 3 in the cache */
   /* replace1006_0 -> 2887 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2887 },
      -1,
   } },

   /* ('ishl', 0, 'a') => 0 */
   /* search1007_0 -> 1242 in the cache */
   /* search1007_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 1242, 16 },
      -1,
   } },

   /* replace1007 -> 1242 in the cache */

   /* ('ishl', 'a', 0) => a */
   /* search1008_0 -> 32 in the cache */
   /* search1008_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 128 },
      -1,
   } },

   /* replace1008 -> 32 in the cache */

   /* ('ishr', 0, 'a') => 0 */
   /* search1009_0 -> 1242 in the cache */
   /* search1009_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 1242, 16 },
      -1,
   } },

   /* replace1009 -> 1242 in the cache */

   /* ('ishr', -1, 'a') => -1 */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* search1010_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 2935, 16 },
      -1,
   } },

   /* replace1010 -> 2935 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search1011_0 -> 32 in the cache */
   /* search1011_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 128 },
      -1,
   } },

   /* replace1011 -> 32 in the cache */

   /* ('ushr', 0, 'a') => 0 */
   /* search1012_0 -> 1242 in the cache */
   /* search1012_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 1242, 16 },
      -1,
   } },

   /* replace1012 -> 1242 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search1013_0 -> 32 in the cache */
   /* search1013_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 128 },
      -1,
   } },

   /* replace1013 -> 32 in the cache */

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search1014_0_0 -> 11 in the cache */
   /* search1014_0_1 -> 36 in the cache */
   /* search1014_0 -> 2332 in the cache */
   /* search1014_1_0 -> 11 in the cache */
   /* search1014_1_1_0 -> 866 in the cache */
   /* search1014_1_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 866, 2940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2941 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2332, 2942 },
      -1,
   } },

   /* replace1014_0 -> 11 in the cache */
   /* replace1014_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b') */
   /* search1015_0_0 -> 11 in the cache */
   /* search1015_0_1 -> 36 in the cache */
   /* search1015_0 -> 2332 in the cache */
   /* search1015_1_0 -> 11 in the cache */
   /* search1015_1_1_0 -> 866 in the cache */
   /* search1015_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 866, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2945 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2332, 2946 },
      -1,
   } },

   /* replace1015_0 -> 11 in the cache */
   /* replace1015_1 -> 36 in the cache */
   /* replace1015 -> 2944 in the cache */

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search1016_0_0 -> 16 in the cache */
   /* search1016_0_1 -> 36 in the cache */
   /* search1016_0 -> 959 in the cache */
   /* search1016_1_0 -> 16 in the cache */
   /* search1016_1_1_0 -> 887 in the cache */
   /* search1016_1_1_1_0 -> 36 in the cache */
   /* search1016_1_1_1 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 887, 2940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2948 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 959, 2949 },
      -1,
   } },

   /* replace1016_0 -> 16 in the cache */
   /* replace1016_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b') */
   /* search1017_0_0 -> 16 in the cache */
   /* search1017_0_1 -> 36 in the cache */
   /* search1017_0 -> 959 in the cache */
   /* search1017_1_0 -> 16 in the cache */
   /* search1017_1_1_0 -> 887 in the cache */
   /* search1017_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 887, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2952 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 959, 2953 },
      -1,
   } },

   /* replace1017_0 -> 16 in the cache */
   /* replace1017_1 -> 36 in the cache */
   /* replace1017 -> 2951 in the cache */

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search1018_0_0 -> 11 in the cache */
   /* search1018_0_1 -> 36 in the cache */
   /* search1018_0 -> 2336 in the cache */
   /* search1018_1_0 -> 11 in the cache */
   /* search1018_1_1_0 -> 866 in the cache */
   /* search1018_1_1_1_0 -> 36 in the cache */
   /* search1018_1_1_1 -> 2940 in the cache */
   /* search1018_1_1 -> 2941 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2941 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2336, 2955 },
      -1,
   } },

   /* replace1018_0 -> 11 in the cache */
   /* replace1018_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b') */
   /* search1019_0_0 -> 11 in the cache */
   /* search1019_0_1 -> 36 in the cache */
   /* search1019_0 -> 2336 in the cache */
   /* search1019_1_0 -> 11 in the cache */
   /* search1019_1_1_0 -> 866 in the cache */
   /* search1019_1_1_1 -> 36 in the cache */
   /* search1019_1_1 -> 2945 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 2945 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2336, 2958 },
      -1,
   } },

   /* replace1019_0 -> 11 in the cache */
   /* replace1019_1 -> 36 in the cache */
   /* replace1019 -> 2957 in the cache */

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search1020_0_0 -> 16 in the cache */
   /* search1020_0_1 -> 36 in the cache */
   /* search1020_0 -> 2365 in the cache */
   /* search1020_1_0 -> 16 in the cache */
   /* search1020_1_1_0 -> 887 in the cache */
   /* search1020_1_1_1_0 -> 36 in the cache */
   /* search1020_1_1_1 -> 2940 in the cache */
   /* search1020_1_1 -> 2948 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2948 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2365, 2960 },
      -1,
   } },

   /* replace1020_0 -> 16 in the cache */
   /* replace1020_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b') */
   /* search1021_0_0 -> 16 in the cache */
   /* search1021_0_1 -> 36 in the cache */
   /* search1021_0 -> 2365 in the cache */
   /* search1021_1_0 -> 16 in the cache */
   /* search1021_1_1_0 -> 887 in the cache */
   /* search1021_1_1_1 -> 36 in the cache */
   /* search1021_1_1 -> 2952 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2952 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2365, 2963 },
      -1,
   } },

   /* replace1021_0 -> 16 in the cache */
   /* replace1021_1 -> 36 in the cache */
   /* replace1021 -> 2962 in the cache */

   /* ('urol@8', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 8, 'b'))) */
   /* search1022_0 -> 6 in the cache */
   /* search1022_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* replace1022_0_0 -> 6 in the cache */
   /* replace1022_0_1 -> 36 in the cache */
   /* replace1022_0 -> 2301 in the cache */
   /* replace1022_1_0 -> 6 in the cache */
   /* replace1022_1_1_0 -> 844 in the cache */
   /* replace1022_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 844, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 2966 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2301, 2967 },
      -1,
   } },

   /* ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b'))) */
   /* search1023_0 -> 11 in the cache */
   /* search1023_1 -> 36 in the cache */
   /* search1023 -> 2944 in the cache */

   /* replace1023_0_0 -> 11 in the cache */
   /* replace1023_0_1 -> 36 in the cache */
   /* replace1023_0 -> 2332 in the cache */
   /* replace1023_1_0 -> 11 in the cache */
   /* replace1023_1_1_0 -> 866 in the cache */
   /* replace1023_1_1_1 -> 36 in the cache */
   /* replace1023_1_1 -> 2945 in the cache */
   /* replace1023_1 -> 2946 in the cache */
   /* replace1023 -> 2947 in the cache */

   /* ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b'))) */
   /* search1024_0 -> 16 in the cache */
   /* search1024_1 -> 36 in the cache */
   /* search1024 -> 2951 in the cache */

   /* replace1024_0_0 -> 16 in the cache */
   /* replace1024_0_1 -> 36 in the cache */
   /* replace1024_0 -> 959 in the cache */
   /* replace1024_1_0 -> 16 in the cache */
   /* replace1024_1_1_0 -> 887 in the cache */
   /* replace1024_1_1_1 -> 36 in the cache */
   /* replace1024_1_1 -> 2952 in the cache */
   /* replace1024_1 -> 2953 in the cache */
   /* replace1024 -> 2954 in the cache */

   /* ('urol@64', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 64, 'b'))) */
   /* search1025_0 -> 21 in the cache */
   /* search1025_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* replace1025_0_0 -> 21 in the cache */
   /* replace1025_0_1 -> 36 in the cache */
   /* replace1025_0 -> 2392 in the cache */
   /* replace1025_1_0 -> 21 in the cache */
   /* replace1025_1_1_0 -> 907 in the cache */
   /* replace1025_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 907, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2970 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2392, 2971 },
      -1,
   } },

   /* ('uror@8', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 8, 'b'))) */
   /* search1026_0 -> 6 in the cache */
   /* search1026_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* replace1026_0_0 -> 6 in the cache */
   /* replace1026_0_1 -> 36 in the cache */
   /* replace1026_0 -> 2305 in the cache */
   /* replace1026_1_0 -> 6 in the cache */
   /* replace1026_1_1_0 -> 844 in the cache */
   /* replace1026_1_1_1 -> 36 in the cache */
   /* replace1026_1_1 -> 2966 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 2966 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2305, 2974 },
      -1,
   } },

   /* ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b'))) */
   /* search1027_0 -> 11 in the cache */
   /* search1027_1 -> 36 in the cache */
   /* search1027 -> 2957 in the cache */

   /* replace1027_0_0 -> 11 in the cache */
   /* replace1027_0_1 -> 36 in the cache */
   /* replace1027_0 -> 2336 in the cache */
   /* replace1027_1_0 -> 11 in the cache */
   /* replace1027_1_1_0 -> 866 in the cache */
   /* replace1027_1_1_1 -> 36 in the cache */
   /* replace1027_1_1 -> 2945 in the cache */
   /* replace1027_1 -> 2958 in the cache */
   /* replace1027 -> 2959 in the cache */

   /* ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b'))) */
   /* search1028_0 -> 16 in the cache */
   /* search1028_1 -> 36 in the cache */
   /* search1028 -> 2962 in the cache */

   /* replace1028_0_0 -> 16 in the cache */
   /* replace1028_0_1 -> 36 in the cache */
   /* replace1028_0 -> 2365 in the cache */
   /* replace1028_1_0 -> 16 in the cache */
   /* replace1028_1_1_0 -> 887 in the cache */
   /* replace1028_1_1_1 -> 36 in the cache */
   /* replace1028_1_1 -> 2952 in the cache */
   /* replace1028_1 -> 2963 in the cache */
   /* replace1028 -> 2964 in the cache */

   /* ('uror@64', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 64, 'b'))) */
   /* search1029_0 -> 21 in the cache */
   /* search1029_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* replace1029_0_0 -> 21 in the cache */
   /* replace1029_0_1 -> 36 in the cache */
   /* replace1029_0 -> 2396 in the cache */
   /* replace1029_1_0 -> 21 in the cache */
   /* replace1029_1_1_0 -> 907 in the cache */
   /* replace1029_1_1_1 -> 36 in the cache */
   /* replace1029_1_1 -> 2970 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 2970 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2396, 2977 },
      -1,
   } },

   /* ('bfi', 4294967295, 'a', 'b') => a */
   /* search1030_0 -> 940 in the cache */
   /* search1030_1 -> 16 in the cache */
   /* search1030_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 940, 16, 36 },
      -1,
   } },

   /* replace1030 -> 16 in the cache */

   /* ('bfi', 0, 'a', 'b') => b */
   /* search1031_0 -> 128 in the cache */
   /* search1031_1 -> 16 in the cache */
   /* search1031_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 128, 16, 36 },
      -1,
   } },

   /* replace1031 -> 36 in the cache */

   /* ('bfi', ('ineg', ('b2i', 'a@1')), 'b', 'c') => ('bcsel', 'a', 'b', 'c') */
   /* search1032_0_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2981 },
      -1,
   } },
   /* search1032_1 -> 36 in the cache */
   /* search1032_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 2982, 36, 141 },
      -1,
   } },

   /* replace1032_0 -> 770 in the cache */
   /* replace1032_1 -> 36 in the cache */
   /* replace1032_2 -> 141 in the cache */
   /* replace1032 -> 1919 in the cache */

   /* ('bfi', '#a(is_odd)', 'a', 'b') => ('ior', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      24,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1033_1 -> 16 in the cache */
   /* search1033_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 2984, 16, 36 },
      -1,
   } },

   /* replace1033_0 -> 16 in the cache */
   /* replace1033_1 -> 36 in the cache */
   /* replace1033 -> 2237 in the cache */

   /* ('bfi', '#a(is_odd)', 'b', 0) => ('iand', 'a', 'b') */
   /* search1034_0 -> 2984 in the cache */
   /* search1034_1 -> 36 in the cache */
   /* search1034_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 2984, 36, 128 },
      -1,
   } },

   /* replace1034_0 -> 16 in the cache */
   /* replace1034_1 -> 36 in the cache */
   /* replace1034 -> 2199 in the cache */

   /* ('u2f32', ('bfi', '#a(is_pos_power_of_two)', 'b', 0)) => ('bcsel', ('ieq', ('iand', 'b', 1), 0), ('iand', 'b', 1), ('u2f', 'a')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1035_0_1 -> 36 in the cache */
   /* search1035_0_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 2987, 36, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 2988 },
      -1,
   } },

   /* replace1035_0_0_0 -> 36 in the cache */
   /* replace1035_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 406 },
      -1,
   } },
   /* replace1035_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2990, 128 },
      -1,
   } },
   /* replace1035_1_0 -> 36 in the cache */
   /* replace1035_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 406 },
      -1,
   } },
   /* replace1035_2_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 2991, 2992, 2993 },
      -1,
   } },

   /* ('u2f', ('bfi', '#a(is_pos_power_of_two)', 'b', 0)) => ('bcsel', ('ieq', ('iand', 'b', 1), 0), 0, ('u2f', 'a')) */
   /* search1036_0_0 -> 2987 in the cache */
   /* search1036_0_1 -> 36 in the cache */
   /* search1036_0_2 -> 128 in the cache */
   /* search1036_0 -> 2988 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 2988 },
      -1,
   } },

   /* replace1036_0_0_0 -> 36 in the cache */
   /* replace1036_0_0_1 -> 406 in the cache */
   /* replace1036_0_0 -> 2990 in the cache */
   /* replace1036_0_1 -> 128 in the cache */
   /* replace1036_0 -> 2991 in the cache */
   /* replace1036_1 -> 1242 in the cache */
   /* replace1036_2_0 -> 16 in the cache */
   /* replace1036_2 -> 1925 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 2991, 1242, 1925 },
      -1,
   } },

   /* ('~fexp2', ('flog2', 'a')) => a */
   /* search1037_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 2997 },
      -1,
   } },

   /* replace1037 -> 32 in the cache */

   /* ('~flog2', ('fexp2', 'a')) => a */
   /* search1038_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2999 },
      -1,
   } },

   /* replace1038 -> 32 in the cache */

   /* ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b')) */
   /* search1039_0 -> 0 in the cache */
   /* search1039_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1039_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace1039_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3002, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3003 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b') */
   /* search1040_0_0_0 -> 0 in the cache */
   /* search1040_0_0 -> 3002 in the cache */
   /* search1040_0_1 -> 3 in the cache */
   /* search1040_0 -> 3003 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3003 },
      -1,
   } },

   /* replace1040_0 -> 0 in the cache */
   /* replace1040_1 -> 3 in the cache */
   /* replace1040 -> 3001 in the cache */

   /* ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd')) */
   /* search1041_0_0_0_0 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 155 },
      -1,
   } },
   /* search1041_0_0_1 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3006, 156 },
      -1,
   } },
   /* search1041_0_1_0_0 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 158 },
      -1,
   } },
   /* search1041_0_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 3008, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 3007, 3009 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 3,
      { 3010 },
      -1,
   } },

   /* replace1041_0_0 -> 155 in the cache */
   /* replace1041_0_1 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 155, 156 },
      -1,
   } },
   /* replace1041_1_0 -> 158 in the cache */
   /* replace1041_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3012, 3013 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a') */
   /* search1042_0_0_0 -> 32 in the cache */
   /* search1042_0_0 -> 2997 in the cache */
   /* search1042_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2997, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3015 },
      -1,
   } },

   /* replace1042_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a') */
   /* search1043_0_0_0 -> 32 in the cache */
   /* search1043_0_0 -> 2997 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2997, 3018 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3019 },
      -1,
   } },

   /* replace1043_0 -> 32 in the cache */
   /* replace1043_1 -> 32 in the cache */
   /* replace1043 -> 261 in the cache */

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search1044_0_0_0 -> 32 in the cache */
   /* search1044_0_0 -> 2997 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4010000000000000 /* 4.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2997, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3022 },
      -1,
   } },

   /* replace1044_0_0 -> 32 in the cache */
   /* replace1044_0_1 -> 32 in the cache */
   /* replace1044_0 -> 261 in the cache */
   /* replace1044_1_0 -> 32 in the cache */
   /* replace1044_1_1 -> 32 in the cache */
   /* replace1044_1 -> 261 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 261, 261 },
      -1,
   } },

   /* ('~fpow', 'a', 1.0) => a */
   /* search1045_0 -> 32 in the cache */
   /* search1045_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 252 },
      -1,
   } },

   /* replace1045 -> 32 in the cache */

   /* ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a') */
   /* search1046_0 -> 32 in the cache */
   /* search1046_1 -> 3018 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 3018 },
      -1,
   } },

   /* replace1046_0 -> 32 in the cache */
   /* replace1046_1 -> 32 in the cache */
   /* replace1046 -> 261 in the cache */

   /* ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search1047_0 -> 32 in the cache */
   /* search1047_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 3021 },
      -1,
   } },

   /* replace1047_0_0 -> 32 in the cache */
   /* replace1047_0_1 -> 32 in the cache */
   /* replace1047_0 -> 261 in the cache */
   /* replace1047_1_0 -> 32 in the cache */
   /* replace1047_1_1 -> 32 in the cache */
   /* replace1047_1 -> 261 in the cache */
   /* replace1047 -> 3024 in the cache */

   /* ('~fpow', 2.0, 'a') => ('fexp2', 'a') */
   /* search1048_0 -> 3018 in the cache */
   /* search1048_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 3018, 32 },
      -1,
   } },

   /* replace1048_0 -> 32 in the cache */
   /* replace1048 -> 2999 in the cache */

   /* ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a */
   /* search1049_0_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x400199999999999a /* 2.2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 3029 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fdd1743e963dc48 /* 0.454545 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 3030, 3031 },
      -1,
   } },

   /* replace1049 -> 32 in the cache */

   /* ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a') */
   /* search1050_0_0_0 -> 32 in the cache */
   /* search1050_0_0_1 -> 3029 in the cache */
   /* search1050_0_0 -> 3030 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3030 },
      -1,
   } },
   /* search1050_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 3033, 3031 },
      -1,
   } },

   /* replace1050_0 -> 32 in the cache */
   /* replace1050 -> 99 in the cache */

   /* ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a')) */
   /* search1051_0_0 -> 32 in the cache */
   /* search1051_0 -> 2999 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 2999 },
      -1,
   } },

   /* replace1051_0_0 -> 100 in the cache */
   /* replace1051_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 100, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3036 },
      -1,
   } },

   /* ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a')) */
   /* search1052_0_0 -> 32 in the cache */
   /* search1052_0 -> 2999 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2999 },
      -1,
   } },

   /* replace1052_0_0 -> 32 in the cache */
   /* replace1052_0 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 107 },
      -1,
   } },

   /* ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a')) */
   /* search1053_0_0 -> 32 in the cache */
   /* search1053_0 -> 2999 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 2999 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbfe0000000000000 /* -0.5 */ },
   } },
   /* replace1053_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3041, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 3042 },
      -1,
   } },

   /* ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a')) */
   /* search1054_0_0 -> 32 in the cache */
   /* search1054_0 -> 3017 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 3017 },
      -1,
   } },

   /* replace1054_0 -> 100 in the cache */
   /* replace1054_1_0 -> 32 in the cache */
   /* replace1054_1 -> 2997 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 100, 2997 },
      -1,
   } },

   /* ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a')) */
   /* search1055_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 3046 },
      -1,
   } },

   /* replace1055_0_0 -> 32 in the cache */
   /* replace1055_0 -> 2997 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2997 },
      -1,
   } },

   /* ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a')) */
   /* search1056_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 3049 },
      -1,
   } },

   /* replace1056_0 -> 3041 in the cache */
   /* replace1056_1_0 -> 32 in the cache */
   /* replace1056_1 -> 2997 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3041, 2997 },
      -1,
   } },

   /* ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a')) */
   /* search1057_0_0 -> 0 in the cache */
   /* search1057_0_1 -> 3 in the cache */
   /* search1057_0 -> 3001 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 3001 },
      -1,
   } },

   /* replace1057_0 -> 3 in the cache */
   /* replace1057_1_0 -> 0 in the cache */
   /* replace1057_1 -> 3002 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3002 },
      -1,
   } },

   /* ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b')) */
   /* search1058_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 0 },
      0,
   } },
   /* search1058_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3054, 3055 },
      -1,
   } },

   /* replace1058_0_0 -> 0 in the cache */
   /* replace1058_0_1 -> 3 in the cache */
   /* replace1058_0 -> 282 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 282 },
      -1,
   } },

   /* ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0)) */
   /* search1059_0_0 -> 32 in the cache */
   /* search1059_0_1 -> 117 in the cache */
   /* search1059_0 -> 1185 in the cache */
   /* search1059_1 -> 117 in the cache */
   /* search1059_2_0 -> 32 in the cache */
   /* search1059_2 -> 3017 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1185, 117, 3017 },
      -1,
   } },

   /* replace1059_0_0 -> 32 in the cache */
   /* replace1059_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 1,
      { 3059 },
      -1,
   } },

   /* ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search1060_0_0 -> 32 in the cache */
   /* search1060_0 -> 3017 in the cache */
   /* search1060_1_0 -> 32 in the cache */
   /* search1060_1 -> 3017 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 3017, 3017 },
      -1,
   } },

   /* replace1060_0 -> 32 in the cache */
   /* replace1060 -> 99 in the cache */

   /* ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search1061_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1061_1_0 -> 16 in the cache */
   /* search1061_1 -> 3062 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 3062, 3062 },
      -1,
   } },

   /* replace1061_0 -> 16 in the cache */
   /* replace1061 -> 706 in the cache */

   /* ('~fdiv', 1.0, 'a') => ('frcp', 'a') */
   /* search1062_0 -> 252 in the cache */
   /* search1062_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 252, 32 },
      -1,
   } },

   /* replace1062_0 -> 32 in the cache */
   /* replace1062 -> 3046 in the cache */

   /* ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b')) */
   /* search1063_0 -> 0 in the cache */
   /* search1063_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1063_0 -> 0 in the cache */
   /* replace1063_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3066 },
      -1,
   } },

   /* ('~frcp', ('frcp', 'a')) => a */
   /* search1064_0_0 -> 32 in the cache */
   /* search1064_0 -> 3046 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3046 },
      -1,
   } },

   /* replace1064 -> 32 in the cache */

   /* ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a') */
   /* search1065_0_0 -> 32 in the cache */
   /* search1065_0 -> 3017 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3017 },
      -1,
   } },

   /* replace1065_0 -> 32 in the cache */
   /* replace1065 -> 3049 in the cache */

   /* ('fsqrt', 'a') => ('frcp', ('frsq', 'a')) */
   /* search1066_0 -> 32 in the cache */
   /* search1066 -> 3017 in the cache */

   /* replace1066_0_0 -> 32 in the cache */
   /* replace1066_0 -> 3049 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3049 },
      -1,
   } },

   /* ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a') */
   /* search1067_0_0 -> 32 in the cache */
   /* search1067_0 -> 3049 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3049 },
      -1,
   } },

   /* replace1067_0 -> 32 in the cache */
   /* replace1067 -> 3017 in the cache */

   /* ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)) */
   /* search1068_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1068_0_0_0_0_0_0 -> 3018 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fc45f306dc9c883 /* 0.15915494309189535 */ },
   } },
   /* replace1068_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      5, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_0_0_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 3074, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3075 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 3018, 3076 },
      -1,
   } },
   /* replace1068_0_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3077, 252 },
      -1,
   } },
   /* replace1068_0_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      9, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_0_0_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 3079, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3080 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 3018, 3081 },
      -1,
   } },
   /* replace1068_0_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3082, 252 },
      -1,
   } },
   /* replace1068_0_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      12, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_0_0_1_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 3084, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3085 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 3018, 3086 },
      -1,
   } },
   /* replace1068_0_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3087, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3088 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 3083, 3089 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3078, 3090 },
      -1,
   } },
   /* replace1068_0_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 3091, 3021 },
      -1,
   } },
   /* replace1068_0_1_0_0_0_0_0 -> 3018 in the cache */
   /* replace1068_0_1_0_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      16, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_1_0_0_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 3093, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3094 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 3018, 3095 },
      -1,
   } },
   /* replace1068_0_1_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3096, 252 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_1_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      20, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 3098, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3099 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 3018, 3100 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3101, 252 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_1_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      23, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 3103, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3104 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 3018, 3105 },
      -1,
   } },
   /* replace1068_0_1_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3106, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3107 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 3102, 3108 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3097, 3109 },
      -1,
   } },
   /* replace1068_0_1_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 3110, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 3111 },
      -1,
   } },
   /* replace1068_0_2_0_0_0_0_0 -> 3018 in the cache */
   /* replace1068_0_2_0_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      27, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_2_0_0_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 3113, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3114 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 3018, 3115 },
      -1,
   } },
   /* replace1068_0_2_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3116, 252 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_2_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      31, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 3118, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 3018, 3120 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3121, 252 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1068_0_2_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      34, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 3123, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 3018, 3125 },
      -1,
   } },
   /* replace1068_0_2_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3126, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 3122, 3128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3117, 3129 },
      -1,
   } },
   /* replace1068_0_2_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 3130, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 3131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 3092, 3112, 3132 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fcccccccccccccd /* 0.225 */ },
   } },
   /* replace1068_2_0_0_0_0 -> 3018 in the cache */
   /* replace1068_2_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_2_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      38, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_2_0_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 3135, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3136 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 3018, 3137 },
      -1,
   } },
   /* replace1068_2_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3138, 252 },
      -1,
   } },
   /* replace1068_2_0_1_0_0_0 -> 3018 in the cache */
   /* replace1068_2_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      42, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_2_0_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 3140, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 3018, 3142 },
      -1,
   } },
   /* replace1068_2_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3143, 252 },
      -1,
   } },
   /* replace1068_2_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1068_2_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1068_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      45, 1,
      { 3073, 32 },
      -1,
   } },
   /* replace1068_2_0_1_1_0_0_1_0_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 3145, 100 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 3018, 3147 },
      -1,
   } },
   /* replace1068_2_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3148, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3149 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 3144, 3150 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3139, 3151 },
      -1,
   } },
   /* replace1068_2_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 3152, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 3133, 3134, 3153 },
      -1,
   } },

   /* ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)) */
   /* search1069_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1069_0_0_0_0_0_0 -> 3018 in the cache */
   /* replace1069_0_0_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_0_0_0_0_1_0_0 -> 3074 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe8000000000000 /* 0.75 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 3074, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3157 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 3018, 3158 },
      -1,
   } },
   /* replace1069_0_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3159, 252 },
      -1,
   } },
   /* replace1069_0_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_0_0_1_0_0_1_0_0 -> 3079 in the cache */
   /* replace1069_0_0_0_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 3079, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3161 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 3018, 3162 },
      -1,
   } },
   /* replace1069_0_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3163, 252 },
      -1,
   } },
   /* replace1069_0_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_0_0_1_1_0_0_1_0_0 -> 3084 in the cache */
   /* replace1069_0_0_0_1_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 3084, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 3018, 3166 },
      -1,
   } },
   /* replace1069_0_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3167, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3168 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 3164, 3169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3160, 3170 },
      -1,
   } },
   /* replace1069_0_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 3171, 3021 },
      -1,
   } },
   /* replace1069_0_1_0_0_0_0_0 -> 3018 in the cache */
   /* replace1069_0_1_0_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_1_0_0_0_0_1_0_0 -> 3093 in the cache */
   /* replace1069_0_1_0_0_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 3093, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3173 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 3018, 3174 },
      -1,
   } },
   /* replace1069_0_1_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3175, 252 },
      -1,
   } },
   /* replace1069_0_1_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_1_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_1_0_0_1_0_0_1_0_0 -> 3098 in the cache */
   /* replace1069_0_1_0_0_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 3098, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3177 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 3018, 3178 },
      -1,
   } },
   /* replace1069_0_1_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3179, 252 },
      -1,
   } },
   /* replace1069_0_1_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_1_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_1_0_0_1_1_0_0_1_0_0 -> 3103 in the cache */
   /* replace1069_0_1_0_0_1_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 3103, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3181 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 3018, 3182 },
      -1,
   } },
   /* replace1069_0_1_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3183, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3184 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 3180, 3185 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3176, 3186 },
      -1,
   } },
   /* replace1069_0_1_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 3187, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 3188 },
      -1,
   } },
   /* replace1069_0_2_0_0_0_0_0 -> 3018 in the cache */
   /* replace1069_0_2_0_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_2_0_0_0_0_1_0_0 -> 3113 in the cache */
   /* replace1069_0_2_0_0_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 3113, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3190 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 3018, 3191 },
      -1,
   } },
   /* replace1069_0_2_0_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3192, 252 },
      -1,
   } },
   /* replace1069_0_2_0_0_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_2_0_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_2_0_0_1_0_0_1_0_0 -> 3118 in the cache */
   /* replace1069_0_2_0_0_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 3118, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3194 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 3018, 3195 },
      -1,
   } },
   /* replace1069_0_2_0_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3196, 252 },
      -1,
   } },
   /* replace1069_0_2_0_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1069_0_2_0_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_0_2_0_0_1_1_0_0_1_0_0 -> 3123 in the cache */
   /* replace1069_0_2_0_0_1_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 3123, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 3018, 3199 },
      -1,
   } },
   /* replace1069_0_2_0_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3200, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3201 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 3197, 3202 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3193, 3203 },
      -1,
   } },
   /* replace1069_0_2_0_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 3204, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 3205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 3172, 3189, 3206 },
      -1,
   } },
   /* replace1069_1 -> 3134 in the cache */
   /* replace1069_2_0_0_0_0 -> 3018 in the cache */
   /* replace1069_2_0_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_2_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_2_0_0_0_1_0_0 -> 3135 in the cache */
   /* replace1069_2_0_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 3135, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3208 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 3018, 3209 },
      -1,
   } },
   /* replace1069_2_0_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3210, 252 },
      -1,
   } },
   /* replace1069_2_0_1_0_0_0 -> 3018 in the cache */
   /* replace1069_2_0_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_2_0_1_0_0_1_0_0 -> 3140 in the cache */
   /* replace1069_2_0_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 3140, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3212 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 3018, 3213 },
      -1,
   } },
   /* replace1069_2_0_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3214, 252 },
      -1,
   } },
   /* replace1069_2_0_1_1_0_0_0 -> 3018 in the cache */
   /* replace1069_2_0_1_1_0_0_1_0_0_0 -> 3073 in the cache */
   /* replace1069_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace1069_2_0_1_1_0_0_1_0_0 -> 3145 in the cache */
   /* replace1069_2_0_1_1_0_0_1_0_1 -> 3156 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 3145, 3156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3216 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 3018, 3217 },
      -1,
   } },
   /* replace1069_2_0_1_1_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3218, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3219 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 3215, 3220 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3211, 3221 },
      -1,
   } },
   /* replace1069_2_1 -> 3021 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 3222, 3021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 3207, 3134, 3223 },
      -1,
   } },

   /* ('ieq', 'a', True) => a */
   /* search1070_0 -> 770 in the cache */
   /* search1070_1 -> 1248 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 770, 1248 },
      -1,
   } },

   /* replace1070 -> 770 in the cache */

   /* ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a') */
   /* search1071_0 -> 770 in the cache */
   /* search1071_1 -> 1248 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 770, 1248 },
      2,
   } },

   /* replace1071_0 -> 770 in the cache */
   /* replace1071 -> 1144 in the cache */

   /* ('ine', 'a', False) => a */
   /* search1072_0 -> 770 in the cache */
   /* search1072_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 770, 1090 },
      -1,
   } },

   /* replace1072 -> 770 in the cache */

   /* ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a') */
   /* search1073_0 -> 770 in the cache */
   /* search1073_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 770, 1090 },
      2,
   } },

   /* replace1073_0 -> 770 in the cache */
   /* replace1073 -> 1144 in the cache */

   /* ('bcsel', 'a', True, False) => a */
   /* search1074_0 -> 770 in the cache */
   /* search1074_1 -> 1248 in the cache */
   /* search1074_2 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1248, 1090 },
      -1,
   } },

   /* replace1074 -> 770 in the cache */

   /* ('bcsel', 'a', False, True) => ('inot', 'a') */
   /* search1075_0 -> 770 in the cache */
   /* search1075_1 -> 1090 in the cache */
   /* search1075_2 -> 1248 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1090, 1248 },
      -1,
   } },

   /* replace1075_0 -> 770 in the cache */
   /* replace1075 -> 1144 in the cache */

   /* ('bcsel', True, 'b', 'c') => b */
   /* search1076_0 -> 1248 in the cache */
   /* search1076_1 -> 1130 in the cache */
   /* search1076_2 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1248, 1130, 694 },
      -1,
   } },

   /* replace1076 -> 1130 in the cache */

   /* ('bcsel', False, 'b', 'c') => c */
   /* search1077_0 -> 1090 in the cache */
   /* search1077_1 -> 1130 in the cache */
   /* search1077_2 -> 694 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 1130, 694 },
      -1,
   } },

   /* replace1077 -> 694 in the cache */

   /* ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search1078_0 -> 770 in the cache */
   /* search1078_1 -> 491 in the cache */
   /* search1078_2 -> 119 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 491, 119 },
      -1,
   } },

   /* replace1078_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search1079_0 -> 770 in the cache */
   /* search1079_1 -> 119 in the cache */
   /* search1079_2 -> 491 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 119, 491 },
      -1,
   } },

   /* replace1079_0_0 -> 770 in the cache */
   /* replace1079_0 -> 1144 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1144 },
      -1,
   } },

   /* ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search1080_0 -> 770 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3237, 3238 },
      -1,
   } },

   /* replace1080_0_0 -> 770 in the cache */
   /* replace1080_0 -> 3234 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3234 },
      -1,
   } },

   /* ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search1081_0 -> 770 in the cache */
   /* search1081_1 -> 3238 in the cache */
   /* search1081_2 -> 3237 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3238, 3237 },
      -1,
   } },

   /* replace1081_0_0_0 -> 770 in the cache */
   /* replace1081_0_0 -> 1144 in the cache */
   /* replace1081_0 -> 3236 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3236 },
      -1,
   } },

   /* ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search1082_0 -> 770 in the cache */
   /* search1082_1 -> 254 in the cache */
   /* search1082_2 -> 121 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 254, 121 },
      -1,
   } },

   /* replace1082_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search1083_0 -> 770 in the cache */
   /* search1083_1 -> 121 in the cache */
   /* search1083_2 -> 254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 121, 254 },
      -1,
   } },

   /* replace1083_0_0 -> 770 in the cache */
   /* replace1083_0 -> 1144 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1144 },
      -1,
   } },

   /* ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search1084_0 -> 770 in the cache */
   /* search1084_1 -> 288 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 288, 3247 },
      -1,
   } },

   /* replace1084_0_0 -> 770 in the cache */
   /* replace1084_0 -> 3244 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3244 },
      -1,
   } },

   /* ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search1085_0 -> 770 in the cache */
   /* search1085_1 -> 3247 in the cache */
   /* search1085_2 -> 288 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3247, 288 },
      -1,
   } },

   /* replace1085_0_0_0 -> 770 in the cache */
   /* replace1085_0_0 -> 1144 in the cache */
   /* replace1085_0 -> 3246 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3246 },
      -1,
   } },

   /* ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search1086_0 -> 770 in the cache */
   /* search1086_1 -> 632 in the cache */
   /* search1086_2 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 632, 711 },
      -1,
   } },

   /* replace1086_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 770 },
      -1,
   } },

   /* ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search1087_0 -> 770 in the cache */
   /* search1087_1 -> 711 in the cache */
   /* search1087_2 -> 632 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 711, 632 },
      -1,
   } },

   /* replace1087_0_0 -> 770 in the cache */
   /* replace1087_0 -> 1144 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1144 },
      -1,
   } },

   /* ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search1088_0 -> 770 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3256, 3257 },
      -1,
   } },

   /* replace1088_0_0 -> 770 in the cache */
   /* replace1088_0 -> 3253 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3253 },
      -1,
   } },

   /* ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search1089_0 -> 770 in the cache */
   /* search1089_1 -> 3257 in the cache */
   /* search1089_2 -> 3256 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3257, 3256 },
      -1,
   } },

   /* replace1089_0_0_0 -> 770 in the cache */
   /* replace1089_0_0 -> 1144 in the cache */
   /* replace1089_0 -> 3255 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3255 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', 'b') => b */
   /* search1090_0 -> 770 in the cache */
   /* search1090_1 -> 3 in the cache */
   /* search1090_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3, 3 },
      -1,
   } },

   /* replace1090 -> 3 in the cache */

   /* ('~fcsel', 'a', 'b', 'b') => b */
   /* search1091_0 -> 16 in the cache */
   /* search1091_1 -> 36 in the cache */
   /* search1091_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 16, 36, 36 },
      -1,
   } },

   /* replace1091 -> 36 in the cache */

   /* ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1')) */
   /* search1092_0 -> 770 in the cache */
   /* search1092_1 -> 2935 in the cache */
   /* search1092_2 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 2935, 1242 },
      -1,
   } },

   /* replace1092_0_0 -> 770 in the cache */
   /* replace1092_0 -> 1241 in the cache */
   /* replace1092 -> 2416 in the cache */

   /* ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a'))) */
   /* search1093_0 -> 770 in the cache */
   /* search1093_1 -> 1242 in the cache */
   /* search1093_2 -> 2935 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1242, 2935 },
      -1,
   } },

   /* replace1093_0_0_0 -> 770 in the cache */
   /* replace1093_0_0 -> 1144 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1144 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3266 },
      -1,
   } },

   /* ('bcsel', 'a', 1, 0) => ('b2i', 'a@1') */
   /* search1094_0 -> 770 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   /* search1094_2 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 3268, 1242 },
      -1,
   } },

   /* replace1094_0 -> 770 in the cache */
   /* replace1094 -> 1241 in the cache */

   /* ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a')) */
   /* search1095_0 -> 770 in the cache */
   /* search1095_1 -> 1242 in the cache */
   /* search1095_2 -> 3268 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1242, 3268 },
      -1,
   } },

   /* replace1095_0_0 -> 770 in the cache */
   /* replace1095_0 -> 1144 in the cache */
   /* replace1095 -> 3266 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1096_0_0_0 -> 770 in the cache */
   /* search1096_0_0 -> 1241 in the cache */
   /* search1096_0 -> 2416 in the cache */
   /* search1096_1_0_0 -> 1147 in the cache */
   /* search1096_1_0 -> 2417 in the cache */
   /* search1096_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1096_0_0_0 -> 770 in the cache */
   /* replace1096_0_0_1 -> 1147 in the cache */
   /* replace1096_0_0 -> 1159 in the cache */
   /* replace1096_0 -> 2844 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2844 },
      -1,
   } },

   /* ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1097_0_0_0 -> 770 in the cache */
   /* search1097_0_0 -> 1241 in the cache */
   /* search1097_0 -> 2416 in the cache */
   /* search1097_1_0_0 -> 1147 in the cache */
   /* search1097_1_0 -> 2417 in the cache */
   /* search1097_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1097_0_0_0 -> 770 in the cache */
   /* replace1097_0_0_1 -> 1147 in the cache */
   /* replace1097_0_0 -> 1151 in the cache */
   /* replace1097_0 -> 2847 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2847 },
      -1,
   } },

   /* ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a */
   /* search1098_0_0_0 -> 770 in the cache */
   /* search1098_0_0 -> 1241 in the cache */
   /* search1098_0 -> 2416 in the cache */
   /* search1098_1 -> 2935 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2416, 2935 },
      -1,
   } },

   /* replace1098 -> 770 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a') */
   /* search1099_0_0_0 -> 770 in the cache */
   /* search1099_0_0 -> 1241 in the cache */
   /* search1099_0 -> 2416 in the cache */
   /* search1099_1 -> 2935 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2416, 2935 },
      -1,
   } },

   /* replace1099_0 -> 770 in the cache */
   /* replace1099 -> 1144 in the cache */

   /* ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a') */
   /* search1100_0_0_0 -> 770 in the cache */
   /* search1100_0_0 -> 1241 in the cache */
   /* search1100_0 -> 2416 in the cache */
   /* search1100_1 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2416, 1242 },
      -1,
   } },

   /* replace1100_0 -> 770 in the cache */
   /* replace1100 -> 1144 in the cache */

   /* ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a */
   /* search1101_0_0_0 -> 770 in the cache */
   /* search1101_0_0 -> 1241 in the cache */
   /* search1101_0 -> 2416 in the cache */
   /* search1101_1 -> 1242 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2416, 1242 },
      -1,
   } },

   /* replace1101 -> 770 in the cache */

   /* ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a */
   /* search1102_0 -> 1242 in the cache */
   /* search1102_1_0_0 -> 770 in the cache */
   /* search1102_1_0 -> 1241 in the cache */
   /* search1102_1 -> 2416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1242, 2416 },
      -1,
   } },

   /* replace1102 -> 770 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a') */
   /* search1103_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3280 },
      -1,
   } },
   /* search1103_1 -> 1152 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3281, 1152 },
      -1,
   } },

   /* replace1103_0 -> 32 in the cache */
   /* replace1103 -> 1541 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a') */
   /* search1104_0_0_0 -> 32 in the cache */
   /* search1104_0_0 -> 3280 in the cache */
   /* search1104_0 -> 3281 in the cache */
   /* search1104_1 -> 3268 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3281, 3268 },
      -1,
   } },

   /* replace1104_0 -> 32 in the cache */
   /* replace1104 -> 3280 in the cache */

   /* ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1105_0_0_0 -> 770 in the cache */
   /* search1105_0_0 -> 1241 in the cache */
   /* search1105_0 -> 2416 in the cache */
   /* search1105_1_0_0 -> 1147 in the cache */
   /* search1105_1_0 -> 2417 in the cache */
   /* search1105_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1105_0_0_0 -> 770 in the cache */
   /* replace1105_0_0_1 -> 1147 in the cache */
   /* replace1105_0_0 -> 1159 in the cache */
   /* replace1105_0 -> 2844 in the cache */
   /* replace1105 -> 3272 in the cache */

   /* ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1106_0_0_0 -> 770 in the cache */
   /* search1106_0_0 -> 1241 in the cache */
   /* search1106_0 -> 2416 in the cache */
   /* search1106_1_0_0 -> 1147 in the cache */
   /* search1106_1_0 -> 2417 in the cache */
   /* search1106_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1106_0_0_0 -> 770 in the cache */
   /* replace1106_0_0_1 -> 1147 in the cache */
   /* replace1106_0_0 -> 1151 in the cache */
   /* replace1106_0 -> 2847 in the cache */
   /* replace1106 -> 3274 in the cache */

   /* ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1107_0_0_0 -> 770 in the cache */
   /* search1107_0_0 -> 1241 in the cache */
   /* search1107_0 -> 2416 in the cache */
   /* search1107_1_0_0 -> 1147 in the cache */
   /* search1107_1_0 -> 2417 in the cache */
   /* search1107_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1107_0_0_0 -> 770 in the cache */
   /* replace1107_0_0_1 -> 1147 in the cache */
   /* replace1107_0_0 -> 1151 in the cache */
   /* replace1107_0 -> 2847 in the cache */
   /* replace1107 -> 3274 in the cache */

   /* ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1108_0_0_0 -> 770 in the cache */
   /* search1108_0_0 -> 1241 in the cache */
   /* search1108_0 -> 2416 in the cache */
   /* search1108_1_0_0 -> 1147 in the cache */
   /* search1108_1_0 -> 2417 in the cache */
   /* search1108_1 -> 2418 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2416, 2418 },
      -1,
   } },

   /* replace1108_0_0_0 -> 770 in the cache */
   /* replace1108_0_0_1 -> 1147 in the cache */
   /* replace1108_0_0 -> 1159 in the cache */
   /* replace1108_0 -> 2844 in the cache */
   /* replace1108 -> 3272 in the cache */

   /* ('umax', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b')) */
   /* search1109_0_0 -> 770 in the cache */
   /* search1109_0 -> 1241 in the cache */
   /* search1109_1_0 -> 1147 in the cache */
   /* search1109_1 -> 2417 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 1241, 2417 },
      -1,
   } },

   /* replace1109_0_0 -> 770 in the cache */
   /* replace1109_0_1 -> 1147 in the cache */
   /* replace1109_0 -> 1151 in the cache */
   /* replace1109 -> 2847 in the cache */

   /* ('umin', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search1110_0_0 -> 770 in the cache */
   /* search1110_0 -> 1241 in the cache */
   /* search1110_1_0 -> 1147 in the cache */
   /* search1110_1 -> 2417 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 1241, 2417 },
      -1,
   } },

   /* replace1110_0_0 -> 770 in the cache */
   /* replace1110_0_1 -> 1147 in the cache */
   /* replace1110_0 -> 1159 in the cache */
   /* replace1110 -> 2844 in the cache */

   /* ('ine', ('umin', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('iand', 'a', ('ine', 'b', 0)) */
   /* search1111_0_0_0_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3290 },
      -1,
   } },
   /* search1111_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 3291, 3 },
      -1,
   } },
   /* search1111_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3292, 86 },
      -1,
   } },

   /* replace1111_0 -> 770 in the cache */
   /* replace1111_1_0 -> 3 in the cache */
   /* replace1111_1_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 3, 86 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 770, 3294 },
      -1,
   } },

   /* ('ine', ('umax', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('ior', 'a', ('ine', 'b', 0)) */
   /* search1112_0_0_0_0 -> 770 in the cache */
   /* search1112_0_0_0 -> 3290 in the cache */
   /* search1112_0_0 -> 3291 in the cache */
   /* search1112_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 3291, 3 },
      -1,
   } },
   /* search1112_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3296, 86 },
      -1,
   } },

   /* replace1112_0 -> 770 in the cache */
   /* replace1112_1_0 -> 3 in the cache */
   /* replace1112_1_1 -> 86 in the cache */
   /* replace1112_1 -> 3294 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 770, 3294 },
      -1,
   } },

   /* ('f2i', ('ftrunc', 'a')) => ('f2i', 'a') */
   /* search1113_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 3299 },
      -1,
   } },

   /* replace1113_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2u', ('ftrunc', 'a')) => ('f2u', 'a') */
   /* search1114_0_0 -> 32 in the cache */
   /* search1114_0 -> 3299 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 3299 },
      -1,
   } },

   /* replace1114_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2fmp', ('f2f32', 'a@16')) => a */
   /* search1115_0_0 -> 11 in the cache */
   /* search1115_0 -> 1845 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace1115 -> 11 in the cache */

   /* ('i2imp', ('i2i32', 'a@16')) => a */
   /* search1116_0_0 -> 11 in the cache */
   /* search1116_0 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace1116 -> 11 in the cache */

   /* ('i2imp', ('u2u32', 'a@16')) => a */
   /* search1117_0_0 -> 11 in the cache */
   /* search1117_0 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace1117 -> 11 in the cache */

   /* ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a') */
   /* search1118_0_0 -> 11 in the cache */
   /* search1118_0 -> 1845 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace1118_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a') */
   /* search1119_0_0 -> 11 in the cache */
   /* search1119_0 -> 1845 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1845 },
      -1,
   } },

   /* replace1119_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a') */
   /* search1120_0_0 -> 11 in the cache */
   /* search1120_0 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace1120_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a') */
   /* search1121_0_0 -> 11 in the cache */
   /* search1121_0 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace1121_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a') */
   /* search1122_0_0 -> 770 in the cache */
   /* search1122_0 -> 1421 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1421 },
      -1,
   } },

   /* replace1122_0 -> 770 in the cache */
   /* replace1122 -> 1417 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search1123_0_0 -> 770 in the cache */
   /* search1123_0 -> 1409 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1409 },
      -1,
   } },

   /* replace1123_0 -> 770 in the cache */
   /* replace1123 -> 1405 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search1124_0_0 -> 770 in the cache */
   /* search1124_0 -> 1409 in the cache */
   /* search1124 -> 3316 in the cache */

   /* replace1124_0 -> 770 in the cache */
   /* replace1124 -> 1405 in the cache */

   /* ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search1125_0_0 -> 770 in the cache */
   /* search1125_0 -> 1421 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1421 },
      -1,
   } },

   /* replace1125_0 -> 770 in the cache */
   /* replace1125 -> 1405 in the cache */

   /* ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search1126_0_0 -> 770 in the cache */
   /* search1126_0 -> 1421 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1421 },
      -1,
   } },

   /* replace1126_0 -> 770 in the cache */
   /* replace1126 -> 1405 in the cache */

   /* ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search1127_0_0 -> 770 in the cache */
   /* search1127_0 -> 1409 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 1409 },
      -1,
   } },

   /* replace1127_0 -> 770 in the cache */
   /* replace1127 -> 1417 in the cache */

   /* ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search1128_0_0 -> 770 in the cache */
   /* search1128_0 -> 1409 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 1409 },
      -1,
   } },

   /* replace1128_0 -> 770 in the cache */
   /* replace1128 -> 1417 in the cache */

   /* ('f2f32', ('f2fmp', 'a@32')) => a */
   /* search1129_0_0 -> 16 in the cache */
   /* search1129_0 -> 2161 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 2161 },
      -1,
   } },

   /* replace1129 -> 16 in the cache */

   /* ('i2i32', ('i2imp', 'a@32')) => a */
   /* search1130_0_0 -> 16 in the cache */
   /* search1130_0 -> 2172 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2172 },
      -1,
   } },

   /* replace1130 -> 16 in the cache */

   /* ('u2u32', ('i2imp', 'a@32')) => a */
   /* search1131_0_0 -> 16 in the cache */
   /* search1131_0 -> 2172 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2172 },
      -1,
   } },

   /* replace1131 -> 16 in the cache */

   /* ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a') */
   /* search1132_0_0 -> 16 in the cache */
   /* search1132_0 -> 2171 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2171 },
      -1,
   } },

   /* replace1132_0 -> 16 in the cache */
   /* replace1132 -> 2169 in the cache */

   /* ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a') */
   /* search1133_0_0 -> 16 in the cache */
   /* search1133_0 -> 2168 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2168 },
      -1,
   } },

   /* replace1133_0 -> 16 in the cache */
   /* replace1133 -> 2166 in the cache */

   /* ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a') */
   /* search1134_0_0 -> 16 in the cache */
   /* search1134_0 -> 2160 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 2160 },
      -1,
   } },

   /* replace1134_0 -> 16 in the cache */
   /* replace1134 -> 2158 in the cache */

   /* ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a') */
   /* search1135_0_0 -> 16 in the cache */
   /* search1135_0 -> 2157 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 2157 },
      -1,
   } },

   /* replace1135_0 -> 16 in the cache */
   /* replace1135 -> 2155 in the cache */

   /* ('f2i32', ('f2fmp', 'a@32')) => ('f2i32', 'a') */
   /* search1136_0_0 -> 16 in the cache */
   /* search1136_0 -> 2161 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 2161 },
      -1,
   } },

   /* replace1136_0 -> 16 in the cache */
   /* replace1136 -> 2169 in the cache */

   /* ('f2u32', ('f2fmp', 'a@32')) => ('f2u32', 'a') */
   /* search1137_0_0 -> 16 in the cache */
   /* search1137_0 -> 2161 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 2161 },
      -1,
   } },

   /* replace1137_0 -> 16 in the cache */
   /* replace1137 -> 2166 in the cache */

   /* ('i2f32', ('i2imp', 'a@32')) => ('i2f32', 'a') */
   /* search1138_0_0 -> 16 in the cache */
   /* search1138_0 -> 2172 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 2172 },
      -1,
   } },

   /* replace1138_0 -> 16 in the cache */
   /* replace1138 -> 2158 in the cache */

   /* ('ffloor', 'a(is_integral)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 3331 },
      -1,
   } },

   /* replace1139 -> 32 in the cache */

   /* ('fceil', 'a(is_integral)') => a */
   /* search1140_0 -> 3331 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 3331 },
      -1,
   } },

   /* replace1140 -> 32 in the cache */

   /* ('ftrunc', 'a(is_integral)') => a */
   /* search1141_0 -> 3331 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 3331 },
      -1,
   } },

   /* replace1141 -> 32 in the cache */

   /* ('fround_even', 'a(is_integral)') => a */
   /* search1142_0 -> 3331 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 3331 },
      -1,
   } },

   /* replace1142 -> 32 in the cache */

   /* ('~ffract', 'a(is_integral)') => 0.0 */
   /* search1143_0 -> 3331 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 3331 },
      -1,
   } },

   /* replace1143 -> 117 in the cache */

   /* ('fabs', 'a(is_not_negative)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3337 },
      -1,
   } },

   /* replace1144 -> 32 in the cache */

   /* ('iabs', 'a(is_not_negative)') => a */
   /* search1145_0 -> 3337 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 3337 },
      -1,
   } },

   /* replace1145 -> 32 in the cache */

   /* ('fsat', 'a(is_not_positive)') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      26,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 3340 },
      -1,
   } },

   /* replace1146 -> 117 in the cache */

   /* ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a') */
   /* search1147_0 -> 3337 in the cache */
   /* search1147_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 3337, 252 },
      -1,
   } },

   /* replace1147_0 -> 32 in the cache */
   /* replace1147 -> 228 in the cache */

   /* ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1148_0_0_0_0 -> 32 in the cache */
   /* search1148_0_0_0 -> 228 in the cache */
   /* search1148_0_0_1_0_0 -> 32 in the cache */
   /* search1148_0_0_1_0 -> 228 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 228 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 228, 3343 },
      -1,
   } },
   /* search1148_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3344, 252 },
      -1,
   } },
   /* search1148_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 3345, 117 },
      -1,
   } },

   /* replace1148 -> 1090 in the cache */

   /* ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1149_0_0_0_0_0 -> 32 in the cache */
   /* search1149_0_0_0_0 -> 228 in the cache */
   /* search1149_0_0_0_1_0 -> 32 in the cache */
   /* search1149_0_0_0_1 -> 228 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 228, 228 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3347 },
      -1,
   } },
   /* search1149_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 3348, 252 },
      -1,
   } },
   /* search1149_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3349, 117 },
      -1,
   } },

   /* replace1149 -> 1090 in the cache */

   /* ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0) */
   /* search1150_0_0_0_0 -> 32 in the cache */
   /* search1150_0_0_0 -> 228 in the cache */
   /* search1150_0_0_1_0_0 -> 32 in the cache */
   /* search1150_0_0_1_0 -> 228 in the cache */
   /* search1150_0_0_1 -> 3343 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 228, 3343 },
      -1,
   } },
   /* search1150_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 3351, 252 },
      -1,
   } },
   /* search1150_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 3352, 117 },
      -1,
   } },

   /* replace1150_0_0_0 -> 32 in the cache */
   /* replace1150_0_0 -> 228 in the cache */
   /* replace1150_0_1_0_0 -> 32 in the cache */
   /* replace1150_0_1_0 -> 228 in the cache */
   /* replace1150_0_1 -> 3343 in the cache */
   /* replace1150_0 -> 3344 in the cache */
   /* replace1150_1 -> 252 in the cache */
   /* replace1150 -> 3345 in the cache */

   /* ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0) */
   /* search1151_0_0_0_0_0 -> 32 in the cache */
   /* search1151_0_0_0_0 -> 228 in the cache */
   /* search1151_0_0_0_1_0 -> 32 in the cache */
   /* search1151_0_0_0_1 -> 228 in the cache */
   /* search1151_0_0_0 -> 3347 in the cache */
   /* search1151_0_0 -> 3348 in the cache */
   /* search1151_0_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3348, 252 },
      -1,
   } },
   /* search1151_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 3354, 117 },
      -1,
   } },

   /* replace1151_0_0_0_0 -> 32 in the cache */
   /* replace1151_0_0_0 -> 228 in the cache */
   /* replace1151_0_0_1_0 -> 32 in the cache */
   /* replace1151_0_0_1 -> 228 in the cache */
   /* replace1151_0_0 -> 3347 in the cache */
   /* replace1151_0 -> 3348 in the cache */
   /* replace1151_1 -> 252 in the cache */
   /* replace1151 -> 3349 in the cache */

   /* ('fneu', 'a(is_not_zero)', 0.0) => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      27,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1152_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3356, 117 },
      -1,
   } },

   /* replace1152 -> 1248 in the cache */

   /* ('feq', 'a(is_not_zero)', 0.0) => False */
   /* search1153_0 -> 3356 in the cache */
   /* search1153_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3356, 117 },
      -1,
   } },

   /* replace1153 -> 1090 in the cache */

   /* ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      28,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3359, 3360 },
      -1,
   } },

   /* replace1154 -> 1248 in the cache */

   /* ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      26,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      30,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3362, 3363 },
      -1,
   } },

   /* replace1155 -> 1090 in the cache */

   /* ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      31,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3365, 3366 },
      -1,
   } },

   /* replace1156 -> 1090 in the cache */

   /* ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1157_0 -> 1599 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      26,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1599, 3368 },
      -1,
   } },

   /* replace1157 -> 1090 in the cache */

   /* ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      32,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3370, 3371 },
      -1,
   } },

   /* replace1158 -> 1248 in the cache */

   /* ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      33,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      28,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3373, 3374 },
      -1,
   } },

   /* replace1159 -> 1248 in the cache */

   /* ('ine', 'a(is_not_zero)', 0) => True */
   /* search1160_0 -> 3356 in the cache */
   /* search1160_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3356, 61 },
      -1,
   } },

   /* replace1160 -> 1248 in the cache */

   /* ('ieq', 'a(is_not_zero)', 0) => False */
   /* search1161_0 -> 3356 in the cache */
   /* search1161_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3356, 61 },
      -1,
   } },

   /* replace1161 -> 1090 in the cache */

   /* ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True */
   /* search1162_0 -> 1599 in the cache */
   /* search1162_1 -> 3368 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1599, 3368 },
      -1,
   } },

   /* replace1162 -> 1248 in the cache */

   /* ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   /* search1163_0 -> 3362 in the cache */
   /* search1163_1 -> 3363 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3362, 3363 },
      -1,
   } },

   /* replace1163 -> 1090 in the cache */

   /* ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   /* search1164_0 -> 3365 in the cache */
   /* search1164_1 -> 3366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3365, 3366 },
      -1,
   } },

   /* replace1164 -> 1090 in the cache */

   /* ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1165_0 -> 1599 in the cache */
   /* search1165_1 -> 3368 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1599, 3368 },
      -1,
   } },

   /* replace1165 -> 1090 in the cache */

   /* ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True */
   /* search1166_0 -> 3362 in the cache */
   /* search1166_1 -> 3363 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3362, 3363 },
      -1,
   } },

   /* replace1166 -> 1248 in the cache */

   /* ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True */
   /* search1167_0 -> 3365 in the cache */
   /* search1167_1 -> 3366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3365, 3366 },
      -1,
   } },

   /* replace1167 -> 1248 in the cache */

   /* ('ult', 0, 'a(is_gt_zero)') => True */
   /* search1168_0 -> 61 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      30,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 3384 },
      -1,
   } },

   /* replace1168 -> 1248 in the cache */

   /* ('ult', 'a', 0) => False */
   /* search1169_0 -> 32 in the cache */
   /* search1169_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace1169 -> 1090 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a */
   /* search1170_0_0 -> 16 in the cache */
   /* search1170_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1170 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b */
   /* search1171_0_0 -> 16 in the cache */
   /* search1171_0_1 -> 36 in the cache */
   /* search1171_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1171 -> 36 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x */
   /* search1172_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 3390 },
      -1,
   } },

   /* replace1172 -> 2595 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y */
   /* search1173_0_0 -> 16 in the cache */
   /* search1173_0 -> 3390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3390 },
      -1,
   } },

   /* replace1173 -> 2598 in the cache */

   /* ('unpack_64_2x32_split_x', ('u2u64', 'a@32')) => a */
   /* search1174_0_0 -> 16 in the cache */
   /* search1174_0 -> 2149 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 2149 },
      -1,
   } },

   /* replace1174 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('u2u64', 'a')) => 0 */
   /* search1175_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1175 -> 128 in the cache */

   /* ('unpack_64_2x32_split_x', ('i2i64', 'a@32')) => a */
   /* search1176_0_0 -> 16 in the cache */
   /* search1176_0 -> 2147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 2147 },
      -1,
   } },

   /* replace1176 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('i2i64(is_used_once)', 'a@32')) => ('ishr', 'a', 31) */
   /* search1177_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 16 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3397 },
      -1,
   } },

   /* replace1177_0 -> 16 in the cache */
   /* replace1177_1 -> 19 in the cache */
   /* replace1177 -> 1473 in the cache */

   /* ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b') */
   /* search1178_0_0 -> 16 in the cache */
   /* search1178_0_1 -> 36 in the cache */
   /* search1178_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1178_0 -> 16 in the cache */
   /* replace1178_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a */
   /* search1179_0_0 -> 16 in the cache */
   /* search1179_0 -> 3390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 3390 },
      -1,
   } },

   /* replace1179 -> 16 in the cache */

   /* ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a */
   /* search1180_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 3402 },
      -1,
   } },

   /* replace1180 -> 16 in the cache */

   /* ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a */
   /* search1181_0_0 -> 21 in the cache */
   /* search1181_0 -> 2900 in the cache */
   /* search1181_1_0 -> 21 in the cache */
   /* search1181_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 2900, 2894 },
      -1,
   } },

   /* replace1181 -> 21 in the cache */

   /* ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a */
   /* search1182_0_0_0 -> 21 in the cache */
   /* search1182_0_0 -> 2900 in the cache */
   /* search1182_0_1_0 -> 21 in the cache */
   /* search1182_0_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2900, 2894 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 3405 },
      -1,
   } },

   /* replace1182 -> 21 in the cache */

   /* ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a */
   /* search1183_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 3407 },
      -1,
   } },

   /* replace1183 -> 21 in the cache */

   /* ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a */
   /* search1184_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 3409 },
      -1,
   } },

   /* replace1184 -> 21 in the cache */

   /* ('unpack_64_4x16', ('pack_64_4x16', 'a')) => a */
   /* search1185_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_4x16,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_64_4x16,
      -1, 0,
      { 3411 },
      -1,
   } },

   /* replace1185 -> 11 in the cache */

   /* ('unpack_64_4x16', ('pack_64_2x32', ('vec2', ('pack_32_2x16_split', 'a', 'b'), ('pack_32_2x16_split', 'c', 'd')))) => ('vec4', 'a', 'b', 'c', 'd') */
   /* search1186_0_0_0_0 -> 11 in the cache */
   /* search1186_0_0_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 11, 268 },
      -1,
   } },
   /* search1186_0_0_1_0 -> 479 in the cache */
   /* search1186_0_0_1_1 -> 485 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 479, 485 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3413, 3414 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 3415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_64_4x16,
      -1, 0,
      { 3416 },
      -1,
   } },

   /* replace1186_0 -> 11 in the cache */
   /* replace1186_1 -> 268 in the cache */
   /* replace1186_2 -> 479 in the cache */
   /* replace1186_3 -> 485 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 11, 268, 479, 485 },
      -1,
   } },

   /* ('unpack_64_4x16', ('pack_64_2x32_split', ('pack_32_2x16_split', 'a', 'b'), ('pack_32_2x16_split', 'c', 'd'))) => ('vec4', 'a', 'b', 'c', 'd') */
   /* search1187_0_0_0 -> 11 in the cache */
   /* search1187_0_0_1 -> 268 in the cache */
   /* search1187_0_0 -> 3413 in the cache */
   /* search1187_0_1_0 -> 479 in the cache */
   /* search1187_0_1_1 -> 485 in the cache */
   /* search1187_0_1 -> 3414 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 3413, 3414 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_64_4x16,
      -1, 0,
      { 3419 },
      -1,
   } },

   /* replace1187_0 -> 11 in the cache */
   /* replace1187_1 -> 268 in the cache */
   /* replace1187_2 -> 479 in the cache */
   /* replace1187_3 -> 485 in the cache */
   /* replace1187 -> 3418 in the cache */

   /* ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c')) */
   /* search1188_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1188_0_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 3421, 2179 },
      -1,
   } },
   /* search1188_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3423, 3424 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3422, 3425 },
      -1,
   } },

   /* replace1188_0 -> 16 in the cache */
   /* replace1188_1_0 -> 268 in the cache */
   /* replace1188_1_1 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 268, 479 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3427 },
      -1,
   } },

   /* ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1) */
   /* search1189_0 -> 11 in the cache */
   /* search1189_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 844 },
      -1,
   } },

   /* replace1189_0 -> 11 in the cache */
   /* replace1189_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 1465 },
      -1,
   } },

   /* ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3) */
   /* search1190_0 -> 16 in the cache */
   /* search1190_1 -> 1025 in the cache */
   /* search1190 -> 1030 in the cache */

   /* replace1190_0 -> 16 in the cache */
   /* replace1190_1 -> 415 in the cache */
   /* replace1190 -> 432 in the cache */

   /* ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7) */
   /* search1191_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x38 /* 56 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3431 },
      -1,
   } },

   /* replace1191_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3433 },
      -1,
   } },

   /* ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1) */
   /* search1192_0 -> 11 in the cache */
   /* search1192_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 844 },
      -1,
   } },

   /* replace1192_0 -> 11 in the cache */
   /* replace1192_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 1465 },
      -1,
   } },

   /* ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3) */
   /* search1193_0 -> 16 in the cache */
   /* search1193_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 1025 },
      -1,
   } },

   /* replace1193_0 -> 16 in the cache */
   /* replace1193_1 -> 415 in the cache */
   /* replace1193 -> 416 in the cache */

   /* ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7) */
   /* search1194_0 -> 21 in the cache */
   /* search1194_1 -> 3431 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3431 },
      -1,
   } },

   /* replace1194_0 -> 21 in the cache */
   /* replace1194_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3433 },
      -1,
   } },

   /* ('iand', 255, 'a') => ('extract_u8', 'a', 0) */
   /* search1195_0 -> 1605 in the cache */
   /* search1195_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1605, 32 },
      -1,
   } },

   /* replace1195_0 -> 32 in the cache */
   /* replace1195_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2)) */
   /* search1196_0_0 -> 32 in the cache */
   /* search1196_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3442 },
      -1,
   } },

   /* replace1196_0_0 -> 32 in the cache */
   /* replace1196_0_1 -> 2406 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 2406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3444 },
      -1,
   } },

   /* ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1)) */
   /* search1197_0_0 -> 32 in the cache */
   /* search1197_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3446 },
      -1,
   } },

   /* replace1197_0_0 -> 32 in the cache */
   /* replace1197_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3448 },
      -1,
   } },

   /* ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b')) */
   /* search1198_0_0_0 -> 0 in the cache */
   /* search1198_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3450 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3451 },
      -1,
   } },

   /* replace1198_0_0 -> 0 in the cache */
   /* replace1198_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3453 },
      -1,
   } },

   /* ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b')) */
   /* search1199_0_0_0 -> 0 in the cache */
   /* search1199_0_0_1 -> 3 in the cache */
   /* search1199_0_0 -> 3450 in the cache */
   /* search1199_0 -> 3451 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3451 },
      -1,
   } },

   /* replace1199_0_0 -> 0 in the cache */
   /* replace1199_0_1 -> 3 in the cache */
   /* replace1199_0 -> 3450 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3450 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0) */
   /* search1200_0 -> 16 in the cache */
   /* search1200_1 -> 128 in the cache */
   /* search1200_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 128, 844 },
      -1,
   } },

   /* replace1200_0 -> 16 in the cache */
   /* replace1200_1 -> 128 in the cache */
   /* replace1200 -> 422 in the cache */

   /* ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1) */
   /* search1201_0 -> 16 in the cache */
   /* search1201_1 -> 844 in the cache */
   /* search1201_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 844, 844 },
      -1,
   } },

   /* replace1201_0 -> 16 in the cache */
   /* replace1201_1 -> 406 in the cache */
   /* replace1201 -> 425 in the cache */

   /* ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2) */
   /* search1202_0 -> 16 in the cache */
   /* search1202_1 -> 866 in the cache */
   /* search1202_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 866, 844 },
      -1,
   } },

   /* replace1202_0 -> 16 in the cache */
   /* replace1202_1 -> 411 in the cache */
   /* replace1202 -> 429 in the cache */

   /* ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3) */
   /* search1203_0 -> 16 in the cache */
   /* search1203_1 -> 1025 in the cache */
   /* search1203_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 1025, 844 },
      -1,
   } },

   /* replace1203_0 -> 16 in the cache */
   /* replace1203_1 -> 415 in the cache */
   /* replace1203 -> 432 in the cache */

   /* ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0) */
   /* search1204_0 -> 16 in the cache */
   /* search1204_1 -> 128 in the cache */
   /* search1204_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 128, 844 },
      -1,
   } },

   /* replace1204_0 -> 16 in the cache */
   /* replace1204_1 -> 128 in the cache */
   /* replace1204 -> 403 in the cache */

   /* ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1) */
   /* search1205_0 -> 16 in the cache */
   /* search1205_1 -> 844 in the cache */
   /* search1205_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 844, 844 },
      -1,
   } },

   /* replace1205_0 -> 16 in the cache */
   /* replace1205_1 -> 406 in the cache */
   /* replace1205 -> 407 in the cache */

   /* ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2) */
   /* search1206_0 -> 16 in the cache */
   /* search1206_1 -> 866 in the cache */
   /* search1206_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 866, 844 },
      -1,
   } },

   /* replace1206_0 -> 16 in the cache */
   /* replace1206_1 -> 411 in the cache */
   /* replace1206 -> 412 in the cache */

   /* ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3) */
   /* search1207_0 -> 16 in the cache */
   /* search1207_1 -> 1025 in the cache */
   /* search1207_2 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 1025, 844 },
      -1,
   } },

   /* replace1207_0 -> 16 in the cache */
   /* replace1207_1 -> 415 in the cache */
   /* replace1207 -> 416 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1208_0_0 -> 0 in the cache */
   /* search1208_0_1 -> 3 in the cache */
   /* search1208_0 -> 3453 in the cache */
   /* search1208_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3453, 86 },
      -1,
   } },

   /* replace1208_0 -> 0 in the cache */
   /* replace1208_1 -> 3 in the cache */
   /* replace1208 -> 3450 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1209_0_0 -> 0 in the cache */
   /* search1209_0_1 -> 3 in the cache */
   /* search1209_0 -> 3450 in the cache */
   /* search1209_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3450, 86 },
      -1,
   } },

   /* replace1209_0 -> 0 in the cache */
   /* replace1209_1 -> 3 in the cache */
   /* replace1209 -> 3450 in the cache */

   /* ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0) */
   /* search1210_0_0 -> 16 in the cache */
   /* search1210_0_1 -> 866 in the cache */
   /* search1210_0 -> 1018 in the cache */
   /* search1210_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 1018, 866 },
      -1,
   } },

   /* replace1210_0 -> 16 in the cache */
   /* replace1210_1 -> 128 in the cache */
   /* replace1210 -> 454 in the cache */

   /* ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1) */
   /* search1211_0 -> 16 in the cache */
   /* search1211_1 -> 866 in the cache */
   /* search1211 -> 1021 in the cache */

   /* replace1211_0 -> 16 in the cache */
   /* replace1211_1 -> 406 in the cache */
   /* replace1211 -> 457 in the cache */

   /* ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0) */
   /* search1212_0_0 -> 16 in the cache */
   /* search1212_0_1 -> 866 in the cache */
   /* search1212_0 -> 1018 in the cache */
   /* search1212_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 1018, 866 },
      -1,
   } },

   /* replace1212_0 -> 16 in the cache */
   /* replace1212_1 -> 128 in the cache */
   /* replace1212 -> 446 in the cache */

   /* ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1) */
   /* search1213_0 -> 16 in the cache */
   /* search1213_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 866 },
      -1,
   } },

   /* replace1213_0 -> 16 in the cache */
   /* replace1213_1 -> 406 in the cache */
   /* replace1213 -> 449 in the cache */

   /* ('iand', 65535, 'a') => ('extract_u16', 'a', 0) */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* search1214_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3470, 32 },
      -1,
   } },

   /* replace1214_0 -> 32 in the cache */
   /* replace1214_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0) */
   /* search1215_0 -> 16 in the cache */
   /* search1215_1 -> 128 in the cache */
   /* search1215_2 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 128, 866 },
      -1,
   } },

   /* replace1215_0 -> 16 in the cache */
   /* replace1215_1 -> 128 in the cache */
   /* replace1215 -> 454 in the cache */

   /* ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1) */
   /* search1216_0 -> 16 in the cache */
   /* search1216_1 -> 866 in the cache */
   /* search1216_2 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 866, 866 },
      -1,
   } },

   /* replace1216_0 -> 16 in the cache */
   /* replace1216_1 -> 406 in the cache */
   /* replace1216 -> 457 in the cache */

   /* ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0) */
   /* search1217_0 -> 16 in the cache */
   /* search1217_1 -> 128 in the cache */
   /* search1217_2 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 128, 866 },
      -1,
   } },

   /* replace1217_0 -> 16 in the cache */
   /* replace1217_1 -> 128 in the cache */
   /* replace1217 -> 446 in the cache */

   /* ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1) */
   /* search1218_0 -> 16 in the cache */
   /* search1218_1 -> 866 in the cache */
   /* search1218_2 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 866, 866 },
      -1,
   } },

   /* replace1218_0 -> 16 in the cache */
   /* replace1218_1 -> 406 in the cache */
   /* replace1218 -> 449 in the cache */

   /* ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a')) */
   /* search1219_0_0_0 -> 6 in the cache */
   /* search1219_0_0 -> 2099 in the cache */
   /* search1219_0_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 2099, 1025 },
      -1,
   } },
   /* search1219_1_0_0_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2280 },
      -1,
   } },
   /* search1219_1_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3478, 866 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3480 },
      -1,
   } },
   /* search1219_1_1_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3481, 844 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3483 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 3482, 3484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 2,
      { 3479, 3485 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3477, 3486 },
      -1,
   } },

   /* replace1219_0_0 -> 3483 in the cache */
   /* replace1219_0_1 -> 3480 in the cache */
   /* replace1219_0_2 -> 2280 in the cache */
   /* replace1219_0_3 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3483, 3480, 2280, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8,
      -1, 0,
      { 3488 },
      -1,
   } },

   /* ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1220_0_0 -> 0 in the cache */
   /* search1220_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   /* search1220_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3490, 86 },
      -1,
   } },

   /* replace1220_0 -> 0 in the cache */
   /* replace1220_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1221_0_0 -> 0 in the cache */
   /* search1221_0_1 -> 3 in the cache */
   /* search1221_0 -> 3492 in the cache */
   /* search1221_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3492, 86 },
      -1,
   } },

   /* replace1221_0 -> 0 in the cache */
   /* replace1221_1 -> 3 in the cache */
   /* replace1221 -> 3492 in the cache */

   /* ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32)) */
   /* search1222_0 -> 16 in the cache */
   /* search1222_1 -> 36 in the cache */
   /* search1222 -> 3387 in the cache */

   /* replace1222_0_0 -> 16 in the cache */
   /* replace1222_0 -> 2149 in the cache */
   /* replace1222_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 36 },
      -1,
   } },
   /* replace1222_1_1 -> 887 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3494, 887 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2149, 3495 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16)) */
   /* search1223_0 -> 11 in the cache */
   /* search1223_1 -> 268 in the cache */
   /* search1223 -> 3413 in the cache */

   /* replace1223_0_0 -> 11 in the cache */
   /* replace1223_0 -> 2129 in the cache */
   /* replace1223_1_0_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 268 },
      -1,
   } },
   /* replace1223_1_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3497, 866 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2129, 3498 },
      -1,
   } },

   /* ('pack_half_2x16_split', 'a', 'b') => ('pack_half_2x16_rtz_split', 'a', 'b') */
   /* search1224_0 -> 16 in the cache */
   /* search1224_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1224_0 -> 16 in the cache */
   /* replace1224_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a') */
   /* search1225_0 -> 21 in the cache */
   /* search1225 -> 2900 in the cache */

   /* replace1225_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32)) */
   /* search1226_0 -> 21 in the cache */
   /* search1226 -> 2894 in the cache */

   /* replace1226_0_0 -> 21 in the cache */
   /* replace1226_0_1 -> 887 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 887 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3503 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a') */
   /* search1227_0 -> 16 in the cache */
   /* search1227 -> 3421 in the cache */

   /* replace1227_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1228_0 -> 16 in the cache */
   /* search1228 -> 3423 in the cache */

   /* replace1228_0_0 -> 16 in the cache */
   /* replace1228_0_1 -> 866 in the cache */
   /* replace1228_0 -> 1021 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1021 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1229_0_0 -> 16 in the cache */
   /* search1229_0_1 -> 1015 in the cache */
   /* search1229_0 -> 1016 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 1,
      { 1016 },
      -1,
   } },

   /* replace1229_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1230_0_0 -> 16 in the cache */
   /* search1230_0_1 -> 1015 in the cache */
   /* search1230_0 -> 1016 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 1,
      { 1016 },
      -1,
   } },

   /* replace1230_0 -> 16 in the cache */
   /* replace1230 -> 3421 in the cache */

   /* ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a') */
   /* search1231_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3510 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 3511 },
      -1,
   } },

   /* replace1231_0 -> 21 in the cache */
   /* replace1231 -> 2900 in the cache */

   /* ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1232_0_0 -> 16 in the cache */
   /* search1232_0_1 -> 1019 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1019 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 1,
      { 3513 },
      -1,
   } },

   /* replace1232_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1233_0_0 -> 16 in the cache */
   /* search1233_0_1 -> 1019 in the cache */
   /* search1233_0 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 1,
      { 3513 },
      -1,
   } },

   /* replace1233_0 -> 16 in the cache */
   /* replace1233 -> 3423 in the cache */

   /* ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a') */
   /* search1234_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff00000000 /* 18446744069414584320 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3517 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 1,
      { 3518 },
      -1,
   } },

   /* replace1234_0 -> 21 in the cache */
   /* replace1234 -> 2894 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1235_0_0 -> 16 in the cache */
   /* search1235_0_1 -> 128 in the cache */
   /* search1235_0 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 454 },
      -1,
   } },

   /* replace1235_0 -> 16 in the cache */
   /* replace1235 -> 3508 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1236_0_0 -> 16 in the cache */
   /* search1236_0_1 -> 406 in the cache */
   /* search1236_0 -> 457 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 457 },
      -1,
   } },

   /* replace1236_0 -> 16 in the cache */
   /* replace1236 -> 3515 in the cache */

   /* ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1237_0_0 -> 16 in the cache */
   /* search1237_0_1 -> 866 in the cache */
   /* search1237_0 -> 1021 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 1021 },
      -1,
   } },

   /* replace1237_0 -> 16 in the cache */
   /* replace1237 -> 3515 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1238_0_0 -> 16 in the cache */
   /* search1238_0_1 -> 128 in the cache */
   /* search1238_0 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 454 },
      -1,
   } },

   /* replace1238_0 -> 16 in the cache */
   /* replace1238 -> 3421 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1239_0_0 -> 16 in the cache */
   /* search1239_0_1 -> 406 in the cache */
   /* search1239_0 -> 457 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 457 },
      -1,
   } },

   /* replace1239_0 -> 16 in the cache */
   /* replace1239 -> 3423 in the cache */

   /* ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a')) */
   /* search1240_0_0_0 -> 16 in the cache */
   /* search1240_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3525 },
      -1,
   } },
   /* search1240_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3526, 866 },
      -1,
   } },

   /* replace1240_0_0 -> 128 in the cache */
   /* replace1240_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 128, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3528 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0)) */
   /* search1241_0_0_0 -> 128 in the cache */
   /* search1241_0_0_1 -> 16 in the cache */
   /* search1241_0_0 -> 3528 in the cache */
   /* search1241_0 -> 3529 in the cache */
   /* search1241_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3529, 866 },
      -1,
   } },

   /* replace1241_0_0 -> 16 in the cache */
   /* replace1241_0_1 -> 128 in the cache */
   /* replace1241_0 -> 3525 in the cache */
   /* replace1241 -> 3526 in the cache */

   /* ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1242_0_0_0 -> 16 in the cache */
   /* search1242_0_0_1 -> 128 in the cache */
   /* search1242_0_0 -> 3525 in the cache */
   /* search1242_0 -> 3526 in the cache */
   /* search1242_1_0_0 -> 128 in the cache */
   /* search1242_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 128, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3526, 3532 },
      -1,
   } },

   /* replace1242_0_0 -> 16 in the cache */
   /* replace1242_0_1 -> 36 in the cache */
   /* replace1242_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1243_0_0_0 -> 16 in the cache */
   /* search1243_0_0_1 -> 128 in the cache */
   /* search1243_0_0 -> 3525 in the cache */
   /* search1243_0 -> 3526 in the cache */
   /* search1243_1_0_0 -> 128 in the cache */
   /* search1243_1_0_1 -> 36 in the cache */
   /* search1243_1_0 -> 3531 in the cache */
   /* search1243_1 -> 3532 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3526, 3532 },
      -1,
   } },

   /* replace1243_0_0 -> 16 in the cache */
   /* replace1243_0_1 -> 36 in the cache */
   /* replace1243_0 -> 3400 in the cache */
   /* replace1243 -> 3534 in the cache */

   /* ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a') */
   /* search1244_0_0 -> 16 in the cache */
   /* search1244_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* search1244_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3536, 866 },
      -1,
   } },

   /* replace1244_0 -> 128 in the cache */
   /* replace1244_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 128, 16 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0) */
   /* search1245_0_0 -> 128 in the cache */
   /* search1245_0_1 -> 16 in the cache */
   /* search1245_0 -> 3538 in the cache */
   /* search1245_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3538, 866 },
      -1,
   } },

   /* replace1245_0 -> 16 in the cache */
   /* replace1245_1 -> 128 in the cache */
   /* replace1245 -> 3536 in the cache */

   /* ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0) */
   /* search1246_0_0 -> 128 in the cache */
   /* search1246_0_1 -> 16 in the cache */
   /* search1246_0 -> 3538 in the cache */
   /* search1246_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3538, 406 },
      -1,
   } },

   /* replace1246_0 -> 16 in the cache */
   /* replace1246_1 -> 128 in the cache */
   /* replace1246 -> 3536 in the cache */

   /* ('ishl', ('pack_half_2x16_rtz_split', 'a', 0), 16) => ('pack_half_2x16_rtz_split', 0, 'a') */
   /* search1247_0_0 -> 16 in the cache */
   /* search1247_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* search1247_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3541, 866 },
      -1,
   } },

   /* replace1247_0 -> 128 in the cache */
   /* replace1247_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 128, 16 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16_rtz_split', 0, 'a'), 16) => ('pack_half_2x16_rtz_split', 'a', 0) */
   /* search1248_0_0 -> 128 in the cache */
   /* search1248_0_1 -> 16 in the cache */
   /* search1248_0 -> 3543 in the cache */
   /* search1248_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3543, 866 },
      -1,
   } },

   /* replace1248_0 -> 16 in the cache */
   /* replace1248_1 -> 128 in the cache */
   /* replace1248 -> 3541 in the cache */

   /* ('extract_u16', ('pack_half_2x16_rtz_split', 0, 'a'), 1) => ('pack_half_2x16_rtz_split', 'a', 0) */
   /* search1249_0_0 -> 128 in the cache */
   /* search1249_0_1 -> 16 in the cache */
   /* search1249_0 -> 3543 in the cache */
   /* search1249_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3543, 406 },
      -1,
   } },

   /* replace1249_0 -> 16 in the cache */
   /* replace1249_1 -> 128 in the cache */
   /* replace1249 -> 3541 in the cache */

   /* ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1250_0_0 -> 16 in the cache */
   /* search1250_0_1 -> 128 in the cache */
   /* search1250_0 -> 3536 in the cache */
   /* search1250_1_0 -> 128 in the cache */
   /* search1250_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 128, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3536, 3546 },
      -1,
   } },

   /* replace1250_0 -> 16 in the cache */
   /* replace1250_1 -> 36 in the cache */
   /* replace1250 -> 3500 in the cache */

   /* ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1251_0_0 -> 16 in the cache */
   /* search1251_0_1 -> 128 in the cache */
   /* search1251_0 -> 3536 in the cache */
   /* search1251_1_0 -> 128 in the cache */
   /* search1251_1_1 -> 36 in the cache */
   /* search1251_1 -> 3546 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3536, 3546 },
      -1,
   } },

   /* replace1251_0 -> 16 in the cache */
   /* replace1251_1 -> 36 in the cache */
   /* replace1251 -> 3500 in the cache */

   /* ('iadd', ('pack_half_2x16_rtz_split', 'a', 0), ('pack_half_2x16_rtz_split', 0, 'b')) => ('pack_half_2x16_rtz_split', 'a', 'b') */
   /* search1252_0_0 -> 16 in the cache */
   /* search1252_0_1 -> 128 in the cache */
   /* search1252_0 -> 3541 in the cache */
   /* search1252_1_0 -> 128 in the cache */
   /* search1252_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 128, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3541, 3549 },
      -1,
   } },

   /* replace1252_0 -> 16 in the cache */
   /* replace1252_1 -> 36 in the cache */
   /* replace1252 -> 3501 in the cache */

   /* ('ior', ('pack_half_2x16_rtz_split', 'a', 0), ('pack_half_2x16_rtz_split', 0, 'b')) => ('pack_half_2x16_rtz_split', 'a', 'b') */
   /* search1253_0_0 -> 16 in the cache */
   /* search1253_0_1 -> 128 in the cache */
   /* search1253_0 -> 3541 in the cache */
   /* search1253_1_0 -> 128 in the cache */
   /* search1253_1_1 -> 36 in the cache */
   /* search1253_1 -> 3549 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3541, 3549 },
      -1,
   } },

   /* replace1253_0 -> 16 in the cache */
   /* replace1253_1 -> 36 in the cache */
   /* replace1253 -> 3501 in the cache */

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a') */
   /* search1254_0_0 -> 6 in the cache */
   /* search1254_0_1 -> 2280 in the cache */
   /* search1254_0_2 -> 3480 in the cache */
   /* search1254_0_3 -> 3483 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8_split,
      -1, 0,
      { 6, 2280, 3480, 3483 },
      -1,
   } },
   /* search1254_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3552, 128 },
      -1,
   } },

   /* replace1254_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b') */
   /* search1255_0_0 -> 6 in the cache */
   /* search1255_0_1 -> 2280 in the cache */
   /* search1255_0_2 -> 3480 in the cache */
   /* search1255_0_3 -> 3483 in the cache */
   /* search1255_0 -> 3552 in the cache */
   /* search1255_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3552, 406 },
      -1,
   } },

   /* replace1255_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 2280 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c') */
   /* search1256_0_0 -> 6 in the cache */
   /* search1256_0_1 -> 2280 in the cache */
   /* search1256_0_2 -> 3480 in the cache */
   /* search1256_0_3 -> 3483 in the cache */
   /* search1256_0 -> 3552 in the cache */
   /* search1256_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3552, 411 },
      -1,
   } },

   /* replace1256_0 -> 3480 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3480 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd') */
   /* search1257_0_0 -> 6 in the cache */
   /* search1257_0_1 -> 2280 in the cache */
   /* search1257_0_2 -> 3480 in the cache */
   /* search1257_0_3 -> 3483 in the cache */
   /* search1257_0 -> 3552 in the cache */
   /* search1257_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3552, 415 },
      -1,
   } },

   /* replace1257_0 -> 3483 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3483 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a') */
   /* search1258_0_0 -> 6 in the cache */
   /* search1258_0_1 -> 2280 in the cache */
   /* search1258_0_2 -> 3480 in the cache */
   /* search1258_0_3 -> 3483 in the cache */
   /* search1258_0 -> 3552 in the cache */
   /* search1258_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3552, 128 },
      -1,
   } },

   /* replace1258_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b') */
   /* search1259_0_0 -> 6 in the cache */
   /* search1259_0_1 -> 2280 in the cache */
   /* search1259_0_2 -> 3480 in the cache */
   /* search1259_0_3 -> 3483 in the cache */
   /* search1259_0 -> 3552 in the cache */
   /* search1259_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3552, 406 },
      -1,
   } },

   /* replace1259_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 2280 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c') */
   /* search1260_0_0 -> 6 in the cache */
   /* search1260_0_1 -> 2280 in the cache */
   /* search1260_0_2 -> 3480 in the cache */
   /* search1260_0_3 -> 3483 in the cache */
   /* search1260_0 -> 3552 in the cache */
   /* search1260_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3552, 411 },
      -1,
   } },

   /* replace1260_0 -> 3480 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3480 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd') */
   /* search1261_0_0 -> 6 in the cache */
   /* search1261_0_1 -> 2280 in the cache */
   /* search1261_0_2 -> 3480 in the cache */
   /* search1261_0_3 -> 3483 in the cache */
   /* search1261_0 -> 3552 in the cache */
   /* search1261_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3552, 415 },
      -1,
   } },

   /* replace1261_0 -> 3483 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3483 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1262_0_0 -> 11 in the cache */
   /* search1262_0_1 -> 844 in the cache */
   /* search1262_0 -> 3429 in the cache */
   /* search1262_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3429, 869 },
      -1,
   } },

   /* replace1262_0 -> 11 in the cache */
   /* replace1262_1 -> 1465 in the cache */
   /* replace1262 -> 3430 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1263_0_0 -> 16 in the cache */
   /* search1263_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 844 },
      -1,
   } },
   /* search1263_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3570, 128 },
      -1,
   } },

   /* replace1263_0 -> 16 in the cache */
   /* replace1263_1 -> 406 in the cache */
   /* replace1263 -> 425 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1264_0_0 -> 16 in the cache */
   /* search1264_0_1 -> 866 in the cache */
   /* search1264_0 -> 1021 in the cache */
   /* search1264_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1021, 128 },
      -1,
   } },

   /* replace1264_0 -> 16 in the cache */
   /* replace1264_1 -> 411 in the cache */
   /* replace1264 -> 429 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1265_0_0 -> 16 in the cache */
   /* search1265_0_1 -> 1025 in the cache */
   /* search1265_0 -> 1030 in the cache */
   /* search1265_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1030, 128 },
      -1,
   } },

   /* replace1265_0 -> 16 in the cache */
   /* replace1265_1 -> 415 in the cache */
   /* replace1265 -> 432 in the cache */

   /* ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1266_0_0 -> 21 in the cache */
   /* search1266_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 844 },
      -1,
   } },
   /* search1266_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3574, 910 },
      -1,
   } },

   /* replace1266_0 -> 21 in the cache */
   /* replace1266_1 -> 1507 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 1507 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1267_0_0 -> 21 in the cache */
   /* search1267_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 866 },
      -1,
   } },
   /* search1267_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3577, 910 },
      -1,
   } },

   /* replace1267_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3579 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1268_0_0 -> 21 in the cache */
   /* search1268_0_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 1025 },
      -1,
   } },
   /* search1268_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3581, 910 },
      -1,
   } },

   /* replace1268_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3583 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1269_0_0 -> 21 in the cache */
   /* search1269_0_1 -> 887 in the cache */
   /* search1269_0 -> 3503 in the cache */
   /* search1269_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3503, 910 },
      -1,
   } },

   /* replace1269_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3586 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1270_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x28 /* 40 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3588 },
      -1,
   } },
   /* search1270_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3589, 910 },
      -1,
   } },

   /* replace1270_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3591 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1271_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x30 /* 48 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3593 },
      -1,
   } },
   /* search1271_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3594, 910 },
      -1,
   } },

   /* replace1271_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x6 /* 6 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3596 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1272_0_0 -> 21 in the cache */
   /* search1272_0_1 -> 3431 in the cache */
   /* search1272_0 -> 3432 in the cache */
   /* search1272_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3432, 910 },
      -1,
   } },

   /* replace1272_0 -> 21 in the cache */
   /* replace1272_1 -> 3433 in the cache */
   /* replace1272 -> 3434 in the cache */

   /* ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1273_0_0 -> 11 in the cache */
   /* search1273_0_1 -> 844 in the cache */
   /* search1273_0 -> 3435 in the cache */
   /* search1273_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3435, 869 },
      -1,
   } },

   /* replace1273_0 -> 11 in the cache */
   /* replace1273_1 -> 1465 in the cache */
   /* replace1273 -> 3430 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1274_0_0 -> 16 in the cache */
   /* search1274_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 844 },
      -1,
   } },
   /* search1274_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3600, 128 },
      -1,
   } },

   /* replace1274_0 -> 16 in the cache */
   /* replace1274_1 -> 406 in the cache */
   /* replace1274 -> 425 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1275_0_0 -> 16 in the cache */
   /* search1275_0_1 -> 866 in the cache */
   /* search1275_0 -> 3469 in the cache */
   /* search1275_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3469, 128 },
      -1,
   } },

   /* replace1275_0 -> 16 in the cache */
   /* replace1275_1 -> 411 in the cache */
   /* replace1275 -> 429 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1276_0_0 -> 16 in the cache */
   /* search1276_0_1 -> 1025 in the cache */
   /* search1276_0 -> 3437 in the cache */
   /* search1276_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3437, 128 },
      -1,
   } },

   /* replace1276_0 -> 16 in the cache */
   /* replace1276_1 -> 415 in the cache */
   /* replace1276 -> 432 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1277_0_0 -> 21 in the cache */
   /* search1277_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 844 },
      -1,
   } },
   /* search1277_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3604, 910 },
      -1,
   } },

   /* replace1277_0 -> 21 in the cache */
   /* replace1277_1 -> 1507 in the cache */
   /* replace1277 -> 3576 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1278_0_0 -> 21 in the cache */
   /* search1278_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 866 },
      -1,
   } },
   /* search1278_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3606, 910 },
      -1,
   } },

   /* replace1278_0 -> 21 in the cache */
   /* replace1278_1 -> 3579 in the cache */
   /* replace1278 -> 3580 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1279_0_0 -> 21 in the cache */
   /* search1279_0_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 1025 },
      -1,
   } },
   /* search1279_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3608, 910 },
      -1,
   } },

   /* replace1279_0 -> 21 in the cache */
   /* replace1279_1 -> 3583 in the cache */
   /* replace1279 -> 3584 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1280_0_0 -> 21 in the cache */
   /* search1280_0_1 -> 887 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 887 },
      -1,
   } },
   /* search1280_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3610, 910 },
      -1,
   } },

   /* replace1280_0 -> 21 in the cache */
   /* replace1280_1 -> 3586 in the cache */
   /* replace1280 -> 3587 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1281_0_0 -> 21 in the cache */
   /* search1281_0_1 -> 3588 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3588 },
      -1,
   } },
   /* search1281_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3612, 910 },
      -1,
   } },

   /* replace1281_0 -> 21 in the cache */
   /* replace1281_1 -> 3591 in the cache */
   /* replace1281 -> 3592 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1282_0_0 -> 21 in the cache */
   /* search1282_0_1 -> 3593 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3593 },
      -1,
   } },
   /* search1282_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3614, 910 },
      -1,
   } },

   /* replace1282_0 -> 21 in the cache */
   /* replace1282_1 -> 3596 in the cache */
   /* replace1282 -> 3597 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1283_0_0 -> 21 in the cache */
   /* search1283_0_1 -> 3431 in the cache */
   /* search1283_0 -> 3438 in the cache */
   /* search1283_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3438, 910 },
      -1,
   } },

   /* replace1283_0 -> 21 in the cache */
   /* replace1283_1 -> 3433 in the cache */
   /* replace1283 -> 3434 in the cache */

   /* ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2) */
   /* search1284_0_0 -> 32 in the cache */
   /* search1284_0_1 -> 35 in the cache */
   /* search1284_0 -> 3442 in the cache */
   /* search1284_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3442, 61 },
      -1,
   } },

   /* replace1284_0 -> 32 in the cache */
   /* replace1284_1 -> 2406 in the cache */
   /* replace1284 -> 3444 in the cache */

   /* ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0) */
   /* search1285_0_0 -> 11 in the cache */
   /* search1285_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 844 },
      -1,
   } },
   /* search1285_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3618, 1465 },
      -1,
   } },

   /* replace1285_0 -> 11 in the cache */
   /* replace1285_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 869 },
      -1,
   } },

   /* ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2) */
   /* search1286_0_0 -> 16 in the cache */
   /* search1286_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 844 },
      -1,
   } },
   /* search1286_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3621, 415 },
      -1,
   } },

   /* replace1286_0 -> 16 in the cache */
   /* replace1286_1 -> 411 in the cache */
   /* replace1286 -> 429 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1) */
   /* search1287_0_0 -> 16 in the cache */
   /* search1287_0_1 -> 866 in the cache */
   /* search1287_0 -> 1018 in the cache */
   /* search1287_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1018, 415 },
      -1,
   } },

   /* replace1287_0 -> 16 in the cache */
   /* replace1287_1 -> 406 in the cache */
   /* replace1287 -> 425 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0) */
   /* search1288_0_0 -> 16 in the cache */
   /* search1288_0_1 -> 1025 in the cache */
   /* search1288_0 -> 1027 in the cache */
   /* search1288_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1027, 415 },
      -1,
   } },

   /* replace1288_0 -> 16 in the cache */
   /* replace1288_1 -> 128 in the cache */
   /* replace1288 -> 422 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6) */
   /* search1289_0_0 -> 21 in the cache */
   /* search1289_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 844 },
      -1,
   } },
   /* search1289_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3625, 3433 },
      -1,
   } },

   /* replace1289_0 -> 21 in the cache */
   /* replace1289_1 -> 3596 in the cache */
   /* replace1289 -> 3597 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5) */
   /* search1290_0_0 -> 21 in the cache */
   /* search1290_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 866 },
      -1,
   } },
   /* search1290_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3627, 3433 },
      -1,
   } },

   /* replace1290_0 -> 21 in the cache */
   /* replace1290_1 -> 3591 in the cache */
   /* replace1290 -> 3592 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4) */
   /* search1291_0_0 -> 21 in the cache */
   /* search1291_0_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 1025 },
      -1,
   } },
   /* search1291_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3629, 3433 },
      -1,
   } },

   /* replace1291_0 -> 21 in the cache */
   /* replace1291_1 -> 3586 in the cache */
   /* replace1291 -> 3587 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3) */
   /* search1292_0_0 -> 21 in the cache */
   /* search1292_0_1 -> 887 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 887 },
      -1,
   } },
   /* search1292_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3631, 3433 },
      -1,
   } },

   /* replace1292_0 -> 21 in the cache */
   /* replace1292_1 -> 3583 in the cache */
   /* replace1292 -> 3584 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2) */
   /* search1293_0_0 -> 21 in the cache */
   /* search1293_0_1 -> 3588 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3588 },
      -1,
   } },
   /* search1293_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3633, 3433 },
      -1,
   } },

   /* replace1293_0 -> 21 in the cache */
   /* replace1293_1 -> 3579 in the cache */
   /* replace1293 -> 3580 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1) */
   /* search1294_0_0 -> 21 in the cache */
   /* search1294_0_1 -> 3593 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3593 },
      -1,
   } },
   /* search1294_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3635, 3433 },
      -1,
   } },

   /* replace1294_0 -> 21 in the cache */
   /* replace1294_1 -> 1507 in the cache */
   /* replace1294 -> 3576 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0) */
   /* search1295_0_0 -> 21 in the cache */
   /* search1295_0_1 -> 3431 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3431 },
      -1,
   } },
   /* search1295_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3637, 3433 },
      -1,
   } },

   /* replace1295_0 -> 21 in the cache */
   /* replace1295_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 910 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0) */
   /* search1296_0_0 -> 11 in the cache */
   /* search1296_0_1 -> 844 in the cache */
   /* search1296_0 -> 3618 in the cache */
   /* search1296_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3618, 1465 },
      -1,
   } },

   /* replace1296_0 -> 11 in the cache */
   /* replace1296_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 869 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2) */
   /* search1297_0_0 -> 16 in the cache */
   /* search1297_0_1 -> 844 in the cache */
   /* search1297_0 -> 3621 in the cache */
   /* search1297_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3621, 415 },
      -1,
   } },

   /* replace1297_0 -> 16 in the cache */
   /* replace1297_1 -> 411 in the cache */
   /* replace1297 -> 412 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1) */
   /* search1298_0_0 -> 16 in the cache */
   /* search1298_0_1 -> 866 in the cache */
   /* search1298_0 -> 1018 in the cache */
   /* search1298_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 1018, 415 },
      -1,
   } },

   /* replace1298_0 -> 16 in the cache */
   /* replace1298_1 -> 406 in the cache */
   /* replace1298 -> 407 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0) */
   /* search1299_0_0 -> 16 in the cache */
   /* search1299_0_1 -> 1025 in the cache */
   /* search1299_0 -> 1027 in the cache */
   /* search1299_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 1027, 415 },
      -1,
   } },

   /* replace1299_0 -> 16 in the cache */
   /* replace1299_1 -> 128 in the cache */
   /* replace1299 -> 403 in the cache */

   /* ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6) */
   /* search1300_0_0 -> 21 in the cache */
   /* search1300_0_1 -> 844 in the cache */
   /* search1300_0 -> 3625 in the cache */
   /* search1300_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3625, 3433 },
      -1,
   } },

   /* replace1300_0 -> 21 in the cache */
   /* replace1300_1 -> 3596 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3596 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5) */
   /* search1301_0_0 -> 21 in the cache */
   /* search1301_0_1 -> 866 in the cache */
   /* search1301_0 -> 3627 in the cache */
   /* search1301_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3627, 3433 },
      -1,
   } },

   /* replace1301_0 -> 21 in the cache */
   /* replace1301_1 -> 3591 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3591 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4) */
   /* search1302_0_0 -> 21 in the cache */
   /* search1302_0_1 -> 1025 in the cache */
   /* search1302_0 -> 3629 in the cache */
   /* search1302_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3629, 3433 },
      -1,
   } },

   /* replace1302_0 -> 21 in the cache */
   /* replace1302_1 -> 3586 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3586 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3) */
   /* search1303_0_0 -> 21 in the cache */
   /* search1303_0_1 -> 887 in the cache */
   /* search1303_0 -> 3631 in the cache */
   /* search1303_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3631, 3433 },
      -1,
   } },

   /* replace1303_0 -> 21 in the cache */
   /* replace1303_1 -> 3583 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3583 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2) */
   /* search1304_0_0 -> 21 in the cache */
   /* search1304_0_1 -> 3588 in the cache */
   /* search1304_0 -> 3633 in the cache */
   /* search1304_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3633, 3433 },
      -1,
   } },

   /* replace1304_0 -> 21 in the cache */
   /* replace1304_1 -> 3579 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3579 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1) */
   /* search1305_0_0 -> 21 in the cache */
   /* search1305_0_1 -> 3593 in the cache */
   /* search1305_0 -> 3635 in the cache */
   /* search1305_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3635, 3433 },
      -1,
   } },

   /* replace1305_0 -> 21 in the cache */
   /* replace1305_1 -> 1507 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 1507 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0) */
   /* search1306_0_0 -> 21 in the cache */
   /* search1306_0_1 -> 3431 in the cache */
   /* search1306_0 -> 3637 in the cache */
   /* search1306_1 -> 3433 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3637, 3433 },
      -1,
   } },

   /* replace1306_0 -> 21 in the cache */
   /* replace1306_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 910 },
      -1,
   } },

   /* ('ussub_4x8_vc4', 'a', 0) => a */
   /* search1307_0 -> 16 in the cache */
   /* search1307_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 128 },
      -1,
   } },

   /* replace1307 -> 16 in the cache */

   /* ('ussub_4x8_vc4', 'a', -1) => 0 */
   /* search1308_0 -> 16 in the cache */
   /* search1308_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 130 },
      -1,
   } },

   /* replace1308 -> 128 in the cache */

   /* ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b')) */
   /* search1309_0 -> 0 in the cache */
   /* search1309_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1309_0 -> 0 in the cache */
   /* replace1309_1_0 -> 3 in the cache */
   /* replace1309_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1100 },
      -1,
   } },

   /* ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b')) */
   /* search1310_0 -> 0 in the cache */
   /* search1310_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1310_0 -> 0 in the cache */
   /* replace1310_1_0 -> 3 in the cache */
   /* replace1310_1 -> 93 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 93 },
      -1,
   } },

   /* ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1311_0 -> 0 in the cache */
   /* search1311_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_usub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1311_0_0 -> 0 in the cache */
   /* replace1311_0_1 -> 3 in the cache */
   /* replace1311_0 -> 1080 in the cache */
   /* replace1311_1_0_0 -> 0 in the cache */
   /* replace1311_1_0_1 -> 3 in the cache */
   /* replace1311_1_0 -> 3663 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3663 },
      -1,
   } },
   /* replace1311_2_0 -> 0 in the cache */
   /* replace1311_2_1 -> 3 in the cache */
   /* replace1311_2 -> 3663 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1080, 3666, 3663 },
      -1,
   } },

   /* ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1312_0 -> 0 in the cache */
   /* search1312_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1312_0_0 -> 0 in the cache */
   /* replace1312_0_1 -> 3 in the cache */
   /* replace1312_0 -> 1077 in the cache */
   /* replace1312_1_0_0 -> 0 in the cache */
   /* replace1312_1_0_1 -> 3 in the cache */
   /* replace1312_1_0 -> 3663 in the cache */
   /* replace1312_1 -> 3666 in the cache */
   /* replace1312_2_0 -> 0 in the cache */
   /* replace1312_2_1 -> 3 in the cache */
   /* replace1312_2 -> 3663 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1077, 3666, 3663 },
      -1,
   } },

   /* ('bitz', 'a', 'b') => ('inot', ('bitnz', 'a', 'b')) */
   /* search1313_0 -> 32 in the cache */
   /* search1313_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 32, 36 },
      -1,
   } },

   /* replace1313_0_0 -> 32 in the cache */
   /* replace1313_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 32, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3671 },
      -1,
   } },

   /* ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b')) */
   /* search1314_0_0 -> 0 in the cache */
   /* search1314_0 -> 223 in the cache */
   /* search1314_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 223, 3 },
      3,
   } },

   /* replace1314_0_0 -> 0 in the cache */
   /* replace1314_0_1 -> 3 in the cache */
   /* replace1314_0 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 274 },
      -1,
   } },

   /* ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b')) */
   /* search1315_0_0 -> 16 in the cache */
   /* search1315_0 -> 290 in the cache */
   /* search1315_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 290, 36 },
      3,
   } },

   /* replace1315_0_0 -> 16 in the cache */
   /* replace1315_0_1 -> 36 in the cache */
   /* replace1315_0 -> 279 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 279 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   /* search1316_0_0 -> 132 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 132 },
      -1,
   } },
   /* search1316_1_0 -> 133 in the cache */
   /* search1316_1 -> 1210 in the cache */
   /* search1316_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 3677, 1210, 135 },
      -1,
   } },

   /* replace1316_0 -> 132 in the cache */
   /* replace1316_1 -> 133 in the cache */
   /* replace1316_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 132, 133, 135 },
      -1,
   } },

   /* ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search1317_0_0 -> 16 in the cache */
   /* search1317_0 -> 290 in the cache */
   /* search1317_1_0 -> 36 in the cache */
   /* search1317_1 -> 610 in the cache */
   /* search1317_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 290, 610, 141 },
      -1,
   } },

   /* replace1317_0 -> 16 in the cache */
   /* replace1317_1 -> 36 in the cache */
   /* replace1317_2 -> 141 in the cache */
   /* replace1317 -> 397 in the cache */

   /* ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b')) */
   /* search1318_0_0 -> 0 in the cache */
   /* search1318_0 -> 218 in the cache */
   /* search1318_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 218, 3 },
      -1,
   } },

   /* replace1318_0_0 -> 0 in the cache */
   /* replace1318_0_1 -> 3 in the cache */
   /* replace1318_0 -> 964 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 964 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3683, 3684 },
      0,
   } },
   /* search1319_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3685, 827 },
      0,
   } },

   /* replace1319_0_0 -> 132 in the cache */
   /* replace1319_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace1319_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3687, 133 },
      -1,
   } },

   /* ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1320_0_0 -> 338 in the cache */
   /* search1320_0_1 -> 348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 338, 348 },
      0,
   } },
   /* search1320_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3689, 319 },
      0,
   } },

   /* replace1320_0_0 -> 16 in the cache */
   /* replace1320_0_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 141 },
      -1,
   } },
   /* replace1320_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3691, 36 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   /* search1321_0_0 -> 338 in the cache */
   /* search1321_0_1 -> 348 in the cache */
   /* search1321_0 -> 3689 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3689, 3693 },
      0,
   } },

   /* replace1321_0_0 -> 16 in the cache */
   /* replace1321_0_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 141 },
      -1,
   } },
   /* replace1321_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3695, 36 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1322_0_0 -> 3683 in the cache */
   /* search1322_0_1 -> 3684 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3683, 3684 },
      0,
   } },
   /* search1322_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 3697, 827 },
      0,
   } },

   /* replace1322_0_0 -> 132 in the cache */
   /* replace1322_0_1 -> 135 in the cache */
   /* replace1322_0 -> 1038 in the cache */
   /* replace1322_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 1038, 133 },
      -1,
   } },

   /* ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3700, 3701 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1323_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3702, 3703, 159 },
      -1,
   } },

   /* replace1323_0_0 -> 155 in the cache */
   /* replace1323_0_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 155, 158 },
      -1,
   } },
   /* replace1323_1 -> 156 in the cache */
   /* replace1323_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3705, 156, 159 },
      -1,
   } },

   /* ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1324_0_0 -> 338 in the cache */
   /* search1324_0_1 -> 348 in the cache */
   /* search1324_0 -> 3689 in the cache */
   /* search1324_1 -> 319 in the cache */
   /* search1324_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3689, 319, 169 },
      -1,
   } },

   /* replace1324_0_0 -> 16 in the cache */
   /* replace1324_0_1 -> 141 in the cache */
   /* replace1324_0 -> 3691 in the cache */
   /* replace1324_1 -> 36 in the cache */
   /* replace1324_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3691, 36, 169 },
      -1,
   } },

   /* ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1325_0_0 -> 338 in the cache */
   /* search1325_0_1 -> 348 in the cache */
   /* search1325_0 -> 3689 in the cache */
   /* search1325_1 -> 3693 in the cache */
   /* search1325_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3689, 3693, 169 },
      -1,
   } },

   /* replace1325_0_0 -> 16 in the cache */
   /* replace1325_0_1 -> 141 in the cache */
   /* replace1325_0 -> 3695 in the cache */
   /* replace1325_1 -> 36 in the cache */
   /* replace1325_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3695, 36, 169 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1326_0_0 -> 3683 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      34,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3683, 3711 },
      0,
   } },
   /* search1326_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3712, 827 },
      0,
   } },

   /* replace1326_0_0 -> 132 in the cache */
   /* replace1326_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace1326_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3714, 133 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1327_0_0 -> 3683 in the cache */
   /* search1327_0_1 -> 3684 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3683, 3684 },
      0,
   } },
   /* search1327_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3716, 827 },
      0,
   } },

   /* replace1327_0_0 -> 132 in the cache */
   /* replace1327_0_1 -> 135 in the cache */
   /* replace1327_0 -> 3714 in the cache */
   /* replace1327_1 -> 133 in the cache */
   /* replace1327 -> 3715 in the cache */

   /* ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c') */
   /* search1328_0_0 -> 3700 in the cache */
   /* search1328_0_1 -> 156 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3700, 156, 3718 },
      0,
   } },
   /* search1328_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3719, 1044 },
      0,
   } },

   /* replace1328_0_0 -> 155 in the cache */
   /* replace1328_0_1 -> 156 in the cache */
   /* replace1328_0_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 155, 156, 159 },
      -1,
   } },
   /* replace1328_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3721, 158 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c') */
   /* search1329_0_0 -> 338 in the cache */
   /* search1329_0_1 -> 36 in the cache */
   /* search1329_0_2 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 338, 36, 357 },
      0,
   } },
   /* search1329_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3723, 321 },
      0,
   } },

   /* replace1329_0_0 -> 16 in the cache */
   /* replace1329_0_1 -> 36 in the cache */
   /* replace1329_0_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 36, 169 },
      -1,
   } },
   /* replace1329_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3725, 141 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1330_0_0 -> 3683 in the cache */
   /* search1330_0_1 -> 3684 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3683, 3684 },
      0,
   } },
   /* search1330_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3727, 827 },
      0,
   } },

   /* replace1330_0_0 -> 132 in the cache */
   /* replace1330_0_1 -> 135 in the cache */
   /* replace1330_0 -> 1052 in the cache */
   /* replace1330_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1052, 133 },
      -1,
   } },

   /* ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b') */
   /* search1331_0 -> 292 in the cache */
   /* search1331_1_0 -> 3684 in the cache */
   /* search1331_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 292, 3730 },
      -1,
   } },

   /* replace1331_0_0 -> 132 in the cache */
   /* replace1331_0_1 -> 135 in the cache */
   /* replace1331_0 -> 3687 in the cache */
   /* replace1331_1 -> 133 in the cache */
   /* replace1331 -> 3688 in the cache */

   /* ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1332_0 -> 296 in the cache */
   /* search1332_1_0 -> 348 in the cache */
   /* search1332_1_1 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 348, 319 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 296, 3732 },
      -1,
   } },

   /* replace1332_0_0 -> 16 in the cache */
   /* replace1332_0_1 -> 141 in the cache */
   /* replace1332_0 -> 3691 in the cache */
   /* replace1332_1 -> 36 in the cache */
   /* replace1332 -> 3692 in the cache */

   /* ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1333_1_0 -> 348 in the cache */
   /* search1333_1_1 -> 319 in the cache */
   /* search1333_1 -> 3732 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3734, 3732 },
      -1,
   } },

   /* replace1333_0_0 -> 16 in the cache */
   /* replace1333_0_1 -> 141 in the cache */
   /* replace1333_0 -> 3695 in the cache */
   /* replace1333_1 -> 36 in the cache */
   /* replace1333 -> 3696 in the cache */

   /* ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1334_1_0 -> 3701 in the cache */
   /* search1334_1_1 -> 3703 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3701, 3703 },
      -1,
   } },
   /* search1334_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3736, 3737, 159 },
      -1,
   } },

   /* replace1334_0_0 -> 155 in the cache */
   /* replace1334_0_1 -> 158 in the cache */
   /* replace1334_0 -> 3705 in the cache */
   /* replace1334_1 -> 156 in the cache */
   /* replace1334_2 -> 159 in the cache */
   /* replace1334 -> 3706 in the cache */

   /* ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1335_0 -> 296 in the cache */
   /* search1335_1_0 -> 348 in the cache */
   /* search1335_1_1 -> 319 in the cache */
   /* search1335_1 -> 3732 in the cache */
   /* search1335_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 296, 3732, 169 },
      -1,
   } },

   /* replace1335_0_0 -> 16 in the cache */
   /* replace1335_0_1 -> 141 in the cache */
   /* replace1335_0 -> 3691 in the cache */
   /* replace1335_1 -> 36 in the cache */
   /* replace1335_2 -> 169 in the cache */
   /* replace1335 -> 3708 in the cache */

   /* ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1336_0 -> 3734 in the cache */
   /* search1336_1_0 -> 348 in the cache */
   /* search1336_1_1 -> 319 in the cache */
   /* search1336_1 -> 3732 in the cache */
   /* search1336_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3734, 3732, 169 },
      -1,
   } },

   /* replace1336_0_0 -> 16 in the cache */
   /* replace1336_0_1 -> 141 in the cache */
   /* replace1336_0 -> 3695 in the cache */
   /* replace1336_1 -> 36 in the cache */
   /* replace1336_2 -> 169 in the cache */
   /* replace1336 -> 3710 in the cache */

   /* ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1337_0 -> 292 in the cache */
   /* search1337_1_0 -> 3684 in the cache */
   /* search1337_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 292, 3741 },
      -1,
   } },

   /* replace1337_0_0 -> 132 in the cache */
   /* replace1337_0_1 -> 135 in the cache */
   /* replace1337_0 -> 1038 in the cache */
   /* replace1337_1 -> 133 in the cache */
   /* replace1337 -> 3699 in the cache */

   /* ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1338_0 -> 292 in the cache */
   /* search1338_1_0 -> 3684 in the cache */
   /* search1338_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 292, 3743 },
      -1,
   } },

   /* replace1338_0_0 -> 132 in the cache */
   /* replace1338_0_1 -> 135 in the cache */
   /* replace1338_0 -> 3714 in the cache */
   /* replace1338_1 -> 133 in the cache */
   /* replace1338 -> 3715 in the cache */

   /* ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b')) */
   /* search1339_0 -> 292 in the cache */
   /* search1339_1_0_0 -> 3684 in the cache */
   /* search1339_1_0_1 -> 827 in the cache */
   /* search1339_1_0 -> 3743 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3743 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 292, 3745 },
      -1,
   } },

   /* replace1339_0_0 -> 132 in the cache */
   /* replace1339_0_1_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 132, 3747 },
      -1,
   } },
   /* replace1339_1_0 -> 133 in the cache */
   /* replace1339_1 -> 1210 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3748, 1210 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1340_0 -> 3736 in the cache */
   /* search1340_1_0 -> 3701 in the cache */
   /* search1340_1_1 -> 3718 in the cache */
   /* search1340_1_2 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3701, 3718, 1044 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3736, 3750 },
      -1,
   } },

   /* replace1340_0 -> 156 in the cache */
   /* replace1340_1 -> 158 in the cache */
   /* replace1340_2_0 -> 155 in the cache */
   /* replace1340_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 155, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 156, 158, 3752 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1341_0 -> 3736 in the cache */
   /* search1341_1_0_0 -> 3701 in the cache */
   /* search1341_1_0_1 -> 3718 in the cache */
   /* search1341_1_0_2 -> 1044 in the cache */
   /* search1341_1_0 -> 3750 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3750 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3736, 3754 },
      -1,
   } },

   /* replace1341_0_0 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 156 },
      -1,
   } },
   /* replace1341_1 -> 158 in the cache */
   /* replace1341_2_0 -> 155 in the cache */
   /* replace1341_2_1_0 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 155, 3757 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3756, 158, 3758 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1342_0 -> 296 in the cache */
   /* search1342_1_0 -> 348 in the cache */
   /* search1342_1_1 -> 357 in the cache */
   /* search1342_1_2 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 348, 357, 321 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 296, 3760 },
      -1,
   } },

   /* replace1342_0 -> 36 in the cache */
   /* replace1342_1 -> 141 in the cache */
   /* replace1342_2_0 -> 16 in the cache */
   /* replace1342_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 36, 141, 3762 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1343_0 -> 296 in the cache */
   /* search1343_1_0_0 -> 348 in the cache */
   /* search1343_1_0_1 -> 357 in the cache */
   /* search1343_1_0_2 -> 321 in the cache */
   /* search1343_1_0 -> 3760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3760 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 296, 3764 },
      -1,
   } },

   /* replace1343_0_0 -> 36 in the cache */
   /* replace1343_0 -> 610 in the cache */
   /* replace1343_1 -> 141 in the cache */
   /* replace1343_2_0 -> 16 in the cache */
   /* replace1343_2_1_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 3766 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 610, 141, 3767 },
      -1,
   } },

   /* ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1344_0 -> 292 in the cache */
   /* search1344_1_0 -> 3684 in the cache */
   /* search1344_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 292, 3769 },
      -1,
   } },

   /* replace1344_0_0 -> 132 in the cache */
   /* replace1344_0_1 -> 135 in the cache */
   /* replace1344_0 -> 1052 in the cache */
   /* replace1344_1 -> 133 in the cache */
   /* replace1344 -> 3729 in the cache */

   /* ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b') */
   /* search1345_0 -> 292 in the cache */
   /* search1345_1_0 -> 3684 in the cache */
   /* search1345_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 292, 3771 },
      -1,
   } },

   /* replace1345_0_0 -> 132 in the cache */
   /* replace1345_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace1345_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3773, 133 },
      -1,
   } },

   /* ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b') */
   /* search1346_0 -> 292 in the cache */
   /* search1346_1_0 -> 3684 in the cache */
   /* search1346_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 292, 3775 },
      -1,
   } },

   /* replace1346_0_0 -> 132 in the cache */
   /* replace1346_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace1346_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 3777, 133 },
      -1,
   } },

   /* ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b') */
   /* search1347_0 -> 292 in the cache */
   /* search1347_1_0 -> 3684 in the cache */
   /* search1347_1_1 -> 827 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 3684, 827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 292, 3779 },
      -1,
   } },

   /* replace1347_0_0 -> 132 in the cache */
   /* replace1347_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 132, 135 },
      -1,
   } },
   /* replace1347_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 3781, 133 },
      -1,
   } },

   /* ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      34,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1348_0_1 -> 3711 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3783, 3711 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3784, 3785 },
      -1,
   } },

   /* replace1348_0_0 -> 132 in the cache */
   /* replace1348_0_1 -> 135 in the cache */
   /* replace1348_0 -> 3714 in the cache */
   /* replace1348_1 -> 133 in the cache */
   /* replace1348 -> 3715 in the cache */

   /* ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a */
   /* search1349_0_0 -> 0 in the cache */
   /* search1349_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      4,
   } },
   /* search1349_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 1,
      { 3787, 3 },
      -1,
   } },

   /* replace1349 -> 0 in the cache */

   /* ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1350_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1350_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3789, 128 },
      -1,
   } },
   /* search1350_1_0 -> 32 in the cache */
   /* search1350_1 -> 3789 in the cache */
   /* search1350_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3790, 3789, 130 },
      -1,
   } },

   /* replace1350_0 -> 32 in the cache */
   /* replace1350 -> 3789 in the cache */

   /* ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1351_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1351_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3792, 128 },
      -1,
   } },
   /* search1351_1_0 -> 16 in the cache */
   /* search1351_1 -> 3792 in the cache */
   /* search1351_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3793, 3792, 130 },
      -1,
   } },

   /* replace1351_0 -> 16 in the cache */
   /* replace1351 -> 3792 in the cache */

   /* ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1352_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1352_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3795, 128 },
      -1,
   } },
   /* search1352_1_0 -> 32 in the cache */
   /* search1352_1 -> 3795 in the cache */
   /* search1352_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3796, 3795, 130 },
      -1,
   } },

   /* replace1352_0 -> 32 in the cache */
   /* replace1352 -> 3795 in the cache */

   /* ('bcsel', ('ige', ('ifind_msb_rev', 'a'), 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1353_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1353_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3798, 128 },
      -1,
   } },
   /* search1353_1_0 -> 16 in the cache */
   /* search1353_1 -> 3798 in the cache */
   /* search1353_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3799, 3798, 130 },
      -1,
   } },

   /* replace1353_0 -> 16 in the cache */
   /* replace1353 -> 3798 in the cache */

   /* ('bcsel', ('ige', ('ufind_msb_rev', 'a'), 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a') */
   /* search1354_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1354_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3801, 128 },
      -1,
   } },
   /* search1354_1_0 -> 32 in the cache */
   /* search1354_1 -> 3801 in the cache */
   /* search1354_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3802, 3801, 130 },
      -1,
   } },

   /* replace1354_0 -> 32 in the cache */
   /* replace1354 -> 3801 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1355_0_0 -> 32 in the cache */
   /* search1355_0_1 -> 61 in the cache */
   /* search1355_0 -> 1123 in the cache */
   /* search1355_1_0 -> 32 in the cache */
   /* search1355_1 -> 3789 in the cache */
   /* search1355_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1123, 3789, 130 },
      -1,
   } },

   /* replace1355_0 -> 32 in the cache */
   /* replace1355 -> 3789 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1356_0_0 -> 16 in the cache */
   /* search1356_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 128 },
      -1,
   } },
   /* search1356_1_0 -> 16 in the cache */
   /* search1356_1 -> 3792 in the cache */
   /* search1356_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3805, 3792, 130 },
      -1,
   } },

   /* replace1356_0 -> 16 in the cache */
   /* replace1356 -> 3792 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1357_0_0 -> 32 in the cache */
   /* search1357_0_1 -> 61 in the cache */
   /* search1357_0 -> 1123 in the cache */
   /* search1357_1_0 -> 32 in the cache */
   /* search1357_1 -> 3795 in the cache */
   /* search1357_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1123, 3795, 130 },
      -1,
   } },

   /* replace1357_0 -> 32 in the cache */
   /* replace1357 -> 3795 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1358_0_0 -> 16 in the cache */
   /* search1358_0_1 -> 128 in the cache */
   /* search1358_0 -> 3805 in the cache */
   /* search1358_1_0 -> 16 in the cache */
   /* search1358_1 -> 3798 in the cache */
   /* search1358_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3805, 3798, 130 },
      -1,
   } },

   /* replace1358_0 -> 16 in the cache */
   /* replace1358 -> 3798 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a') */
   /* search1359_0_0 -> 32 in the cache */
   /* search1359_0_1 -> 61 in the cache */
   /* search1359_0 -> 1123 in the cache */
   /* search1359_1_0 -> 32 in the cache */
   /* search1359_1 -> 3801 in the cache */
   /* search1359_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1123, 3801, 130 },
      -1,
   } },

   /* replace1359_0 -> 32 in the cache */
   /* replace1359 -> 3801 in the cache */

   /* ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1360_0_0 -> 16 in the cache */
   /* search1360_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 130 },
      -1,
   } },
   /* search1360_1_0 -> 16 in the cache */
   /* search1360_1 -> 3792 in the cache */
   /* search1360_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3810, 3792, 130 },
      -1,
   } },

   /* replace1360_0 -> 16 in the cache */
   /* replace1360 -> 3792 in the cache */

   /* ('bcsel', ('ine', 'a', -1), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1361_0_0 -> 16 in the cache */
   /* search1361_0_1 -> 130 in the cache */
   /* search1361_0 -> 3810 in the cache */
   /* search1361_1_0 -> 16 in the cache */
   /* search1361_1 -> 3798 in the cache */
   /* search1361_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3810, 3798, 130 },
      -1,
   } },

   /* replace1361_0 -> 16 in the cache */
   /* replace1361 -> 3798 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), -1) => ('ifind_msb_rev', 'a') */
   /* search1362_0_0_0 -> 16 in the cache */
   /* search1362_0_0 -> 3792 in the cache */
   /* search1362_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3792, 130 },
      -1,
   } },
   /* search1362_1_0 -> 19 in the cache */
   /* search1362_1_1_0_0 -> 16 in the cache */
   /* search1362_1_1_0 -> 3792 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3792 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3814 },
      -1,
   } },
   /* search1362_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3813, 3815, 130 },
      -1,
   } },

   /* replace1362_0 -> 16 in the cache */
   /* replace1362 -> 3798 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a') */
   /* search1363_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1363_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3817, 130 },
      -1,
   } },
   /* search1363_1_0 -> 19 in the cache */
   /* search1363_1_1_0_0 -> 16 in the cache */
   /* search1363_1_1_0 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3817 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3819 },
      -1,
   } },
   /* search1363_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3818, 3820, 130 },
      -1,
   } },

   /* replace1363_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a') */
   /* search1364_0_0_0 -> 16 in the cache */
   /* search1364_0_0 -> 3792 in the cache */
   /* search1364_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3792, 130 },
      -1,
   } },
   /* search1364_1 -> 130 in the cache */
   /* search1364_2_0 -> 19 in the cache */
   /* search1364_2_1_0_0 -> 16 in the cache */
   /* search1364_2_1_0 -> 3792 in the cache */
   /* search1364_2_1 -> 3814 in the cache */
   /* search1364_2 -> 3815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3823, 130, 3815 },
      -1,
   } },

   /* replace1364_0 -> 16 in the cache */
   /* replace1364 -> 3798 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1365_0_0_0 -> 16 in the cache */
   /* search1365_0_0 -> 3817 in the cache */
   /* search1365_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3817, 130 },
      -1,
   } },
   /* search1365_1 -> 130 in the cache */
   /* search1365_2_0 -> 19 in the cache */
   /* search1365_2_1_0_0 -> 16 in the cache */
   /* search1365_2_1_0 -> 3817 in the cache */
   /* search1365_2_1 -> 3819 in the cache */
   /* search1365_2 -> 3820 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3825, 130, 3820 },
      -1,
   } },

   /* replace1365_0 -> 16 in the cache */
   /* replace1365 -> 3822 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), ('ifind_msb', 'a')) => ('ifind_msb_rev', 'a') */
   /* search1366_0_0_0 -> 16 in the cache */
   /* search1366_0_0 -> 3792 in the cache */
   /* search1366_0_1 -> 130 in the cache */
   /* search1366_0 -> 3813 in the cache */
   /* search1366_1_0 -> 19 in the cache */
   /* search1366_1_1_0_0 -> 16 in the cache */
   /* search1366_1_1_0 -> 3792 in the cache */
   /* search1366_1_1 -> 3814 in the cache */
   /* search1366_1 -> 3815 in the cache */
   /* search1366_2_0 -> 16 in the cache */
   /* search1366_2 -> 3792 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3813, 3815, 3792 },
      -1,
   } },

   /* replace1366_0 -> 16 in the cache */
   /* replace1366 -> 3798 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1367_0_0_0 -> 16 in the cache */
   /* search1367_0_0 -> 3817 in the cache */
   /* search1367_0_1 -> 130 in the cache */
   /* search1367_0 -> 3818 in the cache */
   /* search1367_1_0 -> 19 in the cache */
   /* search1367_1_1_0_0 -> 16 in the cache */
   /* search1367_1_1_0 -> 3817 in the cache */
   /* search1367_1_1 -> 3819 in the cache */
   /* search1367_1 -> 3820 in the cache */
   /* search1367_2_0 -> 16 in the cache */
   /* search1367_2 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3818, 3820, 3817 },
      -1,
   } },

   /* replace1367_0 -> 16 in the cache */
   /* replace1367 -> 3822 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), ('ifind_msb', 'a'), ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a') */
   /* search1368_0_0_0 -> 16 in the cache */
   /* search1368_0_0 -> 3792 in the cache */
   /* search1368_0_1 -> 130 in the cache */
   /* search1368_0 -> 3823 in the cache */
   /* search1368_1_0 -> 16 in the cache */
   /* search1368_1 -> 3792 in the cache */
   /* search1368_2_0 -> 19 in the cache */
   /* search1368_2_1_0_0 -> 16 in the cache */
   /* search1368_2_1_0 -> 3792 in the cache */
   /* search1368_2_1 -> 3814 in the cache */
   /* search1368_2 -> 3815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3823, 3792, 3815 },
      -1,
   } },

   /* replace1368_0 -> 16 in the cache */
   /* replace1368 -> 3798 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1369_0_0_0 -> 16 in the cache */
   /* search1369_0_0 -> 3817 in the cache */
   /* search1369_0_1 -> 130 in the cache */
   /* search1369_0 -> 3825 in the cache */
   /* search1369_1_0 -> 16 in the cache */
   /* search1369_1 -> 3817 in the cache */
   /* search1369_2_0 -> 19 in the cache */
   /* search1369_2_1_0_0 -> 16 in the cache */
   /* search1369_2_1_0 -> 3817 in the cache */
   /* search1369_2_1 -> 3819 in the cache */
   /* search1369_2 -> 3820 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3825, 3817, 3820 },
      -1,
   } },

   /* replace1369_0 -> 16 in the cache */
   /* replace1369 -> 3822 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a') */
   /* search1370_0_0 -> 16 in the cache */
   /* search1370_0_1 -> 128 in the cache */
   /* search1370_0 -> 3805 in the cache */
   /* search1370_1_0 -> 19 in the cache */
   /* search1370_1_1_0_0 -> 16 in the cache */
   /* search1370_1_1_0 -> 3817 in the cache */
   /* search1370_1_1 -> 3819 in the cache */
   /* search1370_1 -> 3820 in the cache */
   /* search1370_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3805, 3820, 130 },
      -1,
   } },

   /* replace1370_0 -> 16 in the cache */
   /* replace1370 -> 3822 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1371_0_0 -> 16 in the cache */
   /* search1371_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 128 },
      -1,
   } },
   /* search1371_1 -> 130 in the cache */
   /* search1371_2_0 -> 19 in the cache */
   /* search1371_2_1_0_0 -> 16 in the cache */
   /* search1371_2_1_0 -> 3817 in the cache */
   /* search1371_2_1 -> 3819 in the cache */
   /* search1371_2 -> 3820 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3832, 130, 3820 },
      -1,
   } },

   /* replace1371_0 -> 16 in the cache */
   /* replace1371 -> 3822 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1372_0_0 -> 16 in the cache */
   /* search1372_0_1 -> 128 in the cache */
   /* search1372_0 -> 3805 in the cache */
   /* search1372_1_0 -> 19 in the cache */
   /* search1372_1_1_0_0 -> 16 in the cache */
   /* search1372_1_1_0 -> 3817 in the cache */
   /* search1372_1_1 -> 3819 in the cache */
   /* search1372_1 -> 3820 in the cache */
   /* search1372_2_0 -> 16 in the cache */
   /* search1372_2 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3805, 3820, 3817 },
      -1,
   } },

   /* replace1372_0 -> 16 in the cache */
   /* replace1372 -> 3822 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1373_0_0 -> 16 in the cache */
   /* search1373_0_1 -> 128 in the cache */
   /* search1373_0 -> 3832 in the cache */
   /* search1373_1_0 -> 16 in the cache */
   /* search1373_1 -> 3817 in the cache */
   /* search1373_2_0 -> 19 in the cache */
   /* search1373_2_1_0_0 -> 16 in the cache */
   /* search1373_2_1_0 -> 3817 in the cache */
   /* search1373_2_1 -> 3819 in the cache */
   /* search1373_2 -> 3820 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3832, 3817, 3820 },
      -1,
   } },

   /* replace1373_0 -> 16 in the cache */
   /* replace1373 -> 3822 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), -1) => ('ifind_msb', 'a') */
   /* search1374_0_0_0 -> 16 in the cache */
   /* search1374_0_0 -> 3798 in the cache */
   /* search1374_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3798, 130 },
      -1,
   } },
   /* search1374_1_0 -> 19 in the cache */
   /* search1374_1_1_0_0 -> 16 in the cache */
   /* search1374_1_1_0 -> 3798 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3798 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3837 },
      -1,
   } },
   /* search1374_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3836, 3838, 130 },
      -1,
   } },

   /* replace1374_0 -> 16 in the cache */
   /* replace1374 -> 3792 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a') */
   /* search1375_0_0_0 -> 16 in the cache */
   /* search1375_0_0 -> 3822 in the cache */
   /* search1375_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3822, 130 },
      -1,
   } },
   /* search1375_1_0 -> 19 in the cache */
   /* search1375_1_1_0_0 -> 16 in the cache */
   /* search1375_1_1_0 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3822 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3841 },
      -1,
   } },
   /* search1375_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3840, 3842, 130 },
      -1,
   } },

   /* replace1375_0 -> 16 in the cache */
   /* replace1375 -> 3817 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a') */
   /* search1376_0_0_0 -> 16 in the cache */
   /* search1376_0_0 -> 3798 in the cache */
   /* search1376_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3798, 130 },
      -1,
   } },
   /* search1376_1 -> 130 in the cache */
   /* search1376_2_0 -> 19 in the cache */
   /* search1376_2_1_0_0 -> 16 in the cache */
   /* search1376_2_1_0 -> 3798 in the cache */
   /* search1376_2_1 -> 3837 in the cache */
   /* search1376_2 -> 3838 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3844, 130, 3838 },
      -1,
   } },

   /* replace1376_0 -> 16 in the cache */
   /* replace1376 -> 3792 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1377_0_0_0 -> 16 in the cache */
   /* search1377_0_0 -> 3822 in the cache */
   /* search1377_0_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3822, 130 },
      -1,
   } },
   /* search1377_1 -> 130 in the cache */
   /* search1377_2_0 -> 19 in the cache */
   /* search1377_2_1_0_0 -> 16 in the cache */
   /* search1377_2_1_0 -> 3822 in the cache */
   /* search1377_2_1 -> 3841 in the cache */
   /* search1377_2 -> 3842 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3846, 130, 3842 },
      -1,
   } },

   /* replace1377_0 -> 16 in the cache */
   /* replace1377 -> 3817 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), ('ifind_msb_rev', 'a')) => ('ifind_msb', 'a') */
   /* search1378_0_0_0 -> 16 in the cache */
   /* search1378_0_0 -> 3798 in the cache */
   /* search1378_0_1 -> 130 in the cache */
   /* search1378_0 -> 3836 in the cache */
   /* search1378_1_0 -> 19 in the cache */
   /* search1378_1_1_0_0 -> 16 in the cache */
   /* search1378_1_1_0 -> 3798 in the cache */
   /* search1378_1_1 -> 3837 in the cache */
   /* search1378_1 -> 3838 in the cache */
   /* search1378_2_0 -> 16 in the cache */
   /* search1378_2 -> 3798 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3836, 3838, 3798 },
      -1,
   } },

   /* replace1378_0 -> 16 in the cache */
   /* replace1378 -> 3792 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a') */
   /* search1379_0_0_0 -> 16 in the cache */
   /* search1379_0_0 -> 3822 in the cache */
   /* search1379_0_1 -> 130 in the cache */
   /* search1379_0 -> 3840 in the cache */
   /* search1379_1_0 -> 19 in the cache */
   /* search1379_1_1_0_0 -> 16 in the cache */
   /* search1379_1_1_0 -> 3822 in the cache */
   /* search1379_1_1 -> 3841 in the cache */
   /* search1379_1 -> 3842 in the cache */
   /* search1379_2_0 -> 16 in the cache */
   /* search1379_2 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3840, 3842, 3822 },
      -1,
   } },

   /* replace1379_0 -> 16 in the cache */
   /* replace1379 -> 3817 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), ('ifind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a') */
   /* search1380_0_0_0 -> 16 in the cache */
   /* search1380_0_0 -> 3798 in the cache */
   /* search1380_0_1 -> 130 in the cache */
   /* search1380_0 -> 3844 in the cache */
   /* search1380_1_0 -> 16 in the cache */
   /* search1380_1 -> 3798 in the cache */
   /* search1380_2_0 -> 19 in the cache */
   /* search1380_2_1_0_0 -> 16 in the cache */
   /* search1380_2_1_0 -> 3798 in the cache */
   /* search1380_2_1 -> 3837 in the cache */
   /* search1380_2 -> 3838 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3844, 3798, 3838 },
      -1,
   } },

   /* replace1380_0 -> 16 in the cache */
   /* replace1380 -> 3792 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1381_0_0_0 -> 16 in the cache */
   /* search1381_0_0 -> 3822 in the cache */
   /* search1381_0_1 -> 130 in the cache */
   /* search1381_0 -> 3846 in the cache */
   /* search1381_1_0 -> 16 in the cache */
   /* search1381_1 -> 3822 in the cache */
   /* search1381_2_0 -> 19 in the cache */
   /* search1381_2_1_0_0 -> 16 in the cache */
   /* search1381_2_1_0 -> 3822 in the cache */
   /* search1381_2_1 -> 3841 in the cache */
   /* search1381_2 -> 3842 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3846, 3822, 3842 },
      -1,
   } },

   /* replace1381_0 -> 16 in the cache */
   /* replace1381 -> 3817 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a') */
   /* search1382_0_0 -> 16 in the cache */
   /* search1382_0_1 -> 128 in the cache */
   /* search1382_0 -> 3805 in the cache */
   /* search1382_1_0 -> 19 in the cache */
   /* search1382_1_1_0_0 -> 16 in the cache */
   /* search1382_1_1_0 -> 3822 in the cache */
   /* search1382_1_1 -> 3841 in the cache */
   /* search1382_1 -> 3842 in the cache */
   /* search1382_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3805, 3842, 130 },
      -1,
   } },

   /* replace1382_0 -> 16 in the cache */
   /* replace1382 -> 3817 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1383_0_0 -> 16 in the cache */
   /* search1383_0_1 -> 128 in the cache */
   /* search1383_0 -> 3832 in the cache */
   /* search1383_1 -> 130 in the cache */
   /* search1383_2_0 -> 19 in the cache */
   /* search1383_2_1_0_0 -> 16 in the cache */
   /* search1383_2_1_0 -> 3822 in the cache */
   /* search1383_2_1 -> 3841 in the cache */
   /* search1383_2 -> 3842 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3832, 130, 3842 },
      -1,
   } },

   /* replace1383_0 -> 16 in the cache */
   /* replace1383 -> 3817 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a') */
   /* search1384_0_0 -> 16 in the cache */
   /* search1384_0_1 -> 128 in the cache */
   /* search1384_0 -> 3805 in the cache */
   /* search1384_1_0 -> 19 in the cache */
   /* search1384_1_1_0_0 -> 16 in the cache */
   /* search1384_1_1_0 -> 3822 in the cache */
   /* search1384_1_1 -> 3841 in the cache */
   /* search1384_1 -> 3842 in the cache */
   /* search1384_2_0 -> 16 in the cache */
   /* search1384_2 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3805, 3842, 3822 },
      -1,
   } },

   /* replace1384_0 -> 16 in the cache */
   /* replace1384 -> 3817 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1385_0_0 -> 16 in the cache */
   /* search1385_0_1 -> 128 in the cache */
   /* search1385_0 -> 3832 in the cache */
   /* search1385_1_0 -> 16 in the cache */
   /* search1385_1 -> 3822 in the cache */
   /* search1385_2_0 -> 19 in the cache */
   /* search1385_2_1_0_0 -> 16 in the cache */
   /* search1385_2_1_0 -> 3822 in the cache */
   /* search1385_2_1 -> 3841 in the cache */
   /* search1385_2 -> 3842 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3832, 3822, 3842 },
      -1,
   } },

   /* replace1385_0 -> 16 in the cache */
   /* replace1385 -> 3817 in the cache */

   /* ('find_lsb', ('bitfield_reverse', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1386_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_reverse,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3856 },
      -1,
   } },

   /* replace1386_0 -> 16 in the cache */
   /* replace1386 -> 3822 in the cache */

   /* ('ufind_msb_rev', ('bitfield_reverse', 'a')) => ('find_lsb', 'a') */
   /* search1387_0_0 -> 16 in the cache */
   /* search1387_0 -> 3856 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 3856 },
      -1,
   } },

   /* replace1387_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('ifind_msb', ('f2i32(is_used_once)', 'a')) => ('ufind_msb', ('f2i32', ('fabs', 'a'))) */
   /* search1388_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 32 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 3860 },
      -1,
   } },

   /* replace1388_0_0_0 -> 32 in the cache */
   /* replace1388_0_0 -> 99 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 99 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 3862 },
      -1,
   } },

   /* ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1389_0_0 -> 531 in the cache */
   /* search1389_0_1 -> 285 in the cache */
   /* search1389_0_2 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 285, 280 },
      0,
   } },
   /* search1389_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3864, 3 },
      -1,
   } },

   /* replace1389_0 -> 531 in the cache */
   /* replace1389_1_0 -> 3 in the cache */
   /* replace1389_1 -> 1100 in the cache */
   /* replace1389_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 1100, 3 },
      -1,
   } },

   /* ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1390_0_0 -> 531 in the cache */
   /* search1390_0_1 -> 280 in the cache */
   /* search1390_0_2 -> 285 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 280, 285 },
      0,
   } },
   /* search1390_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3867, 3 },
      -1,
   } },

   /* replace1390_0 -> 531 in the cache */
   /* replace1390_1 -> 3 in the cache */
   /* replace1390_2_0 -> 3 in the cache */
   /* replace1390_2 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 3, 1100 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1391_0_0 -> 531 in the cache */
   /* search1391_0_1 -> 288 in the cache */
   /* search1391_0_2 -> 254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 288, 254 },
      0,
   } },
   /* search1391_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3870, 36 },
      -1,
   } },

   /* replace1391_0 -> 531 in the cache */
   /* replace1391_1_0 -> 36 in the cache */
   /* replace1391_1 -> 610 in the cache */
   /* replace1391_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 610, 36 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1392_0_0 -> 531 in the cache */
   /* search1392_0_1 -> 254 in the cache */
   /* search1392_0_2 -> 288 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 254, 288 },
      0,
   } },
   /* search1392_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3873, 36 },
      -1,
   } },

   /* replace1392_0 -> 531 in the cache */
   /* replace1392_1 -> 36 in the cache */
   /* replace1392_2_0 -> 36 in the cache */
   /* replace1392_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 36, 610 },
      -1,
   } },

   /* ('fabs', ('bcsel(is_used_once)', 'b', ('fneg', 'a'), 'a')) => ('fabs', 'a') */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1393_0_1_0 -> 1127 in the cache */
   /* search1393_0_1 -> 1592 in the cache */
   /* search1393_0_2 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3876, 1592, 1127 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3877 },
      -1,
   } },

   /* replace1393_0 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 1127 },
      -1,
   } },

   /* ('fabs', ('bcsel(is_used_once)', 'b', 'a', ('fneg', 'a'))) => ('fabs', 'a') */
   /* search1394_0_0 -> 3876 in the cache */
   /* search1394_0_1 -> 1127 in the cache */
   /* search1394_0_2_0 -> 1127 in the cache */
   /* search1394_0_2 -> 1592 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3876, 1127, 1592 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3880 },
      -1,
   } },

   /* replace1394_0 -> 1127 in the cache */
   /* replace1394 -> 3879 in the cache */

   /* ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a') */
   /* search1395_0_0 -> 32 in the cache */
   /* search1395_0_1 -> 117 in the cache */
   /* search1395_0 -> 1185 in the cache */
   /* search1395_1_0 -> 32 in the cache */
   /* search1395_1 -> 107 in the cache */
   /* search1395_2 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1185, 107, 32 },
      -1,
   } },

   /* replace1395_0 -> 32 in the cache */
   /* replace1395 -> 99 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd') */
   /* search1396_0 -> 770 in the cache */
   /* search1396_1_0 -> 1147 in the cache */
   /* search1396_1_1 -> 158 in the cache */
   /* search1396_1_2 -> 159 in the cache */
   /* search1396_1 -> 1335 in the cache */
   /* search1396_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1335, 159 },
      -1,
   } },

   /* replace1396_0_0 -> 770 in the cache */
   /* replace1396_0_1 -> 1147 in the cache */
   /* replace1396_0 -> 1159 in the cache */
   /* replace1396_1 -> 158 in the cache */
   /* replace1396_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1159, 158, 159 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd') */
   /* search1397_0 -> 770 in the cache */
   /* search1397_1 -> 156 in the cache */
   /* search1397_2_0 -> 474 in the cache */
   /* search1397_2_1 -> 156 in the cache */
   /* search1397_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 474, 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 156, 3885 },
      -1,
   } },

   /* replace1397_0_0 -> 770 in the cache */
   /* replace1397_0_1 -> 474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 770, 474 },
      -1,
   } },
   /* replace1397_1 -> 156 in the cache */
   /* replace1397_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3887, 156, 159 },
      -1,
   } },

   /* ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b')))) */
   /* search1398_0 -> 0 in the cache */
   /* search1398_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1398_0 -> 0 in the cache */
   /* replace1398_1_0 -> 3 in the cache */
   /* replace1398_1_1_0_0 -> 0 in the cache */
   /* replace1398_1_1_0_1 -> 3 in the cache */
   /* replace1398_1_1_0 -> 3065 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 3065 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3890 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3891 },
      -1,
   } },

   /* ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b')))) */
   /* search1399_0 -> 0 in the cache */
   /* search1399_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1399_0 -> 0 in the cache */
   /* replace1399_1_0 -> 3 in the cache */
   /* replace1399_1_1_0_0 -> 0 in the cache */
   /* replace1399_1_1_0_1 -> 3 in the cache */
   /* replace1399_1_1_0 -> 3065 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 3065 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3894 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3895 },
      -1,
   } },

   /* ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a')) */
   /* search1400_0 -> 0 in the cache */
   /* search1400_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1400_0_0_0 -> 0 in the cache */
   /* replace1400_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1400_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3898, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 3899 },
      -1,
   } },

   /* ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b')) */
   /* search1401_0 -> 0 in the cache */
   /* search1401_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1401_0_0 -> 0 in the cache */
   /* replace1401_0_1 -> 3 in the cache */
   /* replace1401_0 -> 1080 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1080 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* base */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* insert */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 3903, 3904, 3905, 3906 },
      -1,
   } },

   /* replace1402_0_0 -> 19 in the cache */
   /* replace1402_0_1 -> 3906 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 3906 },
      -1,
   } },
   /* replace1402_1 -> 3904 in the cache */
   /* replace1402_2_0_0 -> 3906 in the cache */
   /* replace1402_2_0_1 -> 3905 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3906, 3905 },
      -1,
   } },
   /* replace1402_2_1 -> 3904 in the cache */
   /* replace1402_2_2 -> 3903 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 3909, 3904, 3903 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3908, 3904, 3910 },
      -1,
   } },

   /* ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1403_0 -> 0 in the cache */
   /* search1403_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1403_0_0 -> 0 in the cache */
   /* replace1403_0_1 -> 3 in the cache */
   /* replace1403_0 -> 207 in the cache */
   /* replace1403_1_0_0 -> 0 in the cache */
   /* replace1403_1_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1403_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3913, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 207, 3914 },
      -1,
   } },

   /* ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1404_0 -> 0 in the cache */
   /* search1404_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1404_0_0 -> 0 in the cache */
   /* replace1404_0_1 -> 3 in the cache */
   /* replace1404_0 -> 207 in the cache */
   /* replace1404_1_0_0 -> 0 in the cache */
   /* replace1404_1_0_1 -> 3 in the cache */
   /* replace1404_1_0 -> 3913 in the cache */
   /* replace1404_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3913, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 207, 3917 },
      -1,
   } },

   /* ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1405_0 -> 0 in the cache */
   /* search1405_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1405_0_0 -> 0 in the cache */
   /* replace1405_0_1 -> 3 in the cache */
   /* replace1405_0 -> 2506 in the cache */
   /* replace1405_1_0_0 -> 0 in the cache */
   /* replace1405_1_0_1 -> 3 in the cache */
   /* replace1405_1_0 -> 2879 in the cache */
   /* replace1405_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2879, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2506, 3920 },
      -1,
   } },

   /* ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1406_0 -> 0 in the cache */
   /* search1406_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1406_0_0 -> 0 in the cache */
   /* replace1406_0_1 -> 3 in the cache */
   /* replace1406_0 -> 2506 in the cache */
   /* replace1406_1_0_0 -> 0 in the cache */
   /* replace1406_1_0_1 -> 3 in the cache */
   /* replace1406_1_0 -> 2879 in the cache */
   /* replace1406_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2879, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2506, 3923 },
      -1,
   } },

   /* ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1407_0 -> 21 in the cache */
   /* search1407_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1407_0_0 -> 21 in the cache */
   /* replace1407_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1407_1_0_0 -> 21 in the cache */
   /* replace1407_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1407_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3927, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3926, 3928 },
      -1,
   } },

   /* ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1408_0 -> 21 in the cache */
   /* search1408_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1408_0_0 -> 21 in the cache */
   /* replace1408_0_1 -> 41 in the cache */
   /* replace1408_0 -> 3926 in the cache */
   /* replace1408_1_0_0 -> 21 in the cache */
   /* replace1408_1_0_1 -> 41 in the cache */
   /* replace1408_1_0 -> 3927 in the cache */
   /* replace1408_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3927, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3926, 3931 },
      -1,
   } },

   /* ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1409_0 -> 21 in the cache */
   /* search1409_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1409_0_0 -> 21 in the cache */
   /* replace1409_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1409_1_0_0 -> 21 in the cache */
   /* replace1409_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1409_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3935, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3934, 3936 },
      -1,
   } },

   /* ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1410_0 -> 21 in the cache */
   /* search1410_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1410_0_0 -> 21 in the cache */
   /* replace1410_0_1 -> 41 in the cache */
   /* replace1410_0 -> 3934 in the cache */
   /* replace1410_1_0_0 -> 21 in the cache */
   /* replace1410_1_0_1 -> 41 in the cache */
   /* replace1410_1_0 -> 3935 in the cache */
   /* replace1410_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3935, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3934, 3939 },
      -1,
   } },

   /* ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0)) */
   /* search1411_0 -> 16 in the cache */
   /* search1411_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1411_0 -> 16 in the cache */
   /* replace1411_1_0 -> 36 in the cache */
   /* replace1411_1_1 -> 128 in the cache */
   /* replace1411_1 -> 447 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 447 },
      -1,
   } },

   /* ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0)) */
   /* search1412_0 -> 16 in the cache */
   /* search1412_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1412_0 -> 16 in the cache */
   /* replace1412_1_0 -> 36 in the cache */
   /* replace1412_1_1 -> 128 in the cache */
   /* replace1412_1 -> 455 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 455 },
      -1,
   } },

   /* ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1413_0 -> 21 in the cache */
   /* search1413_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1413_0_0_0 -> 21 in the cache */
   /* replace1413_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1413_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3946, 21 },
      -1,
   } },
   /* replace1413_1 -> 1512 in the cache */
   /* replace1413_2_0 -> 21 in the cache */
   /* replace1413_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3947, 1512, 3948 },
      -1,
   } },

   /* ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1414_0 -> 0 in the cache */
   /* search1414_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1414_0_0_0 -> 0 in the cache */
   /* replace1414_0_0_1 -> 3 in the cache */
   /* replace1414_0_0 -> 3898 in the cache */
   /* replace1414_0_1 -> 0 in the cache */
   /* replace1414_0 -> 3899 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* replace1414_2_0 -> 0 in the cache */
   /* replace1414_2_1 -> 3 in the cache */
   /* replace1414_2 -> 219 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3899, 3951, 219 },
      -1,
   } },

   /* ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1415_0 -> 0 in the cache */
   /* search1415_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1415_0_0 -> 0 in the cache */
   /* replace1415_0_1 -> 3 in the cache */
   /* replace1415_0 -> 1080 in the cache */
   /* replace1415_1 -> 86 in the cache */
   /* replace1415_2_0 -> 0 in the cache */
   /* replace1415_2_1 -> 3 in the cache */
   /* replace1415_2 -> 3663 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1080, 86, 3663 },
      -1,
   } },

   /* ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1416_0 -> 21 in the cache */
   /* search1416_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1416_0_0 -> 21 in the cache */
   /* replace1416_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* replace1416_1 -> 910 in the cache */
   /* replace1416_2_0 -> 21 in the cache */
   /* replace1416_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3956, 910, 3957 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807)) */
   /* search1417_0 -> 21 in the cache */
   /* search1417_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1417_0_0_0_0 -> 21 in the cache */
   /* replace1417_0_0_0_1 -> 910 in the cache */
   /* replace1417_0_0_0 -> 1502 in the cache */
   /* replace1417_0_0_1_0 -> 41 in the cache */
   /* replace1417_0_0_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 41, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1502, 3960 },
      -1,
   } },
   /* replace1417_0_1_0_0 -> 21 in the cache */
   /* replace1417_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1417_0_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3962, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3961, 3963 },
      -1,
   } },
   /* replace1417_1 -> 22 in the cache */
   /* replace1417_2_0_0_0_0 -> 21 in the cache */
   /* replace1417_2_0_0_0_1 -> 910 in the cache */
   /* replace1417_2_0_0_0 -> 1502 in the cache */
   /* replace1417_2_0_0_1_0 -> 41 in the cache */
   /* replace1417_2_0_0_1_1 -> 910 in the cache */
   /* replace1417_2_0_0_1 -> 3960 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      4, 1,
      { 1502, 3960 },
      -1,
   } },
   /* replace1417_2_0_1_0_0 -> 21 in the cache */
   /* replace1417_2_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1417_2_0_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3966, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 3,
      { 3965, 3967 },
      -1,
   } },
   /* replace1417_2_1_0 -> 21 in the cache */
   /* replace1417_2_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      6, 1,
      { 21, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffffffffffff /* 9223372036854775807 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3968, 3969, 3970 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 7,
      { 3964, 22, 3971 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807)) */
   /* search1418_0 -> 21 in the cache */
   /* search1418_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1418_0_0_0_0 -> 21 in the cache */
   /* replace1418_0_0_0_1 -> 910 in the cache */
   /* replace1418_0_0_0 -> 1502 in the cache */
   /* replace1418_0_0_1_0 -> 41 in the cache */
   /* replace1418_0_0_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1502, 3974 },
      -1,
   } },
   /* replace1418_0_1_0_0 -> 21 in the cache */
   /* replace1418_0_1_0_1 -> 41 in the cache */
   /* replace1418_0_1_0 -> 3957 in the cache */
   /* replace1418_0_1_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3957, 910 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3975, 3976 },
      -1,
   } },
   /* replace1418_1 -> 22 in the cache */
   /* replace1418_2_0_0_0_0 -> 21 in the cache */
   /* replace1418_2_0_0_0_1 -> 910 in the cache */
   /* replace1418_2_0_0_0 -> 1502 in the cache */
   /* replace1418_2_0_0_1_0 -> 41 in the cache */
   /* replace1418_2_0_0_1_1 -> 910 in the cache */
   /* replace1418_2_0_0_1 -> 3974 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 1,
      { 1502, 3974 },
      -1,
   } },
   /* replace1418_2_0_1_0_0 -> 21 in the cache */
   /* replace1418_2_0_1_0_1 -> 41 in the cache */
   /* replace1418_2_0_1_0 -> 3957 in the cache */
   /* replace1418_2_0_1_1 -> 910 in the cache */
   /* replace1418_2_0_1 -> 3976 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 3978, 3976 },
      -1,
   } },
   /* replace1418_2_1_0 -> 21 in the cache */
   /* replace1418_2_1_1 -> 41 in the cache */
   /* replace1418_2_1 -> 3957 in the cache */
   /* replace1418_2_2 -> 3970 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3979, 3957, 3970 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3977, 22, 3980 },
      -1,
   } },

   /* ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1419_0_0 -> 21 in the cache */
   /* search1419_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1419_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3982, 910 },
      -1,
   } },

   /* replace1419_0_0_0 -> 21 in the cache */
   /* replace1419_0_0 -> 2894 in the cache */
   /* replace1419_0_1_0 -> 41 in the cache */
   /* replace1419_0_1 -> 2895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2894, 2895 },
      -1,
   } },
   /* replace1419_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3984, 128 },
      -1,
   } },

   /* ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1420_0_0 -> 21 in the cache */
   /* search1420_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1420_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3986, 910 },
      -1,
   } },

   /* replace1420_0_0_0 -> 21 in the cache */
   /* replace1420_0_0 -> 2894 in the cache */
   /* replace1420_0_1_0 -> 41 in the cache */
   /* replace1420_0_1 -> 2895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2894, 2895 },
      -1,
   } },
   /* replace1420_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3988, 128 },
      -1,
   } },

   /* ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1421_0_0 -> 21 in the cache */
   /* search1421_0_1 -> 41 in the cache */
   /* search1421_0 -> 3982 in the cache */
   /* search1421_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3982, 910 },
      -1,
   } },

   /* replace1421_0_0_0 -> 21 in the cache */
   /* replace1421_0_0 -> 2894 in the cache */
   /* replace1421_0_1_0 -> 41 in the cache */
   /* replace1421_0_1 -> 2895 in the cache */
   /* replace1421_0 -> 3984 in the cache */
   /* replace1421_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3984, 128 },
      -1,
   } },

   /* ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1422_0_0 -> 21 in the cache */
   /* search1422_0_1 -> 41 in the cache */
   /* search1422_0 -> 3986 in the cache */
   /* search1422_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3986, 910 },
      -1,
   } },

   /* replace1422_0_0_0 -> 21 in the cache */
   /* replace1422_0_0 -> 2894 in the cache */
   /* replace1422_0_1_0 -> 41 in the cache */
   /* replace1422_0_1 -> 2895 in the cache */
   /* replace1422_0 -> 3988 in the cache */
   /* replace1422_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3988, 128 },
      -1,
   } },

   /* ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1423_0 -> 21 in the cache */
   /* search1423_1 -> 910 in the cache */
   /* search1423 -> 1502 in the cache */

   /* replace1423_0_0 -> 21 in the cache */
   /* replace1423_0 -> 2894 in the cache */
   /* replace1423_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2894, 128 },
      -1,
   } },

   /* ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1424_0 -> 21 in the cache */
   /* search1424_1 -> 910 in the cache */
   /* search1424 -> 1499 in the cache */

   /* replace1424_0_0 -> 21 in the cache */
   /* replace1424_0 -> 2894 in the cache */
   /* replace1424_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2894, 128 },
      -1,
   } },

   /* ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1425_0 -> 21 in the cache */
   /* search1425_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 21, 910 },
      -1,
   } },

   /* replace1425_0_0_0 -> 21 in the cache */
   /* replace1425_0_0 -> 2900 in the cache */
   /* replace1425_0_1_0 -> 21 in the cache */
   /* replace1425_0_1 -> 2894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 2900, 2894 },
      -1,
   } },
   /* replace1425_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3997, 128 },
      -1,
   } },

   /* ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1426_0 -> 21 in the cache */
   /* search1426_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 21, 910 },
      -1,
   } },

   /* replace1426_0_0_0 -> 21 in the cache */
   /* replace1426_0_0 -> 2900 in the cache */
   /* replace1426_0_1_0 -> 21 in the cache */
   /* replace1426_0_1 -> 2894 in the cache */
   /* replace1426_0 -> 3997 in the cache */
   /* replace1426_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 3997, 128 },
      -1,
   } },

   /* ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1427_0 -> 910 in the cache */
   /* search1427_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 910, 21 },
      -1,
   } },

   /* replace1427_0_0_0 -> 21 in the cache */
   /* replace1427_0_0 -> 2900 in the cache */
   /* replace1427_0_1_0 -> 21 in the cache */
   /* replace1427_0_1 -> 2894 in the cache */
   /* replace1427_0 -> 3997 in the cache */
   /* replace1427_1 -> 128 in the cache */
   /* replace1427 -> 3998 in the cache */

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset')))) */
   /* search1428_0 -> 3903 in the cache */
   /* search1428_1 -> 3904 in the cache */
   /* search1428_2 -> 3905 in the cache */
   /* search1428_3 -> 3906 in the cache */
   /* search1428 -> 3907 in the cache */

   /* replace1428_0_0 -> 19 in the cache */
   /* replace1428_0_1 -> 3906 in the cache */
   /* replace1428_0 -> 3908 in the cache */
   /* replace1428_1 -> 3904 in the cache */
   /* replace1428_2_0_0 -> 3903 in the cache */
   /* replace1428_2_0_1_0_0_0_0 -> 406 in the cache */
   /* replace1428_2_0_1_0_0_0_1 -> 3906 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 406, 3906 },
      -1,
   } },
   /* replace1428_2_0_1_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 4002, 406 },
      -1,
   } },
   /* replace1428_2_0_1_0_1 -> 3905 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4003, 3905 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 4004 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3903, 4005 },
      -1,
   } },
   /* replace1428_2_1_0_0 -> 3904 in the cache */
   /* replace1428_2_1_0_1 -> 3905 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3904, 3905 },
      -1,
   } },
   /* replace1428_2_1_1_0_0_0 -> 406 in the cache */
   /* replace1428_2_1_1_0_0_1 -> 3906 in the cache */
   /* replace1428_2_1_1_0_0 -> 4002 in the cache */
   /* replace1428_2_1_1_0_1 -> 406 in the cache */
   /* replace1428_2_1_1_0 -> 4003 in the cache */
   /* replace1428_2_1_1_1 -> 3905 in the cache */
   /* replace1428_2_1_1 -> 4004 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 4007, 4004 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4006, 4008 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 3908, 3904, 4009 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base')) */
   /* search1429_0 -> 3903 in the cache */
   /* search1429_1 -> 3904 in the cache */
   /* search1429_2 -> 3905 in the cache */
   /* search1429_3 -> 3906 in the cache */
   /* search1429 -> 3907 in the cache */

   /* replace1429_0_0 -> 19 in the cache */
   /* replace1429_0_1 -> 3906 in the cache */
   /* replace1429_0 -> 3908 in the cache */
   /* replace1429_1 -> 3904 in the cache */
   /* replace1429_2_0_0 -> 3906 in the cache */
   /* replace1429_2_0_1 -> 3905 in the cache */
   /* replace1429_2_0 -> 3909 in the cache */
   /* replace1429_2_1_0 -> 3904 in the cache */
   /* replace1429_2_1_1 -> 3905 in the cache */
   /* replace1429_2_1 -> 4007 in the cache */
   /* replace1429_2_2 -> 3903 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 3909, 4007, 3903 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3908, 3904, 4011 },
      -1,
   } },

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 0,
      { 4013, 4014, 4015 },
      -1,
   } },

   /* replace1430_0_0 -> 19 in the cache */
   /* replace1430_0_1 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 4015 },
      -1,
   } },
   /* replace1430_1 -> 4013 in the cache */
   /* replace1430_2_0 -> 4013 in the cache */
   /* replace1430_2_1 -> 4014 in the cache */
   /* replace1430_2_2 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 4013, 4014, 4015 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4017, 4013, 4018 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits')) */
   /* search1431_0 -> 4013 in the cache */
   /* search1431_1 -> 4014 in the cache */
   /* search1431_2 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 4013, 4014, 4015 },
      -1,
   } },

   /* replace1431_0_0 -> 19 in the cache */
   /* replace1431_0_1 -> 4015 in the cache */
   /* replace1431_0 -> 4017 in the cache */
   /* replace1431_1 -> 4013 in the cache */
   /* replace1431_2_0 -> 4013 in the cache */
   /* replace1431_2_1 -> 4014 in the cache */
   /* replace1431_2_2 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 4013, 4014, 4015 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4017, 4013, 4021 },
      -1,
   } },

   /* ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b') */
   /* search1432_0 -> 0 in the cache */
   /* search1432_1 -> 3 in the cache */
   /* search1432_2 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 0, 3, 86 },
      -1,
   } },

   /* replace1432_0 -> 0 in the cache */
   /* replace1432_1 -> 3 in the cache */
   /* replace1432 -> 2887 in the cache */

   /* ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c') */
   /* search1433_0 -> 132 in the cache */
   /* search1433_1_0 -> 132 in the cache */
   /* search1433_1_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 132, 133 },
      -1,
   } },
   /* search1433_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 1,
      { 132, 4024, 135 },
      -1,
   } },

   /* replace1433_0 -> 132 in the cache */
   /* replace1433_1 -> 133 in the cache */
   /* replace1433_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 132, 133, 135 },
      -1,
   } },

   /* ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits') */
   /* search1434_0 -> 4013 in the cache */
   /* search1434_1 -> 4014 in the cache */
   /* search1434_2_0 -> 19 in the cache */
   /* search1434_2_1 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 4015 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 4013, 4014, 4027 },
      -1,
   } },

   /* replace1434_0 -> 4013 in the cache */
   /* replace1434_1 -> 4014 in the cache */
   /* replace1434_2 -> 4015 in the cache */
   /* replace1434 -> 4021 in the cache */

   /* ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits') */
   /* search1435_0 -> 4013 in the cache */
   /* search1435_1_0 -> 19 in the cache */
   /* search1435_1_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 4014 },
      -1,
   } },
   /* search1435_2 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 4013, 4029, 4015 },
      -1,
   } },

   /* replace1435_0 -> 4013 in the cache */
   /* replace1435_1 -> 4014 in the cache */
   /* replace1435_2 -> 4015 in the cache */
   /* replace1435 -> 4021 in the cache */

   /* ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits') */
   /* search1436_0 -> 4013 in the cache */
   /* search1436_1 -> 4014 in the cache */
   /* search1436_2_0 -> 19 in the cache */
   /* search1436_2_1 -> 4015 in the cache */
   /* search1436_2 -> 4027 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 4013, 4014, 4027 },
      -1,
   } },

   /* replace1436_0 -> 4013 in the cache */
   /* replace1436_1 -> 4014 in the cache */
   /* replace1436_2 -> 4015 in the cache */
   /* replace1436 -> 4018 in the cache */

   /* ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits') */
   /* search1437_0 -> 4013 in the cache */
   /* search1437_1_0 -> 19 in the cache */
   /* search1437_1_1 -> 4014 in the cache */
   /* search1437_1 -> 4029 in the cache */
   /* search1437_2 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 4013, 4029, 4015 },
      -1,
   } },

   /* replace1437_0 -> 4013 in the cache */
   /* replace1437_1 -> 4014 in the cache */
   /* replace1437_2 -> 4015 in the cache */
   /* replace1437 -> 4018 in the cache */

   /* ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1438_1_0 -> 19 in the cache */
   /* search1438_1_1 -> 4014 in the cache */
   /* search1438_1 -> 4029 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 4033, 4029 },
      -1,
   } },

   /* replace1438_0 -> 4033 in the cache */
   /* replace1438_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 4033, 4014 },
      -1,
   } },

   /* ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset') */
   /* search1439_0_0 -> 19 in the cache */
   /* search1439_0_1 -> 4033 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 4033 },
      -1,
   } },
   /* search1439_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 4036, 4014 },
      -1,
   } },

   /* replace1439_0 -> 4033 in the cache */
   /* replace1439_1 -> 4014 in the cache */
   /* replace1439 -> 4035 in the cache */

   /* ('ult', 'a', ('umin', ('iand', 'a', 'b'), 'c')) => False */
   /* search1440_0 -> 132 in the cache */
   /* search1440_1_0_0 -> 132 in the cache */
   /* search1440_1_0_1 -> 133 in the cache */
   /* search1440_1_0 -> 200 in the cache */
   /* search1440_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 200, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 2,
      { 132, 4038 },
      -1,
   } },

   /* replace1440 -> 1090 in the cache */

   /* ('ult', 31, ('umin', '#bits(is_ult_32)', 'a')) => False */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x1f /* 31 */ },
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* bits */
      true,
      nir_type_invalid,
      35,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1441_1_1 -> 1127 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 4041, 1127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4040, 4042 },
      -1,
   } },

   /* replace1441 -> 1090 in the cache */

   /* ('ubfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ubfe', 'value', 'offset', 'width') */
   /* search1442_0 -> 4013 in the cache */
   /* search1442_1 -> 4014 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* width */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1442_2_1_0 -> 887 in the cache */
   /* search1442_2_1_1_0_0 -> 19 in the cache */
   /* search1442_2_1_1_0_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 19, 4014 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 4045 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 2,
      { 887, 4046 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 4044, 4047 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 3,
      { 4013, 4014, 4048 },
      -1,
   } },

   /* replace1442_0 -> 4013 in the cache */
   /* replace1442_1 -> 4014 in the cache */
   /* replace1442_2 -> 4044 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 4013, 4014, 4044 },
      -1,
   } },

   /* ('ibfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ibfe', 'value', 'offset', 'width') */
   /* search1443_0 -> 4013 in the cache */
   /* search1443_1 -> 4014 in the cache */
   /* search1443_2_0 -> 4044 in the cache */
   /* search1443_2_1_0 -> 887 in the cache */
   /* search1443_2_1_1_0_0 -> 19 in the cache */
   /* search1443_2_1_1_0_1 -> 4014 in the cache */
   /* search1443_2_1_1_0 -> 4045 in the cache */
   /* search1443_2_1_1 -> 4046 in the cache */
   /* search1443_2_1 -> 4047 in the cache */
   /* search1443_2 -> 4048 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 3,
      { 4013, 4014, 4048 },
      -1,
   } },

   /* replace1443_0 -> 4013 in the cache */
   /* replace1443_1 -> 4014 in the cache */
   /* replace1443_2 -> 4044 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 4013, 4014, 4044 },
      -1,
   } },

   /* ('bfm', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset')))), 'offset') => ('bfm', 'width', 'offset') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* width */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1444_0_1_0 -> 887 in the cache */
   /* search1444_0_1_1_0_0 -> 19 in the cache */
   /* search1444_0_1_1_0_1 -> 4014 in the cache */
   /* search1444_0_1_1_0 -> 4045 in the cache */
   /* search1444_0_1_1 -> 4046 in the cache */
   /* search1444_0_1 -> 4047 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 4053, 4047 },
      -1,
   } },
   /* search1444_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 3,
      { 4054, 4014 },
      -1,
   } },

   /* replace1444_0 -> 4053 in the cache */
   /* replace1444_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 4053, 4014 },
      -1,
   } },

   /* ('iadd@32', ('ishl', 1, 'a'), -1) => ('bfm', 'a', 0) */
   /* search1445_0_0 -> 406 in the cache */
   /* search1445_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 406, 16 },
      -1,
   } },
   /* search1445_1 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4057, 130 },
      -1,
   } },

   /* replace1445_0 -> 16 in the cache */
   /* replace1445_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 16, 128 },
      -1,
   } },

   /* ('ishl', ('bfm', 'a', 0), 'b') => ('bfm', 'a', 'b') */
   /* search1446_0_0 -> 16 in the cache */
   /* search1446_0_1 -> 128 in the cache */
   /* search1446_0 -> 4059 in the cache */
   /* search1446_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4059, 36 },
      -1,
   } },

   /* replace1446_0 -> 16 in the cache */
   /* replace1446_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search1447_0 -> 16 in the cache */
   /* search1447_1 -> 36 in the cache */
   /* search1447_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 128 },
      -1,
   } },

   /* replace1447 -> 128 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search1448_0 -> 16 in the cache */
   /* search1448_1 -> 36 in the cache */
   /* search1448_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 128 },
      -1,
   } },

   /* replace1448 -> 128 in the cache */

   /* ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b'))) */
   /* search1449_0 -> 16 in the cache */
   /* search1449_1 -> 128 in the cache */
   /* search1449_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 128, 33 },
      -1,
   } },

   /* replace1449_0 -> 16 in the cache */
   /* replace1449_1_0 -> 940 in the cache */
   /* replace1449_1_1_0 -> 36 in the cache */
   /* replace1449_1_1 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 940, 2940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 4065 },
      -1,
   } },

   /* ('b2i32', ('ine', ('ubfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1) */
   /* search1450_0_0_0 -> 16 in the cache */
   /* search1450_0_0_1 -> 36 in the cache */
   /* search1450_0_0_2 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 406 },
      -1,
   } },
   /* search1450_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4067, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 1,
      { 4068 },
      -1,
   } },

   /* replace1450_0 -> 16 in the cache */
   /* replace1450_1 -> 36 in the cache */
   /* replace1450_2 -> 406 in the cache */
   /* replace1450 -> 4067 in the cache */

   /* ('b2i32', ('ine', ('ibfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1) */
   /* search1451_0_0_0 -> 16 in the cache */
   /* search1451_0_0_1 -> 36 in the cache */
   /* search1451_0_0_2 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 406 },
      -1,
   } },
   /* search1451_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4070, 128 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 1,
      { 4071 },
      -1,
   } },

   /* replace1451_0 -> 16 in the cache */
   /* replace1451_1 -> 36 in the cache */
   /* replace1451_2 -> 406 in the cache */
   /* replace1451 -> 4067 in the cache */

   /* ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1452_0_0 -> 16 in the cache */
   /* search1452_0_1 -> 33 in the cache */
   /* search1452_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 33, 319 },
      0,
   } },
   /* search1452_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4073, 128 },
      -1,
   } },

   /* replace1452_0_0 -> 16 in the cache */
   /* replace1452_0_1_0_0 -> 940 in the cache */
   /* replace1452_0_1_0_1_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 940, 4075 },
      -1,
   } },
   /* replace1452_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4076, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 4077 },
      -1,
   } },
   /* replace1452_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4078, 128 },
      -1,
   } },

   /* ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1453_0_0 -> 16 in the cache */
   /* search1453_0_1 -> 33 in the cache */
   /* search1453_0_2 -> 319 in the cache */
   /* search1453_0 -> 4073 in the cache */
   /* search1453_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4073, 128 },
      -1,
   } },

   /* replace1453_0_0 -> 16 in the cache */
   /* replace1453_0_1_0_0 -> 940 in the cache */
   /* replace1453_0_1_0_1_0 -> 141 in the cache */
   /* replace1453_0_1_0_1 -> 4075 in the cache */
   /* replace1453_0_1_0 -> 4076 in the cache */
   /* replace1453_0_1_1 -> 36 in the cache */
   /* replace1453_0_1 -> 4077 in the cache */
   /* replace1453_0 -> 4078 in the cache */
   /* replace1453_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4078, 128 },
      -1,
   } },

   /* ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1454_0_0 -> 16 in the cache */
   /* search1454_0_1 -> 33 in the cache */
   /* search1454_0_2 -> 319 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 33, 319 },
      0,
   } },
   /* search1454_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4082, 128 },
      -1,
   } },

   /* replace1454_0_0 -> 16 in the cache */
   /* replace1454_0_1_0_0 -> 940 in the cache */
   /* replace1454_0_1_0_1_0 -> 141 in the cache */
   /* replace1454_0_1_0_1 -> 4075 in the cache */
   /* replace1454_0_1_0 -> 4076 in the cache */
   /* replace1454_0_1_1 -> 36 in the cache */
   /* replace1454_0_1 -> 4077 in the cache */
   /* replace1454_0 -> 4078 in the cache */
   /* replace1454_1 -> 128 in the cache */
   /* replace1454 -> 4079 in the cache */

   /* ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1455_0_0 -> 16 in the cache */
   /* search1455_0_1 -> 33 in the cache */
   /* search1455_0_2 -> 319 in the cache */
   /* search1455_0 -> 4082 in the cache */
   /* search1455_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4082, 128 },
      -1,
   } },

   /* replace1455_0_0 -> 16 in the cache */
   /* replace1455_0_1_0_0 -> 940 in the cache */
   /* replace1455_0_1_0_1_0 -> 141 in the cache */
   /* replace1455_0_1_0_1 -> 4075 in the cache */
   /* replace1455_0_1_0 -> 4076 in the cache */
   /* replace1455_0_1_1 -> 36 in the cache */
   /* replace1455_0_1 -> 4077 in the cache */
   /* replace1455_0 -> 4078 in the cache */
   /* replace1455_1 -> 128 in the cache */
   /* replace1455 -> 4081 in the cache */

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits'))) */
   /* search1456_0 -> 4013 in the cache */
   /* search1456_1 -> 4014 in the cache */
   /* search1456_2 -> 4015 in the cache */
   /* search1456 -> 4016 in the cache */

   /* replace1456_0_0 -> 128 in the cache */
   /* replace1456_0_1 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 128, 4015 },
      -1,
   } },
   /* replace1456_1 -> 128 in the cache */
   /* replace1456_2_0_0 -> 4013 in the cache */
   /* replace1456_2_0_1_0_0 -> 887 in the cache */
   /* replace1456_2_0_1_0_1 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 887, 4015 },
      -1,
   } },
   /* replace1456_2_0_1_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 4086, 4014 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4013, 4087 },
      -1,
   } },
   /* replace1456_2_1_0 -> 887 in the cache */
   /* replace1456_2_1_1 -> 4015 in the cache */
   /* replace1456_2_1 -> 4086 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 4088, 4086 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4085, 128, 4089 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1))) */
   /* search1457_0 -> 4013 in the cache */
   /* search1457_1 -> 4014 in the cache */
   /* search1457_2 -> 4015 in the cache */
   /* search1457 -> 4020 in the cache */

   /* replace1457_0_0 -> 4013 in the cache */
   /* replace1457_0_1 -> 4014 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 4013, 4014 },
      -1,
   } },
   /* replace1457_1_0_0 -> 4015 in the cache */
   /* replace1457_1_0_1 -> 887 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 4015, 887 },
      -1,
   } },
   /* replace1457_1_1 -> 940 in the cache */
   /* replace1457_1_2_0_0 -> 406 in the cache */
   /* replace1457_1_2_0_1 -> 4015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 406, 4015 },
      -1,
   } },
   /* replace1457_1_2_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 4093, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4092, 940, 4094 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4091, 4095 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value')) */
   /* search1458_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 4013 },
      -1,
   } },

   /* replace1458_0_0_0 -> 4013 in the cache */
   /* replace1458_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4013, 128 },
      -1,
   } },
   /* replace1458_0_1_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 4013 },
      -1,
   } },
   /* replace1458_0_2 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4098, 4099, 4013 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 4100 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value')) */
   /* search1459_0 -> 4013 in the cache */
   /* search1459 -> 4097 in the cache */

   /* replace1459_0_0_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 4013 },
      -1,
   } },
   /* replace1459_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4102, 128 },
      -1,
   } },
   /* replace1459_1_0 -> 19 in the cache */
   /* replace1459_1_1_0 -> 4013 in the cache */
   /* replace1459_1_1 -> 4102 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 4102 },
      -1,
   } },
   /* replace1459_2_0 -> 4013 in the cache */
   /* replace1459_2 -> 4102 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4103, 4104, 4102 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('isub', 31, ('uclz', ('ixor', 'value', ('ishr', 'value', 31)))) */
   /* search1460_0 -> 4013 in the cache */
   /* search1460 -> 4097 in the cache */

   /* replace1460_0 -> 19 in the cache */
   /* replace1460_1_0_0 -> 4013 in the cache */
   /* replace1460_1_0_1_0 -> 4013 in the cache */
   /* replace1460_1_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 4013, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4013, 4106 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uclz,
      -1, 1,
      { 4107 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 1,
      { 19, 4108 },
      -1,
   } },

   /* ('ufind_msb', 'value@32') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value')) */
   /* search1461_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 4013 },
      -1,
   } },

   /* replace1461_0_0_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 4013 },
      -1,
   } },
   /* replace1461_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4111, 128 },
      -1,
   } },
   /* replace1461_1_0 -> 19 in the cache */
   /* replace1461_1_1_0 -> 4013 in the cache */
   /* replace1461_1_1 -> 4111 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 4111 },
      -1,
   } },
   /* replace1461_2_0 -> 4013 in the cache */
   /* replace1461_2 -> 4111 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4112, 4113, 4111 },
      -1,
   } },

   /* ('ufind_msb', 'value@32') => ('isub', 31, ('uclz', 'value')) */
   /* search1462_0 -> 4013 in the cache */
   /* search1462 -> 4110 in the cache */

   /* replace1462_0 -> 19 in the cache */
   /* replace1462_1_0 -> 4013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uclz,
      -1, 0,
      { 4013 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 4115 },
      -1,
   } },

   /* ('uclz', 'a') => ('umin', 32, ('ufind_msb_rev', 'a')) */
   /* search1463_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uclz,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1463_0 -> 887 in the cache */
   /* replace1463_1_0 -> 16 in the cache */
   /* replace1463_1 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 887, 3822 },
      -1,
   } },

   /* ('find_lsb', 'value@64') => ('ufind_msb', ('iand', 'value', ('ineg', 'value'))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 4119 },
      -1,
   } },

   /* replace1464_0_0 -> 4119 in the cache */
   /* replace1464_0_1_0 -> 4119 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 4119 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4119, 4121 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 1,
      { 4122 },
      -1,
   } },

   /* ('find_lsb', 'value') => ('ufind_msb', ('u2u32', ('iand', 'value', ('ineg', 'value')))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 4124 },
      -1,
   } },

   /* replace1465_0_0_0 -> 4124 in the cache */
   /* replace1465_0_0_1_0 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 4124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4124, 4126 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 1,
      { 4127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 1,
      { 4128 },
      -1,
   } },

   /* ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24) */
   /* search1466_0 -> 16 in the cache */
   /* search1466_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1466_0_0 -> 16 in the cache */
   /* replace1466_0_1_0_0 -> 415 in the cache */
   /* replace1466_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 415, 36 },
      -1,
   } },
   /* replace1466_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4131, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 4132 },
      -1,
   } },
   /* replace1466_1 -> 1025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 4133, 1025 },
      -1,
   } },

   /* ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255) */
   /* search1467_0 -> 16 in the cache */
   /* search1467_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1467_0_0 -> 16 in the cache */
   /* replace1467_0_1_0 -> 36 in the cache */
   /* replace1467_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 4136 },
      -1,
   } },
   /* replace1467_1 -> 1023 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4137, 1023 },
      -1,
   } },

   /* ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16) */
   /* search1468_0 -> 16 in the cache */
   /* search1468_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1468_0_0 -> 16 in the cache */
   /* replace1468_0_1_0_0 -> 406 in the cache */
   /* replace1468_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 406, 36 },
      -1,
   } },
   /* replace1468_0_1_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4140, 866 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 4141 },
      -1,
   } },
   /* replace1468_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 4142, 866 },
      -1,
   } },

   /* ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535) */
   /* search1469_0 -> 16 in the cache */
   /* search1469_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1469_0_0 -> 16 in the cache */
   /* replace1469_0_1_0 -> 36 in the cache */
   /* replace1469_0_1_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 866 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 4145 },
      -1,
   } },
   /* replace1469_1 -> 1015 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4146, 1015 },
      -1,
   } },

   /* ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0)))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* v */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16,
      -1, 0,
      { 4148 },
      -1,
   } },

   /* replace1470_0_0_0_0_0 -> 4148 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 4148 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x40efffe000000000 /* 65535.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 4150, 4151 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 4152 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 4153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 1,
      { 4154 },
      -1,
   } },

   /* ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0)))) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* v */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1471_0_0_0_0_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 4156 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x406fe00000000000 /* 255.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 4158, 4159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 4160 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 4161 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 1,
      { 4162 },
      -1,
   } },

   /* ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0)))) */
   /* search1472_0 -> 4148 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16,
      -1, 0,
      { 4148 },
      -1,
   } },

   /* replace1472_0_0_0_0_0 -> 252 in the cache */
   /* replace1472_0_0_0_0_1_0 -> 258 in the cache */
   /* replace1472_0_0_0_0_1_1 -> 4148 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 258, 4148 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 252, 4165 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x40dfffc000000000 /* 32767.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 4166, 4167 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 4168 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 4169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 3,
      { 4170 },
      -1,
   } },

   /* ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0)))) */
   /* search1473_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1473_0_0_0_0_0 -> 254 in the cache */
   /* replace1473_0_0_0_0_1_0 -> 288 in the cache */
   /* replace1473_0_0_0_0_1_1 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 288, 4156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 254, 4173 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x405fc00000000000 /* 127.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 4174, 4175 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 4176 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 4177 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 3,
      { 4178 },
      -1,
   } },

   /* ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0) */
   /* search1474_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1474_0_0_0_0 -> 4156 in the cache */
   /* replace1474_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 4156, 128 },
      -1,
   } },
   /* replace1474_0_0_1_0 -> 4156 in the cache */
   /* replace1474_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 4156, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 4181, 4182 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 4183 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40efffe000000000 /* 65535.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 4184, 4185 },
      -1,
   } },

   /* ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0) */
   /* search1475_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1475_0_0_0_0 -> 4156 in the cache */
   /* replace1475_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 4156, 128 },
      -1,
   } },
   /* replace1475_0_0_1_0 -> 4156 in the cache */
   /* replace1475_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 4156, 406 },
      -1,
   } },
   /* replace1475_0_0_2_0 -> 4156 in the cache */
   /* replace1475_0_0_2_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 4156, 411 },
      -1,
   } },
   /* replace1475_0_0_3_0 -> 4156 in the cache */
   /* replace1475_0_0_3_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 4156, 415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 4188, 4189, 4190, 4191 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 4192 },
      -1,
   } },
   /* replace1475_1 -> 4159 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 4193, 4159 },
      -1,
   } },

   /* ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0))) */
   /* search1476_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1476_0 -> 254 in the cache */
   /* replace1476_1_0 -> 288 in the cache */
   /* replace1476_1_1_0_0_0_0 -> 4156 in the cache */
   /* replace1476_1_1_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 4156, 128 },
      -1,
   } },
   /* replace1476_1_1_0_0_1_0 -> 4156 in the cache */
   /* replace1476_1_1_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 4156, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 4196, 4197 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 4198 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40dfffc000000000 /* 32767.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 4199, 4200 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 288, 4201 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 254, 4202 },
      -1,
   } },

   /* ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0))) */
   /* search1477_0 -> 4156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8,
      -1, 0,
      { 4156 },
      -1,
   } },

   /* replace1477_0 -> 254 in the cache */
   /* replace1477_1_0 -> 288 in the cache */
   /* replace1477_1_1_0_0_0_0 -> 4156 in the cache */
   /* replace1477_1_1_0_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 4156, 128 },
      -1,
   } },
   /* replace1477_1_1_0_0_1_0 -> 4156 in the cache */
   /* replace1477_1_1_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 4156, 406 },
      -1,
   } },
   /* replace1477_1_1_0_0_2_0 -> 4156 in the cache */
   /* replace1477_1_1_0_0_2_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 4156, 411 },
      -1,
   } },
   /* replace1477_1_1_0_0_3_0 -> 4156 in the cache */
   /* replace1477_1_1_0_0_3_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 4156, 415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 4205, 4206, 4207, 4208 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 4209 },
      -1,
   } },
   /* replace1477_1_1_1 -> 4175 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 4210, 4175 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 288, 4211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 254, 4212 },
      -1,
   } },

   /* ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a'))) */
   /* search1478_0 -> 16 in the cache */
   /* search1478_1 -> 36 in the cache */
   /* search1478 -> 3500 in the cache */

   /* replace1478_0_0_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4214 },
      -1,
   } },
   /* replace1478_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4215, 866 },
      -1,
   } },
   /* replace1478_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4217 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4216, 4218 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a')) */
   /* search1479_0 -> 16 in the cache */
   /* search1479 -> 3508 in the cache */

   /* replace1479_0_0 -> 16 in the cache */
   /* replace1479_0 -> 3505 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3505 },
      -1,
   } },

   /* ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16))) */
   /* search1480_0 -> 16 in the cache */
   /* search1480 -> 3515 in the cache */

   /* replace1480_0_0_0 -> 16 in the cache */
   /* replace1480_0_0_1 -> 866 in the cache */
   /* replace1480_0_0 -> 1021 in the cache */
   /* replace1480_0 -> 3506 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3506 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a')) */
   /* search1481_0 -> 11 in the cache */
   /* search1481_1 -> 268 in the cache */
   /* search1481 -> 3413 in the cache */

   /* replace1481_0_0_0 -> 268 in the cache */
   /* replace1481_0_0 -> 3497 in the cache */
   /* replace1481_0_1 -> 866 in the cache */
   /* replace1481_0 -> 3498 in the cache */
   /* replace1481_1_0 -> 11 in the cache */
   /* replace1481_1 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3498, 2129 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a') */
   /* search1482_0 -> 16 in the cache */
   /* search1482 -> 3421 in the cache */

   /* replace1482_0 -> 16 in the cache */
   /* replace1482 -> 3505 in the cache */

   /* ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1483_0 -> 16 in the cache */
   /* search1483 -> 3423 in the cache */

   /* replace1483_0_0 -> 16 in the cache */
   /* replace1483_0_1 -> 866 in the cache */
   /* replace1483_0 -> 1021 in the cache */
   /* replace1483 -> 3506 in the cache */

   /* ('isign', 'a') => ('imin', ('imax', 'a', -1), 1) */
   /* search1484_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1484_0_0 -> 32 in the cache */
   /* replace1484_0_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 63 },
      -1,
   } },
   /* replace1484_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 4224, 35 },
      -1,
   } },

   /* ('imin', ('imax', 'a', -1), 1) => ('isign', 'a') */
   /* search1485_0_0 -> 32 in the cache */
   /* search1485_0_1 -> 63 in the cache */
   /* search1485_0 -> 4224 in the cache */
   /* search1485_1 -> 35 in the cache */
   /* search1485 -> 4225 in the cache */

   /* replace1485_0 -> 32 in the cache */
   /* replace1485 -> 4223 in the cache */

   /* ('imax', ('imin', 'a', 1), -1) => ('isign', 'a') */
   /* search1486_0_0 -> 32 in the cache */
   /* search1486_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 32, 35 },
      -1,
   } },
   /* search1486_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 4226, 63 },
      -1,
   } },

   /* replace1486_0 -> 32 in the cache */
   /* replace1486 -> 4223 in the cache */

   /* ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0))) */
   /* search1487_0 -> 32 in the cache */
   /* search1487 -> 98 in the cache */

   /* replace1487_0_0_0 -> 117 in the cache */
   /* replace1487_0_0_1 -> 32 in the cache */
   /* replace1487_0_0 -> 1539 in the cache */
   /* replace1487_0 -> 1540 in the cache */
   /* replace1487_1_0_0 -> 32 in the cache */
   /* replace1487_1_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 4228 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 1540, 4229 },
      -1,
   } },

   /* ('amul', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1488_0 -> 0 in the cache */
   /* search1488_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_amul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1488_0 -> 0 in the cache */
   /* replace1488_1 -> 3 in the cache */
   /* replace1488 -> 964 in the cache */

   /* ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)) */
   /* search1489_0 -> 16 in the cache */
   /* search1489_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1489_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffff /* 16777215 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 4233 },
      -1,
   } },
   /* replace1489_1_0 -> 36 in the cache */
   /* replace1489_1_1 -> 4233 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 4233 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 4234, 4235 },
      -1,
   } },

   /* ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c') */
   /* search1490_0 -> 16 in the cache */
   /* search1490_1 -> 36 in the cache */
   /* search1490_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace1490_0_0_0 -> 16 in the cache */
   /* replace1490_0_0_1 -> 4233 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 16, 4233 },
      -1,
   } },
   /* replace1490_0_1_0 -> 36 in the cache */
   /* replace1490_0_1_1 -> 4233 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 36, 4233 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 3,
      { 4238, 4239 },
      -1,
   } },
   /* replace1490_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 4240, 141 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b') */
   /* search1491_0 -> 16 in the cache */
   /* search1491_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1491_0 -> 16 in the cache */
   /* replace1491_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1492_0 -> 16 in the cache */
   /* search1492_1 -> 36 in the cache */
   /* search1492 -> 4242 in the cache */

   /* replace1492_0 -> 16 in the cache */
   /* replace1492_1 -> 36 in the cache */
   /* replace1492 -> 49 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c') */
   /* search1493_0 -> 16 in the cache */
   /* search1493_1 -> 36 in the cache */
   /* search1493_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24_relaxed,
      0, 1,
      { 16, 36, 141 },
      -1,
   } },

   /* replace1493_0 -> 16 in the cache */
   /* replace1493_1 -> 36 in the cache */
   /* replace1493_2 -> 141 in the cache */
   /* replace1493 -> 4237 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c') */
   /* search1494_0 -> 16 in the cache */
   /* search1494_1 -> 36 in the cache */
   /* search1494_2 -> 141 in the cache */
   /* search1494 -> 4244 in the cache */

   /* replace1494_0_0 -> 16 in the cache */
   /* replace1494_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1494_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 4245, 141 },
      -1,
   } },

   /* ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b') */
   /* search1495_0 -> 16 in the cache */
   /* search1495_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1495_0 -> 16 in the cache */
   /* replace1495_1 -> 36 in the cache */
   /* replace1495 -> 4232 in the cache */

   /* ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1496_0 -> 16 in the cache */
   /* search1496_1 -> 36 in the cache */
   /* search1496 -> 4247 in the cache */

   /* replace1496_0 -> 16 in the cache */
   /* replace1496_1 -> 36 in the cache */
   /* replace1496 -> 49 in the cache */

   /* ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b') */
   /* search1497_0 -> 16 in the cache */
   /* search1497_1 -> 36 in the cache */
   /* search1497_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 36, 128 },
      -1,
   } },

   /* replace1497_0 -> 16 in the cache */
   /* replace1497_1 -> 36 in the cache */
   /* replace1497 -> 4243 in the cache */

   /* ('imad24_ir3', 'a', 0, 'c') => c */
   /* search1498_0 -> 16 in the cache */
   /* search1498_1 -> 128 in the cache */
   /* search1498_2 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 128, 565 },
      -1,
   } },

   /* replace1498 -> 565 in the cache */

   /* ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c') */
   /* search1499_0 -> 16 in the cache */
   /* search1499_1 -> 406 in the cache */
   /* search1499_2 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 406, 565 },
      -1,
   } },

   /* replace1499_0 -> 16 in the cache */
   /* replace1499_1 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 565 },
      -1,
   } },

   /* ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c') */
   /* search1500_0 -> 296 in the cache */
   /* search1500_1 -> 33 in the cache */
   /* search1500_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 296, 33, 141 },
      -1,
   } },

   /* replace1500_0_0 -> 16 in the cache */
   /* replace1500_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1500_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 4253, 141 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   /* search1501_0 -> 16 in the cache */
   /* search1501_1 -> 2422 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 2422 },
      -1,
   } },

   /* replace1501_0 -> 16 in the cache */
   /* replace1501_1_0 -> 36 in the cache */
   /* replace1501_1 -> 2428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2428 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search1502_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 4257 },
      -1,
   } },

   /* replace1502_0_0 -> 16 in the cache */
   /* replace1502_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 4259 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 4260 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 4261 },
      -1,
   } },

   /* ('imul24', 'a', 0) => 0 */
   /* search1503_0 -> 16 in the cache */
   /* search1503_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 128 },
      -1,
   } },

   /* replace1503 -> 128 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1504_0_0_0 -> 6 in the cache */
   /* search1504_0_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6, 2280 },
      -1,
   } },
   /* search1504_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4264, 6 },
      -1,
   } },
   /* search1504_1 -> 1448 in the cache */
   /* search1504_2_0 -> 6 in the cache */
   /* search1504_2_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 6, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4265, 1448, 4266 },
      -1,
   } },

   /* replace1504_0 -> 6 in the cache */
   /* replace1504_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 6, 2280 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1505_0_0_0 -> 6 in the cache */
   /* search1505_0_0_1 -> 2280 in the cache */
   /* search1505_0_0 -> 4264 in the cache */
   /* search1505_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4264, 6 },
      -1,
   } },
   /* search1505_1_0 -> 6 in the cache */
   /* search1505_1_1 -> 2280 in the cache */
   /* search1505_1 -> 4266 in the cache */
   /* search1505_2 -> 1448 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4269, 4266, 1448 },
      -1,
   } },

   /* replace1505_0 -> 6 in the cache */
   /* replace1505_1 -> 2280 in the cache */
   /* replace1505 -> 4268 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1506_0_0_0 -> 6 in the cache */
   /* search1506_0_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 6, 2280 },
      -1,
   } },
   /* search1506_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4271, 848 },
      -1,
   } },
   /* search1506_1_0 -> 6 in the cache */
   /* search1506_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 6, 2280 },
      -1,
   } },
   /* search1506_2 -> 1448 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4272, 4273, 1448 },
      -1,
   } },

   /* replace1506_0 -> 6 in the cache */
   /* replace1506_1 -> 2280 in the cache */
   /* replace1506 -> 4268 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1507_0_0_0 -> 6 in the cache */
   /* search1507_0_0_1 -> 2280 in the cache */
   /* search1507_0_0 -> 4271 in the cache */
   /* search1507_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4271, 848 },
      -1,
   } },
   /* search1507_1 -> 1448 in the cache */
   /* search1507_2_0 -> 6 in the cache */
   /* search1507_2_1 -> 2280 in the cache */
   /* search1507_2 -> 4273 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4275, 1448, 4273 },
      -1,
   } },

   /* replace1507_0 -> 6 in the cache */
   /* replace1507_1 -> 2280 in the cache */
   /* replace1507 -> 4268 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1508_0_0_0 -> 11 in the cache */
   /* search1508_0_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 11, 268 },
      -1,
   } },
   /* search1508_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4277, 11 },
      -1,
   } },
   /* search1508_1 -> 1470 in the cache */
   /* search1508_2_0 -> 11 in the cache */
   /* search1508_2_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 11, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4278, 1470, 4279 },
      -1,
   } },

   /* replace1508_0 -> 11 in the cache */
   /* replace1508_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 11, 268 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1509_0_0_0 -> 11 in the cache */
   /* search1509_0_0_1 -> 268 in the cache */
   /* search1509_0_0 -> 4277 in the cache */
   /* search1509_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4277, 11 },
      -1,
   } },
   /* search1509_1_0 -> 11 in the cache */
   /* search1509_1_1 -> 268 in the cache */
   /* search1509_1 -> 4279 in the cache */
   /* search1509_2 -> 1470 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4282, 4279, 1470 },
      -1,
   } },

   /* replace1509_0 -> 11 in the cache */
   /* replace1509_1 -> 268 in the cache */
   /* replace1509 -> 4281 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1510_0_0_0 -> 11 in the cache */
   /* search1510_0_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 11, 268 },
      -1,
   } },
   /* search1510_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4284, 869 },
      -1,
   } },
   /* search1510_1_0 -> 11 in the cache */
   /* search1510_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 11, 268 },
      -1,
   } },
   /* search1510_2 -> 1470 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4285, 4286, 1470 },
      -1,
   } },

   /* replace1510_0 -> 11 in the cache */
   /* replace1510_1 -> 268 in the cache */
   /* replace1510 -> 4281 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1511_0_0_0 -> 11 in the cache */
   /* search1511_0_0_1 -> 268 in the cache */
   /* search1511_0_0 -> 4284 in the cache */
   /* search1511_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4284, 869 },
      -1,
   } },
   /* search1511_1 -> 1470 in the cache */
   /* search1511_2_0 -> 11 in the cache */
   /* search1511_2_1 -> 268 in the cache */
   /* search1511_2 -> 4286 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4288, 1470, 4286 },
      -1,
   } },

   /* replace1511_0 -> 11 in the cache */
   /* replace1511_1 -> 268 in the cache */
   /* replace1511 -> 4281 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1512_0_0_0 -> 16 in the cache */
   /* search1512_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* search1512_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4290, 16 },
      -1,
   } },
   /* search1512_1 -> 130 in the cache */
   /* search1512_2_0 -> 16 in the cache */
   /* search1512_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4291, 130, 4292 },
      -1,
   } },

   /* replace1512_0 -> 16 in the cache */
   /* replace1512_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1513_0_0_0 -> 16 in the cache */
   /* search1513_0_0_1 -> 36 in the cache */
   /* search1513_0_0 -> 4290 in the cache */
   /* search1513_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4290, 16 },
      -1,
   } },
   /* search1513_1_0 -> 16 in the cache */
   /* search1513_1_1 -> 36 in the cache */
   /* search1513_1 -> 4292 in the cache */
   /* search1513_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4295, 4292, 130 },
      -1,
   } },

   /* replace1513_0 -> 16 in the cache */
   /* replace1513_1 -> 36 in the cache */
   /* replace1513 -> 4294 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1514_0_0_0 -> 16 in the cache */
   /* search1514_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* search1514_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4297, 128 },
      -1,
   } },
   /* search1514_1_0 -> 16 in the cache */
   /* search1514_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 16, 36 },
      -1,
   } },
   /* search1514_2 -> 130 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4298, 4299, 130 },
      -1,
   } },

   /* replace1514_0 -> 16 in the cache */
   /* replace1514_1 -> 36 in the cache */
   /* replace1514 -> 4294 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1515_0_0_0 -> 16 in the cache */
   /* search1515_0_0_1 -> 36 in the cache */
   /* search1515_0_0 -> 4297 in the cache */
   /* search1515_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4297, 128 },
      -1,
   } },
   /* search1515_1 -> 130 in the cache */
   /* search1515_2_0 -> 16 in the cache */
   /* search1515_2_1 -> 36 in the cache */
   /* search1515_2 -> 4299 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4301, 130, 4299 },
      -1,
   } },

   /* replace1515_0 -> 16 in the cache */
   /* replace1515_1 -> 36 in the cache */
   /* replace1515 -> 4294 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1516_0_0_0 -> 21 in the cache */
   /* search1516_0_0_1 -> 41 in the cache */
   /* search1516_0_0 -> 3946 in the cache */
   /* search1516_0_1 -> 21 in the cache */
   /* search1516_0 -> 3947 in the cache */
   /* search1516_1 -> 1512 in the cache */
   /* search1516_2_0 -> 21 in the cache */
   /* search1516_2_1 -> 41 in the cache */
   /* search1516_2 -> 3948 in the cache */
   /* search1516 -> 3949 in the cache */

   /* replace1516_0 -> 21 in the cache */
   /* replace1516_1 -> 41 in the cache */
   /* replace1516 -> 3945 in the cache */

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1517_0_0_0 -> 21 in the cache */
   /* search1517_0_0_1 -> 41 in the cache */
   /* search1517_0_0 -> 3946 in the cache */
   /* search1517_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 3946, 21 },
      -1,
   } },
   /* search1517_1_0 -> 21 in the cache */
   /* search1517_1_1 -> 41 in the cache */
   /* search1517_1 -> 3948 in the cache */
   /* search1517_2 -> 1512 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4303, 3948, 1512 },
      -1,
   } },

   /* replace1517_0 -> 21 in the cache */
   /* replace1517_1 -> 41 in the cache */
   /* replace1517 -> 3945 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1518_0_0_0 -> 21 in the cache */
   /* search1518_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* search1518_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4305, 910 },
      -1,
   } },
   /* search1518_1_0 -> 21 in the cache */
   /* search1518_1_1 -> 41 in the cache */
   /* search1518_1 -> 3962 in the cache */
   /* search1518_2 -> 1512 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4306, 3962, 1512 },
      -1,
   } },

   /* replace1518_0 -> 21 in the cache */
   /* replace1518_1 -> 41 in the cache */
   /* replace1518 -> 3945 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1519_0_0_0 -> 21 in the cache */
   /* search1519_0_0_1 -> 41 in the cache */
   /* search1519_0_0 -> 4305 in the cache */
   /* search1519_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4305, 910 },
      -1,
   } },
   /* search1519_1 -> 1512 in the cache */
   /* search1519_2_0 -> 21 in the cache */
   /* search1519_2_1 -> 41 in the cache */
   /* search1519_2 -> 3962 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4308, 1512, 3962 },
      -1,
   } },

   /* replace1519_0 -> 21 in the cache */
   /* replace1519_1 -> 41 in the cache */
   /* replace1519 -> 3945 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1520_0_0 -> 6 in the cache */
   /* search1520_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2280 },
      -1,
   } },
   /* search1520_1 -> 848 in the cache */
   /* search1520_2_0 -> 6 in the cache */
   /* search1520_2_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6, 4311 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4310, 848, 4312 },
      -1,
   } },

   /* replace1520_0 -> 6 in the cache */
   /* replace1520_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 6, 2280 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1521_0_0 -> 6 in the cache */
   /* search1521_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2280 },
      -1,
   } },
   /* search1521_1_0 -> 6 in the cache */
   /* search1521_1_1_0 -> 2280 in the cache */
   /* search1521_1_1 -> 4311 in the cache */
   /* search1521_1 -> 4312 in the cache */
   /* search1521_2 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4315, 4312, 848 },
      -1,
   } },

   /* replace1521_0 -> 6 in the cache */
   /* replace1521_1 -> 2280 in the cache */
   /* replace1521 -> 4314 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1522_0_0_0 -> 6 in the cache */
   /* search1522_0_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 6, 2280 },
      -1,
   } },
   /* search1522_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4317, 848 },
      -1,
   } },
   /* search1522_1_0 -> 6 in the cache */
   /* search1522_1_1_0 -> 2280 in the cache */
   /* search1522_1_1 -> 4311 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 6, 4311 },
      -1,
   } },
   /* search1522_2 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4318, 4319, 848 },
      -1,
   } },

   /* replace1522_0 -> 6 in the cache */
   /* replace1522_1 -> 2280 in the cache */
   /* replace1522 -> 4314 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1523_0_0_0 -> 6 in the cache */
   /* search1523_0_0_1 -> 2280 in the cache */
   /* search1523_0_0 -> 4317 in the cache */
   /* search1523_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4317, 848 },
      -1,
   } },
   /* search1523_1 -> 848 in the cache */
   /* search1523_2_0 -> 6 in the cache */
   /* search1523_2_1_0 -> 2280 in the cache */
   /* search1523_2_1 -> 4311 in the cache */
   /* search1523_2 -> 4319 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4321, 848, 4319 },
      -1,
   } },

   /* replace1523_0 -> 6 in the cache */
   /* replace1523_1 -> 2280 in the cache */
   /* replace1523 -> 4314 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1524_0_0 -> 11 in the cache */
   /* search1524_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 268 },
      -1,
   } },
   /* search1524_1 -> 869 in the cache */
   /* search1524_2_0 -> 11 in the cache */
   /* search1524_2_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 11, 4324 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4323, 869, 4325 },
      -1,
   } },

   /* replace1524_0 -> 11 in the cache */
   /* replace1524_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 11, 268 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1525_0_0 -> 11 in the cache */
   /* search1525_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 268 },
      -1,
   } },
   /* search1525_1_0 -> 11 in the cache */
   /* search1525_1_1_0 -> 268 in the cache */
   /* search1525_1_1 -> 4324 in the cache */
   /* search1525_1 -> 4325 in the cache */
   /* search1525_2 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4328, 4325, 869 },
      -1,
   } },

   /* replace1525_0 -> 11 in the cache */
   /* replace1525_1 -> 268 in the cache */
   /* replace1525 -> 4327 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1526_0_0_0 -> 11 in the cache */
   /* search1526_0_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 11, 268 },
      -1,
   } },
   /* search1526_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4330, 869 },
      -1,
   } },
   /* search1526_1_0 -> 11 in the cache */
   /* search1526_1_1_0 -> 268 in the cache */
   /* search1526_1_1 -> 4324 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 11, 4324 },
      -1,
   } },
   /* search1526_2 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4331, 4332, 869 },
      -1,
   } },

   /* replace1526_0 -> 11 in the cache */
   /* replace1526_1 -> 268 in the cache */
   /* replace1526 -> 4327 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1527_0_0_0 -> 11 in the cache */
   /* search1527_0_0_1 -> 268 in the cache */
   /* search1527_0_0 -> 4330 in the cache */
   /* search1527_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4330, 869 },
      -1,
   } },
   /* search1527_1 -> 869 in the cache */
   /* search1527_2_0 -> 11 in the cache */
   /* search1527_2_1_0 -> 268 in the cache */
   /* search1527_2_1 -> 4324 in the cache */
   /* search1527_2 -> 4332 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4334, 869, 4332 },
      -1,
   } },

   /* replace1527_0 -> 11 in the cache */
   /* replace1527_1 -> 268 in the cache */
   /* replace1527 -> 4327 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1528_0_0 -> 16 in the cache */
   /* search1528_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1528_1 -> 128 in the cache */
   /* search1528_2_0 -> 16 in the cache */
   /* search1528_2_1_0 -> 36 in the cache */
   /* search1528_2_1 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 2940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4336, 128, 4337 },
      -1,
   } },

   /* replace1528_0 -> 16 in the cache */
   /* replace1528_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1529_0_0 -> 16 in the cache */
   /* search1529_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1529_1_0 -> 16 in the cache */
   /* search1529_1_1_0 -> 36 in the cache */
   /* search1529_1_1 -> 2940 in the cache */
   /* search1529_1 -> 4337 in the cache */
   /* search1529_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4340, 4337, 128 },
      -1,
   } },

   /* replace1529_0 -> 16 in the cache */
   /* replace1529_1 -> 36 in the cache */
   /* replace1529 -> 4339 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1530_0_0_0 -> 16 in the cache */
   /* search1530_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1530_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4342, 128 },
      -1,
   } },
   /* search1530_1_0 -> 16 in the cache */
   /* search1530_1_1_0 -> 36 in the cache */
   /* search1530_1_1 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 2940 },
      -1,
   } },
   /* search1530_2 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4343, 4344, 128 },
      -1,
   } },

   /* replace1530_0 -> 16 in the cache */
   /* replace1530_1 -> 36 in the cache */
   /* replace1530 -> 4339 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1531_0_0_0 -> 16 in the cache */
   /* search1531_0_0_1 -> 36 in the cache */
   /* search1531_0_0 -> 4342 in the cache */
   /* search1531_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4342, 128 },
      -1,
   } },
   /* search1531_1 -> 128 in the cache */
   /* search1531_2_0 -> 16 in the cache */
   /* search1531_2_1_0 -> 36 in the cache */
   /* search1531_2_1 -> 2940 in the cache */
   /* search1531_2 -> 4344 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4346, 128, 4344 },
      -1,
   } },

   /* replace1531_0 -> 16 in the cache */
   /* replace1531_1 -> 36 in the cache */
   /* replace1531 -> 4339 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1532_0_0 -> 21 in the cache */
   /* search1532_0_1 -> 41 in the cache */
   /* search1532_0 -> 3956 in the cache */
   /* search1532_1 -> 910 in the cache */
   /* search1532_2_0 -> 21 in the cache */
   /* search1532_2_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 21, 4348 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3956, 910, 4349 },
      -1,
   } },

   /* replace1532_0 -> 21 in the cache */
   /* replace1532_1 -> 41 in the cache */
   /* replace1532 -> 3955 in the cache */

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1533_0_0 -> 21 in the cache */
   /* search1533_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* search1533_1_0 -> 21 in the cache */
   /* search1533_1_1_0 -> 41 in the cache */
   /* search1533_1_1 -> 4348 in the cache */
   /* search1533_1 -> 4349 in the cache */
   /* search1533_2 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4351, 4349, 910 },
      -1,
   } },

   /* replace1533_0 -> 21 in the cache */
   /* replace1533_1 -> 41 in the cache */
   /* replace1533 -> 3955 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1534_0_0_0 -> 21 in the cache */
   /* search1534_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* search1534_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4353, 910 },
      -1,
   } },
   /* search1534_1_0 -> 21 in the cache */
   /* search1534_1_1_0 -> 41 in the cache */
   /* search1534_1_1 -> 4348 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 21, 4348 },
      -1,
   } },
   /* search1534_2 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4354, 4355, 910 },
      -1,
   } },

   /* replace1534_0 -> 21 in the cache */
   /* replace1534_1 -> 41 in the cache */
   /* replace1534 -> 3955 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1535_0_0_0 -> 21 in the cache */
   /* search1535_0_0_1 -> 41 in the cache */
   /* search1535_0_0 -> 4353 in the cache */
   /* search1535_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4353, 910 },
      -1,
   } },
   /* search1535_1 -> 910 in the cache */
   /* search1535_2_0 -> 21 in the cache */
   /* search1535_2_1_0 -> 41 in the cache */
   /* search1535_2_1 -> 4348 in the cache */
   /* search1535_2 -> 4355 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4357, 910, 4355 },
      -1,
   } },

   /* replace1535_0 -> 21 in the cache */
   /* replace1535_1 -> 41 in the cache */
   /* replace1535 -> 3955 in the cache */

   /* ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b'))) */
   /* search1536_0 -> 6 in the cache */
   /* search1536_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 6, 2280 },
      -1,
   } },

   /* replace1536_0_0 -> 2280 in the cache */
   /* replace1536_0_1 -> 1443 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2280, 1443 },
      -1,
   } },
   /* replace1536_1_0_0_0 -> 6 in the cache */
   /* replace1536_1_0_0_1 -> 2280 in the cache */
   /* replace1536_1_0_0 -> 4264 in the cache */
   /* replace1536_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4264, 6 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1536_1_2_0 -> 6 in the cache */
   /* replace1536_1_2_1 -> 2280 in the cache */
   /* replace1536_1_2 -> 4266 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4361, 4362, 4266 },
      -1,
   } },
   /* replace1536_2_0_0 -> 6 in the cache */
   /* replace1536_2_0_1_0 -> 6 in the cache */
   /* replace1536_2_0_1_1 -> 2280 in the cache */
   /* replace1536_2_0_1 -> 4273 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 6, 4273 },
      -1,
   } },
   /* replace1536_2_1 -> 7 in the cache */
   /* replace1536_2_2_0 -> 6 in the cache */
   /* replace1536_2_2_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 6, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4364, 7, 4365 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4360, 4363, 4366 },
      -1,
   } },

   /* ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b'))) */
   /* search1537_0 -> 6 in the cache */
   /* search1537_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 6, 2280 },
      -1,
   } },

   /* replace1537_0_0 -> 2280 in the cache */
   /* replace1537_0_1 -> 848 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2280, 848 },
      -1,
   } },
   /* replace1537_1_0_0_0 -> 6 in the cache */
   /* replace1537_1_0_0_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 6, 2280 },
      -1,
   } },
   /* replace1537_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4370, 6 },
      -1,
   } },
   /* replace1537_1_1 -> 4362 in the cache */
   /* replace1537_1_2_0 -> 6 in the cache */
   /* replace1537_1_2_1 -> 2280 in the cache */
   /* replace1537_1_2 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4371, 4362, 4370 },
      -1,
   } },
   /* replace1537_2_0_0 -> 6 in the cache */
   /* replace1537_2_0_1_0 -> 6 in the cache */
   /* replace1537_2_0_1_1 -> 2280 in the cache */
   /* replace1537_2_0_1 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4370 },
      -1,
   } },
   /* replace1537_2_1 -> 7 in the cache */
   /* replace1537_2_2_0 -> 6 in the cache */
   /* replace1537_2_2_1 -> 2280 in the cache */
   /* replace1537_2_2 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4373, 7, 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4369, 4372, 4374 },
      -1,
   } },

   /* ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b'))) */
   /* search1538_0 -> 11 in the cache */
   /* search1538_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 11, 268 },
      -1,
   } },

   /* replace1538_0_0 -> 268 in the cache */
   /* replace1538_0_1 -> 1465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 268, 1465 },
      -1,
   } },
   /* replace1538_1_0_0_0 -> 11 in the cache */
   /* replace1538_1_0_0_1 -> 268 in the cache */
   /* replace1538_1_0_0 -> 4277 in the cache */
   /* replace1538_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4277, 11 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1538_1_2_0 -> 11 in the cache */
   /* replace1538_1_2_1 -> 268 in the cache */
   /* replace1538_1_2 -> 4279 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4378, 4379, 4279 },
      -1,
   } },
   /* replace1538_2_0_0 -> 11 in the cache */
   /* replace1538_2_0_1_0 -> 11 in the cache */
   /* replace1538_2_0_1_1 -> 268 in the cache */
   /* replace1538_2_0_1 -> 4286 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 11, 4286 },
      -1,
   } },
   /* replace1538_2_1 -> 12 in the cache */
   /* replace1538_2_2_0 -> 11 in the cache */
   /* replace1538_2_2_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 11, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4381, 12, 4382 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4377, 4380, 4383 },
      -1,
   } },

   /* ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b'))) */
   /* search1539_0 -> 11 in the cache */
   /* search1539_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 11, 268 },
      -1,
   } },

   /* replace1539_0_0 -> 268 in the cache */
   /* replace1539_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 268, 869 },
      -1,
   } },
   /* replace1539_1_0_0_0 -> 11 in the cache */
   /* replace1539_1_0_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 11, 268 },
      -1,
   } },
   /* replace1539_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4387, 11 },
      -1,
   } },
   /* replace1539_1_1 -> 4379 in the cache */
   /* replace1539_1_2_0 -> 11 in the cache */
   /* replace1539_1_2_1 -> 268 in the cache */
   /* replace1539_1_2 -> 4387 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4388, 4379, 4387 },
      -1,
   } },
   /* replace1539_2_0_0 -> 11 in the cache */
   /* replace1539_2_0_1_0 -> 11 in the cache */
   /* replace1539_2_0_1_1 -> 268 in the cache */
   /* replace1539_2_0_1 -> 4387 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4387 },
      -1,
   } },
   /* replace1539_2_1 -> 12 in the cache */
   /* replace1539_2_2_0 -> 11 in the cache */
   /* replace1539_2_2_1 -> 268 in the cache */
   /* replace1539_2_2 -> 4387 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4390, 12, 4387 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4386, 4389, 4391 },
      -1,
   } },

   /* ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b'))) */
   /* search1540_0 -> 16 in the cache */
   /* search1540_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1540_0_0 -> 36 in the cache */
   /* replace1540_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 36, 406 },
      -1,
   } },
   /* replace1540_1_0_0_0 -> 16 in the cache */
   /* replace1540_1_0_0_1 -> 36 in the cache */
   /* replace1540_1_0_0 -> 4290 in the cache */
   /* replace1540_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4290, 16 },
      -1,
   } },
   /* replace1540_1_1 -> 1001 in the cache */
   /* replace1540_1_2_0 -> 16 in the cache */
   /* replace1540_1_2_1 -> 36 in the cache */
   /* replace1540_1_2 -> 4292 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4395, 1001, 4292 },
      -1,
   } },
   /* replace1540_2_0_0 -> 16 in the cache */
   /* replace1540_2_0_1_0 -> 16 in the cache */
   /* replace1540_2_0_1_1 -> 36 in the cache */
   /* replace1540_2_0_1 -> 4299 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 16, 4299 },
      -1,
   } },
   /* replace1540_2_1 -> 17 in the cache */
   /* replace1540_2_2_0 -> 16 in the cache */
   /* replace1540_2_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4397, 17, 4398 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4394, 4396, 4399 },
      -1,
   } },

   /* ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b'))) */
   /* search1541_0 -> 16 in the cache */
   /* search1541_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1541_0_0 -> 36 in the cache */
   /* replace1541_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 36, 128 },
      -1,
   } },
   /* replace1541_1_0_0_0 -> 16 in the cache */
   /* replace1541_1_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* replace1541_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4403, 16 },
      -1,
   } },
   /* replace1541_1_1 -> 1001 in the cache */
   /* replace1541_1_2_0 -> 16 in the cache */
   /* replace1541_1_2_1 -> 36 in the cache */
   /* replace1541_1_2 -> 4403 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4404, 1001, 4403 },
      -1,
   } },
   /* replace1541_2_0_0 -> 16 in the cache */
   /* replace1541_2_0_1_0 -> 16 in the cache */
   /* replace1541_2_0_1_1 -> 36 in the cache */
   /* replace1541_2_0_1 -> 4403 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4403 },
      -1,
   } },
   /* replace1541_2_1 -> 17 in the cache */
   /* replace1541_2_2_0 -> 16 in the cache */
   /* replace1541_2_2_1 -> 36 in the cache */
   /* replace1541_2_2 -> 4403 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4406, 17, 4403 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4402, 4405, 4407 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b'))) */
   /* search1542_0 -> 21 in the cache */
   /* search1542_1 -> 41 in the cache */
   /* search1542 -> 3959 in the cache */

   /* replace1542_0_0 -> 41 in the cache */
   /* replace1542_0_1 -> 1507 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 1507 },
      -1,
   } },
   /* replace1542_1_0_0_0 -> 21 in the cache */
   /* replace1542_1_0_0_1 -> 41 in the cache */
   /* replace1542_1_0_0 -> 3946 in the cache */
   /* replace1542_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3946, 21 },
      -1,
   } },
   /* replace1542_1_1 -> 3970 in the cache */
   /* replace1542_1_2_0 -> 21 in the cache */
   /* replace1542_1_2_1 -> 41 in the cache */
   /* replace1542_1_2 -> 3948 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4410, 3970, 3948 },
      -1,
   } },
   /* replace1542_2_0_0 -> 21 in the cache */
   /* replace1542_2_0_1_0 -> 21 in the cache */
   /* replace1542_2_0_1_1 -> 41 in the cache */
   /* replace1542_2_0_1 -> 3962 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 21, 3962 },
      -1,
   } },
   /* replace1542_2_1 -> 22 in the cache */
   /* replace1542_2_2_0 -> 21 in the cache */
   /* replace1542_2_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4412, 22, 4413 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4409, 4411, 4414 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b'))) */
   /* search1543_0 -> 21 in the cache */
   /* search1543_1 -> 41 in the cache */
   /* search1543 -> 3973 in the cache */

   /* replace1543_0_0 -> 41 in the cache */
   /* replace1543_0_1 -> 910 in the cache */
   /* replace1543_0 -> 3960 in the cache */
   /* replace1543_1_0_0_0 -> 21 in the cache */
   /* replace1543_1_0_0_1 -> 41 in the cache */
   /* replace1543_1_0_0 -> 3957 in the cache */
   /* replace1543_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3957, 21 },
      -1,
   } },
   /* replace1543_1_1 -> 3970 in the cache */
   /* replace1543_1_2_0 -> 21 in the cache */
   /* replace1543_1_2_1 -> 41 in the cache */
   /* replace1543_1_2 -> 3957 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4416, 3970, 3957 },
      -1,
   } },
   /* replace1543_2_0_0 -> 21 in the cache */
   /* replace1543_2_0_1_0 -> 21 in the cache */
   /* replace1543_2_0_1_1 -> 41 in the cache */
   /* replace1543_2_0_1 -> 3957 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 3957 },
      -1,
   } },
   /* replace1543_2_1 -> 22 in the cache */
   /* replace1543_2_2_0 -> 21 in the cache */
   /* replace1543_2_2_1 -> 41 in the cache */
   /* replace1543_2_2 -> 3957 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4418, 22, 3957 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3960, 4417, 4419 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1544_0_0_0 -> 0 in the cache */
   /* search1544_0_0_1 -> 3 in the cache */
   /* search1544_0_0 -> 1087 in the cache */
   /* search1544_0_1_0 -> 158 in the cache */
   /* search1544_0_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1087, 4421 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4422 },
      -1,
   } },

   /* replace1544_0_0 -> 0 in the cache */
   /* replace1544_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1544_1_0 -> 158 in the cache */
   /* replace1544_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4424, 4425 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1545_0_0_0 -> 0 in the cache */
   /* search1545_0_0_1 -> 3 in the cache */
   /* search1545_0_0 -> 1087 in the cache */
   /* search1545_0_1_0 -> 158 in the cache */
   /* search1545_0_1_1 -> 159 in the cache */
   /* search1545_0_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1087, 4421 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4427 },
      -1,
   } },

   /* replace1545_0_0 -> 0 in the cache */
   /* replace1545_0_1 -> 3 in the cache */
   /* replace1545_0 -> 4424 in the cache */
   /* replace1545_1_0 -> 158 in the cache */
   /* replace1545_1_1 -> 159 in the cache */
   /* replace1545_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4424, 4425 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1546_0_0_0 -> 0 in the cache */
   /* search1546_0_0_1 -> 3 in the cache */
   /* search1546_0_0 -> 1087 in the cache */
   /* search1546_0_1_0 -> 158 in the cache */
   /* search1546_0_1_1 -> 159 in the cache */
   /* search1546_0_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1087, 4425 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4430 },
      -1,
   } },

   /* replace1546_0_0 -> 0 in the cache */
   /* replace1546_0_1 -> 3 in the cache */
   /* replace1546_0 -> 4424 in the cache */
   /* replace1546_1_0 -> 158 in the cache */
   /* replace1546_1_1 -> 159 in the cache */
   /* replace1546_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4424, 4421 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1547_0_0_0 -> 0 in the cache */
   /* search1547_0_0_1 -> 3 in the cache */
   /* search1547_0_0 -> 1087 in the cache */
   /* search1547_0_1_0 -> 158 in the cache */
   /* search1547_0_1_1 -> 159 in the cache */
   /* search1547_0_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1087, 4425 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4433 },
      -1,
   } },

   /* replace1547_0_0 -> 0 in the cache */
   /* replace1547_0_1 -> 3 in the cache */
   /* replace1547_0 -> 4424 in the cache */
   /* replace1547_1_0 -> 158 in the cache */
   /* replace1547_1_1 -> 159 in the cache */
   /* replace1547_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4424, 4421 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1548_0_0_0 -> 0 in the cache */
   /* search1548_0_0_1 -> 3 in the cache */
   /* search1548_0_0 -> 4424 in the cache */
   /* search1548_0_1_0 -> 158 in the cache */
   /* search1548_0_1_1 -> 159 in the cache */
   /* search1548_0_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4424, 4425 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4436 },
      -1,
   } },

   /* replace1548_0_0 -> 0 in the cache */
   /* replace1548_0_1 -> 3 in the cache */
   /* replace1548_0 -> 1087 in the cache */
   /* replace1548_1_0 -> 158 in the cache */
   /* replace1548_1_1 -> 159 in the cache */
   /* replace1548_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1087, 4421 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1549_0_0_0 -> 0 in the cache */
   /* search1549_0_0_1 -> 3 in the cache */
   /* search1549_0_0 -> 4424 in the cache */
   /* search1549_0_1_0 -> 158 in the cache */
   /* search1549_0_1_1 -> 159 in the cache */
   /* search1549_0_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4424, 4425 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4439 },
      -1,
   } },

   /* replace1549_0_0 -> 0 in the cache */
   /* replace1549_0_1 -> 3 in the cache */
   /* replace1549_0 -> 1087 in the cache */
   /* replace1549_1_0 -> 158 in the cache */
   /* replace1549_1_1 -> 159 in the cache */
   /* replace1549_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1087, 4421 },
      -1,
   } },

   /* ('f2f', ('b2f', 'a')) => ('b2f', 'a') */
   /* search1550_0_0 -> 32 in the cache */
   /* search1550_0 -> 1541 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2f,
      -1, 0,
      { 1541 },
      -1,
   } },

   /* replace1550_0 -> 32 in the cache */
   /* replace1550 -> 1541 in the cache */

   /* ('f2u', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1551_0_0 -> 32 in the cache */
   /* search1551_0 -> 1541 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 1541 },
      -1,
   } },

   /* replace1551_0 -> 32 in the cache */
   /* replace1551 -> 3280 in the cache */

   /* ('f2i', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1552_0_0 -> 32 in the cache */
   /* search1552_0 -> 1541 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 1541 },
      -1,
   } },

   /* replace1552_0 -> 32 in the cache */
   /* replace1552 -> 3280 in the cache */

   /* ('u2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1553_0_0 -> 32 in the cache */
   /* search1553_0 -> 3280 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 3280 },
      -1,
   } },

   /* replace1553_0 -> 32 in the cache */
   /* replace1553 -> 1541 in the cache */

   /* ('u2u', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1554_0_0 -> 32 in the cache */
   /* search1554_0 -> 3280 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3280 },
      -1,
   } },

   /* replace1554_0 -> 32 in the cache */
   /* replace1554 -> 3280 in the cache */

   /* ('i2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1555_0_0 -> 32 in the cache */
   /* search1555_0 -> 3280 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3280 },
      -1,
   } },

   /* replace1555_0 -> 32 in the cache */
   /* replace1555 -> 1541 in the cache */

   /* ('i2i', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1556_0_0 -> 32 in the cache */
   /* search1556_0 -> 3280 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3280 },
      -1,
   } },

   /* replace1556_0 -> 32 in the cache */
   /* replace1556 -> 3280 in the cache */

   /* ('i2i1', 'a@1') => a */
   /* search1557_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 770 },
      -1,
   } },

   /* replace1557 -> 770 in the cache */

   /* ('i2i8', 'a@8') => a */
   /* search1558_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1558 -> 6 in the cache */

   /* ('i2i16', 'a@16') => a */
   /* search1559_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1559 -> 11 in the cache */

   /* ('i2i32', 'a@32') => a */
   /* search1560_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1560 -> 16 in the cache */

   /* ('i2i64', 'a@64') => a */
   /* search1561_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1561 -> 21 in the cache */

   /* ('u2u1', 'a@1') => a */
   /* search1562_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 770 },
      -1,
   } },

   /* replace1562 -> 770 in the cache */

   /* ('u2u8', 'a@8') => a */
   /* search1563_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1563 -> 6 in the cache */

   /* ('u2u16', 'a@16') => a */
   /* search1564_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1564 -> 11 in the cache */

   /* ('u2u32', 'a@32') => a */
   /* search1565_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1565 -> 16 in the cache */

   /* ('u2u64', 'a@64') => a */
   /* search1566_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1566 -> 21 in the cache */

   /* ('f2f16', 'a@16') => a */
   /* search1567_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1567 -> 11 in the cache */

   /* ('f2f32', 'a@32') => a */
   /* search1568_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1568 -> 16 in the cache */

   /* ('f2f64', 'a@64') => a */
   /* search1569_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1569 -> 21 in the cache */

   /* ('b2b1', 'a@1') => a */
   /* search1570_0 -> 770 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_b2b1,
      -1, 0,
      { 770 },
      -1,
   } },

   /* replace1570 -> 770 in the cache */

   /* ('b2b8', 'a@8') => a */
   /* search1571_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2b8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1571 -> 6 in the cache */

   /* ('b2b16', 'a@16') => a */
   /* search1572_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2b16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1572 -> 11 in the cache */

   /* ('b2b32', 'a@32') => a */
   /* search1573_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2b32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1573 -> 16 in the cache */

   /* ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1574_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4466 },
      -1,
   } },

   /* replace1574_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1575_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4469 },
      -1,
   } },

   /* replace1575_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1576_0_0 -> 32 in the cache */
   /* search1576_0 -> 4466 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4466 },
      -1,
   } },

   /* replace1576_0 -> 32 in the cache */
   /* replace1576 -> 4468 in the cache */

   /* ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1577_0_0 -> 32 in the cache */
   /* search1577_0 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4469 },
      -1,
   } },

   /* replace1577_0 -> 32 in the cache */
   /* replace1577 -> 4471 in the cache */

   /* ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1578_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4474 },
      -1,
   } },

   /* replace1578_0 -> 32 in the cache */
   /* replace1578 -> 4468 in the cache */

   /* ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1579_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4476 },
      -1,
   } },

   /* replace1579_0 -> 32 in the cache */
   /* replace1579 -> 4471 in the cache */

   /* ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1580_0_0 -> 32 in the cache */
   /* search1580_0 -> 4474 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4474 },
      -1,
   } },

   /* replace1580_0 -> 32 in the cache */
   /* replace1580 -> 4468 in the cache */

   /* ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1581_0_0 -> 32 in the cache */
   /* search1581_0 -> 4476 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4476 },
      -1,
   } },

   /* replace1581_0 -> 32 in the cache */
   /* replace1581 -> 4471 in the cache */

   /* ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1582_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1582_0 -> 32 in the cache */
   /* replace1582 -> 4468 in the cache */

   /* ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1583_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1583_0 -> 32 in the cache */
   /* replace1583 -> 4471 in the cache */

   /* ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1584_0_0 -> 32 in the cache */
   /* search1584_0 -> 4480 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1584_0 -> 32 in the cache */
   /* replace1584 -> 4468 in the cache */

   /* ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1585_0_0 -> 32 in the cache */
   /* search1585_0 -> 4482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1585_0 -> 32 in the cache */
   /* replace1585 -> 4471 in the cache */

   /* ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1586_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1586_0 -> 32 in the cache */
   /* replace1586 -> 4468 in the cache */

   /* ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1587_0_0 -> 32 in the cache */
   /* search1587_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1587_0 -> 32 in the cache */
   /* replace1587 -> 4471 in the cache */

   /* ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1588_0_0 -> 32 in the cache */
   /* search1588_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1588_0 -> 32 in the cache */
   /* replace1588 -> 4468 in the cache */

   /* ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1589_0_0 -> 32 in the cache */
   /* search1589_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1589_0 -> 32 in the cache */
   /* replace1589 -> 4471 in the cache */

   /* ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a') */
   /* search1590_0_0 -> 770 in the cache */
   /* search1590_0 -> 4449 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4449 },
      -1,
   } },

   /* replace1590_0 -> 770 in the cache */
   /* replace1590 -> 2045 in the cache */

   /* ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a') */
   /* search1591_0_0 -> 770 in the cache */
   /* search1591_0 -> 4454 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4454 },
      -1,
   } },

   /* replace1591_0 -> 770 in the cache */
   /* replace1591 -> 2047 in the cache */

   /* ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1592_0_0 -> 32 in the cache */
   /* search1592_0 -> 4474 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4474 },
      -1,
   } },

   /* replace1592_0 -> 32 in the cache */
   /* replace1592 -> 4466 in the cache */

   /* ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1593_0_0 -> 32 in the cache */
   /* search1593_0 -> 4476 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4476 },
      -1,
   } },

   /* replace1593_0 -> 32 in the cache */
   /* replace1593 -> 4469 in the cache */

   /* ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1594_0_0 -> 32 in the cache */
   /* search1594_0 -> 4474 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4474 },
      -1,
   } },

   /* replace1594_0 -> 32 in the cache */
   /* replace1594 -> 4466 in the cache */

   /* ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1595_0_0 -> 32 in the cache */
   /* search1595_0 -> 4476 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4476 },
      -1,
   } },

   /* replace1595_0 -> 32 in the cache */
   /* replace1595 -> 4469 in the cache */

   /* ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1596_0_0 -> 32 in the cache */
   /* search1596_0 -> 4480 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1596_0 -> 32 in the cache */
   /* replace1596 -> 4466 in the cache */

   /* ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1597_0_0 -> 32 in the cache */
   /* search1597_0 -> 4482 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1597_0 -> 32 in the cache */
   /* replace1597 -> 4469 in the cache */

   /* ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1598_0_0 -> 32 in the cache */
   /* search1598_0 -> 4480 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1598_0 -> 32 in the cache */
   /* replace1598 -> 4466 in the cache */

   /* ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1599_0_0 -> 32 in the cache */
   /* search1599_0 -> 4482 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1599_0 -> 32 in the cache */
   /* replace1599 -> 4469 in the cache */

   /* ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1600_0_0 -> 32 in the cache */
   /* search1600_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1600_0 -> 32 in the cache */
   /* replace1600 -> 4466 in the cache */

   /* ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1601_0_0 -> 32 in the cache */
   /* search1601_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1601_0 -> 32 in the cache */
   /* replace1601 -> 4469 in the cache */

   /* ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1602_0_0 -> 32 in the cache */
   /* search1602_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1602_0 -> 32 in the cache */
   /* replace1602 -> 4466 in the cache */

   /* ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1603_0_0 -> 32 in the cache */
   /* search1603_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1603_0 -> 32 in the cache */
   /* replace1603 -> 4469 in the cache */

   /* ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a') */
   /* search1604_0_0 -> 770 in the cache */
   /* search1604_0 -> 4449 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4449 },
      -1,
   } },

   /* replace1604_0 -> 770 in the cache */
   /* replace1604 -> 2049 in the cache */

   /* ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a') */
   /* search1605_0_0 -> 770 in the cache */
   /* search1605_0 -> 4454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4454 },
      -1,
   } },

   /* replace1605_0 -> 770 in the cache */
   /* replace1605 -> 2051 in the cache */

   /* ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a') */
   /* search1606_0_0 -> 770 in the cache */
   /* search1606_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace1606_0 -> 770 in the cache */
   /* replace1606 -> 2049 in the cache */

   /* ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a') */
   /* search1607_0_0 -> 770 in the cache */
   /* search1607_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace1607_0 -> 770 in the cache */
   /* replace1607 -> 2051 in the cache */

   /* ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a') */
   /* search1608_0_0 -> 6 in the cache */
   /* search1608_0 -> 4450 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4450 },
      -1,
   } },

   /* replace1608_0 -> 6 in the cache */
   /* replace1608 -> 2091 in the cache */

   /* ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a') */
   /* search1609_0_0 -> 6 in the cache */
   /* search1609_0 -> 4455 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4455 },
      -1,
   } },

   /* replace1609_0 -> 6 in the cache */
   /* replace1609 -> 2093 in the cache */

   /* ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1610_0_0 -> 32 in the cache */
   /* search1610_0 -> 4480 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1610_0 -> 32 in the cache */
   /* replace1610 -> 4474 in the cache */

   /* ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1611_0_0 -> 32 in the cache */
   /* search1611_0 -> 4482 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1611_0 -> 32 in the cache */
   /* replace1611 -> 4476 in the cache */

   /* ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1612_0_0 -> 32 in the cache */
   /* search1612_0 -> 4480 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4480 },
      -1,
   } },

   /* replace1612_0 -> 32 in the cache */
   /* replace1612 -> 4474 in the cache */

   /* ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1613_0_0 -> 32 in the cache */
   /* search1613_0 -> 4482 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4482 },
      -1,
   } },

   /* replace1613_0 -> 32 in the cache */
   /* replace1613 -> 4476 in the cache */

   /* ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1614_0_0 -> 32 in the cache */
   /* search1614_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1614_0 -> 32 in the cache */
   /* replace1614 -> 4474 in the cache */

   /* ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1615_0_0 -> 32 in the cache */
   /* search1615_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1615_0 -> 32 in the cache */
   /* replace1615 -> 4476 in the cache */

   /* ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1616_0_0 -> 32 in the cache */
   /* search1616_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1616_0 -> 32 in the cache */
   /* replace1616 -> 4474 in the cache */

   /* ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1617_0_0 -> 32 in the cache */
   /* search1617_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1617_0 -> 32 in the cache */
   /* replace1617 -> 4476 in the cache */

   /* ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a') */
   /* search1618_0_0 -> 770 in the cache */
   /* search1618_0 -> 4449 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4449 },
      -1,
   } },

   /* replace1618_0 -> 770 in the cache */
   /* replace1618 -> 2061 in the cache */

   /* ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a') */
   /* search1619_0_0 -> 770 in the cache */
   /* search1619_0 -> 4454 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4454 },
      -1,
   } },

   /* replace1619_0 -> 770 in the cache */
   /* replace1619 -> 2059 in the cache */

   /* ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a') */
   /* search1620_0_0 -> 770 in the cache */
   /* search1620_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace1620_0 -> 770 in the cache */
   /* replace1620 -> 2061 in the cache */

   /* ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a') */
   /* search1621_0_0 -> 770 in the cache */
   /* search1621_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace1621_0 -> 770 in the cache */
   /* replace1621 -> 2059 in the cache */

   /* ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a') */
   /* search1622_0_0 -> 6 in the cache */
   /* search1622_0 -> 4450 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4450 },
      -1,
   } },

   /* replace1622_0 -> 6 in the cache */
   /* replace1622 -> 2101 in the cache */

   /* ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a') */
   /* search1623_0_0 -> 6 in the cache */
   /* search1623_0 -> 4455 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4455 },
      -1,
   } },

   /* replace1623_0 -> 6 in the cache */
   /* replace1623 -> 2099 in the cache */

   /* ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a') */
   /* search1624_0_0 -> 770 in the cache */
   /* search1624_0 -> 2049 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace1624_0 -> 770 in the cache */
   /* replace1624 -> 2061 in the cache */

   /* ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a') */
   /* search1625_0_0 -> 770 in the cache */
   /* search1625_0 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace1625_0 -> 770 in the cache */
   /* replace1625 -> 2059 in the cache */

   /* ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a') */
   /* search1626_0_0 -> 6 in the cache */
   /* search1626_0 -> 2091 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace1626_0 -> 6 in the cache */
   /* replace1626 -> 2101 in the cache */

   /* ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a') */
   /* search1627_0_0 -> 6 in the cache */
   /* search1627_0 -> 2093 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace1627_0 -> 6 in the cache */
   /* replace1627 -> 2099 in the cache */

   /* ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a') */
   /* search1628_0_0 -> 11 in the cache */
   /* search1628_0 -> 4451 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4451 },
      -1,
   } },

   /* replace1628_0 -> 11 in the cache */
   /* replace1628 -> 2127 in the cache */

   /* ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a') */
   /* search1629_0_0 -> 11 in the cache */
   /* search1629_0 -> 4456 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4456 },
      -1,
   } },

   /* replace1629_0 -> 11 in the cache */
   /* replace1629 -> 2129 in the cache */

   /* ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1630_0_0 -> 32 in the cache */
   /* search1630_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1630_0 -> 32 in the cache */
   /* replace1630 -> 4480 in the cache */

   /* ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1631_0_0 -> 32 in the cache */
   /* search1631_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1631_0 -> 32 in the cache */
   /* replace1631 -> 4482 in the cache */

   /* ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1632_0_0 -> 32 in the cache */
   /* search1632_0 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4486 },
      -1,
   } },

   /* replace1632_0 -> 32 in the cache */
   /* replace1632 -> 4480 in the cache */

   /* ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1633_0_0 -> 32 in the cache */
   /* search1633_0 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3394 },
      -1,
   } },

   /* replace1633_0 -> 32 in the cache */
   /* replace1633 -> 4482 in the cache */

   /* ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a') */
   /* search1634_0_0 -> 770 in the cache */
   /* search1634_0 -> 4449 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4449 },
      -1,
   } },

   /* replace1634_0 -> 770 in the cache */
   /* replace1634 -> 2065 in the cache */

   /* ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a') */
   /* search1635_0_0 -> 770 in the cache */
   /* search1635_0 -> 4454 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4454 },
      -1,
   } },

   /* replace1635_0 -> 770 in the cache */
   /* replace1635 -> 2063 in the cache */

   /* ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a') */
   /* search1636_0_0 -> 770 in the cache */
   /* search1636_0 -> 2045 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2045 },
      -1,
   } },

   /* replace1636_0 -> 770 in the cache */
   /* replace1636 -> 2065 in the cache */

   /* ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a') */
   /* search1637_0_0 -> 770 in the cache */
   /* search1637_0 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2047 },
      -1,
   } },

   /* replace1637_0 -> 770 in the cache */
   /* replace1637 -> 2063 in the cache */

   /* ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a') */
   /* search1638_0_0 -> 6 in the cache */
   /* search1638_0 -> 4450 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4450 },
      -1,
   } },

   /* replace1638_0 -> 6 in the cache */
   /* replace1638 -> 2105 in the cache */

   /* ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a') */
   /* search1639_0_0 -> 6 in the cache */
   /* search1639_0 -> 4455 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4455 },
      -1,
   } },

   /* replace1639_0 -> 6 in the cache */
   /* replace1639 -> 2103 in the cache */

   /* ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a') */
   /* search1640_0_0 -> 770 in the cache */
   /* search1640_0 -> 2049 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2049 },
      -1,
   } },

   /* replace1640_0 -> 770 in the cache */
   /* replace1640 -> 2065 in the cache */

   /* ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a') */
   /* search1641_0_0 -> 770 in the cache */
   /* search1641_0 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2051 },
      -1,
   } },

   /* replace1641_0 -> 770 in the cache */
   /* replace1641 -> 2063 in the cache */

   /* ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a') */
   /* search1642_0_0 -> 6 in the cache */
   /* search1642_0 -> 2091 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2091 },
      -1,
   } },

   /* replace1642_0 -> 6 in the cache */
   /* replace1642 -> 2105 in the cache */

   /* ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a') */
   /* search1643_0_0 -> 6 in the cache */
   /* search1643_0 -> 2093 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2093 },
      -1,
   } },

   /* replace1643_0 -> 6 in the cache */
   /* replace1643 -> 2103 in the cache */

   /* ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a') */
   /* search1644_0_0 -> 11 in the cache */
   /* search1644_0 -> 4451 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4451 },
      -1,
   } },

   /* replace1644_0 -> 11 in the cache */
   /* replace1644 -> 2137 in the cache */

   /* ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a') */
   /* search1645_0_0 -> 11 in the cache */
   /* search1645_0 -> 4456 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4456 },
      -1,
   } },

   /* replace1645_0 -> 11 in the cache */
   /* replace1645 -> 2135 in the cache */

   /* ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a') */
   /* search1646_0_0 -> 770 in the cache */
   /* search1646_0 -> 2061 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2061 },
      -1,
   } },

   /* replace1646_0 -> 770 in the cache */
   /* replace1646 -> 2065 in the cache */

   /* ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a') */
   /* search1647_0_0 -> 770 in the cache */
   /* search1647_0 -> 2059 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2059 },
      -1,
   } },

   /* replace1647_0 -> 770 in the cache */
   /* replace1647 -> 2063 in the cache */

   /* ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a') */
   /* search1648_0_0 -> 6 in the cache */
   /* search1648_0 -> 2101 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2101 },
      -1,
   } },

   /* replace1648_0 -> 6 in the cache */
   /* replace1648 -> 2105 in the cache */

   /* ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a') */
   /* search1649_0_0 -> 6 in the cache */
   /* search1649_0 -> 2099 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2099 },
      -1,
   } },

   /* replace1649_0 -> 6 in the cache */
   /* replace1649 -> 2103 in the cache */

   /* ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a') */
   /* search1650_0_0 -> 11 in the cache */
   /* search1650_0 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2127 },
      -1,
   } },

   /* replace1650_0 -> 11 in the cache */
   /* replace1650 -> 2137 in the cache */

   /* ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a') */
   /* search1651_0_0 -> 11 in the cache */
   /* search1651_0 -> 2129 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace1651_0 -> 11 in the cache */
   /* replace1651 -> 2135 in the cache */

   /* ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a') */
   /* search1652_0_0 -> 16 in the cache */
   /* search1652_0 -> 4452 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4452 },
      -1,
   } },

   /* replace1652_0 -> 16 in the cache */
   /* replace1652 -> 2147 in the cache */

   /* ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a') */
   /* search1653_0_0 -> 16 in the cache */
   /* search1653_0 -> 4457 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4457 },
      -1,
   } },

   /* replace1653_0 -> 16 in the cache */
   /* replace1653 -> 2149 in the cache */

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1654_0_0 -> 16 in the cache */
   /* search1654_0_1 -> 36 in the cache */
   /* search1654_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1654_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1655_0_0 -> 16 in the cache */
   /* search1655_0_1 -> 36 in the cache */
   /* search1655_0 -> 3387 in the cache */
   /* search1655 -> 4555 in the cache */

   /* replace1655_0 -> 16 in the cache */
   /* replace1655 -> 4556 in the cache */

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1656_0_0 -> 16 in the cache */
   /* search1656_0_1 -> 36 in the cache */
   /* search1656_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1656_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1657_0_0 -> 16 in the cache */
   /* search1657_0_1 -> 36 in the cache */
   /* search1657_0 -> 3387 in the cache */
   /* search1657 -> 4557 in the cache */

   /* replace1657_0 -> 16 in the cache */
   /* replace1657 -> 4558 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1658_0_0 -> 16 in the cache */
   /* search1658_0_1 -> 36 in the cache */
   /* search1658_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1658_0 -> 16 in the cache */
   /* replace1658 -> 4452 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1659_0_0 -> 16 in the cache */
   /* search1659_0_1 -> 36 in the cache */
   /* search1659_0 -> 3387 in the cache */
   /* search1659 -> 4559 in the cache */

   /* replace1659_0 -> 16 in the cache */
   /* replace1659 -> 4452 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1660_0_0 -> 16 in the cache */
   /* search1660_0_1 -> 36 in the cache */
   /* search1660_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1660_0 -> 16 in the cache */
   /* replace1660 -> 2193 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1661_0_0 -> 16 in the cache */
   /* search1661_0_1 -> 36 in the cache */
   /* search1661_0 -> 3387 in the cache */
   /* search1661 -> 4560 in the cache */

   /* replace1661_0 -> 16 in the cache */
   /* replace1661 -> 2193 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1662_0_0 -> 16 in the cache */
   /* search1662_0_1 -> 36 in the cache */
   /* search1662_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1662_0 -> 16 in the cache */
   /* replace1662 -> 3505 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1663_0_0 -> 16 in the cache */
   /* search1663_0_1 -> 36 in the cache */
   /* search1663_0 -> 3387 in the cache */
   /* search1663 -> 4561 in the cache */

   /* replace1663_0 -> 16 in the cache */
   /* replace1663 -> 3505 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1664_0_0 -> 16 in the cache */
   /* search1664_0_1 -> 36 in the cache */
   /* search1664_0 -> 3387 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3387 },
      -1,
   } },

   /* replace1664_0 -> 16 in the cache */
   /* replace1664 -> 4457 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1665_0_0 -> 16 in the cache */
   /* search1665_0_1 -> 36 in the cache */
   /* search1665_0 -> 3387 in the cache */
   /* search1665 -> 4562 in the cache */

   /* replace1665_0 -> 16 in the cache */
   /* replace1665 -> 4457 in the cache */

   /* ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1666_0_0 -> 6 in the cache */
   /* search1666_0 -> 2091 in the cache */
   /* search1666_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2091, 4563 },
      -1,
   } },

   /* replace1666_0 -> 6 in the cache */
   /* replace1666_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 4565 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1667_0_0 -> 6 in the cache */
   /* search1667_0 -> 2091 in the cache */
   /* search1667_1_0 -> 2280 in the cache */
   /* search1667_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2091, 4563 },
      -1,
   } },

   /* replace1667_0 -> 6 in the cache */
   /* replace1667_1_0 -> 2280 in the cache */
   /* replace1667_1 -> 4565 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 4565 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1668_0_0 -> 6 in the cache */
   /* search1668_0 -> 2091 in the cache */
   /* search1668_1_0 -> 2280 in the cache */
   /* search1668_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2091, 4563 },
      -1,
   } },

   /* replace1668_0 -> 6 in the cache */
   /* replace1668_1_0 -> 2280 in the cache */
   /* replace1668_1 -> 4565 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4565 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1669_0_0 -> 6 in the cache */
   /* search1669_0 -> 2091 in the cache */
   /* search1669_1_0 -> 2280 in the cache */
   /* search1669_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2091, 4563 },
      -1,
   } },

   /* replace1669_0 -> 6 in the cache */
   /* replace1669_1_0 -> 2280 in the cache */
   /* replace1669_1 -> 4565 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4565 },
      -1,
   } },

   /* ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4573 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4574, 4576 },
      -1,
   } },

   /* replace1670_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* replace1670_1 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4578, 4575 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1671_0_0 -> 4573 in the cache */
   /* search1671_0 -> 4574 in the cache */
   /* search1671_1_0 -> 4575 in the cache */
   /* search1671_1 -> 4576 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4574, 4576 },
      -1,
   } },

   /* replace1671_0_0 -> 4573 in the cache */
   /* replace1671_0 -> 4578 in the cache */
   /* replace1671_1 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4578, 4575 },
      -1,
   } },

   /* ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1672_0_0 -> 6 in the cache */
   /* search1672_0 -> 2091 in the cache */
   /* search1672_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2091, 2179 },
      -1,
   } },

   /* replace1672_0_0 -> 6 in the cache */
   /* replace1672_0_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4583 },
      -1,
   } },
   /* replace1672_1_0_0_0 -> 268 in the cache */
   /* replace1672_1_0_0 -> 4583 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4583 },
      -1,
   } },
   /* replace1672_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4585, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4584, 4586 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1673_0_0 -> 6 in the cache */
   /* search1673_0 -> 2091 in the cache */
   /* search1673_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2091, 2179 },
      -1,
   } },

   /* replace1673_0_0 -> 6 in the cache */
   /* replace1673_0_1_0 -> 268 in the cache */
   /* replace1673_0_1 -> 4583 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4583 },
      -1,
   } },
   /* replace1673_1_0_0_0 -> 268 in the cache */
   /* replace1673_1_0_0 -> 4583 in the cache */
   /* replace1673_1_0 -> 4585 in the cache */
   /* replace1673_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4585, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4589, 4590 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1674_0_0 -> 6 in the cache */
   /* search1674_0 -> 2091 in the cache */
   /* search1674_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2091, 2179 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1674_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4593, 268 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1674_1_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4595, 268 },
      -1,
   } },
   /* replace1674_1_1_0 -> 6 in the cache */
   /* replace1674_1_1_1_0 -> 268 in the cache */
   /* replace1674_1_1_1 -> 4583 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4583 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4596, 4597 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4594, 4598 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1675_1_0 -> 2280 in the cache */
   /* search1675_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4600, 4563 },
      -1,
   } },

   /* replace1675_0_0 -> 11 in the cache */
   /* replace1675_0_1 -> 4595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4595 },
      -1,
   } },
   /* replace1675_1_0_0 -> 11 in the cache */
   /* replace1675_1_0_1 -> 4593 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4593 },
      -1,
   } },
   /* replace1675_1_1_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 11 },
      -1,
   } },
   /* replace1675_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4604, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4603, 4605 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4602, 4606 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1676_0_0 -> 6 in the cache */
   /* search1676_0 -> 2091 in the cache */
   /* search1676_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2091, 2179 },
      -1,
   } },

   /* replace1676_0_0 -> 4595 in the cache */
   /* replace1676_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4595, 268 },
      -1,
   } },
   /* replace1676_1_0_0 -> 4593 in the cache */
   /* replace1676_1_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4593, 268 },
      -1,
   } },
   /* replace1676_1_1_0 -> 6 in the cache */
   /* replace1676_1_1_1_0 -> 268 in the cache */
   /* replace1676_1_1_1 -> 4583 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4583 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4610, 4611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4609, 4612 },
      -1,
   } },

   /* ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1677_0 -> 4600 in the cache */
   /* search1677_1_0 -> 2280 in the cache */
   /* search1677_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4600, 4563 },
      -1,
   } },

   /* replace1677_0_0 -> 11 in the cache */
   /* replace1677_0_1 -> 4593 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4593 },
      -1,
   } },
   /* replace1677_1_0_0 -> 11 in the cache */
   /* replace1677_1_0_1 -> 4595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4595 },
      -1,
   } },
   /* replace1677_1_1_0_0 -> 11 in the cache */
   /* replace1677_1_1_0 -> 4604 in the cache */
   /* replace1677_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4604, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4616, 4617 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4615, 4618 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1678_0_0 -> 6 in the cache */
   /* search1678_0 -> 2101 in the cache */
   /* search1678_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2101, 4620 },
      -1,
   } },

   /* replace1678_0 -> 6 in the cache */
   /* replace1678_1_0 -> 2280 in the cache */
   /* replace1678_1 -> 4565 in the cache */
   /* replace1678 -> 4566 in the cache */

   /* ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1679_0_0 -> 6 in the cache */
   /* search1679_0 -> 2101 in the cache */
   /* search1679_1_0 -> 2280 in the cache */
   /* search1679_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2101, 4620 },
      -1,
   } },

   /* replace1679_0 -> 6 in the cache */
   /* replace1679_1_0 -> 2280 in the cache */
   /* replace1679_1 -> 4565 in the cache */
   /* replace1679 -> 4568 in the cache */

   /* ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1680_0_0 -> 6 in the cache */
   /* search1680_0 -> 2101 in the cache */
   /* search1680_1_0 -> 2280 in the cache */
   /* search1680_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2101, 4620 },
      -1,
   } },

   /* replace1680_0 -> 6 in the cache */
   /* replace1680_1_0 -> 2280 in the cache */
   /* replace1680_1 -> 4565 in the cache */
   /* replace1680 -> 4570 in the cache */

   /* ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1681_0_0 -> 6 in the cache */
   /* search1681_0 -> 2101 in the cache */
   /* search1681_1_0 -> 2280 in the cache */
   /* search1681_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2101, 4620 },
      -1,
   } },

   /* replace1681_0 -> 6 in the cache */
   /* replace1681_1_0 -> 2280 in the cache */
   /* replace1681_1 -> 4565 in the cache */
   /* replace1681 -> 4572 in the cache */

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1682_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* search1682_1_0 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4625, 4626 },
      -1,
   } },

   /* replace1682_0_0 -> 4573 in the cache */
   /* replace1682_0 -> 4578 in the cache */
   /* replace1682_1 -> 4575 in the cache */
   /* replace1682 -> 4579 in the cache */

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1683_0_0 -> 4573 in the cache */
   /* search1683_0 -> 4625 in the cache */
   /* search1683_1_0 -> 4575 in the cache */
   /* search1683_1 -> 4626 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4625, 4626 },
      -1,
   } },

   /* replace1683_0_0 -> 4573 in the cache */
   /* replace1683_0 -> 4578 in the cache */
   /* replace1683_1 -> 4575 in the cache */
   /* replace1683 -> 4581 in the cache */

   /* ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1684_0_0 -> 6 in the cache */
   /* search1684_0 -> 2101 in the cache */
   /* search1684_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2101, 33 },
      -1,
   } },

   /* replace1684_0_0 -> 6 in the cache */
   /* replace1684_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4630 },
      -1,
   } },
   /* replace1684_1_0_0_0 -> 36 in the cache */
   /* replace1684_1_0_0 -> 4630 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4630 },
      -1,
   } },
   /* replace1684_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4632, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4631, 4633 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1685_0_0 -> 6 in the cache */
   /* search1685_0 -> 2101 in the cache */
   /* search1685_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2101, 33 },
      -1,
   } },

   /* replace1685_0_0 -> 6 in the cache */
   /* replace1685_0_1_0 -> 36 in the cache */
   /* replace1685_0_1 -> 4630 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4630 },
      -1,
   } },
   /* replace1685_1_0_0_0 -> 36 in the cache */
   /* replace1685_1_0_0 -> 4630 in the cache */
   /* replace1685_1_0 -> 4632 in the cache */
   /* replace1685_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4632, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4636, 4637 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1686_0_0 -> 6 in the cache */
   /* search1686_0 -> 2101 in the cache */
   /* search1686_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2101, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1686_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4640, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1686_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4642, 36 },
      -1,
   } },
   /* replace1686_1_1_0 -> 6 in the cache */
   /* replace1686_1_1_1_0 -> 36 in the cache */
   /* replace1686_1_1_1 -> 4630 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4643, 4644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4641, 4645 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   /* search1687_0 -> 296 in the cache */
   /* search1687_1_0 -> 2280 in the cache */
   /* search1687_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 296, 4620 },
      -1,
   } },

   /* replace1687_0_0 -> 16 in the cache */
   /* replace1687_0_1 -> 4642 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4642 },
      -1,
   } },
   /* replace1687_1_0_0 -> 16 in the cache */
   /* replace1687_1_0_1 -> 4640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4640 },
      -1,
   } },
   /* replace1687_1_1_0_0 -> 16 in the cache */
   /* replace1687_1_1_0 -> 4556 in the cache */
   /* replace1687_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4556, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4649, 4650 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4648, 4651 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1688_0_0 -> 6 in the cache */
   /* search1688_0 -> 2101 in the cache */
   /* search1688_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2101, 33 },
      -1,
   } },

   /* replace1688_0_0 -> 4642 in the cache */
   /* replace1688_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4642, 36 },
      -1,
   } },
   /* replace1688_1_0_0 -> 4640 in the cache */
   /* replace1688_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4640, 36 },
      -1,
   } },
   /* replace1688_1_1_0 -> 6 in the cache */
   /* replace1688_1_1_1_0 -> 36 in the cache */
   /* replace1688_1_1_1 -> 4630 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4655, 4656 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4654, 4657 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1689_0 -> 296 in the cache */
   /* search1689_1_0 -> 2280 in the cache */
   /* search1689_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 296, 4620 },
      -1,
   } },

   /* replace1689_0_0 -> 16 in the cache */
   /* replace1689_0_1 -> 4640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4640 },
      -1,
   } },
   /* replace1689_1_0_0 -> 16 in the cache */
   /* replace1689_1_0_1 -> 4642 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4642 },
      -1,
   } },
   /* replace1689_1_1_0_0 -> 16 in the cache */
   /* replace1689_1_1_0 -> 4556 in the cache */
   /* replace1689_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4556, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4661, 4662 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4660, 4663 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1690_0_0 -> 6 in the cache */
   /* search1690_0 -> 2105 in the cache */
   /* search1690_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2105, 4665 },
      -1,
   } },

   /* replace1690_0 -> 6 in the cache */
   /* replace1690_1_0 -> 2280 in the cache */
   /* replace1690_1 -> 4565 in the cache */
   /* replace1690 -> 4566 in the cache */

   /* ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1691_0_0 -> 6 in the cache */
   /* search1691_0 -> 2105 in the cache */
   /* search1691_1_0 -> 2280 in the cache */
   /* search1691_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2105, 4665 },
      -1,
   } },

   /* replace1691_0 -> 6 in the cache */
   /* replace1691_1_0 -> 2280 in the cache */
   /* replace1691_1 -> 4565 in the cache */
   /* replace1691 -> 4568 in the cache */

   /* ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1692_0_0 -> 6 in the cache */
   /* search1692_0 -> 2105 in the cache */
   /* search1692_1_0 -> 2280 in the cache */
   /* search1692_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2105, 4665 },
      -1,
   } },

   /* replace1692_0 -> 6 in the cache */
   /* replace1692_1_0 -> 2280 in the cache */
   /* replace1692_1 -> 4565 in the cache */
   /* replace1692 -> 4570 in the cache */

   /* ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1693_0_0 -> 6 in the cache */
   /* search1693_0 -> 2105 in the cache */
   /* search1693_1_0 -> 2280 in the cache */
   /* search1693_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2105, 4665 },
      -1,
   } },

   /* replace1693_0 -> 6 in the cache */
   /* replace1693_1_0 -> 2280 in the cache */
   /* replace1693_1 -> 4565 in the cache */
   /* replace1693 -> 4572 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1694_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* search1694_1_0 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4670, 4671 },
      -1,
   } },

   /* replace1694_0_0 -> 4573 in the cache */
   /* replace1694_0 -> 4578 in the cache */
   /* replace1694_1 -> 4575 in the cache */
   /* replace1694 -> 4579 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1695_0_0 -> 4573 in the cache */
   /* search1695_0 -> 4670 in the cache */
   /* search1695_1_0 -> 4575 in the cache */
   /* search1695_1 -> 4671 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4670, 4671 },
      -1,
   } },

   /* replace1695_0_0 -> 4573 in the cache */
   /* replace1695_0 -> 4578 in the cache */
   /* replace1695_1 -> 4575 in the cache */
   /* replace1695 -> 4581 in the cache */

   /* ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1696_0_0 -> 6 in the cache */
   /* search1696_0 -> 2105 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2105, 4674 },
      -1,
   } },

   /* replace1696_0_0 -> 6 in the cache */
   /* replace1696_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4676 },
      -1,
   } },
   /* replace1696_1_0_0_0 -> 41 in the cache */
   /* replace1696_1_0_0 -> 4676 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4676 },
      -1,
   } },
   /* replace1696_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4678, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4677, 4679 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1697_0_0 -> 6 in the cache */
   /* search1697_0 -> 2105 in the cache */
   /* search1697_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2105, 4674 },
      -1,
   } },

   /* replace1697_0_0 -> 6 in the cache */
   /* replace1697_0_1_0 -> 41 in the cache */
   /* replace1697_0_1 -> 4676 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4676 },
      -1,
   } },
   /* replace1697_1_0_0_0 -> 41 in the cache */
   /* replace1697_1_0_0 -> 4676 in the cache */
   /* replace1697_1_0 -> 4678 in the cache */
   /* replace1697_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4678, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4682, 4683 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1698_0_0 -> 6 in the cache */
   /* search1698_0 -> 2105 in the cache */
   /* search1698_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2105, 4674 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1698_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4686, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1698_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4688, 41 },
      -1,
   } },
   /* replace1698_1_1_0 -> 6 in the cache */
   /* replace1698_1_1_1_0 -> 41 in the cache */
   /* replace1698_1_1_1 -> 4676 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4676 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4689, 4690 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4687, 4691 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1699_1_0 -> 2280 in the cache */
   /* search1699_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4693, 4665 },
      -1,
   } },

   /* replace1699_0_0 -> 21 in the cache */
   /* replace1699_0_1 -> 4688 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4688 },
      -1,
   } },
   /* replace1699_1_0_0 -> 21 in the cache */
   /* replace1699_1_0_1 -> 4686 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4686 },
      -1,
   } },
   /* replace1699_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1699_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4697, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4696, 4698 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4695, 4699 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1700_0_0 -> 6 in the cache */
   /* search1700_0 -> 2105 in the cache */
   /* search1700_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2105, 4674 },
      -1,
   } },

   /* replace1700_0_0 -> 4688 in the cache */
   /* replace1700_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4688, 41 },
      -1,
   } },
   /* replace1700_1_0_0 -> 4686 in the cache */
   /* replace1700_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4686, 41 },
      -1,
   } },
   /* replace1700_1_1_0 -> 6 in the cache */
   /* replace1700_1_1_1_0 -> 41 in the cache */
   /* replace1700_1_1_1 -> 4676 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4676 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4703, 4704 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4702, 4705 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1701_0 -> 4693 in the cache */
   /* search1701_1_0 -> 2280 in the cache */
   /* search1701_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4693, 4665 },
      -1,
   } },

   /* replace1701_0_0 -> 21 in the cache */
   /* replace1701_0_1 -> 4686 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4686 },
      -1,
   } },
   /* replace1701_1_0_0 -> 21 in the cache */
   /* replace1701_1_0_1 -> 4688 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4688 },
      -1,
   } },
   /* replace1701_1_1_0_0 -> 21 in the cache */
   /* replace1701_1_1_0 -> 4697 in the cache */
   /* replace1701_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4697, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4709, 4710 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4708, 4711 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1702_0_0 -> 11 in the cache */
   /* search1702_0 -> 2127 in the cache */
   /* search1702_1_0 -> 2280 in the cache */
   /* search1702_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2127, 4620 },
      -1,
   } },

   /* replace1702_0 -> 11 in the cache */
   /* replace1702_1_0 -> 2280 in the cache */
   /* replace1702_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4563 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1703_0_0 -> 11 in the cache */
   /* search1703_0 -> 2127 in the cache */
   /* search1703_1_0 -> 2280 in the cache */
   /* search1703_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2127, 4620 },
      -1,
   } },

   /* replace1703_0 -> 11 in the cache */
   /* replace1703_1_0 -> 2280 in the cache */
   /* replace1703_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4563 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1704_0_0 -> 11 in the cache */
   /* search1704_0 -> 2127 in the cache */
   /* search1704_1_0 -> 2280 in the cache */
   /* search1704_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2127, 4620 },
      -1,
   } },

   /* replace1704_0 -> 11 in the cache */
   /* replace1704_1_0 -> 2280 in the cache */
   /* replace1704_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4563 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1705_0_0 -> 11 in the cache */
   /* search1705_0 -> 2127 in the cache */
   /* search1705_1_0 -> 2280 in the cache */
   /* search1705_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2127, 4620 },
      -1,
   } },

   /* replace1705_0 -> 11 in the cache */
   /* replace1705_1_0 -> 2280 in the cache */
   /* replace1705_1 -> 4563 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4563 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1706_0_0 -> 4573 in the cache */
   /* search1706_0 -> 4625 in the cache */
   /* search1706_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4625, 4721 },
      -1,
   } },

   /* replace1706_0_0 -> 4573 in the cache */
   /* replace1706_0 -> 4574 in the cache */
   /* replace1706_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4574, 536 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1707_0_0 -> 4573 in the cache */
   /* search1707_0 -> 4625 in the cache */
   /* search1707_1_0 -> 536 in the cache */
   /* search1707_1 -> 4721 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4625, 4721 },
      -1,
   } },

   /* replace1707_0_0 -> 4573 in the cache */
   /* replace1707_0 -> 4574 in the cache */
   /* replace1707_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4574, 536 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1708_0_0 -> 11 in the cache */
   /* search1708_0 -> 2127 in the cache */
   /* search1708_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2127, 4726 },
      -1,
   } },

   /* replace1708_0 -> 11 in the cache */
   /* replace1708_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4728 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1709_0_0 -> 11 in the cache */
   /* search1709_0 -> 2127 in the cache */
   /* search1709_1_0 -> 268 in the cache */
   /* search1709_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2127, 4726 },
      -1,
   } },

   /* replace1709_0 -> 11 in the cache */
   /* replace1709_1_0 -> 268 in the cache */
   /* replace1709_1 -> 4728 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4728 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1710_0_0 -> 11 in the cache */
   /* search1710_0 -> 2127 in the cache */
   /* search1710_1_0 -> 268 in the cache */
   /* search1710_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2127, 4726 },
      -1,
   } },

   /* replace1710_0 -> 11 in the cache */
   /* replace1710_1_0 -> 268 in the cache */
   /* replace1710_1 -> 4728 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4728 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1711_0_0 -> 11 in the cache */
   /* search1711_0 -> 2127 in the cache */
   /* search1711_1_0 -> 268 in the cache */
   /* search1711_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2127, 4726 },
      -1,
   } },

   /* replace1711_0 -> 11 in the cache */
   /* replace1711_1_0 -> 268 in the cache */
   /* replace1711_1 -> 4728 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4728 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1712_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1712_1_0 -> 536 in the cache */
   /* search1712_1 -> 4721 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4736, 4721 },
      -1,
   } },

   /* replace1712_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 525 },
      -1,
   } },
   /* replace1712_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4738, 536 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1713_0_0 -> 525 in the cache */
   /* search1713_0 -> 4736 in the cache */
   /* search1713_1_0 -> 536 in the cache */
   /* search1713_1 -> 4721 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4736, 4721 },
      -1,
   } },

   /* replace1713_0_0 -> 525 in the cache */
   /* replace1713_0 -> 4738 in the cache */
   /* replace1713_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4738, 536 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1714_0_0 -> 11 in the cache */
   /* search1714_0 -> 2127 in the cache */
   /* search1714_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2127, 33 },
      -1,
   } },

   /* replace1714_0_0 -> 11 in the cache */
   /* replace1714_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4743 },
      -1,
   } },
   /* replace1714_1_0_0_0 -> 36 in the cache */
   /* replace1714_1_0_0 -> 4743 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4743 },
      -1,
   } },
   /* replace1714_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4745, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4744, 4746 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1715_0_0 -> 11 in the cache */
   /* search1715_0 -> 2127 in the cache */
   /* search1715_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2127, 33 },
      -1,
   } },

   /* replace1715_0_0 -> 11 in the cache */
   /* replace1715_0_1_0 -> 36 in the cache */
   /* replace1715_0_1 -> 4743 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4743 },
      -1,
   } },
   /* replace1715_1_0_0_0 -> 36 in the cache */
   /* replace1715_1_0_0 -> 4743 in the cache */
   /* replace1715_1_0 -> 4745 in the cache */
   /* replace1715_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4745, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4749, 4750 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1716_0_0 -> 11 in the cache */
   /* search1716_0 -> 2127 in the cache */
   /* search1716_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2127, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1716_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4753, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1716_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4755, 36 },
      -1,
   } },
   /* replace1716_1_1_0 -> 11 in the cache */
   /* replace1716_1_1_1_0 -> 36 in the cache */
   /* replace1716_1_1_1 -> 4743 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4743 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4756, 4757 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4754, 4758 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1717_0 -> 296 in the cache */
   /* search1717_1_0 -> 268 in the cache */
   /* search1717_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 296, 4726 },
      -1,
   } },

   /* replace1717_0_0 -> 16 in the cache */
   /* replace1717_0_1 -> 4755 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4755 },
      -1,
   } },
   /* replace1717_1_0_0 -> 16 in the cache */
   /* replace1717_1_0_1 -> 4753 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4753 },
      -1,
   } },
   /* replace1717_1_1_0_0 -> 16 in the cache */
   /* replace1717_1_1_0 -> 4558 in the cache */
   /* replace1717_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4558, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4762, 4763 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4761, 4764 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1718_0_0 -> 11 in the cache */
   /* search1718_0 -> 2127 in the cache */
   /* search1718_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2127, 33 },
      -1,
   } },

   /* replace1718_0_0 -> 4755 in the cache */
   /* replace1718_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4755, 36 },
      -1,
   } },
   /* replace1718_1_0_0 -> 4753 in the cache */
   /* replace1718_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4753, 36 },
      -1,
   } },
   /* replace1718_1_1_0 -> 11 in the cache */
   /* replace1718_1_1_1_0 -> 36 in the cache */
   /* replace1718_1_1_1 -> 4743 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4743 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4768, 4769 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4767, 4770 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1719_0 -> 296 in the cache */
   /* search1719_1_0 -> 268 in the cache */
   /* search1719_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 296, 4726 },
      -1,
   } },

   /* replace1719_0_0 -> 16 in the cache */
   /* replace1719_0_1 -> 4753 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4753 },
      -1,
   } },
   /* replace1719_1_0_0 -> 16 in the cache */
   /* replace1719_1_0_1 -> 4755 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4755 },
      -1,
   } },
   /* replace1719_1_1_0_0 -> 16 in the cache */
   /* replace1719_1_1_0 -> 4558 in the cache */
   /* replace1719_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4558, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4774, 4775 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4773, 4776 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1720_0_0 -> 11 in the cache */
   /* search1720_0 -> 2137 in the cache */
   /* search1720_1_0 -> 2280 in the cache */
   /* search1720_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2137, 4665 },
      -1,
   } },

   /* replace1720_0 -> 11 in the cache */
   /* replace1720_1_0 -> 2280 in the cache */
   /* replace1720_1 -> 4563 in the cache */
   /* replace1720 -> 4714 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1721_0_0 -> 11 in the cache */
   /* search1721_0 -> 2137 in the cache */
   /* search1721_1_0 -> 2280 in the cache */
   /* search1721_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2137, 4665 },
      -1,
   } },

   /* replace1721_0 -> 11 in the cache */
   /* replace1721_1_0 -> 2280 in the cache */
   /* replace1721_1 -> 4563 in the cache */
   /* replace1721 -> 4716 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1722_0_0 -> 11 in the cache */
   /* search1722_0 -> 2137 in the cache */
   /* search1722_1_0 -> 2280 in the cache */
   /* search1722_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2137, 4665 },
      -1,
   } },

   /* replace1722_0 -> 11 in the cache */
   /* replace1722_1_0 -> 2280 in the cache */
   /* replace1722_1 -> 4563 in the cache */
   /* replace1722 -> 4718 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1723_0_0 -> 11 in the cache */
   /* search1723_0 -> 2137 in the cache */
   /* search1723_1_0 -> 2280 in the cache */
   /* search1723_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2137, 4665 },
      -1,
   } },

   /* replace1723_0 -> 11 in the cache */
   /* replace1723_1_0 -> 2280 in the cache */
   /* replace1723_1 -> 4563 in the cache */
   /* replace1723 -> 4720 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1724_0_0 -> 4573 in the cache */
   /* search1724_0 -> 4670 in the cache */
   /* search1724_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4670, 4782 },
      -1,
   } },

   /* replace1724_0_0 -> 4573 in the cache */
   /* replace1724_0 -> 4574 in the cache */
   /* replace1724_1 -> 536 in the cache */
   /* replace1724 -> 4723 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1725_0_0 -> 4573 in the cache */
   /* search1725_0 -> 4670 in the cache */
   /* search1725_1_0 -> 536 in the cache */
   /* search1725_1 -> 4782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4670, 4782 },
      -1,
   } },

   /* replace1725_0_0 -> 4573 in the cache */
   /* replace1725_0 -> 4574 in the cache */
   /* replace1725_1 -> 536 in the cache */
   /* replace1725 -> 4725 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1726_0_0 -> 11 in the cache */
   /* search1726_0 -> 2137 in the cache */
   /* search1726_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2137, 4785 },
      -1,
   } },

   /* replace1726_0 -> 11 in the cache */
   /* replace1726_1_0 -> 268 in the cache */
   /* replace1726_1 -> 4728 in the cache */
   /* replace1726 -> 4729 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1727_0_0 -> 11 in the cache */
   /* search1727_0 -> 2137 in the cache */
   /* search1727_1_0 -> 268 in the cache */
   /* search1727_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2137, 4785 },
      -1,
   } },

   /* replace1727_0 -> 11 in the cache */
   /* replace1727_1_0 -> 268 in the cache */
   /* replace1727_1 -> 4728 in the cache */
   /* replace1727 -> 4731 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1728_0_0 -> 11 in the cache */
   /* search1728_0 -> 2137 in the cache */
   /* search1728_1_0 -> 268 in the cache */
   /* search1728_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2137, 4785 },
      -1,
   } },

   /* replace1728_0 -> 11 in the cache */
   /* replace1728_1_0 -> 268 in the cache */
   /* replace1728_1 -> 4728 in the cache */
   /* replace1728 -> 4733 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1729_0_0 -> 11 in the cache */
   /* search1729_0 -> 2137 in the cache */
   /* search1729_1_0 -> 268 in the cache */
   /* search1729_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2137, 4785 },
      -1,
   } },

   /* replace1729_0 -> 11 in the cache */
   /* replace1729_1_0 -> 268 in the cache */
   /* replace1729_1 -> 4728 in the cache */
   /* replace1729 -> 4735 in the cache */

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1730_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1730_1_0 -> 536 in the cache */
   /* search1730_1 -> 4782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4790, 4782 },
      -1,
   } },

   /* replace1730_0_0 -> 525 in the cache */
   /* replace1730_0 -> 4738 in the cache */
   /* replace1730_1 -> 536 in the cache */
   /* replace1730 -> 4739 in the cache */

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1731_0_0 -> 525 in the cache */
   /* search1731_0 -> 4790 in the cache */
   /* search1731_1_0 -> 536 in the cache */
   /* search1731_1 -> 4782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4790, 4782 },
      -1,
   } },

   /* replace1731_0_0 -> 525 in the cache */
   /* replace1731_0 -> 4738 in the cache */
   /* replace1731_1 -> 536 in the cache */
   /* replace1731 -> 4741 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1732_0_0 -> 11 in the cache */
   /* search1732_0 -> 2137 in the cache */
   /* search1732_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2137, 4674 },
      -1,
   } },

   /* replace1732_0_0 -> 11 in the cache */
   /* replace1732_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4794 },
      -1,
   } },
   /* replace1732_1_0_0_0 -> 41 in the cache */
   /* replace1732_1_0_0 -> 4794 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4794 },
      -1,
   } },
   /* replace1732_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4796, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4795, 4797 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1733_0_0 -> 11 in the cache */
   /* search1733_0 -> 2137 in the cache */
   /* search1733_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2137, 4674 },
      -1,
   } },

   /* replace1733_0_0 -> 11 in the cache */
   /* replace1733_0_1_0 -> 41 in the cache */
   /* replace1733_0_1 -> 4794 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4794 },
      -1,
   } },
   /* replace1733_1_0_0_0 -> 41 in the cache */
   /* replace1733_1_0_0 -> 4794 in the cache */
   /* replace1733_1_0 -> 4796 in the cache */
   /* replace1733_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4796, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4800, 4801 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1734_0_0 -> 11 in the cache */
   /* search1734_0 -> 2137 in the cache */
   /* search1734_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2137, 4674 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1734_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4804, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1734_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4806, 41 },
      -1,
   } },
   /* replace1734_1_1_0 -> 11 in the cache */
   /* replace1734_1_1_1_0 -> 41 in the cache */
   /* replace1734_1_1_1 -> 4794 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4794 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4807, 4808 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4805, 4809 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1735_0 -> 4693 in the cache */
   /* search1735_1_0 -> 268 in the cache */
   /* search1735_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4693, 4785 },
      -1,
   } },

   /* replace1735_0_0 -> 21 in the cache */
   /* replace1735_0_1 -> 4806 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4806 },
      -1,
   } },
   /* replace1735_1_0_0 -> 21 in the cache */
   /* replace1735_1_0_1 -> 4804 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4804 },
      -1,
   } },
   /* replace1735_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1735_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4814, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4813, 4815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4812, 4816 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1736_0_0 -> 11 in the cache */
   /* search1736_0 -> 2137 in the cache */
   /* search1736_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2137, 4674 },
      -1,
   } },

   /* replace1736_0_0 -> 4806 in the cache */
   /* replace1736_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4806, 41 },
      -1,
   } },
   /* replace1736_1_0_0 -> 4804 in the cache */
   /* replace1736_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4804, 41 },
      -1,
   } },
   /* replace1736_1_1_0 -> 11 in the cache */
   /* replace1736_1_1_1_0 -> 41 in the cache */
   /* replace1736_1_1_1 -> 4794 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4794 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4820, 4821 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4819, 4822 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1737_0 -> 4693 in the cache */
   /* search1737_1_0 -> 268 in the cache */
   /* search1737_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4693, 4785 },
      -1,
   } },

   /* replace1737_0_0 -> 21 in the cache */
   /* replace1737_0_1 -> 4804 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4804 },
      -1,
   } },
   /* replace1737_1_0_0 -> 21 in the cache */
   /* replace1737_1_0_1 -> 4806 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4806 },
      -1,
   } },
   /* replace1737_1_1_0_0 -> 21 in the cache */
   /* replace1737_1_1_0 -> 4814 in the cache */
   /* replace1737_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4814, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4826, 4827 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4825, 4828 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1738_0_0 -> 16 in the cache */
   /* search1738_0 -> 2147 in the cache */
   /* search1738_1_0 -> 2280 in the cache */
   /* search1738_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2147, 4665 },
      -1,
   } },

   /* replace1738_0 -> 16 in the cache */
   /* replace1738_1_0 -> 2280 in the cache */
   /* replace1738_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4620 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1739_0_0 -> 16 in the cache */
   /* search1739_0 -> 2147 in the cache */
   /* search1739_1_0 -> 2280 in the cache */
   /* search1739_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2147, 4665 },
      -1,
   } },

   /* replace1739_0 -> 16 in the cache */
   /* replace1739_1_0 -> 2280 in the cache */
   /* replace1739_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4620 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1740_0_0 -> 16 in the cache */
   /* search1740_0 -> 2147 in the cache */
   /* search1740_1_0 -> 2280 in the cache */
   /* search1740_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2147, 4665 },
      -1,
   } },

   /* replace1740_0 -> 16 in the cache */
   /* replace1740_1_0 -> 2280 in the cache */
   /* replace1740_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4620 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1741_0_0 -> 16 in the cache */
   /* search1741_0 -> 2147 in the cache */
   /* search1741_1_0 -> 2280 in the cache */
   /* search1741_1 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2147, 4665 },
      -1,
   } },

   /* replace1741_0 -> 16 in the cache */
   /* replace1741_1_0 -> 2280 in the cache */
   /* replace1741_1 -> 4620 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4620 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1742_0_0 -> 4573 in the cache */
   /* search1742_0 -> 4670 in the cache */
   /* search1742_1_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4670, 4838 },
      -1,
   } },

   /* replace1742_0_0 -> 4573 in the cache */
   /* replace1742_0 -> 4625 in the cache */
   /* replace1742_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4625, 189 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1743_0_0 -> 4573 in the cache */
   /* search1743_0 -> 4670 in the cache */
   /* search1743_1_0 -> 189 in the cache */
   /* search1743_1 -> 4838 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4670, 4838 },
      -1,
   } },

   /* replace1743_0_0 -> 4573 in the cache */
   /* replace1743_0 -> 4625 in the cache */
   /* replace1743_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4625, 189 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1744_0_0 -> 16 in the cache */
   /* search1744_0 -> 2147 in the cache */
   /* search1744_1_0 -> 268 in the cache */
   /* search1744_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2147, 4785 },
      -1,
   } },

   /* replace1744_0 -> 16 in the cache */
   /* replace1744_1_0 -> 268 in the cache */
   /* replace1744_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4726 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1745_0_0 -> 16 in the cache */
   /* search1745_0 -> 2147 in the cache */
   /* search1745_1_0 -> 268 in the cache */
   /* search1745_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2147, 4785 },
      -1,
   } },

   /* replace1745_0 -> 16 in the cache */
   /* replace1745_1_0 -> 268 in the cache */
   /* replace1745_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4726 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1746_0_0 -> 16 in the cache */
   /* search1746_0 -> 2147 in the cache */
   /* search1746_1_0 -> 268 in the cache */
   /* search1746_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2147, 4785 },
      -1,
   } },

   /* replace1746_0 -> 16 in the cache */
   /* replace1746_1_0 -> 268 in the cache */
   /* replace1746_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4726 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1747_0_0 -> 16 in the cache */
   /* search1747_0 -> 2147 in the cache */
   /* search1747_1_0 -> 268 in the cache */
   /* search1747_1 -> 4785 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2147, 4785 },
      -1,
   } },

   /* replace1747_0 -> 16 in the cache */
   /* replace1747_1_0 -> 268 in the cache */
   /* replace1747_1 -> 4726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4726 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1748_0_0 -> 525 in the cache */
   /* search1748_0 -> 4790 in the cache */
   /* search1748_1_0 -> 189 in the cache */
   /* search1748_1 -> 4838 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4790, 4838 },
      -1,
   } },

   /* replace1748_0_0 -> 525 in the cache */
   /* replace1748_0 -> 4736 in the cache */
   /* replace1748_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4736, 189 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1749_0_0 -> 525 in the cache */
   /* search1749_0 -> 4790 in the cache */
   /* search1749_1_0 -> 189 in the cache */
   /* search1749_1 -> 4838 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4790, 4838 },
      -1,
   } },

   /* replace1749_0_0 -> 525 in the cache */
   /* replace1749_0 -> 4736 in the cache */
   /* replace1749_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4736, 189 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1750_0_0 -> 16 in the cache */
   /* search1750_0 -> 2147 in the cache */
   /* search1750_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2147, 4855 },
      -1,
   } },

   /* replace1750_0 -> 16 in the cache */
   /* replace1750_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4857 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1751_0_0 -> 16 in the cache */
   /* search1751_0 -> 2147 in the cache */
   /* search1751_1_0 -> 36 in the cache */
   /* search1751_1 -> 4855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2147, 4855 },
      -1,
   } },

   /* replace1751_0 -> 16 in the cache */
   /* replace1751_1_0 -> 36 in the cache */
   /* replace1751_1 -> 4857 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4857 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1752_0_0 -> 16 in the cache */
   /* search1752_0 -> 2147 in the cache */
   /* search1752_1_0 -> 36 in the cache */
   /* search1752_1 -> 4855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2147, 4855 },
      -1,
   } },

   /* replace1752_0 -> 16 in the cache */
   /* replace1752_1_0 -> 36 in the cache */
   /* replace1752_1 -> 4857 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4857 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1753_0_0 -> 16 in the cache */
   /* search1753_0 -> 2147 in the cache */
   /* search1753_1_0 -> 36 in the cache */
   /* search1753_1 -> 4855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2147, 4855 },
      -1,
   } },

   /* replace1753_0 -> 16 in the cache */
   /* replace1753_1_0 -> 36 in the cache */
   /* replace1753_1 -> 4857 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4857 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1754_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 383 },
      -1,
   } },
   /* search1754_1_0 -> 189 in the cache */
   /* search1754_1 -> 4838 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4865, 4838 },
      -1,
   } },

   /* replace1754_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 383 },
      -1,
   } },
   /* replace1754_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4867, 189 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1755_0_0 -> 383 in the cache */
   /* search1755_0 -> 4865 in the cache */
   /* search1755_1_0 -> 189 in the cache */
   /* search1755_1 -> 4838 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4865, 4838 },
      -1,
   } },

   /* replace1755_0_0 -> 383 in the cache */
   /* replace1755_0 -> 4867 in the cache */
   /* replace1755_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4867, 189 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1756_0_0 -> 16 in the cache */
   /* search1756_0 -> 2147 in the cache */
   /* search1756_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2147, 4674 },
      -1,
   } },

   /* replace1756_0_0 -> 16 in the cache */
   /* replace1756_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 4872 },
      -1,
   } },
   /* replace1756_1_0_0_0 -> 41 in the cache */
   /* replace1756_1_0_0 -> 4872 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4872 },
      -1,
   } },
   /* replace1756_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4874, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4873, 4875 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1757_0_0 -> 16 in the cache */
   /* search1757_0 -> 2147 in the cache */
   /* search1757_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2147, 4674 },
      -1,
   } },

   /* replace1757_0_0 -> 16 in the cache */
   /* replace1757_0_1_0 -> 41 in the cache */
   /* replace1757_0_1 -> 4872 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 4872 },
      -1,
   } },
   /* replace1757_1_0_0_0 -> 41 in the cache */
   /* replace1757_1_0_0 -> 4872 in the cache */
   /* replace1757_1_0 -> 4874 in the cache */
   /* replace1757_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4874, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4878, 4879 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b')))) */
   /* search1758_0_0 -> 16 in the cache */
   /* search1758_0 -> 2147 in the cache */
   /* search1758_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2147, 4674 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80000000 /* -2147483648 */ },
   } },
   /* replace1758_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4882, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   /* replace1758_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4884, 41 },
      -1,
   } },
   /* replace1758_1_1_0 -> 16 in the cache */
   /* replace1758_1_1_1_0 -> 41 in the cache */
   /* replace1758_1_1_1 -> 4872 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4872 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4885, 4886 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4883, 4887 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b'))) */
   /* search1759_0 -> 4693 in the cache */
   /* search1759_1_0 -> 36 in the cache */
   /* search1759_1 -> 4855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4693, 4855 },
      -1,
   } },

   /* replace1759_0_0 -> 21 in the cache */
   /* replace1759_0_1 -> 4884 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4884 },
      -1,
   } },
   /* replace1759_1_0_0 -> 21 in the cache */
   /* replace1759_1_0_1 -> 4882 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4882 },
      -1,
   } },
   /* replace1759_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1759_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4892, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4891, 4893 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4890, 4894 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b')))) */
   /* search1760_0_0 -> 16 in the cache */
   /* search1760_0 -> 2147 in the cache */
   /* search1760_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2147, 4674 },
      -1,
   } },

   /* replace1760_0_0 -> 4884 in the cache */
   /* replace1760_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4884, 41 },
      -1,
   } },
   /* replace1760_1_0_0 -> 4882 in the cache */
   /* replace1760_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4882, 41 },
      -1,
   } },
   /* replace1760_1_1_0 -> 16 in the cache */
   /* replace1760_1_1_1_0 -> 41 in the cache */
   /* replace1760_1_1_1 -> 4872 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4872 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4898, 4899 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4897, 4900 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b'))) */
   /* search1761_0 -> 4693 in the cache */
   /* search1761_1_0 -> 36 in the cache */
   /* search1761_1 -> 4855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4693, 4855 },
      -1,
   } },

   /* replace1761_0_0 -> 21 in the cache */
   /* replace1761_0_1 -> 4882 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4882 },
      -1,
   } },
   /* replace1761_1_0_0 -> 21 in the cache */
   /* replace1761_1_0_1 -> 4884 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4884 },
      -1,
   } },
   /* replace1761_1_1_0_0 -> 21 in the cache */
   /* replace1761_1_1_0 -> 4892 in the cache */
   /* replace1761_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4892, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4904, 4905 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4903, 4906 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1762_0_0 -> 6 in the cache */
   /* search1762_0 -> 2093 in the cache */
   /* search1762_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2093, 4908 },
      -1,
   } },

   /* replace1762_0 -> 6 in the cache */
   /* replace1762_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 4910 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1763_0_0 -> 6 in the cache */
   /* search1763_0 -> 2093 in the cache */
   /* search1763_1_0 -> 2280 in the cache */
   /* search1763_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2093, 4908 },
      -1,
   } },

   /* replace1763_0 -> 6 in the cache */
   /* replace1763_1_0 -> 2280 in the cache */
   /* replace1763_1 -> 4910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 4910 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1764_0_0 -> 6 in the cache */
   /* search1764_0 -> 2093 in the cache */
   /* search1764_1_0 -> 2280 in the cache */
   /* search1764_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2093, 4908 },
      -1,
   } },

   /* replace1764_0 -> 6 in the cache */
   /* replace1764_1_0 -> 2280 in the cache */
   /* replace1764_1 -> 4910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 4910 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1765_0_0 -> 6 in the cache */
   /* search1765_0 -> 2093 in the cache */
   /* search1765_1_0 -> 2280 in the cache */
   /* search1765_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2093, 4908 },
      -1,
   } },

   /* replace1765_0 -> 6 in the cache */
   /* replace1765_1_0 -> 2280 in the cache */
   /* replace1765_1 -> 4910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 4910 },
      -1,
   } },

   /* ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1766_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* search1766_1_0 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4918, 4919 },
      -1,
   } },

   /* replace1766_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* replace1766_1 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4921, 4575 },
      -1,
   } },

   /* ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1767_0_0 -> 4573 in the cache */
   /* search1767_0 -> 4918 in the cache */
   /* search1767_1_0 -> 4575 in the cache */
   /* search1767_1 -> 4919 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4918, 4919 },
      -1,
   } },

   /* replace1767_0_0 -> 4573 in the cache */
   /* replace1767_0 -> 4921 in the cache */
   /* replace1767_1 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4921, 4575 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1768_0_0 -> 6 in the cache */
   /* search1768_0 -> 2093 in the cache */
   /* search1768_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2093, 2179 },
      -1,
   } },

   /* replace1768_0_0 -> 6 in the cache */
   /* replace1768_0_1_0 -> 268 in the cache */
   /* replace1768_0_1 -> 2187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 2187 },
      -1,
   } },
   /* replace1768_1_0_0_0 -> 268 in the cache */
   /* replace1768_1_0_0 -> 2187 in the cache */
   /* replace1768_1_0 -> 2188 in the cache */
   /* replace1768_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2188, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4926, 4927 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1769_0_0 -> 6 in the cache */
   /* search1769_0 -> 2093 in the cache */
   /* search1769_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2093, 2179 },
      -1,
   } },

   /* replace1769_0_0 -> 6 in the cache */
   /* replace1769_0_1_0 -> 268 in the cache */
   /* replace1769_0_1 -> 2187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 2187 },
      -1,
   } },
   /* replace1769_1_0_0_0 -> 268 in the cache */
   /* replace1769_1_0_0 -> 2187 in the cache */
   /* replace1769_1_0 -> 2188 in the cache */
   /* replace1769_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 2188, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4930, 4931 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1770_0_0 -> 6 in the cache */
   /* search1770_0 -> 2093 in the cache */
   /* search1770_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2093, 2179 },
      -1,
   } },

   /* replace1770_0_0 -> 869 in the cache */
   /* replace1770_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 869, 268 },
      -1,
   } },
   /* replace1770_1_0_0 -> 2181 in the cache */
   /* replace1770_1_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2181, 268 },
      -1,
   } },
   /* replace1770_1_1_0 -> 6 in the cache */
   /* replace1770_1_1_1_0 -> 268 in the cache */
   /* replace1770_1_1_1 -> 2187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2187 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4935, 4936 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4934, 4937 },
      -1,
   } },

   /* ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1771_0 -> 4600 in the cache */
   /* search1771_1_0 -> 2280 in the cache */
   /* search1771_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4600, 4908 },
      -1,
   } },

   /* replace1771_0_0 -> 11 in the cache */
   /* replace1771_0_1 -> 2181 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 2181 },
      -1,
   } },
   /* replace1771_1_0_0 -> 11 in the cache */
   /* replace1771_1_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 869 },
      -1,
   } },
   /* replace1771_1_1_0_0 -> 11 in the cache */
   /* replace1771_1_1_0 -> 2177 in the cache */
   /* replace1771_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2177, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4941, 4942 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4940, 4943 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1772_0_0 -> 6 in the cache */
   /* search1772_0 -> 2093 in the cache */
   /* search1772_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2093, 2179 },
      -1,
   } },

   /* replace1772_0_0 -> 2181 in the cache */
   /* replace1772_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2181, 268 },
      -1,
   } },
   /* replace1772_1_0_0 -> 869 in the cache */
   /* replace1772_1_0_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 869, 268 },
      -1,
   } },
   /* replace1772_1_1_0 -> 6 in the cache */
   /* replace1772_1_1_1_0 -> 268 in the cache */
   /* replace1772_1_1_1 -> 2187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2187 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4947, 4948 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4946, 4949 },
      -1,
   } },

   /* ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1773_0 -> 4600 in the cache */
   /* search1773_1_0 -> 2280 in the cache */
   /* search1773_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4600, 4908 },
      -1,
   } },

   /* replace1773_0_0 -> 11 in the cache */
   /* replace1773_0_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 869 },
      -1,
   } },
   /* replace1773_1_0_0 -> 11 in the cache */
   /* replace1773_1_0_1 -> 2181 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 2181 },
      -1,
   } },
   /* replace1773_1_1_0_0 -> 11 in the cache */
   /* replace1773_1_1_0 -> 2177 in the cache */
   /* replace1773_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2177, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4953, 4954 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4952, 4955 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1774_0_0 -> 6 in the cache */
   /* search1774_0 -> 2099 in the cache */
   /* search1774_1_0 -> 2280 in the cache */
   /* search1774_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2099, 3478 },
      -1,
   } },

   /* replace1774_0 -> 6 in the cache */
   /* replace1774_1_0 -> 2280 in the cache */
   /* replace1774_1 -> 4910 in the cache */
   /* replace1774 -> 4911 in the cache */

   /* ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1775_0_0 -> 6 in the cache */
   /* search1775_0 -> 2099 in the cache */
   /* search1775_1_0 -> 2280 in the cache */
   /* search1775_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2099, 3478 },
      -1,
   } },

   /* replace1775_0 -> 6 in the cache */
   /* replace1775_1_0 -> 2280 in the cache */
   /* replace1775_1 -> 4910 in the cache */
   /* replace1775 -> 4913 in the cache */

   /* ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1776_0_0 -> 6 in the cache */
   /* search1776_0 -> 2099 in the cache */
   /* search1776_1_0 -> 2280 in the cache */
   /* search1776_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2099, 3478 },
      -1,
   } },

   /* replace1776_0 -> 6 in the cache */
   /* replace1776_1_0 -> 2280 in the cache */
   /* replace1776_1 -> 4910 in the cache */
   /* replace1776 -> 4915 in the cache */

   /* ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1777_0_0 -> 6 in the cache */
   /* search1777_0 -> 2099 in the cache */
   /* search1777_1_0 -> 2280 in the cache */
   /* search1777_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2099, 3478 },
      -1,
   } },

   /* replace1777_0 -> 6 in the cache */
   /* replace1777_1_0 -> 2280 in the cache */
   /* replace1777_1 -> 4910 in the cache */
   /* replace1777 -> 4917 in the cache */

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1778_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* search1778_1_0 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4961, 4962 },
      -1,
   } },

   /* replace1778_0_0 -> 4573 in the cache */
   /* replace1778_0 -> 4921 in the cache */
   /* replace1778_1 -> 4575 in the cache */
   /* replace1778 -> 4922 in the cache */

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1779_0_0 -> 4573 in the cache */
   /* search1779_0 -> 4961 in the cache */
   /* search1779_1_0 -> 4575 in the cache */
   /* search1779_1 -> 4962 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4961, 4962 },
      -1,
   } },

   /* replace1779_0_0 -> 4573 in the cache */
   /* replace1779_0 -> 4921 in the cache */
   /* replace1779_1 -> 4575 in the cache */
   /* replace1779 -> 4924 in the cache */

   /* ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1780_0_0 -> 6 in the cache */
   /* search1780_0 -> 2099 in the cache */
   /* search1780_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2099, 33 },
      -1,
   } },

   /* replace1780_0_0 -> 6 in the cache */
   /* replace1780_0_1_0 -> 36 in the cache */
   /* replace1780_0_1 -> 2195 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 2195 },
      -1,
   } },
   /* replace1780_1_0_0_0 -> 36 in the cache */
   /* replace1780_1_0_0 -> 2195 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2195 },
      -1,
   } },
   /* replace1780_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4967, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4966, 4968 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1781_0_0 -> 6 in the cache */
   /* search1781_0 -> 2099 in the cache */
   /* search1781_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2099, 33 },
      -1,
   } },

   /* replace1781_0_0 -> 6 in the cache */
   /* replace1781_0_1_0 -> 36 in the cache */
   /* replace1781_0_1 -> 2195 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 2195 },
      -1,
   } },
   /* replace1781_1_0_0_0 -> 36 in the cache */
   /* replace1781_1_0_0 -> 2195 in the cache */
   /* replace1781_1_0 -> 4967 in the cache */
   /* replace1781_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4967, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4971, 4972 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1782_0_0 -> 6 in the cache */
   /* search1782_0 -> 2099 in the cache */
   /* search1782_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2099, 33 },
      -1,
   } },

   /* replace1782_0_0 -> 128 in the cache */
   /* replace1782_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 128, 36 },
      -1,
   } },
   /* replace1782_1_0_0 -> 1023 in the cache */
   /* replace1782_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1023, 36 },
      -1,
   } },
   /* replace1782_1_1_0 -> 6 in the cache */
   /* replace1782_1_1_1_0 -> 36 in the cache */
   /* replace1782_1_1_1 -> 2195 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2195 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4976, 4977 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4975, 4978 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1783_0 -> 296 in the cache */
   /* search1783_1_0 -> 2280 in the cache */
   /* search1783_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 296, 3478 },
      -1,
   } },

   /* replace1783_0_0 -> 16 in the cache */
   /* replace1783_0_1 -> 1023 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 1023 },
      -1,
   } },
   /* replace1783_1_0_0 -> 16 in the cache */
   /* replace1783_1_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace1783_1_1_0_0 -> 16 in the cache */
   /* replace1783_1_1_0 -> 2193 in the cache */
   /* replace1783_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2193, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4982, 4983 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4981, 4984 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1784_0_0 -> 6 in the cache */
   /* search1784_0 -> 2099 in the cache */
   /* search1784_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2099, 33 },
      -1,
   } },

   /* replace1784_0_0 -> 1023 in the cache */
   /* replace1784_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1023, 36 },
      -1,
   } },
   /* replace1784_1_0_0 -> 128 in the cache */
   /* replace1784_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 128, 36 },
      -1,
   } },
   /* replace1784_1_1_0 -> 6 in the cache */
   /* replace1784_1_1_1_0 -> 36 in the cache */
   /* replace1784_1_1_1 -> 2195 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2195 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4988, 4989 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4987, 4990 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1785_0 -> 296 in the cache */
   /* search1785_1_0 -> 2280 in the cache */
   /* search1785_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 296, 3478 },
      -1,
   } },

   /* replace1785_0_0 -> 16 in the cache */
   /* replace1785_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace1785_1_0_0 -> 16 in the cache */
   /* replace1785_1_0_1 -> 1023 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 1023 },
      -1,
   } },
   /* replace1785_1_1_0_0 -> 16 in the cache */
   /* replace1785_1_1_0 -> 2193 in the cache */
   /* replace1785_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2193, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4994, 4995 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4993, 4996 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1786_0_0 -> 6 in the cache */
   /* search1786_0 -> 2103 in the cache */
   /* search1786_1_0 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2103, 4998 },
      -1,
   } },

   /* replace1786_0 -> 6 in the cache */
   /* replace1786_1_0 -> 2280 in the cache */
   /* replace1786_1 -> 4910 in the cache */
   /* replace1786 -> 4911 in the cache */

   /* ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1787_0_0 -> 6 in the cache */
   /* search1787_0 -> 2103 in the cache */
   /* search1787_1_0 -> 2280 in the cache */
   /* search1787_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2103, 4998 },
      -1,
   } },

   /* replace1787_0 -> 6 in the cache */
   /* replace1787_1_0 -> 2280 in the cache */
   /* replace1787_1 -> 4910 in the cache */
   /* replace1787 -> 4913 in the cache */

   /* ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1788_0_0 -> 6 in the cache */
   /* search1788_0 -> 2103 in the cache */
   /* search1788_1_0 -> 2280 in the cache */
   /* search1788_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2103, 4998 },
      -1,
   } },

   /* replace1788_0 -> 6 in the cache */
   /* replace1788_1_0 -> 2280 in the cache */
   /* replace1788_1 -> 4910 in the cache */
   /* replace1788 -> 4915 in the cache */

   /* ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1789_0_0 -> 6 in the cache */
   /* search1789_0 -> 2103 in the cache */
   /* search1789_1_0 -> 2280 in the cache */
   /* search1789_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2103, 4998 },
      -1,
   } },

   /* replace1789_0 -> 6 in the cache */
   /* replace1789_1_0 -> 2280 in the cache */
   /* replace1789_1 -> 4910 in the cache */
   /* replace1789 -> 4917 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1790_0_0 -> 4573 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4573 },
      -1,
   } },
   /* search1790_1_0 -> 4575 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4575 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5003, 5004 },
      -1,
   } },

   /* replace1790_0_0 -> 4573 in the cache */
   /* replace1790_0 -> 4921 in the cache */
   /* replace1790_1 -> 4575 in the cache */
   /* replace1790 -> 4922 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1791_0_0 -> 4573 in the cache */
   /* search1791_0 -> 5003 in the cache */
   /* search1791_1_0 -> 4575 in the cache */
   /* search1791_1 -> 5004 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5003, 5004 },
      -1,
   } },

   /* replace1791_0_0 -> 4573 in the cache */
   /* replace1791_0 -> 4921 in the cache */
   /* replace1791_1 -> 4575 in the cache */
   /* replace1791 -> 4924 in the cache */

   /* ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1792_0_0 -> 6 in the cache */
   /* search1792_0 -> 2103 in the cache */
   /* search1792_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2103, 4674 },
      -1,
   } },

   /* replace1792_0_0 -> 6 in the cache */
   /* replace1792_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 5008 },
      -1,
   } },
   /* replace1792_1_0_0_0 -> 41 in the cache */
   /* replace1792_1_0_0 -> 5008 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 5008 },
      -1,
   } },
   /* replace1792_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 5010, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 5009, 5011 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1793_0_0 -> 6 in the cache */
   /* search1793_0 -> 2103 in the cache */
   /* search1793_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2103, 4674 },
      -1,
   } },

   /* replace1793_0_0 -> 6 in the cache */
   /* replace1793_0_1_0 -> 41 in the cache */
   /* replace1793_0_1 -> 5008 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 5008 },
      -1,
   } },
   /* replace1793_1_0_0_0 -> 41 in the cache */
   /* replace1793_1_0_0 -> 5008 in the cache */
   /* replace1793_1_0 -> 5010 in the cache */
   /* replace1793_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 5010, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 5014, 5015 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1794_0_0 -> 6 in the cache */
   /* search1794_0 -> 2103 in the cache */
   /* search1794_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2103, 4674 },
      -1,
   } },

   /* replace1794_0_0 -> 910 in the cache */
   /* replace1794_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 910, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace1794_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5019, 41 },
      -1,
   } },
   /* replace1794_1_1_0 -> 6 in the cache */
   /* replace1794_1_1_1_0 -> 41 in the cache */
   /* replace1794_1_1_1 -> 5008 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 5008 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5020, 5021 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5018, 5022 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1795_0 -> 4693 in the cache */
   /* search1795_1_0 -> 2280 in the cache */
   /* search1795_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4693, 4998 },
      -1,
   } },

   /* replace1795_0_0 -> 21 in the cache */
   /* replace1795_0_1 -> 5019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 5019 },
      -1,
   } },
   /* replace1795_1_0_0 -> 21 in the cache */
   /* replace1795_1_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 910 },
      -1,
   } },
   /* replace1795_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1795_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5027, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5026, 5028 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5025, 5029 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1796_0_0 -> 6 in the cache */
   /* search1796_0 -> 2103 in the cache */
   /* search1796_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2103, 4674 },
      -1,
   } },

   /* replace1796_0_0 -> 5019 in the cache */
   /* replace1796_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5019, 41 },
      -1,
   } },
   /* replace1796_1_0_0 -> 910 in the cache */
   /* replace1796_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 910, 41 },
      -1,
   } },
   /* replace1796_1_1_0 -> 6 in the cache */
   /* replace1796_1_1_1_0 -> 41 in the cache */
   /* replace1796_1_1_1 -> 5008 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 5008 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5033, 5034 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5032, 5035 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1797_0 -> 4693 in the cache */
   /* search1797_1_0 -> 2280 in the cache */
   /* search1797_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4693, 4998 },
      -1,
   } },

   /* replace1797_0_0 -> 21 in the cache */
   /* replace1797_0_1 -> 910 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 910 },
      -1,
   } },
   /* replace1797_1_0_0 -> 21 in the cache */
   /* replace1797_1_0_1 -> 5019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 5019 },
      -1,
   } },
   /* replace1797_1_1_0_0 -> 21 in the cache */
   /* replace1797_1_1_0 -> 5027 in the cache */
   /* replace1797_1_1_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5027, 2280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5039, 5040 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5038, 5041 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1798_0_0 -> 11 in the cache */
   /* search1798_0 -> 2129 in the cache */
   /* search1798_1_0 -> 2280 in the cache */
   /* search1798_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2129, 3478 },
      -1,
   } },

   /* replace1798_0 -> 11 in the cache */
   /* replace1798_1_0 -> 2280 in the cache */
   /* replace1798_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4908 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1799_0_0 -> 11 in the cache */
   /* search1799_0 -> 2129 in the cache */
   /* search1799_1_0 -> 2280 in the cache */
   /* search1799_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2129, 3478 },
      -1,
   } },

   /* replace1799_0 -> 11 in the cache */
   /* replace1799_1_0 -> 2280 in the cache */
   /* replace1799_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4908 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1800_0_0 -> 11 in the cache */
   /* search1800_0 -> 2129 in the cache */
   /* search1800_1_0 -> 2280 in the cache */
   /* search1800_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2129, 3478 },
      -1,
   } },

   /* replace1800_0 -> 11 in the cache */
   /* replace1800_1_0 -> 2280 in the cache */
   /* replace1800_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4908 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1801_0_0 -> 11 in the cache */
   /* search1801_0 -> 2129 in the cache */
   /* search1801_1_0 -> 2280 in the cache */
   /* search1801_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2129, 3478 },
      -1,
   } },

   /* replace1801_0 -> 11 in the cache */
   /* replace1801_1_0 -> 2280 in the cache */
   /* replace1801_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4908 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1802_0_0 -> 4573 in the cache */
   /* search1802_0 -> 4961 in the cache */
   /* search1802_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4961, 5051 },
      -1,
   } },

   /* replace1802_0_0 -> 4573 in the cache */
   /* replace1802_0 -> 4918 in the cache */
   /* replace1802_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4918, 536 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1803_0_0 -> 4573 in the cache */
   /* search1803_0 -> 4961 in the cache */
   /* search1803_1_0 -> 536 in the cache */
   /* search1803_1 -> 5051 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4961, 5051 },
      -1,
   } },

   /* replace1803_0_0 -> 4573 in the cache */
   /* replace1803_0 -> 4918 in the cache */
   /* replace1803_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4918, 536 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1804_0_0 -> 11 in the cache */
   /* search1804_0 -> 2129 in the cache */
   /* search1804_1_0 -> 268 in the cache */
   /* search1804_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2129, 3497 },
      -1,
   } },

   /* replace1804_0 -> 11 in the cache */
   /* replace1804_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 5057 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1805_0_0 -> 11 in the cache */
   /* search1805_0 -> 2129 in the cache */
   /* search1805_1_0 -> 268 in the cache */
   /* search1805_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2129, 3497 },
      -1,
   } },

   /* replace1805_0 -> 11 in the cache */
   /* replace1805_1_0 -> 268 in the cache */
   /* replace1805_1 -> 5057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 5057 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1806_0_0 -> 11 in the cache */
   /* search1806_0 -> 2129 in the cache */
   /* search1806_1_0 -> 268 in the cache */
   /* search1806_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2129, 3497 },
      -1,
   } },

   /* replace1806_0 -> 11 in the cache */
   /* replace1806_1_0 -> 268 in the cache */
   /* replace1806_1 -> 5057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 5057 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1807_0_0 -> 11 in the cache */
   /* search1807_0 -> 2129 in the cache */
   /* search1807_1_0 -> 268 in the cache */
   /* search1807_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2129, 3497 },
      -1,
   } },

   /* replace1807_0 -> 11 in the cache */
   /* replace1807_1_0 -> 268 in the cache */
   /* replace1807_1 -> 5057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 5057 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1808_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1808_1_0 -> 536 in the cache */
   /* search1808_1 -> 5051 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5065, 5051 },
      -1,
   } },

   /* replace1808_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 525 },
      -1,
   } },
   /* replace1808_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5067, 536 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1809_0_0 -> 525 in the cache */
   /* search1809_0 -> 5065 in the cache */
   /* search1809_1_0 -> 536 in the cache */
   /* search1809_1 -> 5051 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5065, 5051 },
      -1,
   } },

   /* replace1809_0_0 -> 525 in the cache */
   /* replace1809_0 -> 5067 in the cache */
   /* replace1809_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5067, 536 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1810_0_0 -> 11 in the cache */
   /* search1810_0 -> 2129 in the cache */
   /* search1810_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2129, 33 },
      -1,
   } },

   /* replace1810_0_0 -> 11 in the cache */
   /* replace1810_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 5072 },
      -1,
   } },
   /* replace1810_1_0_0_0 -> 36 in the cache */
   /* replace1810_1_0_0 -> 5072 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5072 },
      -1,
   } },
   /* replace1810_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 5074, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 5073, 5075 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1811_0_0 -> 11 in the cache */
   /* search1811_0 -> 2129 in the cache */
   /* search1811_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2129, 33 },
      -1,
   } },

   /* replace1811_0_0 -> 11 in the cache */
   /* replace1811_0_1_0 -> 36 in the cache */
   /* replace1811_0_1 -> 5072 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 5072 },
      -1,
   } },
   /* replace1811_1_0_0_0 -> 36 in the cache */
   /* replace1811_1_0_0 -> 5072 in the cache */
   /* replace1811_1_0 -> 5074 in the cache */
   /* replace1811_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 5074, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 5078, 5079 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1812_0_0 -> 11 in the cache */
   /* search1812_0 -> 2129 in the cache */
   /* search1812_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2129, 33 },
      -1,
   } },

   /* replace1812_0_0 -> 128 in the cache */
   /* replace1812_0_1 -> 36 in the cache */
   /* replace1812_0 -> 4975 in the cache */
   /* replace1812_1_0_0 -> 1015 in the cache */
   /* replace1812_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1015, 36 },
      -1,
   } },
   /* replace1812_1_1_0 -> 11 in the cache */
   /* replace1812_1_1_1_0 -> 36 in the cache */
   /* replace1812_1_1_1 -> 5072 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 5072 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5082, 5083 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4975, 5084 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1813_0 -> 296 in the cache */
   /* search1813_1_0 -> 268 in the cache */
   /* search1813_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 296, 3497 },
      -1,
   } },

   /* replace1813_0_0 -> 16 in the cache */
   /* replace1813_0_1 -> 1015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 1015 },
      -1,
   } },
   /* replace1813_1_0_0 -> 16 in the cache */
   /* replace1813_1_0_1 -> 128 in the cache */
   /* replace1813_1_0 -> 4982 in the cache */
   /* replace1813_1_1_0_0 -> 16 in the cache */
   /* replace1813_1_1_0 -> 3505 in the cache */
   /* replace1813_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3505, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4982, 5088 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5087, 5089 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1814_0_0 -> 11 in the cache */
   /* search1814_0 -> 2129 in the cache */
   /* search1814_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2129, 33 },
      -1,
   } },

   /* replace1814_0_0 -> 1015 in the cache */
   /* replace1814_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1015, 36 },
      -1,
   } },
   /* replace1814_1_0_0 -> 128 in the cache */
   /* replace1814_1_0_1 -> 36 in the cache */
   /* replace1814_1_0 -> 4988 in the cache */
   /* replace1814_1_1_0 -> 11 in the cache */
   /* replace1814_1_1_1_0 -> 36 in the cache */
   /* replace1814_1_1_1 -> 5072 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 5072 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4988, 5093 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5092, 5094 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1815_0 -> 296 in the cache */
   /* search1815_1_0 -> 268 in the cache */
   /* search1815_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 296, 3497 },
      -1,
   } },

   /* replace1815_0_0 -> 16 in the cache */
   /* replace1815_0_1 -> 128 in the cache */
   /* replace1815_0 -> 4993 in the cache */
   /* replace1815_1_0_0 -> 16 in the cache */
   /* replace1815_1_0_1 -> 1015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 1015 },
      -1,
   } },
   /* replace1815_1_1_0_0 -> 16 in the cache */
   /* replace1815_1_1_0 -> 3505 in the cache */
   /* replace1815_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3505, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5097, 5098 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4993, 5099 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1816_0_0 -> 11 in the cache */
   /* search1816_0 -> 2135 in the cache */
   /* search1816_1_0 -> 2280 in the cache */
   /* search1816_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2135, 4998 },
      -1,
   } },

   /* replace1816_0 -> 11 in the cache */
   /* replace1816_1_0 -> 2280 in the cache */
   /* replace1816_1 -> 4908 in the cache */
   /* replace1816 -> 5044 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1817_0_0 -> 11 in the cache */
   /* search1817_0 -> 2135 in the cache */
   /* search1817_1_0 -> 2280 in the cache */
   /* search1817_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2135, 4998 },
      -1,
   } },

   /* replace1817_0 -> 11 in the cache */
   /* replace1817_1_0 -> 2280 in the cache */
   /* replace1817_1 -> 4908 in the cache */
   /* replace1817 -> 5046 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1818_0_0 -> 11 in the cache */
   /* search1818_0 -> 2135 in the cache */
   /* search1818_1_0 -> 2280 in the cache */
   /* search1818_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2135, 4998 },
      -1,
   } },

   /* replace1818_0 -> 11 in the cache */
   /* replace1818_1_0 -> 2280 in the cache */
   /* replace1818_1 -> 4908 in the cache */
   /* replace1818 -> 5048 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1819_0_0 -> 11 in the cache */
   /* search1819_0 -> 2135 in the cache */
   /* search1819_1_0 -> 2280 in the cache */
   /* search1819_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2135, 4998 },
      -1,
   } },

   /* replace1819_0 -> 11 in the cache */
   /* replace1819_1_0 -> 2280 in the cache */
   /* replace1819_1 -> 4908 in the cache */
   /* replace1819 -> 5050 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1820_0_0 -> 4573 in the cache */
   /* search1820_0 -> 5003 in the cache */
   /* search1820_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5003, 5105 },
      -1,
   } },

   /* replace1820_0_0 -> 4573 in the cache */
   /* replace1820_0 -> 4918 in the cache */
   /* replace1820_1 -> 536 in the cache */
   /* replace1820 -> 5053 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1821_0_0 -> 4573 in the cache */
   /* search1821_0 -> 5003 in the cache */
   /* search1821_1_0 -> 536 in the cache */
   /* search1821_1 -> 5105 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5003, 5105 },
      -1,
   } },

   /* replace1821_0_0 -> 4573 in the cache */
   /* replace1821_0 -> 4918 in the cache */
   /* replace1821_1 -> 536 in the cache */
   /* replace1821 -> 5055 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1822_0_0 -> 11 in the cache */
   /* search1822_0 -> 2135 in the cache */
   /* search1822_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2135, 5108 },
      -1,
   } },

   /* replace1822_0 -> 11 in the cache */
   /* replace1822_1_0 -> 268 in the cache */
   /* replace1822_1 -> 5057 in the cache */
   /* replace1822 -> 5058 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1823_0_0 -> 11 in the cache */
   /* search1823_0 -> 2135 in the cache */
   /* search1823_1_0 -> 268 in the cache */
   /* search1823_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2135, 5108 },
      -1,
   } },

   /* replace1823_0 -> 11 in the cache */
   /* replace1823_1_0 -> 268 in the cache */
   /* replace1823_1 -> 5057 in the cache */
   /* replace1823 -> 5060 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1824_0_0 -> 11 in the cache */
   /* search1824_0 -> 2135 in the cache */
   /* search1824_1_0 -> 268 in the cache */
   /* search1824_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2135, 5108 },
      -1,
   } },

   /* replace1824_0 -> 11 in the cache */
   /* replace1824_1_0 -> 268 in the cache */
   /* replace1824_1 -> 5057 in the cache */
   /* replace1824 -> 5062 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1825_0_0 -> 11 in the cache */
   /* search1825_0 -> 2135 in the cache */
   /* search1825_1_0 -> 268 in the cache */
   /* search1825_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2135, 5108 },
      -1,
   } },

   /* replace1825_0 -> 11 in the cache */
   /* replace1825_1_0 -> 268 in the cache */
   /* replace1825_1 -> 5057 in the cache */
   /* replace1825 -> 5064 in the cache */

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1826_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1826_1_0 -> 536 in the cache */
   /* search1826_1 -> 5105 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5113, 5105 },
      -1,
   } },

   /* replace1826_0_0 -> 525 in the cache */
   /* replace1826_0 -> 5067 in the cache */
   /* replace1826_1 -> 536 in the cache */
   /* replace1826 -> 5068 in the cache */

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1827_0_0 -> 525 in the cache */
   /* search1827_0 -> 5113 in the cache */
   /* search1827_1_0 -> 536 in the cache */
   /* search1827_1 -> 5105 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5113, 5105 },
      -1,
   } },

   /* replace1827_0_0 -> 525 in the cache */
   /* replace1827_0 -> 5067 in the cache */
   /* replace1827_1 -> 536 in the cache */
   /* replace1827 -> 5070 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1828_0_0 -> 11 in the cache */
   /* search1828_0 -> 2135 in the cache */
   /* search1828_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2135, 4674 },
      -1,
   } },

   /* replace1828_0_0 -> 11 in the cache */
   /* replace1828_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 5117 },
      -1,
   } },
   /* replace1828_1_0_0_0 -> 41 in the cache */
   /* replace1828_1_0_0 -> 5117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 5117 },
      -1,
   } },
   /* replace1828_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 5119, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 5118, 5120 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1829_0_0 -> 11 in the cache */
   /* search1829_0 -> 2135 in the cache */
   /* search1829_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2135, 4674 },
      -1,
   } },

   /* replace1829_0_0 -> 11 in the cache */
   /* replace1829_0_1_0 -> 41 in the cache */
   /* replace1829_0_1 -> 5117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 5117 },
      -1,
   } },
   /* replace1829_1_0_0_0 -> 41 in the cache */
   /* replace1829_1_0_0 -> 5117 in the cache */
   /* replace1829_1_0 -> 5119 in the cache */
   /* replace1829_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 5119, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 5123, 5124 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1830_0_0 -> 11 in the cache */
   /* search1830_0 -> 2135 in the cache */
   /* search1830_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2135, 4674 },
      -1,
   } },

   /* replace1830_0_0 -> 910 in the cache */
   /* replace1830_0_1 -> 41 in the cache */
   /* replace1830_0 -> 5018 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace1830_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5127, 41 },
      -1,
   } },
   /* replace1830_1_1_0 -> 11 in the cache */
   /* replace1830_1_1_1_0 -> 41 in the cache */
   /* replace1830_1_1_1 -> 5117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 5117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5128, 5129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5018, 5130 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1831_0 -> 4693 in the cache */
   /* search1831_1_0 -> 268 in the cache */
   /* search1831_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4693, 5108 },
      -1,
   } },

   /* replace1831_0_0 -> 21 in the cache */
   /* replace1831_0_1 -> 5127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 5127 },
      -1,
   } },
   /* replace1831_1_0_0 -> 21 in the cache */
   /* replace1831_1_0_1 -> 910 in the cache */
   /* replace1831_1_0 -> 5026 in the cache */
   /* replace1831_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1831_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5134, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5026, 5135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5133, 5136 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1832_0_0 -> 11 in the cache */
   /* search1832_0 -> 2135 in the cache */
   /* search1832_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2135, 4674 },
      -1,
   } },

   /* replace1832_0_0 -> 5127 in the cache */
   /* replace1832_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5127, 41 },
      -1,
   } },
   /* replace1832_1_0_0 -> 910 in the cache */
   /* replace1832_1_0_1 -> 41 in the cache */
   /* replace1832_1_0 -> 5033 in the cache */
   /* replace1832_1_1_0 -> 11 in the cache */
   /* replace1832_1_1_1_0 -> 41 in the cache */
   /* replace1832_1_1_1 -> 5117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 5117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5033, 5140 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5139, 5141 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1833_0 -> 4693 in the cache */
   /* search1833_1_0 -> 268 in the cache */
   /* search1833_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4693, 5108 },
      -1,
   } },

   /* replace1833_0_0 -> 21 in the cache */
   /* replace1833_0_1 -> 910 in the cache */
   /* replace1833_0 -> 5038 in the cache */
   /* replace1833_1_0_0 -> 21 in the cache */
   /* replace1833_1_0_1 -> 5127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 5127 },
      -1,
   } },
   /* replace1833_1_1_0_0 -> 21 in the cache */
   /* replace1833_1_1_0 -> 5134 in the cache */
   /* replace1833_1_1_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5134, 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5144, 5145 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5038, 5146 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1834_0_0 -> 16 in the cache */
   /* search1834_0 -> 2149 in the cache */
   /* search1834_1_0 -> 2280 in the cache */
   /* search1834_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2149, 4998 },
      -1,
   } },

   /* replace1834_0 -> 16 in the cache */
   /* replace1834_1_0 -> 2280 in the cache */
   /* replace1834_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3478 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1835_0_0 -> 16 in the cache */
   /* search1835_0 -> 2149 in the cache */
   /* search1835_1_0 -> 2280 in the cache */
   /* search1835_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2149, 4998 },
      -1,
   } },

   /* replace1835_0 -> 16 in the cache */
   /* replace1835_1_0 -> 2280 in the cache */
   /* replace1835_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3478 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1836_0_0 -> 16 in the cache */
   /* search1836_0 -> 2149 in the cache */
   /* search1836_1_0 -> 2280 in the cache */
   /* search1836_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2149, 4998 },
      -1,
   } },

   /* replace1836_0 -> 16 in the cache */
   /* replace1836_1_0 -> 2280 in the cache */
   /* replace1836_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3478 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1837_0_0 -> 16 in the cache */
   /* search1837_0 -> 2149 in the cache */
   /* search1837_1_0 -> 2280 in the cache */
   /* search1837_1 -> 4998 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2149, 4998 },
      -1,
   } },

   /* replace1837_0 -> 16 in the cache */
   /* replace1837_1_0 -> 2280 in the cache */
   /* replace1837_1 -> 3478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3478 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1838_0_0 -> 4573 in the cache */
   /* search1838_0 -> 5003 in the cache */
   /* search1838_1_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5003, 5156 },
      -1,
   } },

   /* replace1838_0_0 -> 4573 in the cache */
   /* replace1838_0 -> 4961 in the cache */
   /* replace1838_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4961, 189 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1839_0_0 -> 4573 in the cache */
   /* search1839_0 -> 5003 in the cache */
   /* search1839_1_0 -> 189 in the cache */
   /* search1839_1 -> 5156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5003, 5156 },
      -1,
   } },

   /* replace1839_0_0 -> 4573 in the cache */
   /* replace1839_0 -> 4961 in the cache */
   /* replace1839_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4961, 189 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1840_0_0 -> 16 in the cache */
   /* search1840_0 -> 2149 in the cache */
   /* search1840_1_0 -> 268 in the cache */
   /* search1840_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2149, 5108 },
      -1,
   } },

   /* replace1840_0 -> 16 in the cache */
   /* replace1840_1_0 -> 268 in the cache */
   /* replace1840_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3497 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1841_0_0 -> 16 in the cache */
   /* search1841_0 -> 2149 in the cache */
   /* search1841_1_0 -> 268 in the cache */
   /* search1841_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2149, 5108 },
      -1,
   } },

   /* replace1841_0 -> 16 in the cache */
   /* replace1841_1_0 -> 268 in the cache */
   /* replace1841_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3497 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1842_0_0 -> 16 in the cache */
   /* search1842_0 -> 2149 in the cache */
   /* search1842_1_0 -> 268 in the cache */
   /* search1842_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2149, 5108 },
      -1,
   } },

   /* replace1842_0 -> 16 in the cache */
   /* replace1842_1_0 -> 268 in the cache */
   /* replace1842_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3497 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1843_0_0 -> 16 in the cache */
   /* search1843_0 -> 2149 in the cache */
   /* search1843_1_0 -> 268 in the cache */
   /* search1843_1 -> 5108 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2149, 5108 },
      -1,
   } },

   /* replace1843_0 -> 16 in the cache */
   /* replace1843_1_0 -> 268 in the cache */
   /* replace1843_1 -> 3497 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3497 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1844_0_0 -> 525 in the cache */
   /* search1844_0 -> 5113 in the cache */
   /* search1844_1_0 -> 189 in the cache */
   /* search1844_1 -> 5156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5113, 5156 },
      -1,
   } },

   /* replace1844_0_0 -> 525 in the cache */
   /* replace1844_0 -> 5065 in the cache */
   /* replace1844_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5065, 189 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1845_0_0 -> 525 in the cache */
   /* search1845_0 -> 5113 in the cache */
   /* search1845_1_0 -> 189 in the cache */
   /* search1845_1 -> 5156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5113, 5156 },
      -1,
   } },

   /* replace1845_0_0 -> 525 in the cache */
   /* replace1845_0 -> 5065 in the cache */
   /* replace1845_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5065, 189 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1846_0_0 -> 16 in the cache */
   /* search1846_0 -> 2149 in the cache */
   /* search1846_1_0 -> 36 in the cache */
   /* search1846_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2149, 3494 },
      -1,
   } },

   /* replace1846_0 -> 16 in the cache */
   /* replace1846_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 5174 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1847_0_0 -> 16 in the cache */
   /* search1847_0 -> 2149 in the cache */
   /* search1847_1_0 -> 36 in the cache */
   /* search1847_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2149, 3494 },
      -1,
   } },

   /* replace1847_0 -> 16 in the cache */
   /* replace1847_1_0 -> 36 in the cache */
   /* replace1847_1 -> 5174 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 5174 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1848_0_0 -> 16 in the cache */
   /* search1848_0 -> 2149 in the cache */
   /* search1848_1_0 -> 36 in the cache */
   /* search1848_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2149, 3494 },
      -1,
   } },

   /* replace1848_0 -> 16 in the cache */
   /* replace1848_1_0 -> 36 in the cache */
   /* replace1848_1 -> 5174 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 5174 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1849_0_0 -> 16 in the cache */
   /* search1849_0 -> 2149 in the cache */
   /* search1849_1_0 -> 36 in the cache */
   /* search1849_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2149, 3494 },
      -1,
   } },

   /* replace1849_0 -> 16 in the cache */
   /* replace1849_1_0 -> 36 in the cache */
   /* replace1849_1 -> 5174 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 5174 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1850_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 383 },
      -1,
   } },
   /* search1850_1_0 -> 189 in the cache */
   /* search1850_1 -> 5156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5182, 5156 },
      -1,
   } },

   /* replace1850_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 383 },
      -1,
   } },
   /* replace1850_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5184, 189 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1851_0_0 -> 383 in the cache */
   /* search1851_0 -> 5182 in the cache */
   /* search1851_1_0 -> 189 in the cache */
   /* search1851_1 -> 5156 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5182, 5156 },
      -1,
   } },

   /* replace1851_0_0 -> 383 in the cache */
   /* replace1851_0 -> 5184 in the cache */
   /* replace1851_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5184, 189 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1852_0_0 -> 16 in the cache */
   /* search1852_0 -> 2149 in the cache */
   /* search1852_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2149, 4674 },
      -1,
   } },

   /* replace1852_0_0 -> 16 in the cache */
   /* replace1852_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 5189 },
      -1,
   } },
   /* replace1852_1_0_0_0 -> 41 in the cache */
   /* replace1852_1_0_0 -> 5189 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 5189 },
      -1,
   } },
   /* replace1852_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 5191, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 5190, 5192 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1853_0_0 -> 16 in the cache */
   /* search1853_0 -> 2149 in the cache */
   /* search1853_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2149, 4674 },
      -1,
   } },

   /* replace1853_0_0 -> 16 in the cache */
   /* replace1853_0_1_0 -> 41 in the cache */
   /* replace1853_0_1 -> 5189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 5189 },
      -1,
   } },
   /* replace1853_1_0_0_0 -> 41 in the cache */
   /* replace1853_1_0_0 -> 5189 in the cache */
   /* replace1853_1_0 -> 5191 in the cache */
   /* replace1853_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 5191, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 5195, 5196 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b')))) */
   /* search1854_0_0 -> 16 in the cache */
   /* search1854_0 -> 2149 in the cache */
   /* search1854_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2149, 4674 },
      -1,
   } },

   /* replace1854_0_0 -> 910 in the cache */
   /* replace1854_0_1 -> 41 in the cache */
   /* replace1854_0 -> 5018 in the cache */
   /* replace1854_1_0_0 -> 3510 in the cache */
   /* replace1854_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3510, 41 },
      -1,
   } },
   /* replace1854_1_1_0 -> 16 in the cache */
   /* replace1854_1_1_1_0 -> 41 in the cache */
   /* replace1854_1_1_1 -> 5189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 5189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5199, 5200 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5018, 5201 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b'))) */
   /* search1855_0 -> 4693 in the cache */
   /* search1855_1_0 -> 36 in the cache */
   /* search1855_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4693, 3494 },
      -1,
   } },

   /* replace1855_0_0 -> 21 in the cache */
   /* replace1855_0_1 -> 3510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 3510 },
      -1,
   } },
   /* replace1855_1_0_0 -> 21 in the cache */
   /* replace1855_1_0_1 -> 910 in the cache */
   /* replace1855_1_0 -> 5026 in the cache */
   /* replace1855_1_1_0_0 -> 21 in the cache */
   /* replace1855_1_1_0 -> 3502 in the cache */
   /* replace1855_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3502, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5026, 5205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5204, 5206 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b')))) */
   /* search1856_0_0 -> 16 in the cache */
   /* search1856_0 -> 2149 in the cache */
   /* search1856_1 -> 4674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2149, 4674 },
      -1,
   } },

   /* replace1856_0_0 -> 3510 in the cache */
   /* replace1856_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3510, 41 },
      -1,
   } },
   /* replace1856_1_0_0 -> 910 in the cache */
   /* replace1856_1_0_1 -> 41 in the cache */
   /* replace1856_1_0 -> 5033 in the cache */
   /* replace1856_1_1_0 -> 16 in the cache */
   /* replace1856_1_1_1_0 -> 41 in the cache */
   /* replace1856_1_1_1 -> 5189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 5189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5033, 5210 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5209, 5211 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b'))) */
   /* search1857_0 -> 4693 in the cache */
   /* search1857_1_0 -> 36 in the cache */
   /* search1857_1 -> 3494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4693, 3494 },
      -1,
   } },

   /* replace1857_0_0 -> 21 in the cache */
   /* replace1857_0_1 -> 910 in the cache */
   /* replace1857_0 -> 5038 in the cache */
   /* replace1857_1_0_0 -> 21 in the cache */
   /* replace1857_1_0_1 -> 3510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 3510 },
      -1,
   } },
   /* replace1857_1_1_0_0 -> 21 in the cache */
   /* replace1857_1_1_0 -> 3502 in the cache */
   /* replace1857_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3502, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5214, 5215 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5038, 5216 },
      -1,
   } },

   /* ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1858_0_0 -> 11 in the cache */
   /* search1858_0 -> 1845 in the cache */
   /* search1858_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1845, 5218 },
      -1,
   } },

   /* replace1858_0 -> 11 in the cache */
   /* replace1858_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 5220 },
      -1,
   } },

   /* ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1859_0_0 -> 11 in the cache */
   /* search1859_0 -> 1845 in the cache */
   /* search1859_1_0 -> 268 in the cache */
   /* search1859_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1845, 5218 },
      -1,
   } },

   /* replace1859_0 -> 11 in the cache */
   /* replace1859_1_0 -> 268 in the cache */
   /* replace1859_1 -> 5220 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 11, 5220 },
      -1,
   } },

   /* ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1860_0_0 -> 11 in the cache */
   /* search1860_0 -> 1845 in the cache */
   /* search1860_1_0 -> 268 in the cache */
   /* search1860_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1845, 5218 },
      -1,
   } },

   /* replace1860_0 -> 11 in the cache */
   /* replace1860_1_0 -> 268 in the cache */
   /* replace1860_1 -> 5220 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 5220 },
      -1,
   } },

   /* ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1861_0_0 -> 11 in the cache */
   /* search1861_0 -> 1845 in the cache */
   /* search1861_1_0 -> 268 in the cache */
   /* search1861_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1845, 5218 },
      -1,
   } },

   /* replace1861_0 -> 11 in the cache */
   /* replace1861_1_0 -> 268 in the cache */
   /* replace1861_1 -> 5220 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 5220 },
      -1,
   } },

   /* ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1862_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1862_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5228, 5229 },
      -1,
   } },

   /* replace1862_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 525 },
      -1,
   } },
   /* replace1862_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5231, 536 },
      -1,
   } },

   /* ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1863_0_0 -> 525 in the cache */
   /* search1863_0 -> 5228 in the cache */
   /* search1863_1_0 -> 536 in the cache */
   /* search1863_1 -> 5229 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5228, 5229 },
      -1,
   } },

   /* replace1863_0_0 -> 525 in the cache */
   /* replace1863_0 -> 5231 in the cache */
   /* replace1863_1 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5231, 536 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1864_0_0 -> 11 in the cache */
   /* search1864_0 -> 1855 in the cache */
   /* search1864_1_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1855, 5235 },
      -1,
   } },

   /* replace1864_0 -> 11 in the cache */
   /* replace1864_1_0 -> 268 in the cache */
   /* replace1864_1 -> 5220 in the cache */
   /* replace1864 -> 5221 in the cache */

   /* ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1865_0_0 -> 11 in the cache */
   /* search1865_0 -> 1855 in the cache */
   /* search1865_1_0 -> 268 in the cache */
   /* search1865_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1855, 5235 },
      -1,
   } },

   /* replace1865_0 -> 11 in the cache */
   /* replace1865_1_0 -> 268 in the cache */
   /* replace1865_1 -> 5220 in the cache */
   /* replace1865 -> 5223 in the cache */

   /* ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1866_0_0 -> 11 in the cache */
   /* search1866_0 -> 1855 in the cache */
   /* search1866_1_0 -> 268 in the cache */
   /* search1866_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1855, 5235 },
      -1,
   } },

   /* replace1866_0 -> 11 in the cache */
   /* replace1866_1_0 -> 268 in the cache */
   /* replace1866_1 -> 5220 in the cache */
   /* replace1866 -> 5225 in the cache */

   /* ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1867_0_0 -> 11 in the cache */
   /* search1867_0 -> 1855 in the cache */
   /* search1867_1_0 -> 268 in the cache */
   /* search1867_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1855, 5235 },
      -1,
   } },

   /* replace1867_0 -> 11 in the cache */
   /* replace1867_1_0 -> 268 in the cache */
   /* replace1867_1 -> 5220 in the cache */
   /* replace1867 -> 5227 in the cache */

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1868_0_0 -> 525 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 525 },
      -1,
   } },
   /* search1868_1_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5240, 5241 },
      -1,
   } },

   /* replace1868_0_0 -> 525 in the cache */
   /* replace1868_0 -> 5231 in the cache */
   /* replace1868_1 -> 536 in the cache */
   /* replace1868 -> 5232 in the cache */

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1869_0_0 -> 525 in the cache */
   /* search1869_0 -> 5240 in the cache */
   /* search1869_1_0 -> 536 in the cache */
   /* search1869_1 -> 5241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5240, 5241 },
      -1,
   } },

   /* replace1869_0_0 -> 525 in the cache */
   /* replace1869_0 -> 5231 in the cache */
   /* replace1869_1 -> 536 in the cache */
   /* replace1869 -> 5234 in the cache */

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1870_0_0 -> 16 in the cache */
   /* search1870_0 -> 1951 in the cache */
   /* search1870_1_0 -> 268 in the cache */
   /* search1870_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1951, 5235 },
      -1,
   } },

   /* replace1870_0 -> 16 in the cache */
   /* replace1870_1_0 -> 268 in the cache */
   /* replace1870_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 5218 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1871_0_0 -> 16 in the cache */
   /* search1871_0 -> 1951 in the cache */
   /* search1871_1_0 -> 268 in the cache */
   /* search1871_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1951, 5235 },
      -1,
   } },

   /* replace1871_0 -> 16 in the cache */
   /* replace1871_1_0 -> 268 in the cache */
   /* replace1871_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 5218 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1872_0_0 -> 16 in the cache */
   /* search1872_0 -> 1951 in the cache */
   /* search1872_1_0 -> 268 in the cache */
   /* search1872_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1951, 5235 },
      -1,
   } },

   /* replace1872_0 -> 16 in the cache */
   /* replace1872_1_0 -> 268 in the cache */
   /* replace1872_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 5218 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1873_0_0 -> 16 in the cache */
   /* search1873_0 -> 1951 in the cache */
   /* search1873_1_0 -> 268 in the cache */
   /* search1873_1 -> 5235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1951, 5235 },
      -1,
   } },

   /* replace1873_0 -> 16 in the cache */
   /* replace1873_1_0 -> 268 in the cache */
   /* replace1873_1 -> 5218 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 5218 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1874_0_0 -> 525 in the cache */
   /* search1874_0 -> 5240 in the cache */
   /* search1874_1_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5240, 5252 },
      -1,
   } },

   /* replace1874_0_0 -> 525 in the cache */
   /* replace1874_0 -> 5228 in the cache */
   /* replace1874_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5228, 189 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1875_0_0 -> 525 in the cache */
   /* search1875_0 -> 5240 in the cache */
   /* search1875_1_0 -> 189 in the cache */
   /* search1875_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5240, 5252 },
      -1,
   } },

   /* replace1875_0_0 -> 525 in the cache */
   /* replace1875_0 -> 5228 in the cache */
   /* replace1875_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5228, 189 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1876_0_0 -> 16 in the cache */
   /* search1876_0 -> 1951 in the cache */
   /* search1876_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1951, 5257 },
      -1,
   } },

   /* replace1876_0 -> 16 in the cache */
   /* replace1876_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 5259 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1877_0_0 -> 16 in the cache */
   /* search1877_0 -> 1951 in the cache */
   /* search1877_1_0 -> 36 in the cache */
   /* search1877_1 -> 5257 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1951, 5257 },
      -1,
   } },

   /* replace1877_0 -> 16 in the cache */
   /* replace1877_1_0 -> 36 in the cache */
   /* replace1877_1 -> 5259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 5259 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1878_0_0 -> 16 in the cache */
   /* search1878_0 -> 1951 in the cache */
   /* search1878_1_0 -> 36 in the cache */
   /* search1878_1 -> 5257 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1951, 5257 },
      -1,
   } },

   /* replace1878_0 -> 16 in the cache */
   /* replace1878_1_0 -> 36 in the cache */
   /* replace1878_1 -> 5259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 5259 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1879_0_0 -> 16 in the cache */
   /* search1879_0 -> 1951 in the cache */
   /* search1879_1_0 -> 36 in the cache */
   /* search1879_1 -> 5257 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1951, 5257 },
      -1,
   } },

   /* replace1879_0 -> 16 in the cache */
   /* replace1879_1_0 -> 36 in the cache */
   /* replace1879_1 -> 5259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 5259 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1880_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 383 },
      -1,
   } },
   /* search1880_1_0 -> 189 in the cache */
   /* search1880_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5267, 5252 },
      -1,
   } },

   /* replace1880_0_0 -> 383 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 383 },
      -1,
   } },
   /* replace1880_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5269, 189 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1881_0_0 -> 383 in the cache */
   /* search1881_0 -> 5267 in the cache */
   /* search1881_1_0 -> 189 in the cache */
   /* search1881_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5267, 5252 },
      -1,
   } },

   /* replace1881_0_0 -> 383 in the cache */
   /* replace1881_0 -> 5269 in the cache */
   /* replace1881_1 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5269, 189 },
      -1,
   } },

   /* ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a') */
   /* search1882_0_0 -> 21 in the cache */
   /* search1882_0_1 -> 3510 in the cache */
   /* search1882_0 -> 3511 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 1,
      { 3511 },
      -1,
   } },

   /* replace1882_0 -> 21 in the cache */
   /* replace1882 -> 3502 in the cache */

   /* ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a') */
   /* search1883_0_0 -> 16 in the cache */
   /* search1883_0_1 -> 1015 in the cache */
   /* search1883_0 -> 1016 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 1016 },
      -1,
   } },

   /* replace1883_0 -> 16 in the cache */
   /* replace1883 -> 3505 in the cache */

   /* ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a') */
   /* search1884_0_0 -> 21 in the cache */
   /* search1884_0_1 -> 5127 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 5127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 5275 },
      -1,
   } },

   /* replace1884_0 -> 21 in the cache */
   /* replace1884 -> 5134 in the cache */

   /* ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a') */
   /* search1885_0_0 -> 11 in the cache */
   /* search1885_0_1 -> 2181 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 2181 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 5277 },
      -1,
   } },

   /* replace1885_0 -> 11 in the cache */
   /* replace1885 -> 2177 in the cache */

   /* ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a') */
   /* search1886_0_0 -> 16 in the cache */
   /* search1886_0_1 -> 1023 in the cache */
   /* search1886_0 -> 1024 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 1024 },
      -1,
   } },

   /* replace1886_0 -> 16 in the cache */
   /* replace1886 -> 2193 in the cache */

   /* ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a') */
   /* search1887_0_0 -> 21 in the cache */
   /* search1887_0_1 -> 5019 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 5019 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 5280 },
      -1,
   } },

   /* replace1887_0 -> 21 in the cache */
   /* replace1887 -> 5027 in the cache */

   /* ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1888_0_0 -> 11 in the cache */
   /* search1888_0 -> 2177 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2177 },
      5,
   } },

   /* replace1888 -> 11 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1889_0 -> 11 in the cache */
   /* search1889_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 869 },
      5,
   } },

   /* replace1889 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1890_0 -> 531 in the cache */
   /* search1890_1_0_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5284 },
      -1,
   } },
   /* search1890_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5285, 498 },
      5,
   } },

   /* replace1890_0 -> 531 in the cache */
   /* replace1890_1 -> 536 in the cache */
   /* replace1890_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 536, 498 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1891_0 -> 531 in the cache */
   /* search1891_1_0_0 -> 536 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5288 },
      -1,
   } },
   /* search1891_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5289, 498 },
      5,
   } },

   /* replace1891_0 -> 531 in the cache */
   /* replace1891_1 -> 536 in the cache */
   /* replace1891_2 -> 498 in the cache */
   /* replace1891 -> 5287 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1892_0 -> 531 in the cache */
   /* search1892_1_0 -> 536 in the cache */
   /* search1892_1_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 536, 869 },
      -1,
   } },
   /* search1892_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5291, 498 },
      5,
   } },

   /* replace1892_0 -> 531 in the cache */
   /* replace1892_1 -> 536 in the cache */
   /* replace1892_2 -> 498 in the cache */
   /* replace1892 -> 5287 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1893_0_0_0 -> 11 in the cache */
   /* search1893_0_0 -> 2177 in the cache */
   /* search1893_0 -> 2178 in the cache */
   /* search1893_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 2178, 268 },
      5,
   } },

   /* replace1893_0 -> 11 in the cache */
   /* replace1893_1 -> 268 in the cache */
   /* replace1893 -> 4277 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1894_0_0_0 -> 11 in the cache */
   /* search1894_0_0 -> 4604 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4604 },
      -1,
   } },
   /* search1894_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5294, 268 },
      5,
   } },

   /* replace1894_0 -> 11 in the cache */
   /* replace1894_1 -> 268 in the cache */
   /* replace1894 -> 4277 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1895_0_0 -> 11 in the cache */
   /* search1895_0_1 -> 869 in the cache */
   /* search1895_0 -> 3620 in the cache */
   /* search1895_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3620, 268 },
      5,
   } },

   /* replace1895_0 -> 11 in the cache */
   /* replace1895_1 -> 268 in the cache */
   /* replace1895 -> 4277 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1896_0_0_0 -> 11 in the cache */
   /* search1896_0_0 -> 2177 in the cache */
   /* search1896_0 -> 2178 in the cache */
   /* search1896_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 2178, 268 },
      5,
   } },

   /* replace1896_0 -> 11 in the cache */
   /* replace1896_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 268 },
      -1,
   } },

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1897_0_0_0 -> 11 in the cache */
   /* search1897_0_0 -> 4604 in the cache */
   /* search1897_0 -> 5294 in the cache */
   /* search1897_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5294, 268 },
      5,
   } },

   /* replace1897_0 -> 11 in the cache */
   /* replace1897_1 -> 268 in the cache */
   /* replace1897 -> 5298 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1898_0_0 -> 11 in the cache */
   /* search1898_0_1 -> 869 in the cache */
   /* search1898_0 -> 3620 in the cache */
   /* search1898_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3620, 268 },
      5,
   } },

   /* replace1898_0 -> 11 in the cache */
   /* replace1898_1 -> 268 in the cache */
   /* replace1898 -> 5298 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1899_0_0_0 -> 11 in the cache */
   /* search1899_0_0 -> 2177 in the cache */
   /* search1899_0 -> 2178 in the cache */
   /* search1899_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2178, 268 },
      5,
   } },

   /* replace1899_0 -> 11 in the cache */
   /* replace1899_1 -> 268 in the cache */
   /* replace1899 -> 2191 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1900_0_0_0 -> 11 in the cache */
   /* search1900_0_0 -> 4604 in the cache */
   /* search1900_0 -> 5294 in the cache */
   /* search1900_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5294, 268 },
      5,
   } },

   /* replace1900_0 -> 11 in the cache */
   /* replace1900_1 -> 268 in the cache */
   /* replace1900 -> 2191 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1901_0_0 -> 11 in the cache */
   /* search1901_0_1 -> 869 in the cache */
   /* search1901_0 -> 3620 in the cache */
   /* search1901_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3620, 268 },
      5,
   } },

   /* replace1901_0 -> 11 in the cache */
   /* replace1901_1 -> 268 in the cache */
   /* replace1901 -> 2191 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1902_0_0_0 -> 11 in the cache */
   /* search1902_0_0 -> 2177 in the cache */
   /* search1902_0 -> 2178 in the cache */
   /* search1902_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2178, 268 },
      5,
   } },

   /* replace1902_0 -> 11 in the cache */
   /* replace1902_1 -> 268 in the cache */
   /* replace1902 -> 2233 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1903_0_0_0 -> 11 in the cache */
   /* search1903_0_0 -> 4604 in the cache */
   /* search1903_0 -> 5294 in the cache */
   /* search1903_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5294, 268 },
      5,
   } },

   /* replace1903_0 -> 11 in the cache */
   /* replace1903_1 -> 268 in the cache */
   /* replace1903 -> 2233 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1904_0_0 -> 11 in the cache */
   /* search1904_0_1 -> 869 in the cache */
   /* search1904_0 -> 3620 in the cache */
   /* search1904_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3620, 268 },
      5,
   } },

   /* replace1904_0 -> 11 in the cache */
   /* replace1904_1 -> 268 in the cache */
   /* replace1904 -> 2233 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1905_0_0_0 -> 11 in the cache */
   /* search1905_0_0 -> 2177 in the cache */
   /* search1905_0 -> 2178 in the cache */
   /* search1905_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2178, 268 },
      5,
   } },

   /* replace1905_0 -> 11 in the cache */
   /* replace1905_1 -> 268 in the cache */
   /* replace1905 -> 2257 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1906_0_0_0 -> 11 in the cache */
   /* search1906_0_0 -> 4604 in the cache */
   /* search1906_0 -> 5294 in the cache */
   /* search1906_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5294, 268 },
      5,
   } },

   /* replace1906_0 -> 11 in the cache */
   /* replace1906_1 -> 268 in the cache */
   /* replace1906 -> 2257 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1907_0_0 -> 11 in the cache */
   /* search1907_0_1 -> 869 in the cache */
   /* search1907_0 -> 3620 in the cache */
   /* search1907_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3620, 268 },
      5,
   } },

   /* replace1907_0 -> 11 in the cache */
   /* replace1907_1 -> 268 in the cache */
   /* replace1907 -> 2257 in the cache */

   /* ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1908_0_0 -> 11 in the cache */
   /* search1908_0 -> 2177 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2177 },
      5,
   } },

   /* replace1908 -> 11 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1909_0 -> 11 in the cache */
   /* search1909_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 869 },
      5,
   } },

   /* replace1909 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1910_0 -> 531 in the cache */
   /* search1910_1_0_0 -> 536 in the cache */
   /* search1910_1_0 -> 5284 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5284 },
      -1,
   } },
   /* search1910_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5312, 498 },
      5,
   } },

   /* replace1910_0 -> 531 in the cache */
   /* replace1910_1 -> 536 in the cache */
   /* replace1910_2 -> 498 in the cache */
   /* replace1910 -> 5287 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1911_0 -> 531 in the cache */
   /* search1911_1_0_0 -> 536 in the cache */
   /* search1911_1_0 -> 5288 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5288 },
      -1,
   } },
   /* search1911_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5314, 498 },
      5,
   } },

   /* replace1911_0 -> 531 in the cache */
   /* replace1911_1 -> 536 in the cache */
   /* replace1911_2 -> 498 in the cache */
   /* replace1911 -> 5287 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1912_0 -> 531 in the cache */
   /* search1912_1_0 -> 536 in the cache */
   /* search1912_1_1 -> 869 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 536, 869 },
      -1,
   } },
   /* search1912_2 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5316, 498 },
      5,
   } },

   /* replace1912_0 -> 531 in the cache */
   /* replace1912_1 -> 536 in the cache */
   /* replace1912_2 -> 498 in the cache */
   /* replace1912 -> 5287 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1913_0_0_0 -> 11 in the cache */
   /* search1913_0_0 -> 2177 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2177 },
      -1,
   } },
   /* search1913_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5318, 268 },
      5,
   } },

   /* replace1913_0 -> 11 in the cache */
   /* replace1913_1 -> 268 in the cache */
   /* replace1913 -> 4277 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1914_0_0_0 -> 11 in the cache */
   /* search1914_0_0 -> 4604 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4604 },
      -1,
   } },
   /* search1914_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5320, 268 },
      5,
   } },

   /* replace1914_0 -> 11 in the cache */
   /* replace1914_1 -> 268 in the cache */
   /* replace1914 -> 4277 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1915_0_0 -> 11 in the cache */
   /* search1915_0_1 -> 869 in the cache */
   /* search1915_0 -> 3641 in the cache */
   /* search1915_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3641, 268 },
      5,
   } },

   /* replace1915_0 -> 11 in the cache */
   /* replace1915_1 -> 268 in the cache */
   /* replace1915 -> 4277 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1916_0_0_0 -> 11 in the cache */
   /* search1916_0_0 -> 2177 in the cache */
   /* search1916_0 -> 5318 in the cache */
   /* search1916_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5318, 268 },
      5,
   } },

   /* replace1916_0 -> 11 in the cache */
   /* replace1916_1 -> 268 in the cache */
   /* replace1916 -> 5298 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1917_0_0_0 -> 11 in the cache */
   /* search1917_0_0 -> 4604 in the cache */
   /* search1917_0 -> 5320 in the cache */
   /* search1917_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5320, 268 },
      5,
   } },

   /* replace1917_0 -> 11 in the cache */
   /* replace1917_1 -> 268 in the cache */
   /* replace1917 -> 5298 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1918_0_0 -> 11 in the cache */
   /* search1918_0_1 -> 869 in the cache */
   /* search1918_0 -> 3641 in the cache */
   /* search1918_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3641, 268 },
      5,
   } },

   /* replace1918_0 -> 11 in the cache */
   /* replace1918_1 -> 268 in the cache */
   /* replace1918 -> 5298 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1919_0_0_0 -> 11 in the cache */
   /* search1919_0_0 -> 2177 in the cache */
   /* search1919_0 -> 5318 in the cache */
   /* search1919_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5318, 268 },
      5,
   } },

   /* replace1919_0 -> 11 in the cache */
   /* replace1919_1 -> 268 in the cache */
   /* replace1919 -> 2191 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1920_0_0_0 -> 11 in the cache */
   /* search1920_0_0 -> 4604 in the cache */
   /* search1920_0 -> 5320 in the cache */
   /* search1920_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5320, 268 },
      5,
   } },

   /* replace1920_0 -> 11 in the cache */
   /* replace1920_1 -> 268 in the cache */
   /* replace1920 -> 2191 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1921_0_0 -> 11 in the cache */
   /* search1921_0_1 -> 869 in the cache */
   /* search1921_0 -> 3641 in the cache */
   /* search1921_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3641, 268 },
      5,
   } },

   /* replace1921_0 -> 11 in the cache */
   /* replace1921_1 -> 268 in the cache */
   /* replace1921 -> 2191 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1922_0_0_0 -> 11 in the cache */
   /* search1922_0_0 -> 2177 in the cache */
   /* search1922_0 -> 5318 in the cache */
   /* search1922_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5318, 268 },
      5,
   } },

   /* replace1922_0 -> 11 in the cache */
   /* replace1922_1 -> 268 in the cache */
   /* replace1922 -> 2233 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1923_0_0_0 -> 11 in the cache */
   /* search1923_0_0 -> 4604 in the cache */
   /* search1923_0 -> 5320 in the cache */
   /* search1923_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5320, 268 },
      5,
   } },

   /* replace1923_0 -> 11 in the cache */
   /* replace1923_1 -> 268 in the cache */
   /* replace1923 -> 2233 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1924_0_0 -> 11 in the cache */
   /* search1924_0_1 -> 869 in the cache */
   /* search1924_0 -> 3641 in the cache */
   /* search1924_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3641, 268 },
      5,
   } },

   /* replace1924_0 -> 11 in the cache */
   /* replace1924_1 -> 268 in the cache */
   /* replace1924 -> 2233 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1925_0_0_0 -> 11 in the cache */
   /* search1925_0_0 -> 2177 in the cache */
   /* search1925_0 -> 5318 in the cache */
   /* search1925_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5318, 268 },
      5,
   } },

   /* replace1925_0 -> 11 in the cache */
   /* replace1925_1 -> 268 in the cache */
   /* replace1925 -> 2257 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1926_0_0_0 -> 11 in the cache */
   /* search1926_0_0 -> 4604 in the cache */
   /* search1926_0 -> 5320 in the cache */
   /* search1926_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5320, 268 },
      5,
   } },

   /* replace1926_0 -> 11 in the cache */
   /* replace1926_1 -> 268 in the cache */
   /* replace1926 -> 2257 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1927_0_0 -> 11 in the cache */
   /* search1927_0_1 -> 869 in the cache */
   /* search1927_0 -> 3641 in the cache */
   /* search1927_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3641, 268 },
      5,
   } },

   /* replace1927_0 -> 11 in the cache */
   /* replace1927_1 -> 268 in the cache */
   /* replace1927 -> 2257 in the cache */

   /* ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1928_0_0 -> 16 in the cache */
   /* search1928_0 -> 2193 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2193 },
      5,
   } },

   /* replace1928 -> 16 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1929_0 -> 16 in the cache */
   /* search1929_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 128 },
      5,
   } },

   /* replace1929 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1930_0 -> 531 in the cache */
   /* search1930_1_0_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5337 },
      -1,
   } },
   /* search1930_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5338, 569 },
      5,
   } },

   /* replace1930_0 -> 531 in the cache */
   /* replace1930_1 -> 189 in the cache */
   /* replace1930_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 189, 569 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1931_0 -> 531 in the cache */
   /* search1931_1_0_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5341 },
      -1,
   } },
   /* search1931_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5342, 569 },
      5,
   } },

   /* replace1931_0 -> 531 in the cache */
   /* replace1931_1 -> 189 in the cache */
   /* replace1931_2 -> 569 in the cache */
   /* replace1931 -> 5340 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1932_0 -> 531 in the cache */
   /* search1932_1_0 -> 189 in the cache */
   /* search1932_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 189, 128 },
      -1,
   } },
   /* search1932_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5344, 569 },
      5,
   } },

   /* replace1932_0 -> 531 in the cache */
   /* replace1932_1 -> 189 in the cache */
   /* replace1932_2 -> 569 in the cache */
   /* replace1932 -> 5340 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1933_0_0_0 -> 16 in the cache */
   /* search1933_0_0 -> 2193 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2193 },
      -1,
   } },
   /* search1933_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5346, 36 },
      5,
   } },

   /* replace1933_0 -> 16 in the cache */
   /* replace1933_1 -> 36 in the cache */
   /* replace1933 -> 4290 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1934_0_0_0 -> 16 in the cache */
   /* search1934_0_0 -> 4556 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4556 },
      -1,
   } },
   /* search1934_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5348, 36 },
      5,
   } },

   /* replace1934_0 -> 16 in the cache */
   /* replace1934_1 -> 36 in the cache */
   /* replace1934 -> 4290 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1935_0_0 -> 16 in the cache */
   /* search1935_0_1 -> 128 in the cache */
   /* search1935_0 -> 422 in the cache */
   /* search1935_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 422, 36 },
      5,
   } },

   /* replace1935_0 -> 16 in the cache */
   /* replace1935_1 -> 36 in the cache */
   /* replace1935 -> 4290 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1936_0_0_0 -> 16 in the cache */
   /* search1936_0_0 -> 2193 in the cache */
   /* search1936_0 -> 5346 in the cache */
   /* search1936_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5346, 36 },
      5,
   } },

   /* replace1936_0 -> 16 in the cache */
   /* replace1936_1 -> 36 in the cache */
   /* replace1936 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1937_0_0_0 -> 16 in the cache */
   /* search1937_0_0 -> 4556 in the cache */
   /* search1937_0 -> 5348 in the cache */
   /* search1937_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5348, 36 },
      5,
   } },

   /* replace1937_0 -> 16 in the cache */
   /* replace1937_1 -> 36 in the cache */
   /* replace1937 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1938_0_0 -> 16 in the cache */
   /* search1938_0_1 -> 128 in the cache */
   /* search1938_0 -> 422 in the cache */
   /* search1938_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 422, 36 },
      5,
   } },

   /* replace1938_0 -> 16 in the cache */
   /* replace1938_1 -> 36 in the cache */
   /* replace1938 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1939_0_0_0 -> 16 in the cache */
   /* search1939_0_0 -> 2193 in the cache */
   /* search1939_0 -> 5346 in the cache */
   /* search1939_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5346, 36 },
      5,
   } },

   /* replace1939_0 -> 16 in the cache */
   /* replace1939_1 -> 36 in the cache */
   /* replace1939 -> 2199 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1940_0_0_0 -> 16 in the cache */
   /* search1940_0_0 -> 4556 in the cache */
   /* search1940_0 -> 5348 in the cache */
   /* search1940_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5348, 36 },
      5,
   } },

   /* replace1940_0 -> 16 in the cache */
   /* replace1940_1 -> 36 in the cache */
   /* replace1940 -> 2199 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1941_0_0 -> 16 in the cache */
   /* search1941_0_1 -> 128 in the cache */
   /* search1941_0 -> 422 in the cache */
   /* search1941_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 422, 36 },
      5,
   } },

   /* replace1941_0 -> 16 in the cache */
   /* replace1941_1 -> 36 in the cache */
   /* replace1941 -> 2199 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1942_0_0_0 -> 16 in the cache */
   /* search1942_0_0 -> 2193 in the cache */
   /* search1942_0 -> 5346 in the cache */
   /* search1942_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5346, 36 },
      5,
   } },

   /* replace1942_0 -> 16 in the cache */
   /* replace1942_1 -> 36 in the cache */
   /* replace1942 -> 2237 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1943_0_0_0 -> 16 in the cache */
   /* search1943_0_0 -> 4556 in the cache */
   /* search1943_0 -> 5348 in the cache */
   /* search1943_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5348, 36 },
      5,
   } },

   /* replace1943_0 -> 16 in the cache */
   /* replace1943_1 -> 36 in the cache */
   /* replace1943 -> 2237 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1944_0_0 -> 16 in the cache */
   /* search1944_0_1 -> 128 in the cache */
   /* search1944_0 -> 422 in the cache */
   /* search1944_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 422, 36 },
      5,
   } },

   /* replace1944_0 -> 16 in the cache */
   /* replace1944_1 -> 36 in the cache */
   /* replace1944 -> 2237 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1945_0_0_0 -> 16 in the cache */
   /* search1945_0_0 -> 2193 in the cache */
   /* search1945_0 -> 5346 in the cache */
   /* search1945_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5346, 36 },
      5,
   } },

   /* replace1945_0 -> 16 in the cache */
   /* replace1945_1 -> 36 in the cache */
   /* replace1945 -> 2261 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1946_0_0_0 -> 16 in the cache */
   /* search1946_0_0 -> 4556 in the cache */
   /* search1946_0 -> 5348 in the cache */
   /* search1946_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5348, 36 },
      5,
   } },

   /* replace1946_0 -> 16 in the cache */
   /* replace1946_1 -> 36 in the cache */
   /* replace1946 -> 2261 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1947_0_0 -> 16 in the cache */
   /* search1947_0_1 -> 128 in the cache */
   /* search1947_0 -> 422 in the cache */
   /* search1947_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 422, 36 },
      5,
   } },

   /* replace1947_0 -> 16 in the cache */
   /* replace1947_1 -> 36 in the cache */
   /* replace1947 -> 2261 in the cache */

   /* ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1948_0_0 -> 16 in the cache */
   /* search1948_0 -> 2193 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2193 },
      5,
   } },

   /* replace1948 -> 16 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1949_0 -> 16 in the cache */
   /* search1949_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 128 },
      5,
   } },

   /* replace1949 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1950_0 -> 531 in the cache */
   /* search1950_1_0_0 -> 189 in the cache */
   /* search1950_1_0 -> 5337 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5337 },
      -1,
   } },
   /* search1950_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5365, 569 },
      5,
   } },

   /* replace1950_0 -> 531 in the cache */
   /* replace1950_1 -> 189 in the cache */
   /* replace1950_2 -> 569 in the cache */
   /* replace1950 -> 5340 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1951_0 -> 531 in the cache */
   /* search1951_1_0_0 -> 189 in the cache */
   /* search1951_1_0 -> 5341 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5341 },
      -1,
   } },
   /* search1951_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5367, 569 },
      5,
   } },

   /* replace1951_0 -> 531 in the cache */
   /* replace1951_1 -> 189 in the cache */
   /* replace1951_2 -> 569 in the cache */
   /* replace1951 -> 5340 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1952_0 -> 531 in the cache */
   /* search1952_1_0 -> 189 in the cache */
   /* search1952_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 189, 128 },
      -1,
   } },
   /* search1952_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5369, 569 },
      5,
   } },

   /* replace1952_0 -> 531 in the cache */
   /* replace1952_1 -> 189 in the cache */
   /* replace1952_2 -> 569 in the cache */
   /* replace1952 -> 5340 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1953_0_0_0 -> 16 in the cache */
   /* search1953_0_0 -> 2193 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2193 },
      -1,
   } },
   /* search1953_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5371, 36 },
      5,
   } },

   /* replace1953_0 -> 16 in the cache */
   /* replace1953_1 -> 36 in the cache */
   /* replace1953 -> 4290 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1954_0_0_0 -> 16 in the cache */
   /* search1954_0_0 -> 4556 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4556 },
      -1,
   } },
   /* search1954_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5373, 36 },
      5,
   } },

   /* replace1954_0 -> 16 in the cache */
   /* replace1954_1 -> 36 in the cache */
   /* replace1954 -> 4290 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1955_0_0 -> 16 in the cache */
   /* search1955_0_1 -> 128 in the cache */
   /* search1955_0 -> 403 in the cache */
   /* search1955_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 403, 36 },
      5,
   } },

   /* replace1955_0 -> 16 in the cache */
   /* replace1955_1 -> 36 in the cache */
   /* replace1955 -> 4290 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1956_0_0_0 -> 16 in the cache */
   /* search1956_0_0 -> 2193 in the cache */
   /* search1956_0 -> 5371 in the cache */
   /* search1956_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5371, 36 },
      5,
   } },

   /* replace1956_0 -> 16 in the cache */
   /* replace1956_1 -> 36 in the cache */
   /* replace1956 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1957_0_0_0 -> 16 in the cache */
   /* search1957_0_0 -> 4556 in the cache */
   /* search1957_0 -> 5373 in the cache */
   /* search1957_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5373, 36 },
      5,
   } },

   /* replace1957_0 -> 16 in the cache */
   /* replace1957_1 -> 36 in the cache */
   /* replace1957 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1958_0_0 -> 16 in the cache */
   /* search1958_0_1 -> 128 in the cache */
   /* search1958_0 -> 403 in the cache */
   /* search1958_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 403, 36 },
      5,
   } },

   /* replace1958_0 -> 16 in the cache */
   /* replace1958_1 -> 36 in the cache */
   /* replace1958 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1959_0_0_0 -> 16 in the cache */
   /* search1959_0_0 -> 2193 in the cache */
   /* search1959_0 -> 5371 in the cache */
   /* search1959_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5371, 36 },
      5,
   } },

   /* replace1959_0 -> 16 in the cache */
   /* replace1959_1 -> 36 in the cache */
   /* replace1959 -> 2199 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1960_0_0_0 -> 16 in the cache */
   /* search1960_0_0 -> 4556 in the cache */
   /* search1960_0 -> 5373 in the cache */
   /* search1960_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5373, 36 },
      5,
   } },

   /* replace1960_0 -> 16 in the cache */
   /* replace1960_1 -> 36 in the cache */
   /* replace1960 -> 2199 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1961_0_0 -> 16 in the cache */
   /* search1961_0_1 -> 128 in the cache */
   /* search1961_0 -> 403 in the cache */
   /* search1961_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 403, 36 },
      5,
   } },

   /* replace1961_0 -> 16 in the cache */
   /* replace1961_1 -> 36 in the cache */
   /* replace1961 -> 2199 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1962_0_0_0 -> 16 in the cache */
   /* search1962_0_0 -> 2193 in the cache */
   /* search1962_0 -> 5371 in the cache */
   /* search1962_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5371, 36 },
      5,
   } },

   /* replace1962_0 -> 16 in the cache */
   /* replace1962_1 -> 36 in the cache */
   /* replace1962 -> 2237 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1963_0_0_0 -> 16 in the cache */
   /* search1963_0_0 -> 4556 in the cache */
   /* search1963_0 -> 5373 in the cache */
   /* search1963_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5373, 36 },
      5,
   } },

   /* replace1963_0 -> 16 in the cache */
   /* replace1963_1 -> 36 in the cache */
   /* replace1963 -> 2237 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1964_0_0 -> 16 in the cache */
   /* search1964_0_1 -> 128 in the cache */
   /* search1964_0 -> 403 in the cache */
   /* search1964_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 403, 36 },
      5,
   } },

   /* replace1964_0 -> 16 in the cache */
   /* replace1964_1 -> 36 in the cache */
   /* replace1964 -> 2237 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1965_0_0_0 -> 16 in the cache */
   /* search1965_0_0 -> 2193 in the cache */
   /* search1965_0 -> 5371 in the cache */
   /* search1965_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5371, 36 },
      5,
   } },

   /* replace1965_0 -> 16 in the cache */
   /* replace1965_1 -> 36 in the cache */
   /* replace1965 -> 2261 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1966_0_0_0 -> 16 in the cache */
   /* search1966_0_0 -> 4556 in the cache */
   /* search1966_0 -> 5373 in the cache */
   /* search1966_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5373, 36 },
      5,
   } },

   /* replace1966_0 -> 16 in the cache */
   /* replace1966_1 -> 36 in the cache */
   /* replace1966 -> 2261 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1967_0_0 -> 16 in the cache */
   /* search1967_0_1 -> 128 in the cache */
   /* search1967_0 -> 403 in the cache */
   /* search1967_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 403, 36 },
      5,
   } },

   /* replace1967_0 -> 16 in the cache */
   /* replace1967_1 -> 36 in the cache */
   /* replace1967 -> 2261 in the cache */

   /* ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1968_0_0 -> 16 in the cache */
   /* search1968_0 -> 3505 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3505 },
      6,
   } },

   /* replace1968 -> 16 in the cache */

   /* ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1969_0 -> 16 in the cache */
   /* search1969_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 128 },
      6,
   } },

   /* replace1969 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1970_0 -> 531 in the cache */
   /* search1970_1_0_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5390 },
      -1,
   } },
   /* search1970_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5391, 569 },
      6,
   } },

   /* replace1970_0 -> 531 in the cache */
   /* replace1970_1 -> 189 in the cache */
   /* replace1970_2 -> 569 in the cache */
   /* replace1970 -> 5340 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1971_0 -> 531 in the cache */
   /* search1971_1_0_0 -> 189 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5393 },
      -1,
   } },
   /* search1971_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5394, 569 },
      6,
   } },

   /* replace1971_0 -> 531 in the cache */
   /* replace1971_1 -> 189 in the cache */
   /* replace1971_2 -> 569 in the cache */
   /* replace1971 -> 5340 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1972_0 -> 531 in the cache */
   /* search1972_1_0 -> 189 in the cache */
   /* search1972_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 189, 128 },
      -1,
   } },
   /* search1972_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5396, 569 },
      6,
   } },

   /* replace1972_0 -> 531 in the cache */
   /* replace1972_1 -> 189 in the cache */
   /* replace1972_2 -> 569 in the cache */
   /* replace1972 -> 5340 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1973_0_0_0 -> 16 in the cache */
   /* search1973_0_0 -> 3505 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3505 },
      -1,
   } },
   /* search1973_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5398, 36 },
      6,
   } },

   /* replace1973_0 -> 16 in the cache */
   /* replace1973_1 -> 36 in the cache */
   /* replace1973 -> 4290 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1974_0_0_0 -> 16 in the cache */
   /* search1974_0_0 -> 4558 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4558 },
      -1,
   } },
   /* search1974_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5400, 36 },
      6,
   } },

   /* replace1974_0 -> 16 in the cache */
   /* replace1974_1 -> 36 in the cache */
   /* replace1974 -> 4290 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1975_0_0 -> 16 in the cache */
   /* search1975_0_1 -> 128 in the cache */
   /* search1975_0 -> 454 in the cache */
   /* search1975_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 454, 36 },
      6,
   } },

   /* replace1975_0 -> 16 in the cache */
   /* replace1975_1 -> 36 in the cache */
   /* replace1975 -> 4290 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1976_0_0_0 -> 16 in the cache */
   /* search1976_0_0 -> 3505 in the cache */
   /* search1976_0 -> 5398 in the cache */
   /* search1976_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5398, 36 },
      6,
   } },

   /* replace1976_0 -> 16 in the cache */
   /* replace1976_1 -> 36 in the cache */
   /* replace1976 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1977_0_0_0 -> 16 in the cache */
   /* search1977_0_0 -> 4558 in the cache */
   /* search1977_0 -> 5400 in the cache */
   /* search1977_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5400, 36 },
      6,
   } },

   /* replace1977_0 -> 16 in the cache */
   /* replace1977_1 -> 36 in the cache */
   /* replace1977 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1978_0_0 -> 16 in the cache */
   /* search1978_0_1 -> 128 in the cache */
   /* search1978_0 -> 454 in the cache */
   /* search1978_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 454, 36 },
      6,
   } },

   /* replace1978_0 -> 16 in the cache */
   /* replace1978_1 -> 36 in the cache */
   /* replace1978 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1979_0_0_0 -> 16 in the cache */
   /* search1979_0_0 -> 3505 in the cache */
   /* search1979_0 -> 5398 in the cache */
   /* search1979_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5398, 36 },
      6,
   } },

   /* replace1979_0 -> 16 in the cache */
   /* replace1979_1 -> 36 in the cache */
   /* replace1979 -> 2199 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1980_0_0_0 -> 16 in the cache */
   /* search1980_0_0 -> 4558 in the cache */
   /* search1980_0 -> 5400 in the cache */
   /* search1980_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5400, 36 },
      6,
   } },

   /* replace1980_0 -> 16 in the cache */
   /* replace1980_1 -> 36 in the cache */
   /* replace1980 -> 2199 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1981_0_0 -> 16 in the cache */
   /* search1981_0_1 -> 128 in the cache */
   /* search1981_0 -> 454 in the cache */
   /* search1981_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 454, 36 },
      6,
   } },

   /* replace1981_0 -> 16 in the cache */
   /* replace1981_1 -> 36 in the cache */
   /* replace1981 -> 2199 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1982_0_0_0 -> 16 in the cache */
   /* search1982_0_0 -> 3505 in the cache */
   /* search1982_0 -> 5398 in the cache */
   /* search1982_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5398, 36 },
      6,
   } },

   /* replace1982_0 -> 16 in the cache */
   /* replace1982_1 -> 36 in the cache */
   /* replace1982 -> 2237 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1983_0_0_0 -> 16 in the cache */
   /* search1983_0_0 -> 4558 in the cache */
   /* search1983_0 -> 5400 in the cache */
   /* search1983_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5400, 36 },
      6,
   } },

   /* replace1983_0 -> 16 in the cache */
   /* replace1983_1 -> 36 in the cache */
   /* replace1983 -> 2237 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1984_0_0 -> 16 in the cache */
   /* search1984_0_1 -> 128 in the cache */
   /* search1984_0 -> 454 in the cache */
   /* search1984_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 454, 36 },
      6,
   } },

   /* replace1984_0 -> 16 in the cache */
   /* replace1984_1 -> 36 in the cache */
   /* replace1984 -> 2237 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1985_0_0_0 -> 16 in the cache */
   /* search1985_0_0 -> 3505 in the cache */
   /* search1985_0 -> 5398 in the cache */
   /* search1985_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5398, 36 },
      6,
   } },

   /* replace1985_0 -> 16 in the cache */
   /* replace1985_1 -> 36 in the cache */
   /* replace1985 -> 2261 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1986_0_0_0 -> 16 in the cache */
   /* search1986_0_0 -> 4558 in the cache */
   /* search1986_0 -> 5400 in the cache */
   /* search1986_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5400, 36 },
      6,
   } },

   /* replace1986_0 -> 16 in the cache */
   /* replace1986_1 -> 36 in the cache */
   /* replace1986 -> 2261 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1987_0_0 -> 16 in the cache */
   /* search1987_0_1 -> 128 in the cache */
   /* search1987_0 -> 454 in the cache */
   /* search1987_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 454, 36 },
      6,
   } },

   /* replace1987_0 -> 16 in the cache */
   /* replace1987_1 -> 36 in the cache */
   /* replace1987 -> 2261 in the cache */

   /* ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1988_0_0 -> 16 in the cache */
   /* search1988_0 -> 3505 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3505 },
      6,
   } },

   /* replace1988 -> 16 in the cache */

   /* ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1989_0 -> 16 in the cache */
   /* search1989_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 128 },
      6,
   } },

   /* replace1989 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1990_0 -> 531 in the cache */
   /* search1990_1_0_0 -> 189 in the cache */
   /* search1990_1_0 -> 5390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5390 },
      -1,
   } },
   /* search1990_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5417, 569 },
      6,
   } },

   /* replace1990_0 -> 531 in the cache */
   /* replace1990_1 -> 189 in the cache */
   /* replace1990_2 -> 569 in the cache */
   /* replace1990 -> 5340 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1991_0 -> 531 in the cache */
   /* search1991_1_0_0 -> 189 in the cache */
   /* search1991_1_0 -> 5393 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5393 },
      -1,
   } },
   /* search1991_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5419, 569 },
      6,
   } },

   /* replace1991_0 -> 531 in the cache */
   /* replace1991_1 -> 189 in the cache */
   /* replace1991_2 -> 569 in the cache */
   /* replace1991 -> 5340 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1992_0 -> 531 in the cache */
   /* search1992_1_0 -> 189 in the cache */
   /* search1992_1_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 189, 128 },
      -1,
   } },
   /* search1992_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 5421, 569 },
      6,
   } },

   /* replace1992_0 -> 531 in the cache */
   /* replace1992_1 -> 189 in the cache */
   /* replace1992_2 -> 569 in the cache */
   /* replace1992 -> 5340 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1993_0_0_0 -> 16 in the cache */
   /* search1993_0_0 -> 3505 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3505 },
      -1,
   } },
   /* search1993_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5423, 36 },
      6,
   } },

   /* replace1993_0 -> 16 in the cache */
   /* replace1993_1 -> 36 in the cache */
   /* replace1993 -> 4290 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1994_0_0_0 -> 16 in the cache */
   /* search1994_0_0 -> 4558 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4558 },
      -1,
   } },
   /* search1994_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5425, 36 },
      6,
   } },

   /* replace1994_0 -> 16 in the cache */
   /* replace1994_1 -> 36 in the cache */
   /* replace1994 -> 4290 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1995_0_0 -> 16 in the cache */
   /* search1995_0_1 -> 128 in the cache */
   /* search1995_0 -> 446 in the cache */
   /* search1995_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 446, 36 },
      6,
   } },

   /* replace1995_0 -> 16 in the cache */
   /* replace1995_1 -> 36 in the cache */
   /* replace1995 -> 4290 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1996_0_0_0 -> 16 in the cache */
   /* search1996_0_0 -> 3505 in the cache */
   /* search1996_0 -> 5423 in the cache */
   /* search1996_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5423, 36 },
      6,
   } },

   /* replace1996_0 -> 16 in the cache */
   /* replace1996_1 -> 36 in the cache */
   /* replace1996 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1997_0_0_0 -> 16 in the cache */
   /* search1997_0_0 -> 4558 in the cache */
   /* search1997_0 -> 5425 in the cache */
   /* search1997_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5425, 36 },
      6,
   } },

   /* replace1997_0 -> 16 in the cache */
   /* replace1997_1 -> 36 in the cache */
   /* replace1997 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1998_0_0 -> 16 in the cache */
   /* search1998_0_1 -> 128 in the cache */
   /* search1998_0 -> 446 in the cache */
   /* search1998_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 446, 36 },
      6,
   } },

   /* replace1998_0 -> 16 in the cache */
   /* replace1998_1 -> 36 in the cache */
   /* replace1998 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1999_0_0_0 -> 16 in the cache */
   /* search1999_0_0 -> 3505 in the cache */
   /* search1999_0 -> 5423 in the cache */
   /* search1999_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5423, 36 },
      6,
   } },

   /* replace1999_0 -> 16 in the cache */
   /* replace1999_1 -> 36 in the cache */
   /* replace1999 -> 2199 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search2000_0_0_0 -> 16 in the cache */
   /* search2000_0_0 -> 4558 in the cache */
   /* search2000_0 -> 5425 in the cache */
   /* search2000_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5425, 36 },
      6,
   } },

   /* replace2000_0 -> 16 in the cache */
   /* replace2000_1 -> 36 in the cache */
   /* replace2000 -> 2199 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search2001_0_0 -> 16 in the cache */
   /* search2001_0_1 -> 128 in the cache */
   /* search2001_0 -> 446 in the cache */
   /* search2001_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 446, 36 },
      6,
   } },

   /* replace2001_0 -> 16 in the cache */
   /* replace2001_1 -> 36 in the cache */
   /* replace2001 -> 2199 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search2002_0_0_0 -> 16 in the cache */
   /* search2002_0_0 -> 3505 in the cache */
   /* search2002_0 -> 5423 in the cache */
   /* search2002_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5423, 36 },
      6,
   } },

   /* replace2002_0 -> 16 in the cache */
   /* replace2002_1 -> 36 in the cache */
   /* replace2002 -> 2237 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search2003_0_0_0 -> 16 in the cache */
   /* search2003_0_0 -> 4558 in the cache */
   /* search2003_0 -> 5425 in the cache */
   /* search2003_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5425, 36 },
      6,
   } },

   /* replace2003_0 -> 16 in the cache */
   /* replace2003_1 -> 36 in the cache */
   /* replace2003 -> 2237 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search2004_0_0 -> 16 in the cache */
   /* search2004_0_1 -> 128 in the cache */
   /* search2004_0 -> 446 in the cache */
   /* search2004_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 446, 36 },
      6,
   } },

   /* replace2004_0 -> 16 in the cache */
   /* replace2004_1 -> 36 in the cache */
   /* replace2004 -> 2237 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search2005_0_0_0 -> 16 in the cache */
   /* search2005_0_0 -> 3505 in the cache */
   /* search2005_0 -> 5423 in the cache */
   /* search2005_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5423, 36 },
      6,
   } },

   /* replace2005_0 -> 16 in the cache */
   /* replace2005_1 -> 36 in the cache */
   /* replace2005 -> 2261 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search2006_0_0_0 -> 16 in the cache */
   /* search2006_0_0 -> 4558 in the cache */
   /* search2006_0 -> 5425 in the cache */
   /* search2006_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5425, 36 },
      6,
   } },

   /* replace2006_0 -> 16 in the cache */
   /* replace2006_1 -> 36 in the cache */
   /* replace2006 -> 2261 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search2007_0_0 -> 16 in the cache */
   /* search2007_0_1 -> 128 in the cache */
   /* search2007_0 -> 446 in the cache */
   /* search2007_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 446, 36 },
      6,
   } },

   /* replace2007_0 -> 16 in the cache */
   /* replace2007_1 -> 36 in the cache */
   /* replace2007 -> 2261 in the cache */

   /* ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* exp */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5440, 5441 },
      -1,
   } },

   /* replace2008_0_0 -> 5440 in the cache */
   /* replace2008_0_1_0_0_0_0_0_0 -> 5441 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1e /* -30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5441, 5443 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1e /* 30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5444, 5445 },
      -1,
   } },
   /* replace2008_0_1_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5446, 406 },
      -1,
   } },
   /* replace2008_0_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5447, 14 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xa /* 10 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5448, 5449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 3,
      { 5450 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5440, 5451 },
      -1,
   } },
   /* replace2008_1_0_0_0_0_0_0 -> 5441 in the cache */
   /* replace2008_1_0_0_0_0_0_1 -> 5443 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5441, 5443 },
      -1,
   } },
   /* replace2008_1_0_0_0_0_1 -> 5445 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5453, 5445 },
      -1,
   } },
   /* replace2008_1_0_0_0_1_0_0_0 -> 5441 in the cache */
   /* replace2008_1_0_0_0_1_0_0_1 -> 5443 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5441, 5443 },
      -1,
   } },
   /* replace2008_1_0_0_0_1_0_1 -> 5445 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5455, 5445 },
      -1,
   } },
   /* replace2008_1_0_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5456, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5454, 5457 },
      -1,
   } },
   /* replace2008_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5458, 14 },
      -1,
   } },
   /* replace2008_1_0_1 -> 5449 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5459, 5449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 5,
      { 5460 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5452, 5461 },
      -1,
   } },

   /* ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23)) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2009_1 -> 5441 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5463, 5441 },
      -1,
   } },

   /* replace2009_0_0 -> 5463 in the cache */
   /* replace2009_0_1_0_0_0_0_0 -> 5441 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xfe /* -254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5441, 5465 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfe /* 254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5466, 5467 },
      -1,
   } },
   /* replace2009_0_1_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5468, 406 },
      -1,
   } },
   /* replace2009_0_1_0_1 -> 4642 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5469, 4642 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x17 /* 23 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5470, 5471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5463, 5472 },
      -1,
   } },
   /* replace2009_1_0_0_0_0_0 -> 5441 in the cache */
   /* replace2009_1_0_0_0_0_1 -> 5465 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5441, 5465 },
      -1,
   } },
   /* replace2009_1_0_0_0_1 -> 5467 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5474, 5467 },
      -1,
   } },
   /* replace2009_1_0_0_1_0_0_0 -> 5441 in the cache */
   /* replace2009_1_0_0_1_0_0_1 -> 5465 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5441, 5465 },
      -1,
   } },
   /* replace2009_1_0_0_1_0_1 -> 5467 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5476, 5467 },
      -1,
   } },
   /* replace2009_1_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5477, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5475, 5478 },
      -1,
   } },
   /* replace2009_1_0_1 -> 4642 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5479, 4642 },
      -1,
   } },
   /* replace2009_1_1 -> 5471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5480, 5471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5473, 5481 },
      -1,
   } },

   /* ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2010_1 -> 5441 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5483, 5441 },
      -1,
   } },

   /* replace2010_0_0 -> 5483 in the cache */
   /* replace2010_0_1_0 -> 128 in the cache */
   /* replace2010_0_1_1_0_0_0_0_0 -> 5441 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fe /* -2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5441, 5485 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fe /* 2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5486, 5487 },
      -1,
   } },
   /* replace2010_0_1_1_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5488, 406 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3ff /* 1023 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5489, 5490 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x14 /* 20 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5491, 5492 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 3,
      { 128, 5493 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5483, 5494 },
      -1,
   } },
   /* replace2010_1_0 -> 128 in the cache */
   /* replace2010_1_1_0_0_0_0_0 -> 5441 in the cache */
   /* replace2010_1_1_0_0_0_0_1 -> 5485 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5441, 5485 },
      -1,
   } },
   /* replace2010_1_1_0_0_0_1 -> 5487 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5496, 5487 },
      -1,
   } },
   /* replace2010_1_1_0_0_1_0_0_0 -> 5441 in the cache */
   /* replace2010_1_1_0_0_1_0_0_1 -> 5485 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5441, 5485 },
      -1,
   } },
   /* replace2010_1_1_0_0_1_0_1 -> 5487 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5498, 5487 },
      -1,
   } },
   /* replace2010_1_1_0_0_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5499, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5497, 5500 },
      -1,
   } },
   /* replace2010_1_1_0_1 -> 5490 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5501, 5490 },
      -1,
   } },
   /* replace2010_1_1_1 -> 5492 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5502, 5492 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 5,
      { 128, 5503 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5495, 5504 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x') */
   /* search2011_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 5463, 866 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 5463, 866 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      8, 1,
      { 5506, 5507 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff /* 16711935 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 5508, 5509 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_0_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5510, 844 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      10, 1,
      { 5506, 5507 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff00 /* 4278255360 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 5512, 5513 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_0_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5514, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 5511, 5515 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f /* 252645135 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 5516, 5517 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5518, 2366 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      14, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 5520, 5509 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5521, 844 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      16, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 5523, 5513 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5524, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 5522, 5525 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f0 /* 4042322160 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 5526, 5527 },
      -1,
   } },
   /* search2011_0_0_0_0_0_0_1_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5528, 2366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 5519, 5529 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x33333333 /* 858993459 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 5530, 5531 },
      -1,
   } },
   /* search2011_0_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 5532, 411 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      22, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 5534, 5509 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5535, 844 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      24, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 5537, 5513 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5538, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 5536, 5539 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_0_1 -> 5517 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 5540, 5517 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5541, 2366 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      28, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 5543, 5509 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5544, 844 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      30, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 5546, 5513 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5547, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 5545, 5548 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_0_1 -> 5527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 5549, 5527 },
      -1,
   } },
   /* search2011_0_0_0_1_0_0_1_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5550, 2366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 5542, 5551 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xcccccccc /* 3435973836 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 5552, 5553 },
      -1,
   } },
   /* search2011_0_0_0_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 5554, 411 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 5533, 5555 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x55555555 /* 1431655765 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 5556, 5557 },
      -1,
   } },
   /* search2011_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 30,
      { 5558, 406 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      38, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 5560, 5509 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5561, 844 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      40, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 5563, 5513 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5564, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 5562, 5565 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_0_1 -> 5517 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 5566, 5517 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5567, 2366 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      44, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 5569, 5509 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5570, 844 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      46, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 5572, 5513 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5573, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 5571, 5574 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_0_1 -> 5527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 5575, 5527 },
      -1,
   } },
   /* search2011_1_0_0_0_0_0_1_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5576, 2366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 5568, 5577 },
      -1,
   } },
   /* search2011_1_0_0_0_0_1 -> 5531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 5578, 5531 },
      -1,
   } },
   /* search2011_1_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 5579, 411 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      52, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 5581, 5509 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5582, 844 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      54, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 5584, 5513 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5585, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 5583, 5586 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_0_1 -> 5517 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 5587, 5517 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5588, 2366 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      58, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_0_0_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 5590, 5509 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5591, 844 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      60, 1,
      { 5506, 5507 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_1_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 5593, 5513 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_0_1_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5594, 844 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 5592, 5595 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_0_1 -> 5527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 5596, 5527 },
      -1,
   } },
   /* search2011_1_0_0_1_0_0_1_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5597, 2366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 5589, 5598 },
      -1,
   } },
   /* search2011_1_0_0_1_0_1 -> 5553 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 5599, 5553 },
      -1,
   } },
   /* search2011_1_0_0_1_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 5600, 411 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 5580, 5601 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xaaaaaaaa /* 2863311530 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 5602, 5603 },
      -1,
   } },
   /* search2011_1_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 30,
      { 5604, 406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 5559, 5605 },
      -1,
   } },

   /* replace2011_0 -> 5463 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_reverse,
      -1, 0,
      { 5463 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x') */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_0 -> 5508 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5508, 406 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_0_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 5608, 5603 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_0 -> 5512 in the cache */
   /* search2012_0_0_0_0_0_0_0_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5512, 406 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_0_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 5610, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 5609, 5611 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5612, 411 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_0_1 -> 5553 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 5613, 5553 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_0 -> 5520 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5520, 406 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 5615, 5603 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_0 -> 5523 in the cache */
   /* search2012_0_0_0_0_0_0_1_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5523, 406 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 5617, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 5616, 5618 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5619, 411 },
      -1,
   } },
   /* search2012_0_0_0_0_0_0_1_1 -> 5531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 5620, 5531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 5614, 5621 },
      -1,
   } },
   /* search2012_0_0_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5622, 2366 },
      -1,
   } },
   /* search2012_0_0_0_0_1 -> 5527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 5623, 5527 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_0 -> 5534 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5534, 406 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 5625, 5603 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_0 -> 5537 in the cache */
   /* search2012_0_0_0_1_0_0_0_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5537, 406 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 5627, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 5626, 5628 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5629, 411 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_0_1 -> 5553 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 5630, 5553 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_0 -> 5543 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5543, 406 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 5632, 5603 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_0 -> 5546 in the cache */
   /* search2012_0_0_0_1_0_0_1_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5546, 406 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 5634, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 5633, 5635 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5636, 411 },
      -1,
   } },
   /* search2012_0_0_0_1_0_0_1_1 -> 5531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 5637, 5531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 5631, 5638 },
      -1,
   } },
   /* search2012_0_0_0_1_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5639, 2366 },
      -1,
   } },
   /* search2012_0_0_0_1_1 -> 5517 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 5640, 5517 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 5624, 5641 },
      -1,
   } },
   /* search2012_0_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 29,
      { 5642, 844 },
      -1,
   } },
   /* search2012_0_1 -> 5513 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 5643, 5513 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_0 -> 5560 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5560, 406 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 5645, 5603 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_0 -> 5563 in the cache */
   /* search2012_1_0_0_0_0_0_0_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5563, 406 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 5647, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 5646, 5648 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5649, 411 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_0_1 -> 5553 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 5650, 5553 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_0 -> 5569 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5569, 406 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 5652, 5603 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_0 -> 5572 in the cache */
   /* search2012_1_0_0_0_0_0_1_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5572, 406 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 5654, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 5653, 5655 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5656, 411 },
      -1,
   } },
   /* search2012_1_0_0_0_0_0_1_1 -> 5531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 5657, 5531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 5651, 5658 },
      -1,
   } },
   /* search2012_1_0_0_0_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5659, 2366 },
      -1,
   } },
   /* search2012_1_0_0_0_1 -> 5527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 5660, 5527 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_0 -> 5581 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5581, 406 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 5662, 5603 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_0 -> 5584 in the cache */
   /* search2012_1_0_0_1_0_0_0_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5584, 406 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 5664, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 5663, 5665 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5666, 411 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_0_1 -> 5553 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 5667, 5553 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_0 -> 5590 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_0_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5590, 406 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_0_0_0_1 -> 5603 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 5669, 5603 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 5506 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5463 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 866 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 5507 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_0 -> 5593 in the cache */
   /* search2012_1_0_0_1_0_0_1_0_0_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5593, 406 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_0_0_1_1 -> 5557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 5671, 5557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 5670, 5672 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_0_1 -> 411 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5673, 411 },
      -1,
   } },
   /* search2012_1_0_0_1_0_0_1_1 -> 5531 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 5674, 5531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 5668, 5675 },
      -1,
   } },
   /* search2012_1_0_0_1_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5676, 2366 },
      -1,
   } },
   /* search2012_1_0_0_1_1 -> 5517 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 5677, 5517 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 5661, 5678 },
      -1,
   } },
   /* search2012_1_0_1 -> 844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 29,
      { 5679, 844 },
      -1,
   } },
   /* search2012_1_1 -> 5509 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 5680, 5509 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 5644, 5681 },
      -1,
   } },

   /* replace2012_0 -> 5463 in the cache */
   /* replace2012 -> 5607 in the cache */

   /* ('bcsel', ('iand', ('ine', ('fabs', 'x'), 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', 'x', 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', 'x', 0), 4294967295), ('pack_half_2x16_rtz_split', 'x', 0)) => ('pack_half_2x16_rtz_split', 'x', 0) */
   /* search2013_0_0_0_0 -> 5463 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5463 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7f800000 /* 2139095040 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 5683, 5684 },
      -1,
   } },
   /* search2013_0_1_0_0_0 -> 5463 in the cache */
   /* search2013_0_1_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 5463, 128 },
      -1,
   } },
   /* search2013_0_1_0_1 -> 4755 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 5686, 4755 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7c00 /* 31744 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 5687, 5688 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 5685, 5689 },
      -1,
   } },
   /* search2013_1_0_0 -> 5463 in the cache */
   /* search2013_1_0_1 -> 128 in the cache */
   /* search2013_1_0 -> 5686 in the cache */
   /* search2013_1_1 -> 940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      4, 1,
      { 5686, 940 },
      -1,
   } },
   /* search2013_2_0 -> 5463 in the cache */
   /* search2013_2_1 -> 128 in the cache */
   /* search2013_2 -> 5686 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 5,
      { 5690, 5691, 5686 },
      -1,
   } },

   /* replace2013_0 -> 5463 in the cache */
   /* replace2013_1 -> 128 in the cache */
   /* replace2013 -> 5686 in the cache */

   /* ('bcsel', ('iand', ('ine', 'x', 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0), 4294967295), ('pack_half_2x16_rtz_split', 'x(is_not_negative)', 0)) => ('pack_half_2x16_rtz_split', 'x', 0) */
   /* search2014_0_0_0 -> 5463 in the cache */
   /* search2014_0_0_1 -> 5684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 5463, 5684 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* x */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2014_0_1_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 5694, 128 },
      -1,
   } },
   /* search2014_0_1_0_1 -> 4755 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 5695, 4755 },
      -1,
   } },
   /* search2014_0_1_1 -> 5688 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 5696, 5688 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 5693, 5697 },
      -1,
   } },
   /* search2014_1_0_0 -> 5694 in the cache */
   /* search2014_1_0_1 -> 128 in the cache */
   /* search2014_1_0 -> 5695 in the cache */
   /* search2014_1_1 -> 940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      4, 1,
      { 5695, 940 },
      -1,
   } },
   /* search2014_2_0 -> 5694 in the cache */
   /* search2014_2_1 -> 128 in the cache */
   /* search2014_2 -> 5695 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 5,
      { 5698, 5699, 5695 },
      -1,
   } },

   /* replace2014_0 -> 5463 in the cache */
   /* replace2014_1 -> 128 in the cache */
   /* replace2014 -> 5686 in the cache */

   /* ('bcsel', ('iand', ('ine', ('fabs', 'x'), 2139095040), ('ieq', ('iand', ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0), 32767), 31744)), ('iadd', ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0), 4294967295), ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0)) => ('pack_half_2x16_rtz_split', ('fneg', 'x'), 0) */
   /* search2015_0_0_0_0 -> 5463 in the cache */
   /* search2015_0_0_0 -> 5683 in the cache */
   /* search2015_0_0_1 -> 5684 in the cache */
   /* search2015_0_0 -> 5685 in the cache */
   /* search2015_0_1_0_0_0_0 -> 5463 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5463 },
      -1,
   } },
   /* search2015_0_1_0_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_rtz_split,
      -1, 0,
      { 5701, 128 },
      -1,
   } },
   /* search2015_0_1_0_1 -> 4755 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 5702, 4755 },
      -1,
   } },
   /* search2015_0_1_1 -> 5688 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 5703, 5688 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 5685, 5704 },
      -1,
   } },
   /* search2015_1_0_0_0 -> 5463 in the cache */
   /* search2015_1_0_0 -> 5701 in the cache */
   /* search2015_1_0_1 -> 128 in the cache */
   /* search2015_1_0 -> 5702 in the cache */
   /* search2015_1_1 -> 940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      4, 1,
      { 5702, 940 },
      -1,
   } },
   /* search2015_2_0_0 -> 5463 in the cache */
   /* search2015_2_0 -> 5701 in the cache */
   /* search2015_2_1 -> 128 in the cache */
   /* search2015_2 -> 5702 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 5,
      { 5705, 5706, 5702 },
      -1,
   } },

   /* replace2015_0_0 -> 5463 in the cache */
   /* replace2015_0 -> 5701 in the cache */
   /* replace2015_1 -> 128 in the cache */
   /* replace2015 -> 5702 in the cache */

   /* ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b') */
   /* search2016_0_0 -> 0 in the cache */
   /* search2016_0_1 -> 3 in the cache */
   /* search2016_0 -> 1093 in the cache */
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1093, 5708 },
      -1,
   } },

   /* replace2016_0 -> 0 in the cache */
   /* replace2016_1 -> 3 in the cache */
   /* replace2016 -> 2632 in the cache */

   /* ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b') */
   /* search2017_0_0 -> 0 in the cache */
   /* search2017_0_1 -> 3 in the cache */
   /* search2017_0 -> 1087 in the cache */
   /* search2017_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1087, 5708 },
      -1,
   } },

   /* replace2017_0 -> 0 in the cache */
   /* replace2017_1 -> 3 in the cache */
   /* replace2017 -> 2702 in the cache */

   /* ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b') */
   /* search2018_0_0 -> 0 in the cache */
   /* search2018_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 5711, 5712 },
      -1,
   } },

   /* replace2018_0 -> 0 in the cache */
   /* replace2018_1 -> 3 in the cache */
   /* replace2018 -> 2743 in the cache */

   /* ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b') */
   /* search2019_0_0 -> 0 in the cache */
   /* search2019_0_1 -> 3 in the cache */
   /* search2019_0 -> 4424 in the cache */
   /* search2019_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 4424, 5712 },
      -1,
   } },

   /* replace2019_0 -> 0 in the cache */
   /* replace2019_1 -> 3 in the cache */
   /* replace2019 -> 2784 in the cache */

   /* ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b') */
   /* search2020_0_0 -> 0 in the cache */
   /* search2020_0_1 -> 3 in the cache */
   /* search2020_0 -> 1093 in the cache */
   /* search2020_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1093, 5708 },
      -1,
   } },

   /* replace2020_0 -> 0 in the cache */
   /* replace2020_1 -> 3 in the cache */
   /* replace2020 -> 2638 in the cache */

   /* ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b') */
   /* search2021_0_0 -> 0 in the cache */
   /* search2021_0_1 -> 3 in the cache */
   /* search2021_0 -> 1087 in the cache */
   /* search2021_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1087, 5708 },
      -1,
   } },

   /* replace2021_0 -> 0 in the cache */
   /* replace2021_1 -> 3 in the cache */
   /* replace2021 -> 2706 in the cache */

   /* ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b') */
   /* search2022_0_0 -> 0 in the cache */
   /* search2022_0_1 -> 3 in the cache */
   /* search2022_0 -> 5711 in the cache */
   /* search2022_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 5711, 5712 },
      -1,
   } },

   /* replace2022_0 -> 0 in the cache */
   /* replace2022_1 -> 3 in the cache */
   /* replace2022 -> 2747 in the cache */

   /* ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b') */
   /* search2023_0_0 -> 0 in the cache */
   /* search2023_0_1 -> 3 in the cache */
   /* search2023_0 -> 4424 in the cache */
   /* search2023_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 4424, 5712 },
      -1,
   } },

   /* replace2023_0 -> 0 in the cache */
   /* replace2023_1 -> 3 in the cache */
   /* replace2023 -> 2788 in the cache */

   /* ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b') */
   /* search2024_0_0 -> 0 in the cache */
   /* search2024_0_1 -> 3 in the cache */
   /* search2024_0 -> 1093 in the cache */
   /* search2024_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1093, 5708 },
      -1,
   } },

   /* replace2024_0 -> 0 in the cache */
   /* replace2024_1 -> 3 in the cache */
   /* replace2024 -> 2645 in the cache */

   /* ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b') */
   /* search2025_0_0 -> 0 in the cache */
   /* search2025_0_1 -> 3 in the cache */
   /* search2025_0 -> 1087 in the cache */
   /* search2025_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1087, 5708 },
      -1,
   } },

   /* replace2025_0 -> 0 in the cache */
   /* replace2025_1 -> 3 in the cache */
   /* replace2025 -> 2712 in the cache */

   /* ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b') */
   /* search2026_0_0 -> 0 in the cache */
   /* search2026_0_1 -> 3 in the cache */
   /* search2026_0 -> 5711 in the cache */
   /* search2026_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 5711, 5712 },
      -1,
   } },

   /* replace2026_0 -> 0 in the cache */
   /* replace2026_1 -> 3 in the cache */
   /* replace2026 -> 2753 in the cache */

   /* ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b') */
   /* search2027_0_0 -> 0 in the cache */
   /* search2027_0_1 -> 3 in the cache */
   /* search2027_0 -> 4424 in the cache */
   /* search2027_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 4424, 5712 },
      -1,
   } },

   /* replace2027_0 -> 0 in the cache */
   /* replace2027_1 -> 3 in the cache */
   /* replace2027 -> 2794 in the cache */

   /* ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b') */
   /* search2028_0_0 -> 0 in the cache */
   /* search2028_0_1 -> 3 in the cache */
   /* search2028_0 -> 1093 in the cache */
   /* search2028_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1093, 5708 },
      -1,
   } },

   /* replace2028_0 -> 0 in the cache */
   /* replace2028_1 -> 3 in the cache */
   /* replace2028 -> 2652 in the cache */

   /* ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b') */
   /* search2029_0_0 -> 0 in the cache */
   /* search2029_0_1 -> 3 in the cache */
   /* search2029_0 -> 1087 in the cache */
   /* search2029_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1087, 5708 },
      -1,
   } },

   /* replace2029_0 -> 0 in the cache */
   /* replace2029_1 -> 3 in the cache */
   /* replace2029 -> 2717 in the cache */

   /* ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b') */
   /* search2030_0_0 -> 0 in the cache */
   /* search2030_0_1 -> 3 in the cache */
   /* search2030_0 -> 5711 in the cache */
   /* search2030_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 5711, 5712 },
      -1,
   } },

   /* replace2030_0 -> 0 in the cache */
   /* replace2030_1 -> 3 in the cache */
   /* replace2030 -> 2758 in the cache */

   /* ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b') */
   /* search2031_0_0 -> 0 in the cache */
   /* search2031_0_1 -> 3 in the cache */
   /* search2031_0 -> 4424 in the cache */
   /* search2031_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 4424, 5712 },
      -1,
   } },

   /* replace2031_0 -> 0 in the cache */
   /* replace2031_1 -> 3 in the cache */
   /* replace2031 -> 2799 in the cache */

   /* ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b') */
   /* search2032_0_0 -> 0 in the cache */
   /* search2032_0_1 -> 3 in the cache */
   /* search2032_0 -> 1093 in the cache */
   /* search2032_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1093, 5708 },
      -1,
   } },

   /* replace2032_0 -> 0 in the cache */
   /* replace2032_1 -> 3 in the cache */
   /* replace2032 -> 2669 in the cache */

   /* ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b') */
   /* search2033_0_0 -> 0 in the cache */
   /* search2033_0_1 -> 3 in the cache */
   /* search2033_0 -> 1087 in the cache */
   /* search2033_1 -> 5708 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1087, 5708 },
      -1,
   } },

   /* replace2033_0 -> 0 in the cache */
   /* replace2033_1 -> 3 in the cache */
   /* replace2033 -> 2726 in the cache */

   /* ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b') */
   /* search2034_0_0 -> 0 in the cache */
   /* search2034_0_1 -> 3 in the cache */
   /* search2034_0 -> 5711 in the cache */
   /* search2034_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 5711, 5712 },
      -1,
   } },

   /* replace2034_0 -> 0 in the cache */
   /* replace2034_1 -> 3 in the cache */
   /* replace2034 -> 2767 in the cache */

   /* ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b') */
   /* search2035_0_0 -> 0 in the cache */
   /* search2035_0_1 -> 3 in the cache */
   /* search2035_0 -> 4424 in the cache */
   /* search2035_1 -> 5712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 4424, 5712 },
      -1,
   } },

   /* replace2035_0 -> 0 in the cache */
   /* replace2035_1 -> 3 in the cache */
   /* replace2035 -> 2808 in the cache */

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b') */
   /* search2036_0_0 -> 0 in the cache */
   /* search2036_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search2036_1_0 -> 0 in the cache */
   /* search2036_1_1 -> 3 in the cache */
   /* search2036_1 -> 1062 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5731, 1062 },
      -1,
   } },

   /* replace2036_0 -> 0 in the cache */
   /* replace2036_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a') */
   /* search2037_0_0 -> 32 in the cache */
   /* search2037_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   /* search2037_1_0 -> 2465 in the cache */
   /* search2037_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5734, 5735 },
      -1,
   } },

   /* replace2037_0 -> 2465 in the cache */
   /* replace2037_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 2465, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b') */
   /* search2038_0_0 -> 0 in the cache */
   /* search2038_0_1 -> 0 in the cache */
   /* search2038_0 -> 5731 in the cache */
   /* search2038_1_0 -> 0 in the cache */
   /* search2038_1_1 -> 3 in the cache */
   /* search2038_1 -> 1059 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5731, 1059 },
      -1,
   } },

   /* replace2038_0 -> 0 in the cache */
   /* replace2038_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a') */
   /* search2039_0_0 -> 32 in the cache */
   /* search2039_0_1 -> 32 in the cache */
   /* search2039_0 -> 5734 in the cache */
   /* search2039_1_0 -> 2465 in the cache */
   /* search2039_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5734, 5740 },
      -1,
   } },

   /* replace2039_0 -> 2465 in the cache */
   /* replace2039_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 2465, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b') */
   /* search2040_0_0 -> 0 in the cache */
   /* search2040_0_1 -> 0 in the cache */
   /* search2040_0 -> 5731 in the cache */
   /* search2040_1_0 -> 0 in the cache */
   /* search2040_1_1 -> 3 in the cache */
   /* search2040_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5731, 1087 },
      -1,
   } },

   /* replace2040_0 -> 0 in the cache */
   /* replace2040_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a') */
   /* search2041_0_0 -> 32 in the cache */
   /* search2041_0_1 -> 32 in the cache */
   /* search2041_0 -> 5734 in the cache */
   /* search2041_1_0 -> 2465 in the cache */
   /* search2041_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 2465, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5734, 5745 },
      -1,
   } },

   /* replace2041_0 -> 2465 in the cache */
   /* replace2041_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 2465, 32 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd')) */
   /* search2042_0_0 -> 770 in the cache */
   /* search2042_0_1 -> 1041 in the cache */
   /* search2042_0_2 -> 3703 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1041, 3703 },
      -1,
   } },
   /* search2042_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 5748, 1044 },
      -1,
   } },

   /* replace2042_0 -> 770 in the cache */
   /* replace2042_1_0 -> 156 in the cache */
   /* replace2042_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2042_2_0 -> 158 in the cache */
   /* replace2042_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5750, 5751 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd')) */
   /* search2043_0_0 -> 770 in the cache */
   /* search2043_0_1 -> 1041 in the cache */
   /* search2043_0_2 -> 3703 in the cache */
   /* search2043_0 -> 5748 in the cache */
   /* search2043_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 5748, 1044 },
      -1,
   } },

   /* replace2043_0 -> 770 in the cache */
   /* replace2043_1_0 -> 156 in the cache */
   /* replace2043_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2043_2_0 -> 158 in the cache */
   /* replace2043_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5754, 5755 },
      -1,
   } },

   /* ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd')) */
   /* search2044_0_0 -> 770 in the cache */
   /* search2044_0_1 -> 1041 in the cache */
   /* search2044_0_2 -> 3703 in the cache */
   /* search2044_0 -> 5748 in the cache */
   /* search2044_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 5748, 1044 },
      -1,
   } },

   /* replace2044_0 -> 770 in the cache */
   /* replace2044_1_0 -> 156 in the cache */
   /* replace2044_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2044_2_0 -> 158 in the cache */
   /* replace2044_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5758, 5759 },
      -1,
   } },

   /* ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd')) */
   /* search2045_0_0 -> 770 in the cache */
   /* search2045_0_1 -> 1041 in the cache */
   /* search2045_0_2 -> 3703 in the cache */
   /* search2045_0 -> 5748 in the cache */
   /* search2045_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 5748, 1044 },
      -1,
   } },

   /* replace2045_0 -> 770 in the cache */
   /* replace2045_1_0 -> 156 in the cache */
   /* replace2045_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2045_2_0 -> 158 in the cache */
   /* replace2045_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5762, 5763 },
      -1,
   } },

   /* ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd')) */
   /* search2046_0_0 -> 770 in the cache */
   /* search2046_0_1 -> 1041 in the cache */
   /* search2046_0_2 -> 3703 in the cache */
   /* search2046_0 -> 5748 in the cache */
   /* search2046_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2046_0 -> 770 in the cache */
   /* replace2046_1_0 -> 156 in the cache */
   /* replace2046_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2046_2_0 -> 158 in the cache */
   /* replace2046_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5766, 5767 },
      -1,
   } },

   /* ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2047_1_0 -> 1189 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5770, 5771 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2047_0 -> 1189 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2047_2_0 -> 5774 in the cache */
   /* replace2047_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5776, 5777 },
      -1,
   } },

   /* ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd')) */
   /* search2048_0_0 -> 770 in the cache */
   /* search2048_0_1 -> 1041 in the cache */
   /* search2048_0_2 -> 3703 in the cache */
   /* search2048_0 -> 5748 in the cache */
   /* search2048_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2048_0 -> 770 in the cache */
   /* replace2048_1_0 -> 156 in the cache */
   /* replace2048_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2048_2_0 -> 158 in the cache */
   /* replace2048_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5780, 5781 },
      -1,
   } },

   /* ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c')) */
   /* search2049_0 -> 5769 in the cache */
   /* search2049_1_0 -> 1189 in the cache */
   /* search2049_1_1 -> 5770 in the cache */
   /* search2049_1_2 -> 5771 in the cache */
   /* search2049_1 -> 5772 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2049_0 -> 1189 in the cache */
   /* replace2049_1_0 -> 5774 in the cache */
   /* replace2049_1_1 -> 5775 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2049_2_0 -> 5774 in the cache */
   /* replace2049_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5784, 5785 },
      -1,
   } },

   /* ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd')) */
   /* search2050_0_0 -> 770 in the cache */
   /* search2050_0_1 -> 1041 in the cache */
   /* search2050_0_2 -> 3703 in the cache */
   /* search2050_0 -> 5748 in the cache */
   /* search2050_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2050_0 -> 770 in the cache */
   /* replace2050_1_0 -> 156 in the cache */
   /* replace2050_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2050_2_0 -> 158 in the cache */
   /* replace2050_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5788, 5789 },
      -1,
   } },

   /* ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c')) */
   /* search2051_0 -> 5769 in the cache */
   /* search2051_1_0 -> 1189 in the cache */
   /* search2051_1_1 -> 5770 in the cache */
   /* search2051_1_2 -> 5771 in the cache */
   /* search2051_1 -> 5772 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2051_0 -> 1189 in the cache */
   /* replace2051_1_0 -> 5774 in the cache */
   /* replace2051_1_1 -> 5775 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2051_2_0 -> 5774 in the cache */
   /* replace2051_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5792, 5793 },
      -1,
   } },

   /* ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd')) */
   /* search2052_0_0 -> 770 in the cache */
   /* search2052_0_1 -> 1041 in the cache */
   /* search2052_0_2 -> 3703 in the cache */
   /* search2052_0 -> 5748 in the cache */
   /* search2052_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2052_0 -> 770 in the cache */
   /* replace2052_1_0 -> 156 in the cache */
   /* replace2052_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2052_2_0 -> 158 in the cache */
   /* replace2052_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5796, 5797 },
      -1,
   } },

   /* ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c')) */
   /* search2053_0 -> 5769 in the cache */
   /* search2053_1_0 -> 1189 in the cache */
   /* search2053_1_1 -> 5770 in the cache */
   /* search2053_1_2 -> 5771 in the cache */
   /* search2053_1 -> 5772 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2053_0 -> 1189 in the cache */
   /* replace2053_1_0 -> 5774 in the cache */
   /* replace2053_1_1 -> 5775 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2053_2_0 -> 5774 in the cache */
   /* replace2053_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5800, 5801 },
      -1,
   } },

   /* ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd')) */
   /* search2054_0_0 -> 770 in the cache */
   /* search2054_0_1 -> 1041 in the cache */
   /* search2054_0_2 -> 3703 in the cache */
   /* search2054_0 -> 5748 in the cache */
   /* search2054_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2054_0 -> 770 in the cache */
   /* replace2054_1_0 -> 156 in the cache */
   /* replace2054_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2054_2_0 -> 158 in the cache */
   /* replace2054_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5804, 5805 },
      -1,
   } },

   /* ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c')) */
   /* search2055_0 -> 5769 in the cache */
   /* search2055_1_0 -> 1189 in the cache */
   /* search2055_1_1 -> 5770 in the cache */
   /* search2055_1_2 -> 5771 in the cache */
   /* search2055_1 -> 5772 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2055_0 -> 1189 in the cache */
   /* replace2055_1_0 -> 5774 in the cache */
   /* replace2055_1_1 -> 5775 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2055_2_0 -> 5774 in the cache */
   /* replace2055_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5808, 5809 },
      -1,
   } },

   /* ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd')) */
   /* search2056_0_0 -> 770 in the cache */
   /* search2056_0_1 -> 1041 in the cache */
   /* search2056_0_2 -> 3703 in the cache */
   /* search2056_0 -> 5748 in the cache */
   /* search2056_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5748, 1044 },
      -1,
   } },

   /* replace2056_0 -> 770 in the cache */
   /* replace2056_1_0 -> 156 in the cache */
   /* replace2056_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   /* replace2056_2_0 -> 158 in the cache */
   /* replace2056_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5812, 5813 },
      -1,
   } },

   /* ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c')) */
   /* search2057_0 -> 5769 in the cache */
   /* search2057_1_0 -> 1189 in the cache */
   /* search2057_1_1 -> 5770 in the cache */
   /* search2057_1_2 -> 5771 in the cache */
   /* search2057_1 -> 5772 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5769, 5772 },
      -1,
   } },

   /* replace2057_0 -> 1189 in the cache */
   /* replace2057_1_0 -> 5774 in the cache */
   /* replace2057_1_1 -> 5775 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5774, 5775 },
      -1,
   } },
   /* replace2057_2_0 -> 5774 in the cache */
   /* replace2057_2_1 -> 1741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5774, 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1189, 5816, 5817 },
      -1,
   } },

   /* ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd')) */
   /* search2058_0_0 -> 770 in the cache */
   /* search2058_0_1 -> 1041 in the cache */
   /* search2058_0_2 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 1041, 158 },
      0,
   } },
   /* search2058_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 5819, 1044 },
      -1,
   } },

   /* replace2058_0 -> 770 in the cache */
   /* replace2058_1_0 -> 156 in the cache */
   /* replace2058_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2058_2_0 -> 158 in the cache */
   /* replace2058_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5821, 5822 },
      -1,
   } },

   /* ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd')) */
   /* search2059_0_0 -> 770 in the cache */
   /* search2059_0_1 -> 1041 in the cache */
   /* search2059_0_2 -> 158 in the cache */
   /* search2059_0 -> 5819 in the cache */
   /* search2059_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 5819, 1044 },
      -1,
   } },

   /* replace2059_0 -> 770 in the cache */
   /* replace2059_1_0 -> 156 in the cache */
   /* replace2059_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2059_2_0 -> 158 in the cache */
   /* replace2059_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5825, 5826 },
      -1,
   } },

   /* ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd')) */
   /* search2060_0_0 -> 770 in the cache */
   /* search2060_0_1 -> 33 in the cache */
   /* search2060_0_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 33, 141 },
      0,
   } },
   /* search2060_1 -> 321 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 5828, 321 },
      -1,
   } },

   /* replace2060_0 -> 770 in the cache */
   /* replace2060_1_0 -> 36 in the cache */
   /* replace2060_1_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 169 },
      -1,
   } },
   /* replace2060_2_0 -> 141 in the cache */
   /* replace2060_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5830, 5831 },
      -1,
   } },

   /* ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd')) */
   /* search2061_0_0 -> 770 in the cache */
   /* search2061_0_1 -> 1041 in the cache */
   /* search2061_0_2 -> 158 in the cache */
   /* search2061_0 -> 5819 in the cache */
   /* search2061_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5819, 1044 },
      -1,
   } },

   /* replace2061_0 -> 770 in the cache */
   /* replace2061_1_0 -> 156 in the cache */
   /* replace2061_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2061_2_0 -> 158 in the cache */
   /* replace2061_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5834, 5835 },
      -1,
   } },

   /* ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd')) */
   /* search2062_0_0 -> 770 in the cache */
   /* search2062_0_1 -> 1041 in the cache */
   /* search2062_0_2 -> 158 in the cache */
   /* search2062_0 -> 5819 in the cache */
   /* search2062_1 -> 1044 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5819, 1044 },
      -1,
   } },

   /* replace2062_0 -> 770 in the cache */
   /* replace2062_1_0 -> 156 in the cache */
   /* replace2062_1_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 159 },
      -1,
   } },
   /* replace2062_2_0 -> 158 in the cache */
   /* replace2062_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5838, 5839 },
      -1,
   } },

   /* ('fddx', 'a') => 0.0 */
   /* search2063_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2063 -> 117 in the cache */

   /* ('fddx_fine', 'a') => 0.0 */
   /* search2064_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2064 -> 117 in the cache */

   /* ('fddx_coarse', 'a') => 0.0 */
   /* search2065_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2065 -> 117 in the cache */

   /* ('fddy', 'a') => 0.0 */
   /* search2066_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2066 -> 117 in the cache */

   /* ('fddy_fine', 'a') => 0.0 */
   /* search2067_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2067 -> 117 in the cache */

   /* ('fddy_coarse', 'a') => 0.0 */
   /* search2068_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace2068 -> 117 in the cache */

   /* ('umul_low', '#a(is_lower_half_zero)', 'b') => 0 */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      20,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2069_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_low,
      0, 1,
      { 5847, 36 },
      -1,
   } },

   /* replace2069 -> 128 in the cache */

   /* ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c */
   /* search2070_0 -> 5847 in the cache */
   /* search2070_1 -> 36 in the cache */
   /* search2070_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 5847, 36, 141 },
      -1,
   } },

   /* replace2070 -> 141 in the cache */

   /* ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c */
   /* search2071_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2071_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 16, 5850, 141 },
      -1,
   } },

   /* replace2071 -> 141 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2072_0 -> 770 in the cache */
   /* search2072_1_0 -> 762 in the cache */
   /* search2072_1_1 -> 763 in the cache */
   /* search2072_1_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 763, 772 },
      0,
   } },
   /* search2072_2_0 -> 762 in the cache */
   /* search2072_2_1 -> 763 in the cache */
   /* search2072_2_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 763, 774 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5852, 5853 },
      -1,
   } },

   /* replace2072_0 -> 762 in the cache */
   /* replace2072_1 -> 763 in the cache */
   /* replace2072_2_0 -> 770 in the cache */
   /* replace2072_2_1 -> 772 in the cache */
   /* replace2072_2_2 -> 774 in the cache */
   /* replace2072_2 -> 1330 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 763, 1330 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2073_0 -> 770 in the cache */
   /* search2073_1_0 -> 762 in the cache */
   /* search2073_1_1 -> 763 in the cache */
   /* search2073_1_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 763, 772 },
      -1,
   } },
   /* search2073_2_0 -> 762 in the cache */
   /* search2073_2_1 -> 763 in the cache */
   /* search2073_2_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 763, 774 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5856, 5857 },
      -1,
   } },

   /* replace2073_0 -> 762 in the cache */
   /* replace2073_1 -> 763 in the cache */
   /* replace2073_2_0 -> 770 in the cache */
   /* replace2073_2_1 -> 772 in the cache */
   /* replace2073_2_2 -> 774 in the cache */
   /* replace2073_2 -> 1330 in the cache */
   /* replace2073 -> 5855 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2074_0 -> 770 in the cache */
   /* search2074_1_0 -> 762 in the cache */
   /* search2074_1_1 -> 763 in the cache */
   /* search2074_1_2 -> 772 in the cache */
   /* search2074_1 -> 5852 in the cache */
   /* search2074_2_0 -> 762 in the cache */
   /* search2074_2_1 -> 774 in the cache */
   /* search2074_2_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 774, 772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5852, 5859 },
      -1,
   } },

   /* replace2074_0 -> 762 in the cache */
   /* replace2074_1_0 -> 770 in the cache */
   /* replace2074_1_1 -> 763 in the cache */
   /* replace2074_1_2 -> 774 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 763, 774 },
      -1,
   } },
   /* replace2074_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 5861, 772 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2075_0 -> 770 in the cache */
   /* search2075_1_0 -> 762 in the cache */
   /* search2075_1_1 -> 763 in the cache */
   /* search2075_1_2 -> 772 in the cache */
   /* search2075_1 -> 5856 in the cache */
   /* search2075_2_0 -> 762 in the cache */
   /* search2075_2_1 -> 774 in the cache */
   /* search2075_2_2 -> 772 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 762, 774, 772 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5856, 5863 },
      -1,
   } },

   /* replace2075_0 -> 762 in the cache */
   /* replace2075_1_0 -> 770 in the cache */
   /* replace2075_1_1 -> 763 in the cache */
   /* replace2075_1_2 -> 774 in the cache */
   /* replace2075_1 -> 5861 in the cache */
   /* replace2075_2 -> 772 in the cache */
   /* replace2075 -> 5862 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search2076_0 -> 770 in the cache */
   /* search2076_1_0 -> 156 in the cache */
   /* search2076_1_1 -> 158 in the cache */
   /* search2076_1_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 156, 158, 159 },
      0,
   } },
   /* search2076_2_0 -> 1336 in the cache */
   /* search2076_2_1 -> 158 in the cache */
   /* search2076_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1336, 158, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 5865, 5866 },
      -1,
   } },

   /* replace2076_0_0 -> 770 in the cache */
   /* replace2076_0_1 -> 156 in the cache */
   /* replace2076_0_2 -> 1336 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 156, 1336 },
      -1,
   } },
   /* replace2076_1 -> 158 in the cache */
   /* replace2076_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 5868, 158, 159 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search2077_0 -> 770 in the cache */
   /* search2077_1_0 -> 156 in the cache */
   /* search2077_1_1 -> 158 in the cache */
   /* search2077_1_2 -> 159 in the cache */
   /* search2077_1 -> 692 in the cache */
   /* search2077_2_0 -> 1336 in the cache */
   /* search2077_2_1 -> 158 in the cache */
   /* search2077_2_2 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1336, 158, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 692, 5870 },
      -1,
   } },

   /* replace2077_0_0 -> 770 in the cache */
   /* replace2077_0_1 -> 156 in the cache */
   /* replace2077_0_2 -> 1336 in the cache */
   /* replace2077_0 -> 5868 in the cache */
   /* replace2077_1 -> 158 in the cache */
   /* replace2077_2 -> 159 in the cache */
   /* replace2077 -> 5869 in the cache */

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2078_0 -> 770 in the cache */
   /* search2078_1_0 -> 36 in the cache */
   /* search2078_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 141 },
      0,
   } },
   /* search2078_2_0 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5873 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5872, 5874 },
      -1,
   } },

   /* replace2078_0 -> 36 in the cache */
   /* replace2078_1_0 -> 770 in the cache */
   /* replace2078_1_1 -> 141 in the cache */
   /* replace2078_1_2 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 5876 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2079_0 -> 770 in the cache */
   /* search2079_1_0 -> 36 in the cache */
   /* search2079_1_1 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 357 },
      0,
   } },
   /* search2079_2_0 -> 36 in the cache */
   /* search2079_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5878, 5879 },
      -1,
   } },

   /* replace2079_0 -> 36 in the cache */
   /* replace2079_1_0 -> 770 in the cache */
   /* replace2079_1_1 -> 141 in the cache */
   /* replace2079_1_2 -> 169 in the cache */
   /* replace2079_1 -> 5876 in the cache */
   /* replace2079 -> 5877 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2080_0 -> 770 in the cache */
   /* search2080_1_0 -> 36 in the cache */
   /* search2080_1_1 -> 357 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 357 },
      -1,
   } },
   /* search2080_2_0 -> 36 in the cache */
   /* search2080_2_1 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 169 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5881, 5882 },
      -1,
   } },

   /* replace2080_0 -> 36 in the cache */
   /* replace2080_1_0 -> 770 in the cache */
   /* replace2080_1_1 -> 141 in the cache */
   /* replace2080_1_2 -> 169 in the cache */
   /* replace2080_1 -> 5876 in the cache */
   /* replace2080 -> 5877 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2081_0 -> 770 in the cache */
   /* search2081_1_0 -> 36 in the cache */
   /* search2081_1_1 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 141 },
      -1,
   } },
   /* search2081_2_0 -> 36 in the cache */
   /* search2081_2_1 -> 5873 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5873 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5884, 5885 },
      -1,
   } },

   /* replace2081_0 -> 36 in the cache */
   /* replace2081_1_0 -> 770 in the cache */
   /* replace2081_1_1 -> 141 in the cache */
   /* replace2081_1_2 -> 169 in the cache */
   /* replace2081_1 -> 5876 in the cache */
   /* replace2081 -> 5877 in the cache */

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2082_0 -> 770 in the cache */
   /* search2082_1_0 -> 156 in the cache */
   /* search2082_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2082_2_0 -> 156 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5887, 5889 },
      -1,
   } },

   /* replace2082_0 -> 156 in the cache */
   /* replace2082_1_0 -> 770 in the cache */
   /* replace2082_1_1 -> 158 in the cache */
   /* replace2082_1_2 -> 159 in the cache */
   /* replace2082_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2083_0 -> 770 in the cache */
   /* search2083_1_0 -> 156 in the cache */
   /* search2083_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2083_2_0 -> 156 in the cache */
   /* search2083_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5892, 5893 },
      -1,
   } },

   /* replace2083_0 -> 156 in the cache */
   /* replace2083_1_0 -> 770 in the cache */
   /* replace2083_1_1 -> 158 in the cache */
   /* replace2083_1_2 -> 159 in the cache */
   /* replace2083_1 -> 1325 in the cache */
   /* replace2083 -> 5891 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2084_0 -> 770 in the cache */
   /* search2084_1_0 -> 156 in the cache */
   /* search2084_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2084_2_0 -> 156 in the cache */
   /* search2084_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5895, 5896 },
      -1,
   } },

   /* replace2084_0 -> 156 in the cache */
   /* replace2084_1_0 -> 770 in the cache */
   /* replace2084_1_1 -> 158 in the cache */
   /* replace2084_1_2 -> 159 in the cache */
   /* replace2084_1 -> 1325 in the cache */
   /* replace2084 -> 5891 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2085_0 -> 770 in the cache */
   /* search2085_1_0 -> 156 in the cache */
   /* search2085_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2085_2_0 -> 156 in the cache */
   /* search2085_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5898, 5899 },
      -1,
   } },

   /* replace2085_0 -> 156 in the cache */
   /* replace2085_1_0 -> 770 in the cache */
   /* replace2085_1_1 -> 158 in the cache */
   /* replace2085_1_2 -> 159 in the cache */
   /* replace2085_1 -> 1325 in the cache */
   /* replace2085 -> 5891 in the cache */

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2086_0 -> 770 in the cache */
   /* search2086_1_0 -> 156 in the cache */
   /* search2086_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2086_2_0 -> 156 in the cache */
   /* search2086_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5901, 5902 },
      -1,
   } },

   /* replace2086_0 -> 156 in the cache */
   /* replace2086_1_0 -> 770 in the cache */
   /* replace2086_1_1 -> 158 in the cache */
   /* replace2086_1_2 -> 159 in the cache */
   /* replace2086_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2087_0 -> 770 in the cache */
   /* search2087_1_0 -> 156 in the cache */
   /* search2087_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2087_2_0 -> 156 in the cache */
   /* search2087_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5905, 5906 },
      -1,
   } },

   /* replace2087_0 -> 156 in the cache */
   /* replace2087_1_0 -> 770 in the cache */
   /* replace2087_1_1 -> 158 in the cache */
   /* replace2087_1_2 -> 159 in the cache */
   /* replace2087_1 -> 1325 in the cache */
   /* replace2087 -> 5904 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2088_0 -> 770 in the cache */
   /* search2088_1_0 -> 156 in the cache */
   /* search2088_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2088_2_0 -> 156 in the cache */
   /* search2088_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5908, 5909 },
      -1,
   } },

   /* replace2088_0 -> 156 in the cache */
   /* replace2088_1_0 -> 770 in the cache */
   /* replace2088_1_1 -> 158 in the cache */
   /* replace2088_1_2 -> 159 in the cache */
   /* replace2088_1 -> 1325 in the cache */
   /* replace2088 -> 5904 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2089_0 -> 770 in the cache */
   /* search2089_1_0 -> 156 in the cache */
   /* search2089_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2089_2_0 -> 156 in the cache */
   /* search2089_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5911, 5912 },
      -1,
   } },

   /* replace2089_0 -> 156 in the cache */
   /* replace2089_1_0 -> 770 in the cache */
   /* replace2089_1_1 -> 158 in the cache */
   /* replace2089_1_2 -> 159 in the cache */
   /* replace2089_1 -> 1325 in the cache */
   /* replace2089 -> 5904 in the cache */

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2090_0 -> 770 in the cache */
   /* search2090_1_0 -> 156 in the cache */
   /* search2090_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2090_2_0 -> 156 in the cache */
   /* search2090_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5914, 5915 },
      -1,
   } },

   /* replace2090_0 -> 156 in the cache */
   /* replace2090_1_0 -> 770 in the cache */
   /* replace2090_1_1 -> 158 in the cache */
   /* replace2090_1_2 -> 159 in the cache */
   /* replace2090_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2091_0 -> 770 in the cache */
   /* search2091_1_0 -> 156 in the cache */
   /* search2091_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2091_2_0 -> 156 in the cache */
   /* search2091_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5918, 5919 },
      -1,
   } },

   /* replace2091_0 -> 156 in the cache */
   /* replace2091_1_0 -> 770 in the cache */
   /* replace2091_1_1 -> 158 in the cache */
   /* replace2091_1_2 -> 159 in the cache */
   /* replace2091_1 -> 1325 in the cache */
   /* replace2091 -> 5917 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2092_0 -> 770 in the cache */
   /* search2092_1_0 -> 156 in the cache */
   /* search2092_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2092_2_0 -> 156 in the cache */
   /* search2092_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5921, 5922 },
      -1,
   } },

   /* replace2092_0 -> 156 in the cache */
   /* replace2092_1_0 -> 770 in the cache */
   /* replace2092_1_1 -> 158 in the cache */
   /* replace2092_1_2 -> 159 in the cache */
   /* replace2092_1 -> 1325 in the cache */
   /* replace2092 -> 5917 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2093_0 -> 770 in the cache */
   /* search2093_1_0 -> 156 in the cache */
   /* search2093_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2093_2_0 -> 156 in the cache */
   /* search2093_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5924, 5925 },
      -1,
   } },

   /* replace2093_0 -> 156 in the cache */
   /* replace2093_1_0 -> 770 in the cache */
   /* replace2093_1_1 -> 158 in the cache */
   /* replace2093_1_2 -> 159 in the cache */
   /* replace2093_1 -> 1325 in the cache */
   /* replace2093 -> 5917 in the cache */

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2094_0 -> 770 in the cache */
   /* search2094_1_0 -> 156 in the cache */
   /* search2094_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2094_2_0 -> 156 in the cache */
   /* search2094_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5927, 5928 },
      -1,
   } },

   /* replace2094_0 -> 156 in the cache */
   /* replace2094_1_0 -> 770 in the cache */
   /* replace2094_1_1 -> 158 in the cache */
   /* replace2094_1_2 -> 159 in the cache */
   /* replace2094_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2095_0 -> 770 in the cache */
   /* search2095_1_0 -> 156 in the cache */
   /* search2095_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2095_2_0 -> 156 in the cache */
   /* search2095_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5931, 5932 },
      -1,
   } },

   /* replace2095_0 -> 156 in the cache */
   /* replace2095_1_0 -> 770 in the cache */
   /* replace2095_1_1 -> 158 in the cache */
   /* replace2095_1_2 -> 159 in the cache */
   /* replace2095_1 -> 1325 in the cache */
   /* replace2095 -> 5930 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2096_0 -> 770 in the cache */
   /* search2096_1_0 -> 156 in the cache */
   /* search2096_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2096_2_0 -> 156 in the cache */
   /* search2096_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5934, 5935 },
      -1,
   } },

   /* replace2096_0 -> 156 in the cache */
   /* replace2096_1_0 -> 770 in the cache */
   /* replace2096_1_1 -> 158 in the cache */
   /* replace2096_1_2 -> 159 in the cache */
   /* replace2096_1 -> 1325 in the cache */
   /* replace2096 -> 5930 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2097_0 -> 770 in the cache */
   /* search2097_1_0 -> 156 in the cache */
   /* search2097_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2097_2_0 -> 156 in the cache */
   /* search2097_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5937, 5938 },
      -1,
   } },

   /* replace2097_0 -> 156 in the cache */
   /* replace2097_1_0 -> 770 in the cache */
   /* replace2097_1_1 -> 158 in the cache */
   /* replace2097_1_2 -> 159 in the cache */
   /* replace2097_1 -> 1325 in the cache */
   /* replace2097 -> 5930 in the cache */

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2098_0 -> 770 in the cache */
   /* search2098_1_0 -> 156 in the cache */
   /* search2098_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2098_2_0 -> 156 in the cache */
   /* search2098_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5940, 5941 },
      -1,
   } },

   /* replace2098_0 -> 156 in the cache */
   /* replace2098_1_0 -> 770 in the cache */
   /* replace2098_1_1 -> 158 in the cache */
   /* replace2098_1_2 -> 159 in the cache */
   /* replace2098_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2099_0 -> 770 in the cache */
   /* search2099_1_0 -> 156 in the cache */
   /* search2099_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2099_2_0 -> 156 in the cache */
   /* search2099_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5944, 5945 },
      -1,
   } },

   /* replace2099_0 -> 156 in the cache */
   /* replace2099_1_0 -> 770 in the cache */
   /* replace2099_1_1 -> 158 in the cache */
   /* replace2099_1_2 -> 159 in the cache */
   /* replace2099_1 -> 1325 in the cache */
   /* replace2099 -> 5943 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2100_0 -> 770 in the cache */
   /* search2100_1_0 -> 156 in the cache */
   /* search2100_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2100_2_0 -> 156 in the cache */
   /* search2100_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5947, 5948 },
      -1,
   } },

   /* replace2100_0 -> 156 in the cache */
   /* replace2100_1_0 -> 770 in the cache */
   /* replace2100_1_1 -> 158 in the cache */
   /* replace2100_1_2 -> 159 in the cache */
   /* replace2100_1 -> 1325 in the cache */
   /* replace2100 -> 5943 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2101_0 -> 770 in the cache */
   /* search2101_1_0 -> 156 in the cache */
   /* search2101_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2101_2_0 -> 156 in the cache */
   /* search2101_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5950, 5951 },
      -1,
   } },

   /* replace2101_0 -> 156 in the cache */
   /* replace2101_1_0 -> 770 in the cache */
   /* replace2101_1_1 -> 158 in the cache */
   /* replace2101_1_2 -> 159 in the cache */
   /* replace2101_1 -> 1325 in the cache */
   /* replace2101 -> 5943 in the cache */

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2102_0 -> 770 in the cache */
   /* search2102_1_0 -> 156 in the cache */
   /* search2102_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2102_2_0 -> 156 in the cache */
   /* search2102_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5953, 5954 },
      -1,
   } },

   /* replace2102_0 -> 156 in the cache */
   /* replace2102_1_0 -> 770 in the cache */
   /* replace2102_1_1 -> 158 in the cache */
   /* replace2102_1_2 -> 159 in the cache */
   /* replace2102_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2103_0 -> 770 in the cache */
   /* search2103_1_0 -> 156 in the cache */
   /* search2103_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2103_2_0 -> 156 in the cache */
   /* search2103_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5957, 5958 },
      -1,
   } },

   /* replace2103_0 -> 156 in the cache */
   /* replace2103_1_0 -> 770 in the cache */
   /* replace2103_1_1 -> 158 in the cache */
   /* replace2103_1_2 -> 159 in the cache */
   /* replace2103_1 -> 1325 in the cache */
   /* replace2103 -> 5956 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2104_0 -> 770 in the cache */
   /* search2104_1_0 -> 156 in the cache */
   /* search2104_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2104_2_0 -> 156 in the cache */
   /* search2104_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5960, 5961 },
      -1,
   } },

   /* replace2104_0 -> 156 in the cache */
   /* replace2104_1_0 -> 770 in the cache */
   /* replace2104_1_1 -> 158 in the cache */
   /* replace2104_1_2 -> 159 in the cache */
   /* replace2104_1 -> 1325 in the cache */
   /* replace2104 -> 5956 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2105_0 -> 770 in the cache */
   /* search2105_1_0 -> 156 in the cache */
   /* search2105_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2105_2_0 -> 156 in the cache */
   /* search2105_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5963, 5964 },
      -1,
   } },

   /* replace2105_0 -> 156 in the cache */
   /* replace2105_1_0 -> 770 in the cache */
   /* replace2105_1_1 -> 158 in the cache */
   /* replace2105_1_2 -> 159 in the cache */
   /* replace2105_1 -> 1325 in the cache */
   /* replace2105 -> 5956 in the cache */

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2106_0 -> 770 in the cache */
   /* search2106_1_0 -> 156 in the cache */
   /* search2106_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2106_2_0 -> 156 in the cache */
   /* search2106_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5966, 5967 },
      -1,
   } },

   /* replace2106_0 -> 156 in the cache */
   /* replace2106_1_0 -> 770 in the cache */
   /* replace2106_1_1 -> 158 in the cache */
   /* replace2106_1_2 -> 159 in the cache */
   /* replace2106_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2107_0 -> 770 in the cache */
   /* search2107_1_0 -> 156 in the cache */
   /* search2107_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2107_2_0 -> 156 in the cache */
   /* search2107_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5970, 5971 },
      -1,
   } },

   /* replace2107_0 -> 156 in the cache */
   /* replace2107_1_0 -> 770 in the cache */
   /* replace2107_1_1 -> 158 in the cache */
   /* replace2107_1_2 -> 159 in the cache */
   /* replace2107_1 -> 1325 in the cache */
   /* replace2107 -> 5969 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2108_0 -> 770 in the cache */
   /* search2108_1_0 -> 156 in the cache */
   /* search2108_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2108_2_0 -> 156 in the cache */
   /* search2108_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5973, 5974 },
      -1,
   } },

   /* replace2108_0 -> 156 in the cache */
   /* replace2108_1_0 -> 770 in the cache */
   /* replace2108_1_1 -> 158 in the cache */
   /* replace2108_1_2 -> 159 in the cache */
   /* replace2108_1 -> 1325 in the cache */
   /* replace2108 -> 5969 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2109_0 -> 770 in the cache */
   /* search2109_1_0 -> 156 in the cache */
   /* search2109_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2109_2_0 -> 156 in the cache */
   /* search2109_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5976, 5977 },
      -1,
   } },

   /* replace2109_0 -> 156 in the cache */
   /* replace2109_1_0 -> 770 in the cache */
   /* replace2109_1_1 -> 158 in the cache */
   /* replace2109_1_2 -> 159 in the cache */
   /* replace2109_1 -> 1325 in the cache */
   /* replace2109 -> 5969 in the cache */

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2110_0 -> 770 in the cache */
   /* search2110_1_0 -> 156 in the cache */
   /* search2110_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2110_2_0 -> 156 in the cache */
   /* search2110_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5979, 5980 },
      -1,
   } },

   /* replace2110_0 -> 156 in the cache */
   /* replace2110_1_0 -> 770 in the cache */
   /* replace2110_1_1 -> 158 in the cache */
   /* replace2110_1_2 -> 159 in the cache */
   /* replace2110_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2111_0 -> 770 in the cache */
   /* search2111_1_0 -> 156 in the cache */
   /* search2111_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2111_2_0 -> 156 in the cache */
   /* search2111_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5983, 5984 },
      -1,
   } },

   /* replace2111_0 -> 156 in the cache */
   /* replace2111_1_0 -> 770 in the cache */
   /* replace2111_1_1 -> 158 in the cache */
   /* replace2111_1_2 -> 159 in the cache */
   /* replace2111_1 -> 1325 in the cache */
   /* replace2111 -> 5982 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2112_0 -> 770 in the cache */
   /* search2112_1_0 -> 156 in the cache */
   /* search2112_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2112_2_0 -> 156 in the cache */
   /* search2112_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5986, 5987 },
      -1,
   } },

   /* replace2112_0 -> 156 in the cache */
   /* replace2112_1_0 -> 770 in the cache */
   /* replace2112_1_1 -> 158 in the cache */
   /* replace2112_1_2 -> 159 in the cache */
   /* replace2112_1 -> 1325 in the cache */
   /* replace2112 -> 5982 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2113_0 -> 770 in the cache */
   /* search2113_1_0 -> 156 in the cache */
   /* search2113_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2113_2_0 -> 156 in the cache */
   /* search2113_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5989, 5990 },
      -1,
   } },

   /* replace2113_0 -> 156 in the cache */
   /* replace2113_1_0 -> 770 in the cache */
   /* replace2113_1_1 -> 158 in the cache */
   /* replace2113_1_2 -> 159 in the cache */
   /* replace2113_1 -> 1325 in the cache */
   /* replace2113 -> 5982 in the cache */

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2114_0 -> 770 in the cache */
   /* search2114_1_0 -> 156 in the cache */
   /* search2114_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2114_2_0 -> 156 in the cache */
   /* search2114_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5992, 5993 },
      -1,
   } },

   /* replace2114_0 -> 156 in the cache */
   /* replace2114_1_0 -> 770 in the cache */
   /* replace2114_1_1 -> 158 in the cache */
   /* replace2114_1_2 -> 159 in the cache */
   /* replace2114_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2115_0 -> 770 in the cache */
   /* search2115_1_0 -> 156 in the cache */
   /* search2115_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2115_2_0 -> 156 in the cache */
   /* search2115_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5996, 5997 },
      -1,
   } },

   /* replace2115_0 -> 156 in the cache */
   /* replace2115_1_0 -> 770 in the cache */
   /* replace2115_1_1 -> 158 in the cache */
   /* replace2115_1_2 -> 159 in the cache */
   /* replace2115_1 -> 1325 in the cache */
   /* replace2115 -> 5995 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2116_0 -> 770 in the cache */
   /* search2116_1_0 -> 156 in the cache */
   /* search2116_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2116_2_0 -> 156 in the cache */
   /* search2116_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 5999, 6000 },
      -1,
   } },

   /* replace2116_0 -> 156 in the cache */
   /* replace2116_1_0 -> 770 in the cache */
   /* replace2116_1_1 -> 158 in the cache */
   /* replace2116_1_2 -> 159 in the cache */
   /* replace2116_1 -> 1325 in the cache */
   /* replace2116 -> 5995 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2117_0 -> 770 in the cache */
   /* search2117_1_0 -> 156 in the cache */
   /* search2117_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2117_2_0 -> 156 in the cache */
   /* search2117_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6002, 6003 },
      -1,
   } },

   /* replace2117_0 -> 156 in the cache */
   /* replace2117_1_0 -> 770 in the cache */
   /* replace2117_1_1 -> 158 in the cache */
   /* replace2117_1_2 -> 159 in the cache */
   /* replace2117_1 -> 1325 in the cache */
   /* replace2117 -> 5995 in the cache */

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2118_0 -> 770 in the cache */
   /* search2118_1_0 -> 156 in the cache */
   /* search2118_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2118_2_0 -> 156 in the cache */
   /* search2118_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6005, 6006 },
      -1,
   } },

   /* replace2118_0 -> 156 in the cache */
   /* replace2118_1_0 -> 770 in the cache */
   /* replace2118_1_1 -> 158 in the cache */
   /* replace2118_1_2 -> 159 in the cache */
   /* replace2118_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2119_0 -> 770 in the cache */
   /* search2119_1_0 -> 156 in the cache */
   /* search2119_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2119_2_0 -> 156 in the cache */
   /* search2119_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6009, 6010 },
      -1,
   } },

   /* replace2119_0 -> 156 in the cache */
   /* replace2119_1_0 -> 770 in the cache */
   /* replace2119_1_1 -> 158 in the cache */
   /* replace2119_1_2 -> 159 in the cache */
   /* replace2119_1 -> 1325 in the cache */
   /* replace2119 -> 6008 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2120_0 -> 770 in the cache */
   /* search2120_1_0 -> 156 in the cache */
   /* search2120_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2120_2_0 -> 156 in the cache */
   /* search2120_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6012, 6013 },
      -1,
   } },

   /* replace2120_0 -> 156 in the cache */
   /* replace2120_1_0 -> 770 in the cache */
   /* replace2120_1_1 -> 158 in the cache */
   /* replace2120_1_2 -> 159 in the cache */
   /* replace2120_1 -> 1325 in the cache */
   /* replace2120 -> 6008 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2121_0 -> 770 in the cache */
   /* search2121_1_0 -> 156 in the cache */
   /* search2121_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2121_2_0 -> 156 in the cache */
   /* search2121_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6015, 6016 },
      -1,
   } },

   /* replace2121_0 -> 156 in the cache */
   /* replace2121_1_0 -> 770 in the cache */
   /* replace2121_1_1 -> 158 in the cache */
   /* replace2121_1_2 -> 159 in the cache */
   /* replace2121_1 -> 1325 in the cache */
   /* replace2121 -> 6008 in the cache */

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2122_0 -> 770 in the cache */
   /* search2122_1_0 -> 156 in the cache */
   /* search2122_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2122_2_0 -> 156 in the cache */
   /* search2122_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6018, 6019 },
      -1,
   } },

   /* replace2122_0 -> 156 in the cache */
   /* replace2122_1_0 -> 770 in the cache */
   /* replace2122_1_1 -> 158 in the cache */
   /* replace2122_1_2 -> 159 in the cache */
   /* replace2122_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2123_0 -> 770 in the cache */
   /* search2123_1_0 -> 156 in the cache */
   /* search2123_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2123_2_0 -> 156 in the cache */
   /* search2123_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6022, 6023 },
      -1,
   } },

   /* replace2123_0 -> 156 in the cache */
   /* replace2123_1_0 -> 770 in the cache */
   /* replace2123_1_1 -> 158 in the cache */
   /* replace2123_1_2 -> 159 in the cache */
   /* replace2123_1 -> 1325 in the cache */
   /* replace2123 -> 6021 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2124_0 -> 770 in the cache */
   /* search2124_1_0 -> 156 in the cache */
   /* search2124_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2124_2_0 -> 156 in the cache */
   /* search2124_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6025, 6026 },
      -1,
   } },

   /* replace2124_0 -> 156 in the cache */
   /* replace2124_1_0 -> 770 in the cache */
   /* replace2124_1_1 -> 158 in the cache */
   /* replace2124_1_2 -> 159 in the cache */
   /* replace2124_1 -> 1325 in the cache */
   /* replace2124 -> 6021 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2125_0 -> 770 in the cache */
   /* search2125_1_0 -> 156 in the cache */
   /* search2125_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2125_2_0 -> 156 in the cache */
   /* search2125_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6028, 6029 },
      -1,
   } },

   /* replace2125_0 -> 156 in the cache */
   /* replace2125_1_0 -> 770 in the cache */
   /* replace2125_1_1 -> 158 in the cache */
   /* replace2125_1_2 -> 159 in the cache */
   /* replace2125_1 -> 1325 in the cache */
   /* replace2125 -> 6021 in the cache */

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2126_0 -> 770 in the cache */
   /* search2126_1_0 -> 156 in the cache */
   /* search2126_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 156, 158 },
      0,
   } },
   /* search2126_2_0 -> 156 in the cache */
   /* search2126_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 156, 5888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6031, 6032 },
      -1,
   } },

   /* replace2126_0 -> 156 in the cache */
   /* replace2126_1_0 -> 770 in the cache */
   /* replace2126_1_1 -> 158 in the cache */
   /* replace2126_1_2 -> 159 in the cache */
   /* replace2126_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2127_0 -> 770 in the cache */
   /* search2127_1_0 -> 156 in the cache */
   /* search2127_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 156, 3718 },
      0,
   } },
   /* search2127_2_0 -> 156 in the cache */
   /* search2127_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6035, 6036 },
      -1,
   } },

   /* replace2127_0 -> 156 in the cache */
   /* replace2127_1_0 -> 770 in the cache */
   /* replace2127_1_1 -> 158 in the cache */
   /* replace2127_1_2 -> 159 in the cache */
   /* replace2127_1 -> 1325 in the cache */
   /* replace2127 -> 6034 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2128_0 -> 770 in the cache */
   /* search2128_1_0 -> 156 in the cache */
   /* search2128_1_1 -> 3718 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 156, 3718 },
      -1,
   } },
   /* search2128_2_0 -> 156 in the cache */
   /* search2128_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6038, 6039 },
      -1,
   } },

   /* replace2128_0 -> 156 in the cache */
   /* replace2128_1_0 -> 770 in the cache */
   /* replace2128_1_1 -> 158 in the cache */
   /* replace2128_1_2 -> 159 in the cache */
   /* replace2128_1 -> 1325 in the cache */
   /* replace2128 -> 6034 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2129_0 -> 770 in the cache */
   /* search2129_1_0 -> 156 in the cache */
   /* search2129_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 156, 158 },
      -1,
   } },
   /* search2129_2_0 -> 156 in the cache */
   /* search2129_2_1 -> 5888 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 156, 5888 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 770, 6041, 6042 },
      -1,
   } },

   /* replace2129_0 -> 156 in the cache */
   /* replace2129_1_0 -> 770 in the cache */
   /* replace2129_1_1 -> 158 in the cache */
   /* replace2129_1_2 -> 159 in the cache */
   /* replace2129_1 -> 1325 in the cache */
   /* replace2129 -> 6034 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2130_0 -> 770 in the cache */
   /* search2130_1_0 -> 156 in the cache */
   /* search2130_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 156, 158 },
      0,
   } },
   /* search2130_2_0 -> 156 in the cache */
   /* search2130_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 156, 159 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 6044, 6045 },
      -1,
   } },

   /* replace2130_0 -> 156 in the cache */
   /* replace2130_1_0 -> 770 in the cache */
   /* replace2130_1_1 -> 158 in the cache */
   /* replace2130_1_2 -> 159 in the cache */
   /* replace2130_1 -> 1325 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 156, 1325 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2131_0 -> 770 in the cache */
   /* search2131_1_0 -> 156 in the cache */
   /* search2131_1_1 -> 158 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 156, 158 },
      -1,
   } },
   /* search2131_2_0 -> 156 in the cache */
   /* search2131_2_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 156, 159 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 6048, 6049 },
      -1,
   } },

   /* replace2131_0 -> 156 in the cache */
   /* replace2131_1_0 -> 770 in the cache */
   /* replace2131_1_1 -> 158 in the cache */
   /* replace2131_1_2 -> 159 in the cache */
   /* replace2131_1 -> 1325 in the cache */
   /* replace2131 -> 6047 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search2132_0 -> 770 in the cache */
   /* search2132_1_0 -> 133 in the cache */
   /* search2132_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 133, 135 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2132_2_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 6052, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 6051, 6053 },
      -1,
   } },

   /* replace2132_0_0 -> 770 in the cache */
   /* replace2132_0_1 -> 133 in the cache */
   /* replace2132_0_2 -> 6052 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 133, 6052 },
      -1,
   } },
   /* replace2132_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 6055, 135 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search2133_0 -> 770 in the cache */
   /* search2133_1_0 -> 133 in the cache */
   /* search2133_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 133, 135 },
      -1,
   } },
   /* search2133_2_0 -> 6052 in the cache */
   /* search2133_2_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 6052, 135 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 770, 6057, 6058 },
      -1,
   } },

   /* replace2133_0_0 -> 770 in the cache */
   /* replace2133_0_1 -> 133 in the cache */
   /* replace2133_0_2 -> 6052 in the cache */
   /* replace2133_0 -> 6055 in the cache */
   /* replace2133_1 -> 135 in the cache */
   /* replace2133 -> 6056 in the cache */

   /* ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b')) */
   /* search2134_0 -> 531 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 6060 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6061, 6063 },
      -1,
   } },

   /* replace2134_0_0 -> 531 in the cache */
   /* replace2134_0_1 -> 6060 in the cache */
   /* replace2134_0_2 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6060, 6062 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b')) */
   /* search2135_0 -> 531 in the cache */
   /* search2135_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2135_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6067, 6068 },
      -1,
   } },

   /* replace2135_0_0 -> 531 in the cache */
   /* replace2135_0_1 -> 6060 in the cache */
   /* replace2135_0_2 -> 6062 in the cache */
   /* replace2135_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b')) */
   /* search2136_0 -> 531 in the cache */
   /* search2136_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2136_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6071, 6072 },
      -1,
   } },

   /* replace2136_0_0 -> 531 in the cache */
   /* replace2136_0_1 -> 6060 in the cache */
   /* replace2136_0_2 -> 6062 in the cache */
   /* replace2136_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b')) */
   /* search2137_0 -> 531 in the cache */
   /* search2137_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2137_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6075, 6076 },
      -1,
   } },

   /* replace2137_0_0 -> 531 in the cache */
   /* replace2137_0_1 -> 6060 in the cache */
   /* replace2137_0_2 -> 6062 in the cache */
   /* replace2137_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b')) */
   /* search2138_0 -> 531 in the cache */
   /* search2138_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2138_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6079, 6080 },
      -1,
   } },

   /* replace2138_0_0 -> 531 in the cache */
   /* replace2138_0_1 -> 6060 in the cache */
   /* replace2138_0_2 -> 6062 in the cache */
   /* replace2138_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search2139_0 -> 531 in the cache */
   /* search2139_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2139_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6083, 6084 },
      -1,
   } },

   /* replace2139_0_0 -> 531 in the cache */
   /* replace2139_0_1 -> 6060 in the cache */
   /* replace2139_0_2 -> 6062 in the cache */
   /* replace2139_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b')) */
   /* search2140_0 -> 531 in the cache */
   /* search2140_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2140_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6087, 6088 },
      -1,
   } },

   /* replace2140_0_0 -> 531 in the cache */
   /* replace2140_0_1 -> 6060 in the cache */
   /* replace2140_0_2 -> 6062 in the cache */
   /* replace2140_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b')) */
   /* search2141_0 -> 531 in the cache */
   /* search2141_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2141_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6091, 6092 },
      -1,
   } },

   /* replace2141_0_0 -> 531 in the cache */
   /* replace2141_0_1 -> 6060 in the cache */
   /* replace2141_0_2 -> 6062 in the cache */
   /* replace2141_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search2142_0 -> 531 in the cache */
   /* search2142_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2142_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6095, 6096 },
      -1,
   } },

   /* replace2142_0_0 -> 531 in the cache */
   /* replace2142_0_1 -> 6060 in the cache */
   /* replace2142_0_2 -> 6062 in the cache */
   /* replace2142_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search2143_0 -> 531 in the cache */
   /* search2143_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2143_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6099, 6100 },
      -1,
   } },

   /* replace2143_0_0 -> 531 in the cache */
   /* replace2143_0_1 -> 6060 in the cache */
   /* replace2143_0_2 -> 6062 in the cache */
   /* replace2143_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin_mdg(is_used_once)', 'a'), ('fsin_mdg(is_used_once)', 'b')) => ('fsin_mdg', ('bcsel', 'c', 'a', 'b')) */
   /* search2144_0 -> 531 in the cache */
   /* search2144_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_mdg,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2144_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_mdg,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6103, 6104 },
      -1,
   } },

   /* replace2144_0_0 -> 531 in the cache */
   /* replace2144_0_1 -> 6060 in the cache */
   /* replace2144_0_2 -> 6062 in the cache */
   /* replace2144_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_mdg,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos_mdg(is_used_once)', 'a'), ('fcos_mdg(is_used_once)', 'b')) => ('fcos_mdg', ('bcsel', 'c', 'a', 'b')) */
   /* search2145_0 -> 531 in the cache */
   /* search2145_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_mdg,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2145_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_mdg,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6107, 6108 },
      -1,
   } },

   /* replace2145_0_0 -> 531 in the cache */
   /* replace2145_0_1 -> 6060 in the cache */
   /* replace2145_0_2 -> 6062 in the cache */
   /* replace2145_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_mdg,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin_agx(is_used_once)', 'a'), ('fsin_agx(is_used_once)', 'b')) => ('fsin_agx', ('bcsel', 'c', 'a', 'b')) */
   /* search2146_0 -> 531 in the cache */
   /* search2146_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_agx,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2146_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_agx,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6111, 6112 },
      -1,
   } },

   /* replace2146_0_0 -> 531 in the cache */
   /* replace2146_0_1 -> 6060 in the cache */
   /* replace2146_0_2 -> 6062 in the cache */
   /* replace2146_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_agx,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b')) */
   /* search2147_0 -> 531 in the cache */
   /* search2147_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2147_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6115, 6116 },
      -1,
   } },

   /* replace2147_0_0 -> 531 in the cache */
   /* replace2147_0_1 -> 6060 in the cache */
   /* replace2147_0_2 -> 6062 in the cache */
   /* replace2147_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b')) */
   /* search2148_0 -> 531 in the cache */
   /* search2148_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 6060 },
      0,
   } },
   /* search2148_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 6062 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6119, 6120 },
      -1,
   } },

   /* replace2148_0_0 -> 531 in the cache */
   /* replace2148_0_1 -> 6060 in the cache */
   /* replace2148_0_2 -> 6062 in the cache */
   /* replace2148_0 -> 6065 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 6065 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search2149_0 -> 531 in the cache */
   /* search2149_1_0 -> 6060 in the cache */
   /* search2149_1 -> 6083 in the cache */
   /* search2149_2_0 -> 6062 in the cache */
   /* search2149_2 -> 6084 in the cache */
   /* search2149 -> 6085 in the cache */

   /* replace2149_0_0 -> 531 in the cache */
   /* replace2149_0_1 -> 6060 in the cache */
   /* replace2149_0_2 -> 6062 in the cache */
   /* replace2149_0 -> 6065 in the cache */
   /* replace2149 -> 6086 in the cache */

   /* ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b')) */
   /* search2150_0_0 -> 531 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6123, 6124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 6125 },
      -1,
   } },

   /* replace2150_0 -> 531 in the cache */
   /* replace2150_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 6060 },
      -1,
   } },
   /* replace2150_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 6062 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6127, 6128 },
      -1,
   } },

   /* ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b')) */
   /* search2151_0_0 -> 531 in the cache */
   /* search2151_0_1 -> 6123 in the cache */
   /* search2151_0_2 -> 6124 in the cache */
   /* search2151_0 -> 6125 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 6125 },
      -1,
   } },

   /* replace2151_0 -> 531 in the cache */
   /* replace2151_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 6060 },
      -1,
   } },
   /* replace2151_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 6062 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6131, 6132 },
      -1,
   } },

   /* ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b')) */
   /* search2152_0_0 -> 531 in the cache */
   /* search2152_0_1 -> 6123 in the cache */
   /* search2152_0_2 -> 6124 in the cache */
   /* search2152_0 -> 6125 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 6125 },
      -1,
   } },

   /* replace2152_0 -> 531 in the cache */
   /* replace2152_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 6060 },
      -1,
   } },
   /* replace2152_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 6062 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6135, 6136 },
      -1,
   } },

   /* ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b')) */
   /* search2153_0_0 -> 531 in the cache */
   /* search2153_0_1 -> 6123 in the cache */
   /* search2153_0_2 -> 6124 in the cache */
   /* search2153_0 -> 6125 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 6125 },
      -1,
   } },

   /* replace2153_0 -> 531 in the cache */
   /* replace2153_1_0 -> 6060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 6060 },
      -1,
   } },
   /* replace2153_2_0 -> 6062 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 6062 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 531, 6139, 6140 },
      -1,
   } },

   /* ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048)) */
   /* search2154_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 11 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xfff /* 4095 */ },
   } },
   /* replace2154_1_0_0 -> 11 in the cache */
   /* replace2154_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 406 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x800 /* 2048 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6144, 6145 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 6143, 6146 },
      -1,
   } },

   /* ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216)) */
   /* search2155_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 16 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1ffffff /* 33554431 */ },
   } },
   /* replace2155_1_0_0 -> 16 in the cache */
   /* replace2155_1_0_1 -> 406 in the cache */
   /* replace2155_1_0 -> 1004 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1000000 /* 16777216 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1004, 6150 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 6149, 6151 },
      -1,
   } },

   /* ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992)) */
   /* search2156_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 21 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3fffffffffffff /* 18014398509481983 */ },
   } },
   /* replace2156_1_0_0 -> 21 in the cache */
   /* replace2156_1_0_1 -> 406 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 406 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x20000000000000 /* 9007199254740992 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6155, 6156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 6154, 6157 },
      -1,
   } },

   /* ('bfi', 4294967295, 'a', '#b') => ('iadd', ('ishl', 'a', 0), ('iand', 'b', -4294967296)) */
   /* search2157_0 -> 940 in the cache */
   /* search2157_1 -> 16 in the cache */
   /* search2157_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 940, 16, 33 },
      -1,
   } },

   /* replace2157_0_0 -> 16 in the cache */
   /* replace2157_0_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 128 },
      -1,
   } },
   /* replace2157_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x100000000 /* -4294967296 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6161 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6160, 6162 },
      -1,
   } },

   /* ('bfi', 8589934590, 'a', '#b') => ('iadd', ('ishl', 'a', 1), ('iand', 'b', -8589934591)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe /* 8589934590 */ },
   } },
   /* search2158_1 -> 16 in the cache */
   /* search2158_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6164, 16, 33 },
      -1,
   } },

   /* replace2158_0_0 -> 16 in the cache */
   /* replace2158_0_1 -> 406 in the cache */
   /* replace2158_0 -> 1004 in the cache */
   /* replace2158_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1ffffffff /* -8589934591 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6166 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1004, 6167 },
      -1,
   } },

   /* ('bfi', 17179869180, 'a', '#b') => ('iadd', ('ishl', 'a', 2), ('iand', 'b', -17179869181)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc /* 17179869180 */ },
   } },
   /* search2159_1 -> 16 in the cache */
   /* search2159_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6169, 16, 33 },
      -1,
   } },

   /* replace2159_0_0 -> 16 in the cache */
   /* replace2159_0_1 -> 411 in the cache */
   /* replace2159_0 -> 1011 in the cache */
   /* replace2159_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffd /* -17179869181 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6171 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1011, 6172 },
      -1,
   } },

   /* ('bfi', 34359738360, 'a', '#b') => ('iadd', ('ishl', 'a', 3), ('iand', 'b', -34359738361)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff8 /* 34359738360 */ },
   } },
   /* search2160_1 -> 16 in the cache */
   /* search2160_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6174, 16, 33 },
      -1,
   } },

   /* replace2160_0_0 -> 16 in the cache */
   /* replace2160_0_1 -> 415 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 415 },
      -1,
   } },
   /* replace2160_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff9 /* -34359738361 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6177 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6176, 6178 },
      -1,
   } },

   /* ('bfi', 68719476720, 'a', '#b') => ('iadd', ('ishl', 'a', 4), ('iand', 'b', -68719476721)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff0 /* 68719476720 */ },
   } },
   /* search2161_1 -> 16 in the cache */
   /* search2161_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6180, 16, 33 },
      -1,
   } },

   /* replace2161_0_0 -> 16 in the cache */
   /* replace2161_0_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2366 },
      -1,
   } },
   /* replace2161_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff1 /* -68719476721 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6182, 6184 },
      -1,
   } },

   /* ('bfi', 137438953440, 'a', '#b') => ('iadd', ('ishl', 'a', 5), ('iand', 'b', -137438953441)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe0 /* 137438953440 */ },
   } },
   /* search2162_1 -> 16 in the cache */
   /* search2162_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6186, 16, 33 },
      -1,
   } },

   /* replace2162_0_0 -> 16 in the cache */
   /* replace2162_0_1 -> 2397 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2397 },
      -1,
   } },
   /* replace2162_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe1 /* -137438953441 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6188, 6190 },
      -1,
   } },

   /* ('bfi', 274877906880, 'a', '#b') => ('iadd', ('ishl', 'a', 6), ('iand', 'b', -274877906881)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc0 /* 274877906880 */ },
   } },
   /* search2163_1 -> 16 in the cache */
   /* search2163_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6192, 16, 33 },
      -1,
   } },

   /* replace2163_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x6 /* 6 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6194 },
      -1,
   } },
   /* replace2163_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc1 /* -274877906881 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6196 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6195, 6197 },
      -1,
   } },

   /* ('bfi', 549755813760, 'a', '#b') => ('iadd', ('ishl', 'a', 7), ('iand', 'b', -549755813761)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff80 /* 549755813760 */ },
   } },
   /* search2164_1 -> 16 in the cache */
   /* search2164_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6199, 16, 33 },
      -1,
   } },

   /* replace2164_0_0 -> 16 in the cache */
   /* replace2164_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 9 },
      -1,
   } },
   /* replace2164_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff81 /* -549755813761 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6202 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6201, 6203 },
      -1,
   } },

   /* ('bfi', 1099511627520, 'a', '#b') => ('iadd', ('ishl', 'a', 8), ('iand', 'b', -1099511627521)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff00 /* 1099511627520 */ },
   } },
   /* search2165_1 -> 16 in the cache */
   /* search2165_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6205, 16, 33 },
      -1,
   } },

   /* replace2165_0_0 -> 16 in the cache */
   /* replace2165_0_1 -> 844 in the cache */
   /* replace2165_0 -> 3621 in the cache */
   /* replace2165_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff01 /* -1099511627521 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6207 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3621, 6208 },
      -1,
   } },

   /* ('bfi', 2199023255040, 'a', '#b') => ('iadd', ('ishl', 'a', 9), ('iand', 'b', -2199023255041)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe00 /* 2199023255040 */ },
   } },
   /* search2166_1 -> 16 in the cache */
   /* search2166_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6210, 16, 33 },
      -1,
   } },

   /* replace2166_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x9 /* 9 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6212 },
      -1,
   } },
   /* replace2166_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe01 /* -2199023255041 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6214 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6213, 6215 },
      -1,
   } },

   /* ('bfi', 4398046510080, 'a', '#b') => ('iadd', ('ishl', 'a', 10), ('iand', 'b', -4398046510081)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc00 /* 4398046510080 */ },
   } },
   /* search2167_1 -> 16 in the cache */
   /* search2167_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6217, 16, 33 },
      -1,
   } },

   /* replace2167_0_0 -> 16 in the cache */
   /* replace2167_0_1 -> 5449 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 5449 },
      -1,
   } },
   /* replace2167_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc01 /* -4398046510081 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6220 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6219, 6221 },
      -1,
   } },

   /* ('bfi', 8796093020160, 'a', '#b') => ('iadd', ('ishl', 'a', 11), ('iand', 'b', -8796093020161)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff800 /* 8796093020160 */ },
   } },
   /* search2168_1 -> 16 in the cache */
   /* search2168_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6223, 16, 33 },
      -1,
   } },

   /* replace2168_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xb /* 11 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6225 },
      -1,
   } },
   /* replace2168_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff801 /* -8796093020161 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6227 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6226, 6228 },
      -1,
   } },

   /* ('bfi', 17592186040320, 'a', '#b') => ('iadd', ('ishl', 'a', 12), ('iand', 'b', -17592186040321)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff000 /* 17592186040320 */ },
   } },
   /* search2169_1 -> 16 in the cache */
   /* search2169_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6230, 16, 33 },
      -1,
   } },

   /* replace2169_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xc /* 12 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6232 },
      -1,
   } },
   /* replace2169_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff001 /* -17592186040321 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6234 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6233, 6235 },
      -1,
   } },

   /* ('bfi', 35184372080640, 'a', '#b') => ('iadd', ('ishl', 'a', 13), ('iand', 'b', -35184372080641)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe000 /* 35184372080640 */ },
   } },
   /* search2170_1 -> 16 in the cache */
   /* search2170_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6237, 16, 33 },
      -1,
   } },

   /* replace2170_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xd /* 13 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6239 },
      -1,
   } },
   /* replace2170_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe001 /* -35184372080641 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6240, 6242 },
      -1,
   } },

   /* ('bfi', 70368744161280, 'a', '#b') => ('iadd', ('ishl', 'a', 14), ('iand', 'b', -70368744161281)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc000 /* 70368744161280 */ },
   } },
   /* search2171_1 -> 16 in the cache */
   /* search2171_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6244, 16, 33 },
      -1,
   } },

   /* replace2171_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xe /* 14 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6246 },
      -1,
   } },
   /* replace2171_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc001 /* -70368744161281 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6248 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6247, 6249 },
      -1,
   } },

   /* ('bfi', 140737488322560, 'a', '#b') => ('iadd', ('ishl', 'a', 15), ('iand', 'b', -140737488322561)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff8000 /* 140737488322560 */ },
   } },
   /* search2172_1 -> 16 in the cache */
   /* search2172_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6251, 16, 33 },
      -1,
   } },

   /* replace2172_0_0 -> 16 in the cache */
   /* replace2172_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 14 },
      -1,
   } },
   /* replace2172_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff8001 /* -140737488322561 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6254 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6253, 6255 },
      -1,
   } },

   /* ('bfi', 281474976645120, 'a', '#b') => ('iadd', ('ishl', 'a', 16), ('iand', 'b', -281474976645121)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff0000 /* 281474976645120 */ },
   } },
   /* search2173_1 -> 16 in the cache */
   /* search2173_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6257, 16, 33 },
      -1,
   } },

   /* replace2173_0_0 -> 16 in the cache */
   /* replace2173_0_1 -> 866 in the cache */
   /* replace2173_0 -> 1018 in the cache */
   /* replace2173_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff0001 /* -281474976645121 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6259 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1018, 6260 },
      -1,
   } },

   /* ('bfi', 562949953290240, 'a', '#b') => ('iadd', ('ishl', 'a', 17), ('iand', 'b', -562949953290241)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe0000 /* 562949953290240 */ },
   } },
   /* search2174_1 -> 16 in the cache */
   /* search2174_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6262, 16, 33 },
      -1,
   } },

   /* replace2174_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x11 /* 17 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6264 },
      -1,
   } },
   /* replace2174_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe0001 /* -562949953290241 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6266 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6265, 6267 },
      -1,
   } },

   /* ('bfi', 1125899906580480, 'a', '#b') => ('iadd', ('ishl', 'a', 18), ('iand', 'b', -1125899906580481)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc0000 /* 1125899906580480 */ },
   } },
   /* search2175_1 -> 16 in the cache */
   /* search2175_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6269, 16, 33 },
      -1,
   } },

   /* replace2175_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x12 /* 18 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6271 },
      -1,
   } },
   /* replace2175_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc0001 /* -1125899906580481 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6273 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6272, 6274 },
      -1,
   } },

   /* ('bfi', 2251799813160960, 'a', '#b') => ('iadd', ('ishl', 'a', 19), ('iand', 'b', -2251799813160961)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff80000 /* 2251799813160960 */ },
   } },
   /* search2176_1 -> 16 in the cache */
   /* search2176_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6276, 16, 33 },
      -1,
   } },

   /* replace2176_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x13 /* 19 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6278 },
      -1,
   } },
   /* replace2176_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff80001 /* -2251799813160961 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6279, 6281 },
      -1,
   } },

   /* ('bfi', 4503599626321920, 'a', '#b') => ('iadd', ('ishl', 'a', 20), ('iand', 'b', -4503599626321921)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff00000 /* 4503599626321920 */ },
   } },
   /* search2177_1 -> 16 in the cache */
   /* search2177_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6283, 16, 33 },
      -1,
   } },

   /* replace2177_0_0 -> 16 in the cache */
   /* replace2177_0_1 -> 5492 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 5492 },
      -1,
   } },
   /* replace2177_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff00001 /* -4503599626321921 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6286 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6285, 6287 },
      -1,
   } },

   /* ('bfi', 9007199252643840, 'a', '#b') => ('iadd', ('ishl', 'a', 21), ('iand', 'b', -9007199252643841)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe00000 /* 9007199252643840 */ },
   } },
   /* search2178_1 -> 16 in the cache */
   /* search2178_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6289, 16, 33 },
      -1,
   } },

   /* replace2178_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x15 /* 21 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6291 },
      -1,
   } },
   /* replace2178_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe00001 /* -9007199252643841 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6293 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6292, 6294 },
      -1,
   } },

   /* ('bfi', 18014398505287680, 'a', '#b') => ('iadd', ('ishl', 'a', 22), ('iand', 'b', -18014398505287681)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc00000 /* 18014398505287680 */ },
   } },
   /* search2179_1 -> 16 in the cache */
   /* search2179_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6296, 16, 33 },
      -1,
   } },

   /* replace2179_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x16 /* 22 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6298 },
      -1,
   } },
   /* replace2179_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc00001 /* -18014398505287681 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6300 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6299, 6301 },
      -1,
   } },

   /* ('bfi', 36028797010575360, 'a', '#b') => ('iadd', ('ishl', 'a', 23), ('iand', 'b', -36028797010575361)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff800000 /* 36028797010575360 */ },
   } },
   /* search2180_1 -> 16 in the cache */
   /* search2180_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6303, 16, 33 },
      -1,
   } },

   /* replace2180_0_0 -> 16 in the cache */
   /* replace2180_0_1 -> 5471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 5471 },
      -1,
   } },
   /* replace2180_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff800001 /* -36028797010575361 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6306 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6305, 6307 },
      -1,
   } },

   /* ('bfi', 72057594021150720, 'a', '#b') => ('iadd', ('ishl', 'a', 24), ('iand', 'b', -72057594021150721)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff000000 /* 72057594021150720 */ },
   } },
   /* search2181_1 -> 16 in the cache */
   /* search2181_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6309, 16, 33 },
      -1,
   } },

   /* replace2181_0_0 -> 16 in the cache */
   /* replace2181_0_1 -> 1025 in the cache */
   /* replace2181_0 -> 1027 in the cache */
   /* replace2181_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff000001 /* -72057594021150721 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6311 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1027, 6312 },
      -1,
   } },

   /* ('bfi', 144115188042301440, 'a', '#b') => ('iadd', ('ishl', 'a', 25), ('iand', 'b', -144115188042301441)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe000000 /* 144115188042301440 */ },
   } },
   /* search2182_1 -> 16 in the cache */
   /* search2182_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6314, 16, 33 },
      -1,
   } },

   /* replace2182_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x19 /* 25 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6316 },
      -1,
   } },
   /* replace2182_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe000001 /* -144115188042301441 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6318 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6317, 6319 },
      -1,
   } },

   /* ('bfi', 288230376084602880, 'a', '#b') => ('iadd', ('ishl', 'a', 26), ('iand', 'b', -288230376084602881)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc000000 /* 288230376084602880 */ },
   } },
   /* search2183_1 -> 16 in the cache */
   /* search2183_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6321, 16, 33 },
      -1,
   } },

   /* replace2183_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1a /* 26 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6323 },
      -1,
   } },
   /* replace2183_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc000001 /* -288230376084602881 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6325 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6324, 6326 },
      -1,
   } },

   /* ('bfi', 576460752169205760, 'a', '#b') => ('iadd', ('ishl', 'a', 27), ('iand', 'b', -576460752169205761)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff8000000 /* 576460752169205760 */ },
   } },
   /* search2184_1 -> 16 in the cache */
   /* search2184_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6328, 16, 33 },
      -1,
   } },

   /* replace2184_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1b /* 27 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6330 },
      -1,
   } },
   /* replace2184_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fffffff8000001 /* -576460752169205761 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6332 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6331, 6333 },
      -1,
   } },

   /* ('bfi', 1152921504338411520, 'a', '#b') => ('iadd', ('ishl', 'a', 28), ('iand', 'b', -1152921504338411521)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff0000000 /* 1152921504338411520 */ },
   } },
   /* search2185_1 -> 16 in the cache */
   /* search2185_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6335, 16, 33 },
      -1,
   } },

   /* replace2185_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1c /* 28 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6337 },
      -1,
   } },
   /* replace2185_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xffffffff0000001 /* -1152921504338411521 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6339 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6338, 6340 },
      -1,
   } },

   /* ('bfi', 2305843008676823040, 'a', '#b') => ('iadd', ('ishl', 'a', 29), ('iand', 'b', -2305843008676823041)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1fffffffe0000000 /* 2305843008676823040 */ },
   } },
   /* search2186_1 -> 16 in the cache */
   /* search2186_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6342, 16, 33 },
      -1,
   } },

   /* replace2186_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1d /* 29 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 6344 },
      -1,
   } },
   /* replace2186_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1fffffffe0000001 /* -2305843008676823041 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6346 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6345, 6347 },
      -1,
   } },

   /* ('bfi', 4611686017353646080, 'a', '#b') => ('iadd', ('ishl', 'a', 30), ('iand', 'b', -4611686017353646081)) */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffffc0000000 /* 4611686017353646080 */ },
   } },
   /* search2187_1 -> 16 in the cache */
   /* search2187_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 6349, 16, 33 },
      -1,
   } },

   /* replace2187_0_0 -> 16 in the cache */
   /* replace2187_0_1 -> 5445 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 5445 },
      -1,
   } },
   /* replace2187_1_0 -> 36 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x3fffffffc0000001 /* -4611686017353646081 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 6352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 6351, 6353 },
      -1,
   } },

   /* ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b')) */
   /* search2188_0_0 -> 16 in the cache */
   /* search2188_0_1 -> 36 in the cache */
   /* search2188_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* replace2188_0_0 -> 16 in the cache */
   /* replace2188_0 -> 4217 in the cache */
   /* replace2188_1_0 -> 36 in the cache */
   /* replace2188_1 -> 4214 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 4217, 4214 },
      -1,
   } },

   /* ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b')) */
   /* search2189_0_0 -> 16 in the cache */
   /* search2189_0_1 -> 36 in the cache */
   /* search2189_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* replace2189_0_0 -> 16 in the cache */
   /* replace2189_0 -> 2161 in the cache */
   /* replace2189_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2161, 6358 },
      -1,
   } },

   /* ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b')) */
   /* search2190_0_0 -> 16 in the cache */
   /* search2190_0_1 -> 36 in the cache */
   /* search2190_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* replace2190_0_0 -> 16 in the cache */
   /* replace2190_0 -> 3505 in the cache */
   /* replace2190_1_0 -> 36 in the cache */
   /* replace2190_1 -> 5072 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3505, 5072 },
      -1,
   } },

   /* ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b')) */
   /* search2191_0_0 -> 16 in the cache */
   /* search2191_0_1 -> 36 in the cache */
   /* search2191_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* replace2191_0_0 -> 16 in the cache */
   /* replace2191_0 -> 4558 in the cache */
   /* replace2191_1_0 -> 36 in the cache */
   /* replace2191_1 -> 4743 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 4558, 4743 },
      -1,
   } },

   /* ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b')) */
   /* search2192_0_0 -> 16 in the cache */
   /* search2192_0_1 -> 36 in the cache */
   /* search2192_0 -> 3400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 3400 },
      -1,
   } },

   /* replace2192_0_0 -> 16 in the cache */
   /* replace2192_0 -> 2172 in the cache */
   /* replace2192_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2172, 6365 },
      -1,
   } },

   /* ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c')) */
   /* search2193_0_0 -> 16 in the cache */
   /* search2193_0_1 -> 36 in the cache */
   /* search2193_0_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 16, 36, 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 6367 },
      -1,
   } },

   /* replace2193_0_0 -> 16 in the cache */
   /* replace2193_0 -> 4217 in the cache */
   /* replace2193_1_0 -> 36 in the cache */
   /* replace2193_1 -> 4214 in the cache */
   /* replace2193_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 4217, 4214, 6369 },
      -1,
   } },

   /* ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c')) */
   /* search2194_0_0 -> 16 in the cache */
   /* search2194_0_1 -> 36 in the cache */
   /* search2194_0_2 -> 141 in the cache */
   /* search2194_0 -> 6367 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 6367 },
      -1,
   } },

   /* replace2194_0_0 -> 16 in the cache */
   /* replace2194_0 -> 2161 in the cache */
   /* replace2194_1_0 -> 36 in the cache */
   /* replace2194_1 -> 6358 in the cache */
   /* replace2194_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 2161, 6358, 6372 },
      -1,
   } },

   /* ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c')) */
   /* search2195_0_0 -> 16 in the cache */
   /* search2195_0_1 -> 36 in the cache */
   /* search2195_0_2 -> 141 in the cache */
   /* search2195_0 -> 6367 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 6367 },
      -1,
   } },

   /* replace2195_0_0 -> 16 in the cache */
   /* replace2195_0 -> 3505 in the cache */
   /* replace2195_1_0 -> 36 in the cache */
   /* replace2195_1 -> 5072 in the cache */
   /* replace2195_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3505, 5072, 6375 },
      -1,
   } },

   /* ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c')) */
   /* search2196_0_0 -> 16 in the cache */
   /* search2196_0_1 -> 36 in the cache */
   /* search2196_0_2 -> 141 in the cache */
   /* search2196_0 -> 6367 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 6367 },
      -1,
   } },

   /* replace2196_0_0 -> 16 in the cache */
   /* replace2196_0 -> 4558 in the cache */
   /* replace2196_1_0 -> 36 in the cache */
   /* replace2196_1 -> 4743 in the cache */
   /* replace2196_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 4558, 4743, 6378 },
      -1,
   } },

   /* ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c')) */
   /* search2197_0_0 -> 16 in the cache */
   /* search2197_0_1 -> 36 in the cache */
   /* search2197_0_2 -> 141 in the cache */
   /* search2197_0 -> 6367 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 6367 },
      -1,
   } },

   /* replace2197_0_0 -> 16 in the cache */
   /* replace2197_0 -> 2172 in the cache */
   /* replace2197_1_0 -> 36 in the cache */
   /* replace2197_1 -> 6365 in the cache */
   /* replace2197_2_0 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 141 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 2172, 6365, 6381 },
      -1,
   } },

   /* ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd')) */
   /* search2198_0_0 -> 16 in the cache */
   /* search2198_0_1 -> 36 in the cache */
   /* search2198_0_2 -> 141 in the cache */
   /* search2198_0_3 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 16, 36, 141, 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 6383 },
      -1,
   } },

   /* replace2198_0_0 -> 16 in the cache */
   /* replace2198_0 -> 4217 in the cache */
   /* replace2198_1_0 -> 36 in the cache */
   /* replace2198_1 -> 4214 in the cache */
   /* replace2198_2_0 -> 141 in the cache */
   /* replace2198_2 -> 6369 in the cache */
   /* replace2198_3_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 4217, 4214, 6369, 6385 },
      -1,
   } },

   /* ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd')) */
   /* search2199_0_0 -> 16 in the cache */
   /* search2199_0_1 -> 36 in the cache */
   /* search2199_0_2 -> 141 in the cache */
   /* search2199_0_3 -> 169 in the cache */
   /* search2199_0 -> 6383 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 6383 },
      -1,
   } },

   /* replace2199_0_0 -> 16 in the cache */
   /* replace2199_0 -> 2161 in the cache */
   /* replace2199_1_0 -> 36 in the cache */
   /* replace2199_1 -> 6358 in the cache */
   /* replace2199_2_0 -> 141 in the cache */
   /* replace2199_2 -> 6372 in the cache */
   /* replace2199_3_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 2161, 6358, 6372, 6388 },
      -1,
   } },

   /* ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd')) */
   /* search2200_0_0 -> 16 in the cache */
   /* search2200_0_1 -> 36 in the cache */
   /* search2200_0_2 -> 141 in the cache */
   /* search2200_0_3 -> 169 in the cache */
   /* search2200_0 -> 6383 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 6383 },
      -1,
   } },

   /* replace2200_0_0 -> 16 in the cache */
   /* replace2200_0 -> 3505 in the cache */
   /* replace2200_1_0 -> 36 in the cache */
   /* replace2200_1 -> 5072 in the cache */
   /* replace2200_2_0 -> 141 in the cache */
   /* replace2200_2 -> 6375 in the cache */
   /* replace2200_3_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3505, 5072, 6375, 6391 },
      -1,
   } },

   /* ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd')) */
   /* search2201_0_0 -> 16 in the cache */
   /* search2201_0_1 -> 36 in the cache */
   /* search2201_0_2 -> 141 in the cache */
   /* search2201_0_3 -> 169 in the cache */
   /* search2201_0 -> 6383 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 6383 },
      -1,
   } },

   /* replace2201_0_0 -> 16 in the cache */
   /* replace2201_0 -> 4558 in the cache */
   /* replace2201_1_0 -> 36 in the cache */
   /* replace2201_1 -> 4743 in the cache */
   /* replace2201_2_0 -> 141 in the cache */
   /* replace2201_2 -> 6378 in the cache */
   /* replace2201_3_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 4558, 4743, 6378, 6394 },
      -1,
   } },

   /* ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd')) */
   /* search2202_0_0 -> 16 in the cache */
   /* search2202_0_1 -> 36 in the cache */
   /* search2202_0_2 -> 141 in the cache */
   /* search2202_0_3 -> 169 in the cache */
   /* search2202_0 -> 6383 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 6383 },
      -1,
   } },

   /* replace2202_0_0 -> 16 in the cache */
   /* replace2202_0 -> 2172 in the cache */
   /* replace2202_1_0 -> 36 in the cache */
   /* replace2202_1 -> 6365 in the cache */
   /* replace2202_2_0 -> 141 in the cache */
   /* replace2202_2 -> 6381 in the cache */
   /* replace2202_3_0 -> 169 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 169 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 2172, 6365, 6381, 6397 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_expression_cond[] = {
   (is_used_once),
   (is_only_used_as_float),
   (is_not_used_by_if),
   (is_used_by_non_fsat),
   (no_signed_wrap),
   (only_lower_8_bits_used),
   (only_lower_16_bits_used),
};

static const nir_search_variable_cond nir_opt_algebraic_variable_cond[] = {
   (is_pos_power_of_two),
   (is_neg_power_of_two),
   (is_bitcount2),
   (is_finite_not_zero),
   (is_finite),
   (is_not_const),
   (is_not_const_zero),
   (is_ult_0xfffc07fc),
   (is_not_fmul),
   (is_first_5_bits_uge_2),
   (is_unsigned_multiple_of_4),
   (is_unsigned_multiple_of_2),
   (is_unsigned_multiple_of_8),
   (is_unsigned_multiple_of_16),
   (is_unsigned_multiple_of_32),
   (is_unsigned_multiple_of_64),
   (is_a_number),
   (is_gt_0_and_lt_1),
   (is_zero_to_one),
   (is_not_negative),
   (is_lower_half_zero),
   (is_upper_half_zero),
   (is_lower_half_negative_one),
   (is_upper_half_negative_one),
   (is_odd),
   (is_integral),
   (is_not_positive),
   (is_not_zero),
   (is_a_number_not_negative),
   (is_a_number_not_positive),
   (is_gt_zero),
   (is_lt_zero),
   (is_a_number_gt_zero),
   (is_a_number_lt_zero),
   (is_fmul),
   (is_ult_32),
};

static const struct transform nir_opt_algebraic_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2932, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2932, 1242, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 54, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 57, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 32, 0 },
   { 67, 68, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 61, 0 },
   { 78, 81, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 62, 61, 0 },
   { 64, 61, 0 },
   { 82, 81, 1 },
   { 83, 88, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 61, 0 },
   { 66, 61, 0 },
   { 89, 95, 1 },
   { 96, 97, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 98, 4230, 109 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 717, 719, 23 },
   { 720, 722, 23 },
   { 723, 725, 24 },
   { 3332, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3959, 3972, 82 },
   { 4359, 4367, 119 },
   { 4376, 4384, 119 },
   { 4393, 4400, 119 },
   { 3959, 4415, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 124, 32, 0 },
   { 3959, 3972, 82 },
   { 4359, 4367, 119 },
   { 4376, 4384, 119 },
   { 4393, 4400, 119 },
   { 3959, 4415, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3973, 3981, 82 },
   { 4368, 4375, 119 },
   { 4385, 4392, 119 },
   { 4401, 4408, 119 },
   { 3973, 4420, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 125, 32, 0 },
   { 3973, 3981, 82 },
   { 4368, 4375, 119 },
   { 4385, 4392, 119 },
   { 4401, 4408, 119 },
   { 3973, 4420, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3945, 3949, 78 },
   { 3950, 3952, 79 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 126, 32, 0 },
   { 3945, 3949, 78 },
   { 3950, 3952, 79 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3953, 3954, 80 },
   { 3955, 3958, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 127, 32, 0 },
   { 3953, 3954, 80 },
   { 3955, 3958, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 129, 16, 0 },
   { 131, 130, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 294, 295, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 294, 295, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 278, 279, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 278, 279, 7 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 297, 298, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 278, 279, 7 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 297, 298, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2938, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2938, 1242, 0 },
   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 250, 128, 0 },
   { 251, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 302, 274, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 302, 274, 0 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 358, 360, 0 },
   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 358, 360, 0 },
   { 341, 421, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 36, 0 },
   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 36, 0 },
   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 36, 0 },
   { 361, 363, 0 },
   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 36, 0 },
   { 361, 363, 0 },
   { 344, 437, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 466, 467, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 309, 36, 0 },
   { 466, 467, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 309, 36, 0 },
   { 373, 374, 8 },
   { 466, 467, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 463, 344, 0 },
   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 310, 36, 0 },
   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 310, 36, 0 },
   { 463, 344, 0 },
   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 310, 36, 0 },
   { 375, 376, 9 },
   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 310, 36, 0 },
   { 375, 376, 9 },
   { 463, 344, 0 },
   { 464, 465, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 311, 36, 0 },
   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 311, 36, 0 },
   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 311, 36, 0 },
   { 367, 369, 0 },
   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 311, 36, 0 },
   { 367, 369, 0 },
   { 353, 453, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 312, 36, 0 },
   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 312, 36, 0 },
   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 312, 36, 0 },
   { 370, 372, 0 },
   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 312, 36, 0 },
   { 370, 372, 0 },
   { 356, 461, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 472, 473, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 313, 36, 0 },
   { 472, 473, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 313, 36, 0 },
   { 379, 380, 8 },
   { 472, 473, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 470, 471, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 314, 36, 0 },
   { 470, 471, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 314, 36, 0 },
   { 381, 382, 9 },
   { 470, 471, 14 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 315, 36, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 315, 36, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 316, 36, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 316, 36, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 315, 36, 0 },
   { 316, 36, 0 },
   { 364, 366, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 315, 36, 0 },
   { 316, 36, 0 },
   { 364, 366, 0 },
   { 347, 445, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 468, 469, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 317, 36, 0 },
   { 468, 469, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 318, 36, 0 },
   { 468, 469, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 317, 36, 0 },
   { 318, 36, 0 },
   { 377, 378, 8 },
   { 468, 469, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 698, 703, 21 },
   { 704, 709, 21 },
   { 710, 716, 22 },
   { 3334, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 718, 741, 27 },
   { 721, 742, 27 },
   { 724, 743, 28 },
   { 3336, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 744, 746, 29 },
   { 3333, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 778, 783, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 784, 117, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 785, 117, 0 },
   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 813, 814, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 786, 117, 0 },
   { 809, 812, 0 },
   { 813, 814, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 819, 812, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2937, 32, 0 },
   { 3435, 3436, 60 },
   { 3437, 416, 60 },
   { 3438, 3439, 60 },
   { 3469, 449, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2934, 1242, 0 },
   { 2936, 2935, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2934, 1242, 0 },
   { 2936, 2935, 0 },
   { 2937, 32, 0 },
   { 3435, 3436, 60 },
   { 3437, 416, 60 },
   { 3438, 3439, 60 },
   { 3469, 449, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 3386, 1090, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1356, 32, 0 },
   { 1357, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4135, 4138, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4130, 4134, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4130, 4134, 96 },
   { 5311, 11, 0 },
   { 5364, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4139, 4143, 97 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4139, 4143, 97 },
   { 5416, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4144, 4147, 97 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4144, 4147, 97 },
   { 5389, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2564, 2572, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2566, 2573, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2568, 2574, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2570, 2575, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2594, 2601, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2602, 2604, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2605, 2607, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2608, 2610, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2611, 2613, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2614, 2617, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2618, 2622, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2623, 2625, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2626, 2628, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2629, 2631, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2632, 2637, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2638, 2644, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2645, 2651, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2652, 2668, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2669, 2701, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2702, 2705, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2706, 2711, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2712, 2716, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2717, 2725, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2726, 2742, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2743, 2746, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2747, 2752, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2753, 2757, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2758, 2766, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2767, 2783, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2784, 2787, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2788, 2793, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2794, 2798, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2799, 2807, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2808, 2824, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3663, 3664, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3663, 3664, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2965, 2968, 52 },
   { 2944, 2947, 52 },
   { 2951, 2954, 52 },
   { 2969, 2972, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2973, 2975, 52 },
   { 2957, 2959, 52 },
   { 2962, 2964, 52 },
   { 2976, 2978, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2979, 16, 0 },
   { 2980, 36, 0 },
   { 2985, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2979, 16, 0 },
   { 2980, 36, 0 },
   { 2985, 2237, 0 },
   { 2986, 2199, 0 },
   { 6159, 6163, 126 },
   { 6165, 6168, 126 },
   { 6170, 6173, 126 },
   { 6175, 6179, 126 },
   { 6181, 6185, 126 },
   { 6187, 6191, 126 },
   { 6193, 6198, 126 },
   { 6200, 6204, 126 },
   { 6206, 6209, 126 },
   { 6211, 6216, 126 },
   { 6218, 6222, 126 },
   { 6224, 6229, 126 },
   { 6231, 6236, 126 },
   { 6238, 6243, 126 },
   { 6245, 6250, 126 },
   { 6252, 6256, 126 },
   { 6258, 6261, 126 },
   { 6263, 6268, 126 },
   { 6270, 6275, 126 },
   { 6277, 6282, 126 },
   { 6284, 6288, 126 },
   { 6290, 6295, 126 },
   { 6297, 6302, 126 },
   { 6304, 6308, 126 },
   { 6310, 6313, 126 },
   { 6315, 6320, 126 },
   { 6322, 6327, 126 },
   { 6329, 6334, 126 },
   { 6336, 6341, 126 },
   { 6343, 6348, 126 },
   { 6350, 6354, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { 3025, 32, 0 },
   { 3026, 261, 0 },
   { 3027, 3024, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { 3028, 2999, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { 3025, 32, 0 },
   { 3026, 261, 0 },
   { 3027, 3024, 0 },
   { 3028, 2999, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3017, 3070, 56 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3065, 3067, 55 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3064, 3046, 0 },
   { 3065, 3067, 55 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3072, 3154, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3155, 3224, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3263, 36, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3335, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3387, 3496, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2894, 3504, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3413, 3499, 64 },
   { 3413, 4222, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3421, 3505, 67 },
   { 3421, 3505, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3423, 3506, 67 },
   { 3423, 3506, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4062, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3457, 422, 60 },
   { 3458, 425, 60 },
   { 3459, 429, 60 },
   { 3460, 432, 60 },
   { 3473, 454, 61 },
   { 3474, 457, 61 },
   { 4062, 128, 0 },
   { 4064, 4066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4063, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3461, 403, 60 },
   { 3462, 407, 60 },
   { 3463, 412, 60 },
   { 3464, 416, 60 },
   { 3475, 446, 61 },
   { 3476, 449, 61 },
   { 4063, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3500, 3501, 65 },
   { 3500, 4219, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3500, 3501, 65 },
   { 3500, 4219, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3500, 3501, 65 },
   { 3500, 4219, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3500, 3501, 65 },
   { 3500, 4219, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3508, 4220, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3515, 4221, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3659, 16, 0 },
   { 3660, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3665, 3667, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3668, 3669, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3670, 3672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4120, 4123, 95 },
   { 4125, 4129, 95 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4097, 4101, 90 },
   { 4097, 4105, 91 },
   { 4097, 4109, 92 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4110, 4114, 91 },
   { 4110, 4116, 93 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3889, 3892, 71 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3893, 3896, 71 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3897, 3900, 72 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3901, 3902, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3907, 3911, 74 },
   { 3907, 4010, 85 },
   { 3907, 4012, 86 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3912, 3915, 75 },
   { 3925, 3929, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3916, 3918, 75 },
   { 3930, 3932, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3919, 3921, 75 },
   { 3933, 3937, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3922, 3924, 75 },
   { 3938, 3940, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3941, 3942, 77 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3943, 3944, 77 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4016, 4019, 87 },
   { 4016, 4090, 89 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4020, 4022, 87 },
   { 4020, 4096, 89 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4023, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4117, 4118, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4149, 4155, 98 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4157, 4163, 99 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4164, 4171, 100 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4172, 4179, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4180, 4186, 102 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4187, 4194, 103 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4195, 4203, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4204, 4213, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4223, 4225, 107 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 964, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4232, 4236, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4237, 4241, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4242, 4243, 113 },
   { 4242, 49, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4244, 4237, 114 },
   { 4244, 4246, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4247, 4232, 115 },
   { 4247, 49, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4248, 4243, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4249, 565, 0 },
   { 4250, 4251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4248, 4243, 0 },
   { 4249, 565, 0 },
   { 4250, 4251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4249, 565, 0 },
   { 4250, 4251, 0 },
   { 4252, 4254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4248, 4243, 0 },
   { 4249, 565, 0 },
   { 4250, 4251, 0 },
   { 4252, 4254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4255, 4256, 1 },
   { 4258, 4262, 1 },
   { 4263, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4462, 770, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4463, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4464, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4465, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5442, 5462, 122 },
   { 5464, 5482, 122 },
   { 5484, 5505, 122 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5841, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5842, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5843, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5844, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5845, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5846, 117, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5848, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5851, 141, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5849, 141, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5849, 141, 0 },
   { 5851, 141, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6142, 6147, 125 },
   { 6148, 6152, 125 },
   { 6153, 6158, 125 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5353, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5325, 5298, 0 },
   { 5378, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 3698, 3699, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 3698, 3699, 0 },
   { 3742, 3699, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 3681, 3682, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 1037, 1040, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5837, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5300, 5298, 0 },
   { 5353, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5325, 5298, 0 },
   { 5378, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5300, 5298, 0 },
   { 5353, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5325, 5298, 0 },
   { 5378, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2843, 2844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5325, 5298, 0 },
   { 5353, 49, 0 },
   { 5378, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5353, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5353, 49, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5325, 5298, 0 },
   { 5378, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5325, 5298, 0 },
   { 5378, 49, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5405, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3788, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { 3788, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 4058, 4059, 88 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 3728, 3729, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 3728, 3729, 0 },
   { 3770, 3729, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 322, 324, 0 },
   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 326, 327, 0 },
   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 333, 334, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 336, 337, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 329, 331, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5833, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 185, 187, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 192, 194, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 220, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 220, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5322, 4277, 0 },
   { 5350, 4290, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5402, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3547, 3500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3550, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3804, 3789, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3806, 3792, 0 },
   { 3811, 3792, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3807, 3795, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3808, 3798, 0 },
   { 3812, 3798, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3809, 3801, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3058, 3060, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3882, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3882, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3882, 99, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3882, 99, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3229, 770, 0 },
   { 3230, 1144, 0 },
   { 3233, 3234, 0 },
   { 3235, 3236, 0 },
   { 3239, 3240, 0 },
   { 3241, 3242, 0 },
   { 3243, 3244, 0 },
   { 3245, 3246, 0 },
   { 3248, 3249, 0 },
   { 3250, 3251, 0 },
   { 3252, 3253, 59 },
   { 3254, 3255, 59 },
   { 3258, 3259, 59 },
   { 3260, 3261, 59 },
   { 3262, 3, 0 },
   { 3264, 2416, 0 },
   { 3265, 3267, 0 },
   { 3269, 1241, 0 },
   { 3270, 3266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5916, 5917, 0 },
   { 5920, 5917, 0 },
   { 5923, 5917, 0 },
   { 5926, 5917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5890, 5891, 0 },
   { 5894, 5891, 0 },
   { 5897, 5891, 0 },
   { 5900, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6085, 6086, 0 },
   { 6085, 6086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6121, 6122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6117, 6118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5903, 5904, 0 },
   { 5907, 5904, 0 },
   { 5910, 5904, 0 },
   { 5913, 5904, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5875, 5877, 0 },
   { 5880, 5877, 0 },
   { 5883, 5877, 0 },
   { 5886, 5877, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5929, 5930, 0 },
   { 5933, 5930, 0 },
   { 5936, 5930, 0 },
   { 5939, 5930, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5942, 5943, 0 },
   { 5946, 5943, 0 },
   { 5949, 5943, 0 },
   { 5952, 5943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5854, 5855, 0 },
   { 5858, 5855, 0 },
   { 5860, 5862, 0 },
   { 5864, 5862, 0 },
   { 5867, 5869, 0 },
   { 5871, 5869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1323, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1323, 1324, 0 },
   { 1326, 1324, 0 },
   { 1329, 1331, 0 },
   { 1334, 1331, 0 },
   { 1338, 1340, 0 },
   { 1343, 1340, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3883, 3884, 0 },
   { 3886, 3888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6077, 6078, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1809, 1811, 0 },
   { 1918, 1920, 0 },
   { 2012, 2014, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5968, 5969, 0 },
   { 5972, 5969, 0 },
   { 5975, 5969, 0 },
   { 5978, 5969, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5981, 5982, 0 },
   { 5985, 5982, 0 },
   { 5988, 5982, 0 },
   { 5991, 5982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5994, 5995, 0 },
   { 5998, 5995, 0 },
   { 6001, 5995, 0 },
   { 6004, 5995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6007, 6008, 0 },
   { 6011, 6008, 0 },
   { 6014, 6008, 0 },
   { 6017, 6008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6020, 6021, 0 },
   { 6024, 6021, 0 },
   { 6027, 6021, 0 },
   { 6030, 6021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6033, 6034, 0 },
   { 6037, 6034, 0 },
   { 6040, 6034, 0 },
   { 6043, 6034, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5292, 5287, 0 },
   { 5345, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5317, 5287, 0 },
   { 5370, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5422, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5397, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5955, 5956, 0 },
   { 5959, 5956, 0 },
   { 5962, 5956, 0 },
   { 5965, 5956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6081, 6082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6046, 6047, 0 },
   { 6050, 6047, 0 },
   { 6054, 6056, 0 },
   { 6059, 6056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6073, 6074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6064, 6066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6069, 6070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6089, 6090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6093, 6094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6097, 6098, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6101, 6102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6105, 6106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6109, 6110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 6113, 6114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2300, 2301, 0 },
   { 2331, 2332, 0 },
   { 2361, 959, 0 },
   { 2391, 2392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2300, 2301, 0 },
   { 2331, 2332, 0 },
   { 2361, 959, 0 },
   { 2391, 2392, 0 },
   { 2932, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2932, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 836, 838, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 840, 849, 0 },
   { 862, 870, 0 },
   { 883, 890, 0 },
   { 903, 911, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1033, 1035, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1003, 1004, 0 },
   { 1010, 1011, 0 },
   { 1017, 1018, 0 },
   { 1026, 1027, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 923, 926, 0 },
   { 931, 934, 0 },
   { 939, 942, 0 },
   { 947, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 923, 926, 0 },
   { 931, 934, 0 },
   { 939, 942, 0 },
   { 947, 950, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 923, 926, 0 },
   { 931, 934, 0 },
   { 939, 942, 0 },
   { 947, 950, 0 },
   { 2300, 2301, 0 },
   { 2331, 2332, 0 },
   { 2361, 959, 0 },
   { 2391, 2392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 923, 926, 0 },
   { 931, 934, 0 },
   { 939, 942, 0 },
   { 947, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 927, 926, 0 },
   { 935, 934, 0 },
   { 943, 942, 0 },
   { 951, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 927, 926, 0 },
   { 935, 934, 0 },
   { 943, 942, 0 },
   { 951, 950, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 927, 926, 0 },
   { 935, 934, 0 },
   { 943, 942, 0 },
   { 951, 950, 0 },
   { 2300, 2301, 0 },
   { 2331, 2332, 0 },
   { 2361, 959, 0 },
   { 2391, 2392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 927, 926, 0 },
   { 935, 934, 0 },
   { 943, 942, 0 },
   { 951, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { 3537, 3538, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { 3542, 3543, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4060, 4061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { 4060, 4061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2300, 2301, 0 },
   { 2331, 2332, 0 },
   { 2361, 959, 0 },
   { 2391, 2392, 0 },
   { 4060, 4061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4060, 4061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 963, 964, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 958, 959, 0 },
   { 1005, 1004, 0 },
   { 1012, 1011, 0 },
   { 1020, 1018, 0 },
   { 1029, 1027, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 968, 970, 0 },
   { 974, 976, 0 },
   { 980, 982, 0 },
   { 986, 988, 0 },
   { 992, 994, 0 },
   { 998, 1000, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2886, 2887, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2886, 2887, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 3772, 3774, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1006, 0 },
   { 1014, 1013, 0 },
   { 1022, 1021, 0 },
   { 1031, 1030, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2872, 61, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2229, 2230, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 2890, 2891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2446, 2447, 0 },
   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 197, 199, 0 },
   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2282, 2284, 44 },
   { 2282, 2290, 46 },
   { 2313, 2315, 44 },
   { 2313, 2321, 46 },
   { 2343, 2345, 44 },
   { 2343, 2351, 46 },
   { 2373, 2375, 44 },
   { 2373, 2381, 46 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2294, 2295, 47 },
   { 2325, 2326, 47 },
   { 2355, 2356, 47 },
   { 2385, 2386, 47 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2220, 2222, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5743, 5744, 0 },
   { 5746, 5747, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5732, 5733, 0 },
   { 5736, 5737, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5738, 5739, 0 },
   { 5741, 5742, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1089, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1789, 1790, 0 },
   { 1898, 1899, 0 },
   { 1992, 1993, 0 },
   { 2868, 32, 0 },
   { 5743, 5744, 0 },
   { 5746, 5747, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2845, 2844, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2225, 2226, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2928, 2929, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1659, 1645, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1772, 1773, 0 },
   { 1774, 1773, 0 },
   { 1881, 1882, 0 },
   { 1883, 1882, 0 },
   { 1975, 1976, 0 },
   { 1977, 1976, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1659, 1645, 0 },
   { 1772, 1773, 0 },
   { 1774, 1773, 0 },
   { 1881, 1882, 0 },
   { 1883, 1882, 0 },
   { 1975, 1976, 0 },
   { 1977, 1976, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1664, 1648, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1664, 1648, 0 },
   { 1772, 1773, 0 },
   { 1774, 1773, 0 },
   { 1881, 1882, 0 },
   { 1883, 1882, 0 },
   { 1975, 1976, 0 },
   { 1977, 1976, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1092, 1090, 0 },
   { 1643, 1645, 0 },
   { 1646, 1648, 0 },
   { 1659, 1645, 0 },
   { 1664, 1648, 0 },
   { 1772, 1773, 0 },
   { 1774, 1773, 0 },
   { 1881, 1882, 0 },
   { 1883, 1882, 0 },
   { 1975, 1976, 0 },
   { 1977, 1976, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1667, 1651, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1273, 39 },
   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1775, 1776, 0 },
   { 1777, 1776, 0 },
   { 1884, 1885, 0 },
   { 1886, 1885, 0 },
   { 1978, 1979, 0 },
   { 1980, 1979, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1273, 39 },
   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1667, 1651, 0 },
   { 1775, 1776, 0 },
   { 1777, 1776, 0 },
   { 1884, 1885, 0 },
   { 1886, 1885, 0 },
   { 1978, 1979, 0 },
   { 1980, 1979, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1670, 1654, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1273, 39 },
   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1670, 1654, 0 },
   { 1775, 1776, 0 },
   { 1777, 1776, 0 },
   { 1884, 1885, 0 },
   { 1886, 1885, 0 },
   { 1978, 1979, 0 },
   { 1980, 1979, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1273, 39 },
   { 1649, 1651, 0 },
   { 1652, 1654, 0 },
   { 1667, 1651, 0 },
   { 1670, 1654, 0 },
   { 1775, 1776, 0 },
   { 1777, 1776, 0 },
   { 1884, 1885, 0 },
   { 1886, 1885, 0 },
   { 1978, 1979, 0 },
   { 1980, 1979, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1791, 1792, 0 },
   { 1900, 1901, 0 },
   { 1994, 1995, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1097, 1090, 0 },
   { 1711, 1712, 0 },
   { 1713, 1714, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1090, 0 },
   { 1719, 1720, 0 },
   { 1721, 1722, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1715, 1716, 0 },
   { 1717, 1718, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1723, 1724, 0 },
   { 1725, 1726, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2208, 2209, 0 },
   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5328, 2191, 0 },
   { 5356, 2199, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2203, 2205, 0 },
   { 2868, 32, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2212, 2213, 0 },
   { 2868, 32, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5408, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2216, 2217, 0 },
   { 2868, 32, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2883, 3, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2884, 2506, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2884, 2506, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 3776, 3778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2876, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2253, 2254, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 202, 204, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2443, 2445, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2291, 2293, 47 },
   { 2322, 2324, 47 },
   { 2352, 2354, 47 },
   { 2382, 2384, 47 },
   { 2404, 2405, 0 },
   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2408, 2409, 0 },
   { 2412, 2413, 0 },
   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2287, 2288, 45 },
   { 2287, 2296, 46 },
   { 2318, 2319, 45 },
   { 2318, 2327, 46 },
   { 2348, 2349, 45 },
   { 2348, 2357, 46 },
   { 2378, 2379, 45 },
   { 2378, 2387, 46 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2248, 2250, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1780, 1783, 0 },
   { 1889, 1892, 0 },
   { 1983, 1986, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2846, 2847, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2251, 2252, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 2930, 2931, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1752, 1065, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1631, 1612, 0 },
   { 1752, 1065, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1752, 1065, 0 },
   { 1758, 1760, 0 },
   { 1763, 1760, 0 },
   { 1867, 1869, 0 },
   { 1872, 1869, 0 },
   { 1961, 1963, 0 },
   { 1966, 1963, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1631, 1612, 0 },
   { 1752, 1065, 0 },
   { 1758, 1760, 0 },
   { 1763, 1760, 0 },
   { 1867, 1869, 0 },
   { 1872, 1869, 0 },
   { 1961, 1963, 0 },
   { 1966, 1963, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1636, 1618, 0 },
   { 1752, 1065, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1636, 1618, 0 },
   { 1752, 1065, 0 },
   { 1758, 1760, 0 },
   { 1763, 1760, 0 },
   { 1867, 1869, 0 },
   { 1872, 1869, 0 },
   { 1961, 1963, 0 },
   { 1966, 1963, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1610, 1612, 0 },
   { 1616, 1618, 0 },
   { 1631, 1612, 0 },
   { 1636, 1618, 0 },
   { 1752, 1065, 0 },
   { 1758, 1760, 0 },
   { 1763, 1760, 0 },
   { 1867, 1869, 0 },
   { 1872, 1869, 0 },
   { 1961, 1963, 0 },
   { 1966, 1963, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1639, 1623, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1766, 1768, 0 },
   { 1771, 1768, 0 },
   { 1875, 1877, 0 },
   { 1880, 1877, 0 },
   { 1969, 1971, 0 },
   { 1974, 1971, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1639, 1623, 0 },
   { 1766, 1768, 0 },
   { 1771, 1768, 0 },
   { 1875, 1877, 0 },
   { 1880, 1877, 0 },
   { 1969, 1971, 0 },
   { 1974, 1971, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1642, 1628, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1642, 1628, 0 },
   { 1766, 1768, 0 },
   { 1771, 1768, 0 },
   { 1875, 1877, 0 },
   { 1880, 1877, 0 },
   { 1969, 1971, 0 },
   { 1974, 1971, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1621, 1623, 0 },
   { 1626, 1628, 0 },
   { 1639, 1623, 0 },
   { 1642, 1628, 0 },
   { 1766, 1768, 0 },
   { 1771, 1768, 0 },
   { 1875, 1877, 0 },
   { 1880, 1877, 0 },
   { 1969, 1971, 0 },
   { 1974, 1971, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1786, 1788, 0 },
   { 1895, 1897, 0 },
   { 1989, 1991, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1673, 1675, 0 },
   { 1678, 1680, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1693, 1695, 0 },
   { 1698, 1700, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1683, 1685, 0 },
   { 1688, 1690, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1703, 1705, 0 },
   { 1708, 1710, 0 },
   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5331, 2233, 0 },
   { 5359, 2237, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5359, 2237, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 2241, 0 },
   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5331, 2233, 0 },
   { 5384, 2237, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 2245, 0 },
   { 2873, 32, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5411, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 2247, 0 },
   { 2873, 32, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 3548, 3500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 3551, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2880, 3, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2880, 3, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 3780, 3782, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2277, 2278, 0 },
   { 2877, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2272, 2274, 0 },
   { 2877, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2275, 2276, 0 },
   { 2877, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2266, 2267, 0 },
   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5334, 2257, 0 },
   { 5362, 2261, 0 },
   { 5387, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5309, 2257, 0 },
   { 5362, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2263, 2265, 0 },
   { 2877, 61, 0 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5334, 2257, 0 },
   { 5387, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2268, 2269, 0 },
   { 2877, 61, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5414, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2270, 2271, 0 },
   { 2877, 61, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2304, 2305, 0 },
   { 2335, 2336, 0 },
   { 2364, 2365, 0 },
   { 2395, 2396, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2304, 2305, 0 },
   { 2335, 2336, 0 },
   { 2364, 2365, 0 },
   { 2395, 2396, 0 },
   { 2938, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2938, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 3467, 454, 61 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { 2304, 2305, 0 },
   { 2335, 2336, 0 },
   { 2364, 2365, 0 },
   { 2395, 2396, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 851, 853, 0 },
   { 872, 874, 0 },
   { 892, 894, 0 },
   { 913, 915, 0 },
   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { 3539, 3536, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { 3544, 3541, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2302, 2303, 0 },
   { 2333, 2334, 0 },
   { 2362, 2363, 0 },
   { 2393, 2394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2302, 2303, 0 },
   { 2333, 2334, 0 },
   { 2362, 2363, 0 },
   { 2393, 2394, 0 },
   { 2934, 1242, 0 },
   { 2936, 2935, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2937, 32, 0 },
   { 3435, 3436, 60 },
   { 3437, 416, 60 },
   { 3438, 3439, 60 },
   { 3468, 446, 61 },
   { 3469, 449, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2937, 32, 0 },
   { 3435, 3436, 60 },
   { 3437, 416, 60 },
   { 3438, 3439, 60 },
   { 3469, 449, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2302, 2303, 0 },
   { 2333, 2334, 0 },
   { 2362, 2363, 0 },
   { 2393, 2394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 855, 860, 0 },
   { 876, 881, 0 },
   { 896, 901, 0 },
   { 917, 922, 0 },
   { 2937, 32, 0 },
   { 3435, 3436, 60 },
   { 3437, 416, 60 },
   { 3438, 3439, 60 },
   { 3469, 449, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3619, 3620, 0 },
   { 3622, 429, 0 },
   { 3623, 425, 0 },
   { 3624, 422, 0 },
   { 3626, 3597, 0 },
   { 3628, 3592, 0 },
   { 3630, 3587, 0 },
   { 3632, 3584, 0 },
   { 3634, 3580, 0 },
   { 3636, 3576, 0 },
   { 3638, 3639, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3569, 3430, 0 },
   { 3571, 425, 0 },
   { 3572, 429, 0 },
   { 3573, 432, 0 },
   { 3575, 3576, 0 },
   { 3578, 3580, 0 },
   { 3582, 3584, 0 },
   { 3585, 3587, 0 },
   { 3590, 3592, 0 },
   { 3595, 3597, 0 },
   { 3598, 3434, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3599, 3430, 0 },
   { 3601, 425, 0 },
   { 3602, 429, 0 },
   { 3603, 432, 0 },
   { 3605, 3576, 0 },
   { 3607, 3580, 0 },
   { 3609, 3584, 0 },
   { 3611, 3587, 0 },
   { 3613, 3592, 0 },
   { 3615, 3597, 0 },
   { 3616, 3434, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3466, 3450, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3465, 3450, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3617, 3444, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3561, 3562, 0 },
   { 3563, 3564, 0 },
   { 3565, 3566, 0 },
   { 3567, 3568, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3640, 3641, 0 },
   { 3642, 412, 0 },
   { 3643, 407, 0 },
   { 3644, 403, 0 },
   { 3645, 3646, 0 },
   { 3647, 3648, 0 },
   { 3649, 3650, 0 },
   { 3651, 3652, 0 },
   { 3653, 3654, 0 },
   { 3655, 3656, 0 },
   { 3657, 3658, 0 },
   { 4130, 4134, 96 },
   { 5311, 11, 0 },
   { 5364, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3553, 3554, 0 },
   { 3555, 3556, 0 },
   { 3557, 3558, 0 },
   { 3559, 3560, 0 },
   { 4130, 4134, 96 },
   { 5311, 11, 0 },
   { 5364, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3404, 21, 0 },
   { 3387, 3496, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3387, 3496, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 48, 49, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 51, 49, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2900, 3502, 66 },
   { 3512, 2900, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3396, 16, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3393, 16, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3388, 16, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3391, 2595, 0 },
   { 2900, 3502, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3686, 3688, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3686, 3688, 0 },
   { 3731, 3688, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3694, 3696, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3694, 3696, 0 },
   { 3735, 3696, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 5824, 5827, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { 5824, 5827, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 262, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3056, 3057, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2848, 1279, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3061, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3744, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3720, 3722, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3720, 3722, 0 },
   { 3751, 3753, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3724, 3726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 3724, 3726, 0 },
   { 3761, 3763, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 5820, 5823, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 161, 163, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 177, 172, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 111, 99, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 113, 112, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2565, 2564, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2567, 2566, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2569, 2568, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2571, 2570, 0 },
   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 294, 295, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 294, 295, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { 3738, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { 3738, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { 3738, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3704, 3706, 0 },
   { 3738, 3706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3709, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3709, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3709, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3709, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3678, 3679, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3678, 3679, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3678, 3679, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { 3678, 3679, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 476, 477, 15 },
   { 551, 552, 17 },
   { 617, 618, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 476, 477, 15 },
   { 551, 552, 17 },
   { 617, 618, 19 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 476, 477, 15 },
   { 490, 493, 15 },
   { 551, 552, 17 },
   { 562, 564, 17 },
   { 617, 618, 19 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 553, 555, 17 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 553, 555, 17 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 304, 306, 0 },
   { 476, 477, 15 },
   { 480, 482, 15 },
   { 551, 552, 17 },
   { 553, 555, 17 },
   { 617, 618, 19 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 476, 477, 15 },
   { 551, 552, 17 },
   { 617, 618, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 476, 477, 15 },
   { 551, 552, 17 },
   { 617, 618, 19 },
   { 695, 697, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 302, 274, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 476, 477, 15 },
   { 490, 493, 15 },
   { 551, 552, 17 },
   { 562, 564, 17 },
   { 617, 618, 19 },
   { 631, 634, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 553, 555, 17 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 302, 274, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 553, 555, 17 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 302, 274, 0 },
   { 304, 306, 0 },
   { 476, 477, 15 },
   { 480, 482, 15 },
   { 551, 552, 17 },
   { 553, 555, 17 },
   { 617, 618, 19 },
   { 621, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { 691, 693, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 490, 493, 15 },
   { 562, 564, 17 },
   { 631, 634, 19 },
   { 691, 693, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 476, 477, 15 },
   { 490, 493, 15 },
   { 551, 552, 17 },
   { 562, 564, 17 },
   { 617, 618, 19 },
   { 631, 634, 19 },
   { 691, 693, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 486, 488, 15 },
   { 553, 555, 17 },
   { 558, 560, 17 },
   { 621, 623, 19 },
   { 627, 629, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 0, 0 },
   { 300, 3, 0 },
   { 301, 0, 0 },
   { 304, 306, 0 },
   { 480, 482, 15 },
   { 486, 488, 15 },
   { 553, 555, 17 },
   { 558, 560, 17 },
   { 621, 623, 19 },
   { 627, 629, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 0, 0 },
   { 304, 306, 0 },
   { 476, 477, 15 },
   { 480, 482, 15 },
   { 486, 488, 15 },
   { 551, 552, 17 },
   { 553, 555, 17 },
   { 558, 560, 17 },
   { 617, 618, 19 },
   { 621, 623, 19 },
   { 627, 629, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2998, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1538, 1540, 0 },
   { 228, 1530, 41 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 1543, 228, 0 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1542, 1541, 0 },
   { 228, 1530, 41 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1817, 11, 0 },
   { 1926, 16, 0 },
   { 2020, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4444, 3280, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3300, 3301, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1815, 11, 0 },
   { 1924, 16, 0 },
   { 2018, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1841, 1842, 0 },
   { 1849, 1850, 0 },
   { 1859, 1860, 0 },
   { 1947, 1948, 0 },
   { 1955, 1956, 0 },
   { 2041, 2042, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2164, 2165, 0 },
   { 3328, 2169, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1204, 1205, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1204, 1205, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1204, 1205, 0 },
   { 1223, 1225, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1118, 1119, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1246, 1121, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1135, 1068, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { 5749, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1144, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1256, 1121, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2827, 1062, 0 },
   { 2835, 1059, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2828, 1059, 0 },
   { 2836, 1062, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2825, 1068, 0 },
   { 2833, 1065, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2826, 1065, 0 },
   { 2834, 1068, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1204, 1205, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1104, 1105, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1176, 1175, 0 },
   { 1419, 1175, 0 },
   { 1423, 1175, 0 },
   { 1427, 1175, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2866, 1248, 0 },
   { 5219, 5221, 121 },
   { 5236, 5221, 121 },
   { 5244, 5245, 0 },
   { 5258, 5260, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1184, 1185, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1265, 1266, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1110, 1111, 0 },
   { 1184, 1185, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1252, 1253, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1129, 1131, 0 },
   { 1139, 1140, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { 5773, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1263, 1264, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1184, 1185, 0 },
   { 1200, 1201, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 2473, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 2467, 2468, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 1209, 1212, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1109, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1101, 1091, 0 },
   { 1184, 1185, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1251, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { 5765, 5768, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2469, 1091, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2469, 1091, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1184, 1185, 0 },
   { 2469, 1091, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2469, 1091, 0 },
   { 2473, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2469, 1091, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1261, 1262, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1184, 1185, 0 },
   { 2474, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1090, 0 },
   { 2474, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2474, 1090, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { 5226, 5227, 121 },
   { 5233, 5234, 121 },
   { 5239, 5227, 121 },
   { 5243, 5234, 121 },
   { 5250, 5251, 0 },
   { 5255, 5256, 0 },
   { 5265, 5266, 0 },
   { 5271, 5272, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1186, 1187, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2471, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 2476, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1269, 1270, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1114, 1115, 0 },
   { 1186, 1187, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1249, 1250, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { 5783, 5786, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1143, 1144, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2471, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1259, 1260, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 2476, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1202, 1203, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1202, 1203, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1186, 1187, 0 },
   { 1202, 1203, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1202, 1203, 0 },
   { 2471, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1202, 1203, 0 },
   { 2475, 2476, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1202, 1203, 0 },
   { 1218, 1219, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1113, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1102, 1103, 0 },
   { 1186, 1187, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1134, 1059, 0 },
   { 1138, 1115, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { 5779, 5782, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1103, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1103, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1186, 1187, 0 },
   { 2477, 1103, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2471, 1248, 0 },
   { 2477, 1103, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 2476, 0 },
   { 2477, 1103, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1257, 1258, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2472, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2472, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1186, 1187, 0 },
   { 2472, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2471, 1248, 0 },
   { 2472, 1248, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2472, 1248, 0 },
   { 2475, 2476, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { 5224, 5225, 121 },
   { 5230, 5232, 121 },
   { 5238, 5225, 121 },
   { 5242, 5232, 121 },
   { 5248, 5249, 0 },
   { 5253, 5254, 0 },
   { 5263, 5264, 0 },
   { 5268, 5270, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1206, 1207, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1206, 1207, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1206, 1207, 0 },
   { 1231, 1232, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1116, 1117, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1123, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1136, 1065, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { 5753, 5756, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1240, 770, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1255, 1123, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2831, 1062, 0 },
   { 2839, 1059, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2832, 1059, 0 },
   { 2840, 1062, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2829, 1068, 0 },
   { 2837, 1065, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2830, 1065, 0 },
   { 2838, 1068, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1206, 1207, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1106, 1107, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1165, 1164, 0 },
   { 1420, 1404, 0 },
   { 1424, 1404, 0 },
   { 1428, 1404, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 1090, 0 },
   { 5222, 5223, 121 },
   { 5237, 5223, 121 },
   { 5246, 5247, 0 },
   { 5261, 5262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1237, 1238, 0 },
   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 209, 210, 0 },
   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1122, 1123, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1125, 1123, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1237, 1238, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 209, 210, 0 },
   { 1436, 1435, 0 },
   { 1437, 1438, 0 },
   { 1458, 1457, 0 },
   { 1459, 1460, 0 },
   { 1480, 1479, 0 },
   { 1481, 1482, 0 },
   { 1500, 1499, 0 },
   { 1501, 1502, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 214, 215, 0 },
   { 1441, 1438, 0 },
   { 1445, 1435, 0 },
   { 1463, 1460, 0 },
   { 1467, 1457, 0 },
   { 1485, 1482, 0 },
   { 1488, 1479, 0 },
   { 1505, 1502, 0 },
   { 1509, 1499, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { 5761, 5764, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1244, 770, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1446, 1438, 0 },
   { 1450, 1435, 0 },
   { 1468, 1460, 0 },
   { 1472, 1457, 0 },
   { 1489, 1482, 0 },
   { 1492, 1479, 0 },
   { 1510, 1502, 0 },
   { 1514, 1499, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { 4588, 4591, 120 },
   { 4635, 4638, 120 },
   { 4681, 4684, 120 },
   { 4748, 4751, 121 },
   { 4799, 4802, 121 },
   { 4877, 4880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { 4929, 4932, 120 },
   { 4970, 4973, 120 },
   { 5013, 5016, 120 },
   { 5077, 5080, 121 },
   { 5122, 5125, 121 },
   { 5194, 5197, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { 4083, 4079, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { 4074, 4079, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 209, 210, 0 },
   { 1237, 1238, 0 },
   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1403, 1404, 0 },
   { 1408, 1404, 0 },
   { 1412, 1404, 0 },
   { 1416, 1404, 0 },
   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 4567, 4568, 120 },
   { 4622, 4568, 120 },
   { 4667, 4568, 120 },
   { 4715, 4716, 121 },
   { 4730, 4731, 121 },
   { 4779, 4716, 121 },
   { 4787, 4731, 121 },
   { 4832, 4833, 0 },
   { 4845, 4846, 0 },
   { 4859, 4860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 4912, 4913, 120 },
   { 4958, 4913, 120 },
   { 5000, 4913, 120 },
   { 5045, 5046, 121 },
   { 5059, 5060, 121 },
   { 5102, 5046, 121 },
   { 5110, 5060, 121 },
   { 5150, 5151, 0 },
   { 5163, 5164, 0 },
   { 5176, 5177, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1520, 32, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1586, 1587, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1285, 1288, 0 },
   { 1348, 32, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1530, 228, 39 },
   { 1534, 1535, 39 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1281, 1279, 0 },
   { 1348, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1524, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1524, 99, 0 },
   { 1528, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1582, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1532, 228, 39 },
   { 1537, 1535, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1524, 99, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1524, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1585, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 1276, 0 },
   { 1349, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 3675, 3676, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 3871, 3872, 0 },
   { 3874, 3875, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { 3675, 3676, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { 3690, 3692, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { 3690, 3692, 0 },
   { 3733, 3692, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { 5829, 5832, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 121, 0 },
   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 255, 16, 0 },
   { 3871, 3872, 0 },
   { 3874, 3875, 0 },
   { 5829, 5832, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 3675, 3676, 7 },
   { 3871, 3872, 0 },
   { 3874, 3875, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 7 },
   { 241, 238, 0 },
   { 242, 243, 0 },
   { 3063, 706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 180, 182, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 297, 298, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 297, 298, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 278, 279, 7 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 180, 182, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { 3739, 3708, 0 },
   { 3740, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 278, 279, 7 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { 3739, 3708, 0 },
   { 3740, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 180, 182, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { 3739, 3708, 0 },
   { 3740, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 271, 272, 0 },
   { 283, 284, 7 },
   { 289, 291, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3707, 3708, 0 },
   { 3739, 3708, 0 },
   { 3740, 3710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3680, 397, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 278, 279, 7 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3680, 397, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 145, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3680, 397, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 172, 0 },
   { 244, 245, 7 },
   { 246, 245, 0 },
   { 247, 248, 0 },
   { 397, 298, 31 },
   { 758, 298, 34 },
   { 3680, 397, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 110, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6126, 6129, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2298, 6, 0 },
   { 2329, 11, 0 },
   { 2359, 16, 0 },
   { 2389, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 116, 114, 0 },
   { 3339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 115, 114, 0 },
   { 3339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3339, 32, 0 },
   { 6130, 6133, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 2859, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 208, 0 },
   { 2859, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1120, 1121, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1124, 1121, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 208, 0 },
   { 1434, 1435, 0 },
   { 1439, 1438, 0 },
   { 1456, 1457, 0 },
   { 1461, 1460, 0 },
   { 1478, 1479, 0 },
   { 1483, 1482, 0 },
   { 1498, 1499, 0 },
   { 1503, 1502, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 212, 213, 0 },
   { 1442, 1435, 0 },
   { 1444, 1438, 0 },
   { 1464, 1457, 0 },
   { 1466, 1460, 0 },
   { 1486, 1479, 0 },
   { 1487, 1482, 0 },
   { 1506, 1499, 0 },
   { 1508, 1502, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { 5757, 5760, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1243, 1144, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1447, 1435, 0 },
   { 1449, 1438, 0 },
   { 1469, 1457, 0 },
   { 1471, 1460, 0 },
   { 1490, 1479, 0 },
   { 1491, 1482, 0 },
   { 1511, 1499, 0 },
   { 1513, 1502, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { 4582, 4587, 120 },
   { 4629, 4634, 120 },
   { 4675, 4680, 120 },
   { 4742, 4747, 121 },
   { 4793, 4798, 121 },
   { 4871, 4876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { 4925, 4928, 120 },
   { 4965, 4969, 120 },
   { 5007, 5012, 120 },
   { 5071, 5076, 121 },
   { 5116, 5121, 121 },
   { 5188, 5193, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { 4084, 4081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { 4080, 4081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 208, 0 },
   { 1235, 1236, 0 },
   { 2859, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1402, 1175, 0 },
   { 1407, 1175, 0 },
   { 1411, 1175, 0 },
   { 1415, 1175, 0 },
   { 2859, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 4564, 4566, 120 },
   { 4621, 4566, 120 },
   { 4666, 4566, 120 },
   { 4713, 4714, 121 },
   { 4727, 4729, 121 },
   { 4778, 4714, 121 },
   { 4786, 4729, 121 },
   { 4830, 4831, 0 },
   { 4843, 4844, 0 },
   { 4856, 4858, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 1248, 0 },
   { 4909, 4911, 120 },
   { 4957, 4911, 120 },
   { 4999, 4911, 120 },
   { 5043, 5044, 121 },
   { 5056, 5058, 121 },
   { 5101, 5044, 121 },
   { 5109, 5058, 121 },
   { 5148, 5149, 0 },
   { 5161, 5162, 0 },
   { 5173, 5175, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 4225, 4223, 108 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1516, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1525, 114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1525, 114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 4227, 4223, 108 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1525, 114, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1525, 114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1819, 11, 0 },
   { 1928, 16, 0 },
   { 2022, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4443, 3280, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3302, 3303, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1818, 11, 0 },
   { 1927, 16, 0 },
   { 2021, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1839, 1840, 0 },
   { 1847, 1848, 0 },
   { 1857, 1858, 0 },
   { 1945, 1946, 0 },
   { 1953, 1954, 0 },
   { 2039, 2040, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2162, 2163, 0 },
   { 3329, 2166, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1835, 1836, 0 },
   { 1852, 1836, 0 },
   { 1862, 1836, 0 },
   { 1943, 1851, 0 },
   { 1957, 1851, 0 },
   { 2037, 1861, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4442, 1541, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { 6384, 6386, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { 6368, 6370, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { 6355, 6356, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1837, 1838, 0 },
   { 1854, 1838, 0 },
   { 1864, 1838, 0 },
   { 1944, 1853, 0 },
   { 1958, 1853, 0 },
   { 2038, 1863, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1833, 1834, 0 },
   { 1846, 11, 0 },
   { 1856, 11, 0 },
   { 1941, 1942, 0 },
   { 1952, 16, 0 },
   { 2035, 2036, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3321, 16, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3326, 2158, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 2155, 0 },
   { 4459, 11, 0 },
   { 4460, 16, 0 },
   { 4461, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2156, 2157, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3315, 1417, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6387, 6389, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6371, 6373, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6357, 6359, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2159, 2160, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3304, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2505, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2505, 1090, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { 5807, 5810, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2550, 2551, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { 4043, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2527, 2529, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { 4939, 4944, 120 },
   { 4980, 4985, 120 },
   { 5024, 5030, 120 },
   { 5086, 5090, 121 },
   { 5132, 5137, 121 },
   { 5203, 5207, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2505, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2505, 1090, 0 },
   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { 5803, 5806, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2505, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2488, 1098, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2531, 2533, 0 },
   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2499, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2499, 1090, 0 },
   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2499, 1090, 0 },
   { 2505, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2499, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2486, 2487, 0 },
   { 2499, 1090, 0 },
   { 2861, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2499, 1090, 0 },
   { 2552, 2553, 0 },
   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3386, 1090, 0 },
   { 4933, 4938, 120 },
   { 4974, 4979, 120 },
   { 5017, 5023, 120 },
   { 5081, 5085, 121 },
   { 5126, 5131, 121 },
   { 5198, 5202, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 4916, 4917, 120 },
   { 4923, 4924, 120 },
   { 4960, 4917, 120 },
   { 4964, 4924, 120 },
   { 5002, 4917, 120 },
   { 5006, 4924, 120 },
   { 5049, 5050, 121 },
   { 5054, 5055, 121 },
   { 5063, 5064, 121 },
   { 5069, 5070, 121 },
   { 5104, 5050, 121 },
   { 5107, 5055, 121 },
   { 5112, 5064, 121 },
   { 5115, 5070, 121 },
   { 5154, 5155, 0 },
   { 5159, 5160, 0 },
   { 5167, 5168, 0 },
   { 5171, 5172, 0 },
   { 5180, 5181, 0 },
   { 5186, 5187, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { 5815, 5818, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2535, 2537, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2554, 2555, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { 4951, 4956, 120 },
   { 4992, 4997, 120 },
   { 5037, 5042, 120 },
   { 5096, 5100, 121 },
   { 5143, 5147, 121 },
   { 5213, 5217, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2509, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1248, 0 },
   { 2509, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2509, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2509, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { 5811, 5814, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2502, 2503, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2502, 2503, 0 },
   { 2508, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2502, 2503, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2500, 2501, 0 },
   { 2502, 2503, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2502, 2503, 0 },
   { 2556, 2557, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2491, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2491, 1248, 0 },
   { 2508, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 1248, 0 },
   { 2491, 1248, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2491, 1248, 0 },
   { 2500, 2501, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2491, 1248, 0 },
   { 2539, 2541, 0 },
   { 2862, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { 4945, 4950, 120 },
   { 4986, 4991, 120 },
   { 5031, 5036, 120 },
   { 5091, 5095, 121 },
   { 5138, 5142, 121 },
   { 5208, 5212, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 1248, 0 },
   { 4914, 4915, 120 },
   { 4920, 4922, 120 },
   { 4959, 4915, 120 },
   { 4963, 4922, 120 },
   { 5001, 4915, 120 },
   { 5005, 4922, 120 },
   { 5047, 5048, 121 },
   { 5052, 5053, 121 },
   { 5061, 5062, 121 },
   { 5066, 5068, 121 },
   { 5103, 5048, 121 },
   { 5106, 5053, 121 },
   { 5111, 5062, 121 },
   { 5114, 5068, 121 },
   { 5152, 5153, 0 },
   { 5157, 5158, 0 },
   { 5165, 5166, 0 },
   { 5169, 5170, 0 },
   { 5178, 5179, 0 },
   { 5183, 5185, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1754, 1755, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 3289, 2844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4025, 4026, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4023, 2887, 0 },
   { 4025, 4026, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 5273, 3502, 0 },
   { 5274, 3505, 0 },
   { 5276, 5134, 0 },
   { 5278, 2177, 0 },
   { 5279, 2193, 0 },
   { 5281, 5027, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4448, 3280, 0 },
   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 6393, 6395, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 6377, 6379, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 6362, 6363, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2046, 770, 0 },
   { 2050, 770, 0 },
   { 2067, 770, 0 },
   { 2079, 770, 0 },
   { 2092, 6, 0 },
   { 2107, 6, 0 },
   { 2117, 6, 0 },
   { 2128, 11, 0 },
   { 2139, 11, 0 },
   { 2148, 16, 0 },
   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 4467, 4468, 0 },
   { 4475, 4468, 0 },
   { 4481, 4468, 0 },
   { 4487, 4468, 0 },
   { 4491, 2045, 0 },
   { 4493, 4466, 0 },
   { 4497, 4466, 0 },
   { 4501, 4466, 0 },
   { 4505, 2049, 0 },
   { 4507, 2049, 0 },
   { 4509, 2091, 0 },
   { 4511, 4474, 0 },
   { 4515, 4474, 0 },
   { 4519, 2061, 0 },
   { 4521, 2061, 0 },
   { 4523, 2101, 0 },
   { 4525, 2061, 0 },
   { 4527, 2101, 0 },
   { 4529, 2127, 0 },
   { 4531, 4480, 0 },
   { 4535, 2065, 0 },
   { 4537, 2065, 0 },
   { 4539, 2105, 0 },
   { 4541, 2065, 0 },
   { 4543, 2105, 0 },
   { 4545, 2137, 0 },
   { 4547, 2065, 0 },
   { 4549, 2105, 0 },
   { 4551, 2137, 0 },
   { 4553, 2147, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 4470, 4471, 0 },
   { 4477, 4471, 0 },
   { 4483, 4471, 0 },
   { 4488, 4471, 0 },
   { 4494, 4469, 0 },
   { 4498, 4469, 0 },
   { 4502, 4469, 0 },
   { 4512, 4476, 0 },
   { 4516, 4476, 0 },
   { 4532, 4482, 0 },
   { 5310, 11, 0 },
   { 5363, 16, 0 },
   { 5415, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3322, 16, 0 },
   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3324, 2169, 0 },
   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 4555, 4556, 0 },
   { 4555, 4556, 0 },
   { 4557, 4558, 0 },
   { 4557, 4558, 0 },
   { 4559, 4452, 0 },
   { 4559, 4452, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3447, 3449, 60 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4446, 3280, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 6390, 6392, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 6374, 6376, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 6360, 6361, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 4472, 4468, 0 },
   { 4478, 4468, 0 },
   { 4484, 4468, 0 },
   { 4489, 4468, 0 },
   { 4495, 4466, 0 },
   { 4499, 4466, 0 },
   { 4503, 4466, 0 },
   { 4513, 4474, 0 },
   { 4517, 4474, 0 },
   { 4533, 4480, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2048, 770, 0 },
   { 2052, 770, 0 },
   { 2068, 770, 0 },
   { 2080, 770, 0 },
   { 2094, 6, 0 },
   { 2108, 6, 0 },
   { 2118, 6, 0 },
   { 2130, 11, 0 },
   { 2140, 11, 0 },
   { 2150, 16, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 4473, 4471, 0 },
   { 4479, 4471, 0 },
   { 4485, 4471, 0 },
   { 4490, 4471, 0 },
   { 4492, 2047, 0 },
   { 4496, 4469, 0 },
   { 4500, 4469, 0 },
   { 4504, 4469, 0 },
   { 4506, 2051, 0 },
   { 4508, 2051, 0 },
   { 4510, 2093, 0 },
   { 4514, 4476, 0 },
   { 4518, 4476, 0 },
   { 4520, 2059, 0 },
   { 4522, 2059, 0 },
   { 4524, 2099, 0 },
   { 4526, 2059, 0 },
   { 4528, 2099, 0 },
   { 4530, 2129, 0 },
   { 4534, 4482, 0 },
   { 4536, 2063, 0 },
   { 4538, 2063, 0 },
   { 4540, 2103, 0 },
   { 4542, 2063, 0 },
   { 4544, 2103, 0 },
   { 4546, 2135, 0 },
   { 4548, 2063, 0 },
   { 4550, 2103, 0 },
   { 4552, 2135, 0 },
   { 4554, 2149, 0 },
   { 5282, 11, 0 },
   { 5335, 16, 0 },
   { 5388, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3323, 16, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3443, 3445, 60 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3325, 2166, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 4560, 2193, 0 },
   { 4560, 2193, 0 },
   { 4561, 3505, 0 },
   { 4561, 3505, 0 },
   { 4562, 4457, 0 },
   { 4562, 4457, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2894, 3504, 66 },
   { 3519, 2894, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3398, 1473, 0 },
   { 2894, 3504, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3395, 128, 0 },
   { 2894, 3504, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3389, 36, 0 },
   { 2894, 3504, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3392, 2598, 0 },
   { 2894, 3504, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3421, 3505, 67 },
   { 3509, 3421, 0 },
   { 3421, 3505, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3421, 3505, 67 },
   { 3523, 3421, 0 },
   { 3524, 3423, 0 },
   { 3421, 3505, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3423, 3506, 67 },
   { 3516, 3423, 0 },
   { 3423, 3506, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4028, 4021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4030, 4021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4030, 4021, 0 },
   { 4062, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4028, 4021, 0 },
   { 4030, 4021, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4031, 4018, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4032, 4018, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4032, 4018, 0 },
   { 4063, 128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4031, 4018, 0 },
   { 4032, 4018, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3507, 3508, 0 },
   { 3508, 4220, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3522, 3515, 0 },
   { 3508, 4220, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3520, 3508, 0 },
   { 3521, 3515, 0 },
   { 3508, 4220, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3514, 3515, 0 },
   { 3515, 4221, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4034, 4035, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4037, 4035, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4037, 4035, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4034, 4035, 0 },
   { 4037, 4035, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1082, 1083, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1085, 1086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6134, 6137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1064, 1065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2882, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1058, 1059, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1067, 1068, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1070, 1071, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1073, 1074, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1076, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1080, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2632, 2637, 49 },
   { 5709, 2632, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2632, 2637, 49 },
   { 5710, 2702, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2638, 2644, 49 },
   { 5715, 2638, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2638, 2644, 49 },
   { 5716, 2706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2645, 2651, 49 },
   { 5719, 2645, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2645, 2651, 49 },
   { 5720, 2712, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2652, 2668, 49 },
   { 5723, 2652, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2652, 2668, 49 },
   { 5724, 2717, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2669, 2701, 49 },
   { 5727, 2669, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2669, 2701, 49 },
   { 5728, 2726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2420, 2421, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2743, 2746, 49 },
   { 5713, 2743, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2743, 2746, 49 },
   { 5714, 2784, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2747, 2752, 49 },
   { 5717, 2747, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2747, 2752, 49 },
   { 5718, 2788, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2753, 2757, 49 },
   { 5721, 2753, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2753, 2757, 49 },
   { 5722, 2794, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2758, 2766, 49 },
   { 5725, 2758, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2758, 2766, 49 },
   { 5726, 2799, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2767, 2783, 49 },
   { 5729, 2767, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2767, 2783, 49 },
   { 5730, 2808, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2492, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 2480, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 5791, 5794, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2492, 1090, 0 },
   { 2542, 2543, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 2480, 0 },
   { 2511, 2513, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 4601, 4607, 120 },
   { 4647, 4652, 120 },
   { 4694, 4700, 120 },
   { 4760, 4765, 121 },
   { 4811, 4817, 121 },
   { 4889, 4895, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 1502, 3994, 84 },
   { 5787, 5790, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1096, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1096, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 3987, 3989, 83 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1096, 0 },
   { 2492, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 2480, 0 },
   { 2481, 1096, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1096, 0 },
   { 2515, 2517, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 3987, 3989, 83 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2493, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2493, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 3983, 3985, 83 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2492, 1090, 0 },
   { 2493, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 2480, 0 },
   { 2493, 1090, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2493, 1090, 0 },
   { 2544, 2545, 0 },
   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 3983, 3985, 83 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 1502, 3994, 84 },
   { 4592, 4599, 120 },
   { 4639, 4646, 120 },
   { 4685, 4692, 120 },
   { 4752, 4759, 121 },
   { 4803, 4810, 121 },
   { 4881, 4888, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 4571, 4572, 120 },
   { 4580, 4581, 120 },
   { 4624, 4572, 120 },
   { 4628, 4581, 120 },
   { 4669, 4572, 120 },
   { 4673, 4581, 120 },
   { 4719, 4720, 121 },
   { 4724, 4725, 121 },
   { 4734, 4735, 121 },
   { 4740, 4741, 121 },
   { 4781, 4720, 121 },
   { 4784, 4725, 121 },
   { 4789, 4735, 121 },
   { 4792, 4741, 121 },
   { 4836, 4837, 0 },
   { 4841, 4842, 0 },
   { 4849, 4850, 0 },
   { 4853, 4854, 0 },
   { 4863, 4864, 0 },
   { 4869, 4870, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1248, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2494, 2495, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 5799, 5802, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1248, 0 },
   { 2519, 2521, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2494, 2495, 0 },
   { 2546, 2547, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 4614, 4619, 120 },
   { 4659, 4664, 120 },
   { 4707, 4712, 120 },
   { 4772, 4777, 121 },
   { 4824, 4829, 121 },
   { 4902, 4907, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { 5795, 5798, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2496, 2497, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2496, 2497, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 3992, 3993, 83 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1248, 0 },
   { 2496, 2497, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2494, 2495, 0 },
   { 2496, 2497, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2496, 2497, 0 },
   { 2548, 2549, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 3992, 3993, 83 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 1248, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 1248, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 3990, 3991, 83 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1248, 0 },
   { 2484, 1248, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 1248, 0 },
   { 2494, 2495, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 1248, 0 },
   { 2523, 2525, 0 },
   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 3990, 3991, 83 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { 4608, 4613, 120 },
   { 4653, 4658, 120 },
   { 4701, 4706, 120 },
   { 4766, 4771, 121 },
   { 4818, 4823, 121 },
   { 4896, 4901, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 4569, 4570, 120 },
   { 4577, 4579, 120 },
   { 4623, 4570, 120 },
   { 4627, 4579, 120 },
   { 4668, 4570, 120 },
   { 4672, 4579, 120 },
   { 4717, 4718, 121 },
   { 4722, 4723, 121 },
   { 4732, 4733, 121 },
   { 4737, 4739, 121 },
   { 4780, 4718, 121 },
   { 4783, 4723, 121 },
   { 4788, 4733, 121 },
   { 4791, 4739, 121 },
   { 4834, 4835, 0 },
   { 4839, 4840, 0 },
   { 4847, 4848, 0 },
   { 4851, 4852, 0 },
   { 4861, 4862, 0 },
   { 4866, 4868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4223, 4225, 107 },
   { 6138, 6141, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3000, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3052, 3053, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3044, 3045, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3047, 3048, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3050, 3051, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3035, 3037, 0 },
   { 3017, 3070, 56 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3038, 3039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3069, 3049, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3068, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3071, 3017, 57 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3040, 3043, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3317, 1405, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3307, 3308, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3318, 1405, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3309, 3310, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4445, 1541, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2053, 2054, 0 },
   { 2057, 2054, 0 },
   { 2060, 2054, 0 },
   { 2064, 2054, 0 },
   { 2069, 2070, 0 },
   { 2073, 2070, 0 },
   { 2075, 2070, 0 },
   { 2077, 2070, 0 },
   { 2081, 2082, 0 },
   { 2085, 2082, 0 },
   { 2087, 2082, 0 },
   { 2089, 2082, 0 },
   { 2095, 2096, 0 },
   { 2100, 2096, 0 },
   { 2104, 2096, 0 },
   { 2109, 2110, 0 },
   { 2113, 2110, 0 },
   { 2115, 2110, 0 },
   { 2119, 2120, 0 },
   { 2123, 2120, 0 },
   { 2125, 2120, 0 },
   { 2131, 2132, 0 },
   { 2136, 2132, 0 },
   { 2141, 2142, 0 },
   { 2145, 2142, 0 },
   { 2151, 2152, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2173, 2174, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2989, 2994, 0 },
   { 2995, 2996, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4447, 1541, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1813, 698, 0 },
   { 1922, 704, 0 },
   { 2016, 710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2055, 2056, 0 },
   { 2058, 2056, 0 },
   { 2062, 2056, 0 },
   { 2066, 2056, 0 },
   { 2071, 2072, 0 },
   { 2074, 2072, 0 },
   { 2076, 2072, 0 },
   { 2078, 2072, 0 },
   { 2083, 2084, 0 },
   { 2086, 2084, 0 },
   { 2088, 2084, 0 },
   { 2090, 2084, 0 },
   { 2097, 2098, 0 },
   { 2102, 2098, 0 },
   { 2106, 2098, 0 },
   { 2111, 2112, 0 },
   { 2114, 2112, 0 },
   { 2116, 2112, 0 },
   { 2121, 2122, 0 },
   { 2124, 2122, 0 },
   { 2126, 2122, 0 },
   { 2133, 2134, 0 },
   { 2138, 2134, 0 },
   { 2143, 2144, 0 },
   { 2146, 2144, 0 },
   { 2153, 2154, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2175, 2176, 0 },
   { 3330, 2158, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1356, 32, 0 },
   { 1357, 63, 0 },
   { 1361, 1304, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 3288, 2847, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3316, 1405, 0 },
   { 3316, 1405, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6396, 6398, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6380, 6382, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6364, 6366, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2170, 2171, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2167, 2168, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3305, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3306, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3319, 1417, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3311, 3312, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3320, 1417, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3313, 3314, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 789, 791, 37 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 789, 791, 37 },
   { 796, 798, 0 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 789, 791, 37 },
   { 793, 274, 0 },
   { 796, 798, 0 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 784, 117, 0 },
   { 789, 791, 37 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 784, 117, 0 },
   { 789, 791, 37 },
   { 796, 798, 0 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 784, 117, 0 },
   { 789, 791, 37 },
   { 793, 274, 0 },
   { 796, 798, 0 },
   { 801, 802, 0 },
   { 817, 818, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 806, 798, 0 },
   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 804, 274, 0 },
   { 806, 798, 0 },
   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 785, 117, 0 },
   { 806, 798, 0 },
   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 785, 117, 0 },
   { 804, 274, 0 },
   { 806, 798, 0 },
   { 815, 816, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 808, 274, 0 },
   { 813, 814, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 786, 117, 0 },
   { 808, 274, 0 },
   { 809, 812, 0 },
   { 813, 814, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3861, 3863, 0 },
   { 4097, 4101, 90 },
   { 4097, 4105, 91 },
   { 4097, 4109, 92 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3491, 3492, 0 },
   { 4144, 4147, 97 },
   { 5389, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3493, 3492, 0 },
   { 4144, 4147, 97 },
   { 5389, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3540, 3536, 0 },
   { 4144, 4147, 97 },
   { 5389, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3545, 3541, 0 },
   { 4144, 4147, 97 },
   { 5389, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2568, 2574, 48 },
   { 2580, 2564, 0 },
   { 2588, 2566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2568, 2574, 48 },
   { 2581, 2566, 0 },
   { 2589, 2564, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2568, 2574, 48 },
   { 2577, 2568, 0 },
   { 2586, 2570, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2568, 2574, 48 },
   { 2579, 2570, 0 },
   { 2587, 2568, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2570, 2575, 48 },
   { 2584, 2564, 0 },
   { 2592, 2566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2570, 2575, 48 },
   { 2585, 2566, 0 },
   { 2593, 2564, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2570, 2575, 48 },
   { 2582, 2568, 0 },
   { 2590, 2570, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2570, 2575, 48 },
   { 2583, 2570, 0 },
   { 2591, 2568, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { 3025, 32, 0 },
   { 3026, 261, 0 },
   { 3027, 3024, 0 },
   { 3032, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3399, 3400, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3401, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3408, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3410, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3403, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3412, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3857, 3822, 68 },
   { 4120, 4123, 95 },
   { 4125, 4129, 95 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3858, 3859, 70 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 222, 3, 0 },
   { 3728, 3729, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 222, 3, 0 },
   { 3728, 3729, 0 },
   { 3770, 3729, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 217, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 220, 3, 0 },
   { 222, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 220, 3, 0 },
   { 222, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 340, 341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 340, 341, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 343, 344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 344, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 352, 353, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 352, 353, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 355, 356, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 355, 356, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 346, 347, 0 },
   { 350, 347, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 347, 0 },
   { 350, 347, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5296, 4277, 0 },
   { 5350, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5321, 4277, 0 },
   { 5350, 4290, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5296, 4277, 0 },
   { 5319, 4277, 0 },
   { 5350, 4290, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5296, 4277, 0 },
   { 5349, 4290, 0 },
   { 5350, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5296, 4277, 0 },
   { 5347, 4290, 0 },
   { 5350, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5322, 4277, 0 },
   { 5375, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5321, 4277, 0 },
   { 5322, 4277, 0 },
   { 5374, 4290, 0 },
   { 5375, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5322, 4277, 0 },
   { 5372, 4290, 0 },
   { 5375, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5322, 4277, 0 },
   { 5349, 4290, 0 },
   { 5375, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5322, 4277, 0 },
   { 5347, 4290, 0 },
   { 5375, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { 5427, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5402, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5402, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { 5402, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5321, 4277, 0 },
   { 5374, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5319, 4277, 0 },
   { 5372, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5295, 4277, 0 },
   { 5349, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 3, 0 },
   { 5293, 4277, 0 },
   { 5347, 4290, 0 },
   { 5399, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5319, 4277, 0 },
   { 5321, 4277, 0 },
   { 5372, 4290, 0 },
   { 5374, 4290, 0 },
   { 5424, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5321, 4277, 0 },
   { 5349, 4290, 0 },
   { 5374, 4290, 0 },
   { 5401, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5321, 4277, 0 },
   { 5347, 4290, 0 },
   { 5374, 4290, 0 },
   { 5399, 4290, 0 },
   { 5426, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5295, 4277, 0 },
   { 5319, 4277, 0 },
   { 5349, 4290, 0 },
   { 5372, 4290, 0 },
   { 5401, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5319, 4277, 0 },
   { 5347, 4290, 0 },
   { 5372, 4290, 0 },
   { 5399, 4290, 0 },
   { 5424, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5293, 4277, 0 },
   { 5295, 4277, 0 },
   { 5347, 4290, 0 },
   { 5349, 4290, 0 },
   { 5399, 4290, 0 },
   { 5401, 4290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3533, 3534, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3231, 1130, 0 },
   { 3232, 694, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 401, 402, 10 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1320, 1321, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1291, 1292, 0 },
   { 1293, 1294, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1295, 1296, 0 },
   { 1298, 1299, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1311, 1312, 0 },
   { 1313, 1314, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 2560, 2561, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4313, 4314, 117 },
   { 4326, 4327, 117 },
   { 4338, 4339, 117 },
   { 4350, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4316, 4314, 117 },
   { 4329, 4327, 117 },
   { 4341, 4339, 117 },
   { 4352, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5315, 5287, 0 },
   { 5368, 5340, 0 },
   { 5420, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5313, 5287, 0 },
   { 5366, 5340, 0 },
   { 5418, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5290, 5287, 0 },
   { 5343, 5340, 0 },
   { 5395, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1326, 1324, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3886, 3888, 0 },
   { 5286, 5287, 0 },
   { 5339, 5340, 0 },
   { 5392, 5340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4276, 4268, 9 },
   { 4289, 4281, 9 },
   { 4302, 4294, 9 },
   { 4309, 3945, 116 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4322, 4314, 117 },
   { 4335, 4327, 117 },
   { 4347, 4339, 117 },
   { 4358, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4274, 4268, 9 },
   { 4287, 4281, 9 },
   { 4300, 4294, 9 },
   { 4307, 3945, 116 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4320, 4314, 117 },
   { 4333, 4327, 117 },
   { 4345, 4339, 117 },
   { 4356, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4267, 4268, 9 },
   { 4280, 4281, 9 },
   { 4293, 4294, 9 },
   { 3949, 3945, 116 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1301, 1302, 0 },
   { 1303, 1304, 0 },
   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4267, 4268, 9 },
   { 4280, 4281, 9 },
   { 4293, 4294, 9 },
   { 3949, 3945, 116 },
   { 4313, 4314, 117 },
   { 4326, 4327, 117 },
   { 4338, 4339, 117 },
   { 4350, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4270, 4268, 9 },
   { 4283, 4281, 9 },
   { 4296, 4294, 9 },
   { 4304, 3945, 116 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1305, 1306, 0 },
   { 1308, 1309, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 4270, 4268, 9 },
   { 4283, 4281, 9 },
   { 4296, 4294, 9 },
   { 4304, 3945, 116 },
   { 4316, 4314, 117 },
   { 4329, 4327, 117 },
   { 4341, 4339, 117 },
   { 4352, 3955, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3791, 3789, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3794, 3792, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3797, 3795, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3800, 3798, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1315, 1316, 0 },
   { 1318, 1319, 0 },
   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3803, 3801, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5324, 5298, 0 },
   { 5377, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5323, 5298, 0 },
   { 5376, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5352, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5351, 49, 0 },
   { 5403, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 1045, 1049, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 1037, 1040, 0 },
   { 1045, 1049, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5324, 5298, 0 },
   { 5377, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5323, 5298, 0 },
   { 5376, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5299, 5298, 0 },
   { 5352, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 249, 61, 0 },
   { 257, 32, 0 },
   { 260, 109, 0 },
   { 5297, 5298, 0 },
   { 5351, 49, 0 },
   { 5403, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5324, 5298, 0 },
   { 5377, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5323, 5298, 0 },
   { 5376, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5299, 5298, 0 },
   { 5352, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3681, 3682, 0 },
   { 5297, 5298, 0 },
   { 5351, 49, 0 },
   { 5403, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5324, 5298, 0 },
   { 5353, 49, 0 },
   { 5377, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5300, 5298, 0 },
   { 5323, 5298, 0 },
   { 5353, 49, 0 },
   { 5376, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5300, 5298, 0 },
   { 5352, 49, 0 },
   { 5353, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5300, 5298, 0 },
   { 5351, 49, 0 },
   { 5353, 49, 0 },
   { 5403, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5324, 5298, 0 },
   { 5325, 5298, 0 },
   { 5377, 49, 0 },
   { 5378, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5323, 5298, 0 },
   { 5325, 5298, 0 },
   { 5376, 49, 0 },
   { 5378, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5325, 5298, 0 },
   { 5352, 49, 0 },
   { 5378, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5325, 5298, 0 },
   { 5351, 49, 0 },
   { 5378, 49, 0 },
   { 5403, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5324, 5298, 0 },
   { 5377, 49, 0 },
   { 5429, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5323, 5298, 0 },
   { 5376, 49, 0 },
   { 5428, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5352, 49, 0 },
   { 5404, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5351, 49, 0 },
   { 5403, 49, 0 },
   { 5430, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5324, 5298, 0 },
   { 5377, 49, 0 },
   { 5405, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5323, 5298, 0 },
   { 5376, 49, 0 },
   { 5405, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5352, 49, 0 },
   { 5404, 49, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5351, 49, 0 },
   { 5403, 49, 0 },
   { 5405, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5323, 5298, 0 },
   { 5324, 5298, 0 },
   { 5376, 49, 0 },
   { 5377, 49, 0 },
   { 5428, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5324, 5298, 0 },
   { 5352, 49, 0 },
   { 5377, 49, 0 },
   { 5404, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5324, 5298, 0 },
   { 5351, 49, 0 },
   { 5377, 49, 0 },
   { 5403, 49, 0 },
   { 5429, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5299, 5298, 0 },
   { 5323, 5298, 0 },
   { 5352, 49, 0 },
   { 5376, 49, 0 },
   { 5404, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5323, 5298, 0 },
   { 5351, 49, 0 },
   { 5376, 49, 0 },
   { 5403, 49, 0 },
   { 5428, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5297, 5298, 0 },
   { 5299, 5298, 0 },
   { 5351, 49, 0 },
   { 5352, 49, 0 },
   { 5403, 49, 0 },
   { 5404, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2932, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 923, 926, 0 },
   { 931, 934, 0 },
   { 939, 942, 0 },
   { 947, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 927, 926, 0 },
   { 935, 934, 0 },
   { 943, 942, 0 },
   { 951, 950, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4060, 4061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1051, 1054, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 829, 833, 0 },
   { 1033, 1035, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1033, 1035, 0 },
   { 1051, 1054, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { 3527, 3529, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2308, 2310, 0 },
   { 2338, 2340, 0 },
   { 2368, 2370, 0 },
   { 2399, 2401, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2938, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2308, 2310, 0 },
   { 2338, 2340, 0 },
   { 2368, 2370, 0 },
   { 2399, 2401, 0 },
   { 2938, 1242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 928, 930, 0 },
   { 936, 938, 0 },
   { 944, 946, 0 },
   { 952, 954, 0 },
   { 2308, 2310, 0 },
   { 2338, 2340, 0 },
   { 2368, 2370, 0 },
   { 2399, 2401, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { 3530, 3526, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { 3881, 3879, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { 3878, 3879, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 32, 0 },
   { 3878, 3879, 0 },
   { 3881, 3879, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1169, 1168, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1173, 1172, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1171, 1172, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1167, 1168, 0 },
   { 1204, 1205, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1118, 1119, 0 },
   { 1227, 1230, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1170, 1168, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1104, 1105, 0 },
   { 1177, 1175, 0 },
   { 2842, 1270, 0 },
   { 2866, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1154, 1151, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1161, 1159, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1158, 1159, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1150, 1151, 0 },
   { 1206, 1207, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1116, 1117, 0 },
   { 1233, 1234, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1156, 1151, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1106, 1107, 0 },
   { 1166, 1164, 0 },
   { 2841, 1266, 0 },
   { 2867, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { 5824, 5827, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 265, 263, 0 },
   { 3686, 3688, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 265, 263, 0 },
   { 3686, 3688, 0 },
   { 3731, 3688, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 262, 263, 0 },
   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3673, 3674, 0 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 388, 389, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 388, 389, 10 },
   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 529, 530, 0 },
   { 597, 598, 0 },
   { 669, 670, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 529, 530, 0 },
   { 597, 598, 0 },
   { 669, 670, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 529, 530, 0 },
   { 597, 598, 0 },
   { 669, 670, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 396, 397, 10 },
   { 529, 530, 0 },
   { 597, 598, 0 },
   { 669, 670, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 534, 535, 0 },
   { 601, 602, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 534, 535, 0 },
   { 601, 602, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 534, 535, 0 },
   { 601, 602, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 534, 535, 0 },
   { 601, 602, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 529, 530, 0 },
   { 534, 535, 0 },
   { 597, 598, 0 },
   { 601, 602, 0 },
   { 669, 670, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 540, 542, 0 },
   { 606, 608, 0 },
   { 679, 681, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 394, 395, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 394, 395, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 394, 395, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 394, 395, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 394, 395, 10 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 274, 0 },
   { 275, 276, 6 },
   { 277, 238, 7 },
   { 394, 395, 10 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 394, 395, 10 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 151, 0 },
   { 394, 395, 10 },
   { 396, 397, 10 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2874, 32, 0 },
   { 2875, 63, 0 },
   { 2910, 2912, 50 },
   { 2913, 2915, 50 },
   { 2916, 2917, 50 },
   { 2918, 2919, 50 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2947, 2944, 51 },
   { 2954, 2951, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 1248, 0 },
   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2959, 2957, 51 },
   { 2964, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2876, 63, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1731, 1734, 0 },
   { 1738, 1734, 0 },
   { 1743, 1746, 0 },
   { 1750, 1746, 0 },
   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5330, 2233, 0 },
   { 5359, 2237, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5306, 2233, 0 },
   { 5329, 2233, 0 },
   { 5359, 2237, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5306, 2233, 0 },
   { 5358, 2237, 0 },
   { 5359, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5306, 2233, 0 },
   { 5357, 2237, 0 },
   { 5359, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5331, 2233, 0 },
   { 5383, 2237, 0 },
   { 5384, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5331, 2233, 0 },
   { 5382, 2237, 0 },
   { 5384, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5331, 2233, 0 },
   { 5358, 2237, 0 },
   { 5384, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5331, 2233, 0 },
   { 5357, 2237, 0 },
   { 5384, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5435, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5434, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { 5436, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5330, 2233, 0 },
   { 5383, 2237, 0 },
   { 5411, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5382, 2237, 0 },
   { 5411, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { 5411, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2947, 2944, 51 },
   { 2954, 2951, 51 },
   { 2959, 2957, 51 },
   { 2964, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5358, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 3273, 3274, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5329, 2233, 0 },
   { 5330, 2233, 0 },
   { 5382, 2237, 0 },
   { 5383, 2237, 0 },
   { 5434, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5330, 2233, 0 },
   { 5358, 2237, 0 },
   { 5383, 2237, 0 },
   { 5410, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5330, 2233, 0 },
   { 5357, 2237, 0 },
   { 5383, 2237, 0 },
   { 5409, 2237, 0 },
   { 5435, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5305, 2233, 0 },
   { 5329, 2233, 0 },
   { 5358, 2237, 0 },
   { 5382, 2237, 0 },
   { 5410, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5329, 2233, 0 },
   { 5357, 2237, 0 },
   { 5382, 2237, 0 },
   { 5409, 2237, 0 },
   { 5434, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5305, 2233, 0 },
   { 5357, 2237, 0 },
   { 5358, 2237, 0 },
   { 5409, 2237, 0 },
   { 5410, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5304, 2233, 0 },
   { 5357, 2237, 0 },
   { 5409, 2237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 3535, 3534, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 208, 0 },
   { 1434, 1435, 0 },
   { 1439, 1438, 0 },
   { 1456, 1457, 0 },
   { 1461, 1460, 0 },
   { 1478, 1479, 0 },
   { 1483, 1482, 0 },
   { 1498, 1499, 0 },
   { 1503, 1502, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1120, 1121, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3275, 770, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4428, 4429, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4434, 4435, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4440, 4441, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4423, 4426, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4431, 4432, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4437, 4438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3282, 1541, 0 },
   { 3283, 3280, 0 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2180, 2183, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2872, 61, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2886, 2887, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2886, 2887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 197, 199, 0 },
   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2282, 2284, 44 },
   { 2282, 2290, 46 },
   { 2313, 2315, 44 },
   { 2313, 2321, 46 },
   { 2343, 2345, 44 },
   { 2343, 2351, 46 },
   { 2373, 2375, 44 },
   { 2373, 2381, 46 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2928, 2929, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5327, 2191, 0 },
   { 5356, 2199, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5303, 2191, 0 },
   { 5326, 2191, 0 },
   { 5356, 2199, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5303, 2191, 0 },
   { 5355, 2199, 0 },
   { 5356, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5303, 2191, 0 },
   { 5354, 2199, 0 },
   { 5356, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5303, 2191, 0 },
   { 5356, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5327, 2191, 0 },
   { 5328, 2191, 0 },
   { 5380, 2199, 0 },
   { 5381, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5328, 2191, 0 },
   { 5379, 2199, 0 },
   { 5381, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5328, 2191, 0 },
   { 5355, 2199, 0 },
   { 5381, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5328, 2191, 0 },
   { 5354, 2199, 0 },
   { 5381, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5328, 2191, 0 },
   { 5381, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5433, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5408, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5408, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5408, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 2885, 3, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 2885, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 3271, 3272, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2282, 2284, 44 },
   { 2282, 2290, 46 },
   { 2313, 2315, 44 },
   { 2313, 2321, 46 },
   { 2343, 2345, 44 },
   { 2343, 2351, 46 },
   { 2373, 2375, 44 },
   { 2373, 2381, 46 },
   { 2868, 32, 0 },
   { 3426, 3428, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5326, 2191, 0 },
   { 5327, 2191, 0 },
   { 5379, 2199, 0 },
   { 5380, 2199, 0 },
   { 5431, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5327, 2191, 0 },
   { 5355, 2199, 0 },
   { 5380, 2199, 0 },
   { 5407, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5327, 2191, 0 },
   { 5354, 2199, 0 },
   { 5380, 2199, 0 },
   { 5406, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5327, 2191, 0 },
   { 5380, 2199, 0 },
   { 5432, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5302, 2191, 0 },
   { 5326, 2191, 0 },
   { 5355, 2199, 0 },
   { 5379, 2199, 0 },
   { 5407, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5326, 2191, 0 },
   { 5354, 2199, 0 },
   { 5379, 2199, 0 },
   { 5406, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5326, 2191, 0 },
   { 5379, 2199, 0 },
   { 5431, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5302, 2191, 0 },
   { 5354, 2199, 0 },
   { 5355, 2199, 0 },
   { 5406, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5302, 2191, 0 },
   { 5355, 2199, 0 },
   { 5407, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 32, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1431, 1432, 0 },
   { 1453, 1454, 0 },
   { 1475, 1476, 0 },
   { 1495, 1496, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { 5301, 2191, 0 },
   { 5354, 2199, 0 },
   { 5406, 2199, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2563, 2562, 0 },
   { 3339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3406, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3420, 3418, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 227, 3, 0 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 227, 3, 0 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3744, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 3746, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 3755, 3759, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 3765, 3768, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 823, 825, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 227, 3, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 230, 231, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 227, 3, 0 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3005, 3001, 54 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3005, 3001, 54 },
   { 3016, 3017, 0 },
   { 3020, 261, 0 },
   { 3023, 3024, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1181, 1151, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 1290, 1062, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 1209, 1212, 0 },
   { 1290, 1062, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1109, 0 },
   { 1214, 1217, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1109, 0 },
   { 1268, 1266, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1109, 0 },
   { 2856, 770, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 717, 719, 23 },
   { 720, 722, 23 },
   { 723, 725, 24 },
   { 3332, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 2849, 1276, 0 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 1560, 1564, 39 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 1551, 1553, 39 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 1546, 1548, 39 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 1530, 41 },
   { 1556, 1558, 42 },
   { 3341, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1602, 1603, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1532, 228, 39 },
   { 1537, 1535, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1522, 32, 0 },
   { 1524, 99, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1524, 99, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1524, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { 1524, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { 1522, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1280, 0 },
   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { 1526, 99, 0 },
   { 1590, 1593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1359, 1294, 0 },
   { 1372, 1374, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1183, 1168, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1113, 0 },
   { 1220, 1221, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1113, 0 },
   { 1271, 1270, 0 },
   { 1271, 1270, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1113, 0 },
   { 1146, 1144, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3001, 3004, 53 },
   { 3025, 32, 0 },
   { 3026, 261, 0 },
   { 3027, 3024, 0 },
   { 3034, 99, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1520, 32, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1813, 698, 0 },
   { 1844, 1824, 0 },
   { 1922, 704, 0 },
   { 1950, 1933, 0 },
   { 2016, 710, 0 },
   { 2044, 2027, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2420, 2421, 0 },
   { 2425, 2440, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2420, 2421, 0 },
   { 4069, 4067, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2420, 2421, 0 },
   { 4072, 4067, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { 1516, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { 1521, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 3285, 3274, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1523, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1523, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 4227, 4223, 108 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1523, 32, 0 },
   { 1525, 114, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1525, 114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { 1525, 114, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1523, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1523, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { 1523, 32, 0 },
   { 1527, 114, 40 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1527, 114, 40 },
   { 3284, 3272, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1351, 32, 0 },
   { 1363, 1314, 0 },
   { 1382, 1384, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1122, 1123, 0 },
   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3276, 1144, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2419, 1404, 0 },
   { 2860, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 1090, 0 },
   { 3278, 770, 0 },
   { 3381, 1090, 0 },
   { 3382, 1248, 0 },
   { 3383, 1248, 0 },
   { 1502, 3994, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 1090, 0 },
   { 3279, 770, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2550, 2551, 0 },
   { 2861, 1090, 0 },
   { 3385, 1248, 0 },
   { 4001, 3998, 84 },
   { 4039, 1090, 0 },
   { 4043, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2507, 1090, 0 },
   { 2861, 1090, 0 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2488, 1098, 0 },
   { 2498, 1090, 0 },
   { 2861, 1090, 0 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2498, 1090, 0 },
   { 2499, 1090, 0 },
   { 2861, 1090, 0 },
   { 4039, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2858, 1248, 0 },
   { 3277, 1144, 0 },
   { 3378, 1248, 0 },
   { 3379, 1090, 0 },
   { 3380, 1090, 0 },
   { 1499, 3995, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 3287, 3272, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { 1377, 1379, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1356, 32, 0 },
   { 1357, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1356, 32, 0 },
   { 1357, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1356, 32, 0 },
   { 1357, 63, 0 },
   { 1361, 1304, 0 },
   { 1377, 1379, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 3286, 3274, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { 1377, 1379, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1355, 32, 0 },
   { 1361, 1304, 0 },
   { 1377, 1379, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2983, 1919, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1607, 1606, 0 },
   { 4135, 4138, 96 },
   { 5283, 11, 0 },
   { 5336, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5333, 2257, 0 },
   { 5386, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5332, 2257, 0 },
   { 5385, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5361, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5333, 2257, 0 },
   { 5386, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5332, 2257, 0 },
   { 5385, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5308, 2257, 0 },
   { 5361, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2878, 32, 0 },
   { 2881, 2871, 0 },
   { 2920, 2922, 50 },
   { 2923, 2925, 50 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5333, 2257, 0 },
   { 5362, 2261, 0 },
   { 5386, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5309, 2257, 0 },
   { 5332, 2257, 0 },
   { 5362, 2261, 0 },
   { 5385, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5309, 2257, 0 },
   { 5361, 2261, 0 },
   { 5362, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5309, 2257, 0 },
   { 5360, 2261, 0 },
   { 5362, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5333, 2257, 0 },
   { 5334, 2257, 0 },
   { 5386, 2261, 0 },
   { 5387, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5332, 2257, 0 },
   { 5334, 2257, 0 },
   { 5385, 2261, 0 },
   { 5387, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5334, 2257, 0 },
   { 5361, 2261, 0 },
   { 5387, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5334, 2257, 0 },
   { 5360, 2261, 0 },
   { 5387, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5333, 2257, 0 },
   { 5386, 2261, 0 },
   { 5438, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5332, 2257, 0 },
   { 5385, 2261, 0 },
   { 5437, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5361, 2261, 0 },
   { 5413, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { 5439, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5333, 2257, 0 },
   { 5386, 2261, 0 },
   { 5414, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5332, 2257, 0 },
   { 5385, 2261, 0 },
   { 5414, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5361, 2261, 0 },
   { 5413, 2261, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { 5414, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5333, 2257, 0 },
   { 5386, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5332, 2257, 0 },
   { 5385, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5308, 2257, 0 },
   { 5361, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 2880, 3, 0 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5332, 2257, 0 },
   { 5333, 2257, 0 },
   { 5385, 2261, 0 },
   { 5386, 2261, 0 },
   { 5437, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5333, 2257, 0 },
   { 5361, 2261, 0 },
   { 5386, 2261, 0 },
   { 5413, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5333, 2257, 0 },
   { 5360, 2261, 0 },
   { 5386, 2261, 0 },
   { 5412, 2261, 0 },
   { 5438, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5308, 2257, 0 },
   { 5332, 2257, 0 },
   { 5361, 2261, 0 },
   { 5385, 2261, 0 },
   { 5413, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5332, 2257, 0 },
   { 5360, 2261, 0 },
   { 5385, 2261, 0 },
   { 5412, 2261, 0 },
   { 5437, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5308, 2257, 0 },
   { 5360, 2261, 0 },
   { 5361, 2261, 0 },
   { 5412, 2261, 0 },
   { 5413, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2877, 61, 0 },
   { 5307, 2257, 0 },
   { 5360, 2261, 0 },
   { 5412, 2261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2048, 770, 0 },
   { 2052, 770, 0 },
   { 2068, 770, 0 },
   { 2080, 770, 0 },
   { 2094, 6, 0 },
   { 2108, 6, 0 },
   { 2118, 6, 0 },
   { 2130, 11, 0 },
   { 2140, 11, 0 },
   { 2150, 16, 0 },
   { 2186, 2183, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 4473, 4471, 0 },
   { 4479, 4471, 0 },
   { 4485, 4471, 0 },
   { 4490, 4471, 0 },
   { 4492, 2047, 0 },
   { 4496, 4469, 0 },
   { 4500, 4469, 0 },
   { 4504, 4469, 0 },
   { 4506, 2051, 0 },
   { 4508, 2051, 0 },
   { 4510, 2093, 0 },
   { 4514, 4476, 0 },
   { 4518, 4476, 0 },
   { 4520, 2059, 0 },
   { 4522, 2059, 0 },
   { 4524, 2099, 0 },
   { 4526, 2059, 0 },
   { 4528, 2099, 0 },
   { 4530, 2129, 0 },
   { 4534, 4482, 0 },
   { 4536, 2063, 0 },
   { 4538, 2063, 0 },
   { 4540, 2103, 0 },
   { 4542, 2063, 0 },
   { 4544, 2103, 0 },
   { 4546, 2135, 0 },
   { 4548, 2063, 0 },
   { 4550, 2103, 0 },
   { 4552, 2135, 0 },
   { 4554, 2149, 0 },
   { 5282, 11, 0 },
   { 5335, 16, 0 },
   { 5388, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2048, 770, 0 },
   { 2052, 770, 0 },
   { 2068, 770, 0 },
   { 2080, 770, 0 },
   { 2094, 6, 0 },
   { 2108, 6, 0 },
   { 2118, 6, 0 },
   { 2130, 11, 0 },
   { 2140, 11, 0 },
   { 2150, 16, 0 },
   { 3455, 3456, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { 4473, 4471, 0 },
   { 4479, 4471, 0 },
   { 4485, 4471, 0 },
   { 4490, 4471, 0 },
   { 4492, 2047, 0 },
   { 4496, 4469, 0 },
   { 4500, 4469, 0 },
   { 4504, 4469, 0 },
   { 4506, 2051, 0 },
   { 4508, 2051, 0 },
   { 4510, 2093, 0 },
   { 4514, 4476, 0 },
   { 4518, 4476, 0 },
   { 4520, 2059, 0 },
   { 4522, 2059, 0 },
   { 4524, 2099, 0 },
   { 4526, 2059, 0 },
   { 4528, 2099, 0 },
   { 4530, 2129, 0 },
   { 4534, 4482, 0 },
   { 4536, 2063, 0 },
   { 4538, 2063, 0 },
   { 4540, 2103, 0 },
   { 4542, 2063, 0 },
   { 4544, 2103, 0 },
   { 4546, 2135, 0 },
   { 4548, 2063, 0 },
   { 4550, 2103, 0 },
   { 4552, 2135, 0 },
   { 4554, 2149, 0 },
   { 5282, 11, 0 },
   { 5335, 16, 0 },
   { 5388, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3452, 3454, 0 },
   { 4449, 770, 0 },
   { 4450, 6, 0 },
   { 4451, 11, 0 },
   { 4452, 16, 0 },
   { 4453, 21, 0 },
   { 4470, 4471, 0 },
   { 4477, 4471, 0 },
   { 4483, 4471, 0 },
   { 4488, 4471, 0 },
   { 4494, 4469, 0 },
   { 4498, 4469, 0 },
   { 4502, 4469, 0 },
   { 4512, 4476, 0 },
   { 4516, 4476, 0 },
   { 4532, 4482, 0 },
   { 5310, 11, 0 },
   { 5363, 16, 0 },
   { 5415, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2298, 6, 0 },
   { 2329, 11, 0 },
   { 2359, 16, 0 },
   { 2389, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3833, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3853, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3835, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3855, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3831, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3834, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3852, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3854, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3816, 3798, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3827, 3798, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3821, 3822, 68 },
   { 3831, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3828, 3822, 68 },
   { 3834, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3839, 3792, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3848, 3792, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1346, 1159, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3843, 3817, 69 },
   { 3852, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3849, 3817, 69 },
   { 3854, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3824, 3798, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3829, 3798, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3826, 3822, 68 },
   { 3833, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3830, 3822, 68 },
   { 3835, 3822, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3845, 3792, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3850, 3792, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3847, 3817, 69 },
   { 3853, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 3851, 3817, 69 },
   { 3855, 3817, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1353, 61, 0 },
   { 1354, 32, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1367, 1306, 0 },
   { 1392, 1394, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1352, 32, 0 },
   { 1578, 1579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2943, 2944, 51 },
   { 2950, 2951, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2956, 2957, 51 },
   { 2961, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2943, 2944, 51 },
   { 2950, 2951, 51 },
   { 2959, 2957, 51 },
   { 2964, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2947, 2944, 51 },
   { 2954, 2951, 51 },
   { 2956, 2957, 51 },
   { 2961, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2943, 2944, 51 },
   { 2950, 2951, 51 },
   { 2956, 2957, 51 },
   { 2961, 2962, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 958, 959, 0 },
   { 1005, 1004, 0 },
   { 1012, 1011, 0 },
   { 1020, 1018, 0 },
   { 1029, 1027, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1006, 0 },
   { 1014, 1013, 0 },
   { 1022, 1021, 0 },
   { 1031, 1030, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 227, 3, 0 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 227, 3, 0 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3713, 3715, 0 },
   { 3717, 3715, 0 },
   { 3744, 3715, 0 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 227, 3, 0 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 161, 163, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 161, 163, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 225, 3, 0 },
   { 227, 3, 0 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 738, 717, 25 },
   { 739, 720, 25 },
   { 740, 723, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 117, 0 },
   { 1823, 1824, 43 },
   { 1932, 1933, 43 },
   { 2026, 2027, 43 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 511, 501, 16 },
   { 581, 572, 18 },
   { 652, 642, 20 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 3, 0 },
   { 511, 501, 16 },
   { 521, 519, 15 },
   { 581, 572, 18 },
   { 590, 588, 17 },
   { 652, 642, 20 },
   { 661, 659, 19 },
   { 728, 729, 25 },
   { 732, 733, 25 },
   { 736, 737, 26 },
   { 3786, 3715, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2232, 2234, 0 },
   { 2236, 2238, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2190, 2192, 0 },
   { 2198, 2200, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2258, 0 },
   { 2260, 2262, 0 },
   { 4454, 770, 0 },
   { 4455, 6, 0 },
   { 4456, 11, 0 },
   { 4457, 16, 0 },
   { 4458, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1003, 1004, 0 },
   { 1010, 1011, 0 },
   { 1017, 1018, 0 },
   { 1026, 1027, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2426, 2439, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 208, 0 },
   { 1434, 1435, 0 },
   { 1439, 1438, 0 },
   { 1456, 1457, 0 },
   { 1461, 1460, 0 },
   { 1478, 1479, 0 },
   { 1483, 1482, 0 },
   { 1498, 1499, 0 },
   { 1503, 1502, 0 },
   { 2859, 1248, 0 },
   { 3225, 770, 0 },
   { 3228, 1144, 0 },
   { 3377, 1090, 0 },
   { 3999, 4000, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3417, 3418, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 103, 106, 5 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 279, 10 },
   { 3865, 3866, 0 },
   { 3868, 3869, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 265, 263, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3673, 3674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 524, 519, 0 },
   { 593, 588, 0 },
   { 664, 659, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 267, 3, 0 },
   { 269, 268, 6 },
   { 270, 36, 7 },
   { 281, 282, 0 },
   { 286, 287, 0 },
   { 524, 519, 0 },
   { 593, 588, 0 },
   { 664, 659, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 166, 168, 0 },
   { 524, 519, 0 },
   { 593, 588, 0 },
   { 664, 659, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 524, 519, 0 },
   { 593, 588, 0 },
   { 664, 659, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 524, 519, 0 },
   { 534, 535, 0 },
   { 593, 588, 0 },
   { 601, 602, 0 },
   { 664, 659, 0 },
   { 673, 674, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 154, 0 },
   { 396, 397, 10 },
   { 524, 519, 0 },
   { 593, 588, 0 },
   { 664, 659, 0 },
   { 747, 749, 30 },
   { 245, 751, 31 },
   { 752, 754, 32 },
   { 755, 749, 33 },
   { 756, 751, 34 },
   { 757, 754, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1174, 1175, 0 },
   { 1204, 1205, 0 },
   { 2866, 1248, 0 },
   { 3358, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1163, 1164, 0 },
   { 1206, 1207, 0 },
   { 2867, 1090, 0 },
   { 3357, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3011, 3014, 54 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1109, 0 },
   { 1180, 1151, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1113, 0 },
   { 1182, 1168, 0 },
   { 2865, 1248, 0 },
   { 3361, 1248, 0 },
   { 3364, 1090, 0 },
   { 3367, 1090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1568, 1569, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1530, 228, 39 },
   { 1534, 1535, 39 },
   { 1568, 1569, 0 },
   { 3342, 228, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1520, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1520, 32, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1518, 1267, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1568, 1569, 0 },
   { 1596, 1598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1515, 1267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1365, 1296, 0 },
   { 1387, 1389, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1348, 32, 0 },
   { 1515, 1267, 0 },
   { 1568, 1569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1573, 1574, 0 },
   { 4225, 4223, 108 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1521, 32, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1516, 1517, 0 },
   { 1519, 1517, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1350, 32, 0 },
   { 1369, 1316, 0 },
   { 1397, 1399, 0 },
   { 1573, 1574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3293, 3295, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2860, 1090, 0 },
   { 3226, 1144, 0 },
   { 3227, 770, 0 },
   { 3297, 3298, 0 },
   { 3376, 1248, 0 },
   { 3996, 3998, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 61, 0 },
   { 1830, 1831, 43 },
   { 1939, 1940, 43 },
   { 2033, 2034, 43 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4049, 4050, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4030, 4021, 0 },
   { 4049, 4050, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4051, 4052, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4032, 4018, 0 },
   { 4051, 4052, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4055, 4056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4055, 4056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4034, 4035, 0 },
   { 4055, 4056, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 2884, 2506, 0 },
   { 3487, 3489, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 202, 204, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5700, 5686, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5692, 5686, 0 },
   { 5700, 5686, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 1209, 1212, 0 },
   { 2863, 1090, 0 },
   { 3346, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1201, 0 },
   { 1209, 1212, 0 },
   { 1290, 1062, 0 },
   { 2863, 1090, 0 },
   { 3350, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1196, 1199, 0 },
   { 2469, 1091, 0 },
   { 2863, 1090, 0 },
   { 3369, 1090, 0 },
   { 3372, 1248, 0 },
   { 3375, 1248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 3353, 3345, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1349, 32, 0 },
   { 1602, 1603, 39 },
   { 3355, 3349, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 117, 0 },
   { 233, 119, 6 },
   { 234, 121, 7 },
   { 253, 32, 0 },
   { 256, 32, 0 },
   { 259, 107, 0 },
   { 769, 777, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 120, 11, 6 },
   { 122, 16, 7 },
   { 216, 117, 0 },
   { 548, 549, 16 },
   { 614, 615, 18 },
   { 687, 688, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 506, 507, 16 },
   { 577, 578, 18 },
   { 647, 648, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 506, 507, 16 },
   { 511, 501, 16 },
   { 577, 578, 18 },
   { 581, 572, 18 },
   { 647, 648, 20 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 506, 507, 16 },
   { 511, 501, 16 },
   { 571, 572, 18 },
   { 577, 578, 18 },
   { 581, 572, 18 },
   { 641, 642, 20 },
   { 647, 648, 20 },
   { 652, 642, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 518, 519, 15 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 587, 588, 17 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { 658, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 518, 519, 15 },
   { 521, 519, 15 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 587, 588, 17 },
   { 590, 588, 17 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { 658, 659, 19 },
   { 661, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 506, 507, 16 },
   { 511, 501, 16 },
   { 518, 519, 15 },
   { 571, 572, 18 },
   { 577, 578, 18 },
   { 581, 572, 18 },
   { 587, 588, 17 },
   { 641, 642, 20 },
   { 647, 648, 20 },
   { 652, 642, 20 },
   { 658, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 506, 507, 16 },
   { 577, 578, 18 },
   { 647, 648, 20 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { 500, 501, 16 },
   { 511, 501, 16 },
   { 518, 519, 15 },
   { 571, 572, 18 },
   { 581, 572, 18 },
   { 587, 588, 17 },
   { 641, 642, 20 },
   { 652, 642, 20 },
   { 658, 659, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2452, 2455, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2460, 2464, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2452, 2455, 0 },
   { 2460, 2464, 0 },
   { 2868, 32, 0 },
   { 2872, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1090, 0 },
   { 2868, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1344, 1151, 0 },
   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 1799, 1806, 0 },
   { 1908, 1915, 0 },
   { 2002, 2009, 0 },
   { 3262, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1345, 1151, 0 },
   { 1347, 1159, 0 },
   { 3262, 3, 0 },
   { 5692, 5686, 0 },
   { 5700, 5686, 0 },
   { 5707, 5702, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 958, 959, 0 },
   { 1005, 1004, 0 },
   { 1012, 1011, 0 },
   { 1020, 1018, 0 },
   { 1029, 1027, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1006, 0 },
   { 1014, 1013, 0 },
   { 1022, 1021, 0 },
   { 1031, 1030, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1003, 1004, 0 },
   { 1010, 1011, 0 },
   { 1017, 1018, 0 },
   { 1026, 1027, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 202, 204, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 958, 959, 0 },
   { 1005, 1004, 0 },
   { 1012, 1011, 0 },
   { 1020, 1018, 0 },
   { 1029, 1027, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1006, 0 },
   { 1014, 1013, 0 },
   { 1022, 1021, 0 },
   { 1031, 1030, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1003, 1004, 0 },
   { 1010, 1011, 0 },
   { 1017, 1018, 0 },
   { 1026, 1027, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 202, 204, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2885, 3, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 958, 959, 0 },
   { 1005, 1004, 0 },
   { 1012, 1011, 0 },
   { 1020, 1018, 0 },
   { 1029, 1027, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1006, 0 },
   { 1014, 1013, 0 },
   { 1022, 1021, 0 },
   { 1031, 1030, 0 },
   { 2851, 2852, 0 },
   { 2854, 2855, 0 },
   { 2868, 32, 0 },
   { 2869, 61, 0 },
   { 2870, 32, 0 },
   { 2893, 2897, 50 },
   { 2899, 2903, 50 },
   { 2905, 2906, 50 },
   { 2908, 2909, 50 },
   { 3440, 3441, 60 },
   { 3471, 3472, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1003, 1004, 0 },
   { 1010, 1011, 0 },
   { 1017, 1018, 0 },
   { 1026, 1027, 0 },
   { 2933, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 32, 0 },
   { 3429, 3430, 60 },
   { 1030, 432, 60 },
   { 3432, 3434, 60 },
   { 1021, 457, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 202, 204, 0 },
   { 2873, 32, 0 },
   { 2883, 3, 0 },
   { 5682, 5607, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 32, 0 },
   { 5606, 5607, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_pass_op_table[nir_num_search_ops] = {
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         13,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         5,
         13,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         12,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         15,
         4,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         2,
         3,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         3,
         3,
         371,
         372,
         373,
         374,
         375,
         3,
         376,
         377,
         378,
         379,
         2299,
         2300,
         2301,
         2302,
         2303,
         2304,
         2,
         371,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         2,
         372,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         366,
         373,
         366,
         366,
         366,
         366,
         366,
         366,
         380,
         381,
         382,
         383,
         366,
         366,
         2305,
         2306,
         2307,
         2308,
         2,
         374,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         2,
         375,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         2,
         3,
         2,
         2,
         366,
         2,
         2,
         384,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         367,
         376,
         367,
         367,
         380,
         367,
         367,
         367,
         367,
         385,
         386,
         387,
         367,
         367,
         2309,
         2310,
         2311,
         2312,
         368,
         377,
         368,
         368,
         381,
         368,
         368,
         368,
         385,
         368,
         388,
         389,
         368,
         368,
         2313,
         2314,
         2315,
         2316,
         369,
         378,
         369,
         369,
         382,
         369,
         369,
         369,
         386,
         388,
         369,
         390,
         369,
         369,
         2317,
         2318,
         2319,
         2320,
         370,
         379,
         370,
         370,
         383,
         370,
         370,
         370,
         387,
         389,
         390,
         370,
         370,
         370,
         2321,
         2322,
         2323,
         2324,
         2,
         2299,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         2,
         2300,
         2,
         2,
         366,
         2,
         2,
         2,
         367,
         368,
         369,
         370,
         2,
         2,
         2295,
         2296,
         2297,
         2298,
         2295,
         2301,
         2295,
         2295,
         2305,
         2295,
         2295,
         2295,
         2309,
         2313,
         2317,
         2321,
         2295,
         2295,
         2295,
         2325,
         2326,
         2327,
         2296,
         2302,
         2296,
         2296,
         2306,
         2296,
         2296,
         2296,
         2310,
         2314,
         2318,
         2322,
         2296,
         2296,
         2325,
         2296,
         2328,
         2329,
         2297,
         2303,
         2297,
         2297,
         2307,
         2297,
         2297,
         2297,
         2311,
         2315,
         2319,
         2323,
         2297,
         2297,
         2326,
         2328,
         2297,
         2330,
         2298,
         2304,
         2298,
         2298,
         2308,
         2298,
         2298,
         2298,
         2312,
         2316,
         2320,
         2324,
         2298,
         2298,
         2327,
         2329,
         2330,
         2298,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         11,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         4,
         15,
         4,
         16,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         4,
         16,
         4,
         15,
         15,
         15,
         15,
         15,
         4,
         4,
         4,
         4,
         0,
         13,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         19,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         3,
         3,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         21,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         22,
         0,
         0,
         5,
         5,
         3,
         6,
         23,
         17,
         3,
         6,
         24,
         5,
         5,
         3,
         6,
         25,
         17,
         3,
         6,
         26,
         5,
         5,
         3,
         6,
         0,
         17,
      },
      
      .num_filtered_states = 27,
      .table = (const uint16_t []) {
      
         4,
         5,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         6,
         7,
         6,
         6,
         6,
         1065,
         6,
         6,
         6,
         1066,
         6,
         6,
         6,
         6,
         6,
         2332,
         6,
         6,
         6,
         1065,
         6,
         6,
         1065,
         6,
         1065,
         6,
         1065,
         4,
         1067,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1068,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1069,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1070,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         1071,
         1072,
         1071,
         1071,
         1071,
         1073,
         1071,
         1071,
         1071,
         1074,
         1071,
         1071,
         1071,
         1071,
         1071,
         2333,
         1071,
         1071,
         1071,
         1073,
         1071,
         1071,
         1073,
         1071,
         1073,
         1071,
         1073,
         1075,
         1076,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1078,
         1075,
         1075,
         1075,
         1075,
         1075,
         2334,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1077,
         1075,
         1077,
         1075,
         1077,
         4,
         1079,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         5,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1080,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1081,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         1082,
         1083,
         1082,
         1082,
         1082,
         1084,
         1082,
         1082,
         1082,
         1085,
         1082,
         1082,
         1082,
         1082,
         1082,
         2335,
         1082,
         1082,
         1082,
         1084,
         1082,
         1082,
         1084,
         1082,
         1084,
         1082,
         1084,
         4,
         2336,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2337,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         1069,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2338,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2339,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2340,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2883,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2884,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2986,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2993,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         2997,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         3002,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         3006,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
         4,
         3011,
         4,
         4,
         4,
         1063,
         4,
         4,
         4,
         1064,
         4,
         4,
         4,
         4,
         4,
         2331,
         4,
         4,
         4,
         1063,
         4,
         4,
         1063,
         4,
         1063,
         4,
         1063,
      },
   },
   [nir_op_unpack_64_2x32_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         8,
         1321,
         1322,
         1323,
         1324,
         1325,
         1326,
         1327,
      },
   },
   [nir_op_imul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_umul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_udiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         11,
         0,
         11,
      },
   },
   [nir_op_idiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         12,
         0,
         0,
         12,
         0,
         391,
         392,
         391,
      },
   },
   [nir_op_umod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         13,
         0,
         13,
      },
   },
   [nir_op_imod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         14,
         0,
         14,
      },
   },
   [nir_op_irem] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         15,
         0,
         15,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         9,
         10,
         0,
         0,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         16,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         18,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         17,
         17,
         17,
         17,
         17,
         17,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         19,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         21,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         17,
         17,
         17,
         17,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         0,
         0,
         20,
         20,
         20,
         20,
         22,
         20,
         19,
         19,
         0,
         19,
         0,
         19,
         11,
         11,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         24,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         25,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         19,
         19,
         19,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         17,
         17,
         17,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 26,
      .table = (const uint16_t []) {
      
         16,
         17,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         17,
         17,
         1334,
         1335,
         1336,
         1337,
         1338,
         1339,
         1340,
         1341,
         1342,
         17,
         1343,
         1344,
         17,
         1341,
         2370,
         2371,
         2372,
         2373,
         2374,
         1337,
         2890,
         17,
         2891,
         2971,
         16,
         1334,
         16,
         16,
         1345,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         16,
         1335,
         16,
         16,
         1345,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         1328,
         1336,
         1345,
         1345,
         1328,
         1346,
         1328,
         1328,
         1347,
         1328,
         1348,
         1328,
         1349,
         1350,
         1328,
         1328,
         2375,
         2376,
         2376,
         2377,
         2378,
         1346,
         2892,
         2893,
         2894,
         1328,
         1329,
         1337,
         1329,
         1329,
         1346,
         1329,
         1329,
         1329,
         1351,
         1329,
         1352,
         1329,
         1353,
         1354,
         1329,
         1329,
         2379,
         2380,
         2380,
         2381,
         2382,
         1329,
         2895,
         1329,
         2896,
         1329,
         16,
         1338,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         16,
         1339,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         1330,
         1340,
         1330,
         1330,
         1347,
         1351,
         1330,
         1330,
         1355,
         1330,
         1356,
         1330,
         1357,
         1358,
         1330,
         1330,
         2383,
         2384,
         2384,
         2385,
         2386,
         2387,
         2897,
         1330,
         2898,
         1330,
         16,
         1341,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         1331,
         1342,
         1331,
         1331,
         1348,
         1352,
         1331,
         1331,
         1356,
         1331,
         1331,
         1331,
         1359,
         1360,
         1331,
         1331,
         2366,
         2388,
         2388,
         2389,
         2390,
         1352,
         2899,
         1331,
         2900,
         1331,
         16,
         17,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         1361,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         1332,
         1343,
         1332,
         1332,
         1349,
         1353,
         1332,
         1332,
         1357,
         1332,
         1359,
         1332,
         1362,
         1363,
         1332,
         1332,
         2391,
         2392,
         2392,
         2393,
         2394,
         1353,
         2901,
         1332,
         2902,
         1332,
         1333,
         1344,
         1333,
         1333,
         1350,
         1354,
         1333,
         1333,
         1358,
         1333,
         1360,
         1333,
         1363,
         1333,
         1333,
         1333,
         2395,
         2396,
         2396,
         2397,
         2398,
         1354,
         2903,
         1333,
         2904,
         1333,
         16,
         17,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         1364,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         16,
         1341,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         2399,
         2400,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         2366,
         2370,
         2366,
         2366,
         2375,
         2379,
         2366,
         2366,
         2383,
         2366,
         2366,
         2366,
         2391,
         2395,
         2366,
         2400,
         2400,
         2401,
         2401,
         2402,
         2403,
         2379,
         2905,
         2366,
         2906,
         2366,
         2367,
         2371,
         2367,
         2367,
         2376,
         2380,
         2367,
         2367,
         2384,
         2367,
         2388,
         2367,
         2392,
         2396,
         2367,
         2367,
         2401,
         2367,
         2367,
         2404,
         2405,
         2380,
         2907,
         2367,
         2908,
         2367,
         2367,
         2372,
         2367,
         2367,
         2376,
         2380,
         2367,
         2367,
         2384,
         2367,
         2388,
         2367,
         2392,
         2396,
         2367,
         2367,
         2401,
         2367,
         2367,
         2404,
         2405,
         2380,
         2907,
         2367,
         2908,
         2367,
         2368,
         2373,
         2368,
         2368,
         2377,
         2381,
         2368,
         2368,
         2385,
         2368,
         2389,
         2368,
         2393,
         2397,
         2368,
         2368,
         2402,
         2404,
         2404,
         2368,
         2369,
         2381,
         2888,
         2368,
         2909,
         2368,
         2369,
         2374,
         2369,
         2369,
         2378,
         2382,
         2369,
         2369,
         2386,
         2369,
         2390,
         2369,
         2394,
         2398,
         2369,
         2369,
         2403,
         2405,
         2405,
         2369,
         2406,
         2382,
         2910,
         2369,
         2911,
         2369,
         1329,
         1337,
         1329,
         1329,
         1346,
         1329,
         1329,
         1329,
         2387,
         1329,
         1352,
         1329,
         1353,
         1354,
         1329,
         1329,
         2379,
         2380,
         2380,
         2381,
         2382,
         1329,
         2895,
         1329,
         2912,
         1329,
         2888,
         2890,
         2888,
         2888,
         2892,
         2895,
         2888,
         2888,
         2897,
         2888,
         2899,
         2888,
         2901,
         2903,
         2888,
         2888,
         2905,
         2907,
         2907,
         2888,
         2910,
         2895,
         2910,
         2888,
         2913,
         2888,
         16,
         17,
         16,
         16,
         2893,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
         2889,
         2891,
         2889,
         2889,
         2894,
         2896,
         2889,
         2889,
         2898,
         2889,
         2900,
         2889,
         2902,
         2904,
         2889,
         2889,
         2906,
         2908,
         2908,
         2909,
         2911,
         2912,
         2913,
         2889,
         2898,
         2889,
         16,
         2971,
         16,
         16,
         1328,
         1329,
         16,
         16,
         1330,
         16,
         1331,
         16,
         1332,
         1333,
         16,
         16,
         2366,
         2367,
         2367,
         2368,
         2369,
         1329,
         2888,
         16,
         2889,
         16,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         18,
      },
   },
   [nir_op_ffloor] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         19,
         2630,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         8,
         9,
         8,
         9,
         8,
         9,
         10,
         11,
         10,
         11,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         0,
         0,
         13,
         13,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         15,
         2,
         16,
         17,
         2,
         2,
         2,
         2,
         2,
         2,
         15,
         2,
         2,
         16,
         17,
         2,
         2,
         15,
         2,
         16,
         17,
         15,
         2,
         16,
         17,
         15,
         15,
         18,
         19,
         16,
         17,
         2,
         16,
         20,
         2,
         3,
         21,
         4,
         4,
         4,
         22,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         21,
         3,
         3,
         3,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         23,
         6,
         6,
         6,
         24,
         25,
         26,
         27,
         28,
         6,
         6,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         29,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         11,
         10,
         10,
         10,
         11,
         10,
         10,
         10,
         11,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         30,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         0,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         31,
         32,
         2,
         2,
         2,
         31,
         32,
         2,
         2,
         31,
         32,
         2,
         2,
         31,
         32,
         15,
         15,
         33,
         34,
         35,
         2,
         31,
         32,
         16,
         16,
         36,
         37,
         17,
         17,
         38,
         39,
         2,
         2,
         2,
         31,
         32,
         31,
         32,
         32,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         8,
         8,
         9,
         8,
         8,
         8,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         41,
         4,
         4,
         22,
         22,
         22,
         22,
         22,
         42,
         41,
         21,
         3,
         21,
         3,
         41,
         0,
         0,
         43,
         6,
         44,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         22,
         3,
         3,
         3,
         3,
         3,
         3,
         21,
         21,
         21,
         3,
         3,
         3,
         3,
         3,
         3,
         41,
         41,
         41,
         21,
         3,
         3,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         46,
         47,
         46,
         47,
         46,
         2,
         47,
         46,
         47,
         48,
         47,
         46,
         47,
         49,
         50,
         51,
         52,
         46,
         47,
         46,
         47,
         53,
         46,
         54,
         55,
         56,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         22,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 57,
      .table = (const uint16_t []) {
      
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         21,
         21,
         21,
         1367,
         1368,
         1369,
         1370,
         21,
         1371,
         1372,
         1373,
         1374,
         21,
         1375,
         21,
         21,
         21,
         21,
         21,
         21,
         21,
         2605,
         2606,
         2607,
         2608,
         2609,
         2610,
         2611,
         2612,
         21,
         21,
         2848,
         2848,
         2848,
         2848,
         2849,
         2850,
         2850,
         2848,
         2848,
         2851,
         2852,
         2853,
         2854,
         2611,
         2972,
         2848,
         21,
         2848,
         2850,
         21,
         2848,
         21,
         2848,
         2848,
         2849,
         2848,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         1376,
         1376,
         1376,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         1376,
         2856,
         2858,
         1376,
         2856,
         1376,
         2855,
         2856,
         1376,
         2856,
         1365,
         1367,
         1365,
         1365,
         1365,
         1365,
         1378,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         2603,
         2603,
         1378,
         1378,
         1378,
         1378,
         1378,
         2613,
         1365,
         1365,
         2859,
         2859,
         2859,
         2859,
         1365,
         2860,
         2860,
         2859,
         2859,
         1365,
         2847,
         2847,
         1378,
         1378,
         1378,
         2859,
         1365,
         2859,
         2860,
         1365,
         2859,
         1365,
         2859,
         2859,
         1365,
         2859,
         1365,
         1368,
         1365,
         1365,
         1365,
         1365,
         1378,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         1365,
         2603,
         2603,
         1378,
         1378,
         1378,
         1378,
         1378,
         2613,
         1365,
         1365,
         2859,
         2859,
         2859,
         2859,
         1365,
         2860,
         2860,
         2859,
         2859,
         1365,
         2847,
         2847,
         1378,
         1378,
         1378,
         2859,
         1365,
         2859,
         2860,
         1365,
         2859,
         1365,
         2859,
         2859,
         1365,
         2859,
         20,
         1369,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         1366,
         1370,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         1379,
         20,
         20,
         1380,
         1380,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         1371,
         1377,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         1377,
         1377,
         1377,
         1377,
         1377,
         1377,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2864,
         2864,
         2864,
         2864,
         1377,
         2865,
         2865,
         2864,
         2864,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2864,
         1377,
         2864,
         2865,
         1377,
         2864,
         1377,
         2864,
         2864,
         1377,
         2864,
         20,
         1372,
         1377,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         1377,
         1377,
         1377,
         1377,
         1377,
         1377,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2864,
         2864,
         2864,
         2864,
         1377,
         2865,
         2865,
         2864,
         2864,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2864,
         1377,
         2864,
         2865,
         1377,
         2864,
         1377,
         2864,
         2864,
         1377,
         2864,
         20,
         1373,
         20,
         1365,
         1365,
         20,
         1366,
         1380,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         1374,
         20,
         1365,
         1365,
         20,
         1366,
         1380,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         2615,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         1375,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         1381,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         2616,
         2604,
         20,
         1381,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         2616,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         1376,
         1376,
         1376,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2856,
         2858,
         2973,
         2856,
         2973,
         2974,
         2975,
         2973,
         2975,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         1376,
         1376,
         1376,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2976,
         1376,
         2976,
         2977,
         1376,
         2976,
         1376,
         2855,
         2856,
         1376,
         2976,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         2617,
         1376,
         2617,
         2617,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2866,
         2867,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         1376,
         2856,
         2858,
         1376,
         2867,
         2617,
         2855,
         2856,
         1376,
         2856,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         1376,
         1376,
         1376,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2976,
         2973,
         2976,
         2977,
         2973,
         2976,
         2973,
         2974,
         2975,
         2973,
         2978,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         2617,
         1376,
         2617,
         2617,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2866,
         2867,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2856,
         2858,
         2973,
         2867,
         2979,
         2974,
         2975,
         2973,
         2975,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         2617,
         1376,
         2617,
         2617,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2866,
         2867,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2976,
         1376,
         2976,
         2977,
         1376,
         2980,
         2617,
         2855,
         2856,
         1376,
         2976,
         2603,
         2605,
         2603,
         2603,
         2603,
         2603,
         2614,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2614,
         2614,
         2614,
         2614,
         2614,
         2618,
         2603,
         2603,
         2868,
         2868,
         2868,
         2868,
         2603,
         2869,
         2869,
         2868,
         2868,
         2603,
         2847,
         2847,
         2614,
         2614,
         2614,
         2868,
         2603,
         2868,
         2869,
         2603,
         2868,
         2603,
         2868,
         2868,
         2603,
         2868,
         2603,
         2606,
         2603,
         2603,
         2603,
         2603,
         2614,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2603,
         2614,
         2614,
         2614,
         2614,
         2614,
         2618,
         2603,
         2603,
         2868,
         2868,
         2868,
         2868,
         2603,
         2869,
         2869,
         2868,
         2868,
         2603,
         2847,
         2847,
         2614,
         2614,
         2614,
         2868,
         2603,
         2868,
         2869,
         2603,
         2868,
         2603,
         2868,
         2868,
         2603,
         2868,
         1366,
         2607,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2608,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2609,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2610,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2611,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2616,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         2616,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         2604,
         2612,
         2604,
         2613,
         2613,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2618,
         2618,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2604,
         2870,
         2870,
         2870,
         2870,
         2604,
         2871,
         2871,
         2870,
         2870,
         2604,
         2872,
         2872,
         2604,
         2604,
         2604,
         2870,
         2604,
         2870,
         2871,
         2604,
         2870,
         2604,
         2870,
         2870,
         2604,
         2870,
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         2615,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         2615,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         20,
         21,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         1381,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         2616,
         2604,
         20,
         1381,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         2873,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         2845,
         2848,
         2855,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2855,
         2856,
         2855,
         2856,
         2855,
         2857,
         2858,
         2855,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2855,
         2856,
         2858,
         2855,
         2856,
         2855,
         2855,
         2856,
         2855,
         2856,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2856,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2858,
         2858,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2856,
         2856,
         2858,
         2856,
         2856,
         2856,
         2856,
         2856,
         2856,
         2856,
         2845,
         2848,
         2855,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2855,
         2856,
         2855,
         2856,
         2855,
         2857,
         2858,
         2855,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2974,
         2856,
         2858,
         2974,
         2856,
         2974,
         2974,
         2975,
         2974,
         2975,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2856,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2858,
         2858,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2975,
         2856,
         2858,
         2975,
         2856,
         2975,
         2975,
         2975,
         2975,
         2975,
         20,
         2849,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         1376,
         1376,
         1376,
         1376,
         1376,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2855,
         2856,
         1376,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2856,
         2858,
         2973,
         2856,
         2973,
         2974,
         2975,
         2973,
         2975,
         2846,
         2850,
         2857,
         2860,
         2860,
         2846,
         2862,
         2846,
         2865,
         2865,
         2846,
         2846,
         2846,
         2846,
         2846,
         2857,
         2857,
         2857,
         2857,
         2857,
         2857,
         2869,
         2869,
         2862,
         2862,
         2862,
         2862,
         2862,
         2871,
         2846,
         2846,
         2857,
         2858,
         2857,
         2858,
         2857,
         2857,
         2858,
         2857,
         2858,
         2846,
         2875,
         2875,
         2862,
         2862,
         2862,
         2977,
         2857,
         2977,
         2977,
         2857,
         2977,
         2857,
         2857,
         2858,
         2857,
         2977,
         2846,
         2850,
         2858,
         2860,
         2860,
         2846,
         2862,
         2846,
         2865,
         2865,
         2846,
         2846,
         2846,
         2846,
         2846,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2869,
         2869,
         2862,
         2862,
         2862,
         2862,
         2862,
         2871,
         2846,
         2846,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2846,
         2875,
         2875,
         2862,
         2862,
         2862,
         2977,
         2858,
         2977,
         2977,
         2858,
         2977,
         2858,
         2858,
         2858,
         2858,
         2977,
         2845,
         2848,
         2855,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2855,
         2855,
         2866,
         2855,
         2866,
         2866,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2855,
         2856,
         2855,
         2856,
         2855,
         2857,
         2858,
         2866,
         2867,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2855,
         2856,
         2858,
         2855,
         2867,
         2866,
         2855,
         2856,
         2855,
         2856,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2856,
         2867,
         2856,
         2867,
         2867,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2858,
         2858,
         2867,
         2867,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2856,
         2856,
         2858,
         2856,
         2867,
         2867,
         2856,
         2856,
         2856,
         2856,
         20,
         2851,
         20,
         1365,
         1365,
         20,
         1366,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2845,
         2845,
         2845,
         2845,
         20,
         2846,
         2846,
         2845,
         2845,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2845,
         20,
         2845,
         2846,
         20,
         2845,
         20,
         2845,
         2845,
         20,
         2845,
         2847,
         2852,
         2847,
         2847,
         2847,
         2847,
         2863,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2863,
         2863,
         2863,
         2863,
         2863,
         2872,
         2847,
         2847,
         2874,
         2874,
         2874,
         2874,
         2847,
         2875,
         2875,
         2874,
         2874,
         2847,
         2847,
         2847,
         2863,
         2863,
         2863,
         2874,
         2847,
         2874,
         2875,
         2847,
         2874,
         2847,
         2874,
         2874,
         2847,
         2874,
         2847,
         2853,
         2847,
         2847,
         2847,
         2847,
         2863,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2847,
         2863,
         2863,
         2863,
         2863,
         2863,
         2872,
         2847,
         2847,
         2874,
         2874,
         2874,
         2874,
         2847,
         2875,
         2875,
         2874,
         2874,
         2847,
         2847,
         2847,
         2863,
         2863,
         2863,
         2874,
         2847,
         2874,
         2875,
         2847,
         2874,
         2847,
         2874,
         2874,
         2847,
         2874,
         1366,
         2854,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2611,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2616,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         2873,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         1366,
         2972,
         1366,
         1378,
         1378,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         1366,
         2614,
         2614,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         1366,
         1366,
         2861,
         2861,
         2861,
         2861,
         1366,
         2862,
         2862,
         2861,
         2861,
         1366,
         2863,
         2863,
         1366,
         1366,
         1366,
         2861,
         1366,
         2861,
         2862,
         1366,
         2861,
         1366,
         2861,
         2861,
         1366,
         2861,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2976,
         2856,
         2976,
         2856,
         2976,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2977,
         2977,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2856,
         2856,
         2977,
         2976,
         2856,
         2856,
         2856,
         2856,
         2856,
         2856,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         2973,
         1376,
         1376,
         2973,
         2973,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2974,
         2975,
         2973,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2975,
         2858,
         2973,
         2856,
         2973,
         2974,
         2975,
         2973,
         2975,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2976,
         2856,
         2976,
         2856,
         2976,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2977,
         2977,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2975,
         2856,
         2977,
         2978,
         2856,
         2975,
         2975,
         2975,
         2975,
         2975,
         2846,
         2850,
         2858,
         2860,
         2860,
         2846,
         2862,
         2846,
         2865,
         2865,
         2846,
         2846,
         2846,
         2846,
         2846,
         2858,
         2977,
         2858,
         2977,
         2858,
         2977,
         2869,
         2869,
         2862,
         2862,
         2862,
         2862,
         2862,
         2871,
         2846,
         2846,
         2858,
         2858,
         2858,
         2858,
         2858,
         2977,
         2977,
         2858,
         2858,
         2846,
         2875,
         2875,
         2862,
         2862,
         2862,
         2977,
         2858,
         2977,
         2977,
         2977,
         2977,
         2858,
         2858,
         2858,
         2858,
         2977,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         2973,
         1376,
         1376,
         2973,
         2973,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2974,
         2975,
         2973,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2976,
         2973,
         2978,
         2977,
         2973,
         2976,
         2973,
         2974,
         2975,
         2973,
         2978,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2856,
         2976,
         2867,
         2976,
         2867,
         2980,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2856,
         2856,
         2856,
         2977,
         2977,
         2867,
         2867,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2856,
         2856,
         2977,
         2976,
         2867,
         2867,
         2856,
         2856,
         2856,
         2856,
         20,
         21,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         2973,
         1376,
         2617,
         2973,
         2979,
         2617,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2974,
         2975,
         2973,
         2857,
         2858,
         2866,
         2867,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2975,
         2858,
         2973,
         2867,
         2979,
         2974,
         2975,
         2973,
         2975,
         2845,
         2848,
         2855,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2974,
         2855,
         2855,
         2974,
         2974,
         2855,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2855,
         2856,
         2974,
         2975,
         2974,
         2857,
         2858,
         2855,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2974,
         2975,
         2858,
         2974,
         2856,
         2974,
         2974,
         2975,
         2974,
         2975,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2975,
         2856,
         2856,
         2975,
         2975,
         2856,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2975,
         2975,
         2975,
         2858,
         2858,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2975,
         2975,
         2858,
         2975,
         2856,
         2975,
         2975,
         2975,
         2975,
         2975,
         20,
         2849,
         1376,
         1365,
         1365,
         20,
         1366,
         20,
         1377,
         1377,
         20,
         20,
         20,
         20,
         20,
         2973,
         1376,
         1376,
         2973,
         2973,
         1376,
         2603,
         2603,
         1366,
         1366,
         1366,
         1366,
         1366,
         2604,
         20,
         20,
         2855,
         2856,
         2974,
         2975,
         2973,
         2857,
         2858,
         2855,
         2856,
         20,
         2847,
         2847,
         1366,
         1366,
         1366,
         2856,
         2973,
         2975,
         2858,
         2973,
         2856,
         2973,
         2974,
         2975,
         2973,
         2975,
         2845,
         2848,
         2856,
         2859,
         2859,
         2845,
         2861,
         2845,
         2864,
         2864,
         2845,
         2845,
         2845,
         2845,
         2845,
         2975,
         2976,
         2856,
         2978,
         2975,
         2976,
         2868,
         2868,
         2861,
         2861,
         2861,
         2861,
         2861,
         2870,
         2845,
         2845,
         2856,
         2856,
         2975,
         2975,
         2975,
         2977,
         2977,
         2856,
         2856,
         2845,
         2874,
         2874,
         2861,
         2861,
         2861,
         2856,
         2975,
         2975,
         2977,
         2978,
         2856,
         2975,
         2975,
         2975,
         2975,
         2975,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         22,
         1382,
         1383,
         1384,
         1385,
         1386,
         1387,
         1388,
         1389,
         2349,
         2350,
         2351,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         8,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         14,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         15,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         10,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         17,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         14,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         17,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         1,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         23,
         1390,
         1391,
         1392,
         1393,
         1394,
         1395,
         1396,
         1397,
         1398,
         1399,
         1400,
         1401,
         1402,
         2621,
         2622,
         2623,
         2876,
      },
   },
   [nir_op_ineg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         11,
         5,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         24,
         1664,
         1665,
         1666,
         1667,
         1668,
         1669,
         1670,
         1671,
         1672,
         1673,
         2796,
      },
   },
   [nir_search_op_u2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         25,
         1974,
         1975,
         1976,
         1977,
      },
   },
   [nir_op_iabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         26,
         1674,
         1675,
         1676,
         2599,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         11,
         0,
         11,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         0,
         0,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         17,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         20,
         21,
         0,
         0,
         22,
         23,
         24,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         25,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         26,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         27,
         6,
         6,
         28,
         6,
         6,
         6,
         29,
         30,
         31,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         37,
         37,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         38,
         39,
         40,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         26,
         26,
         26,
         26,
         26,
         8,
         8,
         8,
         8,
         25,
         25,
         25,
         25,
         25,
         25,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         0,
         0,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         4,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         36,
         36,
         34,
         41,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         14,
         0,
         14,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         14,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 42,
      .table = (const uint16_t []) {
      
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         28,
         28,
         28,
         404,
         28,
         405,
         406,
         407,
         408,
         409,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         417,
         418,
         28,
         28,
         28,
         419,
         28,
         419,
         2043,
         2044,
         2045,
         2046,
         2047,
         2048,
         2049,
         2050,
         2051,
         2052,
         2053,
         2054,
         28,
         28,
         28,
         28,
         406,
         27,
         28,
         420,
         420,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         404,
         420,
         420,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         28,
         27,
         27,
         421,
         421,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         405,
         27,
         27,
         421,
         421,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         393,
         406,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         27,
         407,
         27,
         27,
         27,
         27,
         422,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         422,
         422,
         422,
         422,
         422,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         422,
         394,
         408,
         394,
         394,
         394,
         394,
         423,
         394,
         394,
         433,
         434,
         435,
         436,
         437,
         394,
         438,
         439,
         440,
         441,
         394,
         394,
         394,
         394,
         394,
         394,
         2038,
         2038,
         394,
         423,
         423,
         423,
         423,
         423,
         2061,
         2062,
         2063,
         2064,
         394,
         394,
         394,
         394,
         423,
         395,
         409,
         395,
         395,
         395,
         395,
         424,
         395,
         433,
         395,
         442,
         443,
         444,
         445,
         395,
         446,
         447,
         448,
         449,
         395,
         395,
         395,
         395,
         395,
         395,
         2065,
         2066,
         395,
         424,
         424,
         424,
         424,
         424,
         2067,
         2068,
         2069,
         2070,
         395,
         395,
         395,
         395,
         424,
         396,
         410,
         396,
         396,
         396,
         396,
         425,
         396,
         434,
         442,
         396,
         450,
         451,
         452,
         396,
         453,
         454,
         455,
         456,
         396,
         396,
         396,
         396,
         396,
         396,
         2071,
         2072,
         396,
         425,
         425,
         425,
         425,
         425,
         2073,
         2074,
         2075,
         2076,
         396,
         396,
         396,
         396,
         425,
         397,
         411,
         397,
         397,
         397,
         397,
         426,
         397,
         435,
         443,
         450,
         397,
         457,
         458,
         397,
         459,
         460,
         461,
         462,
         397,
         397,
         397,
         397,
         397,
         397,
         2077,
         2078,
         397,
         426,
         426,
         426,
         426,
         426,
         2079,
         2080,
         2081,
         2082,
         397,
         397,
         397,
         397,
         426,
         398,
         412,
         398,
         398,
         398,
         398,
         427,
         398,
         436,
         444,
         451,
         457,
         398,
         463,
         398,
         464,
         465,
         466,
         467,
         398,
         398,
         398,
         398,
         398,
         398,
         2083,
         2084,
         398,
         427,
         427,
         427,
         427,
         427,
         2085,
         2086,
         2087,
         2088,
         398,
         398,
         398,
         398,
         427,
         399,
         413,
         399,
         399,
         399,
         399,
         428,
         399,
         437,
         445,
         452,
         458,
         463,
         399,
         399,
         468,
         469,
         470,
         471,
         399,
         399,
         399,
         399,
         399,
         399,
         2089,
         2090,
         399,
         428,
         428,
         428,
         428,
         428,
         2091,
         2092,
         2093,
         2094,
         399,
         399,
         399,
         399,
         428,
         27,
         414,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         400,
         415,
         400,
         400,
         400,
         400,
         429,
         400,
         438,
         446,
         453,
         459,
         464,
         468,
         400,
         400,
         472,
         473,
         474,
         400,
         400,
         400,
         400,
         400,
         400,
         2095,
         2096,
         400,
         429,
         429,
         429,
         429,
         429,
         2097,
         2098,
         2099,
         2100,
         400,
         400,
         400,
         400,
         429,
         401,
         416,
         401,
         401,
         401,
         401,
         430,
         401,
         439,
         447,
         454,
         460,
         465,
         469,
         401,
         472,
         401,
         475,
         476,
         401,
         401,
         401,
         401,
         401,
         401,
         2101,
         2102,
         401,
         430,
         430,
         430,
         430,
         430,
         2103,
         2104,
         2105,
         2106,
         401,
         401,
         401,
         401,
         430,
         402,
         417,
         402,
         402,
         402,
         402,
         431,
         402,
         440,
         448,
         455,
         461,
         466,
         470,
         402,
         473,
         475,
         402,
         477,
         402,
         402,
         402,
         402,
         402,
         402,
         2107,
         2108,
         402,
         431,
         431,
         431,
         431,
         431,
         2109,
         2110,
         2111,
         2112,
         402,
         402,
         402,
         402,
         431,
         403,
         418,
         403,
         403,
         403,
         403,
         432,
         403,
         441,
         449,
         456,
         462,
         467,
         471,
         403,
         474,
         476,
         477,
         403,
         403,
         403,
         403,
         403,
         403,
         403,
         2113,
         2114,
         403,
         432,
         432,
         432,
         432,
         432,
         2115,
         2116,
         2117,
         2118,
         403,
         403,
         403,
         403,
         432,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         478,
         478,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         478,
         27,
         478,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         478,
         478,
         478,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         419,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         479,
         479,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         479,
         27,
         479,
         2037,
         2038,
         479,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         27,
         419,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         479,
         479,
         479,
         2037,
         2038,
         479,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         2037,
         2043,
         2037,
         2037,
         2037,
         2037,
         2055,
         2037,
         2038,
         2065,
         2071,
         2077,
         2083,
         2089,
         2037,
         2095,
         2101,
         2107,
         2113,
         2037,
         2037,
         2037,
         2037,
         2037,
         2037,
         2037,
         2038,
         2037,
         2055,
         2055,
         2055,
         2055,
         2055,
         2119,
         2120,
         2121,
         2122,
         2037,
         2037,
         2037,
         2037,
         2055,
         2038,
         2044,
         2038,
         2038,
         2038,
         2038,
         2056,
         2038,
         2038,
         2066,
         2072,
         2078,
         2084,
         2090,
         2038,
         2096,
         2102,
         2108,
         2114,
         2038,
         2038,
         2038,
         2038,
         2038,
         2038,
         2038,
         2038,
         2038,
         2056,
         2056,
         2056,
         2056,
         2056,
         2123,
         2124,
         2125,
         2126,
         2038,
         2038,
         2038,
         2038,
         2056,
         27,
         2045,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         479,
         479,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         393,
         393,
         2046,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         393,
         2047,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         393,
         2048,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         393,
         2049,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         393,
         2050,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         393,
         393,
         393,
         393,
         393,
         2039,
         2051,
         2039,
         2039,
         2039,
         2039,
         2057,
         2039,
         2061,
         2067,
         2073,
         2079,
         2085,
         2091,
         2039,
         2097,
         2103,
         2109,
         2115,
         2039,
         2039,
         2039,
         2039,
         2039,
         2039,
         2119,
         2123,
         2039,
         2057,
         2057,
         2057,
         2057,
         2057,
         2039,
         2127,
         2128,
         2129,
         2039,
         2039,
         2039,
         2039,
         2057,
         2040,
         2052,
         2040,
         2040,
         2040,
         2040,
         2058,
         2040,
         2062,
         2068,
         2074,
         2080,
         2086,
         2092,
         2040,
         2098,
         2104,
         2110,
         2116,
         2040,
         2040,
         2040,
         2040,
         2040,
         2040,
         2120,
         2124,
         2040,
         2058,
         2058,
         2058,
         2058,
         2058,
         2127,
         2040,
         2130,
         2131,
         2040,
         2040,
         2040,
         2040,
         2058,
         2041,
         2053,
         2041,
         2041,
         2041,
         2041,
         2059,
         2041,
         2063,
         2069,
         2075,
         2081,
         2087,
         2093,
         2041,
         2099,
         2105,
         2111,
         2117,
         2041,
         2041,
         2041,
         2041,
         2041,
         2041,
         2121,
         2125,
         2041,
         2059,
         2059,
         2059,
         2059,
         2059,
         2128,
         2130,
         2041,
         2132,
         2041,
         2041,
         2041,
         2041,
         2059,
         2042,
         2054,
         2042,
         2042,
         2042,
         2042,
         2060,
         2042,
         2064,
         2070,
         2076,
         2082,
         2088,
         2094,
         2042,
         2100,
         2106,
         2112,
         2118,
         2042,
         2042,
         2042,
         2042,
         2042,
         2042,
         2122,
         2126,
         2042,
         2060,
         2060,
         2060,
         2060,
         2060,
         2129,
         2131,
         2132,
         2042,
         2042,
         2042,
         2042,
         2042,
         2060,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         27,
         27,
         2953,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         27,
         2133,
         2133,
         393,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         2133,
         27,
         2133,
         393,
         27,
         28,
         27,
         27,
         27,
         27,
         393,
         27,
         394,
         395,
         396,
         397,
         398,
         399,
         27,
         400,
         401,
         402,
         403,
         27,
         27,
         27,
         27,
         27,
         27,
         2037,
         2038,
         27,
         393,
         393,
         393,
         393,
         393,
         2039,
         2040,
         2041,
         2042,
         27,
         2133,
         2133,
         2133,
         393,
         393,
         406,
         393,
         393,
         393,
         393,
         393,
         422,
         423,
         424,
         425,
         426,
         427,
         428,
         393,
         429,
         430,
         431,
         432,
         393,
         393,
         393,
         393,
         393,
         393,
         2055,
         2056,
         393,
         393,
         393,
         393,
         393,
         393,
         2057,
         2058,
         2059,
         2060,
         2953,
         393,
         393,
         393,
         393,
      },
   },
   [nir_op_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         29,
         30,
         30,
         30,
      },
   },
   [nir_op_isub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         31,
         32,
         31,
         32,
      },
   },
   [nir_op_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         33,
         34,
         34,
         34,
      },
   },
   [nir_op_usub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         35,
         36,
         35,
         36,
      },
   },
   [nir_op_usadd_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         37,
         37,
         37,
      },
   },
   [nir_op_fmulz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         38,
         39,
         1635,
         38,
         38,
         38,
         1636,
         38,
         39,
         39,
         1637,
         1638,
         1639,
         1640,
         1641,
         39,
         1635,
         1637,
         1635,
         1635,
         1635,
         1635,
         1642,
         1635,
         38,
         1638,
         1635,
         38,
         38,
         38,
         1636,
         38,
         38,
         1639,
         1635,
         38,
         38,
         38,
         1636,
         38,
         38,
         1640,
         1635,
         38,
         38,
         38,
         1636,
         38,
         1636,
         1641,
         1642,
         1636,
         1636,
         1636,
         1636,
         1636,
         38,
         39,
         1635,
         38,
         38,
         38,
         1636,
         1643,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         10,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         11,
         3,
         2,
         2,
         2,
         2,
         10,
         2,
         2,
         2,
         10,
         2,
         2,
         2,
         10,
         2,
         10,
         2,
         2,
         10,
         10,
         2,
         0,
         12,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         10,
         10,
         10,
         10,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         14,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         12,
         12,
         12,
         12,
         15,
         12,
         12,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         10,
         10,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         6,
         6,
         6,
         13,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         44,
         45,
         1412,
         1412,
         44,
         44,
         1413,
         44,
         44,
         44,
         1412,
         1412,
         44,
         1413,
         1413,
         44,
         1414,
         1415,
         1416,
         1416,
         1414,
         1414,
         1417,
         1414,
         1414,
         1414,
         1416,
         1416,
         1414,
         1417,
         1417,
         1414,
         1418,
         1419,
         1420,
         1420,
         1418,
         1418,
         1421,
         1418,
         1418,
         1418,
         1420,
         1420,
         1418,
         1421,
         1421,
         1418,
         1422,
         1423,
         1424,
         1424,
         1422,
         1422,
         1425,
         1422,
         1422,
         1422,
         1424,
         1424,
         1422,
         1425,
         1425,
         1422,
         1426,
         1427,
         1428,
         1428,
         1426,
         1426,
         1429,
         1426,
         1426,
         1426,
         1428,
         1428,
         1426,
         1429,
         1429,
         1426,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         2412,
         42,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         2413,
         2414,
         2415,
         2415,
         2413,
         2413,
         2416,
         2413,
         2413,
         2413,
         2415,
         2415,
         2413,
         2416,
         2416,
         2413,
         1414,
         1415,
         1416,
         1416,
         1414,
         1414,
         1417,
         1414,
         1414,
         1414,
         1416,
         1416,
         1414,
         1417,
         1417,
         1414,
         1418,
         1419,
         1420,
         1420,
         1418,
         1418,
         1421,
         1418,
         1418,
         1418,
         1420,
         1420,
         1418,
         1421,
         1421,
         1418,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         2915,
         2915,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1414,
         1415,
         1416,
         1416,
         1414,
         1414,
         1417,
         1414,
         1414,
         1414,
         1416,
         1416,
         1414,
         1417,
         1417,
         1414,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1418,
         1419,
         1420,
         1420,
         1418,
         1418,
         1421,
         1418,
         1418,
         1418,
         1420,
         1420,
         1418,
         1421,
         1421,
         1418,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1422,
         1423,
         1424,
         1424,
         1422,
         1422,
         1425,
         1422,
         1422,
         1422,
         1424,
         1424,
         1422,
         1425,
         1425,
         1422,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         1431,
         1432,
         1433,
         1433,
         1431,
         1431,
         1434,
         1431,
         1431,
         1431,
         1433,
         1433,
         1431,
         1434,
         1434,
         1431,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1435,
         1436,
         1437,
         1437,
         1435,
         1435,
         1438,
         1435,
         1435,
         1435,
         1437,
         1437,
         1435,
         1438,
         2418,
         1435,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         2419,
         2420,
         2421,
         2421,
         2419,
         2419,
         2422,
         2419,
         2419,
         2419,
         2421,
         2421,
         2419,
         2422,
         2422,
         2419,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         2917,
         2917,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         1426,
         1427,
         1428,
         1428,
         1426,
         1426,
         1429,
         1426,
         1426,
         1426,
         1428,
         1428,
         1426,
         1429,
         1429,
         1426,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         2412,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         1435,
         1436,
         1437,
         1437,
         1435,
         1435,
         1438,
         1435,
         1435,
         1435,
         1437,
         1437,
         1435,
         1438,
         2418,
         1435,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2423,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         2424,
         2425,
         2426,
         2426,
         2424,
         2424,
         2427,
         2424,
         2424,
         2424,
         2426,
         2426,
         2424,
         2427,
         2428,
         2424,
         40,
         41,
         2429,
         2429,
         40,
         40,
         1404,
         40,
         40,
         40,
         2429,
         2429,
         40,
         1404,
         2407,
         40,
         40,
         41,
         2429,
         2429,
         40,
         40,
         1404,
         40,
         40,
         40,
         2429,
         2429,
         40,
         1404,
         2407,
         40,
         2424,
         2425,
         2426,
         2426,
         2424,
         2424,
         2427,
         2424,
         2424,
         2424,
         2918,
         2918,
         2424,
         2427,
         2428,
         2424,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         2407,
         40,
         2430,
         2431,
         2432,
         2432,
         2430,
         2430,
         2433,
         2430,
         2430,
         2430,
         2432,
         2432,
         2430,
         2433,
         2433,
         2430,
         2434,
         2435,
         2436,
         2436,
         2434,
         2434,
         2437,
         2434,
         2434,
         2434,
         2436,
         2436,
         2434,
         2437,
         2437,
         2434,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2413,
         2414,
         2415,
         2415,
         2413,
         2413,
         2416,
         2413,
         2413,
         2413,
         2415,
         2415,
         2413,
         2416,
         2416,
         2413,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         2419,
         2420,
         2421,
         2421,
         2419,
         2419,
         2422,
         2419,
         2419,
         2419,
         2421,
         2421,
         2419,
         2422,
         2422,
         2419,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2423,
         2408,
         2434,
         2435,
         2436,
         2436,
         2434,
         2434,
         2437,
         2434,
         2434,
         2434,
         2436,
         2436,
         2434,
         2437,
         2437,
         2434,
         2434,
         2435,
         2436,
         2436,
         2434,
         2434,
         2437,
         2434,
         2434,
         2434,
         2436,
         2436,
         2434,
         2437,
         2437,
         2434,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2919,
         2919,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1414,
         1415,
         1416,
         1416,
         1414,
         1414,
         1417,
         1414,
         1414,
         1414,
         1416,
         1416,
         1414,
         1417,
         1417,
         1414,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1418,
         1419,
         1420,
         1420,
         1418,
         1418,
         1421,
         1418,
         1418,
         1418,
         1420,
         1420,
         1418,
         1421,
         1421,
         1418,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1430,
         1430,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1430,
         1430,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2417,
         2417,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2417,
         2417,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2424,
         2425,
         2426,
         2426,
         2424,
         2424,
         2427,
         2424,
         2424,
         2424,
         2426,
         2426,
         2424,
         2427,
         2428,
         2424,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         2429,
         2429,
         40,
         40,
         1404,
         40,
         40,
         40,
         2429,
         2429,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         1405,
         1405,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         1410,
         1410,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         2429,
         2429,
         40,
         40,
         1404,
         40,
         40,
         40,
         2429,
         2429,
         40,
         1404,
         2407,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2410,
         2410,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         1407,
         1407,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         1403,
         1403,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         42,
         43,
         1405,
         1405,
         42,
         42,
         1406,
         42,
         42,
         42,
         2915,
         2915,
         42,
         1406,
         1406,
         42,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         1408,
         1409,
         1410,
         1410,
         1408,
         1408,
         1411,
         1408,
         1408,
         1408,
         2917,
         2917,
         1408,
         1411,
         1411,
         1408,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         2424,
         2425,
         2426,
         2426,
         2424,
         2424,
         2427,
         2424,
         2424,
         2424,
         2918,
         2918,
         2424,
         2427,
         2428,
         2424,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         2408,
         2409,
         2410,
         2410,
         2408,
         2408,
         2411,
         2408,
         2408,
         2408,
         2919,
         2919,
         2408,
         2411,
         2411,
         2408,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1407,
         1407,
         40,
         40,
         1404,
         40,
         40,
         40,
         2916,
         2916,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
         40,
         41,
         1403,
         1403,
         40,
         40,
         1404,
         40,
         40,
         40,
         2914,
         2914,
         40,
         1404,
         1404,
         40,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         46,
         47,
         46,
         1644,
         1644,
         1645,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         50,
         51,
         50,
         1649,
         1649,
         1650,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         1651,
         1652,
         1651,
         1653,
         1653,
         1654,
         1655,
         1656,
         1655,
         1657,
         1657,
         1658,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         1659,
         1660,
         1659,
         1661,
         1661,
         1662,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         1651,
         1652,
         1651,
         1653,
         1653,
         1654,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         1655,
         1656,
         1655,
         1657,
         1657,
         1658,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         48,
         49,
         48,
         1646,
         1646,
         1647,
         46,
         47,
         46,
         1644,
         1644,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1648,
         1648,
         1645,
         46,
         47,
         46,
         1644,
         1644,
         1645,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         3,
         4,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         9,
         9,
         10,
         11,
         12,
         13,
         14,
         15,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         20,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         26,
         27,
         28,
         29,
         30,
         31,
         32,
         33,
         34,
         34,
         34,
         34,
         35,
         35,
         36,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         37,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         40,
         41,
         42,
         43,
         44,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         0,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         3,
         4,
         3,
         3,
         3,
         4,
         3,
         3,
         4,
         4,
         4,
         3,
         4,
         3,
         3,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         9,
         9,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         47,
         9,
         9,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         46,
         46,
         46,
         46,
         46,
         46,
         46,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         15,
         14,
         14,
         15,
         15,
         15,
         19,
         19,
         20,
         19,
         20,
         19,
         20,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         40,
         40,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         22,
         22,
         22,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         22,
         23,
         22,
         22,
         22,
         23,
         23,
         22,
         23,
         23,
         22,
         23,
         22,
         22,
         22,
         23,
         22,
         23,
         22,
         22,
         22,
         22,
         26,
         26,
         26,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         26,
         27,
         26,
         26,
         26,
         27,
         27,
         26,
         27,
         27,
         27,
         26,
         27,
         26,
         26,
         26,
         27,
         26,
         27,
         26,
         26,
         26,
         26,
         30,
         30,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         30,
         30,
         30,
         30,
         12,
         12,
         48,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         49,
         0,
         0,
         0,
         0,
         50,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         51,
         11,
         11,
         52,
         53,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         54,
         55,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         56,
         57,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         21,
         21,
         21,
         21,
         58,
         21,
         21,
         21,
         21,
         21,
         21,
         21,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         32,
         32,
         32,
         32,
         32,
         32,
         33,
         32,
         33,
         32,
         32,
         33,
         32,
         33,
         32,
         32,
         33,
         33,
         32,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         44,
         44,
         44,
         44,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         59,
         4,
         14,
         14,
         14,
         14,
         14,
         14,
         60,
         15,
         0,
         0,
         0,
         17,
         17,
         17,
         17,
         17,
         16,
         16,
         31,
         31,
         31,
         31,
         31,
         30,
         30,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         46,
         46,
         46,
         46,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         61,
         11,
         21,
         21,
         21,
         21,
         21,
         21,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         23,
         23,
         23,
         23,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         27,
         27,
         27,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         12,
         33,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         57,
         57,
         55,
         50,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         8,
         8,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         62,
         63,
         64,
         4,
         4,
         15,
         65,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         31,
         0,
         23,
         27,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         66,
         0,
         0,
         23,
         23,
         23,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         67,
         68,
         8,
         0,
         0,
         8,
         8,
         4,
         15,
         9,
         69,
         70,
         71,
         8,
         8,
         8,
         4,
         15,
         9,
         72,
         73,
         74,
         8,
         8,
         8,
         4,
         15,
         9,
         72,
      },
      
      .num_filtered_states = 75,
      .table = (const uint16_t []) {
      
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         53,
         53,
         1093,
         53,
         1094,
         53,
         1095,
         1096,
         1097,
         1098,
         53,
         53,
         53,
         53,
         53,
         1099,
         53,
         53,
         53,
         53,
         53,
         1100,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1101,
         53,
         1102,
         53,
         1103,
         53,
         1104,
         1105,
         2513,
         2514,
         53,
         2515,
         2516,
         2517,
         53,
         53,
         2518,
         2519,
         2520,
         2521,
         2522,
         2841,
         2842,
         53,
         1100,
         1100,
         1100,
         53,
         2988,
         2991,
         2992,
         2999,
         3000,
         3001,
         3008,
         3009,
         3010,
         52,
         1093,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         1094,
         52,
         1106,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         2523,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         1095,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1086,
         1096,
         1086,
         1086,
         1086,
         1086,
         1086,
         1107,
         1107,
         1108,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1109,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1110,
         1086,
         1111,
         1086,
         1112,
         1086,
         1113,
         1086,
         2524,
         1108,
         1086,
         1086,
         1086,
         2525,
         1086,
         1086,
         2526,
         2527,
         2528,
         2529,
         2530,
         1086,
         1086,
         1086,
         1109,
         1109,
         1109,
         1086,
         1108,
         1086,
         1086,
         1108,
         1086,
         1086,
         1108,
         1086,
         1086,
         1086,
         1097,
         1086,
         1086,
         1086,
         1086,
         1086,
         1107,
         1107,
         1108,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1114,
         1086,
         1086,
         1109,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1086,
         1110,
         1086,
         1111,
         1086,
         1112,
         1086,
         1113,
         1086,
         2524,
         1108,
         1086,
         1086,
         1086,
         2525,
         1086,
         1086,
         2526,
         2527,
         2528,
         2529,
         2530,
         1086,
         1086,
         1086,
         1109,
         1109,
         1109,
         1086,
         1108,
         1086,
         1086,
         1108,
         1086,
         1086,
         1108,
         1086,
         1086,
         1087,
         1098,
         1087,
         1087,
         1087,
         1087,
         1087,
         1108,
         1108,
         1115,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1116,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1117,
         1087,
         1118,
         1087,
         1119,
         1087,
         1120,
         1087,
         2531,
         1115,
         1087,
         1087,
         1087,
         2532,
         1087,
         1087,
         2533,
         2534,
         2535,
         2536,
         2537,
         1087,
         1087,
         1087,
         1116,
         1116,
         1116,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         1121,
         1121,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1121,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         1122,
         1121,
         1121,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1121,
         52,
         52,
         2538,
         1122,
         1122,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1122,
         1088,
         1088,
         1088,
         1122,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         1121,
         1121,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2539,
         1121,
         1121,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1121,
         1088,
         1088,
         1088,
         1121,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         1121,
         1121,
         52,
         1123,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1123,
         52,
         52,
         2539,
         1121,
         1121,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         2843,
         1088,
         1088,
         1088,
         2843,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         52,
         1099,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         1124,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1125,
         1125,
         52,
         52,
         52,
         1088,
         1126,
         1126,
         1126,
         1126,
         1127,
         1127,
         1127,
         1127,
         1128,
         1128,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1125,
         1129,
         52,
         52,
         52,
         1088,
         1126,
         1126,
         1126,
         1126,
         1127,
         1127,
         1127,
         1127,
         1128,
         1128,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1114,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1130,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1131,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1131,
         1131,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1088,
         1100,
         1088,
         1088,
         1088,
         1088,
         1088,
         1109,
         1109,
         1116,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1132,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1133,
         1088,
         1134,
         1088,
         1135,
         1088,
         1136,
         1088,
         2540,
         1116,
         1088,
         1088,
         1088,
         2541,
         1088,
         1088,
         2542,
         2543,
         2544,
         2545,
         2546,
         1088,
         1088,
         1088,
         1132,
         1132,
         1132,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1126,
         1126,
         52,
         52,
         52,
         1088,
         1137,
         1137,
         1137,
         1137,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1126,
         1126,
         52,
         52,
         52,
         1088,
         1137,
         1138,
         1139,
         1140,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1126,
         1126,
         52,
         52,
         52,
         1088,
         1137,
         1139,
         1141,
         1142,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1126,
         1126,
         52,
         52,
         52,
         1088,
         1137,
         1140,
         1142,
         1143,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1127,
         1127,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         1144,
         1144,
         1144,
         1144,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1127,
         1127,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         1144,
         1145,
         1146,
         1147,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1127,
         1127,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         1144,
         1146,
         1148,
         1149,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1127,
         1127,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         1144,
         1147,
         1149,
         1150,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1128,
         1128,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1151,
         1151,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         1128,
         1128,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1151,
         1152,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1153,
         1153,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1153,
         1153,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         2981,
         2982,
         2983,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1154,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1155,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1156,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1157,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1089,
         1101,
         1089,
         1089,
         1089,
         1089,
         1089,
         1110,
         1110,
         1117,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1133,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1089,
         1157,
         1157,
         1089,
         1158,
         1089,
         1159,
         1089,
         1160,
         1089,
         2547,
         1117,
         1089,
         1089,
         1089,
         2548,
         1089,
         1089,
         2549,
         2550,
         2551,
         2552,
         2553,
         1089,
         1089,
         1089,
         1133,
         1133,
         1133,
         1089,
         1117,
         1089,
         1089,
         1117,
         1089,
         1089,
         1117,
         1089,
         1089,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1161,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1090,
         1102,
         1090,
         1090,
         1090,
         1090,
         1090,
         1111,
         1111,
         1118,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1134,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1090,
         1158,
         1161,
         1161,
         1090,
         1162,
         1090,
         1163,
         1090,
         2554,
         1118,
         1090,
         1090,
         1090,
         2555,
         1090,
         1090,
         2556,
         2557,
         2558,
         2559,
         2560,
         1090,
         1090,
         1090,
         1134,
         1134,
         1134,
         1090,
         1118,
         1090,
         1090,
         1118,
         1090,
         1090,
         1118,
         1090,
         1090,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1164,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1091,
         1103,
         1091,
         1091,
         1091,
         1091,
         1091,
         1112,
         1112,
         1119,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1135,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1091,
         1159,
         1091,
         1162,
         1164,
         1164,
         1091,
         1165,
         1091,
         2561,
         1119,
         1091,
         1091,
         1091,
         2562,
         1091,
         1091,
         2563,
         2564,
         2565,
         2566,
         2567,
         1091,
         1091,
         1091,
         1135,
         1135,
         1135,
         1091,
         1119,
         1091,
         1091,
         1119,
         1091,
         1091,
         1119,
         1091,
         1091,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1166,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1092,
         1104,
         1092,
         1092,
         1092,
         1092,
         1092,
         1113,
         1113,
         1120,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1136,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1160,
         1092,
         1163,
         1092,
         1165,
         1166,
         1166,
         1092,
         2568,
         1120,
         1092,
         1092,
         1092,
         2569,
         1092,
         1092,
         2570,
         2571,
         2572,
         2573,
         2574,
         1092,
         1092,
         1092,
         1136,
         1136,
         1136,
         1092,
         1120,
         1092,
         1092,
         1120,
         1092,
         1092,
         1120,
         1092,
         1092,
         52,
         1105,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         2506,
         2513,
         2506,
         2506,
         2506,
         2506,
         2506,
         2524,
         2524,
         2531,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2540,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2506,
         2547,
         2506,
         2554,
         2506,
         2561,
         2506,
         2568,
         2506,
         2531,
         2531,
         2506,
         2506,
         2506,
         2575,
         2506,
         2506,
         2576,
         2577,
         2578,
         2579,
         2580,
         2506,
         2506,
         2506,
         2540,
         2540,
         2540,
         2506,
         2531,
         2506,
         2506,
         2531,
         2506,
         2506,
         2531,
         2506,
         2506,
         1087,
         2514,
         1087,
         1087,
         1087,
         1087,
         1087,
         1108,
         1108,
         1115,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1116,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1117,
         1087,
         1118,
         1087,
         1119,
         1087,
         1120,
         1087,
         2531,
         1115,
         1087,
         1087,
         1087,
         2532,
         1087,
         1087,
         2533,
         2534,
         2535,
         2536,
         2537,
         1087,
         1087,
         1087,
         1116,
         1116,
         1116,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         1121,
         1121,
         52,
         1123,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1123,
         52,
         52,
         2539,
         1121,
         1121,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         2844,
         1088,
         1088,
         1088,
         2984,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         2515,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         2516,
         52,
         52,
         52,
         2523,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         2581,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         2507,
         2517,
         2507,
         2507,
         2507,
         2507,
         2507,
         2525,
         2525,
         2532,
         2507,
         2538,
         2539,
         2539,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2541,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2507,
         2548,
         2507,
         2555,
         2507,
         2562,
         2507,
         2569,
         2507,
         2575,
         2532,
         2539,
         2507,
         2507,
         2538,
         2538,
         2538,
         2582,
         2583,
         2584,
         2585,
         2586,
         2507,
         2507,
         2538,
         2541,
         2541,
         2541,
         2538,
         2532,
         2507,
         2507,
         2532,
         2507,
         2507,
         2532,
         2507,
         2507,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         1122,
         1121,
         1121,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1121,
         52,
         52,
         2538,
         1122,
         2587,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1122,
         1088,
         1088,
         1088,
         1122,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         1122,
         1121,
         1121,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         1121,
         52,
         52,
         2538,
         2587,
         1122,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1122,
         1088,
         1088,
         1088,
         1122,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         2508,
         2518,
         2508,
         2508,
         2508,
         2508,
         2508,
         2526,
         2526,
         2533,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2542,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2508,
         2549,
         2508,
         2556,
         2508,
         2563,
         2508,
         2570,
         2508,
         2576,
         2533,
         2508,
         2508,
         2508,
         2582,
         2508,
         2508,
         2508,
         2588,
         2589,
         2590,
         2591,
         2508,
         2508,
         2508,
         2542,
         2542,
         2542,
         2508,
         2533,
         2508,
         2508,
         2533,
         2508,
         2508,
         2533,
         2508,
         2508,
         2509,
         2519,
         2509,
         2509,
         2509,
         2509,
         2509,
         2527,
         2527,
         2534,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2543,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2509,
         2550,
         2509,
         2557,
         2509,
         2564,
         2509,
         2571,
         2509,
         2577,
         2534,
         2509,
         2509,
         2509,
         2583,
         2509,
         2509,
         2588,
         2509,
         2592,
         2593,
         2594,
         2509,
         2509,
         2509,
         2543,
         2543,
         2543,
         2509,
         2534,
         2509,
         2509,
         2534,
         2509,
         2509,
         2534,
         2509,
         2509,
         2510,
         2520,
         2510,
         2510,
         2510,
         2510,
         2510,
         2528,
         2528,
         2535,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2544,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2510,
         2551,
         2510,
         2558,
         2510,
         2565,
         2510,
         2572,
         2510,
         2578,
         2535,
         2510,
         2510,
         2510,
         2584,
         2510,
         2510,
         2589,
         2592,
         2510,
         2595,
         2596,
         2510,
         2510,
         2510,
         2544,
         2544,
         2544,
         2510,
         2535,
         2510,
         2510,
         2535,
         2510,
         2510,
         2535,
         2510,
         2510,
         2511,
         2521,
         2511,
         2511,
         2511,
         2511,
         2511,
         2529,
         2529,
         2536,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2545,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2511,
         2552,
         2511,
         2559,
         2511,
         2566,
         2511,
         2573,
         2511,
         2579,
         2536,
         2511,
         2511,
         2511,
         2585,
         2511,
         2511,
         2590,
         2593,
         2595,
         2597,
         2598,
         2511,
         2511,
         2511,
         2545,
         2545,
         2545,
         2511,
         2536,
         2511,
         2511,
         2536,
         2511,
         2511,
         2536,
         2511,
         2511,
         2512,
         2522,
         2512,
         2512,
         2512,
         2512,
         2512,
         2530,
         2530,
         2537,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2546,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2512,
         2553,
         2512,
         2560,
         2512,
         2567,
         2512,
         2574,
         2512,
         2580,
         2537,
         2512,
         2512,
         2512,
         2586,
         2512,
         2512,
         2591,
         2594,
         2596,
         2598,
         2546,
         2512,
         2512,
         2512,
         2546,
         2546,
         2546,
         2512,
         2537,
         2512,
         2512,
         2537,
         2512,
         2512,
         2537,
         2512,
         2512,
         52,
         2841,
         52,
         1106,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         2842,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         1124,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         1122,
         1121,
         2843,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         2844,
         52,
         52,
         2538,
         1122,
         1122,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1122,
         1088,
         1088,
         1088,
         1122,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1088,
         1100,
         1088,
         1088,
         1088,
         1088,
         1088,
         1109,
         1109,
         1116,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1132,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         2981,
         1088,
         1088,
         1088,
         1088,
         1133,
         1088,
         1134,
         1088,
         1135,
         1088,
         1136,
         1088,
         2540,
         1116,
         1088,
         1088,
         1088,
         2541,
         1088,
         1088,
         2542,
         2543,
         2544,
         2545,
         2546,
         1088,
         1088,
         1088,
         1132,
         1132,
         1132,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1088,
         1100,
         1088,
         1088,
         1088,
         1088,
         1088,
         1109,
         1109,
         1116,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1132,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         2982,
         1088,
         1088,
         1088,
         1088,
         1133,
         1088,
         1134,
         1088,
         1135,
         1088,
         1136,
         1088,
         2540,
         1116,
         1088,
         1088,
         1088,
         2541,
         1088,
         1088,
         2542,
         2543,
         2544,
         2545,
         2546,
         1088,
         1088,
         1088,
         1132,
         1132,
         1132,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1088,
         1100,
         1088,
         1088,
         1088,
         1088,
         1088,
         1109,
         1109,
         1116,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1132,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         1088,
         2983,
         1088,
         1088,
         1088,
         1088,
         1133,
         1088,
         1134,
         1088,
         1135,
         1088,
         1136,
         1088,
         2540,
         1116,
         1088,
         1088,
         1088,
         2541,
         1088,
         1088,
         2542,
         2543,
         2544,
         2545,
         2546,
         1088,
         1088,
         1088,
         1132,
         1132,
         1132,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         1116,
         1088,
         1088,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         1122,
         1121,
         2843,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         2984,
         52,
         52,
         2538,
         1122,
         1122,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         52,
         1122,
         1088,
         1088,
         1088,
         1122,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         52,
         52,
         1087,
         2988,
         1087,
         1087,
         1087,
         1087,
         1087,
         1108,
         1108,
         1115,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1116,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1117,
         1087,
         1118,
         1087,
         1119,
         1087,
         1120,
         1087,
         2531,
         1115,
         1087,
         1087,
         1087,
         2532,
         1087,
         1087,
         2533,
         2534,
         2535,
         2536,
         2537,
         1087,
         1087,
         1087,
         1116,
         1116,
         1116,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         52,
         2991,
         52,
         1106,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         2992,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         1124,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         2999,
         1087,
         1087,
         1087,
         1087,
         1087,
         1108,
         1108,
         1115,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1116,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1117,
         1087,
         1118,
         1087,
         1119,
         1087,
         1120,
         1087,
         2531,
         1115,
         1087,
         1087,
         1087,
         2532,
         1087,
         1087,
         2533,
         2534,
         2535,
         2536,
         2537,
         1087,
         1087,
         1087,
         1116,
         1116,
         1116,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         52,
         3000,
         52,
         1106,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         3001,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         1124,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         3008,
         1087,
         1087,
         1087,
         1087,
         1087,
         1108,
         1108,
         1115,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1116,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1087,
         1117,
         1087,
         1118,
         1087,
         1119,
         1087,
         1120,
         1087,
         2531,
         1115,
         1087,
         1087,
         1087,
         2532,
         1087,
         1087,
         2533,
         2534,
         2535,
         2536,
         2537,
         1087,
         1087,
         1087,
         1116,
         1116,
         1116,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         1115,
         1087,
         1087,
         52,
         3009,
         52,
         1106,
         1106,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         1106,
         52,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         1087,
         1106,
         52,
         52,
         3010,
         52,
         52,
         52,
         52,
         52,
         1086,
         1086,
         1087,
         52,
         52,
         52,
         52,
         1124,
         1124,
         52,
         52,
         52,
         52,
         52,
         1088,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1089,
         52,
         1090,
         52,
         1091,
         52,
         1092,
         52,
         2506,
         1087,
         52,
         52,
         52,
         2507,
         52,
         52,
         2508,
         2509,
         2510,
         2511,
         2512,
         52,
         1124,
         52,
         1088,
         1088,
         1088,
         52,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
         1087,
         52,
         1124,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         6,
         7,
         0,
         8,
         0,
         9,
         10,
         11,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         16,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         26,
         27,
         28,
         28,
         29,
         29,
         29,
         29,
         30,
         30,
         31,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         34,
         0,
         0,
         0,
         0,
         35,
         0,
         0,
         36,
         37,
         38,
         39,
         40,
         41,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         42,
         43,
         44,
         0,
         0,
         45,
         46,
         47,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         48,
         2,
         48,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         48,
         2,
         3,
         2,
         48,
         49,
         3,
         3,
         2,
         3,
         2,
         48,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         6,
         6,
         6,
         6,
         50,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         51,
         7,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         50,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         50,
         6,
         6,
         6,
         6,
         6,
         6,
         50,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         50,
         6,
         6,
         6,
         6,
         50,
         50,
         50,
         50,
         50,
         50,
         50,
         50,
         50,
         50,
         50,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         52,
         10,
         52,
         10,
         11,
         10,
         52,
         11,
         11,
         11,
         15,
         15,
         16,
         15,
         16,
         15,
         16,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         37,
         37,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         12,
         12,
         12,
         12,
         18,
         18,
         18,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         18,
         19,
         18,
         18,
         18,
         19,
         19,
         18,
         19,
         19,
         18,
         19,
         18,
         18,
         18,
         19,
         18,
         19,
         18,
         18,
         18,
         18,
         22,
         22,
         22,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         22,
         23,
         22,
         22,
         22,
         23,
         23,
         22,
         23,
         23,
         23,
         22,
         23,
         22,
         22,
         22,
         23,
         22,
         23,
         22,
         22,
         22,
         22,
         26,
         26,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         26,
         26,
         26,
         26,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         0,
         0,
         0,
         0,
         53,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         31,
         31,
         31,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         54,
         55,
         0,
         0,
         0,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         56,
         57,
         35,
         35,
         35,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         58,
         59,
         60,
         0,
         0,
         0,
         0,
         41,
         41,
         41,
         41,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         61,
         61,
         61,
         61,
         61,
         3,
         3,
         3,
         3,
         3,
         62,
         10,
         62,
         10,
         62,
         10,
         11,
         11,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         12,
         12,
         27,
         27,
         27,
         27,
         27,
         26,
         26,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         50,
         50,
         50,
         50,
         50,
         6,
         6,
         6,
         6,
         63,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         63,
         63,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         8,
         8,
         17,
         17,
         17,
         17,
         17,
         17,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         19,
         19,
         19,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         23,
         23,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         28,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         57,
         57,
         55,
         53,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         64,
         6,
         6,
         65,
         66,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         35,
         35,
         17,
         17,
         17,
         67,
         3,
         68,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         27,
         0,
         19,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         19,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         3,
         11,
         5,
         0,
         0,
         69,
         70,
         71,
         72,
         6,
         6,
         3,
         11,
         5,
         73,
         74,
         75,
         76,
         6,
         6,
         3,
         11,
         5,
         77,
         78,
         79,
         80,
         6,
         6,
      },
      
      .num_filtered_states = 81,
      .table = (const uint16_t []) {
      
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         55,
         55,
         55,
         55,
         1176,
         1176,
         1177,
         1178,
         1179,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1180,
         1181,
         1181,
         1181,
         1181,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1182,
         55,
         55,
         1183,
         55,
         1184,
         55,
         1185,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1177,
         1178,
         55,
         55,
         2442,
         2443,
         2444,
         2445,
         55,
         55,
         55,
         55,
         55,
         1177,
         1177,
         1176,
         1176,
         55,
         55,
         1176,
         1176,
         55,
         55,
         1176,
         1176,
         55,
         55,
         1176,
         1176,
         55,
         55,
         54,
         55,
         54,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         54,
         1167,
         1167,
         1186,
         54,
         1167,
         1167,
         1186,
         54,
         1167,
         1167,
         1186,
         54,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         1186,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1168,
         1177,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         1200,
         1200,
         1200,
         1200,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1168,
         1178,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         1200,
         1200,
         1200,
         1200,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1169,
         1179,
         1169,
         1169,
         1190,
         1190,
         1198,
         1198,
         1205,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1206,
         1207,
         1207,
         1207,
         1207,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1208,
         1169,
         1209,
         1169,
         1169,
         1210,
         1169,
         1211,
         1169,
         1212,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1169,
         1198,
         1198,
         1169,
         1169,
         2455,
         2456,
         2457,
         2458,
         1169,
         1169,
         1169,
         1169,
         1169,
         1198,
         1198,
         1190,
         1190,
         1169,
         1169,
         1190,
         1190,
         1169,
         1169,
         1190,
         1190,
         1169,
         1169,
         1190,
         1190,
         1169,
         1169,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         1213,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         1214,
         1167,
         1167,
         54,
         1214,
         1167,
         1167,
         54,
         1214,
         1167,
         1167,
         54,
         1214,
         54,
         55,
         54,
         1187,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         1214,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         1187,
         1168,
         1168,
         1214,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         1214,
         1168,
         1168,
         1167,
         1167,
         1187,
         1214,
         1167,
         1167,
         1187,
         1214,
         1167,
         1167,
         1187,
         1214,
         1167,
         1167,
         1187,
         1214,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         1215,
         1215,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         1216,
         1216,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         1215,
         1217,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         1216,
         1216,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1197,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         1218,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1197,
         1197,
         54,
         54,
         1197,
         1197,
         54,
         54,
         1197,
         1197,
         54,
         54,
         1197,
         1197,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1219,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         1219,
         1219,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1170,
         1180,
         1170,
         1170,
         1191,
         1191,
         1199,
         1199,
         1206,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1220,
         1221,
         1221,
         1221,
         1221,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1222,
         1170,
         1170,
         1223,
         1170,
         1224,
         1170,
         1225,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1170,
         1199,
         1199,
         1170,
         1170,
         2460,
         2461,
         2462,
         2463,
         1170,
         1170,
         1170,
         1170,
         1170,
         1199,
         1199,
         1191,
         1191,
         1170,
         1170,
         1191,
         1191,
         1170,
         1170,
         1191,
         1191,
         1170,
         1170,
         1191,
         1191,
         1170,
         1170,
         1171,
         1181,
         1171,
         1171,
         1192,
         1192,
         1200,
         1200,
         1207,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1221,
         1226,
         1226,
         1226,
         1226,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1227,
         1171,
         1171,
         1228,
         1171,
         1229,
         1171,
         1230,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         2464,
         2464,
         1171,
         1171,
         2465,
         2466,
         2467,
         2468,
         1171,
         1171,
         1171,
         1171,
         1171,
         1200,
         1200,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1171,
         1181,
         1171,
         1171,
         1192,
         1192,
         1200,
         1200,
         1207,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1221,
         1226,
         1231,
         1232,
         1233,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1227,
         1171,
         1171,
         1228,
         1171,
         1229,
         1171,
         1230,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         2464,
         2464,
         1171,
         1171,
         2465,
         2466,
         2467,
         2468,
         1171,
         1171,
         1171,
         1171,
         1171,
         1200,
         1200,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1171,
         1181,
         1171,
         1171,
         1192,
         1192,
         1200,
         1200,
         1207,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1221,
         1226,
         1232,
         1234,
         1235,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1227,
         1171,
         1171,
         1228,
         1171,
         1229,
         1171,
         1230,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         2464,
         2464,
         1171,
         1171,
         2465,
         2466,
         2467,
         2468,
         1171,
         1171,
         1171,
         1171,
         1171,
         1200,
         1200,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1171,
         1181,
         1171,
         1171,
         1192,
         1192,
         1200,
         1200,
         1207,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1221,
         1226,
         1233,
         1235,
         1236,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1227,
         1171,
         1171,
         1228,
         1171,
         1229,
         1171,
         1230,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         1171,
         2464,
         2464,
         1171,
         1171,
         2465,
         2466,
         2467,
         2468,
         1171,
         1171,
         1171,
         1171,
         1171,
         1200,
         1200,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         1192,
         1192,
         1171,
         1171,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         1237,
         1237,
         1237,
         1237,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         1237,
         1238,
         1239,
         1240,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         1237,
         1239,
         1241,
         1242,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         1237,
         1240,
         1242,
         1243,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         1216,
         1216,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         1244,
         1244,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         1216,
         1216,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         1244,
         1245,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         1246,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1247,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1248,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1249,
         1249,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1208,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1249,
         1249,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1250,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1172,
         1182,
         1172,
         1172,
         1193,
         1193,
         1201,
         1201,
         1209,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1222,
         1227,
         1227,
         1227,
         1227,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1250,
         1250,
         1172,
         1172,
         1251,
         1172,
         1252,
         1172,
         1253,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1172,
         1201,
         1201,
         1172,
         1172,
         2469,
         2470,
         2471,
         2472,
         1172,
         1172,
         1172,
         1172,
         1172,
         1201,
         1201,
         1193,
         1193,
         1172,
         1172,
         1193,
         1193,
         1172,
         1172,
         1193,
         1193,
         1172,
         1172,
         1193,
         1193,
         1172,
         1172,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         2473,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1254,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1173,
         1183,
         1173,
         1173,
         1194,
         1194,
         1202,
         1202,
         1210,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1223,
         1228,
         1228,
         1228,
         1228,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1251,
         1173,
         1254,
         1254,
         1173,
         1255,
         1173,
         1256,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1173,
         1202,
         1202,
         1173,
         1173,
         2474,
         2475,
         2476,
         2477,
         1173,
         1173,
         1173,
         1173,
         1173,
         1202,
         1202,
         1194,
         1194,
         1173,
         1173,
         1194,
         1194,
         1173,
         1173,
         1194,
         1194,
         1173,
         1173,
         1194,
         1194,
         1173,
         1173,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1257,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1174,
         1184,
         1174,
         1174,
         1195,
         1195,
         1203,
         1203,
         1211,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1224,
         1229,
         1229,
         1229,
         1229,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1252,
         1174,
         1174,
         1255,
         1257,
         1257,
         1174,
         1258,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1174,
         1203,
         1203,
         1174,
         1174,
         2478,
         2479,
         2480,
         2481,
         1174,
         1174,
         1174,
         1174,
         1174,
         1203,
         1203,
         1195,
         1195,
         1174,
         1174,
         1195,
         1195,
         1174,
         1174,
         1195,
         1195,
         1174,
         1174,
         1195,
         1195,
         1174,
         1174,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1259,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1175,
         1185,
         1175,
         1175,
         1196,
         1196,
         1204,
         1204,
         1212,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1225,
         1230,
         1230,
         1230,
         1230,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1253,
         1175,
         1175,
         1256,
         1175,
         1258,
         1259,
         1259,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1175,
         1204,
         1204,
         1175,
         1175,
         2482,
         2483,
         2484,
         2485,
         1175,
         1175,
         1175,
         1175,
         1175,
         1204,
         1204,
         1196,
         1196,
         1175,
         1175,
         1196,
         1196,
         1175,
         1175,
         1196,
         1196,
         1175,
         1175,
         1196,
         1196,
         1175,
         1175,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         1260,
         1260,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         1260,
         54,
         1260,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         1260,
         1260,
         1260,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         1261,
         1261,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         1261,
         54,
         1261,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         1261,
         1261,
         1261,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         2459,
         2459,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1168,
         1168,
         2486,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         2837,
         1168,
         1168,
         1167,
         1167,
         1186,
         2459,
         1167,
         1167,
         1186,
         2459,
         1167,
         1167,
         1186,
         2459,
         1167,
         1167,
         1186,
         2459,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         2473,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2487,
         2488,
         54,
         54,
         54,
         1186,
         2835,
         2838,
         2961,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1167,
         1167,
         1186,
         1187,
         1168,
         1177,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         2464,
         2464,
         2464,
         2464,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1168,
         1178,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         2464,
         2464,
         2464,
         2464,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         54,
         55,
         2446,
         2446,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2486,
         2446,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2839,
         54,
         1168,
         1168,
         1167,
         1167,
         2446,
         1214,
         1167,
         1167,
         2446,
         1214,
         1167,
         1167,
         2446,
         1214,
         1167,
         1167,
         2446,
         1214,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         2489,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         2438,
         2442,
         2438,
         2438,
         2447,
         2447,
         2451,
         2451,
         2455,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2460,
         2465,
         2465,
         2465,
         2465,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2469,
         2438,
         2438,
         2474,
         2438,
         2478,
         2438,
         2482,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2438,
         2451,
         2451,
         2438,
         2438,
         2438,
         2490,
         2491,
         2492,
         2438,
         2438,
         2438,
         2438,
         2438,
         2451,
         2451,
         2447,
         2447,
         2438,
         2438,
         2447,
         2447,
         2438,
         2438,
         2447,
         2447,
         2438,
         2438,
         2447,
         2447,
         2438,
         2438,
         2439,
         2443,
         2439,
         2439,
         2448,
         2448,
         2452,
         2452,
         2456,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2461,
         2466,
         2466,
         2466,
         2466,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2470,
         2439,
         2439,
         2475,
         2439,
         2479,
         2439,
         2483,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2439,
         2452,
         2452,
         2439,
         2439,
         2490,
         2439,
         2493,
         2494,
         2439,
         2439,
         2439,
         2439,
         2439,
         2452,
         2452,
         2448,
         2448,
         2439,
         2439,
         2448,
         2448,
         2439,
         2439,
         2448,
         2448,
         2439,
         2439,
         2448,
         2448,
         2439,
         2439,
         2440,
         2444,
         2440,
         2440,
         2449,
         2449,
         2453,
         2453,
         2457,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2462,
         2467,
         2467,
         2467,
         2467,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2471,
         2440,
         2440,
         2476,
         2440,
         2480,
         2440,
         2484,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2440,
         2487,
         2453,
         2453,
         2440,
         2440,
         2491,
         2493,
         2440,
         2495,
         2440,
         2440,
         2440,
         2440,
         2440,
         2453,
         2453,
         2449,
         2449,
         2440,
         2440,
         2449,
         2449,
         2440,
         2440,
         2449,
         2449,
         2440,
         2440,
         2449,
         2449,
         2440,
         2440,
         2441,
         2445,
         2441,
         2441,
         2450,
         2450,
         2454,
         2454,
         2458,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2463,
         2468,
         2468,
         2468,
         2468,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2472,
         2441,
         2441,
         2477,
         2441,
         2481,
         2441,
         2485,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2441,
         2488,
         2454,
         2454,
         2441,
         2441,
         2492,
         2494,
         2495,
         2496,
         2441,
         2441,
         2441,
         2441,
         2441,
         2454,
         2454,
         2450,
         2450,
         2441,
         2441,
         2450,
         2450,
         2441,
         2441,
         2450,
         2450,
         2441,
         2441,
         2450,
         2450,
         2441,
         2441,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         2497,
         2497,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         2497,
         54,
         2497,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         54,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         2497,
         2497,
         2497,
         54,
         54,
         1168,
         1168,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         1167,
         1167,
         54,
         54,
         54,
         55,
         54,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         2836,
         2836,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1168,
         1168,
         2839,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         54,
         2840,
         1168,
         1168,
         1167,
         1167,
         1186,
         2836,
         1167,
         1167,
         1186,
         2836,
         1167,
         1167,
         1186,
         2836,
         1167,
         1167,
         1186,
         2836,
         54,
         55,
         2835,
         2835,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2837,
         2835,
         1168,
         1168,
         54,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2840,
         54,
         1168,
         1168,
         1167,
         1167,
         2835,
         1214,
         1167,
         1167,
         2835,
         1214,
         1167,
         1167,
         2835,
         1214,
         1167,
         1167,
         2835,
         1214,
         1168,
         1177,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         1200,
         1200,
         1200,
         1200,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2838,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1168,
         1177,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1198,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1199,
         1200,
         1200,
         1200,
         1200,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1201,
         1168,
         1168,
         1202,
         1168,
         1203,
         1168,
         1204,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         2961,
         1168,
         1168,
         1168,
         1168,
         2451,
         2452,
         2453,
         2454,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1189,
         1189,
         1168,
         1168,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         2962,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         2962,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         1186,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2963,
         54,
         55,
         54,
         1187,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         1214,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         1187,
         1168,
         1168,
         1214,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         1214,
         1168,
         1168,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2963,
         1214,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         2995,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         2995,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         1186,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2996,
         1167,
         1167,
         1186,
         2996,
         1167,
         1167,
         1186,
         2996,
         54,
         55,
         54,
         1187,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         1214,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         1187,
         1168,
         1168,
         1214,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         1214,
         1168,
         1168,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2996,
         1214,
         1167,
         1167,
         2996,
         1214,
         1167,
         1167,
         2996,
         1214,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         3004,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         3004,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         1186,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2996,
         1167,
         1167,
         1186,
         3005,
         1167,
         1167,
         1186,
         3005,
         54,
         55,
         54,
         1187,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         1214,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         1187,
         1168,
         1168,
         1214,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         1214,
         1168,
         1168,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2996,
         1214,
         1167,
         1167,
         3005,
         1214,
         1167,
         1167,
         3005,
         1214,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         3013,
         1167,
         1167,
         1167,
         1176,
         1167,
         1167,
         1188,
         1188,
         1189,
         1189,
         1190,
         1167,
         1167,
         1167,
         1167,
         1167,
         1197,
         1167,
         1167,
         1191,
         1192,
         1192,
         1192,
         1192,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1193,
         1167,
         1167,
         1194,
         1167,
         1195,
         1167,
         1196,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1167,
         1167,
         2447,
         2448,
         2449,
         2450,
         1167,
         1167,
         1167,
         1167,
         1167,
         1189,
         1189,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         1188,
         1188,
         1167,
         1167,
         3013,
         1188,
         1167,
         1167,
         54,
         55,
         1186,
         1186,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         54,
         1187,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         1186,
         1186,
         1168,
         1168,
         2446,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         1186,
         2835,
         1168,
         1168,
         1167,
         1167,
         1186,
         2963,
         1167,
         1167,
         1186,
         2996,
         1167,
         1167,
         1186,
         3005,
         1167,
         1167,
         1186,
         3014,
         54,
         55,
         54,
         1187,
         1167,
         1167,
         1168,
         1168,
         1169,
         54,
         1214,
         1214,
         54,
         54,
         54,
         54,
         54,
         1170,
         1171,
         1171,
         1171,
         1171,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1172,
         54,
         54,
         1173,
         54,
         1174,
         54,
         1175,
         54,
         54,
         54,
         54,
         54,
         54,
         2459,
         1187,
         1168,
         1168,
         1214,
         54,
         2438,
         2439,
         2440,
         2441,
         54,
         54,
         54,
         2836,
         1214,
         1168,
         1168,
         1167,
         1167,
         2963,
         1214,
         1167,
         1167,
         2996,
         1214,
         1167,
         1167,
         3005,
         1214,
         1167,
         1167,
         3014,
         1214,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         14,
         0,
         15,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         19,
         20,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         24,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         25,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         11,
         24,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         5,
         0,
         0,
         5,
         5,
         0,
         6,
         0,
         0,
         0,
         6,
         5,
         5,
         5,
         0,
         6,
         0,
         0,
         0,
         6,
         5,
         5,
         5,
         0,
         6,
         0,
         0,
      },
      
      .num_filtered_states = 26,
      .table = (const uint16_t []) {
      
         56,
         57,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         57,
         57,
         1679,
         1680,
         1681,
         1682,
         1683,
         1684,
         1685,
         1686,
         1687,
         1688,
         1689,
         1690,
         1691,
         1692,
         1693,
         1694,
         1695,
         1696,
         1697,
         1698,
         1699,
         2498,
         2499,
         2886,
         56,
         1679,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1680,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         1677,
         1681,
         1677,
         1677,
         1677,
         1700,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1677,
         1700,
         1677,
         1700,
         1678,
         1682,
         1678,
         1678,
         1700,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         56,
         1683,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1684,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1685,
         56,
         56,
         1677,
         1678,
         56,
         56,
         1701,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1686,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1687,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1688,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         1702,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1689,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         1703,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1690,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1691,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1692,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1693,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1694,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1695,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1696,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1697,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1698,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         56,
         1699,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         1678,
         2498,
         1678,
         1678,
         1700,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         56,
         2499,
         56,
         56,
         1677,
         1678,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         56,
         1678,
         56,
         1678,
         1678,
         2886,
         1678,
         1678,
         1700,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
         1678,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         14,
         15,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         19,
         20,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         23,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         0,
         24,
         0,
         24,
         0,
         24,
         24,
         0,
         24,
         24,
         25,
         0,
         0,
         25,
         0,
         0,
         25,
         0,
         25,
         25,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         11,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         24,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         6,
         0,
         0,
         6,
         6,
         0,
         7,
         0,
         0,
         0,
         7,
         6,
         6,
         6,
         0,
         7,
         0,
         0,
         0,
         7,
         6,
         6,
         6,
         0,
         7,
         0,
         0,
      },
      
      .num_filtered_states = 26,
      .table = (const uint16_t []) {
      
         58,
         59,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         59,
         59,
         1579,
         1580,
         1581,
         1582,
         1583,
         1584,
         1585,
         1586,
         1587,
         1588,
         1589,
         1590,
         1591,
         1592,
         1593,
         1594,
         1595,
         1596,
         1597,
         1598,
         1599,
         2721,
         2951,
         2952,
         58,
         1579,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1580,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1581,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1577,
         1582,
         1577,
         1577,
         1577,
         1577,
         1600,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1577,
         1578,
         1583,
         1578,
         1578,
         1578,
         1600,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         1578,
         58,
         1584,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1585,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1586,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         1601,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1587,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1588,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         1602,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1589,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         1603,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1590,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1591,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1592,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1593,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1594,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1595,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1596,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1597,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1598,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         1599,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         2721,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         2722,
         58,
         58,
         58,
         2951,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         2952,
         58,
         58,
         58,
         1577,
         1578,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
         58,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         9,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         12,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         2,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         4,
         14,
         0,
         0,
         3,
         3,
         9,
         4,
         15,
         10,
         2,
         4,
         16,
         3,
         3,
         9,
         4,
         17,
         10,
         2,
         4,
         18,
         3,
         3,
         9,
         4,
         0,
         10,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         60,
         61,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         62,
         63,
         62,
         1292,
         62,
         1293,
         62,
         62,
         2343,
         2344,
         62,
         62,
         1292,
         62,
         1292,
         62,
         1292,
         62,
         1292,
         1294,
         1295,
         1294,
         1296,
         1294,
         1297,
         1294,
         1294,
         2345,
         2346,
         1294,
         1294,
         1296,
         1294,
         1296,
         1294,
         1296,
         1294,
         1296,
         60,
         61,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1298,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         61,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1299,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1300,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         61,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         1294,
         1295,
         1294,
         1296,
         1294,
         1297,
         1294,
         1294,
         2345,
         2346,
         1294,
         1294,
         1296,
         1294,
         1296,
         1294,
         1296,
         1294,
         1296,
         60,
         2347,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         2348,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         2885,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         2987,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         2994,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         2998,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         3003,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         3007,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         3012,
         60,
         1290,
         60,
         1291,
         60,
         60,
         2341,
         2342,
         60,
         60,
         1290,
         60,
         1290,
         60,
         1290,
         60,
         1290,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         9,
         2,
         2,
         10,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         64,
         1463,
         1464,
         1465,
         1466,
         2631,
         2632,
         2633,
         2634,
         2635,
         2636,
      },
   },
   [nir_op_umul_unorm_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         65,
         65,
         65,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         66,
         67,
         66,
         66,
         1439,
         68,
         69,
         68,
         68,
         1440,
         1441,
         1442,
         1441,
         1441,
         1443,
         1444,
         1445,
         1444,
         1444,
         1446,
         66,
         67,
         66,
         66,
         1439,
         70,
         71,
         70,
         70,
         1447,
         72,
         73,
         72,
         72,
         1448,
         1449,
         1450,
         1449,
         1449,
         1451,
         1452,
         1453,
         1452,
         1452,
         1454,
         70,
         71,
         70,
         70,
         1447,
         66,
         67,
         66,
         66,
         1439,
         68,
         69,
         68,
         68,
         1440,
         1455,
         1456,
         1455,
         1455,
         1457,
         1444,
         1445,
         1444,
         1444,
         1446,
         66,
         67,
         66,
         66,
         1439,
         66,
         67,
         66,
         66,
         1439,
         68,
         69,
         68,
         68,
         1440,
         1441,
         1442,
         1441,
         1441,
         1443,
         1458,
         1459,
         1458,
         1458,
         1460,
         66,
         67,
         66,
         66,
         1439,
         66,
         67,
         66,
         66,
         1439,
         68,
         69,
         68,
         68,
         1440,
         1441,
         1442,
         1441,
         1441,
         1443,
         1444,
         1445,
         1444,
         1444,
         1446,
         66,
         67,
         66,
         66,
         1439,
      },
   },
   [nir_op_sdot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         74,
         75,
         76,
         77,
         76,
         77,
         78,
         79,
      },
   },
   [nir_op_udot_4x8_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         80,
         81,
         82,
         83,
         82,
         83,
         84,
         85,
      },
   },
   [nir_op_sdot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         86,
         86,
         87,
         87,
         87,
         87,
         88,
         88,
      },
   },
   [nir_op_udot_4x8_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         89,
         90,
         91,
         92,
         91,
         92,
         93,
         94,
      },
   },
   [nir_op_sdot_2x16_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         95,
         96,
         97,
         98,
         97,
         98,
         99,
         100,
      },
   },
   [nir_op_udot_2x16_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         101,
         102,
         103,
         104,
         103,
         104,
         105,
         106,
      },
   },
   [nir_op_sdot_2x16_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         107,
         107,
         108,
         108,
         108,
         108,
         109,
         109,
      },
   },
   [nir_op_udot_2x16_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         110,
         110,
         111,
         111,
         111,
         111,
         112,
         112,
      },
   },
   [nir_op_sudot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         113,
         114,
         115,
         116,
         117,
         118,
         119,
         120,
      },
   },
   [nir_op_sudot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         121,
         121,
         122,
         122,
         123,
         123,
         124,
         124,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         10,
         0,
         11,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         19,
         20,
         21,
         20,
         21,
         22,
         22,
         22,
         22,
         0,
         0,
         23,
         24,
         25,
         25,
         25,
         25,
         26,
         26,
         27,
         27,
         28,
         28,
         0,
         29,
         29,
         30,
         30,
         31,
         31,
         32,
         32,
         33,
         33,
         0,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         0,
         36,
         0,
         37,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         40,
         40,
         40,
         40,
         41,
         42,
         43,
         0,
         0,
         44,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         46,
         0,
         46,
         0,
         0,
         0,
         0,
         0,
         0,
         47,
         48,
         49,
         50,
         51,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         54,
         55,
         56,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         57,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         57,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         58,
         7,
         7,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         18,
         18,
         18,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         35,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         59,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         21,
         20,
         20,
         20,
         21,
         21,
         20,
         21,
         21,
         20,
         21,
         20,
         20,
         20,
         21,
         20,
         21,
         20,
         20,
         20,
         20,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         60,
         61,
         62,
         63,
         64,
         65,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         66,
         0,
         0,
         0,
         0,
         0,
         67,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         68,
         69,
         70,
         71,
         72,
         73,
         0,
         0,
         0,
         0,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         74,
         74,
         74,
         74,
         74,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         75,
         75,
         75,
         75,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         76,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         23,
         23,
         23,
         23,
         23,
         24,
         23,
         24,
         23,
         23,
         24,
         23,
         24,
         23,
         23,
         24,
         24,
         23,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         80,
         81,
         82,
         83,
         84,
         0,
         39,
         39,
         39,
         39,
         39,
         41,
         42,
         42,
         42,
         42,
         43,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         33,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         17,
         48,
         37,
         37,
         37,
         37,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         0,
         0,
         0,
         0,
         0,
         0,
         47,
         51,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         85,
         57,
         86,
         87,
         88,
         89,
         7,
         7,
         7,
         7,
         57,
         57,
         57,
         57,
         57,
         57,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         15,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         11,
         11,
         19,
         19,
         19,
         19,
         19,
         19,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         59,
         59,
         6,
         6,
         6,
         20,
         21,
         21,
         21,
         21,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         22,
         22,
         22,
         22,
         40,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         0,
         0,
         0,
         0,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         12,
         0,
         24,
         25,
         25,
         25,
         25,
         74,
         25,
         25,
         25,
         26,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         0,
         34,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         79,
         79,
         77,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         10,
         10,
         10,
         10,
         10,
         10,
         9,
         9,
         90,
         91,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         19,
         19,
         19,
         0,
         0,
         0,
         11,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         16,
         21,
         22,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         12,
         12,
         57,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         14,
         14,
         21,
         21,
         21,
         29,
         29,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         92,
         93,
         0,
         0,
         9,
         14,
         14,
         9,
         9,
         0,
         0,
         10,
         10,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         10,
         10,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         10,
         10,
      },
      
      .num_filtered_states = 94,
      .table = (const uint16_t []) {
      
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         131,
         132,
         131,
         131,
         131,
         131,
         542,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         543,
         131,
         131,
         544,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         131,
         129,
         130,
         545,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         546,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         547,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         548,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         549,
         550,
         549,
         549,
         549,
         549,
         551,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         552,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         549,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         554,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         555,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         555,
         555,
         555,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         556,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         557,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         558,
         559,
         558,
         558,
         558,
         558,
         560,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         561,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         558,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         562,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         562,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         563,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         564,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         565,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         566,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         567,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         568,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         569,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         570,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         571,
         572,
         571,
         571,
         571,
         571,
         573,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         574,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         571,
         575,
         576,
         575,
         575,
         575,
         575,
         577,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         578,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         575,
         579,
         580,
         579,
         579,
         579,
         579,
         581,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         582,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         579,
         583,
         584,
         583,
         583,
         583,
         583,
         585,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         586,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         583,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         587,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         588,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         589,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         590,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         591,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         592,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         593,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         594,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         595,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         596,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         597,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         598,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         599,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         562,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         562,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         2150,
         2151,
         2150,
         2150,
         2150,
         2150,
         2152,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2153,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2150,
         2154,
         2155,
         2154,
         2154,
         2154,
         2154,
         2156,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2157,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2154,
         2158,
         2159,
         2158,
         2158,
         2158,
         2158,
         2160,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2161,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2158,
         2162,
         2163,
         2162,
         2162,
         2162,
         2162,
         2164,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2165,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         2162,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         553,
         553,
         553,
         553,
         553,
         129,
         129,
         129,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         555,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         555,
         555,
         555,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         555,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         555,
         555,
         555,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         555,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         555,
         555,
         555,
         129,
         129,
         130,
         129,
         129,
         129,
         129,
         540,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         541,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2797,
         127,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         605,
         606,
         605,
         605,
         605,
         605,
         607,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         608,
         605,
         605,
         609,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         2166,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         605,
         2989,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         614,
         615,
         614,
         614,
         614,
         614,
         616,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         617,
         614,
         614,
         618,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         614,
         610,
         611,
         619,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         620,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         621,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         622,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         623,
         624,
         623,
         623,
         623,
         623,
         625,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         626,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         623,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         628,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         629,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         629,
         629,
         629,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         630,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         631,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         632,
         633,
         632,
         632,
         632,
         632,
         634,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         635,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         632,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         636,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         636,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         637,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         638,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         639,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         640,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         641,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         642,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         643,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         644,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         645,
         646,
         645,
         645,
         645,
         645,
         647,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         648,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         649,
         650,
         649,
         649,
         649,
         649,
         651,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         652,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         649,
         653,
         654,
         653,
         653,
         653,
         653,
         655,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         656,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         653,
         657,
         658,
         657,
         657,
         657,
         657,
         659,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         660,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         657,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         661,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         662,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         663,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         664,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         665,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         666,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         667,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         668,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         669,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         670,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         671,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         672,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         673,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         636,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         636,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         2167,
         2168,
         2167,
         2167,
         2167,
         2167,
         2169,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2170,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2167,
         2171,
         2172,
         2171,
         2171,
         2171,
         2171,
         2173,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2174,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2175,
         2176,
         2175,
         2175,
         2175,
         2175,
         2177,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2178,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2179,
         2180,
         2179,
         2179,
         2179,
         2179,
         2181,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2182,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         627,
         627,
         627,
         627,
         627,
         610,
         610,
         610,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         629,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         629,
         629,
         629,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         629,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         629,
         629,
         629,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         629,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         629,
         629,
         629,
         610,
         610,
         611,
         610,
         610,
         610,
         610,
         612,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         613,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         610,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         678,
         679,
         678,
         678,
         678,
         678,
         680,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         681,
         678,
         678,
         682,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         674,
         675,
         683,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         684,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         685,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         686,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         687,
         688,
         687,
         687,
         687,
         687,
         689,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         690,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         692,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         694,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         695,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         696,
         697,
         696,
         696,
         696,
         696,
         698,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         699,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         701,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         702,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         703,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         704,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         705,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         706,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         707,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         708,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         709,
         710,
         709,
         709,
         709,
         709,
         711,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         712,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         713,
         714,
         713,
         713,
         713,
         713,
         715,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         716,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         717,
         718,
         717,
         717,
         717,
         717,
         719,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         720,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         721,
         722,
         721,
         721,
         721,
         721,
         723,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         724,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         725,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         726,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         727,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         728,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         729,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         730,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         731,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         732,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         733,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         734,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         735,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         736,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         737,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         2183,
         2184,
         2183,
         2183,
         2183,
         2183,
         2185,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2186,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2187,
         2188,
         2187,
         2187,
         2187,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2191,
         2191,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2195,
         2195,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         678,
         679,
         678,
         678,
         678,
         678,
         680,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         681,
         678,
         678,
         682,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         738,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         678,
         674,
         675,
         683,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         684,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         685,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         686,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         739,
         740,
         739,
         739,
         739,
         739,
         741,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         742,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         739,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         692,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         694,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         695,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         696,
         697,
         696,
         696,
         696,
         696,
         698,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         699,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         701,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         702,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         703,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         704,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         705,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         706,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         707,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         708,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         709,
         710,
         709,
         709,
         709,
         709,
         711,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         712,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         709,
         713,
         714,
         713,
         713,
         713,
         713,
         715,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         716,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         717,
         718,
         717,
         717,
         717,
         717,
         719,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         720,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         717,
         721,
         722,
         721,
         721,
         721,
         721,
         723,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         724,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         721,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         725,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         726,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         727,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         728,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         729,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         730,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         731,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         732,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         733,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         734,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         735,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         736,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         737,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         700,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         2183,
         2184,
         2183,
         2183,
         2183,
         2183,
         2185,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2186,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2187,
         2188,
         2187,
         2187,
         2187,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2191,
         2191,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2195,
         2195,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         691,
         691,
         691,
         691,
         691,
         674,
         674,
         674,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         693,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         693,
         693,
         693,
         674,
         674,
         675,
         674,
         674,
         674,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         747,
         748,
         747,
         747,
         747,
         747,
         749,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         750,
         747,
         747,
         751,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         747,
         743,
         744,
         752,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         753,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         754,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         755,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         756,
         757,
         756,
         756,
         756,
         756,
         758,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         759,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         756,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         761,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         762,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         762,
         762,
         762,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         763,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         764,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         765,
         766,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         768,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         769,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         769,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         770,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         771,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         772,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         773,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         774,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         775,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         776,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         777,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         778,
         779,
         778,
         778,
         778,
         778,
         780,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         781,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         778,
         782,
         783,
         782,
         782,
         782,
         782,
         784,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         785,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         786,
         787,
         786,
         786,
         786,
         786,
         788,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         789,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         786,
         790,
         791,
         790,
         790,
         790,
         790,
         792,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         793,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         790,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         794,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         795,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         796,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         797,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         798,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         799,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         800,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         801,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         802,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         803,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         804,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         805,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         806,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         769,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         769,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         2199,
         2200,
         2199,
         2199,
         2199,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2203,
         2204,
         2203,
         2203,
         2203,
         2203,
         2205,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2206,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2203,
         2207,
         2208,
         2207,
         2207,
         2207,
         2207,
         2209,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2210,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2207,
         2211,
         2212,
         2211,
         2211,
         2211,
         2211,
         2213,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2214,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         2211,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         760,
         760,
         760,
         760,
         760,
         743,
         743,
         743,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         762,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         762,
         762,
         762,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         762,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         762,
         762,
         762,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         762,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         762,
         762,
         762,
         743,
         743,
         744,
         743,
         743,
         743,
         743,
         745,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         746,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         743,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         811,
         812,
         811,
         811,
         811,
         811,
         813,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         814,
         811,
         811,
         815,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         807,
         808,
         816,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         817,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         818,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         819,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         820,
         821,
         820,
         820,
         820,
         820,
         822,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         823,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         825,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         827,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         828,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         829,
         830,
         829,
         829,
         829,
         829,
         831,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         832,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         834,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         835,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         836,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         837,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         838,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         839,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         840,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         841,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         842,
         843,
         842,
         842,
         842,
         842,
         844,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         845,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         846,
         847,
         846,
         846,
         846,
         846,
         848,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         849,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         850,
         851,
         850,
         850,
         850,
         850,
         852,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         853,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         854,
         855,
         854,
         854,
         854,
         854,
         856,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         857,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         858,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         859,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         860,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         861,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         862,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         863,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         864,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         865,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         866,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         867,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         868,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         869,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         870,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         2215,
         2216,
         2215,
         2215,
         2215,
         2215,
         2217,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2218,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2219,
         2220,
         2219,
         2219,
         2219,
         2219,
         2221,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2222,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2223,
         2224,
         2223,
         2223,
         2223,
         2223,
         2225,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2226,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2227,
         2228,
         2227,
         2227,
         2227,
         2227,
         2229,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2230,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         811,
         812,
         811,
         811,
         811,
         811,
         813,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         814,
         811,
         811,
         815,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         811,
         807,
         808,
         816,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         817,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         818,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         819,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         820,
         821,
         820,
         820,
         820,
         820,
         822,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         823,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         820,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         825,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         827,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         828,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         829,
         830,
         829,
         829,
         829,
         829,
         831,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         832,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         829,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         834,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         835,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         836,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         837,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         838,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         839,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         840,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         841,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         842,
         843,
         842,
         842,
         842,
         842,
         844,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         845,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         842,
         846,
         847,
         846,
         846,
         846,
         846,
         848,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         849,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         846,
         850,
         851,
         850,
         850,
         850,
         850,
         852,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         853,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         854,
         855,
         854,
         854,
         854,
         854,
         856,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         857,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         858,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         859,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         860,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         861,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         862,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         863,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         864,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         865,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         866,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         867,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         868,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         869,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         870,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         833,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         2231,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         2215,
         2216,
         2215,
         2215,
         2215,
         2215,
         2217,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2218,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2215,
         2219,
         2220,
         2219,
         2219,
         2219,
         2219,
         2221,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2222,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2219,
         2223,
         2224,
         2223,
         2223,
         2223,
         2223,
         2225,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2226,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2223,
         2227,
         2228,
         2227,
         2227,
         2227,
         2227,
         2229,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2230,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         2227,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         824,
         824,
         824,
         824,
         824,
         807,
         807,
         807,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         826,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         826,
         826,
         826,
         807,
         807,
         808,
         807,
         807,
         807,
         807,
         809,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         810,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         807,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         875,
         876,
         875,
         875,
         875,
         875,
         877,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         878,
         875,
         875,
         879,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         2232,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         2232,
         2232,
         2232,
         2232,
         875,
         875,
         875,
         875,
         871,
         872,
         880,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         881,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         882,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         883,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         884,
         885,
         884,
         884,
         884,
         884,
         886,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         887,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         889,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         891,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         892,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         893,
         894,
         893,
         893,
         893,
         893,
         895,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         896,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         898,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         899,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         900,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         901,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         902,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         903,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         904,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         905,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         906,
         907,
         906,
         906,
         906,
         906,
         908,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         909,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         910,
         911,
         910,
         910,
         910,
         910,
         912,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         913,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         914,
         915,
         914,
         914,
         914,
         914,
         916,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         917,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         918,
         919,
         918,
         918,
         918,
         918,
         920,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         921,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         922,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         923,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         924,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         925,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         926,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         927,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         928,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         929,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         930,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         931,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         932,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         933,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         934,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         2233,
         2234,
         2233,
         2233,
         2233,
         2233,
         2235,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2236,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2237,
         2238,
         2237,
         2237,
         2237,
         2237,
         2239,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2240,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2241,
         2242,
         2241,
         2241,
         2241,
         2241,
         2243,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2244,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2245,
         2246,
         2245,
         2245,
         2245,
         2245,
         2247,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2248,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1003,
         1004,
         1003,
         1003,
         1003,
         1003,
         1005,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1006,
         1003,
         1003,
         1007,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         999,
         1000,
         1008,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         1009,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         1010,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         1011,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1012,
         1013,
         1012,
         1012,
         1012,
         1012,
         1014,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1015,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         1017,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         1019,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         1020,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1021,
         1022,
         1021,
         1021,
         1021,
         1021,
         1023,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1024,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1026,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         1027,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1028,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1029,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1030,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1031,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1032,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1033,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1034,
         1035,
         1034,
         1034,
         1034,
         1034,
         1036,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1037,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1038,
         1039,
         1038,
         1038,
         1038,
         1038,
         1040,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1041,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1042,
         1043,
         1042,
         1042,
         1042,
         1042,
         1044,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1045,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1046,
         1047,
         1046,
         1046,
         1046,
         1046,
         1048,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1049,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1050,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1051,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1052,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1053,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1054,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1055,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1056,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1057,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1058,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1059,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1060,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1061,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1062,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         2265,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         2266,
         2267,
         2266,
         2266,
         2266,
         2266,
         2268,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2269,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2270,
         2271,
         2270,
         2270,
         2270,
         2270,
         2272,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2273,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2274,
         2275,
         2274,
         2274,
         2274,
         2274,
         2276,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2277,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2278,
         2279,
         2278,
         2278,
         2278,
         2278,
         2280,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2281,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2265,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2265,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2265,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2265,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2805,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2806,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2807,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2808,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2809,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2810,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2811,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2812,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         2282,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2282,
         2282,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2282,
         2282,
         2282,
         2282,
         2282,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2283,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2283,
         2283,
         2283,
         2283,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         600,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         601,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         602,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         603,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         604,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2801,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2802,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2803,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2804,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2813,
         2797,
         127,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2814,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2815,
         127,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2816,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2797,
         2817,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2818,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2797,
         127,
         2819,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2820,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2797,
         127,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2284,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         2797,
         127,
         2798,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2799,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2800,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2285,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2285,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2285,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2285,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         2285,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         875,
         876,
         875,
         875,
         875,
         875,
         877,
         2286,
         875,
         875,
         875,
         875,
         875,
         875,
         878,
         875,
         875,
         879,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         2287,
         2286,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         2286,
         2287,
         2287,
         2287,
         2287,
         875,
         875,
         875,
         875,
         871,
         872,
         880,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         881,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         882,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         883,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         884,
         885,
         884,
         884,
         884,
         884,
         886,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         887,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         884,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         889,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         891,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         892,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         893,
         894,
         893,
         893,
         893,
         893,
         895,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         896,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         893,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         898,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         899,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         900,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         901,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         902,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         903,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         904,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         905,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         906,
         907,
         906,
         906,
         906,
         906,
         908,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         909,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         906,
         910,
         911,
         910,
         910,
         910,
         910,
         912,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         913,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         910,
         914,
         915,
         914,
         914,
         914,
         914,
         916,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         917,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         914,
         918,
         919,
         918,
         918,
         918,
         918,
         920,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         921,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         918,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         922,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         923,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         924,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         925,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         926,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         927,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         928,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         929,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         930,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         931,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         932,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         933,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         934,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         2233,
         2234,
         2233,
         2233,
         2233,
         2233,
         2235,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2236,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2233,
         2237,
         2238,
         2237,
         2237,
         2237,
         2237,
         2239,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2240,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2237,
         2241,
         2242,
         2241,
         2241,
         2241,
         2241,
         2243,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2244,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2241,
         2245,
         2246,
         2245,
         2245,
         2245,
         2245,
         2247,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2248,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         2245,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         888,
         888,
         888,
         888,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         890,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         890,
         890,
         890,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1003,
         1004,
         1003,
         1003,
         1003,
         1003,
         1005,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1006,
         1003,
         1003,
         1007,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         1003,
         999,
         1000,
         1008,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         1009,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         1010,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         1011,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1012,
         1013,
         1012,
         1012,
         1012,
         1012,
         1014,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1015,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         1012,
         999,
         2288,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         1017,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         1019,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         1020,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1021,
         1022,
         1021,
         1021,
         1021,
         1021,
         1023,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1024,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         1021,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1026,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         1027,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1028,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1029,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1030,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1031,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1032,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1033,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1034,
         1035,
         1034,
         1034,
         1034,
         1034,
         1036,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1037,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1034,
         1038,
         1039,
         1038,
         1038,
         1038,
         1038,
         1040,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1041,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1038,
         1042,
         1043,
         1042,
         1042,
         1042,
         1042,
         1044,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1045,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1042,
         1046,
         1047,
         1046,
         1046,
         1046,
         1046,
         1048,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1049,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1050,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1051,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1052,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1053,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1054,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1055,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1056,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1057,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1058,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1059,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1060,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1061,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1062,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         2289,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2288,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1025,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         2266,
         2267,
         2266,
         2266,
         2266,
         2266,
         2268,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2269,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2266,
         2270,
         2271,
         2270,
         2270,
         2270,
         2270,
         2272,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2273,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2270,
         2274,
         2275,
         2274,
         2274,
         2274,
         2274,
         2276,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2277,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2274,
         2278,
         2279,
         2278,
         2278,
         2278,
         2278,
         2280,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2281,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         2278,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         2288,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2289,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2289,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2289,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         2289,
         999,
         999,
         999,
         999,
         1001,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1016,
         1016,
         1016,
         1016,
         1016,
         999,
         999,
         999,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         1018,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1018,
         1018,
         1018,
         999,
         999,
         1000,
         999,
         999,
         999,
         999,
         1001,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         1002,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         999,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2290,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2291,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2292,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2293,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         939,
         940,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         942,
         939,
         939,
         943,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         935,
         936,
         944,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         945,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         946,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         947,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         948,
         949,
         948,
         948,
         948,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         953,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         955,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         956,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         957,
         958,
         957,
         957,
         957,
         957,
         959,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         960,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         957,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         962,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         963,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         964,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         965,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         966,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         967,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         968,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         969,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         970,
         971,
         970,
         970,
         970,
         970,
         972,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         973,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         970,
         974,
         975,
         974,
         974,
         974,
         974,
         976,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         977,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         974,
         978,
         979,
         978,
         978,
         978,
         978,
         980,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         981,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         978,
         982,
         983,
         982,
         982,
         982,
         982,
         984,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         985,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         982,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         986,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         987,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         988,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         989,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         990,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         991,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         992,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         993,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2294,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         994,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         995,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         996,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         997,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         998,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         961,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         2249,
         2250,
         2249,
         2249,
         2249,
         2249,
         2251,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2252,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2249,
         2253,
         2254,
         2253,
         2253,
         2253,
         2253,
         2255,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2256,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2253,
         2257,
         2258,
         2257,
         2257,
         2257,
         2257,
         2259,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2260,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2257,
         2261,
         2262,
         2261,
         2261,
         2261,
         2261,
         2263,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2264,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         952,
         952,
         952,
         952,
         952,
         935,
         935,
         935,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         954,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         954,
         954,
         954,
         935,
         935,
         936,
         935,
         935,
         935,
         935,
         937,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         938,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         935,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2964,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2964,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2964,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2964,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2965,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2990,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         127,
         127,
         127,
         482,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         483,
         127,
         127,
         484,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         485,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         486,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         487,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         488,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         489,
         490,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         494,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         496,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         497,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         498,
         499,
         498,
         498,
         498,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         503,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         504,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         505,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         506,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         507,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         508,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         509,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         510,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         511,
         512,
         511,
         511,
         511,
         511,
         513,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         514,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         511,
         515,
         516,
         515,
         515,
         515,
         515,
         517,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         518,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         515,
         519,
         520,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         522,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         523,
         524,
         523,
         523,
         523,
         523,
         525,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         526,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         523,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         527,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         528,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         529,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         531,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         532,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         533,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         502,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2134,
         2135,
         2134,
         2134,
         2134,
         2134,
         2136,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2137,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2134,
         2138,
         2139,
         2138,
         2138,
         2138,
         2138,
         2140,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2141,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2138,
         2142,
         2143,
         2142,
         2142,
         2142,
         2142,
         2144,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2145,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2142,
         2146,
         2147,
         2146,
         2146,
         2146,
         2146,
         2148,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2149,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         2146,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         493,
         493,
         493,
         493,
         493,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         495,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         495,
         495,
         495,
         125,
         125,
         126,
         125,
         125,
         125,
         125,
         480,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         481,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         19,
         4,
         4,
         4,
         4,
         4,
         4,
         20,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         21,
         2,
         2,
         0,
         0,
         4,
         4,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         21,
         2,
         2,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         133,
         134,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         134,
         134,
         1476,
         1477,
         1478,
         1479,
         1480,
         1481,
         1482,
         1483,
         134,
         1484,
         1485,
         1486,
         1487,
         2352,
         2353,
         2354,
         2355,
         2356,
         1478,
         2920,
         1474,
         1476,
         1474,
         1474,
         1488,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1488,
         1488,
         1474,
         1474,
         1477,
         1474,
         1474,
         1488,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1488,
         1488,
         1474,
         1475,
         1478,
         1488,
         1488,
         1489,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1488,
         1489,
         1489,
         1488,
         133,
         1479,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1480,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1481,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1482,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         1490,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         2357,
         1474,
         133,
         1483,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         134,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         1491,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1484,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1485,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1486,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         1487,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         2352,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         2353,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         133,
         2354,
         1474,
         1474,
         1475,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         133,
         1474,
         1475,
         1475,
         1474,
         1474,
         2355,
         1474,
         1474,
         1488,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1488,
         1488,
         1474,
         1475,
         2356,
         1488,
         1488,
         1489,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1488,
         1489,
         1489,
         1488,
         1475,
         1478,
         1488,
         1488,
         1489,
         1475,
         1475,
         1475,
         2357,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1488,
         1489,
         2358,
         1488,
         1474,
         2920,
         1474,
         1474,
         1488,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1488,
         1488,
         1474,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         4,
         1,
         1,
         1,
         1,
         1,
         3,
         1,
         1,
         1,
         3,
         1,
         1,
         1,
         3,
         1,
         3,
         1,
         1,
         3,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         3,
         3,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         135,
         1461,
         1462,
         2619,
         2620,
         2922,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         1,
         0,
         0,
         0,
         1,
         1,
         0,
         1,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         136,
         2012,
         2013,
         2014,
      },
   },
   [nir_op_ftrunc] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         137,
      },
   },
   [nir_op_ffract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         138,
      },
   },
   [nir_op_fceil] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         139,
      },
   },
   [nir_search_op_b2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         2,
         0,
         0,
         0,
         2,
         2,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         5,
         1,
         1,
         1,
         1,
         1,
         1,
         6,
         6,
         7,
         7,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         140,
         1901,
         1902,
         1903,
         1904,
         2688,
         2689,
         2690,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         141,
      },
   },
   [nir_op_fdot4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         142,
         143,
         1997,
         1998,
         1999,
         143,
         143,
         2000,
         2001,
         2002,
         1997,
         2000,
         1997,
         1998,
         1999,
         1998,
         2001,
         1998,
         1998,
         1999,
         1999,
         2002,
         1999,
         1999,
         1999,
      },
   },
   [nir_op_fdot3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         144,
         145,
         2003,
         2004,
         145,
         145,
         2005,
         2006,
         2003,
         2005,
         2003,
         2004,
         2004,
         2006,
         2004,
         2004,
      },
   },
   [nir_op_fdot2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         146,
         147,
         2007,
         147,
         147,
         2008,
         2007,
         2008,
         2007,
      },
   },
   [nir_op_vec4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         148,
         149,
         148,
         150,
         148,
         149,
         148,
         151,
         148,
         149,
         148,
         150,
         148,
         149,
         148,
         151,
      },
   },
   [nir_op_vec3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         152,
         153,
         152,
         154,
         152,
         153,
         152,
         154,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         155,
         156,
         155,
         155,
         155,
         157,
         158,
         157,
         157,
         157,
         155,
         156,
         155,
         1317,
         155,
         155,
         156,
         155,
         155,
         155,
         155,
         156,
         155,
         155,
         1318,
      },
   },
   [nir_op_fsum2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         159,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         4,
         0,
         0,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         160,
         161,
         160,
         160,
         1301,
         160,
         162,
         163,
         162,
         162,
         1302,
         162,
         160,
         1303,
         160,
         160,
         1301,
         160,
         1304,
         1305,
         1304,
         1304,
         1306,
         1304,
         160,
         161,
         160,
         160,
         1301,
         160,
         160,
         1307,
         160,
         160,
         1301,
         160,
      },
   },
   [nir_op_inot] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         6,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         9,
         9,
         10,
         10,
         11,
         11,
         11,
         11,
         12,
         12,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         15,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         6,
         6,
         6,
         6,
         6,
         6,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         21,
         21,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         22,
         20,
         20,
         20,
         20,
         20,
         0,
         21,
         21,
         21,
         21,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         10,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         9,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 23,
      .table = (const uint16_t []) {
      
         164,
         1878,
         1879,
         1880,
         1881,
         1882,
         1883,
         1884,
         1885,
         1886,
         1887,
         1888,
         1889,
         1890,
         2500,
         2501,
         2502,
         2503,
         2504,
         2505,
         2880,
         2881,
         2882,
      },
   },
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         13,
         3,
         3,
         3,
         14,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         13,
         2,
         2,
         15,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         16,
         17,
         5,
         5,
         5,
         5,
         5,
         18,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         13,
         3,
         3,
         14,
         14,
         14,
         14,
         14,
         14,
         13,
         13,
         2,
         13,
         2,
         13,
         0,
         0,
         5,
         19,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         20,
         3,
         3,
         3,
         3,
         21,
         2,
         2,
         2,
         2,
         2,
         2,
         13,
         13,
         13,
         2,
         2,
         2,
         2,
         2,
         2,
         13,
         13,
         13,
         13,
         2,
         2,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         22,
         9,
         22,
         9,
         9,
         22,
         9,
         9,
         22,
         9,
         22,
         22,
         9,
         22,
         9,
         22,
         9,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         0,
         14,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 23,
      .table = (const uint16_t []) {
      
         165,
         166,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         167,
         168,
         167,
         167,
         1495,
         1496,
         1497,
         1498,
         1499,
         1500,
         1501,
         1502,
         167,
         167,
         167,
         2624,
         1496,
         1496,
         1496,
         1496,
         167,
         167,
         1500,
         1503,
         1504,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         1503,
         1508,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         165,
         166,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         165,
         1509,
         165,
         165,
         165,
         1510,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1510,
         1510,
         1510,
         1510,
         165,
         165,
         1493,
         165,
         1511,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         165,
         166,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         165,
         1512,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         1513,
         1514,
         1513,
         1513,
         1513,
         1515,
         1513,
         1513,
         1513,
         1516,
         1513,
         1517,
         1513,
         1513,
         1513,
         1513,
         1515,
         1515,
         1515,
         1515,
         1513,
         1513,
         1516,
         165,
         1518,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         1519,
         1520,
         1519,
         1519,
         1519,
         1521,
         1519,
         1519,
         1519,
         1522,
         1519,
         1523,
         1519,
         1519,
         1519,
         1519,
         1521,
         1521,
         1521,
         1521,
         1519,
         1519,
         1522,
         165,
         166,
         165,
         165,
         165,
         1492,
         165,
         165,
         165,
         1493,
         165,
         1494,
         1524,
         165,
         165,
         165,
         1492,
         1492,
         1492,
         1492,
         165,
         165,
         1493,
         1503,
         2625,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         1503,
         2626,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         1503,
         1504,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         165,
         2627,
         165,
         165,
         165,
         1510,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1510,
         1510,
         1510,
         1510,
         165,
         165,
         1493,
         165,
         2628,
         165,
         165,
         165,
         1510,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1510,
         1510,
         1510,
         1510,
         165,
         165,
         1493,
         165,
         2629,
         165,
         165,
         165,
         1510,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1510,
         1510,
         1510,
         1510,
         165,
         165,
         1493,
         165,
         2923,
         165,
         165,
         165,
         1510,
         165,
         165,
         165,
         1493,
         165,
         1494,
         165,
         165,
         165,
         165,
         1510,
         1510,
         1510,
         1510,
         165,
         165,
         1493,
         1503,
         2966,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         1503,
         2967,
         1503,
         1503,
         1503,
         1505,
         1503,
         1503,
         1503,
         1506,
         1503,
         1507,
         1503,
         1503,
         1503,
         1503,
         1505,
         1505,
         1505,
         1505,
         1503,
         1503,
         1506,
         1513,
         2968,
         1513,
         1513,
         1513,
         1515,
         1513,
         1513,
         1513,
         1516,
         1513,
         1517,
         1513,
         1513,
         1513,
         1513,
         1515,
         1515,
         1515,
         1515,
         1513,
         1513,
         1516,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         14,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         15,
         16,
         5,
         5,
         5,
         5,
         5,
         17,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         5,
         18,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         169,
         170,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         171,
         172,
         171,
         171,
         1528,
         1529,
         1530,
         171,
         1531,
         1532,
         1533,
         1534,
         1535,
         171,
         2663,
         1529,
         1529,
         1529,
         1529,
         1536,
         1537,
         1536,
         1536,
         1536,
         1538,
         1536,
         1536,
         1536,
         1536,
         1539,
         1536,
         1540,
         1536,
         1536,
         1538,
         1538,
         1538,
         1538,
         1536,
         1541,
         1536,
         1536,
         1536,
         1538,
         1536,
         1536,
         1536,
         1536,
         1539,
         1536,
         1540,
         1536,
         1536,
         1538,
         1538,
         1538,
         1538,
         169,
         170,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         169,
         1542,
         169,
         169,
         169,
         1543,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1543,
         1543,
         1543,
         1543,
         169,
         1544,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         169,
         1545,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         169,
         1546,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         169,
         170,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         1547,
         1548,
         1547,
         1547,
         1547,
         1549,
         1547,
         1547,
         1547,
         1547,
         1550,
         1547,
         1551,
         1547,
         1547,
         1549,
         1549,
         1549,
         1549,
         169,
         1552,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1525,
         1525,
         1525,
         1525,
         1553,
         1554,
         1553,
         1553,
         1553,
         1555,
         1553,
         1553,
         1553,
         1553,
         1556,
         1553,
         1557,
         1553,
         1553,
         1555,
         1555,
         1555,
         1555,
         169,
         170,
         169,
         169,
         169,
         1525,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         1558,
         169,
         1525,
         1525,
         1525,
         1525,
         1536,
         1537,
         1536,
         1536,
         1536,
         1538,
         1536,
         1536,
         1536,
         1536,
         1539,
         1536,
         1540,
         1536,
         1536,
         1538,
         1538,
         1538,
         1538,
         169,
         2664,
         169,
         169,
         169,
         1543,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1543,
         1543,
         1543,
         1543,
         169,
         2665,
         169,
         169,
         169,
         1543,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1543,
         1543,
         1543,
         1543,
         169,
         2666,
         169,
         169,
         169,
         1543,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1543,
         1543,
         1543,
         1543,
         169,
         2924,
         169,
         169,
         169,
         1543,
         169,
         169,
         169,
         169,
         1526,
         169,
         1527,
         169,
         169,
         1543,
         1543,
         1543,
         1543,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         19,
         4,
         4,
         4,
         4,
         4,
         4,
         20,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         21,
         2,
         2,
         0,
         0,
         4,
         4,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         21,
         2,
         2,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         173,
         174,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         174,
         174,
         1561,
         1562,
         1563,
         1564,
         1565,
         1566,
         1567,
         1568,
         174,
         1569,
         1570,
         1571,
         1572,
         2359,
         2360,
         2361,
         2362,
         2363,
         1563,
         2921,
         1559,
         1561,
         1559,
         1559,
         1573,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1573,
         1573,
         1559,
         1559,
         1562,
         1559,
         1559,
         1573,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1573,
         1573,
         1559,
         1560,
         1563,
         1573,
         1573,
         1574,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1573,
         1574,
         1574,
         1573,
         173,
         1564,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1565,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1566,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1567,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         1575,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         2364,
         1559,
         173,
         1568,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         174,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         1576,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1569,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1570,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1571,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         1572,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         2359,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         2360,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         173,
         2361,
         1559,
         1559,
         1560,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         173,
         1559,
         1560,
         1560,
         1559,
         1559,
         2362,
         1559,
         1559,
         1573,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1573,
         1573,
         1559,
         1560,
         2363,
         1573,
         1573,
         1574,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1573,
         1574,
         1574,
         1573,
         1560,
         1563,
         1573,
         1573,
         1574,
         1560,
         1560,
         1560,
         2364,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1560,
         1573,
         1574,
         2365,
         1573,
         1559,
         2921,
         1559,
         1559,
         1573,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1559,
         1573,
         1573,
         1559,
      },
   },
   [nir_op_ilt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         175,
         176,
         175,
         1915,
         1915,
         1916,
         1916,
         175,
         175,
         175,
         176,
         1917,
         1915,
         1918,
         1916,
         1919,
         1920,
         175,
         175,
         1921,
         175,
         1915,
         1915,
         1916,
         1916,
         175,
         175,
         1922,
         1923,
         1922,
         1924,
         1924,
         1925,
         1925,
         1922,
         1922,
         1922,
         1926,
         1922,
         1924,
         1924,
         1925,
         1925,
         1922,
         1922,
         1927,
         1928,
         1927,
         1929,
         1929,
         1930,
         1930,
         1927,
         1927,
         1927,
         1931,
         1927,
         1929,
         1929,
         1930,
         1930,
         1927,
         1927,
         175,
         1932,
         175,
         1915,
         1915,
         1916,
         1916,
         1933,
         175,
         175,
         2723,
         175,
         1915,
         1915,
         1916,
         1916,
         175,
         175,
      },
   },
   [nir_op_ult] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         6,
         13,
         7,
         7,
         12,
         12,
         12,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         8,
         8,
         8,
         9,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
      },
      
      .num_filtered_states = 14,
      .table = (const uint16_t []) {
      
         177,
         178,
         177,
         1752,
         177,
         177,
         1753,
         1753,
         1754,
         1754,
         177,
         177,
         2724,
         2724,
         179,
         180,
         179,
         1755,
         179,
         1756,
         1757,
         1758,
         1759,
         1760,
         1761,
         2725,
         2726,
         2727,
         1762,
         1763,
         1762,
         1764,
         1762,
         1762,
         1765,
         1765,
         1766,
         1766,
         1762,
         1762,
         2728,
         2728,
         177,
         178,
         177,
         1752,
         177,
         177,
         1753,
         1753,
         1754,
         1754,
         177,
         177,
         2724,
         2724,
         1767,
         1768,
         1767,
         1769,
         1767,
         1767,
         1770,
         1770,
         1771,
         1771,
         1767,
         1767,
         2729,
         2729,
         177,
         1772,
         177,
         1752,
         177,
         177,
         1753,
         1753,
         1754,
         1754,
         177,
         177,
         2724,
         2724,
         1773,
         1774,
         1773,
         1775,
         1773,
         1773,
         1776,
         1776,
         1777,
         1777,
         1773,
         1773,
         2730,
         2730,
         1773,
         1778,
         1773,
         1775,
         1773,
         1773,
         1776,
         1776,
         1777,
         1777,
         1773,
         1773,
         2730,
         2730,
         1779,
         1780,
         1779,
         1781,
         1779,
         1779,
         1782,
         1782,
         1783,
         1783,
         1779,
         1779,
         2731,
         2731,
         1779,
         1784,
         1779,
         1781,
         1779,
         1779,
         1782,
         1782,
         1783,
         1783,
         1779,
         1779,
         2731,
         2731,
         177,
         1785,
         177,
         1752,
         177,
         177,
         1753,
         1753,
         1754,
         1754,
         1786,
         177,
         2724,
         2724,
         177,
         178,
         177,
         1752,
         177,
         177,
         1753,
         1753,
         1754,
         1754,
         177,
         177,
         2724,
         2724,
         1773,
         1774,
         1773,
         1775,
         1773,
         1773,
         1776,
         1776,
         1777,
         1777,
         1773,
         1773,
         2730,
         2730,
         1773,
         1778,
         1773,
         1775,
         1773,
         1773,
         1776,
         1776,
         1777,
         1777,
         1773,
         1773,
         2730,
         2730,
      },
   },
   [nir_op_ige] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         10,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 14,
      .table = (const uint16_t []) {
      
         181,
         182,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         182,
         1936,
         1934,
         1937,
         1935,
         1938,
         1939,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1940,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1941,
         1942,
         1941,
         1943,
         1943,
         1944,
         1944,
         1941,
         1941,
         1941,
         1941,
         1941,
         1941,
         1941,
         1941,
         1945,
         1941,
         1943,
         1943,
         1944,
         1944,
         1941,
         1941,
         1941,
         1941,
         1941,
         1941,
         1941,
         1946,
         1947,
         1946,
         1948,
         1948,
         1949,
         1949,
         1946,
         1946,
         1946,
         1946,
         1946,
         1946,
         1946,
         1946,
         1950,
         1946,
         1948,
         1948,
         1949,
         1949,
         1946,
         1946,
         1946,
         1946,
         1946,
         1946,
         1946,
         181,
         1951,
         181,
         1934,
         1934,
         1935,
         1935,
         1952,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1953,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1954,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1955,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1956,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         1957,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
         2732,
         181,
         1934,
         1934,
         1935,
         1935,
         181,
         181,
         181,
         181,
         181,
         181,
         181,
      },
   },
   [nir_op_uge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         8,
         8,
         8,
         9,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         183,
         183,
         183,
         1787,
         183,
         183,
         1788,
         1788,
         1789,
         1789,
         183,
         184,
         184,
         184,
         1790,
         184,
         1791,
         1792,
         1793,
         1794,
         1795,
         1796,
         1797,
         1797,
         1797,
         1798,
         1797,
         1797,
         1799,
         1799,
         1800,
         1800,
         1797,
         183,
         183,
         183,
         1787,
         183,
         183,
         1788,
         1788,
         1789,
         1789,
         183,
         1801,
         1801,
         1801,
         1802,
         1801,
         1801,
         1803,
         1803,
         1804,
         1804,
         1801,
         183,
         1805,
         183,
         1787,
         183,
         183,
         1788,
         1788,
         1789,
         1789,
         183,
         1806,
         1806,
         1806,
         1807,
         1806,
         1806,
         1808,
         1808,
         1809,
         1809,
         1806,
         1806,
         1810,
         1806,
         1807,
         1806,
         1806,
         1808,
         1808,
         1809,
         1809,
         1806,
         1811,
         1811,
         1811,
         1812,
         1811,
         1811,
         1813,
         1813,
         1814,
         1814,
         1811,
         1811,
         1815,
         1811,
         1812,
         1811,
         1811,
         1813,
         1813,
         1814,
         1814,
         1811,
         183,
         1816,
         183,
         1787,
         183,
         183,
         1788,
         1788,
         1789,
         1789,
         1817,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         9,
         3,
         3,
         3,
         3,
         3,
         10,
         3,
         3,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         6,
         6,
         6,
         12,
         6,
         12,
         6,
         6,
         12,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         13,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         7,
         7,
         14,
         7,
         8,
         8,
         8,
         15,
         15,
         8,
         7,
         7,
         14,
         7,
         14,
         7,
         7,
         7,
         14,
         7,
         14,
         7,
         7,
         14,
         7,
         14,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         12,
         6,
         6,
         12,
         12,
         12,
         12,
         6,
         6,
         12,
         6,
         12,
         6,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         185,
         186,
         1604,
         1605,
         185,
         185,
         1606,
         1607,
         1607,
         2668,
         1605,
         2669,
         2670,
         2925,
         2926,
         2926,
         186,
         186,
         1608,
         1609,
         1610,
         1611,
         1612,
         1613,
         1614,
         2671,
         1609,
         2672,
         2673,
         2927,
         2928,
         2929,
         1604,
         1608,
         1604,
         1615,
         1604,
         1604,
         1616,
         1617,
         1617,
         2674,
         1615,
         2675,
         2676,
         2930,
         2931,
         2931,
         1605,
         1609,
         1615,
         1605,
         1605,
         1605,
         1618,
         1619,
         1619,
         2668,
         1605,
         2669,
         2677,
         2925,
         2932,
         2932,
         185,
         1610,
         1604,
         1605,
         185,
         185,
         1606,
         1607,
         1607,
         2668,
         1605,
         2669,
         2670,
         2925,
         2926,
         2926,
         185,
         1611,
         1604,
         1605,
         185,
         1620,
         1606,
         1607,
         1607,
         2668,
         1605,
         2669,
         2670,
         2925,
         2926,
         2926,
         1606,
         1612,
         1616,
         1618,
         1606,
         1606,
         1606,
         1621,
         1621,
         2678,
         1618,
         2679,
         2670,
         2933,
         2934,
         2934,
         1607,
         1613,
         1617,
         1619,
         1607,
         1607,
         1621,
         1607,
         1607,
         2680,
         1619,
         2681,
         2682,
         2935,
         2926,
         2926,
         1607,
         1614,
         1617,
         1619,
         1607,
         1607,
         1621,
         1607,
         1607,
         2680,
         1619,
         2681,
         2682,
         2935,
         2926,
         2926,
         2668,
         2671,
         2674,
         2668,
         2668,
         2668,
         2678,
         2680,
         2680,
         2668,
         2668,
         2683,
         2684,
         2936,
         2937,
         2937,
         1605,
         1609,
         1615,
         1605,
         1605,
         1605,
         1618,
         1619,
         1619,
         2668,
         2685,
         2669,
         2677,
         2925,
         2932,
         2932,
         2669,
         2672,
         2675,
         2669,
         2669,
         2669,
         2679,
         2681,
         2681,
         2683,
         2669,
         2669,
         2686,
         2938,
         2939,
         2939,
         2670,
         2673,
         2676,
         2677,
         2670,
         2670,
         2670,
         2682,
         2682,
         2684,
         2677,
         2686,
         2670,
         2940,
         2941,
         2941,
         2925,
         2927,
         2930,
         2925,
         2925,
         2925,
         2933,
         2935,
         2935,
         2936,
         2925,
         2938,
         2940,
         2925,
         2942,
         2942,
         2926,
         2928,
         2931,
         2932,
         2926,
         2926,
         2934,
         2926,
         2926,
         2937,
         2932,
         2939,
         2941,
         2942,
         2926,
         2926,
         2926,
         2929,
         2931,
         2932,
         2926,
         2926,
         2934,
         2926,
         2926,
         2937,
         2932,
         2939,
         2941,
         2942,
         2926,
         2926,
      },
   },
   [nir_search_op_i2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         187,
         1978,
         1979,
         1980,
         1981,
         2687,
         2985,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         9,
         3,
         3,
         3,
         3,
         3,
         10,
         3,
         11,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         0,
         0,
         12,
         0,
         12,
         0,
         12,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         14,
         7,
         14,
         14,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         7,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         7,
         14,
         7,
         7,
         7,
         14,
         7,
         14,
         14,
         14,
         14,
         7,
         7,
         14,
         7,
         7,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         13,
         13,
         0,
         12,
         12,
         0,
         12,
         12,
         0,
         12,
         0,
         12,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 17,
      .table = (const uint16_t []) {
      
         188,
         189,
         1622,
         1623,
         188,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         189,
         189,
         1625,
         1626,
         1627,
         189,
         1628,
         1629,
         2641,
         2642,
         1626,
         2643,
         2644,
         2645,
         2646,
         2969,
         2970,
         1622,
         1625,
         1622,
         1630,
         1622,
         1622,
         1622,
         1631,
         1622,
         2647,
         1630,
         2648,
         2649,
         2649,
         2650,
         1622,
         1622,
         1623,
         1626,
         1630,
         1623,
         1623,
         1623,
         1623,
         1632,
         1623,
         2637,
         1623,
         2638,
         2651,
         2651,
         2652,
         1623,
         1623,
         188,
         1627,
         1622,
         1623,
         1633,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         188,
         189,
         1622,
         1623,
         188,
         1634,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         188,
         1628,
         1622,
         1623,
         188,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         1624,
         1629,
         1631,
         1632,
         1624,
         1624,
         1624,
         1624,
         1624,
         2653,
         1632,
         2654,
         2655,
         2655,
         2640,
         1624,
         1624,
         188,
         2641,
         1622,
         1623,
         188,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         2637,
         2642,
         2647,
         2637,
         2637,
         2637,
         2637,
         2653,
         2637,
         2637,
         2637,
         2656,
         2657,
         2657,
         2658,
         2637,
         2637,
         1623,
         1626,
         1630,
         1623,
         1623,
         1623,
         1623,
         1632,
         1623,
         2637,
         2659,
         2638,
         2651,
         2651,
         2652,
         1623,
         1623,
         2638,
         2643,
         2648,
         2638,
         2638,
         2638,
         2638,
         2654,
         2638,
         2656,
         2638,
         2638,
         2660,
         2660,
         2661,
         2638,
         2638,
         2639,
         2644,
         2649,
         2651,
         2639,
         2639,
         2639,
         2655,
         2639,
         2657,
         2651,
         2660,
         2639,
         2639,
         2662,
         2639,
         2639,
         2639,
         2645,
         2649,
         2651,
         2639,
         2639,
         2639,
         2655,
         2639,
         2657,
         2651,
         2660,
         2639,
         2639,
         2662,
         2639,
         2639,
         2640,
         2646,
         2650,
         2652,
         2640,
         2640,
         2640,
         2640,
         2640,
         2658,
         2652,
         2661,
         2662,
         2662,
         2640,
         2640,
         2640,
         188,
         2969,
         1622,
         1623,
         188,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
         188,
         2970,
         1622,
         1623,
         188,
         188,
         188,
         1624,
         188,
         2637,
         1623,
         2638,
         2639,
         2639,
         2640,
         188,
         188,
      },
   },
   [nir_op_imin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         7,
         2,
         2,
         8,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         9,
         4,
         9,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         5,
         10,
         5,
         6,
         11,
         11,
         6,
         10,
         5,
         5,
         10,
         5,
         5,
         10,
         10,
         5,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         9,
         4,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         190,
         191,
         1704,
         1705,
         1706,
         1707,
         1707,
         2691,
         1704,
         2692,
         2943,
         2943,
         191,
         191,
         1708,
         1709,
         1710,
         1711,
         1712,
         2693,
         1708,
         2694,
         2944,
         2945,
         1704,
         1708,
         1704,
         1713,
         1714,
         1715,
         1715,
         2691,
         1704,
         2695,
         2946,
         2946,
         1705,
         1709,
         1713,
         1705,
         1716,
         1717,
         1717,
         2696,
         1713,
         2697,
         2947,
         2947,
         1706,
         1710,
         1714,
         1716,
         1706,
         1718,
         1718,
         2698,
         1714,
         2692,
         2948,
         2948,
         1707,
         1711,
         1715,
         1717,
         1718,
         1707,
         1707,
         2699,
         1715,
         2700,
         2943,
         2943,
         1707,
         1712,
         1715,
         1717,
         1718,
         1707,
         1707,
         2699,
         1715,
         2700,
         2943,
         2943,
         2691,
         2693,
         2691,
         2696,
         2698,
         2699,
         2699,
         2691,
         2691,
         2701,
         2949,
         2949,
         1704,
         1708,
         1704,
         1713,
         1714,
         1715,
         1715,
         2691,
         2702,
         2695,
         2946,
         2946,
         2692,
         2694,
         2695,
         2697,
         2692,
         2700,
         2700,
         2701,
         2695,
         2692,
         2950,
         2950,
         2943,
         2944,
         2946,
         2947,
         2948,
         2943,
         2943,
         2949,
         2946,
         2950,
         2943,
         2943,
         2943,
         2945,
         2946,
         2947,
         2948,
         2943,
         2943,
         2949,
         2946,
         2950,
         2943,
         2943,
      },
   },
   [nir_op_imax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         6,
         2,
         2,
         7,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         4,
         4,
         4,
         9,
         9,
         0,
         0,
         8,
         0,
         8,
         8,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         5,
         5,
         10,
         10,
         10,
         5,
         10,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         192,
         193,
         1719,
         1720,
         192,
         1721,
         2703,
         1719,
         2704,
         2704,
         2705,
         193,
         193,
         1722,
         1723,
         1724,
         1725,
         2706,
         1722,
         2707,
         2708,
         2709,
         1719,
         1722,
         1719,
         1726,
         1719,
         1727,
         2703,
         1719,
         2710,
         2710,
         2711,
         1720,
         1723,
         1726,
         1720,
         1720,
         1728,
         2712,
         1726,
         2713,
         2713,
         2714,
         192,
         1724,
         1719,
         1720,
         192,
         1721,
         2703,
         1719,
         2704,
         2704,
         2705,
         1721,
         1725,
         1727,
         1728,
         1721,
         1721,
         2715,
         1727,
         2716,
         2716,
         2705,
         2703,
         2706,
         2703,
         2712,
         2703,
         2715,
         2703,
         2703,
         2717,
         2717,
         2718,
         1719,
         1722,
         1719,
         1726,
         1719,
         1727,
         2703,
         2719,
         2710,
         2710,
         2711,
         2704,
         2707,
         2710,
         2713,
         2704,
         2716,
         2717,
         2710,
         2704,
         2704,
         2720,
         2704,
         2708,
         2710,
         2713,
         2704,
         2716,
         2717,
         2710,
         2704,
         2704,
         2720,
         2705,
         2709,
         2711,
         2714,
         2705,
         2705,
         2718,
         2711,
         2720,
         2720,
         2705,
      },
   },
   [nir_op_umin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         8,
         5,
         5,
         8,
         5,
         5,
         8,
         5,
         5,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         8,
         5,
         8,
         5,
         8,
         8,
         5,
         8,
         5,
         8,
         6,
         10,
         6,
         6,
         10,
         6,
         6,
         10,
         6,
         10,
         6,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         8,
         8,
         5,
         5,
         5,
         8,
         8,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         194,
         195,
         1818,
         1818,
         194,
         1819,
         1820,
         2733,
         2734,
         2821,
         2822,
         195,
         195,
         1821,
         1821,
         195,
         1822,
         1823,
         2735,
         2736,
         2823,
         2824,
         1818,
         1821,
         1818,
         1824,
         1818,
         1825,
         1826,
         2737,
         2738,
         2825,
         2826,
         1818,
         1821,
         1824,
         1824,
         1818,
         1825,
         1826,
         2737,
         2738,
         2825,
         2826,
         194,
         195,
         1818,
         1818,
         1827,
         1819,
         1820,
         2733,
         2734,
         2821,
         2822,
         1819,
         1822,
         1825,
         1825,
         1819,
         1819,
         1828,
         2739,
         2734,
         2827,
         2828,
         1820,
         1823,
         1826,
         1826,
         1820,
         1828,
         1820,
         2740,
         2741,
         2829,
         2822,
         2733,
         2735,
         2737,
         2737,
         2733,
         2739,
         2740,
         2742,
         2743,
         2830,
         2831,
         2734,
         2736,
         2738,
         2738,
         2734,
         2734,
         2741,
         2743,
         2734,
         2832,
         2833,
         2821,
         2823,
         2825,
         2825,
         2821,
         2827,
         2829,
         2830,
         2832,
         2821,
         2834,
         2822,
         2824,
         2826,
         2826,
         2822,
         2828,
         2822,
         2831,
         2833,
         2834,
         2822,
      },
   },
   [nir_op_umax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         5,
         0,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         3,
         3,
         6,
         3,
         3,
         6,
         6,
         6,
         3,
         3,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         5,
         5,
         0,
         5,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         196,
         197,
         196,
         1982,
         2744,
         2745,
         2746,
         197,
         197,
         197,
         1983,
         2747,
         2748,
         2749,
         196,
         197,
         1984,
         1982,
         2744,
         2745,
         2746,
         1982,
         1983,
         1982,
         1982,
         2750,
         2751,
         2746,
         2744,
         2747,
         2744,
         2750,
         2752,
         2753,
         2754,
         2745,
         2748,
         2745,
         2751,
         2753,
         2745,
         2755,
         2746,
         2749,
         2746,
         2746,
         2754,
         2755,
         2746,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         0,
         4,
         0,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
      },
      
      .num_filtered_states = 10,
      .table = (const uint16_t []) {
      
         198,
         199,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         199,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1308,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1309,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1310,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1311,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1312,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1313,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         1314,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         2757,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
         198,
      },
   },
   [nir_search_op_f2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         200,
         1467,
         1468,
         1469,
         1470,
         1471,
         1472,
         1473,
      },
   },
   [nir_search_op_f2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         201,
         1729,
         1730,
         1731,
         1732,
         1733,
         1734,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         202,
         1735,
         1736,
         1737,
         1738,
         1739,
         1740,
         1741,
         1742,
         1743,
         1744,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         11,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         203,
         1831,
         1832,
         1833,
         1834,
         1835,
         1836,
         1837,
         1838,
         1839,
         1840,
         2795,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         8,
         9,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         6,
         6,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         14,
         9,
         9,
         9,
         9,
         9,
         15,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         9,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         0,
         0,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         204,
         1841,
         1842,
         1843,
         1844,
         1845,
         1846,
         1847,
         1848,
         1849,
         1850,
         1851,
         1852,
         1853,
         2793,
         2794,
         2877,
         2878,
         2879,
      },
   },
   [nir_op_f2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         205,
         1745,
         1746,
         1747,
         1748,
         1749,
         1750,
         1751,
      },
   },
   [nir_op_i2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         206,
         1985,
         1986,
         1987,
         1988,
         1989,
         1990,
         1991,
         1992,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         207,
         208,
         207,
         207,
         207,
         208,
         207,
         207,
         207,
         1315,
         207,
         207,
         207,
         1316,
         207,
         207,
      },
   },
   [nir_op_extract_i16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         209,
         210,
         209,
         210,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         211,
         212,
         211,
         211,
         211,
         211,
         211,
         212,
         211,
         211,
         211,
         211,
         211,
         2016,
         211,
         211,
         211,
         211,
         211,
         2017,
         211,
         211,
         211,
         211,
         211,
         2018,
         211,
         211,
         211,
         211,
         211,
         2019,
         211,
         211,
         211,
         211,
      },
   },
   [nir_op_ixor] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         2,
         2,
         3,
         2,
         2,
         3,
         3,
         3,
         2,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         4,
         4,
         4,
         4,
         4,
         5,
         4,
         4,
         5,
         5,
         5,
         6,
         6,
         7,
         6,
         7,
         6,
         7,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         21,
         21,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         214,
         214,
         214,
         214,
         214,
         214,
         214,
         214,
         214,
         1267,
         214,
         1268,
         214,
         1269,
         214,
         1270,
         1271,
         1272,
         2762,
         2763,
         2764,
         2765,
         213,
         214,
         213,
         1273,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         1273,
         1273,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         213,
         213,
         213,
         1274,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         213,
         213,
         1274,
         1274,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         1275,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         213,
         213,
         213,
         213,
         1275,
         1275,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1276,
         213,
         1263,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         1262,
         1267,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1276,
         1276,
         1262,
         1277,
         1262,
         1278,
         1262,
         1279,
         1280,
         1280,
         2766,
         2767,
         2768,
         2769,
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1281,
         213,
         1264,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         1263,
         1268,
         1263,
         1263,
         1263,
         1263,
         1263,
         1263,
         1263,
         1277,
         1281,
         1281,
         1263,
         1282,
         1263,
         1283,
         1284,
         1284,
         2770,
         2771,
         2772,
         2773,
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1285,
         213,
         1265,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         1264,
         1269,
         1264,
         1264,
         1264,
         1264,
         1264,
         1264,
         1264,
         1278,
         1264,
         1282,
         1285,
         1285,
         1264,
         1286,
         1287,
         1287,
         2774,
         2775,
         2776,
         2777,
         213,
         214,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1262,
         213,
         1263,
         213,
         1264,
         213,
         1288,
         1266,
         1266,
         2758,
         2759,
         2760,
         2761,
         1265,
         1270,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1279,
         1265,
         1283,
         1265,
         1286,
         1288,
         1288,
         1289,
         1289,
         2778,
         2779,
         2780,
         2781,
         1266,
         1271,
         1266,
         1266,
         1266,
         1266,
         1266,
         1266,
         1266,
         1280,
         1266,
         1284,
         1266,
         1287,
         1266,
         1289,
         1266,
         1266,
         2782,
         2783,
         2784,
         2785,
         1266,
         1272,
         1266,
         1266,
         1266,
         1266,
         1266,
         1266,
         1266,
         1280,
         1266,
         1284,
         1266,
         1287,
         1266,
         1289,
         1266,
         1266,
         2782,
         2783,
         2784,
         2785,
         2758,
         2762,
         2758,
         2758,
         2758,
         2758,
         2758,
         2758,
         2758,
         2766,
         2758,
         2770,
         2758,
         2774,
         2758,
         2778,
         2782,
         2782,
         2758,
         2786,
         2787,
         2788,
         2759,
         2763,
         2759,
         2759,
         2759,
         2759,
         2759,
         2759,
         2759,
         2767,
         2759,
         2771,
         2759,
         2775,
         2759,
         2779,
         2783,
         2783,
         2786,
         2759,
         2789,
         2790,
         2760,
         2764,
         2760,
         2760,
         2760,
         2760,
         2760,
         2760,
         2760,
         2768,
         2760,
         2772,
         2760,
         2776,
         2760,
         2780,
         2784,
         2784,
         2787,
         2789,
         2760,
         2791,
         2761,
         2765,
         2761,
         2761,
         2761,
         2761,
         2761,
         2761,
         2761,
         2769,
         2761,
         2773,
         2761,
         2777,
         2761,
         2781,
         2785,
         2785,
         2788,
         2790,
         2791,
         2792,
      },
   },
   [nir_op_slt] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         215,
      },
   },
   [nir_op_sge] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         216,
      },
   },
   [nir_op_seq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         2020,
         2021,
         2022,
         2023,
         217,
         2020,
         217,
         217,
         217,
         217,
         217,
         2021,
         217,
         217,
         217,
         217,
         217,
         2022,
         217,
         217,
         217,
         217,
         217,
         2023,
         217,
         217,
         217,
         217,
      },
   },
   [nir_op_sne] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         218,
         218,
         218,
         218,
         218,
         218,
         218,
         218,
         2024,
         2025,
         2026,
         2027,
         218,
         2024,
         218,
         218,
         218,
         218,
         218,
         2025,
         218,
         218,
         218,
         218,
         218,
         2026,
         218,
         218,
         218,
         218,
         218,
         2027,
         218,
         218,
         218,
         218,
      },
   },
   [nir_op_fall_equal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         219,
      },
   },
   [nir_op_fall_equal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         220,
      },
   },
   [nir_op_fall_equal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         221,
      },
   },
   [nir_op_fall_equal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         222,
      },
   },
   [nir_op_fall_equal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         223,
      },
   },
   [nir_op_fany_nequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         224,
      },
   },
   [nir_op_fany_nequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         225,
      },
   },
   [nir_op_fany_nequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         226,
      },
   },
   [nir_op_fany_nequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         227,
      },
   },
   [nir_op_fany_nequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         228,
      },
   },
   [nir_op_ball_iequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         229,
         229,
         229,
         229,
         229,
         229,
         1891,
         1892,
         229,
         1891,
         229,
         229,
         229,
         1892,
         229,
         229,
      },
   },
   [nir_op_ball_iequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         230,
         230,
         230,
         230,
         230,
         230,
         1893,
         1894,
         230,
         1893,
         230,
         230,
         230,
         1894,
         230,
         230,
      },
   },
   [nir_op_ball_iequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         231,
         231,
         231,
         231,
         231,
         231,
         1895,
         1896,
         231,
         1895,
         231,
         231,
         231,
         1896,
         231,
         231,
      },
   },
   [nir_op_ball_iequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         232,
         232,
         232,
         232,
         232,
         232,
         1897,
         1898,
         232,
         1897,
         232,
         232,
         232,
         1898,
         232,
         232,
      },
   },
   [nir_op_ball_iequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         233,
         233,
         233,
         233,
         233,
         233,
         1899,
         1900,
         233,
         1899,
         233,
         233,
         233,
         1900,
         233,
         233,
      },
   },
   [nir_op_ball_fequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         234,
      },
   },
   [nir_op_ball_fequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         235,
      },
   },
   [nir_op_ball_fequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         236,
      },
   },
   [nir_op_ball_fequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         237,
      },
   },
   [nir_op_ball_fequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         238,
      },
   },
   [nir_op_bany_inequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         239,
         239,
         239,
         239,
         239,
         239,
         1905,
         1906,
         239,
         1905,
         239,
         239,
         239,
         1906,
         239,
         239,
      },
   },
   [nir_op_bany_inequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         240,
         240,
         240,
         240,
         240,
         240,
         1907,
         1908,
         240,
         1907,
         240,
         240,
         240,
         1908,
         240,
         240,
      },
   },
   [nir_op_bany_inequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         241,
         241,
         241,
         241,
         241,
         241,
         1909,
         1910,
         241,
         1909,
         241,
         241,
         241,
         1910,
         241,
         241,
      },
   },
   [nir_op_bany_inequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         242,
         242,
         242,
         242,
         242,
         242,
         1911,
         1912,
         242,
         1911,
         242,
         242,
         242,
         1912,
         242,
         242,
      },
   },
   [nir_op_bany_inequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         243,
         243,
         243,
         243,
         243,
         243,
         1913,
         1914,
         243,
         1913,
         243,
         243,
         243,
         1914,
         243,
         243,
      },
   },
   [nir_op_bany_fnequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         244,
      },
   },
   [nir_op_bany_fnequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         245,
      },
   },
   [nir_op_bany_fnequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         246,
      },
   },
   [nir_op_bany_fnequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         247,
      },
   },
   [nir_op_bany_fnequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         248,
      },
   },
   [nir_op_isub] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         249,
         249,
         250,
         250,
      },
   },
   [nir_op_urol] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         251,
      },
   },
   [nir_op_uror] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         252,
      },
   },
   [nir_op_bfi] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         253,
         254,
         253,
         253,
         254,
         253,
         253,
         254,
         253,
         2756,
         2756,
         2756,
         2756,
         2756,
         2756,
         2756,
         2756,
         2756,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         4,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         255,
         1959,
         1960,
         1961,
         1962,
         1963,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         256,
         257,
         256,
         256,
         258,
         259,
         258,
         258,
         256,
         2028,
         256,
         256,
         256,
         2667,
         256,
         256,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         260,
         1964,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         3,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         261,
         1965,
         1966,
         1967,
         1968,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         262,
         1969,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         263,
         263,
         264,
         264,
      },
   },
   [nir_op_fsin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         265,
      },
   },
   [nir_op_fcos] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         266,
      },
   },
   [nir_op_fcsel] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         267,
      },
   },
   [nir_op_f2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         268,
         1970,
         1971,
      },
   },
   [nir_op_f2ump] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         269,
         1972,
         1973,
      },
   },
   [nir_op_i2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         270,
         1993,
         1994,
      },
   },
   [nir_op_u2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         271,
         1995,
         1996,
      },
   },
   [nir_op_fround_even] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         272,
      },
   },
   [nir_op_pack_64_2x32_split] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         273,
         273,
         273,
         273,
         273,
         273,
         1319,
         273,
         273,
         273,
         273,
         273,
         273,
         273,
         273,
         1320,
      },
   },
   [nir_op_unpack_64_2x32_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         274,
         1854,
         1855,
         1856,
         1857,
         1858,
      },
   },
   [nir_op_pack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         275,
         2031,
         2600,
         2601,
      },
   },
   [nir_op_unpack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         276,
         2029,
         2030,
      },
   },
   [nir_op_unpack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         277,
         2033,
      },
   },
   [nir_op_pack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         278,
         2032,
      },
   },
   [nir_op_unpack_64_4x16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         2034,
         2602,
         2887,
      },
   },
   [nir_op_pack_64_4x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         279,
      },
   },
   [nir_op_pack_32_2x16_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         280,
      },
   },
   [nir_op_unpack_32_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         281,
         1859,
         1860,
      },
   },
   [nir_op_unpack_32_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         282,
         1861,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         283,
         1862,
         2954,
         0,
         284,
         1862,
         2954,
         1863,
         1864,
         1865,
         2955,
         0,
         283,
         1862,
         2954,
         0,
         283,
         1862,
         2954,
         0,
         284,
         1862,
         2954,
         1863,
         1864,
         1865,
         2955,
         0,
         283,
         1862,
         2954,
         0,
         283,
         1862,
         2954,
         0,
         284,
         1862,
         2954,
         1863,
         1864,
         1865,
         2955,
         0,
         283,
         1862,
         2954,
         0,
         283,
         1862,
         2954,
         0,
         284,
         1862,
         2954,
         1863,
         1864,
         1865,
         2955,
         0,
         283,
         1862,
         2954,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         285,
         1866,
         2956,
         0,
         286,
         1866,
         2956,
         1867,
         1868,
         1869,
         2957,
         0,
         285,
         1866,
         2956,
         0,
         285,
         1866,
         2956,
         0,
         286,
         1866,
         2956,
         1867,
         1868,
         1869,
         2957,
         0,
         285,
         1866,
         2956,
         0,
         285,
         1866,
         2956,
         0,
         286,
         1866,
         2956,
         1867,
         1868,
         1869,
         2957,
         0,
         285,
         1866,
         2956,
         0,
         285,
         1866,
         2956,
         0,
         286,
         1866,
         2956,
         1867,
         1868,
         1869,
         2957,
         0,
         285,
         1866,
         2956,
      },
   },
   [nir_op_pack_half_2x16_split] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         287,
         288,
         289,
         290,
      },
   },
   [nir_op_unpack_half_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         0,
         0,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         291,
         1870,
         1871,
         1872,
      },
   },
   [nir_op_unpack_half_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         292,
         1873,
      },
   },
   [nir_op_pack_half_2x16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         2009,
         2010,
         2011,
      },
   },
   [nir_op_pack_half_2x16_rtz_split] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         293,
         0,
         294,
         295,
         294,
         0,
         1663,
         0,
      },
   },
   [nir_op_pack_32_4x8_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         296,
      },
   },
   [nir_op_ussub_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         297,
         0,
         297,
      },
   },
   [nir_op_fsub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         298,
      },
   },
   [nir_op_uabs_usub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         299,
      },
   },
   [nir_op_uabs_isub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         300,
      },
   },
   [nir_op_bitz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         301,
      },
   },
   [nir_op_find_lsb] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         302,
         2035,
      },
   },
   [nir_op_ifind_msb] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         303,
         2015,
      },
   },
   [nir_op_ufind_msb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         304,
      },
   },
   [nir_op_ifind_msb_rev] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         305,
      },
   },
   [nir_op_ufind_msb_rev] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         306,
         2036,
      },
   },
   [nir_op_bitfield_reverse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         307,
      },
   },
   [nir_op_fmod] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         308,
      },
   },
   [nir_op_frem] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         309,
      },
   },
   [nir_op_uadd_carry] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         310,
      },
   },
   [nir_op_usub_borrow] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         311,
      },
   },
   [nir_op_bitfield_insert] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         312,
      },
   },
   [nir_op_ihadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         313,
      },
   },
   [nir_op_uhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         314,
      },
   },
   [nir_op_irhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         315,
      },
   },
   [nir_op_urhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         316,
      },
   },
   [nir_op_imul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         317,
      },
   },
   [nir_op_umul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         318,
      },
   },
   [nir_op_ibitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         319,
      },
   },
   [nir_op_ubitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         320,
      },
   },
   [nir_op_bitfield_select] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         321,
         0,
         0,
         321,
         0,
         1829,
         1830,
         1829,
         0,
         321,
         0,
         0,
         321,
         0,
         1829,
         1830,
         1829,
         0,
         321,
         0,
         0,
         321,
         0,
         1829,
         1830,
         1829,
      },
   },
   [nir_op_bfm] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         322,
         1874,
         0,
         0,
         322,
         1874,
         0,
         1875,
         1876,
         1877,
         1875,
         2958,
         2959,
         2960,
         2958,
      },
   },
   [nir_op_uclz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         323,
      },
   },
   [nir_op_pack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         324,
      },
   },
   [nir_op_pack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         325,
      },
   },
   [nir_op_pack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         326,
      },
   },
   [nir_op_pack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         327,
      },
   },
   [nir_op_unpack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         328,
      },
   },
   [nir_op_unpack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         329,
      },
   },
   [nir_op_unpack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         330,
      },
   },
   [nir_op_unpack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         331,
      },
   },
   [nir_op_isign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         332,
         1958,
      },
   },
   [nir_op_amul] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         333,
      },
   },
   [nir_op_umul24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         334,
      },
   },
   [nir_op_umad24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         335,
      },
   },
   [nir_op_imul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         336,
      },
   },
   [nir_op_umad24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         337,
      },
   },
   [nir_op_umul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         338,
      },
   },
   [nir_op_imad24_ir3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         339,
         340,
         341,
         340,
         341,
         342,
         343,
      },
   },
   [nir_op_imul24] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         344,
         344,
         344,
      },
   },
   [nir_op_b2b1] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         345,
      },
   },
   [nir_op_b2b8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         346,
      },
   },
   [nir_op_b2b16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         347,
      },
   },
   [nir_op_b2b32] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         348,
      },
   },
   [nir_op_ldexp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         349,
      },
   },
   [nir_op_fddx] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         350,
      },
   },
   [nir_op_fddx_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         351,
      },
   },
   [nir_op_fddx_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         352,
      },
   },
   [nir_op_fddy] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         353,
      },
   },
   [nir_op_fddy_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         354,
      },
   },
   [nir_op_fddy_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         355,
      },
   },
   [nir_op_umul_low] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         356,
         356,
         356,
      },
   },
   [nir_op_imadsh_mix16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         357,
         357,
         358,
         358,
         359,
         359,
      },
   },
   [nir_op_fsin_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         360,
      },
   },
   [nir_op_fcos_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         361,
      },
   },
   [nir_op_fsin_mdg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         362,
      },
   },
   [nir_op_fcos_mdg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         363,
      },
   },
   [nir_op_fsin_agx] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         364,
      },
   },
   [nir_op_fisnormal] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         365,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_transform_offsets[] = {
   0,
   0,
   0,
   1,
   0,
   12,
   15,
   17,
   21,
   23,
   25,
   27,
   30,
   34,
   37,
   42,
   0,
   47,
   54,
   56,
   0,
   61,
   65,
   0,
   0,
   0,
   67,
   0,
   69,
   71,
   77,
   84,
   90,
   97,
   100,
   104,
   107,
   111,
   114,
   118,
   124,
   131,
   141,
   153,
   168,
   181,
   197,
   203,
   210,
   219,
   229,
   239,
   250,
   252,
   264,
   266,
   274,
   276,
   282,
   284,
   0,
   290,
   296,
   298,
   305,
   308,
   311,
   313,
   317,
   320,
   325,
   328,
   333,
   337,
   343,
   345,
   347,
   350,
   353,
   357,
   361,
   363,
   365,
   368,
   371,
   375,
   379,
   381,
   384,
   388,
   390,
   393,
   396,
   400,
   404,
   409,
   411,
   413,
   416,
   419,
   423,
   427,
   429,
   431,
   434,
   437,
   441,
   445,
   447,
   450,
   454,
   456,
   459,
   463,
   465,
   467,
   470,
   473,
   476,
   479,
   484,
   489,
   491,
   494,
   497,
   502,
   506,
   511,
   516,
   540,
   546,
   553,
   560,
   586,
   588,
   0,
   0,
   591,
   596,
   601,
   0,
   604,
   606,
   608,
   611,
   613,
   616,
   618,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   622,
   0,
   624,
   630,
   633,
   0,
   641,
   646,
   651,
   656,
   661,
   666,
   671,
   676,
   681,
   683,
   686,
   691,
   697,
   699,
   702,
   706,
   711,
   716,
   722,
   724,
   726,
   728,
   0,
   731,
   733,
   735,
   737,
   739,
   741,
   743,
   745,
   749,
   751,
   755,
   757,
   0,
   0,
   761,
   765,
   771,
   0,
   0,
   777,
   779,
   783,
   785,
   788,
   790,
   793,
   795,
   801,
   803,
   805,
   807,
   809,
   811,
   813,
   815,
   817,
   819,
   821,
   823,
   825,
   827,
   829,
   831,
   833,
   835,
   837,
   839,
   841,
   843,
   845,
   847,
   849,
   851,
   853,
   855,
   857,
   859,
   861,
   863,
   865,
   867,
   869,
   871,
   873,
   878,
   883,
   887,
   0,
   923,
   925,
   930,
   933,
   939,
   0,
   0,
   941,
   943,
   946,
   948,
   950,
   0,
   0,
   0,
   0,
   952,
   954,
   956,
   0,
   0,
   0,
   0,
   0,
   958,
   961,
   964,
   967,
   969,
   978,
   980,
   988,
   991,
   994,
   997,
   1000,
   1002,
   0,
   0,
   0,
   0,
   1004,
   1007,
   1009,
   1011,
   1013,
   1015,
   1018,
   1022,
   0,
   0,
   0,
   1025,
   1027,
   1029,
   1031,
   1033,
   1037,
   1040,
   1043,
   1046,
   1049,
   1051,
   1053,
   1056,
   1059,
   0,
   1061,
   1063,
   1065,
   1067,
   1069,
   1071,
   1073,
   1075,
   1077,
   1079,
   1081,
   1083,
   1085,
   1087,
   1090,
   1093,
   1096,
   1098,
   1101,
   1105,
   1109,
   1114,
   1118,
   1120,
   1122,
   1124,
   1126,
   1130,
   1132,
   1134,
   1136,
   1138,
   1140,
   1142,
   1144,
   1146,
   1148,
   0,
   0,
   0,
   0,
   0,
   1151,
   1155,
   1157,
   1160,
   1163,
   1165,
   1167,
   1179,
   1192,
   1204,
   1216,
   1228,
   1241,
   1254,
   1266,
   1278,
   1282,
   1286,
   1289,
   1292,
   1294,
   1299,
   1303,
   1307,
   1311,
   1315,
   1318,
   1320,
   1325,
   0,
   1327,
   1329,
   1331,
   1333,
   1335,
   1338,
   1341,
   1344,
   1346,
   1348,
   1350,
   1353,
   1356,
   1359,
   1363,
   1367,
   1371,
   1375,
   1379,
   1384,
   1387,
   1391,
   1395,
   1398,
   1401,
   1403,
   1405,
   1407,
   1410,
   1413,
   1416,
   1419,
   1422,
   1425,
   1429,
   1433,
   1437,
   1440,
   1443,
   1445,
   1447,
   1449,
   1451,
   1454,
   1457,
   1460,
   1462,
   1464,
   1467,
   1470,
   1473,
   1477,
   1481,
   1485,
   1488,
   1491,
   1494,
   1497,
   1501,
   1505,
   1509,
   1512,
   1515,
   1518,
   1522,
   1526,
   1530,
   1533,
   1536,
   1540,
   1544,
   1548,
   1551,
   1554,
   1559,
   1564,
   1568,
   1572,
   1577,
   1581,
   1585,
   1589,
   1593,
   1596,
   1598,
   1600,
   1604,
   1610,
   1615,
   1622,
   1627,
   1635,
   1643,
   1649,
   1654,
   1658,
   1663,
   1668,
   1674,
   1682,
   1690,
   1698,
   1706,
   1716,
   1722,
   1729,
   1735,
   1747,
   1752,
   1757,
   1764,
   1772,
   1780,
   1788,
   1796,
   1804,
   1812,
   1818,
   1825,
   1831,
   1839,
   1845,
   1852,
   1858,
   1866,
   1871,
   1877,
   1882,
   1889,
   1894,
   1900,
   1905,
   1912,
   1920,
   1925,
   1933,
   1938,
   1943,
   1948,
   1953,
   1958,
   1963,
   1968,
   1973,
   1978,
   1983,
   1989,
   1997,
   2004,
   2013,
   2020,
   2030,
   2040,
   2048,
   2055,
   2061,
   2068,
   2075,
   2083,
   2093,
   2103,
   2113,
   2123,
   2135,
   2143,
   2152,
   2160,
   2174,
   2181,
   2188,
   2197,
   2207,
   2217,
   2227,
   2237,
   2247,
   2257,
   2265,
   2274,
   2282,
   2292,
   2300,
   2309,
   2317,
   2327,
   2334,
   2342,
   2349,
   2358,
   2365,
   2373,
   2380,
   2389,
   2399,
   2406,
   2416,
   2423,
   2430,
   2437,
   2444,
   2451,
   2458,
   2465,
   2472,
   2479,
   2486,
   2492,
   2499,
   2505,
   2512,
   2518,
   2523,
   2547,
   2552,
   2559,
   2564,
   2569,
   2575,
   2580,
   2587,
   2593,
   2618,
   2624,
   2632,
   2638,
   2647,
   2656,
   2663,
   2669,
   2674,
   2680,
   2686,
   2693,
   2702,
   2711,
   2720,
   2729,
   2740,
   2747,
   2755,
   2762,
   2775,
   2781,
   2787,
   2795,
   2804,
   2813,
   2822,
   2831,
   2840,
   2849,
   2856,
   2864,
   2871,
   2880,
   2887,
   2895,
   2902,
   2911,
   2917,
   2924,
   2930,
   2938,
   2944,
   2951,
   2957,
   2965,
   2974,
   2980,
   2989,
   2995,
   3001,
   3007,
   3013,
   3019,
   3025,
   3031,
   3037,
   3043,
   3049,
   3055,
   3062,
   3068,
   3076,
   3083,
   3109,
   3116,
   3125,
   3132,
   3142,
   3152,
   3160,
   3167,
   3173,
   3180,
   3187,
   3195,
   3205,
   3215,
   3225,
   3235,
   3247,
   3255,
   3264,
   3272,
   3286,
   3293,
   3300,
   3309,
   3319,
   3329,
   3339,
   3349,
   3359,
   3369,
   3377,
   3386,
   3394,
   3404,
   3412,
   3421,
   3429,
   3439,
   3446,
   3454,
   3461,
   3470,
   3477,
   3485,
   3492,
   3501,
   3511,
   3518,
   3528,
   3535,
   3542,
   3549,
   3556,
   3563,
   3570,
   3577,
   3584,
   3591,
   3598,
   3606,
   3613,
   3621,
   3629,
   3638,
   3644,
   3651,
   3657,
   3665,
   3672,
   3698,
   3705,
   3714,
   3721,
   3731,
   3741,
   3749,
   3756,
   3762,
   3769,
   3776,
   3784,
   3794,
   3804,
   3814,
   3824,
   3836,
   3844,
   3853,
   3861,
   3875,
   3882,
   3889,
   3898,
   3908,
   3918,
   3928,
   3938,
   3948,
   3958,
   3966,
   3975,
   3983,
   3993,
   4001,
   4010,
   4018,
   4028,
   4035,
   4043,
   4050,
   4059,
   4066,
   4074,
   4081,
   4090,
   4100,
   4107,
   4117,
   4124,
   4131,
   4138,
   4145,
   4152,
   4159,
   4166,
   4173,
   4180,
   4187,
   4193,
   4200,
   4206,
   4214,
   4221,
   4247,
   4254,
   4263,
   4270,
   4280,
   4290,
   4298,
   4305,
   4311,
   4318,
   4325,
   4333,
   4343,
   4353,
   4363,
   4373,
   4385,
   4393,
   4402,
   4410,
   4424,
   4431,
   4438,
   4447,
   4457,
   4467,
   4477,
   4487,
   4497,
   4507,
   4515,
   4524,
   4532,
   4542,
   4550,
   4559,
   4567,
   4577,
   4584,
   4592,
   4599,
   4608,
   4615,
   4623,
   4630,
   4639,
   4649,
   4656,
   4666,
   4673,
   4680,
   4687,
   4694,
   4701,
   4708,
   4715,
   4722,
   4729,
   4736,
   4742,
   4749,
   4755,
   4763,
   4770,
   4796,
   4803,
   4812,
   4819,
   4829,
   4839,
   4847,
   4854,
   4860,
   4867,
   4874,
   4882,
   4892,
   4902,
   4912,
   4922,
   4934,
   4942,
   4951,
   4959,
   4973,
   4980,
   4987,
   4996,
   5006,
   5016,
   5026,
   5036,
   5046,
   5056,
   5064,
   5073,
   5081,
   5091,
   5099,
   5108,
   5116,
   5126,
   5133,
   5141,
   5148,
   5157,
   5164,
   5172,
   5179,
   5188,
   5198,
   5205,
   5215,
   5222,
   5229,
   5236,
   5243,
   5250,
   5257,
   5264,
   5271,
   5278,
   5285,
   5291,
   5298,
   5304,
   5312,
   5319,
   5345,
   5352,
   5361,
   5368,
   5378,
   5388,
   5396,
   5403,
   5409,
   5416,
   5423,
   5431,
   5441,
   5451,
   5461,
   5471,
   5483,
   5491,
   5500,
   5508,
   5522,
   5529,
   5536,
   5545,
   5555,
   5565,
   5575,
   5585,
   5595,
   5605,
   5613,
   5622,
   5630,
   5640,
   5648,
   5657,
   5665,
   5675,
   5682,
   5690,
   5697,
   5706,
   5713,
   5721,
   5728,
   5737,
   5747,
   5754,
   5764,
   5771,
   5778,
   5785,
   5792,
   5799,
   5806,
   5813,
   5820,
   5827,
   5834,
   5840,
   5847,
   5853,
   5861,
   5868,
   5894,
   5901,
   5910,
   5917,
   5927,
   5937,
   5945,
   5952,
   5958,
   5965,
   5972,
   5980,
   5990,
   6000,
   6010,
   6020,
   6032,
   6040,
   6049,
   6057,
   6071,
   6078,
   6085,
   6094,
   6104,
   6114,
   6124,
   6134,
   6144,
   6154,
   6162,
   6171,
   6179,
   6189,
   6197,
   6206,
   6214,
   6224,
   6231,
   6239,
   6246,
   6255,
   6262,
   6270,
   6277,
   6286,
   6296,
   6303,
   6313,
   6320,
   6327,
   6334,
   6341,
   6348,
   6355,
   6362,
   6369,
   6376,
   6383,
   0,
   6388,
   6394,
   6396,
   6400,
   6407,
   6411,
   6418,
   6423,
   6430,
   6439,
   6444,
   6449,
   6456,
   6465,
   6470,
   6473,
   6477,
   6481,
   6483,
   6487,
   6493,
   6495,
   6498,
   6501,
   6504,
   6508,
   6512,
   6515,
   6518,
   6531,
   6548,
   6566,
   6579,
   6593,
   6606,
   6622,
   6635,
   6649,
   6663,
   6676,
   6689,
   6701,
   6704,
   6708,
   6712,
   6716,
   6721,
   6726,
   6730,
   6734,
   6738,
   6742,
   6746,
   6751,
   6756,
   6760,
   6764,
   6767,
   6777,
   6783,
   6786,
   6790,
   6794,
   6798,
   6801,
   6808,
   6811,
   6814,
   6818,
   6823,
   6828,
   6832,
   6836,
   6841,
   6847,
   6858,
   6870,
   6876,
   6888,
   6901,
   6905,
   6910,
   6921,
   6933,
   6938,
   6950,
   6963,
   6965,
   6970,
   6975,
   6980,
   6984,
   6988,
   6993,
   6999,
   7004,
   7009,
   7014,
   7019,
   7024,
   7028,
   7032,
   7036,
   7039,
   7042,
   7045,
   7048,
   7050,
   7054,
   7058,
   7061,
   7064,
   7073,
   7082,
   7092,
   7101,
   7110,
   7118,
   7128,
   7138,
   7147,
   7156,
   7159,
   7161,
   7165,
   7169,
   7173,
   7177,
   7180,
   7185,
   7190,
   7194,
   7198,
   7202,
   7206,
   7210,
   7213,
   7218,
   7223,
   7227,
   7231,
   7239,
   7243,
   7246,
   7251,
   7256,
   7261,
   7265,
   7269,
   7279,
   7282,
   7284,
   7286,
   7291,
   7294,
   7297,
   7301,
   7304,
   7309,
   7314,
   7318,
   7322,
   7327,
   7331,
   7335,
   7338,
   7341,
   7347,
   7358,
   7370,
   7376,
   7388,
   7401,
   7405,
   7410,
   7420,
   7431,
   7436,
   7447,
   7459,
   7461,
   7466,
   7470,
   7474,
   7478,
   7482,
   7487,
   7493,
   7498,
   7503,
   7508,
   7513,
   7518,
   7522,
   7526,
   7530,
   7533,
   7536,
   7540,
   7544,
   7547,
   7550,
   7553,
   7561,
   7569,
   7576,
   7583,
   7590,
   7598,
   7601,
   7604,
   7607,
   7612,
   7618,
   7623,
   7628,
   7633,
   7638,
   7643,
   7648,
   7653,
   7657,
   7661,
   7665,
   7669,
   7673,
   0,
   7678,
   7684,
   7686,
   7691,
   7702,
   7711,
   7716,
   7726,
   7733,
   7740,
   7745,
   7752,
   0,
   7759,
   7765,
   7770,
   7780,
   7795,
   7810,
   7825,
   7830,
   7835,
   7840,
   7848,
   7863,
   0,
   0,
   7871,
   7874,
   7876,
   7879,
   7882,
   7885,
   7888,
   7891,
   7894,
   0,
   7897,
   0,
   7899,
   0,
   0,
   7902,
   7910,
   7919,
   7926,
   7934,
   7942,
   7951,
   7958,
   7966,
   7976,
   7983,
   7990,
   7992,
   0,
   7994,
   0,
   0,
   7997,
   7999,
   8003,
   8005,
   0,
   8007,
   0,
   0,
   8010,
   8013,
   8016,
   8018,
   0,
   8020,
   8022,
   8024,
   8026,
   8033,
   8041,
   8045,
   8050,
   8055,
   8061,
   8066,
   8072,
   8077,
   8079,
   8081,
   8085,
   8087,
   0,
   8089,
   8091,
   8094,
   8097,
   8100,
   8103,
   8106,
   8109,
   0,
   0,
   0,
   0,
   8111,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   8113,
   8121,
   8129,
   8142,
   8155,
   8164,
   8171,
   8181,
   8189,
   8197,
   8211,
   8225,
   8238,
   8254,
   8269,
   8283,
   8297,
   8314,
   8330,
   8345,
   8357,
   8372,
   8385,
   8398,
   8411,
   8427,
   8441,
   8455,
   8464,
   8472,
   8483,
   8492,
   8501,
   8508,
   8518,
   8526,
   8534,
   8539,
   8545,
   8550,
   8557,
   8565,
   8571,
   8579,
   8588,
   8594,
   8601,
   8607,
   8615,
   8624,
   8631,
   8640,
   8650,
   8656,
   8664,
   8673,
   8682,
   8693,
   0,
   8705,
   8707,
   8711,
   8714,
   8718,
   0,
   8722,
   8726,
   8728,
   8730,
   8734,
   8741,
   8744,
   8747,
   8750,
   8754,
   8759,
   8764,
   8768,
   8772,
   8776,
   8780,
   8784,
   8789,
   8794,
   8799,
   8804,
   8808,
   8812,
   8818,
   8824,
   8830,
   8836,
   8842,
   8848,
   8855,
   8861,
   8868,
   8874,
   8880,
   8886,
   8892,
   8898,
   8904,
   8911,
   8918,
   8925,
   8932,
   8938,
   8945,
   8951,
   8957,
   8963,
   8969,
   8976,
   8983,
   8990,
   8996,
   9002,
   9008,
   9015,
   9022,
   9029,
   9042,
   9048,
   9054,
   9060,
   9066,
   9073,
   9079,
   9085,
   9091,
   9097,
   9103,
   9109,
   9115,
   9121,
   9128,
   9135,
   9142,
   9149,
   9155,
   9162,
   9168,
   9175,
   9181,
   9187,
   9193,
   9200,
   9207,
   9214,
   9220,
   9226,
   9232,
   9239,
   9246,
   9253,
   9266,
   9269,
   9272,
   9276,
   9281,
   9286,
   9290,
   9294,
   9298,
   9302,
   9306,
   9311,
   9316,
   9321,
   9326,
   9330,
   9334,
   9340,
   9346,
   9349,
   9352,
   9358,
   9365,
   9372,
   9379,
   9394,
   9409,
   9416,
   9423,
   9437,
   9449,
   9461,
   9467,
   9474,
   9480,
   9487,
   9493,
   9499,
   9505,
   9511,
   9517,
   9523,
   9527,
   9533,
   9545,
   9557,
   9560,
   9563,
   9566,
   9568,
   9572,
   9576,
   9580,
   9584,
   9588,
   9591,
   9596,
   9600,
   9604,
   9607,
   9611,
   9614,
   9617,
   9620,
   9623,
   9626,
   9629,
   9633,
   9636,
   9639,
   9643,
   9646,
   9650,
   9654,
   9658,
   9661,
   9664,
   9669,
   9675,
   9682,
   9689,
   9697,
   9704,
   9713,
   9720,
   9725,
   9732,
   9739,
   9749,
   9759,
   9767,
   9778,
   9789,
   9799,
   9810,
   9822,
   9833,
   9845,
   9858,
   9872,
   9885,
   9892,
   9900,
   9908,
   0,
   9916,
   0,
   0,
   9918,
   9920,
   0,
   0,
   0,
   0,
   0,
   9925,
   9928,
   9931,
   9934,
   9937,
   9940,
   9947,
   9954,
   9961,
   9976,
   9991,
   9998,
   10005,
   10019,
   10025,
   10037,
   10049,
   10055,
   10061,
   10068,
   10075,
   10081,
   10087,
   10093,
   10099,
   10105,
   10111,
   10115,
   10121,
   10133,
   10145,
   10148,
   10151,
   10154,
   10156,
   10159,
   10162,
   10165,
   10167,
   10170,
   10174,
   10178,
   10181,
   10185,
   10188,
   10191,
   10194,
   10197,
   10200,
   10203,
   10206,
   10209,
   10212,
   10216,
   10220,
   10224,
   10228,
   10230,
   10232,
   10236,
   10243,
   10246,
   10256,
   10261,
   10266,
   10271,
   10276,
   10286,
   10296,
   10301,
   10306,
   10311,
   10313,
   10315,
   10317,
   10319,
   10321,
   10323,
   10325,
   10328,
   10331,
   10334,
   10339,
   10344,
   10349,
   10356,
   10361,
   10367,
   10377,
   10379,
   10382,
   10385,
   10388,
   10391,
   10394,
   10398,
   10402,
   10406,
   10410,
   10414,
   10417,
   10421,
   10425,
   10429,
   10433,
   10438,
   10441,
   10445,
   10449,
   10453,
   10457,
   10462,
   10471,
   10493,
   10496,
   10499,
   10502,
   10505,
   10508,
   10511,
   10515,
   10518,
   10522,
   10530,
   10532,
   10535,
   10538,
   10541,
   10544,
   10548,
   10552,
   10556,
   10559,
   10562,
   10566,
   10570,
   10574,
   10578,
   10581,
   10585,
   10589,
   10593,
   10597,
   10605,
   10627,
   10629,
   10632,
   10634,
   10638,
   10643,
   10647,
   10650,
   10653,
   10655,
   10658,
   10661,
   10663,
   10666,
   10678,
   10685,
   10692,
   10699,
   10706,
   10752,
   10771,
   10778,
   10785,
   10797,
   10803,
   10810,
   10817,
   10824,
   10831,
   10838,
   10844,
   10860,
   10909,
   10916,
   10923,
   10930,
   10942,
   10945,
   10948,
   10951,
   10954,
   10957,
   10961,
   10966,
   10970,
   10972,
   10974,
   10977,
   10980,
   10982,
   10984,
   10987,
   10990,
   10993,
   10996,
   11000,
   11003,
   11005,
   11007,
   11009,
   11012,
   11014,
   11016,
   11018,
   0,
   11020,
   11022,
   11024,
   11026,
   11028,
   11030,
   11032,
   11034,
   11036,
   11039,
   11042,
   11045,
   11048,
   11051,
   11054,
   11057,
   11060,
   11063,
   11066,
   0,
   0,
   0,
   11068,
   11071,
   11074,
   11077,
   11080,
   11083,
   11086,
   11089,
   11092,
   11095,
   11098,
   11104,
   11110,
   11116,
   11123,
   11130,
   11141,
   11148,
   11154,
   11162,
   11169,
   11176,
   11185,
   11191,
   11199,
   11206,
   11213,
   11222,
   11234,
   11259,
   11265,
   11271,
   11277,
   11284,
   11291,
   11302,
   11309,
   11315,
   11323,
   11330,
   11337,
   11346,
   11352,
   11360,
   11367,
   11374,
   11383,
   11395,
   11420,
   11426,
   11432,
   11438,
   11444,
   11450,
   11453,
   11455,
   11457,
   11459,
   11461,
   11463,
   11466,
   11468,
   11470,
   11472,
   11474,
   11476,
   11478,
   11480,
   11482,
   11484,
   11486,
   11513,
   11515,
   11518,
   11520,
   11524,
   11551,
   11554,
   11557,
   11562,
   11565,
   11568,
   11570,
   11572,
   11574,
   11576,
   11578,
   11580,
   11582,
   11584,
   11586,
   11588,
   11590,
   11594,
   11599,
   11605,
   11610,
   11616,
   11623,
   11626,
   11630,
   11634,
   11639,
   11642,
   0,
   0,
   0,
   0,
   0,
   0,
   11647,
   11652,
   11656,
   11660,
   11664,
   11668,
   11672,
   11676,
   11680,
   11684,
   11688,
   11692,
   11696,
   11700,
   11706,
   11708,
   11710,
   11712,
   11714,
   11716,
   11718,
   11722,
   11724,
   11726,
   11728,
   11732,
   11736,
   11740,
   11744,
   11748,
   11753,
   11755,
   11758,
   11761,
   11764,
   11767,
   11770,
   11775,
   11780,
   11785,
   11790,
   11794,
   11798,
   11803,
   11808,
   11813,
   11818,
   11822,
   11826,
   11830,
   11834,
   11837,
   11840,
   11845,
   11850,
   11855,
   11860,
   11863,
   11866,
   11871,
   11876,
   11881,
   11886,
   11889,
   11892,
   11897,
   11902,
   11907,
   11912,
   11915,
   11918,
   11923,
   11928,
   11933,
   11938,
   11942,
   11946,
   11952,
   11958,
   11964,
   11970,
   11974,
   11978,
   11984,
   11990,
   11996,
   12002,
   12006,
   12010,
   12016,
   12022,
   12028,
   12034,
   12037,
   12040,
   12045,
   12050,
   12055,
   12060,
   12063,
   12066,
   12071,
   12076,
   12081,
   12086,
   12091,
   12096,
   12101,
   12106,
   12111,
   12116,
   12121,
   12126,
   12133,
   12140,
   12147,
   12154,
   12161,
   12168,
   12170,
   12177,
   12185,
   12192,
   12201,
   12208,
   12216,
   12223,
   12232,
   12239,
   12247,
   12254,
   12263,
   12270,
   12278,
   12285,
   12294,
   12303,
   12313,
   12322,
   12333,
   12342,
   12352,
   12361,
   12372,
   12381,
   12391,
   12400,
   12411,
   12420,
   12430,
   12439,
   12450,
   12456,
   12464,
   12473,
   12481,
   12491,
   12499,
   12508,
   12516,
   12526,
   12534,
   12543,
   12551,
   12561,
   12569,
   12578,
   12586,
   12596,
   12605,
   12615,
   12624,
   12635,
   12644,
   12654,
   12663,
   12674,
   12683,
   12693,
   12702,
   12713,
   12722,
   12732,
   12741,
   12752,
   12761,
   12771,
   12780,
   12791,
   12800,
   12810,
   12819,
   12830,
   12839,
   12849,
   12858,
   12869,
   12878,
   12888,
   12897,
   12908,
   12917,
   12927,
   12936,
   12947,
   12956,
   12966,
   12975,
   12986,
   12995,
   13005,
   13014,
   13025,
   13034,
   13044,
   13053,
   13064,
   13071,
   13082,
   13091,
   13101,
   13110,
   13121,
   13130,
   13140,
   13149,
   13160,
   13169,
   13179,
   13188,
   13199,
   13208,
   13218,
   13227,
   13238,
   13247,
   13257,
   13266,
   13277,
   13286,
   13296,
   13305,
   13316,
   13325,
   13335,
   13344,
   13355,
   13364,
   13374,
   13383,
   13394,
   13405,
   13414,
   13424,
   13433,
   13444,
   13453,
   13463,
   13472,
   13483,
   13492,
   13502,
   13511,
   13522,
   13531,
   13541,
   13550,
   13561,
   13570,
   13579,
   13588,
   13597,
   13608,
   13623,
   13634,
   13649,
   13657,
   13665,
   13673,
   13681,
   13689,
   13693,
   13697,
   13701,
   13705,
   13717,
   13730,
   13744,
   13758,
   13772,
   13786,
   13791,
   13796,
   13801,
   13806,
   13812,
   13818,
   13824,
   13830,
   13836,
   13842,
   13848,
   13854,
   13859,
   13864,
   13869,
   13874,
   13879,
   13884,
   13889,
   13894,
   13901,
   13908,
   13915,
   13922,
   13929,
   0,
   13936,
   13938,
   13943,
   13948,
   13950,
   13954,
   13959,
   13964,
   13967,
   0,
   13971,
   13976,
   13978,
   13984,
   13989,
   13998,
   14004,
   14011,
   14014,
   14017,
   14021,
   14025,
   14029,
   14033,
   14038,
   14044,
   14048,
   14053,
   14057,
   14061,
   14065,
   14070,
   14076,
   14080,
   14085,
   14089,
   0,
   0,
   14091,
   14102,
   14111,
   14121,
   14128,
   14135,
   14139,
   0,
   0,
   14142,
   14147,
   14150,
   14152,
   14154,
   14158,
   0,
   0,
   14160,
   14162,
   14166,
   14169,
   14172,
   14176,
   0,
   0,
   14178,
   14182,
   0,
   0,
   14184,
   14186,
   14191,
   14196,
   14200,
   14204,
   14206,
   0,
   14208,
   14219,
   14227,
   14238,
   14247,
   14256,
   14272,
   14285,
   14301,
   14315,
   14329,
   14339,
   14350,
   14358,
   14369,
   14378,
   14387,
   14399,
   14409,
   14422,
   14433,
   14444,
   14458,
   14469,
   14477,
   14488,
   14497,
   14506,
   14515,
   14527,
   14537,
   14547,
   14552,
   14557,
   14562,
   14567,
   14578,
   14589,
   14600,
   14611,
   14615,
   14621,
   14627,
   14633,
   14639,
   14645,
   14651,
   14657,
   14663,
   14669,
   14675,
   14681,
   14687,
   14691,
   14697,
   14703,
   14709,
   14715,
   14722,
   14727,
   14732,
   14737,
   14742,
   14749,
   14756,
   14763,
   14770,
   14772,
   14779,
   14786,
   14793,
   14800,
   14806,
   14812,
   14818,
   14824,
   14830,
   14836,
   14842,
   14848,
   14854,
   14859,
   14864,
   14867,
   14875,
   14883,
   14891,
   14899,
   14907,
   14915,
   14920,
   14923,
   14938,
   14946,
   14948,
   14950,
   14952,
   14954,
   14956,
   14958,
   14961,
   14963,
   14968,
   14973,
   14978,
   14983,
   14990,
   15003,
   15016,
   15028,
   15042,
   15054,
   15069,
   15084,
   15099,
   15115,
   15132,
   15134,
   15138,
   15141,
   15147,
   15153,
   15159,
   15165,
   15173,
   15177,
   15180,
   15186,
   15192,
   15198,
   15204,
   15212,
   15222,
   15225,
   15229,
   15232,
   15238,
   15244,
   15250,
   15256,
   15264,
   15269,
   15273,
   15280,
   15287,
   15294,
   15301,
   15310,
   15315,
   15319,
   15326,
   15333,
   15340,
   15347,
   15356,
   15360,
   15363,
   15369,
   15375,
   15381,
   15387,
   15395,
   15399,
   15402,
   15408,
   15414,
   15420,
   15426,
   15434,
   15437,
   15443,
   15449,
   15455,
   15461,
   15469,
   15472,
   15477,
   15482,
   15487,
   15492,
   15499,
   15510,
   15518,
   15526,
   15534,
   15544,
   15552,
   15560,
   15570,
   15578,
   15588,
   15593,
   15603,
   15606,
   0,
   15608,
   15610,
   15613,
   15618,
   15626,
   15635,
   15641,
   15647,
   15653,
   15659,
   15664,
   15672,
   15679,
   15684,
   15686,
   15688,
   15690,
   15698,
   15700,
   0,
   0,
   0,
   15705,
   15711,
   15718,
   15726,
   15733,
   15740,
   15747,
   15752,
   15755,
   15759,
   15763,
   15767,
   15771,
   15775,
   15779,
   15783,
   15785,
   15789,
   15792,
   15796,
   15800,
   15802,
   15806,
   15810,
   15815,
   15820,
   15823,
   15828,
   15831,
   15836,
   15841,
   15846,
   15849,
   15854,
   15858,
   15864,
   15868,
   15872,
   15878,
   15882,
   15888,
   15895,
   15903,
   15910,
   15916,
   15920,
   15924,
   15928,
   15933,
   15938,
   15943,
   15948,
   15953,
   15958,
   15963,
   15968,
   15973,
   15977,
   15981,
   15985,
   15990,
   15996,
   16000,
   16006,
   16013,
   16016,
   16019,
   16022,
   16026,
   16030,
   16034,
   16038,
   16043,
   16048,
   16053,
   16058,
   16062,
   16066,
   16072,
   16076,
   16080,
   16082,
   16086,
   16090,
   16092,
   16095,
   16099,
   16102,
   16107,
   16112,
   16115,
   16120,
   16125,
   16128,
   16132,
   16138,
   16142,
   16146,
   16154,
   16157,
   16164,
   16168,
   16173,
   16179,
   16187,
   16191,
   16196,
   16201,
   16206,
   16213,
   16215,
   16219,
   16223,
   16229,
   16231,
   16235,
   16238,
   16240,
   16244,
   16247,
   16251,
   16253,
   16255,
   16259,
   16263,
   16267,
   16273,
   16276,
   16279,
   16282,
   16284,
   16288,
   16292,
   16294,
   16299,
   16304,
   16309,
   16314,
   16319,
   16328,
   16337,
   16346,
   16355,
   16362,
   16369,
   16376,
   16383,
   16390,
   16397,
   16404,
   16411,
   16417,
   16423,
   16429,
   16435,
   16441,
   16447,
   16453,
   16459,
   16465,
   16471,
   16477,
   16483,
   16491,
   16499,
   16507,
   16515,
   16523,
   16531,
   16536,
   16586,
   16636,
   16656,
   16661,
   16667,
   16673,
   16678,
   16683,
   16689,
   16694,
   16700,
   16705,
   16711,
   16716,
   16723,
   16729,
   16735,
   16740,
   16747,
   16753,
   16759,
   16764,
   16771,
   16777,
   16783,
   16788,
   16795,
   16801,
   16803,
   16806,
   16810,
   16815,
   16817,
   16820,
   16823,
   16827,
   16829,
   16831,
   16834,
   16838,
   16843,
   16846,
   16850,
   16854,
   16860,
   16863,
   16869,
   16875,
   16892,
   16908,
   16911,
   16914,
   16918,
   16925,
   16931,
   16938,
   16942,
   16952,
   16956,
   16967,
   16979,
   16984,
   16989,
   16997,
   17005,
   17016,
   17021,
   17029,
   17034,
   17042,
   17050,
   17055,
   17063,
   17068,
   17076,
   17082,
   17091,
   17099,
   17110,
   17121,
   17126,
   17135,
   0,
   17147,
   17155,
   17163,
   0,
   0,
   0,
   17171,
   17178,
   17182,
   17188,
   17203,
   0,
   0,
   17205,
   17212,
   0,
   17219,
   0,
   17221,
   17223,
   0,
   0,
   17225,
   17228,
   0,
   0,
   0,
   0,
   17231,
   17235,
   17239,
   17241,
   0,
   0,
   0,
   17243,
   0,
   17245,
   17256,
   17272,
   17284,
   17295,
   17309,
   17321,
   17326,
   17331,
   17333,
   17340,
   17347,
   17350,
   17353,
   17357,
   17361,
   17367,
   17371,
   17375,
   17379,
   17383,
   17387,
   17390,
   17394,
   17399,
   17403,
   17408,
   17413,
   17418,
   17422,
   17425,
   17428,
   17432,
   17436,
   17440,
   17444,
   17449,
   17454,
   17461,
   17468,
   17473,
   17475,
   17478,
   17480,
   17483,
   17485,
   17487,
   17490,
   17494,
   17498,
   17500,
   17505,
   17511,
   17519,
   17528,
   17535,
   17538,
   17542,
   17550,
   17558,
   17563,
   17571,
   17582,
   17593,
   17607,
   17621,
   17626,
   17637,
   17641,
   17645,
   17650,
   0,
   17653,
   17656,
   17662,
   17675,
   17683,
   17690,
   17707,
   17723,
   17730,
   17736,
   17740,
   17742,
   17745,
   17751,
   17764,
   17781,
   17797,
   17804,
   17810,
   17814,
   17816,
   17819,
   17825,
   17838,
   17855,
   17871,
   17878,
   17884,
   17889,
};

static const nir_algebraic_table nir_opt_algebraic_table = {
   .transforms = nir_opt_algebraic_transforms,
   .transform_offsets = nir_opt_algebraic_transform_offsets,
   .pass_op_table = nir_opt_algebraic_pass_op_table,
   .values = nir_opt_algebraic_values,
   .expression_cond = nir_opt_algebraic_expression_cond,
   .variable_cond = nir_opt_algebraic_variable_cond,
};

bool
nir_opt_algebraic(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[128];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(6399 == ARRAY_SIZE(nir_opt_algebraic_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !options->lower_ftrunc || options->lower_ffloor;
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[7] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[8] = !options->lower_iadd_sat;
   condition_flags[9] = !options->lower_uadd_sat;
   condition_flags[10] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[11] = !options->has_sdot_4x8;
   condition_flags[12] = !options->has_udot_4x8;
   condition_flags[13] = !options->has_sudot_4x8;
   condition_flags[14] = !options->has_dot_2x16;
   condition_flags[15] = options->lower_flrp16;
   condition_flags[16] = !options->lower_flrp16;
   condition_flags[17] = options->lower_flrp32;
   condition_flags[18] = !options->lower_flrp32;
   condition_flags[19] = options->lower_flrp64;
   condition_flags[20] = !options->lower_flrp64;
   condition_flags[21] = options->lower_ftrunc;
   condition_flags[22] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[23] = options->lower_ffloor;
   condition_flags[24] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[25] = !options->lower_ffloor;
   condition_flags[26] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[27] = options->lower_ffract;
   condition_flags[28] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[29] = options->lower_fceil;
   condition_flags[30] = options->lower_ffma16;
   condition_flags[31] = options->lower_ffma32;
   condition_flags[32] = options->lower_ffma64;
   condition_flags[33] = options->fuse_ffma16;
   condition_flags[34] = options->fuse_ffma32;
   condition_flags[35] = options->fuse_ffma64;
   condition_flags[36] = options->lower_fdph;
   condition_flags[37] = !options->lower_fdph;
   condition_flags[38] = options->lower_fdot;
   condition_flags[39] = !options->lower_fsat;
   condition_flags[40] = !options->lower_iabs;
   condition_flags[41] = options->lower_fsat;
   condition_flags[42] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[43] = !options->lower_fsign;
   condition_flags[44] = options->lower_umax;
   condition_flags[45] = options->lower_umin;
   condition_flags[46] = !options->lower_umax;
   condition_flags[47] = !options->lower_umin;
   condition_flags[48] = options->lower_scmp;
   condition_flags[49] = options->lower_vector_cmp;
   condition_flags[50] = !options->lower_pack_64_2x32_split;
   condition_flags[51] = !options->lower_rotate;
   condition_flags[52] = options->lower_rotate;
   condition_flags[53] = options->lower_fpow;
   condition_flags[54] = !options->lower_fpow;
   condition_flags[55] = options->lower_fdiv;
   condition_flags[56] = options->lower_fsqrt;
   condition_flags[57] = !options->lower_fsqrt;
   condition_flags[58] = options->lower_sincos;
   condition_flags[59] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[60] = !options->lower_extract_byte;
   condition_flags[61] = !options->lower_extract_word;
   condition_flags[62] = options->has_pack_32_4x8;
   condition_flags[63] = options->lower_pack_64_2x32_split;
   condition_flags[64] = options->lower_pack_32_2x16_split;
   condition_flags[65] = options->has_pack_half_2x16_rtz;
   condition_flags[66] = options->lower_unpack_64_2x32_split;
   condition_flags[67] = options->lower_unpack_32_2x16_split;
   condition_flags[68] = options->has_find_msb_rev;
   condition_flags[69] = !options->lower_find_msb_to_reverse;
   condition_flags[70] = !options->lower_find_lsb;
   condition_flags[71] = options->lower_fmod;
   condition_flags[72] = options->lower_uadd_carry;
   condition_flags[73] = options->lower_usub_borrow;
   condition_flags[74] = options->lower_bitfield_insert;
   condition_flags[75] = options->lower_hadd;
   condition_flags[76] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[77] = options->lower_mul_32x16;
   condition_flags[78] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[79] = options->lower_uadd_sat;
   condition_flags[80] = options->lower_usub_sat;
   condition_flags[81] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[82] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[83] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[84] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[85] = options->lower_bitfield_insert_to_shifts;
   condition_flags[86] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[87] = options->lower_bitfield_extract;
   condition_flags[88] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[89] = options->lower_bitfield_extract_to_shifts;
   condition_flags[90] = options->lower_ifind_msb;
   condition_flags[91] = options->lower_find_msb_to_reverse;
   condition_flags[92] = options->lower_ifind_msb_to_uclz;
   condition_flags[93] = options->lower_ufind_msb_to_uclz;
   condition_flags[94] = options->lower_uclz;
   condition_flags[95] = options->lower_find_lsb;
   condition_flags[96] = options->lower_extract_byte;
   condition_flags[97] = options->lower_extract_word;
   condition_flags[98] = options->lower_pack_unorm_2x16;
   condition_flags[99] = options->lower_pack_unorm_4x8;
   condition_flags[100] = options->lower_pack_snorm_2x16;
   condition_flags[101] = options->lower_pack_snorm_4x8;
   condition_flags[102] = options->lower_unpack_unorm_2x16;
   condition_flags[103] = options->lower_unpack_unorm_4x8;
   condition_flags[104] = options->lower_unpack_snorm_2x16;
   condition_flags[105] = options->lower_unpack_snorm_4x8;
   condition_flags[106] = options->lower_pack_split;
   condition_flags[107] = options->lower_isign;
   condition_flags[108] = !options->lower_isign;
   condition_flags[109] = options->lower_fsign;
   condition_flags[110] = !options->has_imul24;
   condition_flags[111] = !options->has_umul24;
   condition_flags[112] = !options->has_umad24;
   condition_flags[113] = options->has_imul24;
   condition_flags[114] = options->has_umad24;
   condition_flags[115] = options->has_umul24;
   condition_flags[116] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[117] = !options->lower_usub_sat;
   condition_flags[118] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[119] = options->lower_iadd_sat;
   condition_flags[120] = options->support_8bit_alu;
   condition_flags[121] = options->support_16bit_alu;
   condition_flags[122] = options->lower_ldexp;
   condition_flags[123] = !options->lower_bitfield_reverse;
   condition_flags[124] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[125] = options->lower_fisnormal;
   condition_flags[126] = options->avoid_ternary_with_two_constants;
   condition_flags[127] = !options->vectorize_vec2_16bit;

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &nir_opt_algebraic_table);
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 15
 * transforms:
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b'))
 */


static const nir_search_value_union nir_opt_algebraic_before_ffma_values[] = {
   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 1 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 2, 3 },
      0,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 7, 8 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search2204_0_0 -> 0 in the cache */
   /* search2204_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2204_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 10, 3 },
      0,
   } },

   /* replace2204_0_0 -> 5 in the cache */
   /* replace2204_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2204_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 12, 8 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search2205_0_0 -> 0 in the cache */
   /* search2205_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2205_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 14, 3 },
      0,
   } },

   /* replace2205_0_0 -> 5 in the cache */
   /* replace2205_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2205_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 16, 8 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search2206_0_0 -> 0 in the cache */
   /* search2206_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2206_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 18, 3 },
      0,
   } },

   /* replace2206_0_0 -> 5 in the cache */
   /* replace2206_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2206_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 20, 8 },
      -1,
   } },

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search2207_0_0 -> 5 in the cache */
   /* search2207_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search2207_1_0 -> 5 in the cache */
   /* search2207_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 22, 23 },
      -1,
   } },

   /* replace2207_0 -> 5 in the cache */
   /* replace2207_1_0 -> 8 in the cache */
   /* replace2207_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 5, 25 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search2208_0_0 -> 5 in the cache */
   /* search2208_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search2208_1_0 -> 5 in the cache */
   /* search2208_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 27, 28 },
      -1,
   } },

   /* replace2208_0 -> 5 in the cache */
   /* replace2208_1_0 -> 8 in the cache */
   /* replace2208_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 5, 30 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search2209_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 33, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search2210_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search2210_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 36, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search2211_1_0 -> 39 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 40, 42 },
      -1,
   } },

   /* replace2211 -> 41 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search2212_0 -> 39 in the cache */
   /* search2212_1_0_0 -> 39 in the cache */
   /* search2212_1_0 -> 40 in the cache */
   /* search2212_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 40, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 39, 44 },
      -1,
   } },

   /* replace2212 -> 41 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search2213_0_0 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search2213_1_0 -> 39 in the cache */
   /* search2213_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 46, 47 },
      -1,
   } },

   /* replace2213 -> 41 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search2214_0 -> 39 in the cache */
   /* search2214_1_0_0 -> 39 in the cache */
   /* search2214_1_0 -> 46 in the cache */
   /* search2214_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 46, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 39, 49 },
      -1,
   } },

   /* replace2214 -> 41 in the cache */

   /* ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a') */
   /* search2215_0_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 51 },
      0,
   } },
   /* search2215_1_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 53 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 52, 54, 55 },
      -1,
   } },

   /* replace2215_0_0 -> 51 in the cache */
   /* replace2215_0_1 -> 53 in the cache */
   /* replace2215_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 51, 53, 55 },
      -1,
   } },
   /* replace2215_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 57, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a') */
   /* search2216_0_0 -> 39 in the cache */
   /* search2216_0_1 -> 53 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 53 },
      0,
   } },
   /* search2216_1_0 -> 39 in the cache */
   /* search2216_1_1 -> 51 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 51 },
      0,
   } },
   /* search2216_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 59, 60, 55 },
      -1,
   } },

   /* replace2216_0_0 -> 53 in the cache */
   /* replace2216_0_1 -> 51 in the cache */
   /* replace2216_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 53, 51, 55 },
      -1,
   } },
   /* replace2216_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 62, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 64, 65 },
      0,
   } },
   /* search2217_1_0 -> 64 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 64, 67 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 66, 68, 69 },
      -1,
   } },

   /* replace2217_0_0 -> 69 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 71, 73 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 69, 74 },
      -1,
   } },
   /* replace2217_1_0 -> 64 in the cache */
   /* replace2217_1_1 -> 72 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 64, 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 75, 76 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_before_ffma_expression_cond[] = {
   (is_used_once),
};

static const nir_search_variable_cond nir_opt_algebraic_before_ffma_variable_cond[] = {
   (is_not_const),
};

static const struct transform nir_opt_algebraic_before_ffma_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 24, 26, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 29, 31, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 56, 58, 0 },
   { 61, 63, 0 },
   { 70, 77, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_before_ffma_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         2,
         3,
         19,
         3,
         3,
         20,
         19,
         20,
         19,
      },
   },
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         4,
         5,
         28,
         5,
         5,
         29,
         28,
         29,
         28,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         0,
         5,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         6,
         7,
         6,
         6,
         12,
         13,
         31,
         32,
         7,
         7,
         7,
         7,
         14,
         15,
         33,
         34,
         6,
         7,
         16,
         6,
         12,
         13,
         31,
         32,
         6,
         7,
         6,
         6,
         12,
         17,
         31,
         32,
         12,
         14,
         12,
         12,
         12,
         18,
         32,
         32,
         13,
         15,
         13,
         17,
         18,
         13,
         35,
         36,
         31,
         33,
         31,
         31,
         32,
         35,
         31,
         32,
         32,
         34,
         32,
         32,
         32,
         36,
         32,
         32,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         8,
         9,
         8,
         8,
         21,
         22,
         37,
         38,
         9,
         9,
         9,
         9,
         23,
         24,
         39,
         40,
         8,
         9,
         25,
         8,
         21,
         22,
         37,
         38,
         8,
         9,
         8,
         8,
         21,
         26,
         37,
         38,
         21,
         23,
         21,
         21,
         21,
         27,
         38,
         38,
         22,
         24,
         22,
         26,
         27,
         22,
         41,
         42,
         37,
         39,
         37,
         37,
         38,
         41,
         37,
         38,
         38,
         40,
         38,
         38,
         38,
         42,
         38,
         38,
      },
   },
   [nir_op_fneg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_ineg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         30,
         30,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_before_ffma_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   14,
   18,
   20,
   22,
   24,
   26,
   28,
   30,
   32,
   35,
   39,
   41,
   43,
   47,
   49,
   52,
   54,
   57,
   61,
   66,
   68,
   71,
   73,
   76,
   80,
};

static const nir_algebraic_table nir_opt_algebraic_before_ffma_table = {
   .transforms = nir_opt_algebraic_before_ffma_transforms,
   .transform_offsets = nir_opt_algebraic_before_ffma_transform_offsets,
   .pass_op_table = nir_opt_algebraic_before_ffma_pass_op_table,
   .values = nir_opt_algebraic_before_ffma_values,
   .expression_cond = nir_opt_algebraic_before_ffma_expression_cond,
   .variable_cond = nir_opt_algebraic_before_ffma_variable_cond,
};

bool
nir_opt_algebraic_before_ffma(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[128];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(78 == ARRAY_SIZE(nir_opt_algebraic_before_ffma_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !options->lower_ftrunc || options->lower_ffloor;
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[7] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[8] = !options->lower_iadd_sat;
   condition_flags[9] = !options->lower_uadd_sat;
   condition_flags[10] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[11] = !options->has_sdot_4x8;
   condition_flags[12] = !options->has_udot_4x8;
   condition_flags[13] = !options->has_sudot_4x8;
   condition_flags[14] = !options->has_dot_2x16;
   condition_flags[15] = options->lower_flrp16;
   condition_flags[16] = !options->lower_flrp16;
   condition_flags[17] = options->lower_flrp32;
   condition_flags[18] = !options->lower_flrp32;
   condition_flags[19] = options->lower_flrp64;
   condition_flags[20] = !options->lower_flrp64;
   condition_flags[21] = options->lower_ftrunc;
   condition_flags[22] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[23] = options->lower_ffloor;
   condition_flags[24] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[25] = !options->lower_ffloor;
   condition_flags[26] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[27] = options->lower_ffract;
   condition_flags[28] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[29] = options->lower_fceil;
   condition_flags[30] = options->lower_ffma16;
   condition_flags[31] = options->lower_ffma32;
   condition_flags[32] = options->lower_ffma64;
   condition_flags[33] = options->fuse_ffma16;
   condition_flags[34] = options->fuse_ffma32;
   condition_flags[35] = options->fuse_ffma64;
   condition_flags[36] = options->lower_fdph;
   condition_flags[37] = !options->lower_fdph;
   condition_flags[38] = options->lower_fdot;
   condition_flags[39] = !options->lower_fsat;
   condition_flags[40] = !options->lower_iabs;
   condition_flags[41] = options->lower_fsat;
   condition_flags[42] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[43] = !options->lower_fsign;
   condition_flags[44] = options->lower_umax;
   condition_flags[45] = options->lower_umin;
   condition_flags[46] = !options->lower_umax;
   condition_flags[47] = !options->lower_umin;
   condition_flags[48] = options->lower_scmp;
   condition_flags[49] = options->lower_vector_cmp;
   condition_flags[50] = !options->lower_pack_64_2x32_split;
   condition_flags[51] = !options->lower_rotate;
   condition_flags[52] = options->lower_rotate;
   condition_flags[53] = options->lower_fpow;
   condition_flags[54] = !options->lower_fpow;
   condition_flags[55] = options->lower_fdiv;
   condition_flags[56] = options->lower_fsqrt;
   condition_flags[57] = !options->lower_fsqrt;
   condition_flags[58] = options->lower_sincos;
   condition_flags[59] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[60] = !options->lower_extract_byte;
   condition_flags[61] = !options->lower_extract_word;
   condition_flags[62] = options->has_pack_32_4x8;
   condition_flags[63] = options->lower_pack_64_2x32_split;
   condition_flags[64] = options->lower_pack_32_2x16_split;
   condition_flags[65] = options->has_pack_half_2x16_rtz;
   condition_flags[66] = options->lower_unpack_64_2x32_split;
   condition_flags[67] = options->lower_unpack_32_2x16_split;
   condition_flags[68] = options->has_find_msb_rev;
   condition_flags[69] = !options->lower_find_msb_to_reverse;
   condition_flags[70] = !options->lower_find_lsb;
   condition_flags[71] = options->lower_fmod;
   condition_flags[72] = options->lower_uadd_carry;
   condition_flags[73] = options->lower_usub_borrow;
   condition_flags[74] = options->lower_bitfield_insert;
   condition_flags[75] = options->lower_hadd;
   condition_flags[76] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[77] = options->lower_mul_32x16;
   condition_flags[78] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[79] = options->lower_uadd_sat;
   condition_flags[80] = options->lower_usub_sat;
   condition_flags[81] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[82] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[83] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[84] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[85] = options->lower_bitfield_insert_to_shifts;
   condition_flags[86] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[87] = options->lower_bitfield_extract;
   condition_flags[88] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[89] = options->lower_bitfield_extract_to_shifts;
   condition_flags[90] = options->lower_ifind_msb;
   condition_flags[91] = options->lower_find_msb_to_reverse;
   condition_flags[92] = options->lower_ifind_msb_to_uclz;
   condition_flags[93] = options->lower_ufind_msb_to_uclz;
   condition_flags[94] = options->lower_uclz;
   condition_flags[95] = options->lower_find_lsb;
   condition_flags[96] = options->lower_extract_byte;
   condition_flags[97] = options->lower_extract_word;
   condition_flags[98] = options->lower_pack_unorm_2x16;
   condition_flags[99] = options->lower_pack_unorm_4x8;
   condition_flags[100] = options->lower_pack_snorm_2x16;
   condition_flags[101] = options->lower_pack_snorm_4x8;
   condition_flags[102] = options->lower_unpack_unorm_2x16;
   condition_flags[103] = options->lower_unpack_unorm_4x8;
   condition_flags[104] = options->lower_unpack_snorm_2x16;
   condition_flags[105] = options->lower_unpack_snorm_4x8;
   condition_flags[106] = options->lower_pack_split;
   condition_flags[107] = options->lower_isign;
   condition_flags[108] = !options->lower_isign;
   condition_flags[109] = options->lower_fsign;
   condition_flags[110] = !options->has_imul24;
   condition_flags[111] = !options->has_umul24;
   condition_flags[112] = !options->has_umad24;
   condition_flags[113] = options->has_imul24;
   condition_flags[114] = options->has_umad24;
   condition_flags[115] = options->has_umul24;
   condition_flags[116] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[117] = !options->lower_usub_sat;
   condition_flags[118] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[119] = options->lower_iadd_sat;
   condition_flags[120] = options->support_8bit_alu;
   condition_flags[121] = options->support_16bit_alu;
   condition_flags[122] = options->lower_ldexp;
   condition_flags[123] = !options->lower_bitfield_reverse;
   condition_flags[124] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[125] = options->lower_fisnormal;
   condition_flags[126] = options->avoid_ternary_with_two_constants;
   condition_flags[127] = !options->vectorize_vec2_16bit;

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &nir_opt_algebraic_before_ffma_table);
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 232
 * transforms:
 *    ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b'))
 *    ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b'))
 *    ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))
 *    ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a'))
 *    ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')))
 *    ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a')))
 *    ('fneg', ('fneg', 'a')) => a
 *    ('~fadd@16', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@16', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fmulz(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmulz(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffmaz', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@64', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@64', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fneg', 'a') => ('fmul', 'a', -1.0)
 *    ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b')
 *    ('ineg', 'a') => ('isub', 0, 'a')
 *    ('iabs', 'a') => ('imax', 'a', ('ineg', 'a'))
 *    ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c')
 *    ('iadd', ('iadd(is_used_once)', '#a(is_16_bits)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c')
 *    ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_16_bits)') => ('iadd3', 'a', 'b', 'c')
 *    ('iadd', ('ineg', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)')), 'c(is_not_const)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c')
 *    ('iadd', ('ineg', ('iadd(is_used_once)', '#a(is_16_bits)', 'b(is_not_const)')), 'c(is_not_const)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c')
 *    ('iadd', ('ineg', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)')), '#c(is_16_bits)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c')
 *    ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0))
 *    ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0))
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd')))
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('ior', 'a', 'a') => a
 *    ('iand', 'a', 'a') => a
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b')
 *    ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5)))
 *    ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a')
 *    ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b'))
 *    ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b'))
 *    ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38)
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g')))
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g')))
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b'))
 *    ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c'))
 *    ('ishl', 'a', 0) => a
 *    ('ishl', 'a', -32) => a
 *    ('ishr', 'a', 0) => a
 *    ('ishr', 'a', -32) => a
 *    ('ushr', 'a', 0) => a
 *    ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 0) => ('bitnz', 'a', ('find_lsb', 'b'))
 *    ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 0) => ('bitz', 'a', ('find_lsb', 'b'))
 *    ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('bitz', 'a', ('find_lsb', 'b'))
 *    ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('bitnz', 'a', ('find_lsb', 'b'))
 *    ('ine', ('iand', 'a', ('ishl', 1, 'b')), 0) => ('bitnz', 'a', 'b')
 *    ('ieq', ('iand', 'a', ('ishl', 1, 'b')), 0) => ('bitz', 'a', 'b')
 *    ('ine', ('iand', 'a', ('ishl', 1, 'b')), ('ishl', 1, 'b')) => ('bitz', 'a', 'b')
 *    ('ieq', ('iand', 'a', ('ishl', 1, 'b')), ('ishl', 1, 'b')) => ('bitnz', 'a', 'b')
 *    ('bitz', ('ushr', 'a', 'b'), 0) => ('bitz', 'a', 'b')
 *    ('bitz', ('ishr', 'a', 'b'), 0) => ('bitz', 'a', 'b')
 *    ('bitnz', ('ushr', 'a', 'b'), 0) => ('bitnz', 'a', 'b')
 *    ('bitnz', ('ishr', 'a', 'b'), 0) => ('bitnz', 'a', 'b')
 *    ('ine', ('ubfe', 'a', 'b', 1), 0) => ('bitnz', 'a', 'b')
 *    ('ieq', ('ubfe', 'a', 'b', 1), 0) => ('bitz', 'a', 'b')
 *    ('ine', ('ubfe', 'a', 'b', 1), 1) => ('bitz', 'a', 'b')
 *    ('ieq', ('ubfe', 'a', 'b', 1), 1) => ('bitnz', 'a', 'b')
 *    ('ine', ('ibfe', 'a', 'b', 1), 0) => ('bitnz', 'a', 'b')
 *    ('ieq', ('ibfe', 'a', 'b', 1), 0) => ('bitz', 'a', 'b')
 *    ('ine', ('ibfe', 'a', 'b', 1), -1) => ('bitz', 'a', 'b')
 *    ('ieq', ('ibfe', 'a', 'b', 1), -1) => ('bitnz', 'a', 'b')
 *    ('inot', ('bitnz', 'a', 'b')) => ('bitz', 'a', 'b')
 *    ('inot', ('bitz', 'a', 'b')) => ('bitnz', 'a', 'b')
 *    ('bitnz', ('inot', 'a'), 'b') => ('bitz', 'a', 'b')
 *    ('bitz', ('inot', 'a'), 'b') => ('bitnz', 'a', 'b')
 *    ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1)
 *    ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2)
 *    ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3)
 *    ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1)
 *    ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b')
 *    ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b')
 *    ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b')))
 *    ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a')
 *    ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a')
 *    ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a')
 *    ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a')
 *    ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a')
 *    ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a')
 *    ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a')
 *    ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a')
 *    ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a')
 *    ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a')
 *    ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a')
 *    ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a')
 *    ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a')
 *    ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a')
 *    ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a')
 *    ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b')
 *    ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b')
 *    ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b')
 *    ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b')
 *    ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b')
 *    ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b')
 *    ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b')
 *    ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b')
 *    ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c')
 *    ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c')
 *    ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b')
 *    ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b')
 *    ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b')
 *    ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b')
 *    ('f2fmp', 'a') => ('f2f16', 'a')
 *    ('f2imp', 'a') => ('f2i16', 'a')
 *    ('f2ump', 'a') => ('f2u16', 'a')
 *    ('i2imp', 'a') => ('i2i16', 'a')
 *    ('i2fmp', 'a') => ('i2f16', 'a')
 *    ('i2imp', 'a') => ('u2u16', 'a')
 *    ('u2fmp', 'a') => ('u2f16', 'a')
 *    ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf)
 *    ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b')
 *    ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b')
 *    ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 */


static const nir_search_value_union nir_opt_algebraic_late_values[] = {
   /* ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace2218_0 -> 0 in the cache */
   /* replace2218_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b') */
   /* search2219_0_0_0 -> 0 in the cache */
   /* search2219_0_0_1 -> 1 in the cache */
   /* search2219_0_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 2 },
      0,
   } },
   /* search2219_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace2219_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace2219_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b') */
   /* search2220_0 -> 3 in the cache */
   /* search2220_1_0 -> 0 in the cache */
   /* search2220_1_1 -> 1 in the cache */
   /* search2220_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace2220_0_0 -> 0 in the cache */
   /* replace2220_0 -> 9 in the cache */
   /* replace2220_1 -> 1 in the cache */
   /* replace2220 -> 10 in the cache */

   /* ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b')) */
   /* search2221_0 -> 3 in the cache */
   /* search2221_1_0_0 -> 0 in the cache */
   /* search2221_1_0_1 -> 1 in the cache */
   /* search2221_1_0 -> 2 in the cache */
   /* search2221_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace2221_0 -> 0 in the cache */
   /* replace2221_1_0 -> 1 in the cache */
   /* replace2221_1 -> 5 in the cache */
   /* replace2221 -> 6 in the cache */

   /* ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   /* search2222_0_0 -> 0 in the cache */
   /* search2222_0_1 -> 1 in the cache */
   /* search2222_0 -> 2 in the cache */
   /* search2222_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace2222_0 -> 0 in the cache */
   /* replace2222_1_0 -> 1 in the cache */
   /* replace2222_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search2223_0_0_0 -> 0 in the cache */
   /* search2223_0_0_1 -> 1 in the cache */
   /* search2223_0_0 -> 2 in the cache */
   /* search2223_0 -> 7 in the cache */
   /* search2223_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace2223_0_0 -> 0 in the cache */
   /* replace2223_0 -> 9 in the cache */
   /* replace2223_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search2224_0 -> 3 in the cache */
   /* search2224_1_0 -> 0 in the cache */
   /* search2224_1_1 -> 1 in the cache */
   /* search2224_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace2224_0_0 -> 0 in the cache */
   /* replace2224_0 -> 9 in the cache */
   /* replace2224_1 -> 1 in the cache */
   /* replace2224 -> 16 in the cache */

   /* ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search2225_0 -> 3 in the cache */
   /* search2225_1_0_0 -> 0 in the cache */
   /* search2225_1_0_1 -> 1 in the cache */
   /* search2225_1_0 -> 2 in the cache */
   /* search2225_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace2225_0 -> 0 in the cache */
   /* replace2225_1_0 -> 1 in the cache */
   /* replace2225_1 -> 5 in the cache */
   /* replace2225 -> 14 in the cache */

   /* ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search2226_0_0 -> 0 in the cache */
   /* search2226_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2226_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace2226_0 -> 0 in the cache */
   /* replace2226_1_0 -> 1 in the cache */
   /* replace2226_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search2227_0_0 -> 0 in the cache */
   /* search2227_0_1 -> 1 in the cache */
   /* search2227_0 -> 19 in the cache */
   /* search2227_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace2227_0 -> 0 in the cache */
   /* replace2227_1_0 -> 1 in the cache */
   /* replace2227_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2228_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 24, 1 },
      0,
   } },
   /* search2228_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 25, 3 },
      -1,
   } },

   /* replace2228_0 -> 0 in the cache */
   /* replace2228_1_0 -> 1 in the cache */
   /* replace2228_1 -> 5 in the cache */
   /* replace2228 -> 14 in the cache */

   /* ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search2229_0_0_0 -> 24 in the cache */
   /* search2229_0_0_1 -> 1 in the cache */
   /* search2229_0_0 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 25 },
      0,
   } },
   /* search2229_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 27, 3 },
      -1,
   } },

   /* replace2229_0_0 -> 0 in the cache */
   /* replace2229_0 -> 9 in the cache */
   /* replace2229_1 -> 1 in the cache */
   /* replace2229 -> 16 in the cache */

   /* ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search2230_0 -> 3 in the cache */
   /* search2230_1_0 -> 24 in the cache */
   /* search2230_1_1 -> 1 in the cache */
   /* search2230_1 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 25 },
      -1,
   } },

   /* replace2230_0_0 -> 0 in the cache */
   /* replace2230_0 -> 9 in the cache */
   /* replace2230_1 -> 1 in the cache */
   /* replace2230 -> 16 in the cache */

   /* ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search2231_0 -> 3 in the cache */
   /* search2231_1_0_0 -> 24 in the cache */
   /* search2231_1_0_1 -> 1 in the cache */
   /* search2231_1_0 -> 25 in the cache */
   /* search2231_1 -> 27 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 27 },
      -1,
   } },

   /* replace2231_0 -> 0 in the cache */
   /* replace2231_1_0 -> 1 in the cache */
   /* replace2231_1 -> 5 in the cache */
   /* replace2231 -> 14 in the cache */

   /* ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search2232_0_0 -> 24 in the cache */
   /* search2232_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 24, 1 },
      0,
   } },
   /* search2232_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace2232_0 -> 0 in the cache */
   /* replace2232_1_0 -> 1 in the cache */
   /* replace2232_1 -> 5 in the cache */
   /* replace2232 -> 21 in the cache */

   /* ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search2233_0_0 -> 24 in the cache */
   /* search2233_0_1 -> 1 in the cache */
   /* search2233_0 -> 31 in the cache */
   /* search2233_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace2233_0 -> 0 in the cache */
   /* replace2233_1_0 -> 1 in the cache */
   /* replace2233_1 -> 5 in the cache */
   /* replace2233 -> 23 in the cache */

   /* ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')) */
   /* search2234_0_0 -> 0 in the cache */
   /* search2234_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search2234_1_0_0 -> 0 in the cache */
   /* search2234_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search2234_1_1_0 -> 1 in the cache */
   /* search2234_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 34, 37 },
      -1,
   } },

   /* replace2234_0_0 -> 0 in the cache */
   /* replace2234_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* replace2234_1_0 -> 1 in the cache */
   /* replace2234_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 39, 40 },
      -1,
   } },

   /* ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search2235_1_0 -> 42 in the cache */
   /* search2235_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 44, 45 },
      -1,
   } },

   /* replace2235_0 -> 43 in the cache */
   /* replace2235_1_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 43, 47 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))) */
   /* search2236_0_0 -> 0 in the cache */
   /* search2236_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search2236_1_0_0 -> 0 in the cache */
   /* search2236_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search2236_1_1_0 -> 1 in the cache */
   /* search2236_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 50, 51 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 49, 52 },
      -1,
   } },

   /* replace2236_0_0_0 -> 0 in the cache */
   /* replace2236_0_0_1 -> 1 in the cache */
   /* replace2236_0_0 -> 39 in the cache */
   /* replace2236_0_1_0 -> 1 in the cache */
   /* replace2236_0_1_1 -> 0 in the cache */
   /* replace2236_0_1 -> 40 in the cache */
   /* replace2236_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 41 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a'))) */
   /* search2237_0_0 -> 42 in the cache */
   /* search2237_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search2237_1_0 -> 42 in the cache */
   /* search2237_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 55, 56 },
      -1,
   } },

   /* replace2237_0_0 -> 43 in the cache */
   /* replace2237_0_1_0 -> 42 in the cache */
   /* replace2237_0_1 -> 47 in the cache */
   /* replace2237_0 -> 48 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 48 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   /* search2238_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 59 },
      -1,
   } },

   /* replace2238 -> 42 in the cache */

   /* ('~fadd@16', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 61, 62 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 63, 64 },
      -1,
   } },

   /* replace2239_0 -> 61 in the cache */
   /* replace2239_1 -> 62 in the cache */
   /* replace2239_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 61, 62, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2240_0_0_0 -> 61 in the cache */
   /* search2240_0_0_1 -> 62 in the cache */
   /* search2240_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 63 },
      1,
   } },
   /* search2240_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 67, 64 },
      -1,
   } },

   /* replace2240_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 61 },
      -1,
   } },
   /* replace2240_1 -> 62 in the cache */
   /* replace2240_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 69, 62, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2241_0_0_0 -> 61 in the cache */
   /* search2241_0_0_1 -> 62 in the cache */
   /* search2241_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 63 },
      1,
   } },
   /* search2241_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 71, 64 },
      -1,
   } },

   /* replace2241_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 61 },
      -1,
   } },
   /* replace2241_1_0 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 62 },
      -1,
   } },
   /* replace2241_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 73, 74, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2242_0_0_0_0 -> 61 in the cache */
   /* search2242_0_0_0_1 -> 62 in the cache */
   /* search2242_0_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 63 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 76 },
      1,
   } },
   /* search2242_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 77, 64 },
      -1,
   } },

   /* replace2242_0_0_0 -> 61 in the cache */
   /* replace2242_0_0 -> 73 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 73 },
      -1,
   } },
   /* replace2242_1_0 -> 62 in the cache */
   /* replace2242_1 -> 74 in the cache */
   /* replace2242_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 79, 74, 64 },
      -1,
   } },

   /* ('~fadd@32', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 81, 82 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 83, 84 },
      -1,
   } },

   /* replace2243_0 -> 81 in the cache */
   /* replace2243_1 -> 82 in the cache */
   /* replace2243_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2244_0_0_0 -> 81 in the cache */
   /* search2244_0_0_1 -> 82 in the cache */
   /* search2244_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 83 },
      1,
   } },
   /* search2244_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 87, 84 },
      -1,
   } },

   /* replace2244_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 81 },
      -1,
   } },
   /* replace2244_1 -> 82 in the cache */
   /* replace2244_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 89, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2245_0_0_0 -> 81 in the cache */
   /* search2245_0_0_1 -> 82 in the cache */
   /* search2245_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 83 },
      1,
   } },
   /* search2245_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 91, 84 },
      -1,
   } },

   /* replace2245_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 81 },
      -1,
   } },
   /* replace2245_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 82 },
      -1,
   } },
   /* replace2245_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 93, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2246_0_0_0_0 -> 81 in the cache */
   /* search2246_0_0_0_1 -> 82 in the cache */
   /* search2246_0_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 83 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 96 },
      1,
   } },
   /* search2246_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 97, 84 },
      -1,
   } },

   /* replace2246_0_0_0 -> 81 in the cache */
   /* replace2246_0_0 -> 93 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 93 },
      -1,
   } },
   /* replace2246_1_0 -> 82 in the cache */
   /* replace2246_1 -> 94 in the cache */
   /* replace2246_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 99, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fmulz(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search2247_0_0 -> 81 in the cache */
   /* search2247_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 81, 82 },
      1,
   } },
   /* search2247_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 101, 84 },
      -1,
   } },

   /* replace2247_0 -> 81 in the cache */
   /* replace2247_1 -> 82 in the cache */
   /* replace2247_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fneg', 'a'), 'b', 'c') */
   /* search2248_0_0_0 -> 81 in the cache */
   /* search2248_0_0_1 -> 82 in the cache */
   /* search2248_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 101 },
      1,
   } },
   /* search2248_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 104, 84 },
      -1,
   } },

   /* replace2248_0_0 -> 81 in the cache */
   /* replace2248_0 -> 89 in the cache */
   /* replace2248_1 -> 82 in the cache */
   /* replace2248_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 89, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2249_0_0_0 -> 81 in the cache */
   /* search2249_0_0_1 -> 82 in the cache */
   /* search2249_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 101 },
      1,
   } },
   /* search2249_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 107, 84 },
      -1,
   } },

   /* replace2249_0_0 -> 81 in the cache */
   /* replace2249_0 -> 93 in the cache */
   /* replace2249_1_0 -> 82 in the cache */
   /* replace2249_1 -> 94 in the cache */
   /* replace2249_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 93, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmulz(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffmaz', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2250_0_0_0_0 -> 81 in the cache */
   /* search2250_0_0_0_1 -> 82 in the cache */
   /* search2250_0_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 101 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 110 },
      1,
   } },
   /* search2250_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 111, 84 },
      -1,
   } },

   /* replace2250_0_0_0 -> 81 in the cache */
   /* replace2250_0_0 -> 93 in the cache */
   /* replace2250_0 -> 99 in the cache */
   /* replace2250_1_0 -> 82 in the cache */
   /* replace2250_1 -> 94 in the cache */
   /* replace2250_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 99, 94, 84 },
      -1,
   } },

   /* ('~fadd@64', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 114, 115 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 116, 117 },
      -1,
   } },

   /* replace2251_0 -> 114 in the cache */
   /* replace2251_1 -> 115 in the cache */
   /* replace2251_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 114, 115, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2252_0_0_0 -> 114 in the cache */
   /* search2252_0_0_1 -> 115 in the cache */
   /* search2252_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 116 },
      1,
   } },
   /* search2252_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 120, 117 },
      -1,
   } },

   /* replace2252_0_0 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 114 },
      -1,
   } },
   /* replace2252_1 -> 115 in the cache */
   /* replace2252_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 122, 115, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2253_0_0_0 -> 114 in the cache */
   /* search2253_0_0_1 -> 115 in the cache */
   /* search2253_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 116 },
      1,
   } },
   /* search2253_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 124, 117 },
      -1,
   } },

   /* replace2253_0_0 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 114 },
      -1,
   } },
   /* replace2253_1_0 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 115 },
      -1,
   } },
   /* replace2253_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 126, 127, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2254_0_0_0_0 -> 114 in the cache */
   /* search2254_0_0_0_1 -> 115 in the cache */
   /* search2254_0_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 116 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 129 },
      1,
   } },
   /* search2254_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 130, 117 },
      -1,
   } },

   /* replace2254_0_0_0 -> 114 in the cache */
   /* replace2254_0_0 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 126 },
      -1,
   } },
   /* replace2254_1_0 -> 115 in the cache */
   /* replace2254_1 -> 127 in the cache */
   /* replace2254_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 132, 127, 117 },
      -1,
   } },

   /* ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 134, 136 },
      -1,
   } },

   /* replace2255_0 -> 134 in the cache */
   /* replace2255_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 134, 135 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2256_0 -> 61 in the cache */
   /* search2256_1_0 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 62 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 61, 139 },
      -1,
   } },

   /* replace2256_0 -> 61 in the cache */
   /* replace2256_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 61, 62 },
      -1,
   } },

   /* ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2257_0 -> 81 in the cache */
   /* search2257_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 81, 142 },
      -1,
   } },

   /* replace2257_0 -> 81 in the cache */
   /* replace2257_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2258_0 -> 114 in the cache */
   /* search2258_1_0 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 114, 145 },
      -1,
   } },

   /* replace2258_0 -> 114 in the cache */
   /* replace2258_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 114, 115 },
      -1,
   } },

   /* ('fneg', 'a') => ('fmul', 'a', -1.0) */
   /* search2259_0 -> 42 in the cache */
   /* search2259 -> 59 in the cache */

   /* replace2259_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 42, 148 },
      -1,
   } },

   /* ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b') */
   /* search2260_0 -> 0 in the cache */
   /* search2260_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 150 },
      -1,
   } },

   /* replace2260_0 -> 0 in the cache */
   /* replace2260_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('ineg', 'a') => ('isub', 0, 'a') */
   /* search2261_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 42 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace2261_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 154, 42 },
      -1,
   } },

   /* ('iabs', 'a') => ('imax', 'a', ('ineg', 'a')) */
   /* search2262_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2262_0 -> 42 in the cache */
   /* replace2262_1_0 -> 42 in the cache */
   /* replace2262_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 42, 153 },
      -1,
   } },

   /* ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 158, 159 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 160, 161 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 163, 164, 165 },
      -1,
   } },

   /* ('iadd', ('iadd(is_used_once)', '#a(is_16_bits)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2264_0_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 167, 159 },
      0,
   } },
   /* search2264_1 -> 161 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 168, 161 },
      -1,
   } },

   /* replace2264_0 -> 163 in the cache */
   /* replace2264_1 -> 164 in the cache */
   /* replace2264_2 -> 165 in the cache */
   /* replace2264 -> 166 in the cache */

   /* ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_16_bits)') => ('iadd3', 'a', 'b', 'c') */
   /* search2265_0_0 -> 158 in the cache */
   /* search2265_0_1 -> 159 in the cache */
   /* search2265_0 -> 160 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 160, 170 },
      -1,
   } },

   /* replace2265_0 -> 163 in the cache */
   /* replace2265_1 -> 164 in the cache */
   /* replace2265_2 -> 165 in the cache */
   /* replace2265 -> 166 in the cache */

   /* ('iadd', ('ineg', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)')), 'c(is_not_const)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c') */
   /* search2266_0_0_0 -> 158 in the cache */
   /* search2266_0_0_1 -> 159 in the cache */
   /* search2266_0_0 -> 160 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 160 },
      -1,
   } },
   /* search2266_1 -> 161 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 172, 161 },
      -1,
   } },

   /* replace2266_0_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 163 },
      -1,
   } },
   /* replace2266_1_0 -> 164 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 164 },
      -1,
   } },
   /* replace2266_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 174, 175, 165 },
      -1,
   } },

   /* ('iadd', ('ineg', ('iadd(is_used_once)', '#a(is_16_bits)', 'b(is_not_const)')), 'c(is_not_const)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c') */
   /* search2267_0_0_0 -> 167 in the cache */
   /* search2267_0_0_1 -> 159 in the cache */
   /* search2267_0_0 -> 168 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 168 },
      -1,
   } },
   /* search2267_1 -> 161 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 177, 161 },
      -1,
   } },

   /* replace2267_0_0 -> 163 in the cache */
   /* replace2267_0 -> 174 in the cache */
   /* replace2267_1_0 -> 164 in the cache */
   /* replace2267_1 -> 175 in the cache */
   /* replace2267_2 -> 165 in the cache */
   /* replace2267 -> 176 in the cache */

   /* ('iadd', ('ineg', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)')), '#c(is_16_bits)') => ('iadd3', ('ineg', 'a'), ('ineg', 'b'), 'c') */
   /* search2268_0_0_0 -> 158 in the cache */
   /* search2268_0_0_1 -> 159 in the cache */
   /* search2268_0_0 -> 160 in the cache */
   /* search2268_0 -> 172 in the cache */
   /* search2268_1 -> 170 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 172, 170 },
      -1,
   } },

   /* replace2268_0_0 -> 163 in the cache */
   /* replace2268_0 -> 174 in the cache */
   /* replace2268_1_0 -> 164 in the cache */
   /* replace2268_1 -> 175 in the cache */
   /* replace2268_2 -> 165 in the cache */
   /* replace2268 -> 176 in the cache */

   /* ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0)) */
   /* search2269_0_0 -> 61 in the cache */
   /* search2269_0 -> 69 in the cache */
   /* search2269_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 69, 62 },
      2,
   } },

   /* replace2269_0_0 -> 61 in the cache */
   /* replace2269_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 62 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 182, 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 181, 184 },
      -1,
   } },

   /* ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0)) */
   /* search2270_0 -> 61 in the cache */
   /* search2270_1_0 -> 62 in the cache */
   /* search2270_1 -> 139 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 139 },
      2,
   } },

   /* replace2270_0_0 -> 61 in the cache */
   /* replace2270_0_1 -> 62 in the cache */
   /* replace2270_0 -> 181 in the cache */
   /* replace2270_1_0 -> 183 in the cache */
   /* replace2270_1_1 -> 182 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 183, 182 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 181, 187 },
      -1,
   } },

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 190 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 189, 191 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2271_1 -> 190 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 193, 190 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search2272_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 195, 196 },
      -1,
   } },

   /* replace2272_0 -> 0 in the cache */
   /* replace2272_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search2273_0_0 -> 0 in the cache */
   /* search2273_0 -> 195 in the cache */
   /* search2273_1 -> 196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 195, 196 },
      -1,
   } },

   /* replace2273_0 -> 0 in the cache */
   /* replace2273_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search2274_0_0 -> 0 in the cache */
   /* search2274_0 -> 195 in the cache */
   /* search2274_1 -> 196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 195, 196 },
      -1,
   } },

   /* replace2274_0 -> 0 in the cache */
   /* replace2274_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search2275_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 42 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 203, 204 },
      -1,
   } },

   /* replace2275_0 -> 42 in the cache */
   /* replace2275_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 42, 204 },
      -1,
   } },

   /* ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd'))) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 207, 208 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 210, 211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 209, 212 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 3,
      { 213, 214 },
      -1,
   } },

   /* replace2276_0_0 -> 207 in the cache */
   /* replace2276_0_1_0 -> 208 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 208 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 207, 216 },
      -1,
   } },
   /* replace2276_1_0 -> 210 in the cache */
   /* replace2276_1_1_0 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 210, 218 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 217, 219 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search2277_0_0 -> 0 in the cache */
   /* search2277_0 -> 9 in the cache */
   /* search2277_1_0 -> 1 in the cache */
   /* search2277_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2277_0 -> 1 in the cache */
   /* replace2277_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search2278_0_0 -> 0 in the cache */
   /* search2278_0 -> 9 in the cache */
   /* search2278_1_0 -> 1 in the cache */
   /* search2278_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2278_0 -> 1 in the cache */
   /* replace2278_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search2279_0_0 -> 0 in the cache */
   /* search2279_0 -> 9 in the cache */
   /* search2279_1_0 -> 1 in the cache */
   /* search2279_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2279_0 -> 1 in the cache */
   /* replace2279_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search2280_0_0 -> 0 in the cache */
   /* search2280_0 -> 9 in the cache */
   /* search2280_1_0 -> 1 in the cache */
   /* search2280_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2280_0 -> 1 in the cache */
   /* replace2280_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search2281_0_0 -> 42 in the cache */
   /* search2281_0 -> 59 in the cache */
   /* search2281_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 59, 148 },
      -1,
   } },

   /* replace2281_0 -> 204 in the cache */
   /* replace2281_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 204, 42 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search2282_0 -> 148 in the cache */
   /* search2282_1_0 -> 42 in the cache */
   /* search2282_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 148, 59 },
      -1,
   } },

   /* replace2282_0 -> 42 in the cache */
   /* replace2282_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 42, 204 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search2283_0_0 -> 42 in the cache */
   /* search2283_0 -> 59 in the cache */
   /* search2283_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 59, 148 },
      -1,
   } },

   /* replace2283_0 -> 204 in the cache */
   /* replace2283_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 204, 42 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search2284_0 -> 148 in the cache */
   /* search2284_1_0 -> 42 in the cache */
   /* search2284_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 148, 59 },
      -1,
   } },

   /* replace2284_0 -> 42 in the cache */
   /* replace2284_1 -> 204 in the cache */
   /* replace2284 -> 206 in the cache */

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search2285_0_0 -> 42 in the cache */
   /* search2285_0 -> 59 in the cache */
   /* search2285_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 59, 148 },
      -1,
   } },

   /* replace2285_0 -> 204 in the cache */
   /* replace2285_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 204, 42 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search2286_0 -> 148 in the cache */
   /* search2286_1_0 -> 42 in the cache */
   /* search2286_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 148, 59 },
      -1,
   } },

   /* replace2286_0 -> 42 in the cache */
   /* replace2286_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 42, 204 },
      -1,
   } },

   /* ('ior', 'a', 'a') => a */
   /* search2287_0 -> 42 in the cache */
   /* search2287_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2287 -> 42 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search2288_0 -> 42 in the cache */
   /* search2288_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2288 -> 42 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 242 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 243 },
      0,
   } },
   /* search2289_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 244, 204 },
      -1,
   } },

   /* replace2289_0_0 -> 204 in the cache */
   /* replace2289_0_1_0 -> 42 in the cache */
   /* replace2289_0_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 204, 59 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 246 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b') */
   /* search2290_0 -> 0 in the cache */
   /* search2290_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2290_0 -> 0 in the cache */
   /* replace2290_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b') */
   /* search2291_0 -> 0 in the cache */
   /* search2291_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2291_0 -> 0 in the cache */
   /* replace2291_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b') */
   /* search2292_0 -> 0 in the cache */
   /* search2292_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2292_0 -> 0 in the cache */
   /* replace2292_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b') */
   /* search2293_0 -> 0 in the cache */
   /* search2293_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace2293_0 -> 0 in the cache */
   /* replace2293_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search2294_0_0 -> 207 in the cache */
   /* search2294_0_1 -> 208 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 207, 208 },
      0,
   } },
   /* search2294_1_0 -> 207 in the cache */
   /* search2294_1_1 -> 210 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 207, 210 },
      0,
   } },
   /* search2294_2 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 256, 257, 211 },
      -1,
   } },

   /* replace2294_0_0 -> 208 in the cache */
   /* replace2294_0_1 -> 210 in the cache */
   /* replace2294_0_2 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 208, 210, 211 },
      -1,
   } },
   /* replace2294_1 -> 207 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 259, 207 },
      -1,
   } },

   /* ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5))) */
   /* search2295_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2295_0_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 42 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 262, 263 },
      -1,
   } },
   /* replace2295_1_0_0_0 -> 42 in the cache */
   /* replace2295_1_0_0_1 -> 263 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 42, 263 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 265 },
      -1,
   } },
   /* replace2295_1_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 266, 204 },
      -1,
   } },
   /* replace2295_2_0_0 -> 42 in the cache */
   /* replace2295_2_0_1 -> 263 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 42, 263 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 268 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 264, 267, 269 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2296_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 271, 272 },
      -1,
   } },

   /* replace2296_0 -> 272 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace2296_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 272, 274, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2297_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xc000000000000000 /* -2.0 */ },
   } },
   /* search2297_2 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 276, 272 },
      -1,
   } },

   /* replace2297_0 -> 272 in the cache */
   /* replace2297_1 -> 274 in the cache */
   /* replace2297_2_0 -> 81 in the cache */
   /* replace2297_2 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 272, 274, 89 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2298_0 -> 81 in the cache */
   /* search2298_1 -> 276 in the cache */
   /* search2298_2 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 276, 274 },
      -1,
   } },

   /* replace2298_0 -> 274 in the cache */
   /* replace2298_1 -> 272 in the cache */
   /* replace2298_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 274, 272, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2299_0 -> 81 in the cache */
   /* search2299_1 -> 271 in the cache */
   /* search2299_2 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 271, 274 },
      -1,
   } },

   /* replace2299_0 -> 274 in the cache */
   /* replace2299_1 -> 272 in the cache */
   /* replace2299_2_0 -> 81 in the cache */
   /* replace2299_2 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 274, 272, 89 },
      -1,
   } },

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2300_0_0 -> 271 in the cache */
   /* search2300_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 271, 81 },
      0,
   } },
   /* search2300_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 283, 272 },
      -1,
   } },

   /* replace2300_0 -> 272 in the cache */
   /* replace2300_1 -> 274 in the cache */
   /* replace2300_2 -> 81 in the cache */
   /* replace2300 -> 275 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2301_0_0 -> 276 in the cache */
   /* search2301_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 276, 81 },
      0,
   } },
   /* search2301_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 285, 272 },
      -1,
   } },

   /* replace2301_0 -> 272 in the cache */
   /* replace2301_1 -> 274 in the cache */
   /* replace2301_2_0 -> 81 in the cache */
   /* replace2301_2 -> 89 in the cache */
   /* replace2301 -> 278 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2302_0_0 -> 276 in the cache */
   /* search2302_0_1 -> 81 in the cache */
   /* search2302_0 -> 285 in the cache */
   /* search2302_1 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 285, 274 },
      -1,
   } },

   /* replace2302_0 -> 274 in the cache */
   /* replace2302_1 -> 272 in the cache */
   /* replace2302_2 -> 81 in the cache */
   /* replace2302 -> 280 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2303_0_0 -> 271 in the cache */
   /* search2303_0_1 -> 81 in the cache */
   /* search2303_0 -> 283 in the cache */
   /* search2303_1 -> 274 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 283, 274 },
      -1,
   } },

   /* replace2303_0 -> 274 in the cache */
   /* replace2303_1 -> 272 in the cache */
   /* replace2303_2_0 -> 81 in the cache */
   /* replace2303_2 -> 89 in the cache */
   /* replace2303 -> 282 in the cache */

   /* ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 290 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 289, 291 },
      -1,
   } },
   /* search2304_1 -> 290 in the cache */
   /* search2304_2 -> 290 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 292, 290, 290 },
      -1,
   } },

   /* replace2304_0 -> 290 in the cache */
   /* replace2304_1 -> 289 in the cache */
   /* replace2304_2 -> 290 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 290, 289, 290 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2305_0 -> 81 in the cache */
   /* search2305_1 -> 271 in the cache */
   /* search2305_2_0_0 -> 81 in the cache */
   /* search2305_2_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 81, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 295 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 271, 296 },
      -1,
   } },

   /* replace2305_0 -> 81 in the cache */
   /* replace2305_1 -> 274 in the cache */
   /* replace2305_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 81, 274, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2306_0 -> 81 in the cache */
   /* search2306_1 -> 271 in the cache */
   /* search2306_2_0_0 -> 81 in the cache */
   /* search2306_2_0 -> 89 in the cache */
   /* search2306_2_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 89, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 81, 271, 299 },
      -1,
   } },

   /* replace2306_0 -> 81 in the cache */
   /* replace2306_1 -> 274 in the cache */
   /* replace2306_2 -> 81 in the cache */
   /* replace2306 -> 298 in the cache */

   /* ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2307_0 -> 81 in the cache */
   /* search2307_1_0 -> 81 in the cache */
   /* search2307_1 -> 89 in the cache */
   /* search2307_2_0 -> 271 in the cache */
   /* search2307_2_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 271, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 81, 89, 301 },
      -1,
   } },

   /* replace2307_0 -> 81 in the cache */
   /* replace2307_1 -> 274 in the cache */
   /* replace2307_2 -> 81 in the cache */
   /* replace2307 -> 298 in the cache */

   /* ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2308_0 -> 81 in the cache */
   /* search2308_1_0 -> 271 in the cache */
   /* search2308_1_1_0 -> 81 in the cache */
   /* search2308_1_1 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 271, 89 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 81, 303 },
      -1,
   } },

   /* replace2308_0 -> 81 in the cache */
   /* replace2308_1 -> 274 in the cache */
   /* replace2308_2 -> 81 in the cache */
   /* replace2308 -> 298 in the cache */

   /* ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 305, 306 },
      0,
   } },
   /* search2309_1_0 -> 305 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 305, 308 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 307, 309 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2309_1_0 -> 306 in the cache */
   /* replace2309_1_1 -> 308 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 306, 308 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 311, 312 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b')) */
   /* search2310_0_0 -> 305 in the cache */
   /* search2310_0_1 -> 306 in the cache */
   /* search2310_0 -> 307 in the cache */
   /* search2310_1_0 -> 305 in the cache */
   /* search2310_1_1 -> 308 in the cache */
   /* search2310_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 307, 309 },
      -1,
   } },

   /* replace2310_0 -> 311 in the cache */
   /* replace2310_1_0 -> 306 in the cache */
   /* replace2310_1_1 -> 308 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 306, 308 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 311, 315 },
      -1,
   } },

   /* ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 317 },
      -1,
   } },
   /* search2311_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 318, 43 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x47efffffe0000000 /* 3.4028234663852886e+38 */ },
   } },
   /* search2311_2_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 319, 320, 321 },
      -1,
   } },

   /* replace2311_0_0 -> 42 in the cache */
   /* replace2311_0 -> 321 in the cache */
   /* replace2311_1 -> 320 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 321, 320 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g'))) */
   { .variable = {
      { nir_search_value_variable, -7 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 328, 329 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 2,
      { 326, 327, 330 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 3,
      { 324, 325, 331 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 332, 333 },
      -1,
   } },

   /* replace2312_0 -> 324 in the cache */
   /* replace2312_1 -> 325 in the cache */
   /* replace2312_2_0 -> 326 in the cache */
   /* replace2312_2_1 -> 327 in the cache */
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 335, 336, 337 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 326, 327, 338 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 324, 325, 339 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, -5 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 343, 344 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 341, 342, 345 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 346, 347 },
      -1,
   } },

   /* replace2313_0 -> 341 in the cache */
   /* replace2313_1 -> 342 in the cache */
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 349, 350, 351 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 341, 342, 352 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g'))) */
   /* search2314_0_0_0 -> 324 in the cache */
   /* search2314_0_0_1 -> 325 in the cache */
   /* search2314_0_0_2_0 -> 326 in the cache */
   /* search2314_0_0_2_1 -> 327 in the cache */
   /* search2314_0_0_2_2_0 -> 328 in the cache */
   /* search2314_0_0_2_2_1 -> 329 in the cache */
   /* search2314_0_0_2_2 -> 330 in the cache */
   /* search2314_0_0_2 -> 331 in the cache */
   /* search2314_0_0 -> 332 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 332 },
      -1,
   } },
   /* search2314_1 -> 333 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 354, 333 },
      -1,
   } },

   /* replace2314_0_0 -> 324 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 324 },
      -1,
   } },
   /* replace2314_1 -> 325 in the cache */
   /* replace2314_2_0_0 -> 326 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 326 },
      -1,
   } },
   /* replace2314_2_1 -> 327 in the cache */
   /* replace2314_2_2_0_0 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 335 },
      -1,
   } },
   /* replace2314_2_2_1 -> 336 in the cache */
   /* replace2314_2_2_2 -> 337 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 358, 336, 337 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 357, 327, 359 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 356, 325, 360 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g'))) */
   /* search2315_0_0 -> 81 in the cache */
   /* search2315_0_1 -> 82 in the cache */
   /* search2315_0_2_0 -> 84 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      3, 1,
      { 363, 364 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 2,
      { 84, 362, 365 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 3,
      { 81, 82, 366 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 367, 368 },
      -1,
   } },

   /* replace2315_0 -> 81 in the cache */
   /* replace2315_1 -> 82 in the cache */
   /* replace2315_2_0 -> 84 in the cache */
   /* replace2315_2_1 -> 362 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 370, 371, 372 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 84, 362, 373 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 81, 82, 374 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e')) */
   /* search2316_0_0 -> 81 in the cache */
   /* search2316_0_1 -> 82 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 376, 377 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 81, 82, 378 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 379, 380 },
      -1,
   } },

   /* replace2316_0 -> 81 in the cache */
   /* replace2316_1 -> 82 in the cache */
   /* replace2316_2_0 -> 84 in the cache */
   /* replace2316_2_1 -> 362 in the cache */
   /* replace2316_2_2 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 84, 362, 370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 81, 82, 382 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g'))) */
   /* search2317_0_0_0 -> 81 in the cache */
   /* search2317_0_0_1 -> 82 in the cache */
   /* search2317_0_0_2_0 -> 84 in the cache */
   /* search2317_0_0_2_1 -> 362 in the cache */
   /* search2317_0_0_2_2_0 -> 363 in the cache */
   /* search2317_0_0_2_2_1 -> 364 in the cache */
   /* search2317_0_0_2_2 -> 365 in the cache */
   /* search2317_0_0_2 -> 366 in the cache */
   /* search2317_0_0 -> 367 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 367 },
      -1,
   } },
   /* search2317_1 -> 368 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 384, 368 },
      -1,
   } },

   /* replace2317_0_0 -> 81 in the cache */
   /* replace2317_0 -> 89 in the cache */
   /* replace2317_1 -> 82 in the cache */
   /* replace2317_2_0_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 84 },
      -1,
   } },
   /* replace2317_2_1 -> 362 in the cache */
   /* replace2317_2_2_0_0 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 370 },
      -1,
   } },
   /* replace2317_2_2_1 -> 371 in the cache */
   /* replace2317_2_2_2 -> 372 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 387, 371, 372 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 386, 362, 388 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 89, 82, 389 },
      -1,
   } },

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search2318_0 -> 81 in the cache */
   /* search2318_1 -> 82 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 81, 82, 391 },
      -1,
   } },

   /* replace2318 -> 391 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search2319_0 -> 81 in the cache */
   /* search2319_1 -> 82 in the cache */
   /* search2319_2 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 81, 82, 391 },
      -1,
   } },

   /* replace2319 -> 391 in the cache */

   /* ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b')) */
   /* search2320_0 -> 81 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 81, 394, 395 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace2320_0_1_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 397, 398 },
      -1,
   } },
   /* replace2320_1_0 -> 81 in the cache */
   /* replace2320_1_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 81, 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 399, 400 },
      -1,
   } },

   /* ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c')) */
   /* search2321_0 -> 81 in the cache */
   /* search2321_1 -> 394 in the cache */
   /* search2321_2 -> 395 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 81, 394, 395 },
      -1,
   } },

   /* replace2321_0_0 -> 81 in the cache */
   /* replace2321_0_1_0_0 -> 82 in the cache */
   /* replace2321_0_1_0_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 82, 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 403 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 81, 404 },
      -1,
   } },
   /* replace2321_1_0 -> 84 in the cache */
   /* replace2321_1 -> 398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 405, 398 },
      -1,
   } },

   /* ('ishl', 'a', 0) => a */
   /* search2322_0 -> 42 in the cache */
   /* search2322_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 391 },
      -1,
   } },

   /* replace2322 -> 42 in the cache */

   /* ('ishl', 'a', -32) => a */
   /* search2323_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x20 /* -32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 408 },
      -1,
   } },

   /* replace2323 -> 42 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search2324_0 -> 42 in the cache */
   /* search2324_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 391 },
      -1,
   } },

   /* replace2324 -> 42 in the cache */

   /* ('ishr', 'a', -32) => a */
   /* search2325_0 -> 42 in the cache */
   /* search2325_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 408 },
      -1,
   } },

   /* replace2325 -> 42 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search2326_0 -> 42 in the cache */
   /* search2326_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 42, 391 },
      -1,
   } },

   /* replace2326 -> 42 in the cache */

   /* ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2327_0_0 -> 0 in the cache */
   /* search2327_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 413, 414 },
      -1,
   } },

   /* replace2327_0 -> 0 in the cache */
   /* replace2327_1 -> 1 in the cache */
   /* replace2327 -> 413 in the cache */

   /* ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2328_0_0 -> 0 in the cache */
   /* search2328_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* search2328_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 416, 414 },
      -1,
   } },

   /* replace2328_0 -> 0 in the cache */
   /* replace2328_1 -> 1 in the cache */
   /* replace2328 -> 413 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2329_0_0 -> 0 in the cache */
   /* search2329_0_1 -> 1 in the cache */
   /* search2329_0 -> 413 in the cache */
   /* search2329_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 413, 414 },
      -1,
   } },

   /* replace2329_0 -> 0 in the cache */
   /* replace2329_1 -> 1 in the cache */
   /* replace2329 -> 416 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2330_0_0 -> 0 in the cache */
   /* search2330_0_1 -> 1 in the cache */
   /* search2330_0 -> 416 in the cache */
   /* search2330_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 416, 414 },
      -1,
   } },

   /* replace2330_0 -> 0 in the cache */
   /* replace2330_1 -> 1 in the cache */
   /* replace2330 -> 416 in the cache */

   /* ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 0) => ('bitnz', 'a', ('find_lsb', 'b')) */
   /* search2331_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      7,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 420 },
      -1,
   } },
   /* search2331_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 421, 414 },
      -1,
   } },

   /* replace2331_0 -> 0 in the cache */
   /* replace2331_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 0, 423 },
      -1,
   } },

   /* ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 0) => ('bitz', 'a', ('find_lsb', 'b')) */
   /* search2332_0_0 -> 0 in the cache */
   /* search2332_0_1 -> 420 in the cache */
   /* search2332_0 -> 421 in the cache */
   /* search2332_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 421, 414 },
      -1,
   } },

   /* replace2332_0 -> 0 in the cache */
   /* replace2332_1_0 -> 1 in the cache */
   /* replace2332_1 -> 423 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 0, 423 },
      -1,
   } },

   /* ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('bitz', 'a', ('find_lsb', 'b')) */
   /* search2333_0_0 -> 0 in the cache */
   /* search2333_0_1 -> 420 in the cache */
   /* search2333_0 -> 421 in the cache */
   /* search2333_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 421, 1 },
      -1,
   } },

   /* replace2333_0 -> 0 in the cache */
   /* replace2333_1_0 -> 1 in the cache */
   /* replace2333_1 -> 423 in the cache */
   /* replace2333 -> 426 in the cache */

   /* ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('bitnz', 'a', ('find_lsb', 'b')) */
   /* search2334_0_0 -> 0 in the cache */
   /* search2334_0_1 -> 420 in the cache */
   /* search2334_0 -> 421 in the cache */
   /* search2334_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 421, 1 },
      -1,
   } },

   /* replace2334_0 -> 0 in the cache */
   /* replace2334_1_0 -> 1 in the cache */
   /* replace2334_1 -> 423 in the cache */
   /* replace2334 -> 424 in the cache */

   /* ('ine', ('iand', 'a', ('ishl', 1, 'b')), 0) => ('bitnz', 'a', 'b') */
   /* search2335_0_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   /* search2335_0_1_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 429, 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 42, 430 },
      -1,
   } },
   /* search2335_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 431, 154 },
      -1,
   } },

   /* replace2335_0 -> 42 in the cache */
   /* replace2335_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 42, 82 },
      -1,
   } },

   /* ('ieq', ('iand', 'a', ('ishl', 1, 'b')), 0) => ('bitz', 'a', 'b') */
   /* search2336_0_0 -> 42 in the cache */
   /* search2336_0_1_0 -> 429 in the cache */
   /* search2336_0_1_1 -> 82 in the cache */
   /* search2336_0_1 -> 430 in the cache */
   /* search2336_0 -> 431 in the cache */
   /* search2336_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 431, 154 },
      -1,
   } },

   /* replace2336_0 -> 42 in the cache */
   /* replace2336_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 42, 82 },
      -1,
   } },

   /* ('ine', ('iand', 'a', ('ishl', 1, 'b')), ('ishl', 1, 'b')) => ('bitz', 'a', 'b') */
   /* search2337_0_0 -> 42 in the cache */
   /* search2337_0_1_0 -> 429 in the cache */
   /* search2337_0_1_1 -> 82 in the cache */
   /* search2337_0_1 -> 430 in the cache */
   /* search2337_0 -> 431 in the cache */
   /* search2337_1_0 -> 429 in the cache */
   /* search2337_1_1 -> 82 in the cache */
   /* search2337_1 -> 430 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 431, 430 },
      -1,
   } },

   /* replace2337_0 -> 42 in the cache */
   /* replace2337_1 -> 82 in the cache */
   /* replace2337 -> 435 in the cache */

   /* ('ieq', ('iand', 'a', ('ishl', 1, 'b')), ('ishl', 1, 'b')) => ('bitnz', 'a', 'b') */
   /* search2338_0_0 -> 42 in the cache */
   /* search2338_0_1_0 -> 429 in the cache */
   /* search2338_0_1_1 -> 82 in the cache */
   /* search2338_0_1 -> 430 in the cache */
   /* search2338_0 -> 431 in the cache */
   /* search2338_1_0 -> 429 in the cache */
   /* search2338_1_1 -> 82 in the cache */
   /* search2338_1 -> 430 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 431, 430 },
      -1,
   } },

   /* replace2338_0 -> 42 in the cache */
   /* replace2338_1 -> 82 in the cache */
   /* replace2338 -> 433 in the cache */

   /* ('bitz', ('ushr', 'a', 'b'), 0) => ('bitz', 'a', 'b') */
   /* search2339_0_0 -> 42 in the cache */
   /* search2339_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 42, 82 },
      -1,
   } },
   /* search2339_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 438, 391 },
      -1,
   } },

   /* replace2339_0 -> 42 in the cache */
   /* replace2339_1 -> 82 in the cache */
   /* replace2339 -> 435 in the cache */

   /* ('bitz', ('ishr', 'a', 'b'), 0) => ('bitz', 'a', 'b') */
   /* search2340_0_0 -> 42 in the cache */
   /* search2340_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 82 },
      -1,
   } },
   /* search2340_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 440, 391 },
      -1,
   } },

   /* replace2340_0 -> 42 in the cache */
   /* replace2340_1 -> 82 in the cache */
   /* replace2340 -> 435 in the cache */

   /* ('bitnz', ('ushr', 'a', 'b'), 0) => ('bitnz', 'a', 'b') */
   /* search2341_0_0 -> 42 in the cache */
   /* search2341_0_1 -> 82 in the cache */
   /* search2341_0 -> 438 in the cache */
   /* search2341_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 438, 391 },
      -1,
   } },

   /* replace2341_0 -> 42 in the cache */
   /* replace2341_1 -> 82 in the cache */
   /* replace2341 -> 433 in the cache */

   /* ('bitnz', ('ishr', 'a', 'b'), 0) => ('bitnz', 'a', 'b') */
   /* search2342_0_0 -> 42 in the cache */
   /* search2342_0_1 -> 82 in the cache */
   /* search2342_0 -> 440 in the cache */
   /* search2342_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 440, 391 },
      -1,
   } },

   /* replace2342_0 -> 42 in the cache */
   /* replace2342_1 -> 82 in the cache */
   /* replace2342 -> 433 in the cache */

   /* ('ine', ('ubfe', 'a', 'b', 1), 0) => ('bitnz', 'a', 'b') */
   /* search2343_0_0 -> 81 in the cache */
   /* search2343_0_1 -> 82 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 81, 82, 444 },
      -1,
   } },
   /* search2343_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 445, 391 },
      -1,
   } },

   /* replace2343_0 -> 81 in the cache */
   /* replace2343_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('ieq', ('ubfe', 'a', 'b', 1), 0) => ('bitz', 'a', 'b') */
   /* search2344_0_0 -> 81 in the cache */
   /* search2344_0_1 -> 82 in the cache */
   /* search2344_0_2 -> 444 in the cache */
   /* search2344_0 -> 445 in the cache */
   /* search2344_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 445, 391 },
      -1,
   } },

   /* replace2344_0 -> 81 in the cache */
   /* replace2344_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('ine', ('ubfe', 'a', 'b', 1), 1) => ('bitz', 'a', 'b') */
   /* search2345_0_0 -> 81 in the cache */
   /* search2345_0_1 -> 82 in the cache */
   /* search2345_0_2 -> 444 in the cache */
   /* search2345_0 -> 445 in the cache */
   /* search2345_1 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 445, 444 },
      -1,
   } },

   /* replace2345_0 -> 81 in the cache */
   /* replace2345_1 -> 82 in the cache */
   /* replace2345 -> 449 in the cache */

   /* ('ieq', ('ubfe', 'a', 'b', 1), 1) => ('bitnz', 'a', 'b') */
   /* search2346_0_0 -> 81 in the cache */
   /* search2346_0_1 -> 82 in the cache */
   /* search2346_0_2 -> 444 in the cache */
   /* search2346_0 -> 445 in the cache */
   /* search2346_1 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 445, 444 },
      -1,
   } },

   /* replace2346_0 -> 81 in the cache */
   /* replace2346_1 -> 82 in the cache */
   /* replace2346 -> 447 in the cache */

   /* ('ine', ('ibfe', 'a', 'b', 1), 0) => ('bitnz', 'a', 'b') */
   /* search2347_0_0 -> 81 in the cache */
   /* search2347_0_1 -> 82 in the cache */
   /* search2347_0_2 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 81, 82, 444 },
      -1,
   } },
   /* search2347_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 452, 391 },
      -1,
   } },

   /* replace2347_0 -> 81 in the cache */
   /* replace2347_1 -> 82 in the cache */
   /* replace2347 -> 447 in the cache */

   /* ('ieq', ('ibfe', 'a', 'b', 1), 0) => ('bitz', 'a', 'b') */
   /* search2348_0_0 -> 81 in the cache */
   /* search2348_0_1 -> 82 in the cache */
   /* search2348_0_2 -> 444 in the cache */
   /* search2348_0 -> 452 in the cache */
   /* search2348_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 452, 391 },
      -1,
   } },

   /* replace2348_0 -> 81 in the cache */
   /* replace2348_1 -> 82 in the cache */
   /* replace2348 -> 449 in the cache */

   /* ('ine', ('ibfe', 'a', 'b', 1), -1) => ('bitz', 'a', 'b') */
   /* search2349_0_0 -> 81 in the cache */
   /* search2349_0_1 -> 82 in the cache */
   /* search2349_0_2 -> 444 in the cache */
   /* search2349_0 -> 452 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 452, 455 },
      -1,
   } },

   /* replace2349_0 -> 81 in the cache */
   /* replace2349_1 -> 82 in the cache */
   /* replace2349 -> 449 in the cache */

   /* ('ieq', ('ibfe', 'a', 'b', 1), -1) => ('bitnz', 'a', 'b') */
   /* search2350_0_0 -> 81 in the cache */
   /* search2350_0_1 -> 82 in the cache */
   /* search2350_0_2 -> 444 in the cache */
   /* search2350_0 -> 452 in the cache */
   /* search2350_1 -> 455 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 452, 455 },
      -1,
   } },

   /* replace2350_0 -> 81 in the cache */
   /* replace2350_1 -> 82 in the cache */
   /* replace2350 -> 447 in the cache */

   /* ('inot', ('bitnz', 'a', 'b')) => ('bitz', 'a', 'b') */
   /* search2351_0_0 -> 42 in the cache */
   /* search2351_0_1 -> 82 in the cache */
   /* search2351_0 -> 433 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 433 },
      -1,
   } },

   /* replace2351_0 -> 42 in the cache */
   /* replace2351_1 -> 82 in the cache */
   /* replace2351 -> 435 in the cache */

   /* ('inot', ('bitz', 'a', 'b')) => ('bitnz', 'a', 'b') */
   /* search2352_0_0 -> 42 in the cache */
   /* search2352_0_1 -> 82 in the cache */
   /* search2352_0 -> 435 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 435 },
      -1,
   } },

   /* replace2352_0 -> 42 in the cache */
   /* replace2352_1 -> 82 in the cache */
   /* replace2352 -> 433 in the cache */

   /* ('bitnz', ('inot', 'a'), 'b') => ('bitz', 'a', 'b') */
   /* search2353_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 42 },
      -1,
   } },
   /* search2353_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitnz,
      -1, 0,
      { 460, 82 },
      -1,
   } },

   /* replace2353_0 -> 42 in the cache */
   /* replace2353_1 -> 82 in the cache */
   /* replace2353 -> 435 in the cache */

   /* ('bitz', ('inot', 'a'), 'b') => ('bitnz', 'a', 'b') */
   /* search2354_0_0 -> 42 in the cache */
   /* search2354_0 -> 460 in the cache */
   /* search2354_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bitz,
      -1, 0,
      { 460, 82 },
      -1,
   } },

   /* replace2354_0 -> 42 in the cache */
   /* replace2354_1 -> 82 in the cache */
   /* replace2354 -> 433 in the cache */

   /* ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2355_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 463 },
      -1,
   } },

   /* replace2355_0 -> 61 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 61, 465 },
      -1,
   } },

   /* ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2356_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 467 },
      -1,
   } },

   /* replace2356_0 -> 61 in the cache */
   /* replace2356_1 -> 465 in the cache */
   /* replace2356 -> 466 in the cache */

   /* ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2357_0_0 -> 61 in the cache */
   /* search2357_0 -> 463 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 463 },
      -1,
   } },

   /* replace2357_0 -> 61 in the cache */
   /* replace2357_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 61, 465 },
      -1,
   } },

   /* ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2358_0_0 -> 61 in the cache */
   /* search2358_0 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 467 },
      -1,
   } },

   /* replace2358_0 -> 61 in the cache */
   /* replace2358_1 -> 465 in the cache */
   /* replace2358 -> 470 in the cache */

   /* ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2359_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 472 },
      -1,
   } },

   /* replace2359_0 -> 81 in the cache */
   /* replace2359_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 81, 391 },
      -1,
   } },

   /* ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2360_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 475 },
      -1,
   } },

   /* replace2360_0 -> 81 in the cache */
   /* replace2360_1 -> 391 in the cache */
   /* replace2360 -> 474 in the cache */

   /* ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2361_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 477 },
      -1,
   } },

   /* replace2361_0 -> 81 in the cache */
   /* replace2361_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 81, 391 },
      -1,
   } },

   /* ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2362_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 480 },
      -1,
   } },

   /* replace2362_0 -> 81 in the cache */
   /* replace2362_1 -> 391 in the cache */
   /* replace2362 -> 479 in the cache */

   /* ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2363_0_0 -> 81 in the cache */
   /* search2363_0 -> 472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 472 },
      -1,
   } },

   /* replace2363_0 -> 81 in the cache */
   /* replace2363_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 81, 391 },
      -1,
   } },

   /* ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2364_0_0 -> 81 in the cache */
   /* search2364_0 -> 475 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 475 },
      -1,
   } },

   /* replace2364_0 -> 81 in the cache */
   /* replace2364_1 -> 391 in the cache */
   /* replace2364 -> 483 in the cache */

   /* ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2365_0_0 -> 81 in the cache */
   /* search2365_0 -> 477 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 477 },
      -1,
   } },

   /* replace2365_0 -> 81 in the cache */
   /* replace2365_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 81, 391 },
      -1,
   } },

   /* ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2366_0_0 -> 81 in the cache */
   /* search2366_0 -> 480 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 480 },
      -1,
   } },

   /* replace2366_0 -> 81 in the cache */
   /* replace2366_1 -> 391 in the cache */
   /* replace2366 -> 486 in the cache */

   /* ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1) */
   /* search2367_0_0 -> 81 in the cache */
   /* search2367_0_1 -> 391 in the cache */
   /* search2367_0 -> 474 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 474, 488 },
      -1,
   } },

   /* replace2367_0 -> 81 in the cache */
   /* replace2367_1 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 444 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2) */
   /* search2368_0_0 -> 81 in the cache */
   /* search2368_0_1 -> 391 in the cache */
   /* search2368_0 -> 474 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 474, 491 },
      -1,
   } },

   /* replace2368_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 493 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3) */
   /* search2369_0_0 -> 81 in the cache */
   /* search2369_0_1 -> 391 in the cache */
   /* search2369_0 -> 474 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 474, 495 },
      -1,
   } },

   /* replace2369_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 497 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1) */
   /* search2370_0_0 -> 81 in the cache */
   /* search2370_0_1 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 488 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00 /* 65280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 499, 500 },
      -1,
   } },

   /* replace2370_0 -> 81 in the cache */
   /* replace2370_1 -> 444 in the cache */
   /* replace2370 -> 490 in the cache */

   /* ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2) */
   /* search2371_0_0 -> 81 in the cache */
   /* search2371_0_1 -> 491 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 491 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff0000 /* 16711680 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 502, 503 },
      -1,
   } },

   /* replace2371_0 -> 81 in the cache */
   /* replace2371_1 -> 493 in the cache */
   /* replace2371 -> 494 in the cache */

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3) */
   /* search2372_0_0 -> 81 in the cache */
   /* search2372_0_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 495 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 505, 506 },
      -1,
   } },

   /* replace2372_0 -> 81 in the cache */
   /* replace2372_1 -> 497 in the cache */
   /* replace2372 -> 498 in the cache */

   /* ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3) */
   /* search2373_0 -> 81 in the cache */
   /* search2373_1 -> 495 in the cache */
   /* search2373 -> 505 in the cache */

   /* replace2373_0 -> 81 in the cache */
   /* replace2373_1 -> 497 in the cache */
   /* replace2373 -> 498 in the cache */

   /* ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1) */
   /* search2374_0 -> 81 in the cache */
   /* search2374_1 -> 491 in the cache */
   /* search2374 -> 502 in the cache */

   /* replace2374_0 -> 81 in the cache */
   /* replace2374_1 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 81, 444 },
      -1,
   } },

   /* ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b') */
   /* search2375_0_0 -> 0 in the cache */
   /* search2375_0_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 414 },
      -1,
   } },
   /* search2375_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 509, 1 },
      -1,
   } },

   /* replace2375_0 -> 0 in the cache */
   /* replace2375_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b') */
   /* search2376_0_0 -> 0 in the cache */
   /* search2376_0_1 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 414 },
      -1,
   } },
   /* search2376_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 512, 1 },
      -1,
   } },

   /* replace2376_0 -> 0 in the cache */
   /* replace2376_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2377_0_0 -> 134 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 134, 515 },
      0,
   } },
   /* search2377_1_0 -> 135 in the cache */
   /* search2377_1_1 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 135, 515 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 516, 517 },
      -1,
   } },

   /* replace2377_0_0 -> 134 in the cache */
   /* replace2377_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 134, 135 },
      -1,
   } },
   /* replace2377_1 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 519, 515 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2378_0_0 -> 134 in the cache */
   /* search2378_0_1 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 134, 515 },
      0,
   } },
   /* search2378_1_0 -> 135 in the cache */
   /* search2378_1_1 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 135, 515 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 521, 522 },
      -1,
   } },

   /* replace2378_0_0 -> 134 in the cache */
   /* replace2378_0_1 -> 135 in the cache */
   /* replace2378_0 -> 519 in the cache */
   /* replace2378_1 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 519, 515 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2379_0_0 -> 61 in the cache */
   /* search2379_0_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 61, 465 },
      0,
   } },
   /* search2379_1_0 -> 62 in the cache */
   /* search2379_1_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 62, 465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 525, 526 },
      -1,
   } },

   /* replace2379_0_0 -> 61 in the cache */
   /* replace2379_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 61, 62 },
      -1,
   } },
   /* replace2379_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 528, 465 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2380_0_0 -> 61 in the cache */
   /* search2380_0_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 61, 465 },
      0,
   } },
   /* search2380_1_0 -> 62 in the cache */
   /* search2380_1_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 62, 465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 530, 531 },
      -1,
   } },

   /* replace2380_0_0 -> 61 in the cache */
   /* replace2380_0_1 -> 62 in the cache */
   /* replace2380_0 -> 528 in the cache */
   /* replace2380_1 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 528, 465 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2381_0_0 -> 81 in the cache */
   /* search2381_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 81, 391 },
      0,
   } },
   /* search2381_1_0 -> 82 in the cache */
   /* search2381_1_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 82, 391 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 534, 535 },
      -1,
   } },

   /* replace2381_0_0 -> 81 in the cache */
   /* replace2381_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 81, 82 },
      -1,
   } },
   /* replace2381_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 537, 391 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2382_0_0 -> 81 in the cache */
   /* search2382_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 81, 391 },
      0,
   } },
   /* search2382_1_0 -> 82 in the cache */
   /* search2382_1_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 82, 391 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 539, 540 },
      -1,
   } },

   /* replace2382_0_0 -> 81 in the cache */
   /* replace2382_0_1 -> 82 in the cache */
   /* replace2382_0 -> 537 in the cache */
   /* replace2382_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 537, 391 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2383_0_0 -> 114 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 114, 543 },
      0,
   } },
   /* search2383_1_0 -> 115 in the cache */
   /* search2383_1_1 -> 543 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 115, 543 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 544, 545 },
      -1,
   } },

   /* replace2383_0_0 -> 114 in the cache */
   /* replace2383_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 114, 115 },
      -1,
   } },
   /* replace2383_1 -> 543 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 547, 543 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2384_0_0 -> 114 in the cache */
   /* search2384_0_1 -> 543 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 114, 543 },
      0,
   } },
   /* search2384_1_0 -> 115 in the cache */
   /* search2384_1_1 -> 543 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 115, 543 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 549, 550 },
      -1,
   } },

   /* replace2384_0_0 -> 114 in the cache */
   /* replace2384_0_1 -> 115 in the cache */
   /* replace2384_0 -> 547 in the cache */
   /* replace2384_1 -> 543 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 547, 543 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2385_0 -> 183 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2385_1_1_0 -> 62 in the cache */
   /* search2385_1_1_1 -> 182 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 62, 182 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 553, 554 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 183, 555 },
      -1,
   } },

   /* replace2385_0_0 -> 183 in the cache */
   /* replace2385_0_1_0 -> 553 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 553 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 183, 557 },
      -1,
   } },
   /* replace2385_1_0 -> 62 in the cache */
   /* replace2385_1_1 -> 553 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 62, 553 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 558, 559 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b'))) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2386_1 -> 465 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 562 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 563 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 465, 564 },
      -1,
   } },

   /* replace2386_0_0_0 -> 561 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 561, 566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 567 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 1,
      { 568 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2387_0 -> 274 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2387_1_1_0 -> 82 in the cache */
   /* search2387_1_1_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 82, 272 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 570, 571 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 274, 572 },
      -1,
   } },

   /* replace2387_0_0 -> 274 in the cache */
   /* replace2387_0_1_0 -> 570 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 274, 574 },
      -1,
   } },
   /* replace2387_1_0 -> 82 in the cache */
   /* replace2387_1_1 -> 570 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 82, 570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 575, 576 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b'))) */
   /* search2388_0 -> 561 in the cache */
   /* search2388_1 -> 391 in the cache */
   /* search2388_2_0_0 -> 562 in the cache */
   /* search2388_2_0 -> 563 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 563 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 391, 578 },
      -1,
   } },

   /* replace2388_0_0_0 -> 561 in the cache */
   /* replace2388_0_0_1 -> 566 in the cache */
   /* replace2388_0_0 -> 567 in the cache */
   /* replace2388_0 -> 568 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 1,
      { 568 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2389_1_1_0 -> 115 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 115, 583 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 582, 584 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 581, 585 },
      -1,
   } },

   /* replace2389_0_0 -> 581 in the cache */
   /* replace2389_0_1_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 581, 587 },
      -1,
   } },
   /* replace2389_1_0 -> 115 in the cache */
   /* replace2389_1_1 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 115, 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 588, 589 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b'))) */
   /* search2390_0 -> 561 in the cache */
   /* search2390_1 -> 543 in the cache */
   /* search2390_2_0_0 -> 562 in the cache */
   /* search2390_2_0 -> 563 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 563 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 543, 591 },
      -1,
   } },

   /* replace2390_0_0_0 -> 561 in the cache */
   /* replace2390_0_0_1 -> 566 in the cache */
   /* replace2390_0_0 -> 567 in the cache */
   /* replace2390_0 -> 568 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 1,
      { 568 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2391_0 -> 561 in the cache */
   /* search2391_1_0 -> 208 in the cache */
   /* search2391_1_1 -> 210 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 208, 210 },
      0,
   } },
   /* search2391_2_0 -> 208 in the cache */
   /* search2391_2_1 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 208, 211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 594, 595 },
      -1,
   } },

   /* replace2391_0 -> 208 in the cache */
   /* replace2391_1_0 -> 561 in the cache */
   /* replace2391_1_1 -> 210 in the cache */
   /* replace2391_1_2 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 210, 211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 208, 597 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2392_0 -> 561 in the cache */
   /* search2392_1_0 -> 208 in the cache */
   /* search2392_1_1 -> 210 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 208, 210 },
      -1,
   } },
   /* search2392_2_0 -> 208 in the cache */
   /* search2392_2_1 -> 211 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 208, 211 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 599, 600 },
      -1,
   } },

   /* replace2392_0 -> 208 in the cache */
   /* replace2392_1_0 -> 561 in the cache */
   /* replace2392_1_1 -> 210 in the cache */
   /* replace2392_1_2 -> 211 in the cache */
   /* replace2392_1 -> 597 in the cache */
   /* replace2392 -> 598 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2393_0 -> 561 in the cache */
   /* search2393_1_0 -> 342 in the cache */
   /* search2393_1_1 -> 349 in the cache */
   /* search2393_1_2 -> 350 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 342, 349, 350 },
      0,
   } },
   /* search2393_2_0 -> 342 in the cache */
   /* search2393_2_1 -> 349 in the cache */
   /* search2393_2_2 -> 351 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 342, 349, 351 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 602, 603 },
      -1,
   } },

   /* replace2393_0 -> 342 in the cache */
   /* replace2393_1 -> 349 in the cache */
   /* replace2393_2_0 -> 561 in the cache */
   /* replace2393_2_1 -> 350 in the cache */
   /* replace2393_2_2 -> 351 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 350, 351 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 342, 349, 605 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2394_0 -> 561 in the cache */
   /* search2394_1_0 -> 342 in the cache */
   /* search2394_1_1 -> 349 in the cache */
   /* search2394_1_2 -> 350 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 342, 349, 350 },
      -1,
   } },
   /* search2394_2_0 -> 342 in the cache */
   /* search2394_2_1 -> 349 in the cache */
   /* search2394_2_2 -> 351 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 342, 349, 351 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 607, 608 },
      -1,
   } },

   /* replace2394_0 -> 342 in the cache */
   /* replace2394_1 -> 349 in the cache */
   /* replace2394_2_0 -> 561 in the cache */
   /* replace2394_2_1 -> 350 in the cache */
   /* replace2394_2_2 -> 351 in the cache */
   /* replace2394_2 -> 605 in the cache */
   /* replace2394 -> 606 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2395_0 -> 561 in the cache */
   /* search2395_1_0 -> 342 in the cache */
   /* search2395_1_1 -> 349 in the cache */
   /* search2395_1_2 -> 350 in the cache */
   /* search2395_1 -> 602 in the cache */
   /* search2395_2_0 -> 342 in the cache */
   /* search2395_2_1 -> 351 in the cache */
   /* search2395_2_2 -> 350 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 342, 351, 350 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 602, 610 },
      -1,
   } },

   /* replace2395_0 -> 342 in the cache */
   /* replace2395_1_0 -> 561 in the cache */
   /* replace2395_1_1 -> 349 in the cache */
   /* replace2395_1_2 -> 351 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 349, 351 },
      -1,
   } },
   /* replace2395_2 -> 350 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 342, 612, 350 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2396_0 -> 561 in the cache */
   /* search2396_1_0 -> 342 in the cache */
   /* search2396_1_1 -> 349 in the cache */
   /* search2396_1_2 -> 350 in the cache */
   /* search2396_1 -> 607 in the cache */
   /* search2396_2_0 -> 342 in the cache */
   /* search2396_2_1 -> 351 in the cache */
   /* search2396_2_2 -> 350 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 342, 351, 350 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 607, 614 },
      -1,
   } },

   /* replace2396_0 -> 342 in the cache */
   /* replace2396_1_0 -> 561 in the cache */
   /* replace2396_1_1 -> 349 in the cache */
   /* replace2396_1_2 -> 351 in the cache */
   /* replace2396_1 -> 612 in the cache */
   /* replace2396_2 -> 350 in the cache */
   /* replace2396 -> 613 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2397_0 -> 561 in the cache */
   /* search2397_1_0 -> 82 in the cache */
   /* search2397_1_1 -> 84 in the cache */
   /* search2397_1_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 362 },
      0,
   } },
   /* search2397_2_0 -> 82 in the cache */
   /* search2397_2_1 -> 84 in the cache */
   /* search2397_2_2 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 84, 370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 616, 617 },
      -1,
   } },

   /* replace2397_0 -> 82 in the cache */
   /* replace2397_1 -> 84 in the cache */
   /* replace2397_2_0 -> 561 in the cache */
   /* replace2397_2_1 -> 362 in the cache */
   /* replace2397_2_2 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 362, 370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 619 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2398_0 -> 561 in the cache */
   /* search2398_1_0 -> 82 in the cache */
   /* search2398_1_1 -> 84 in the cache */
   /* search2398_1_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 362 },
      -1,
   } },
   /* search2398_2_0 -> 82 in the cache */
   /* search2398_2_1 -> 84 in the cache */
   /* search2398_2_2 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 84, 370 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 621, 622 },
      -1,
   } },

   /* replace2398_0 -> 82 in the cache */
   /* replace2398_1 -> 84 in the cache */
   /* replace2398_2_0 -> 561 in the cache */
   /* replace2398_2_1 -> 362 in the cache */
   /* replace2398_2_2 -> 370 in the cache */
   /* replace2398_2 -> 619 in the cache */
   /* replace2398 -> 620 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2399_0 -> 561 in the cache */
   /* search2399_1_0 -> 82 in the cache */
   /* search2399_1_1 -> 84 in the cache */
   /* search2399_1_2 -> 362 in the cache */
   /* search2399_1 -> 616 in the cache */
   /* search2399_2_0 -> 82 in the cache */
   /* search2399_2_1 -> 370 in the cache */
   /* search2399_2_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 370, 362 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 616, 624 },
      -1,
   } },

   /* replace2399_0 -> 82 in the cache */
   /* replace2399_1_0 -> 561 in the cache */
   /* replace2399_1_1 -> 84 in the cache */
   /* replace2399_1_2 -> 370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 561, 84, 370 },
      -1,
   } },
   /* replace2399_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 626, 362 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2400_0 -> 561 in the cache */
   /* search2400_1_0 -> 82 in the cache */
   /* search2400_1_1 -> 84 in the cache */
   /* search2400_1_2 -> 362 in the cache */
   /* search2400_1 -> 621 in the cache */
   /* search2400_2_0 -> 82 in the cache */
   /* search2400_2_1 -> 370 in the cache */
   /* search2400_2_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 370, 362 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 561, 621, 628 },
      -1,
   } },

   /* replace2400_0 -> 82 in the cache */
   /* replace2400_1_0 -> 561 in the cache */
   /* replace2400_1_1 -> 84 in the cache */
   /* replace2400_1_2 -> 370 in the cache */
   /* replace2400_1 -> 626 in the cache */
   /* replace2400_2 -> 362 in the cache */
   /* replace2400 -> 627 in the cache */

   /* ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a') */
   /* search2401_0_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 631 },
      -1,
   } },

   /* replace2401_0 -> 81 in the cache */
   /* replace2401 -> 93 in the cache */

   /* ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a') */
   /* search2402_0_0_0 -> 81 in the cache */
   /* search2402_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 633 },
      -1,
   } },

   /* replace2402_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a') */
   /* search2403_0_0_0 -> 81 in the cache */
   /* search2403_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 636 },
      -1,
   } },

   /* replace2403_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a') */
   /* search2404_0_0_0 -> 81 in the cache */
   /* search2404_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 639 },
      -1,
   } },

   /* replace2404_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a') */
   /* search2405_0_0_0 -> 81 in the cache */
   /* search2405_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 642 },
      -1,
   } },

   /* replace2405_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a') */
   /* search2406_0_0_0 -> 81 in the cache */
   /* search2406_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 645 },
      -1,
   } },

   /* replace2406_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a') */
   /* search2407_0_0_0 -> 81 in the cache */
   /* search2407_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 648 },
      -1,
   } },

   /* replace2407_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a') */
   /* search2408_0_0_0 -> 81 in the cache */
   /* search2408_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 651 },
      -1,
   } },

   /* replace2408_0 -> 81 in the cache */
   /* replace2408 -> 89 in the cache */

   /* ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a') */
   /* search2409_0_0_0 -> 81 in the cache */
   /* search2409_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 653 },
      -1,
   } },

   /* replace2409_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a') */
   /* search2410_0_0_0 -> 81 in the cache */
   /* search2410_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 656 },
      -1,
   } },

   /* replace2410_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a') */
   /* search2411_0_0_0 -> 81 in the cache */
   /* search2411_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 659 },
      -1,
   } },

   /* replace2411_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a') */
   /* search2412_0_0_0 -> 81 in the cache */
   /* search2412_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 662 },
      -1,
   } },

   /* replace2412_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a') */
   /* search2413_0_0_0 -> 81 in the cache */
   /* search2413_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 665 },
      -1,
   } },

   /* replace2413_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a') */
   /* search2414_0_0_0 -> 81 in the cache */
   /* search2414_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 668 },
      -1,
   } },

   /* replace2414_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a') */
   /* search2415_0_0_0 -> 81 in the cache */
   /* search2415_0_0 -> 630 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 671 },
      -1,
   } },

   /* replace2415_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b') */
   /* search2416_0_0_0 -> 81 in the cache */
   /* search2416_0_0 -> 630 in the cache */
   /* search2416_0_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 675 },
      -1,
   } },

   /* replace2416_0 -> 81 in the cache */
   /* replace2416_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b') */
   /* search2417_0_0_0 -> 81 in the cache */
   /* search2417_0_0 -> 630 in the cache */
   /* search2417_0_1_0 -> 82 in the cache */
   /* search2417_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 678 },
      -1,
   } },

   /* replace2417_0 -> 81 in the cache */
   /* replace2417_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b') */
   /* search2418_0_0_0 -> 81 in the cache */
   /* search2418_0_0 -> 630 in the cache */
   /* search2418_0_1_0 -> 82 in the cache */
   /* search2418_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 681 },
      -1,
   } },

   /* replace2418_0 -> 81 in the cache */
   /* replace2418_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b') */
   /* search2419_0_0_0 -> 81 in the cache */
   /* search2419_0_0 -> 630 in the cache */
   /* search2419_0_1_0 -> 82 in the cache */
   /* search2419_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 684 },
      -1,
   } },

   /* replace2419_0 -> 81 in the cache */
   /* replace2419_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b') */
   /* search2420_0_0_0 -> 81 in the cache */
   /* search2420_0_0 -> 630 in the cache */
   /* search2420_0_1_0 -> 82 in the cache */
   /* search2420_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 687 },
      -1,
   } },

   /* replace2420_0 -> 81 in the cache */
   /* replace2420_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b') */
   /* search2421_0_0_0 -> 81 in the cache */
   /* search2421_0_0 -> 630 in the cache */
   /* search2421_0_1_0 -> 82 in the cache */
   /* search2421_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 690 },
      -1,
   } },

   /* replace2421_0 -> 81 in the cache */
   /* replace2421_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b') */
   /* search2422_0_0_0 -> 81 in the cache */
   /* search2422_0_0 -> 630 in the cache */
   /* search2422_0_1_0 -> 82 in the cache */
   /* search2422_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 693 },
      -1,
   } },

   /* replace2422_0 -> 81 in the cache */
   /* replace2422_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b') */
   /* search2423_0_0_0 -> 81 in the cache */
   /* search2423_0_0 -> 630 in the cache */
   /* search2423_0_1_0 -> 82 in the cache */
   /* search2423_0_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 630, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 696 },
      -1,
   } },

   /* replace2423_0 -> 81 in the cache */
   /* replace2423_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c') */
   /* search2424_0_0_0 -> 81 in the cache */
   /* search2424_0_0 -> 630 in the cache */
   /* search2424_0_1_0 -> 82 in the cache */
   /* search2424_0_1 -> 674 in the cache */
   /* search2424_0_2_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 630, 674, 699 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 700 },
      -1,
   } },

   /* replace2424_0 -> 81 in the cache */
   /* replace2424_1 -> 82 in the cache */
   /* replace2424_2 -> 84 in the cache */
   /* replace2424 -> 86 in the cache */

   /* ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c') */
   /* search2425_0_0_0 -> 81 in the cache */
   /* search2425_0_0 -> 630 in the cache */
   /* search2425_0_1_0 -> 82 in the cache */
   /* search2425_0_1 -> 674 in the cache */
   /* search2425_0_2_0 -> 84 in the cache */
   /* search2425_0_2 -> 699 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 630, 674, 699 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 702 },
      -1,
   } },

   /* replace2425_0 -> 81 in the cache */
   /* replace2425_1 -> 82 in the cache */
   /* replace2425_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b') */
   /* search2426_0_0 -> 81 in the cache */
   /* search2426_0 -> 630 in the cache */
   /* search2426_1_0 -> 82 in the cache */
   /* search2426_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 630, 674 },
      -1,
   } },

   /* replace2426_0 -> 81 in the cache */
   /* replace2426_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b') */
   /* search2427_0_0 -> 81 in the cache */
   /* search2427_0 -> 630 in the cache */
   /* search2427_1_0 -> 82 in the cache */
   /* search2427_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 630, 674 },
      -1,
   } },

   /* replace2427_0 -> 81 in the cache */
   /* replace2427_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b') */
   /* search2428_0_0 -> 81 in the cache */
   /* search2428_0 -> 630 in the cache */
   /* search2428_1_0 -> 82 in the cache */
   /* search2428_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 630, 674 },
      -1,
   } },

   /* replace2428_0 -> 81 in the cache */
   /* replace2428_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b') */
   /* search2429_0_0 -> 81 in the cache */
   /* search2429_0 -> 630 in the cache */
   /* search2429_1_0 -> 82 in the cache */
   /* search2429_1 -> 674 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 630, 674 },
      -1,
   } },

   /* replace2429_0 -> 81 in the cache */
   /* replace2429_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('f2fmp', 'a') => ('f2f16', 'a') */
   /* search2430_0 -> 81 in the cache */
   /* search2430 -> 630 in the cache */

   /* replace2430_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('f2imp', 'a') => ('f2i16', 'a') */
   /* search2431_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2431_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('f2ump', 'a') => ('f2u16', 'a') */
   /* search2432_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2432_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('i2imp', 'a') => ('i2i16', 'a') */
   /* search2433_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2433_0 -> 81 in the cache */
   /* replace2433 -> 480 in the cache */

   /* ('i2fmp', 'a') => ('i2f16', 'a') */
   /* search2434_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2434_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('i2imp', 'a') => ('u2u16', 'a') */
   /* search2435_0 -> 81 in the cache */
   /* search2435 -> 718 in the cache */

   /* replace2435_0 -> 81 in the cache */
   /* replace2435 -> 477 in the cache */

   /* ('u2fmp', 'a') => ('u2f16', 'a') */
   /* search2436_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2436_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf) */
   /* search2437_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisfinite,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2437_0_0 -> 42 in the cache */
   /* replace2437_0 -> 47 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x7ff0000000000000 /* inf */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 47, 724 },
      -1,
   } },

   /* ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search2438_0_0 -> 391 in the cache */
   /* search2438_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2438_1 -> 82 in the cache */
   /* search2438_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 726, 82, 84 },
      -1,
   } },

   /* replace2438_0 -> 81 in the cache */
   /* replace2438_1 -> 82 in the cache */
   /* replace2438_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search2439_0_0 -> 81 in the cache */
   /* search2439_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2439_1 -> 82 in the cache */
   /* search2439_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 729, 82, 84 },
      -1,
   } },

   /* replace2439_0_0 -> 81 in the cache */
   /* replace2439_0 -> 89 in the cache */
   /* replace2439_1 -> 82 in the cache */
   /* replace2439_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 89, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search2440_0_0 -> 81 in the cache */
   /* search2440_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2440_1 -> 82 in the cache */
   /* search2440_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 732, 82, 84 },
      -1,
   } },

   /* replace2440_0 -> 81 in the cache */
   /* replace2440_1 -> 82 in the cache */
   /* replace2440_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search2441_0_0 -> 391 in the cache */
   /* search2441_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2441_1 -> 82 in the cache */
   /* search2441_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 735, 82, 84 },
      -1,
   } },

   /* replace2441_0_0 -> 81 in the cache */
   /* replace2441_0 -> 89 in the cache */
   /* replace2441_1 -> 82 in the cache */
   /* replace2441_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 89, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c') */
   /* search2442_0_0 -> 391 in the cache */
   /* search2442_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2442_1 -> 82 in the cache */
   /* search2442_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 738, 82, 84 },
      -1,
   } },

   /* replace2442_0 -> 81 in the cache */
   /* replace2442_1 -> 82 in the cache */
   /* replace2442_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b') */
   /* search2443_0_0 -> 81 in the cache */
   /* search2443_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2443_1 -> 82 in the cache */
   /* search2443_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 741, 82, 84 },
      -1,
   } },

   /* replace2443_0 -> 81 in the cache */
   /* replace2443_1 -> 84 in the cache */
   /* replace2443_2 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 81, 84, 82 },
      -1,
   } },

   /* ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c') */
   /* search2444_0_0 -> 81 in the cache */
   /* search2444_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2444_1 -> 82 in the cache */
   /* search2444_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 744, 82, 84 },
      -1,
   } },

   /* replace2444_0 -> 81 in the cache */
   /* replace2444_1 -> 82 in the cache */
   /* replace2444_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b') */
   /* search2445_0_0 -> 391 in the cache */
   /* search2445_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2445_1 -> 82 in the cache */
   /* search2445_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 747, 82, 84 },
      -1,
   } },

   /* replace2445_0 -> 81 in the cache */
   /* replace2445_1 -> 84 in the cache */
   /* replace2445_2 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 81, 84, 82 },
      -1,
   } },

   /* ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search2446_0_0 -> 391 in the cache */
   /* search2446_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2446_1 -> 82 in the cache */
   /* search2446_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 750, 82, 84 },
      -1,
   } },

   /* replace2446_0 -> 81 in the cache */
   /* replace2446_1 -> 82 in the cache */
   /* replace2446_2 -> 84 in the cache */
   /* replace2446 -> 728 in the cache */

   /* ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search2447_0_0 -> 81 in the cache */
   /* search2447_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2447_1 -> 82 in the cache */
   /* search2447_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 752, 82, 84 },
      -1,
   } },

   /* replace2447_0_0 -> 81 in the cache */
   /* replace2447_0 -> 89 in the cache */
   /* replace2447_1 -> 82 in the cache */
   /* replace2447_2 -> 84 in the cache */
   /* replace2447 -> 731 in the cache */

   /* ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search2448_0_0 -> 81 in the cache */
   /* search2448_0_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 81, 391 },
      -1,
   } },
   /* search2448_1 -> 82 in the cache */
   /* search2448_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 754, 82, 84 },
      -1,
   } },

   /* replace2448_0 -> 81 in the cache */
   /* replace2448_1 -> 82 in the cache */
   /* replace2448_2 -> 84 in the cache */
   /* replace2448 -> 734 in the cache */

   /* ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search2449_0_0 -> 391 in the cache */
   /* search2449_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 391, 81 },
      -1,
   } },
   /* search2449_1 -> 82 in the cache */
   /* search2449_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 756, 82, 84 },
      -1,
   } },

   /* replace2449_0_0 -> 81 in the cache */
   /* replace2449_0 -> 89 in the cache */
   /* replace2449_1 -> 82 in the cache */
   /* replace2449_2 -> 84 in the cache */
   /* replace2449 -> 737 in the cache */

};

static const nir_search_expression_cond nir_opt_algebraic_late_expression_cond[] = {
   (is_used_once),
   (is_only_used_by_fadd),
   (is_only_used_as_float),
};

static const nir_search_variable_cond nir_opt_algebraic_late_variable_cond[] = {
   (is_finite),
   (is_not_const),
   (is_16_bits),
   (is_gt_0_and_lt_1),
   (is_not_fmul),
   (is_not_negative),
   (is_not_const_and_not_fsign),
   (is_pos_power_of_two),
};

static const struct transform nir_opt_algebraic_late_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 240, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 155, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 156, 157, 133 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 248, 249, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 250, 251, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 254, 255, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 261, 270, 137 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 275, 18 },
   { 277, 278, 18 },
   { 279, 280, 18 },
   { 281, 282, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 392, 391, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 392, 391, 0 },
   { 396, 401, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 391, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 393, 391, 0 },
   { 402, 406, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 407, 42, 0 },
   { 409, 42, 0 },
   { 505, 498, 140 },
   { 502, 508, 141 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 407, 42, 0 },
   { 409, 42, 0 },
   { 505, 498, 140 },
   { 502, 508, 141 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 410, 42, 0 },
   { 411, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 412, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 630, 713, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 714, 715, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 716, 717, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 718, 480, 142 },
   { 718, 477, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 719, 720, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 721, 722, 142 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 723, 725, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 751, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 751, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 751, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 751, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 751, 728, 143 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 751, 728, 143 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 751, 728, 143 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 751, 728, 143 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 755, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 755, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 755, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 755, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 755, 734, 143 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 755, 734, 143 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 755, 734, 143 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 755, 734, 143 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 739, 740, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 739, 740, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 739, 740, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 739, 740, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 739, 740, 144 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 739, 740, 144 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 739, 740, 144 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 739, 740, 144 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 745, 746, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 745, 746, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 745, 746, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 745, 746, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 745, 746, 144 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 596, 598, 0 },
   { 601, 598, 0 },
   { 745, 746, 144 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 606, 0 },
   { 609, 606, 0 },
   { 611, 613, 0 },
   { 615, 613, 0 },
   { 745, 746, 144 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 620, 0 },
   { 623, 620, 0 },
   { 625, 627, 0 },
   { 629, 627, 0 },
   { 745, 746, 144 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 10, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 231, 232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 192, 194, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 229, 230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 222, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 709, 710, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 42, 0 },
   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 16, 0 },
   { 29, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 206, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { 26, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 223, 224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 197, 198, 0 },
   { 205, 206, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 707, 708, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 20, 21, 0 },
   { 32, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 238, 239, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 199, 200, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 225, 226, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 705, 706, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 22, 23, 0 },
   { 33, 23, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 236, 237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 201, 202, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 711, 712, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 310, 313, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 258, 260, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 314, 316, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 284, 275, 18 },
   { 286, 278, 18 },
   { 287, 280, 18 },
   { 288, 282, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 102, 103, 34 },
   { 118, 119, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 186, 188, 135 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 180, 185, 135 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 180, 185, 135 },
   { 186, 188, 135 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 302, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { 501, 490, 140 },
   { 504, 494, 140 },
   { 507, 498, 140 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { 501, 490, 140 },
   { 504, 494, 140 },
   { 507, 498, 140 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 48, 0 },
   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 42, 0 },
   { 518, 520, 0 },
   { 527, 529, 0 },
   { 536, 538, 0 },
   { 546, 548, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 57, 58, 0 },
   { 240, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 240, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 240, 42, 0 },
   { 523, 524, 0 },
   { 532, 533, 0 },
   { 541, 542, 0 },
   { 551, 552, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 427, 426, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 422, 424, 139 },
   { 427, 426, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 446, 447, 139 },
   { 450, 449, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 453, 447, 139 },
   { 456, 449, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 428, 424, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 425, 426, 139 },
   { 428, 424, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 448, 449, 139 },
   { 451, 447, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 454, 449, 139 },
   { 457, 447, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 134 },
   { 169, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 134 },
   { 171, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 134 },
   { 169, 166, 134 },
   { 171, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { 169, 166, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 155, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 155, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 441, 435, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 439, 435, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 462, 433, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 443, 433, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 442, 433, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 461, 435, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 415, 413, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 417, 413, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 418, 416, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 419, 416, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 407, 42, 0 },
   { 409, 42, 0 },
   { 489, 490, 140 },
   { 492, 494, 140 },
   { 496, 498, 140 },
   { 505, 498, 140 },
   { 502, 508, 141 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 510, 511, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 459, 433, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 458, 435, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 464, 466, 60 },
   { 473, 474, 60 },
   { 478, 479, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 468, 466, 60 },
   { 476, 474, 60 },
   { 481, 479, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 469, 470, 60 },
   { 482, 483, 60 },
   { 485, 486, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 471, 470, 60 },
   { 484, 483, 60 },
   { 487, 486, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 513, 514, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 261, 270, 137 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 730, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 727, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 727, 728, 143 },
   { 730, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 733, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 736, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 733, 734, 143 },
   { 736, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 12, 6, 0 },
   { 231, 232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 8, 10, 0 },
   { 229, 230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 18, 14, 0 },
   { 30, 14, 0 },
   { 235, 206, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { 28, 16, 0 },
   { 233, 234, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 220, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 245, 247, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 556, 560, 15 },
   { 573, 577, 17 },
   { 586, 590, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 284, 275, 18 },
   { 286, 278, 18 },
   { 287, 280, 18 },
   { 288, 282, 18 },
   { 556, 560, 15 },
   { 573, 577, 17 },
   { 586, 590, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 68, 70, 33 },
   { 85, 86, 34 },
   { 88, 90, 34 },
   { 118, 119, 35 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 105, 106, 34 },
   { 118, 119, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 72, 75, 33 },
   { 85, 86, 34 },
   { 92, 95, 34 },
   { 118, 119, 35 },
   { 125, 128, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 108, 109, 34 },
   { 118, 119, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 102, 103, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 102, 103, 34 },
   { 125, 128, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 108, 109, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 105, 106, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 72, 75, 33 },
   { 88, 90, 34 },
   { 92, 95, 34 },
   { 121, 123, 35 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 108, 109, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 105, 106, 34 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 108, 109, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 348, 353, 138 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 108, 109, 34 },
   { 125, 128, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 298, 18 },
   { 302, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 273, 275, 18 },
   { 277, 278, 18 },
   { 279, 280, 18 },
   { 281, 282, 18 },
   { 293, 294, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { 297, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { 300, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 294, 18 },
   { 302, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 652, 89, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 685, 686, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 703, 704, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 691, 692, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 682, 683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 676, 677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 701, 86, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 632, 93, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 663, 664, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 657, 658, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 672, 673, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 660, 661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 634, 635, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 637, 638, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 640, 641, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 643, 644, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 646, 647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 649, 650, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 654, 655, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 666, 667, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 669, 670, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 679, 680, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 688, 689, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 694, 695, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 697, 698, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 751, 728, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 751, 728, 143 },
   { 753, 731, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 755, 734, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 755, 734, 143 },
   { 757, 737, 143 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 739, 740, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 739, 740, 144 },
   { 742, 743, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 745, 746, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 565, 569, 0 },
   { 579, 580, 0 },
   { 592, 593, 0 },
   { 745, 746, 144 },
   { 748, 749, 144 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 298, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 432, 433, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 422, 424, 139 },
   { 427, 426, 139 },
   { 432, 433, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 436, 435, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 427, 426, 139 },
   { 436, 435, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 434, 435, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 425, 426, 139 },
   { 428, 424, 139 },
   { 434, 435, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 437, 433, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 428, 424, 139 },
   { 437, 433, 139 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 38, 41, 0 },
   { 241, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 53, 54, 0 },
   { 240, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 173, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 173, 176, 134 },
   { 178, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 173, 176, 134 },
   { 179, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 173, 176, 134 },
   { 178, 176, 134 },
   { 179, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { 173, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { 173, 176, 134 },
   { 178, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { 169, 166, 134 },
   { 173, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 131 },
   { 162, 166, 134 },
   { 169, 166, 134 },
   { 173, 176, 134 },
   { 178, 176, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 78, 80, 33 },
   { 85, 86, 34 },
   { 98, 100, 34 },
   { 118, 119, 35 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 112, 113, 34 },
   { 118, 119, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 102, 103, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 78, 80, 33 },
   { 88, 90, 34 },
   { 98, 100, 34 },
   { 121, 123, 35 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 112, 113, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 105, 106, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 340, 138 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 348, 353, 138 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 78, 80, 33 },
   { 92, 95, 34 },
   { 98, 100, 34 },
   { 125, 128, 35 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 112, 113, 34 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 108, 109, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 340, 138 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 369, 375, 138 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 340, 138 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 112, 113, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 33 },
   { 85, 86, 34 },
   { 118, 119, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 33 },
   { 88, 90, 34 },
   { 121, 123, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 348, 353, 138 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 33 },
   { 92, 95, 34 },
   { 125, 128, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { 381, 383, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 381, 383, 138 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 334, 340, 138 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 33 },
   { 98, 100, 34 },
   { 131, 133, 35 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 34 },
   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { 369, 375, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 369, 375, 138 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 128 },
   { 140, 141, 128 },
   { 143, 144, 128 },
   { 146, 147, 129 },
   { 355, 361, 138 },
   { 385, 390, 138 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_late_pass_op_table[nir_num_search_ops] = {
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         128,
         129,
         130,
         3,
         264,
         0,
         131,
         0,
         0,
         0,
         0,
         0,
         0,
         132,
         0,
         133,
         0,
         0,
         133,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         134,
         0,
         0,
         265,
         0,
         133,
         0,
         0,
         133,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         7,
         8,
         9,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         3,
         11,
         4,
         10,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         14,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         12,
         0,
         12,
         16,
         12,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         5,
         6,
         172,
         173,
         173,
         174,
         5,
         269,
         270,
         172,
         173,
         173,
         271,
         272,
         273,
         274,
         389,
         390,
         391,
         392,
         439,
         440,
         6,
         6,
         175,
         176,
         177,
         178,
         6,
         275,
         276,
         277,
         278,
         279,
         280,
         281,
         282,
         283,
         393,
         394,
         395,
         396,
         441,
         442,
         172,
         175,
         172,
         179,
         179,
         180,
         172,
         269,
         270,
         172,
         179,
         179,
         284,
         285,
         286,
         287,
         397,
         390,
         391,
         398,
         439,
         440,
         173,
         176,
         179,
         173,
         173,
         181,
         173,
         288,
         289,
         179,
         173,
         173,
         290,
         291,
         292,
         293,
         399,
         400,
         401,
         402,
         443,
         444,
         173,
         177,
         179,
         173,
         173,
         181,
         173,
         288,
         289,
         179,
         173,
         173,
         290,
         291,
         292,
         293,
         399,
         400,
         401,
         402,
         443,
         444,
         174,
         178,
         180,
         181,
         181,
         174,
         174,
         294,
         295,
         180,
         181,
         181,
         296,
         297,
         298,
         299,
         403,
         404,
         405,
         406,
         445,
         446,
         5,
         6,
         172,
         173,
         173,
         174,
         182,
         269,
         270,
         172,
         173,
         173,
         271,
         272,
         273,
         274,
         389,
         390,
         391,
         392,
         439,
         440,
         269,
         275,
         269,
         288,
         288,
         294,
         269,
         269,
         300,
         269,
         288,
         288,
         301,
         302,
         303,
         304,
         407,
         408,
         409,
         410,
         447,
         448,
         270,
         276,
         270,
         289,
         289,
         295,
         270,
         300,
         270,
         270,
         289,
         289,
         305,
         306,
         307,
         308,
         411,
         412,
         413,
         414,
         449,
         450,
         172,
         277,
         172,
         179,
         179,
         180,
         172,
         269,
         270,
         172,
         179,
         179,
         284,
         285,
         286,
         287,
         397,
         390,
         391,
         398,
         439,
         440,
         173,
         278,
         179,
         173,
         173,
         181,
         173,
         288,
         289,
         179,
         173,
         173,
         290,
         291,
         292,
         293,
         399,
         400,
         401,
         402,
         443,
         444,
         173,
         279,
         179,
         173,
         173,
         181,
         173,
         288,
         289,
         179,
         173,
         173,
         290,
         291,
         292,
         293,
         399,
         400,
         401,
         402,
         443,
         444,
         271,
         280,
         284,
         290,
         290,
         296,
         271,
         301,
         305,
         284,
         290,
         290,
         271,
         309,
         310,
         311,
         415,
         416,
         417,
         418,
         451,
         452,
         272,
         281,
         285,
         291,
         291,
         297,
         272,
         302,
         306,
         285,
         291,
         291,
         309,
         272,
         312,
         313,
         419,
         420,
         421,
         422,
         453,
         454,
         273,
         282,
         286,
         292,
         292,
         298,
         273,
         303,
         307,
         286,
         292,
         292,
         310,
         312,
         273,
         314,
         423,
         424,
         425,
         426,
         455,
         456,
         274,
         283,
         287,
         293,
         293,
         299,
         274,
         304,
         308,
         287,
         293,
         293,
         311,
         313,
         314,
         274,
         427,
         428,
         429,
         430,
         457,
         458,
         389,
         393,
         397,
         399,
         399,
         403,
         389,
         407,
         411,
         397,
         399,
         399,
         415,
         419,
         423,
         427,
         389,
         431,
         432,
         433,
         459,
         460,
         390,
         394,
         390,
         400,
         400,
         404,
         390,
         408,
         412,
         390,
         400,
         400,
         416,
         420,
         424,
         428,
         431,
         390,
         434,
         435,
         461,
         462,
         391,
         395,
         391,
         401,
         401,
         405,
         391,
         409,
         413,
         391,
         401,
         401,
         417,
         421,
         425,
         429,
         432,
         434,
         391,
         436,
         463,
         464,
         392,
         396,
         398,
         402,
         402,
         406,
         392,
         410,
         414,
         398,
         402,
         402,
         418,
         422,
         426,
         430,
         433,
         435,
         436,
         392,
         465,
         466,
         439,
         441,
         439,
         443,
         443,
         445,
         439,
         447,
         449,
         439,
         443,
         443,
         451,
         453,
         455,
         457,
         459,
         461,
         463,
         465,
         439,
         467,
         440,
         442,
         440,
         444,
         444,
         446,
         440,
         448,
         450,
         440,
         444,
         444,
         452,
         454,
         456,
         458,
         460,
         462,
         464,
         466,
         467,
         440,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         7,
         135,
         136,
         137,
         138,
         139,
         140,
         378,
         379,
         437,
         438,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         141,
         142,
         9,
         9,
         266,
         9,
         0,
         143,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         144,
         0,
         145,
         0,
         0,
         145,
         0,
         0,
         146,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         147,
         0,
         0,
         0,
         267,
         0,
         145,
         0,
         0,
         145,
         0,
         0,
         268,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         11,
         12,
         11,
         11,
         11,
         11,
         11,
         12,
         12,
         148,
         149,
         150,
         151,
         12,
         11,
         148,
         11,
         11,
         11,
         11,
         11,
         11,
         149,
         11,
         152,
         11,
         11,
         11,
         11,
         150,
         11,
         11,
         11,
         11,
         11,
         11,
         151,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         11,
         11,
         11,
         11,
         153,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         13,
         14,
         13,
         13,
         13,
         13,
         14,
         14,
         154,
         155,
         156,
         14,
         13,
         154,
         13,
         13,
         13,
         13,
         13,
         155,
         13,
         157,
         13,
         13,
         13,
         156,
         13,
         13,
         13,
         13,
         13,
         14,
         13,
         13,
         13,
         158,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 10,
      .table = (const uint16_t []) {
      
         15,
         16,
         15,
         15,
         15,
         15,
         189,
         189,
         15,
         15,
         16,
         16,
         16,
         16,
         16,
         190,
         191,
         192,
         16,
         16,
         15,
         16,
         193,
         15,
         194,
         15,
         189,
         189,
         15,
         15,
         15,
         16,
         15,
         15,
         15,
         15,
         189,
         189,
         15,
         376,
         15,
         16,
         194,
         15,
         15,
         15,
         189,
         189,
         15,
         376,
         15,
         190,
         15,
         15,
         15,
         15,
         189,
         189,
         15,
         15,
         189,
         191,
         189,
         189,
         189,
         189,
         189,
         189,
         189,
         189,
         189,
         192,
         189,
         189,
         189,
         189,
         189,
         189,
         189,
         189,
         15,
         16,
         15,
         15,
         15,
         15,
         189,
         189,
         195,
         15,
         15,
         16,
         15,
         376,
         376,
         15,
         189,
         189,
         15,
         15,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         1,
         1,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         377,
         17,
         17,
         17,
         196,
         17,
         377,
         17,
         17,
         196,
         197,
         17,
         17,
         17,
         17,
         17,
         17,
         198,
         17,
         17,
         377,
         377,
         17,
         17,
         17,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         5,
         5,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         18,
         19,
         164,
         165,
         18,
         365,
         19,
         19,
         166,
         167,
         19,
         366,
         164,
         166,
         164,
         168,
         164,
         365,
         165,
         167,
         168,
         165,
         165,
         367,
         18,
         19,
         164,
         165,
         169,
         365,
         365,
         366,
         365,
         367,
         365,
         365,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         207,
         208,
         209,
      },
   },
   [nir_op_fmulz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         20,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         3,
         3,
         3,
         2,
         2,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         21,
         22,
         211,
         212,
         213,
         381,
         382,
         22,
         22,
         214,
         215,
         216,
         383,
         384,
         211,
         214,
         211,
         212,
         217,
         385,
         386,
         212,
         215,
         212,
         212,
         218,
         387,
         388,
         213,
         216,
         217,
         218,
         213,
         381,
         382,
         381,
         383,
         385,
         387,
         381,
         381,
         382,
         382,
         384,
         386,
         388,
         382,
         382,
         382,
      },
   },
   [nir_op_ineg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         2,
         2,
         2,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         23,
         219,
         220,
      },
   },
   [nir_op_iabs] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         24,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         183,
         184,
         185,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         25,
         241,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         159,
         159,
         0,
         0,
         159,
         160,
         0,
         0,
         0,
         0,
         161,
      },
   },
   [nir_op_fdot2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         26,
      },
   },
   [nir_op_fdot3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         27,
      },
   },
   [nir_op_fdot4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         28,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         29,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         162,
         162,
         162,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         163,
      },
   },
   [nir_op_fround_even] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         30,
         242,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         6,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         7,
         4,
         8,
         7,
         3,
         0,
         0,
         9,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         9,
         9,
         9,
         2,
         2,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         187,
         31,
         316,
         317,
         321,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         187,
         31,
         316,
         317,
         321,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         318,
         322,
         318,
         319,
         319,
         318,
         323,
         324,
         324,
         318,
         320,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         32,
         31,
         186,
         187,
         31,
         316,
         317,
         321,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         318,
         318,
         318,
         319,
         325,
         318,
         318,
         319,
         325,
         318,
         320,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         188,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         32,
         31,
         186,
         187,
         31,
         316,
         317,
         321,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         318,
         318,
         318,
         319,
         325,
         318,
         318,
         319,
         325,
         318,
         320,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         322,
         318,
         319,
         319,
         318,
         323,
         324,
         324,
         318,
         320,
         318,
         318,
         318,
         319,
         325,
         318,
         318,
         319,
         325,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         325,
         318,
         318,
         319,
         325,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         32,
         31,
         186,
         186,
         31,
         316,
         317,
         317,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         187,
         31,
         31,
         186,
         187,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
         318,
         318,
         318,
         319,
         319,
         318,
         318,
         319,
         319,
         318,
         320,
         31,
         31,
         31,
         186,
         186,
         31,
         31,
         186,
         186,
         31,
         315,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         170,
         0,
         0,
         0,
         171,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         4,
         5,
         5,
         0,
         6,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         14,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         6,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         7,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
      },
      
      .num_filtered_states = 20,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         352,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         81,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         82,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         83,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         80,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         353,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         85,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         86,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         87,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         84,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         354,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         89,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         90,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         91,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         355,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         93,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         94,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         95,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         92,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         356,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         97,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         98,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         99,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         96,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         357,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         101,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         102,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         103,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         100,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         358,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         105,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         106,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         107,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         104,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         359,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         109,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         110,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         111,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         108,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         360,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         113,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         114,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         115,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         112,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         361,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         117,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         118,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         119,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         116,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         362,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         122,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         123,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         363,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         125,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         126,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         127,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         364,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         351,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         77,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         78,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         79,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         33,
         243,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         34,
         244,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
         35,
         210,
         380,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         36,
         0,
         37,
         0,
         36,
         0,
         37,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         38,
         0,
         39,
         0,
         38,
         0,
         39,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         40,
         0,
         41,
         42,
         41,
         0,
         231,
         0,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         43,
         44,
         43,
         44,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         45,
         46,
         45,
         46,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         47,
         47,
         47,
         47,
         47,
         47,
         47,
         47,
         47,
         227,
         47,
         47,
         47,
         228,
         47,
         47,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         48,
         49,
         48,
         48,
         48,
         49,
         48,
         48,
         48,
         229,
         48,
         48,
         48,
         230,
         48,
         48,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         2,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         50,
         199,
         0,
         0,
         0,
         0,
         199,
         50,
         50,
         200,
         201,
         202,
         50,
         368,
         369,
         199,
         200,
         199,
         199,
         199,
         199,
         199,
         199,
         0,
         201,
         199,
         0,
         0,
         0,
         0,
         199,
         0,
         202,
         199,
         0,
         0,
         0,
         0,
         199,
         0,
         50,
         199,
         0,
         0,
         0,
         370,
         371,
         0,
         368,
         199,
         0,
         0,
         370,
         0,
         199,
         199,
         369,
         199,
         199,
         199,
         371,
         199,
         199,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         2,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         51,
         203,
         0,
         0,
         0,
         0,
         203,
         51,
         51,
         204,
         205,
         206,
         51,
         372,
         373,
         203,
         204,
         203,
         203,
         203,
         203,
         203,
         203,
         0,
         205,
         203,
         0,
         0,
         0,
         0,
         203,
         0,
         206,
         203,
         0,
         0,
         0,
         0,
         203,
         0,
         51,
         203,
         0,
         0,
         0,
         374,
         375,
         0,
         372,
         203,
         0,
         0,
         374,
         0,
         203,
         203,
         373,
         203,
         203,
         203,
         375,
         203,
         203,
      },
   },
   [nir_op_bitz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         221,
         52,
         52,
         52,
         52,
         222,
         52,
         52,
         52,
         223,
         223,
         223,
         223,
         223,
      },
   },
   [nir_op_bitnz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         224,
         53,
         53,
         53,
         53,
         225,
         53,
         53,
         53,
         226,
         226,
         226,
         226,
         226,
      },
   },
   [nir_op_inot] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         54,
         233,
         234,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         55,
         236,
         237,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         56,
         238,
         239,
      },
   },
   [nir_op_insert_u8] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         232,
         232,
      },
   },
   [nir_op_insert_u16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         240,
         240,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         57,
         0,
         57,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         235,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 26,
      .table = (const uint16_t []) {
      
         0,
         326,
         327,
         328,
         329,
         330,
         331,
         332,
         333,
         334,
         335,
         336,
         337,
         338,
         339,
         340,
         341,
         342,
         343,
         344,
         345,
         346,
         347,
         348,
         349,
         350,
      },
   },
   [nir_op_f2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         58,
      },
   },
   [nir_op_fceil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         245,
      },
   },
   [nir_op_fcos] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         246,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         247,
      },
   },
   [nir_op_ffloor] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         248,
      },
   },
   [nir_op_ffract] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         249,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         250,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         251,
      },
   },
   [nir_op_fsign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         252,
      },
   },
   [nir_op_fsin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         253,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         254,
      },
   },
   [nir_op_fmod] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         255,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         256,
      },
   },
   [nir_op_frem] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         257,
      },
   },
   [nir_op_f2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         59,
      },
   },
   [nir_op_f2ump] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         60,
      },
   },
   [nir_op_i2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         61,
      },
   },
   [nir_op_i2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         62,
      },
   },
   [nir_op_u2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         63,
      },
   },
   [nir_op_fisfinite] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         64,
      },
   },
   [nir_op_fcsel] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         258,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         259,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         260,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         261,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         262,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
      },
   },
   [nir_op_slt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         65,
         66,
         67,
      },
   },
   [nir_op_sge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         68,
         69,
         70,
      },
   },
   [nir_op_ilt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         71,
         72,
         73,
      },
   },
   [nir_op_ige] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         74,
         75,
         76,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_late_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   3,
   5,
   7,
   0,
   0,
   0,
   0,
   0,
   9,
   11,
   0,
   13,
   15,
   17,
   19,
   21,
   0,
   23,
   0,
   0,
   0,
   28,
   30,
   33,
   35,
   38,
   0,
   43,
   0,
   48,
   0,
   51,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   53,
   55,
   57,
   59,
   62,
   64,
   66,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   68,
   71,
   76,
   81,
   83,
   87,
   93,
   99,
   101,
   105,
   111,
   117,
   120,
   125,
   132,
   139,
   141,
   145,
   151,
   157,
   159,
   163,
   169,
   175,
   178,
   183,
   190,
   197,
   199,
   203,
   209,
   215,
   217,
   221,
   227,
   233,
   236,
   241,
   248,
   255,
   257,
   261,
   267,
   273,
   275,
   279,
   285,
   291,
   294,
   299,
   306,
   313,
   315,
   317,
   319,
   321,
   323,
   325,
   327,
   329,
   332,
   334,
   336,
   338,
   340,
   343,
   345,
   348,
   350,
   352,
   355,
   357,
   360,
   362,
   0,
   364,
   366,
   368,
   371,
   373,
   375,
   377,
   0,
   379,
   0,
   381,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   383,
   0,
   385,
   390,
   394,
   396,
   401,
   405,
   413,
   415,
   423,
   429,
   0,
   434,
   436,
   438,
   0,
   441,
   0,
   443,
   445,
   450,
   452,
   457,
   459,
   462,
   468,
   471,
   473,
   479,
   481,
   484,
   487,
   490,
   492,
   495,
   498,
   0,
   0,
   0,
   0,
   501,
   503,
   506,
   508,
   511,
   515,
   517,
   520,
   524,
   526,
   528,
   530,
   532,
   534,
   536,
   538,
   540,
   542,
   544,
   546,
   548,
   556,
   558,
   560,
   0,
   562,
   566,
   570,
   574,
   578,
   0,
   580,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   582,
   584,
   586,
   589,
   591,
   593,
   596,
   599,
   602,
   606,
   610,
   612,
   620,
   626,
   628,
   632,
   634,
   636,
   644,
   650,
   656,
   663,
   674,
   676,
   680,
   682,
   684,
   690,
   698,
   704,
   710,
   721,
   730,
   735,
   742,
   747,
   752,
   761,
   768,
   771,
   776,
   779,
   782,
   791,
   800,
   811,
   820,
   829,
   836,
   845,
   852,
   859,
   864,
   867,
   870,
   875,
   880,
   0,
   883,
   885,
   887,
   889,
   891,
   893,
   896,
   902,
   905,
   908,
   911,
   913,
   915,
   917,
   919,
   921,
   923,
   925,
   927,
   929,
   931,
   933,
   935,
   937,
   939,
   941,
   943,
   945,
   947,
   949,
   951,
   953,
   955,
   957,
   959,
   961,
   965,
   970,
   975,
   981,
   986,
   991,
   997,
   1002,
   1007,
   1013,
   1018,
   1023,
   1029,
   1031,
   1033,
   1035,
   1037,
   1039,
   1043,
   1045,
   1048,
   1050,
   1054,
   1056,
   1059,
   1062,
   1065,
   1067,
   0,
   1069,
   1072,
   1076,
   1080,
   1085,
   1089,
   1094,
   1099,
   1105,
   1107,
   1115,
   1121,
   1123,
   1125,
   1133,
   1139,
   1141,
   1147,
   1153,
   1158,
   1169,
   1178,
   1183,
   1186,
   1195,
   1202,
   1205,
   1214,
   1225,
   1234,
   1243,
   1250,
   1259,
   1266,
   1273,
   1276,
   1285,
   1292,
   1295,
   1300,
   1311,
   1320,
   1325,
   1328,
   1337,
   1344,
   1347,
   1350,
   1359,
   1366,
   1369,
   1378,
   1385,
   1388,
   1397,
   1406,
   1413,
   1415,
   1417,
   1423,
   1429,
   1435,
   1441,
   1450,
   1459,
   1466,
   1473,
   1482,
   1491,
   1498,
   1505,
   1512,
   1519,
   1528,
   1537,
   1544,
   1551,
   1558,
   1565,
   1572,
   1579,
   1588,
   1597,
   1604,
   1611,
   1618,
   1625,
};

static const nir_algebraic_table nir_opt_algebraic_late_table = {
   .transforms = nir_opt_algebraic_late_transforms,
   .transform_offsets = nir_opt_algebraic_late_transform_offsets,
   .pass_op_table = nir_opt_algebraic_late_pass_op_table,
   .values = nir_opt_algebraic_late_values,
   .expression_cond = nir_opt_algebraic_late_expression_cond,
   .variable_cond = nir_opt_algebraic_late_variable_cond,
};

bool
nir_opt_algebraic_late(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[145];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(758 == ARRAY_SIZE(nir_opt_algebraic_late_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !options->lower_ftrunc || options->lower_ffloor;
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[7] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[8] = !options->lower_iadd_sat;
   condition_flags[9] = !options->lower_uadd_sat;
   condition_flags[10] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[11] = !options->has_sdot_4x8;
   condition_flags[12] = !options->has_udot_4x8;
   condition_flags[13] = !options->has_sudot_4x8;
   condition_flags[14] = !options->has_dot_2x16;
   condition_flags[15] = options->lower_flrp16;
   condition_flags[16] = !options->lower_flrp16;
   condition_flags[17] = options->lower_flrp32;
   condition_flags[18] = !options->lower_flrp32;
   condition_flags[19] = options->lower_flrp64;
   condition_flags[20] = !options->lower_flrp64;
   condition_flags[21] = options->lower_ftrunc;
   condition_flags[22] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[23] = options->lower_ffloor;
   condition_flags[24] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[25] = !options->lower_ffloor;
   condition_flags[26] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[27] = options->lower_ffract;
   condition_flags[28] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[29] = options->lower_fceil;
   condition_flags[30] = options->lower_ffma16;
   condition_flags[31] = options->lower_ffma32;
   condition_flags[32] = options->lower_ffma64;
   condition_flags[33] = options->fuse_ffma16;
   condition_flags[34] = options->fuse_ffma32;
   condition_flags[35] = options->fuse_ffma64;
   condition_flags[36] = options->lower_fdph;
   condition_flags[37] = !options->lower_fdph;
   condition_flags[38] = options->lower_fdot;
   condition_flags[39] = !options->lower_fsat;
   condition_flags[40] = !options->lower_iabs;
   condition_flags[41] = options->lower_fsat;
   condition_flags[42] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[43] = !options->lower_fsign;
   condition_flags[44] = options->lower_umax;
   condition_flags[45] = options->lower_umin;
   condition_flags[46] = !options->lower_umax;
   condition_flags[47] = !options->lower_umin;
   condition_flags[48] = options->lower_scmp;
   condition_flags[49] = options->lower_vector_cmp;
   condition_flags[50] = !options->lower_pack_64_2x32_split;
   condition_flags[51] = !options->lower_rotate;
   condition_flags[52] = options->lower_rotate;
   condition_flags[53] = options->lower_fpow;
   condition_flags[54] = !options->lower_fpow;
   condition_flags[55] = options->lower_fdiv;
   condition_flags[56] = options->lower_fsqrt;
   condition_flags[57] = !options->lower_fsqrt;
   condition_flags[58] = options->lower_sincos;
   condition_flags[59] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[60] = !options->lower_extract_byte;
   condition_flags[61] = !options->lower_extract_word;
   condition_flags[62] = options->has_pack_32_4x8;
   condition_flags[63] = options->lower_pack_64_2x32_split;
   condition_flags[64] = options->lower_pack_32_2x16_split;
   condition_flags[65] = options->has_pack_half_2x16_rtz;
   condition_flags[66] = options->lower_unpack_64_2x32_split;
   condition_flags[67] = options->lower_unpack_32_2x16_split;
   condition_flags[68] = options->has_find_msb_rev;
   condition_flags[69] = !options->lower_find_msb_to_reverse;
   condition_flags[70] = !options->lower_find_lsb;
   condition_flags[71] = options->lower_fmod;
   condition_flags[72] = options->lower_uadd_carry;
   condition_flags[73] = options->lower_usub_borrow;
   condition_flags[74] = options->lower_bitfield_insert;
   condition_flags[75] = options->lower_hadd;
   condition_flags[76] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[77] = options->lower_mul_32x16;
   condition_flags[78] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[79] = options->lower_uadd_sat;
   condition_flags[80] = options->lower_usub_sat;
   condition_flags[81] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[82] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[83] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[84] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[85] = options->lower_bitfield_insert_to_shifts;
   condition_flags[86] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[87] = options->lower_bitfield_extract;
   condition_flags[88] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[89] = options->lower_bitfield_extract_to_shifts;
   condition_flags[90] = options->lower_ifind_msb;
   condition_flags[91] = options->lower_find_msb_to_reverse;
   condition_flags[92] = options->lower_ifind_msb_to_uclz;
   condition_flags[93] = options->lower_ufind_msb_to_uclz;
   condition_flags[94] = options->lower_uclz;
   condition_flags[95] = options->lower_find_lsb;
   condition_flags[96] = options->lower_extract_byte;
   condition_flags[97] = options->lower_extract_word;
   condition_flags[98] = options->lower_pack_unorm_2x16;
   condition_flags[99] = options->lower_pack_unorm_4x8;
   condition_flags[100] = options->lower_pack_snorm_2x16;
   condition_flags[101] = options->lower_pack_snorm_4x8;
   condition_flags[102] = options->lower_unpack_unorm_2x16;
   condition_flags[103] = options->lower_unpack_unorm_4x8;
   condition_flags[104] = options->lower_unpack_snorm_2x16;
   condition_flags[105] = options->lower_unpack_snorm_4x8;
   condition_flags[106] = options->lower_pack_split;
   condition_flags[107] = options->lower_isign;
   condition_flags[108] = !options->lower_isign;
   condition_flags[109] = options->lower_fsign;
   condition_flags[110] = !options->has_imul24;
   condition_flags[111] = !options->has_umul24;
   condition_flags[112] = !options->has_umad24;
   condition_flags[113] = options->has_imul24;
   condition_flags[114] = options->has_umad24;
   condition_flags[115] = options->has_umul24;
   condition_flags[116] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[117] = !options->lower_usub_sat;
   condition_flags[118] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[119] = options->lower_iadd_sat;
   condition_flags[120] = options->support_8bit_alu;
   condition_flags[121] = options->support_16bit_alu;
   condition_flags[122] = options->lower_ldexp;
   condition_flags[123] = !options->lower_bitfield_reverse;
   condition_flags[124] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[125] = options->lower_fisnormal;
   condition_flags[126] = options->avoid_ternary_with_two_constants;
   condition_flags[127] = !options->vectorize_vec2_16bit;
   condition_flags[128] = options->has_fsub;
   condition_flags[129] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[130] = options->lower_fneg;
   condition_flags[131] = options->has_isub || options->lower_ineg;
   condition_flags[132] = options->lower_ineg;
   condition_flags[133] = options->lower_iabs;
   condition_flags[134] = options->has_iadd3;
   condition_flags[135] = options->vectorize_vec2_16bit;
   condition_flags[136] = options->fdot_replicates;
   condition_flags[137] = options->lower_fround_even;
   condition_flags[138] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[139] = options->has_bit_test;
   condition_flags[140] = !options->lower_insert_byte;
   condition_flags[141] = !options->lower_insert_word;
   condition_flags[142] = !options->preserve_mediump;
   condition_flags[143] = options->has_fused_comp_and_csel;
   condition_flags[144] = options->has_fused_comp_and_csel && !options->no_integers;

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &nir_opt_algebraic_late_table);
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 19
 * transforms:
 *    ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b')
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b')
 *    ('fneg', ('fneg', 'a')) => a
 *    ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b')
 *    ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b'))
 *    ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c'))
 *    ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c')
 *    ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b'))
 *    ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a'))
 *    ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a'))
 */


static const nir_search_value_union nir_opt_algebraic_distribute_src_mods_values[] = {
   /* ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2450_0 -> 0 in the cache */
   /* replace2450_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 8 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 9, 10 },
      -1,
   } },

   /* replace2451_0 -> 6 in the cache */
   /* replace2451_1 -> 8 in the cache */
   /* replace2451_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      -1,
   } },

   /* ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b') */
   /* search2452_0_0 -> 0 in the cache */
   /* search2452_0 -> 1 in the cache */
   /* search2452_1_0 -> 2 in the cache */
   /* search2452_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2452_0 -> 0 in the cache */
   /* replace2452_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b') */
   /* search2453_0_0 -> 0 in the cache */
   /* search2453_0 -> 1 in the cache */
   /* search2453_1_0 -> 2 in the cache */
   /* search2453_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2453_0 -> 0 in the cache */
   /* replace2453_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b') */
   /* search2454_0_0 -> 0 in the cache */
   /* search2454_0 -> 1 in the cache */
   /* search2454_1_0 -> 2 in the cache */
   /* search2454_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2454_0 -> 0 in the cache */
   /* replace2454_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 20 },
      -1,
   } },

   /* replace2455 -> 19 in the cache */

   /* ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b') */
   /* search2456_0_0 -> 0 in the cache */
   /* search2456_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2456_0_0 -> 0 in the cache */
   /* replace2456_0 -> 1 in the cache */
   /* replace2456_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b')) */
   /* search2457_0_0 -> 0 in the cache */
   /* search2457_0_1 -> 2 in the cache */
   /* search2457_0 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2457_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace2457_1_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 26, 27 },
      -1,
   } },

   /* ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c')) */
   /* search2458_0_0 -> 6 in the cache */
   /* search2458_0_1 -> 8 in the cache */
   /* search2458_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 29 },
      -1,
   } },

   /* replace2458_0_0 -> 6 in the cache */
   /* replace2458_0 -> 7 in the cache */
   /* replace2458_1 -> 8 in the cache */
   /* replace2458_2_0 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 10 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 8, 31 },
      -1,
   } },

   /* ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c') */
   /* search2459_0_0 -> 6 in the cache */
   /* search2459_0_1 -> 8 in the cache */
   /* search2459_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 33 },
      -1,
   } },

   /* replace2459_0_0 -> 6 in the cache */
   /* replace2459_0 -> 7 in the cache */
   /* replace2459_1_0 -> 8 in the cache */
   /* replace2459_1 -> 9 in the cache */
   /* replace2459_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 7, 9, 10 },
      -1,
   } },

   /* ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2460_0_0 -> 0 in the cache */
   /* search2460_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 36 },
      -1,
   } },

   /* replace2460_0_0 -> 0 in the cache */
   /* replace2460_0 -> 1 in the cache */
   /* replace2460_1_0 -> 2 in the cache */
   /* replace2460_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2461_0_0 -> 0 in the cache */
   /* search2461_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 39 },
      -1,
   } },

   /* replace2461_0_0 -> 0 in the cache */
   /* replace2461_0 -> 1 in the cache */
   /* replace2461_1_0 -> 2 in the cache */
   /* replace2461_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2462_0_0 -> 0 in the cache */
   /* search2462_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 42 },
      -1,
   } },

   /* replace2462_0_0 -> 0 in the cache */
   /* replace2462_0 -> 1 in the cache */
   /* replace2462_1_0 -> 2 in the cache */
   /* replace2462_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b') */
   /* search2463_0_0 -> 0 in the cache */
   /* search2463_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 45 },
      -1,
   } },

   /* replace2463_0_0 -> 0 in the cache */
   /* replace2463_0 -> 1 in the cache */
   /* replace2463_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b') */
   /* search2464_0_0 -> 0 in the cache */
   /* search2464_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 48 },
      -1,
   } },

   /* replace2464_0_0 -> 0 in the cache */
   /* replace2464_0 -> 1 in the cache */
   /* replace2464_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b') */
   /* search2465_0_0 -> 0 in the cache */
   /* search2465_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 51 },
      -1,
   } },

   /* replace2465_0_0 -> 0 in the cache */
   /* replace2465_0 -> 1 in the cache */
   /* replace2465_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b')) */
   /* search2466_0_0 -> 0 in the cache */
   /* search2466_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 54 },
      -1,
   } },

   /* replace2466_0 -> 0 in the cache */
   /* replace2466_1_0 -> 2 in the cache */
   /* replace2466_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a')) */
   /* search2467_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 19 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2467_0_0 -> 19 in the cache */
   /* replace2467_0 -> 20 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 20 },
      -1,
   } },

   /* ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a')) */
   /* search2468_0_0 -> 19 in the cache */
   /* search2468_0 -> 57 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2468_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 61 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_distribute_src_mods_expression_cond[] = {
   (is_used_once),
};


static const struct transform nir_opt_algebraic_distribute_src_mods_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 23, 24, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 21, 19, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 30, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 47, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 49, 50, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 53, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 40, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 43, 44, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 56, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 59, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 25, 28, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 62, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 5, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 12, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 18, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_distribute_src_mods_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         2,
         2,
         2,
         28,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         7,
         8,
         9,
         10,
         11,
         12,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         1,
         3,
         4,
         5,
         6,
      },
      
      .num_filtered_states = 13,
      .table = (const uint16_t []) {
      
         3,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         4,
         4,
         4,
         4,
         4,
         4,
         29,
         29,
      },
   },
   [nir_op_fdot2_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         5,
         5,
         5,
         30,
      },
   },
   [nir_op_fdot3_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         6,
         6,
         6,
         31,
      },
   },
   [nir_op_fdot4_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         7,
         7,
         7,
         32,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         26,
         27,
      },
   },
   [nir_op_flrp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         8,
      },
   },
   [nir_op_fadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_fmin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_fmax] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_fdph_replicated] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         12,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         13,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_distribute_src_mods_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   13,
   15,
   17,
   19,
   21,
   23,
   25,
   27,
   29,
   31,
   33,
   35,
   37,
};

static const nir_algebraic_table nir_opt_algebraic_distribute_src_mods_table = {
   .transforms = nir_opt_algebraic_distribute_src_mods_transforms,
   .transform_offsets = nir_opt_algebraic_distribute_src_mods_transform_offsets,
   .pass_op_table = nir_opt_algebraic_distribute_src_mods_pass_op_table,
   .values = nir_opt_algebraic_distribute_src_mods_values,
   .expression_cond = nir_opt_algebraic_distribute_src_mods_expression_cond,
   .variable_cond = NULL,
};

bool
nir_opt_algebraic_distribute_src_mods(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[145];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(63 == ARRAY_SIZE(nir_opt_algebraic_distribute_src_mods_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !options->lower_ftrunc || options->lower_ffloor;
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[7] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[8] = !options->lower_iadd_sat;
   condition_flags[9] = !options->lower_uadd_sat;
   condition_flags[10] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[11] = !options->has_sdot_4x8;
   condition_flags[12] = !options->has_udot_4x8;
   condition_flags[13] = !options->has_sudot_4x8;
   condition_flags[14] = !options->has_dot_2x16;
   condition_flags[15] = options->lower_flrp16;
   condition_flags[16] = !options->lower_flrp16;
   condition_flags[17] = options->lower_flrp32;
   condition_flags[18] = !options->lower_flrp32;
   condition_flags[19] = options->lower_flrp64;
   condition_flags[20] = !options->lower_flrp64;
   condition_flags[21] = options->lower_ftrunc;
   condition_flags[22] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[23] = options->lower_ffloor;
   condition_flags[24] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[25] = !options->lower_ffloor;
   condition_flags[26] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[27] = options->lower_ffract;
   condition_flags[28] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[29] = options->lower_fceil;
   condition_flags[30] = options->lower_ffma16;
   condition_flags[31] = options->lower_ffma32;
   condition_flags[32] = options->lower_ffma64;
   condition_flags[33] = options->fuse_ffma16;
   condition_flags[34] = options->fuse_ffma32;
   condition_flags[35] = options->fuse_ffma64;
   condition_flags[36] = options->lower_fdph;
   condition_flags[37] = !options->lower_fdph;
   condition_flags[38] = options->lower_fdot;
   condition_flags[39] = !options->lower_fsat;
   condition_flags[40] = !options->lower_iabs;
   condition_flags[41] = options->lower_fsat;
   condition_flags[42] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[43] = !options->lower_fsign;
   condition_flags[44] = options->lower_umax;
   condition_flags[45] = options->lower_umin;
   condition_flags[46] = !options->lower_umax;
   condition_flags[47] = !options->lower_umin;
   condition_flags[48] = options->lower_scmp;
   condition_flags[49] = options->lower_vector_cmp;
   condition_flags[50] = !options->lower_pack_64_2x32_split;
   condition_flags[51] = !options->lower_rotate;
   condition_flags[52] = options->lower_rotate;
   condition_flags[53] = options->lower_fpow;
   condition_flags[54] = !options->lower_fpow;
   condition_flags[55] = options->lower_fdiv;
   condition_flags[56] = options->lower_fsqrt;
   condition_flags[57] = !options->lower_fsqrt;
   condition_flags[58] = options->lower_sincos;
   condition_flags[59] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[60] = !options->lower_extract_byte;
   condition_flags[61] = !options->lower_extract_word;
   condition_flags[62] = options->has_pack_32_4x8;
   condition_flags[63] = options->lower_pack_64_2x32_split;
   condition_flags[64] = options->lower_pack_32_2x16_split;
   condition_flags[65] = options->has_pack_half_2x16_rtz;
   condition_flags[66] = options->lower_unpack_64_2x32_split;
   condition_flags[67] = options->lower_unpack_32_2x16_split;
   condition_flags[68] = options->has_find_msb_rev;
   condition_flags[69] = !options->lower_find_msb_to_reverse;
   condition_flags[70] = !options->lower_find_lsb;
   condition_flags[71] = options->lower_fmod;
   condition_flags[72] = options->lower_uadd_carry;
   condition_flags[73] = options->lower_usub_borrow;
   condition_flags[74] = options->lower_bitfield_insert;
   condition_flags[75] = options->lower_hadd;
   condition_flags[76] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[77] = options->lower_mul_32x16;
   condition_flags[78] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[79] = options->lower_uadd_sat;
   condition_flags[80] = options->lower_usub_sat;
   condition_flags[81] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[82] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[83] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[84] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[85] = options->lower_bitfield_insert_to_shifts;
   condition_flags[86] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[87] = options->lower_bitfield_extract;
   condition_flags[88] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[89] = options->lower_bitfield_extract_to_shifts;
   condition_flags[90] = options->lower_ifind_msb;
   condition_flags[91] = options->lower_find_msb_to_reverse;
   condition_flags[92] = options->lower_ifind_msb_to_uclz;
   condition_flags[93] = options->lower_ufind_msb_to_uclz;
   condition_flags[94] = options->lower_uclz;
   condition_flags[95] = options->lower_find_lsb;
   condition_flags[96] = options->lower_extract_byte;
   condition_flags[97] = options->lower_extract_word;
   condition_flags[98] = options->lower_pack_unorm_2x16;
   condition_flags[99] = options->lower_pack_unorm_4x8;
   condition_flags[100] = options->lower_pack_snorm_2x16;
   condition_flags[101] = options->lower_pack_snorm_4x8;
   condition_flags[102] = options->lower_unpack_unorm_2x16;
   condition_flags[103] = options->lower_unpack_unorm_4x8;
   condition_flags[104] = options->lower_unpack_snorm_2x16;
   condition_flags[105] = options->lower_unpack_snorm_4x8;
   condition_flags[106] = options->lower_pack_split;
   condition_flags[107] = options->lower_isign;
   condition_flags[108] = !options->lower_isign;
   condition_flags[109] = options->lower_fsign;
   condition_flags[110] = !options->has_imul24;
   condition_flags[111] = !options->has_umul24;
   condition_flags[112] = !options->has_umad24;
   condition_flags[113] = options->has_imul24;
   condition_flags[114] = options->has_umad24;
   condition_flags[115] = options->has_umul24;
   condition_flags[116] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[117] = !options->lower_usub_sat;
   condition_flags[118] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[119] = options->lower_iadd_sat;
   condition_flags[120] = options->support_8bit_alu;
   condition_flags[121] = options->support_16bit_alu;
   condition_flags[122] = options->lower_ldexp;
   condition_flags[123] = !options->lower_bitfield_reverse;
   condition_flags[124] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[125] = options->lower_fisnormal;
   condition_flags[126] = options->avoid_ternary_with_two_constants;
   condition_flags[127] = !options->vectorize_vec2_16bit;
   condition_flags[128] = options->has_fsub;
   condition_flags[129] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[130] = options->lower_fneg;
   condition_flags[131] = options->has_isub || options->lower_ineg;
   condition_flags[132] = options->lower_ineg;
   condition_flags[133] = options->lower_iabs;
   condition_flags[134] = options->has_iadd3;
   condition_flags[135] = options->vectorize_vec2_16bit;
   condition_flags[136] = options->fdot_replicates;
   condition_flags[137] = options->lower_fround_even;
   condition_flags[138] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[139] = options->has_bit_test;
   condition_flags[140] = !options->lower_insert_byte;
   condition_flags[141] = !options->lower_insert_word;
   condition_flags[142] = !options->preserve_mediump;
   condition_flags[143] = options->has_fused_comp_and_csel;
   condition_flags[144] = options->has_fused_comp_and_csel && !options->no_integers;

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &nir_opt_algebraic_distribute_src_mods_table);
   }

   return progress;
}

